/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

//
// This module allow works with bounds (polygons) using google maps module
// opts = {
//   map:
//   options:
// }
//
redBorder.boundMap = function (map, opts) {
  var polygons = [];
  var selectedPolygon;
  var latlngbounds;
  var infoWindows = [];

  opts = opts || {};
  opts.autoinit = (opts.autoinit !== false);
  opts.polygons = (opts.polygons) || [];
  opts.center = (opts.center) || [];

  if(opts.input != undefined && !opts.input.startsWith('#'))
    opts.input = '#' + opts.input;

  if (opts.autoinit === true) {
    initialize();
  }

  function initialize() {
    drawingManager = initializeDrawingManager();
    latlngbounds = new google.maps.LatLngBounds();
    // Add polygons
    if(opts.polygons != undefined && opts.polygons.length > 0)
      addPolygons(opts.polygons);
    // Center
    if(opts.center != undefined && opts.center.length > 0) {
      for(var i = 0; i < opts.center.length; i++){
        getPolygon(opts.center[i]);
      }
    }
    // Fit bounds
    fitPolygons();

    // Load DOM listeners
    bindEvents();
  }

  function initializeDrawingManager(){
    dm = new google.maps.drawing.DrawingManager({
      drawingMode: google.maps.drawing.OverlayType.POLYGON,
      drawingControl: true,
      drawingControlOptions: {
        position: google.maps.ControlPosition.TOP_CENTER,
        drawingModes: [
          google.maps.drawing.OverlayType.POLYGON
        ]
      },
      polygonOptions: {
        strokeWeight: 0,
        fillOpacity: 0.45,
        draggable: true,
        editable: true
      }
    });

    dm.setMap(map);
    dm.setDrawingMode(null);

    google.maps.event.addListener(dm, 'drawingmode_changed', clearSelection);
    google.maps.event.addListener(map, 'click', clearSelection);

    google.maps.event.addListener(dm, 'overlaycomplete', function(event) {
      if (event.type != google.maps.drawing.OverlayType.MARKER) {
        // Switch back to non-drawing mode after drawing a shape.
        drawingManager.setDrawingMode(null);

        // Add an event listener that selects the newly-drawn shape when the user
        // mouses down on it.
        var newShape = event.overlay;
        newShape.type = event.type;

        // id
        var id = randomId();
        polygons.push({
          id: id,
          polygon: newShape
        });

        polygonsSave();

        setSelection(newShape);

        create_listener(newShape);
      };
    });

    return dm;
  }

  /*
   * Create listeners
   */
  function bindEvents() {
    $(document).on('click', '.delete-polygon', function(e) {
      e.preventDefault();

      var id = $(this).data('polygon-id');
      deletePolygon(id);
    });
  }

  /*
   * Delete listeners
   */
  function unbindEvents() {
    $(document).off('click', '.delete-polygon');
  }

  function clearSelection() {
    if (selectedPolygon) {
      selectedPolygon = null;
    }
  }

  function polygonsSave() {
    var _polygons = jsonMake();
    $(opts.input).val(_polygons);
  }

  function addPolygons(data){
    for(var i = 0; i < data.length; i++){
      var polygon = getPolygon(data[i]);
      var id = randomId();

      create_listener(polygon, id);

      polygon.setMap(map);
      setSelection(polygon);
      polygons.push(
        {
          id: id,
          polygon: polygon
        });
    }
    polygonsSave();
  }

  /*
   * Create a GoogleMaps Polygon
   */
  function getPolygon(polygon_data) {
    var coordinates = Array();

    for(var j = 0; j < polygon_data.length; j++){
      var latlng = new google.maps.LatLng(polygon_data[j][0], polygon_data[j][1]);
      coordinates.push(latlng);
      latlngbounds.extend(latlng);
    }

    var polygon = new google.maps.Polygon({
      paths: coordinates,
      strokeWeight: 0,
      fillOpacity: 0.45,
      draggable: true,
      editable: true
    });

    return polygon;
  }

  function fitPolygons(){
    google.maps.event.trigger(map, 'resize');
    //Fit polygons
    if (!latlngbounds.isEmpty()) {
      map.fitBounds(latlngbounds);
    } else {
      map.setCenter(new google.maps.LatLng(0, 0));
      map.setZoom(2);
    }
  }

  function setSelection(polygon) {
    clearSelection();
    selectedPolygon = polygon;
  }

  function deleteSelectedPolygon(){
    if(selectedPolygon !== undefined && selectedPolygon !== null) {
      selectedPolygon.setMap(null);
      selectedPolygon = null;
      polygonsSave();
    }
  }

  function deletePolygon(id){
    var l = polygons.length;

    for (var i = 0; i < l; i++) {
      if (polygons[i].id === id) {
        selectedPolygon = polygons.splice(i, 1)[0].polygon;
        break;
      }
    }

    deleteSelectedPolygon();
    _closeAllInfoWindows();
    return false;
  }

  function getMap() {
    return map;
  }

  function randomId() {
    return Math.random().toString(36).substring(2,15);
  }

  function create_listener(polygon, id){
    google.maps.event.addListener(polygon, 'click', function (this_zone) {
      return function(){

        setSelection(this_zone);

        var vertices = this_zone.getPath();
        var bounds = new google.maps.LatLngBounds();

        vertices.forEach(function(xy,i){
          bounds.extend(xy);
        });

        var contentString = '<div class="clearfix">' +
                              '<div class="map-icon">' +
                                '<i class=' + opts.domain_icon + '></i>' +
                              '</div>' +
                              '<div class="map-marker-info">' +
                                '<p>' + opts.domain_name + '</p>' +
                                '<p>' +
                                  '<a href="#" data-polygon-id="' + id + '" class="delete-polygon">' + I18n.t('javascript.delete_zone') + '</a>' +
                                '</p>' +
                              '</div>' +
                            '</div>';

        var info = new google.maps.InfoWindow({
          content: contentString,
          position: bounds.getCenter()
        });

        infoWindows.push(info);
        _closeAllInfoWindows();
        info.open(map);

      };
    }(polygon));
    google.maps.event.addListener(polygon, 'dragend', function() {
      polygonsSave();
    });

    google.maps.event.addListener(polygon.getPath(), 'set_at', function(){
      polygonsSave();
    });

    google.maps.event.addListener(polygon.getPath(), 'insertAt', function(){
      polygonsSave();
    });
  }

  function jsonMake() {
    var buf = '[';
    for (i = 0; i < polygons.length; i++) {
      buf += _comma(i) + jsonMakePaths(polygons[i].polygon.getPath().getArray());
    }
    buf += ']';
    return buf;
  }

  function jsonMakePaths(paths) {
    var n = paths.length;

    var buf = '[';
    for (var i = 0; i < n; i++) {
      var path = paths[i];

      buf += _comma(i) + '[' + path.lat() + ',' + path.lng() + ']';
    }
    buf += ']';

    return buf;
  }

  function _comma(i) {
    return (i > 0) ? ',' : '';
  }

  function _closeAllInfoWindows() {
    for (var i = 0; i < infoWindows.length; i++) {
      infoWindows[i].close();
    }
  }

  return {
    initialize: initialize,
    polygons: polygons,
    fitPolygons: fitPolygons,
    getMap: getMap,
    deletePolygon: deletePolygon,
    unbindEvents: unbindEvents
  }
}
