/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

//
// This module renders a heatMap using google maps module
// opts = {
//   container: The ID of the container of the map
//   data: The data that the heatmap should display
//   aggregation: The unit that we're assigning
//   type: satellite or roadmap
//   scrollwheel: specifies if the map should attend scroll wheel events
//   autoinit: Call initialize automatically
//   onLoad: event fired when the map is loaded, and after setting the data
//   grid_heatmap: true/false if grid heatmap wanted
//   styles: Style of the map. Check redBorder.mapStyles
// }
//
redBorder.heatMap = function (opts) {
  var data = {max: (opts.force_max || 1), data: []},
      map, $container,
      gridData = {},
      colors = generateColorGradient(),
      currentSquareId,
      TILE_SIZE = 32,
      heatmapOverlay;

  opts = opts || {};
  opts.autoinit = (opts.autoinit !== false);
  opts.clustered = (opts.clustered == true);
  opts.radius = (opts.radius || 15);
  opts.shadowBlur = (opts.shadowBlur || 15);
  opts.type = (opts.type || 'roads');

  if (opts.data !== undefined && opts.data.length > 0) {
    data = parseData(opts.data);
  }

  if (opts.autoinit === true) {
    initialize();
  }

  /* Generate green to yellow to red hex colors array */
  function generateColorGradient(){
    var c = []
    var n;
    for (var i = 0; i <= 255; i++){
      n = i.toString(16)
      if (n.length == 1)
        n = "0" + n;
      c.push("#ff" + n + "00" );
    }
    for (var i = 255; i >= 0; i--){
      n = i.toString(16)
      if (n.length == 1)
        n = "0" + n;
      c.push("#" + n + "ff" + "00" );
    }
    return c;
  }

  function setQualityData(){
    if (currentSquareId != undefined){
      $("#" + currentSquareId).parent().remove()
    }

    map.center();
    map.getMap().overlayMapTypes.insertAt(0,
      new CoordMapType(new google.maps.Size(TILE_SIZE, TILE_SIZE)));
  }

  function initialize() {
    $container = $('#' + opts.container);
    map = redBorder.map({
      container: opts.container,
      searchBox: false,
      bounds: data.data,
      scrollwheel: opts.scrollwheel,
      mirrorControls: opts.mirrorControls,
      clustered: opts.clustered,
      calculator: opts.calculator,
      styles: opts.styles,
      type: opts.type,
      onLoad: function (map, mouseEvent) {
        if (opts.grid_heatmap == true){
          setQualityData();
        }
        else
          setData();
        if (opts.onLoad !== undefined) opts.onLoad(map, mouseEvent);
      },
      onDrag: function () {
        if (opts.grid_heatmap != true) heatmapOverlay.draw();
      }
    });

    google.maps.event.addListener(map.getMap(), 'zoom_changed', function() {
      if (opts.grid_heatmap == true) gridData = dataToGrid();
    });

    heatmapOverlay = new HeatmapOverlay(getMap(), {
      radius: opts.radius,
      shadowBlur: opts.shadowBlur,
      visible: true,
      opacity: 60,
      logscale: true
    });

    bindEvents();
  }

  CoordMapType.prototype.getTile = function(coord, zoom, ownerDocument) {
    var div = ownerDocument.createElement('div');
    div.id = coord.x + "-" + coord.y
    div.style.width = this.tileSize.width + 'px';
    div.style.height = this.tileSize.height + 'px';
    div.style.opacity = 0.6;

    for (var i = 0; i < gridData.length; i++){
      if (gridData[i].x == coord.x && gridData[i].y == coord.y ){
        div.style.background = colors[Math.floor(gridData[i].count * (colors.length - 1) / 100)]
        i = gridData.length;
      }
    }

    currentSquareId = div.id;
    return div;
  };

  function CoordMapType(tileSize) {
    this.tileSize = tileSize;
  }

  function latLngToTile(LatLng){
    var c = [];
    var pix_from_origin_x = map.getMap().getProjection().fromLatLngToPoint(LatLng).x * Math.pow(2, map.getMap().getZoom());
    var pix_from_origin_y = map.getMap().getProjection().fromLatLngToPoint(LatLng).y * Math.pow(2, map.getMap().getZoom());
    c.push({
      x: Math.floor(pix_from_origin_x / TILE_SIZE),
      y: Math.floor(pix_from_origin_y / TILE_SIZE)
    });
    return c
  }

  function dataToGrid(){
    var gridData = []

    for (var i = 0; i < data.data.length; i++){
      var loc = new google.maps.LatLng(data.data[i].lat, data.data[i].lng);
      var coord = latLngToTile(loc);
      var flag = false;

      for (var j = 0; j < gridData.length; j++){
        if (gridData[j].x == coord[0].x && gridData[j].y == coord[0].y ){
          gridData[j].count = gridData[j].count + data.data[i].count;
          gridData[j].n = gridData[j].n + 1;
          flag = true;
          j = gridData.length;
        }
      }
      if(!flag){
        gridData.push({
          x: coord[0].x,
          y: coord[0].y,
          n: 1,
          count: data.data[i].count
        })
      }
    }

    for (var i = 0; i < gridData.length; i++){
      if (gridData[i].n > 1){
        gridData[i].count = gridData[i].count / gridData[i].n;
      }
    }

    return gridData;
  }

  function parseData(unparsedData) {
    var mapData = {max: (opts.force_max || 1), data: []},
        parsedData;

    parsedData = unparsedData.filter(function (d) {
      return (d.client_latlong !== undefined &&
        d.client_latlong !== null &&
        d.client_latlong.split(",").length === 2);
    });

    parsedData.forEach(function (element) {
      var res = element.client_latlong.split(",");

      // Calculate max
      if (element[opts.aggregation] > mapData.max) mapData.max = element[opts.aggregation];

      mapData.data.push({
        lat: parseFloat(res[0]),
        lng: parseFloat(res[1]),
        count: element[opts.aggregation]
      });
    });

    return mapData;
  }

  function setMax(max){
    opts.force_max = max;
    data.max = max;
  }

  function bindEvents() {
    if (opts.tab == false) return;

    $(document).off('click', '#tab_' + opts.tab + ' table td.serie a.coord_serie');
    $(document).on('click', '#tab_' + opts.tab + ' table td.serie a.coord_serie', (function (markers, heatmapOverlay) {
      return function() {
        var position = $(this).data('coord');

        if (position !== "") {
          if ($(this).hasClass('markerAdded')) {
            $(this).removeClass('markerAdded');

            markers[position].setMap(null);
          } else {
            var latLngArray = position.split(',');

            if (latLngArray.length === 2) {
              $(this).addClass('markerAdded');

              var latLng = new google.maps.LatLng(latLngArray[0], latLngArray[1]),
                element_for_title = $(this).parents('.selectable-serie').find('td.last');

              markers[position] = new google.maps.Marker({
                  position: latLng,
                  map: heatmapOverlay.map,
                  title: element_for_title.attr('title') || element_for_title.text()
              });
            }
          }
        }
      }
    })({}, heatmapOverlay));
  }

  function wasInitialize() {
    return (map !== undefined);
  }

  function getRbMap() {
    return map;
  }

  function getMap() {
    return map.getMap();
  }

  function resize() {
    if (!wasInitialize()) return;
    map.reload();
    reloadData();
  }

  function setData(nocenter) {
    if(map !== undefined) {
      data.data.forEach(function (d) {
        map.extendBounds(d.lat, d.lng);
      });

      if(nocenter === undefined || nocenter === false){
        map.center();
      }

      heatmapOverlay.setDataSet(data);
    }
  }

  function reloadData(_data, nocenter) {
    if (_data !== undefined){
      data = parseData(_data);
    }
    heatmapOverlay.toggle();
    if (opts.grid_heatmap == true)
      setQualityData();
    else
      setData(nocenter);

    // Force the draw of heatmap
    setTimeout(function(){
      heatmapOverlay.draw();
      heatmapOverlay.toggle();
    }, 400);
  }

  // TODO. Mix with simple reloadData()
  function reloadData_dynamic(_data, nocenter){
    if (_data !== undefined){
      data = parseData(_data);
    }

    setData(nocenter);

    // Force the draw of heatmap
    heatmapOverlay.draw();
  }

  function addData(unparsedData) {
    if (unparsedData === undefined || unparsedData.length === 0) return;
    var parsedData = parseData(unparsedData);

    if (parsedData.max > data.max) data.max = parsedData.max;
    data.data = data.data.concat(parsedData.data);

    if (opts.grid_heatmap == true)
      setQualityData();
    else
      setData();
  }

  function getData() {
    return data;
  }

  function setOptions(opts) {
    map.setOptions(opts);
  }

  return {
    initialize: initialize,
    wasInitialize: wasInitialize,
    resize: resize,
    addData: addData,
    getData: getData,
    reloadData: reloadData,
    reloadData_dynamic: reloadData_dynamic,
    getRbMap: getRbMap,
    getMap: getMap,
    setOptions: setOptions,
    setMax: setMax
  };
}
