/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

/*
This module renders draw aps, zones, domains, clients... in a given map.
The target map can be a redBorder heatmap o a redborder map. When you add floors,
the library try to load images with access points. If the image is not present,
a map will be showed.

@param $container [JQuery Selector] Container of the map
@param $floorContainer [JQuery Selector] Container of the floor image
@param map [redBorder.map, redBorder.heatmap] Map to render the view.
@param opts [Object] Options for the library:
  * autoinit: Call initialize automatically
  * heatmap: Indicates if the given map is a heatmap
  * domains: Domains to represent.
  * floors: Floors to represent.
  * bindEvents: Set events for domains.
  * onClickDomain: Callback to call when the user click in a domain
  * domainFillColor: Color to fill domain zones
  * domainStrokeColor: Color for border domain zones
*/

redBorder.infrastructureMap = function ($parent, $mapContainer, $floorContainer, map, opts) {
  var apsAdded = [],
      $floorWrap,
      floorsAdded = [],
      floor_aps = [],
      activeFloor,
      isFloor = false,
      ref = -1,
      // For Floor view
      buildingBounds,
      dimensions,
      radius;

  // Color palette for access points
  var palette = ['#90A860', '#789937', '#486018', '#304800', '#183000'];

  // Set the options
  opts = opts || {};
  opts.autoinit         = (opts.autoinit !== false);
  opts.heatmap          = (opts.heatmap !== false);
  opts.domains          = (opts.domains) || [];
  opts.parentDomains    = (opts.parentDomains) || [];
  opts.showDomainsAsMarkers   = opts.showDomainsAsMarkers === true;
  opts.floors           = (opts.floors) || [];
  opts.floors.reverse();
  opts.bindEvents       = (opts.bindEvents !== false);
  opts.onClickDomain    = opts.onClickDomain;
  opts.onChangeFloor    = opts.onChangeFloor;
  // Color for polygons
  opts.domainFillColor = (opts.domainFillColor || '#f9cbb8')
  opts.domainStrokeColor = (opts.domainStrokeColor || '#f9cbb8')

  // Initialize the instance
  if (opts.autoinit === true) {
    initialize();
  }

  // Initialize the library. We need to change the behaviour depends the
  // type of domain to represent.
  function initialize() {
    if (opts.floors.length > 0) {
      // Render the floors
      // Get the first element
      opts.floors = opts.floors.sort(sortFloors).reverse();
      $mapContainer.hide();
      $floorContainer.append('<div class="infrastructure-map-wrap"></div>');
      $floorWrap = $floorContainer.find('.infrastructure-map-wrap');
      addFloors(opts.floors);
      levelSelector(opts.floors);
      bindFloorEvents();

      // Main length
      var l = opts.floors.length;
      // Store floor aps
      for(var i = 0; i < l; i++){
        if (opts.floors[i].access_points && opts.floors[i].access_points.length > 0) {
          floor_aps = floor_aps.concat(opts.floors[i].access_points);
        }
      }
    } else {
      // Render the domains
      if ($floorContainer && $floorContainer.length > 0) {
        $floorContainer.hide();
      }

      if (opts.domains && opts.domains.length > 0) {
        addDomains(opts.domains);
      } else {
        centerMapInBuilding();
      }
    }
  }

  /* -------------------
   *  Bind events
   * ------------------- */

  function bindFloorEvents() {
    $(document).on('click', '.infrastructure-floor-level', function(e){
      var index = $(this).data('index');
      if(opts.floors[index]){
        floor_uuid = opts.floors[index].uuid
      } else {
        floor_uuid = 'all'
      }
      changeFloor(floor_uuid);
      $('.overview-time a').attr('data-floor', floor_uuid);
    });

    $(document).on('mouseenter', 'svg.rb-aps circle', function(e){
      $tooltip.css({
        "top": $(this).position().top - $tooltip.parent().offset().top - 25,
        "left": $(this).position().left - $tooltip.parent().offset().left + 40
      });
      fill_aps_tooltip($(this).data('data'));
      $tooltip.show();
    });

    $(document).on('mouseout', 'svg.rb-aps circle', function(e){
      $tooltip.hide();
    });
  }

  function unbindEvents() {
    $(document).off('click', '.infrastructure-floor-level');
    $(document).off('mouseout', 'svg.rb-aps circle');
    $(document).off('mouseenter', 'svg.rb-aps circle');
  }

  /* -------------------
   *  AccessPoints
   * ------------------- */

  function fill_aps_tooltip(element) {
    var last_check_in, tooltip;

    tooltip = $('<div class="rb-aps-tooltip-field">' + I18n.t('javascript.name') + '<span>' + element.name + '</span></div>' +
      '<div class="rb-aps-tooltip-field">MAC: <span>' + element.mac_address + '</span></div>');

    $tooltip.html(tooltip);

    if (element.last_check_in != undefined && element.last_check_in != null) {
      last_check_in = new Date(element.last_check_in).toLocaleString();
        $tooltip.append('<div class="rb-aps-tooltip-field">' +
          I18n.t('javascript.last_check') + last_check_in + '</span>' +
        '</div>');
    }
  }

  /*
   * Calculate the reference of the number of clients in APS.
   *
   * @param aps [Array] Access Points to calculate the reference
   */
  function calculateReference(aps){
    if (aps === undefined || aps.length === 0) {
      ref = 1;
      return;
    }

    // Get the max number of APs
    aps.forEach(function (d) {
      if (d !== undefined) {
        ref = (d.clients > ref) ? d.clients : ref;
      }
    });

    // Calc the reference
    if (ref > 0){
      ref = (ref + 1)/palette.length;
    } else {
      ref = 1;
    }
  }

  /*
   * Add access points in a map
   */
  function addAccessPoints(apList, nocenter){
    if (apList === undefined) {
      return;
    }

    calculateReference(apsAdded.concat(apList));

    apList.forEach(function (d) {
      if(d.latitude != undefined && d.longitude != undefined){
        addAccessPoint(d, nocenter);
      }
    });

    if(nocenter){
      getRbMap().repaintCluster();
    }
  }

  function addAccessPoint(ap, nocenter){
    var status = ap.status;
    var clients = ap.clients;
    var last_check_in = new Date(ap.last_check_in);

    if(status === 1){
      // Change the color based on the number of Clients
      var color = palette[parseInt(clients/ref)];
    } else if(status === 0) {
      var color = 'orange';
    } else {
      var color = 'red';
    }

    if(clients > 0){
      var text = clients;
    }else{
      var text = '';
    }
    getRbMap().addCustomMarker(
      ap.latitude,
      ap.longitude,
      { text: text,
        backgroundColor: color,
        content: '<b>' + I18n.t('javascript.name') + '</b> ' + ap.name + '<br>' +
                 '<b>' + I18n.t('javascript.mac_address') + '</b> ' + ap.mac_address + '<br>' +
                 '<b>' + I18n.t('javascript.clients') + '</b> ' + clients + '<br>' +
                 '<b>' + I18n.t('javascript.last_check') + '</b> ' + last_check_in.toLocaleString() + '<br>' +
                 '<b>' + I18n.t('javascript.domain') + '</b> <a href="/dashboard/sensors/' + ap.sensor_id + '/overview"><u>' +
                 ap.sensor_name + '</u></a><br>'
      },
      nocenter
    );

    apsAdded.push(ap);
  }

  /*
   * Update AccessPoints. If we are showing access points in a single map,
   * we delete all access point and add it. If we are in building view,
   * we update all access points of floors and redisplay current floor.
   */
  function updateAccessPoints(data) {
    if (floorsAdded.length > 0) {
      // Update Aps in the floors
      var l = data.length;

      for(var i = 0; i < l; i++) {
        var name = data[i].name,
            aps = data[i].access_points,
            floor;

        floor = floorsAdded.filter(function(el) { return el.name === name })[0];

        if (floor !== undefined) {
          floor.access_points = aps;
        }
      }
      // Remove and add
      levelSelector(floorsAdded.sort(sortFloors).reverse());
      removeFloorAccessPoints();
      displayFloor(activeFloor, false);
    } else {
      // Remove and add
      removeAccessPoints();
      addAccessPoints(data, true);
    }
  }

  /*
   * Remove access points in Floor image
   */
  function removeFloorAccessPoints() {
    $floorWrap.find('.rb-aps').remove();
  }

  /*
   * Remove access points in map
   */
  function removeAccessPoints(){
    getRbMap().removeCustomMarkers();
    apsAdded = [];
  }

  /*
   * Add Aps to an image
   */
  function addFloorAps(access_points) {
    var l = access_points.length,
        x, y,
        scale_x = dimensions.width / dimensions.originalWidth,
        scale_y = dimensions.height / dimensions.originalHeight;
    // Set the width and height of the circles
    radius = calcRadius();

    for (var i = 0; i < l; i++) {
      var ap = access_points[i];

      if(ap.x !== undefined && ap.y !== undefined) {
        var xFull = (ap.x * dimensions.originalWidth / dimensions.widthFeet),
            yFull = (ap.y * dimensions.originalHeight / dimensions.heightFeet);
        // Get real X and Y
        x = xFull * scale_x;
        y = yFull * scale_y;

        // Render the AP
        displayFloorAp(ap, x, y);
      }
    }
  }

  /*
   * Calculate circle radius
   */
  function calcRadius() {
    var radius;

    if (dimensions.width > 1200) {
      radius = 18;
    } else if (dimensions.width > 900) {
      radius = 16;
    } else if (dimensions.width > 700) {
      radius = 14;
    } else if (dimensions.width > 500) {
      radius = 12;
    } else {
      radius = 10;
    }

    return radius;
  }

  /*
   * Render an Access Point as SVG.
   */
  function displayFloorAp(ap, x, y) {
    y = (y < 15) ? 15 : parseInt(y);
    x = (x < 15) ? 15 : parseInt(x);

    var size = (radius * 2 + 4),
        pos  = (radius * 2 + 4)/2;
        x_text = 4,
        font_size = 16;

    if (ap.clients == 0){
      var clients_text = '';
    } else {
      var clients_text = ap.clients;
    }

    if (ap.clients > 100) {
      font_size = 13;
      x_text = 11;
    } else if(ap.clients > 10){
      x_text = 8;
    }

    if(ap.clients > 0 || ap.status === 1){
      // Change the color based on the number of Clients
      var color = palette[parseInt(ap.clients/ref)];
    } else if(ap.status === 0) {
      var color = 'orange';
    } else {
      var color = '#d80808';
    }

    var marker = $(' \
      <svg height="'+size+'" width="'+size+'" class="rb-aps zone-ap" data-rb-aps-id="' + ap.id + '" \
        style="position: absolute; top: ' + (y - pos) + 'px; left: ' + (x - pos) + 'px" \
        data-rb-aps-name="' + ap.name + '" > \
        <circle cx="'+pos+'" cy="'+pos+'" r="'+ radius +'" class="rb-aps-ap" data-rb-status="' + ap.status + '" style="fill: '+ color +'" /> \
        <text fill="#fff" x="'+(pos - x_text)+'" y="'+(pos + 5)+'" font-size="' + font_size + '">'+ clients_text +'</text> \
      </svg> \
    ');

    $('circle', marker).data('data', ap);
    $floorWrap.append(marker);
  }

  /*
   * Show a level selector to change the floor
   */
  function levelSelector(floors) {
    var html,
        l = floors.length;

    // Create the list
    html = '<ul>';
    for (var i = 0; i < l; i++) {
      var active = floors[i] === activeFloor ? 'active' : '';
      html += '<li class="infrastructure-floor-level ' + active + '" data-index="'+ i +'" data-uuid="' + floors[i].uuid + '">' +
                levelIcon(floors[i]) + floors[i].name + '</li>';
    }
    // Include ALL element
    html += '<li class="infrastructure-floor-level" data-index="all">' + I18n.t('javascript.all') + '</li>';
    html += '</ul>';

    // Insert selector
    if ($parent.find('.floor-selector').length > 0){
      $parent.find('.floor-selector').replaceWith('<div class="floor-selector">' + html + '</div>');
    } else {
      $parent.append('<div class="floor-selector">' + html + '</div>');
      // Scroll to bottom
      $('.floor-selector').scrollTop($('.floor-selector')[0].scrollHeight);
    }
  }

  /* -------------------
   *  Floors
   * ------------------- */

  /*
   * Add floors to the view
   */
  function addFloors(floors){
    if (floors.length === 0) { return; }
    floorsAdded = floors;
    // Display first element
    displayFloor(floors[floors.length - 1], true);
  }

  /*
   * Display given floor. First we need to try to load the image associated
   * to the floor. If no image present, we will use the map.
   */
  function displayFloor(floor, centerMap) {
    // Remove last floor
    $floorWrap.html('');
    activeFloor = floor;

    if (floor.image !== undefined && floor.image !== '') {
      // Show the image
      var $image;

      // Show and hide elements
      $mapContainer.hide();
      $floorContainer.show();

      // Init image and calc aspect ratio
      $image = $('<img class="infrastructure-image" src="' + floor.image + '" alt=' + I18n.t('javascript.floor_image') + '/>');
      $floorWrap.append($image);

      // Load image with a fixed aspect ratio
      redBorder.utils.imageAspectRatio($image, $floorWrap, {
        widthFeet: floor.dimensions[0],
        heightFeet: floor.dimensions[1]
      }).load().then(function(_dimensions) {
        // Build the tooltip
        $tooltip = $('<div class="rb-aps-tooltip"></div>');
        $tooltip.hide();
        $floorWrap.append($tooltip);
        // Store dimensions
        dimensions = _dimensions;
        // Add access points
        if (floor.access_points !== undefined && floor.access_points.length > 0) {
          addFloorAps(floor.access_points);
        }
      });
    } else {
      // Show and hide elements
      $mapContainer.show();
      $floorContainer.hide();
      // Reload the map
      map.reload();
      removeAccessPoints();
      if (floor.access_points !== undefined && floor.access_points.length > 0) {
        // Add access points
        addAccessPoints(floor.access_points, !centerMap);
      } else {
        // Center in building
        centerMapInBuilding()
      }
    }
  }

  function changeFloor(floor_uuid) {
    var onChangeFloor = opts.onChangeFloor;
    var floor = opts.floors.find(function (elem, i, arr) {
      if (elem.uuid === floor_uuid) {
        return true;
      } else {
        return false;
      }
    });
    if (floor === undefined && floor_uuid !== 'all') { return; }
    if (floor_uuid === 'all') {
      if (activeFloor === 'all') { return; }
      // Show and hide elements
      $mapContainer.show();
      $floorContainer.hide();
      // Reload the map
      map.reload();
      removeAccessPoints();
      if (floor_aps.length > 0){
        // Add access points
        addAccessPoints(floor_aps, false);
      } else {
        // Center map in building zone
        centerMapInBuilding();
      }
      // Callback
      if ( typeof onChangeFloor === 'function' ) {
        // Callback when run the code
        onChangeFloor.call(this, 'all');
      }
      activeFloor = 'all';
    } else {
      // Check if it's active
      if (floor === activeFloor) { return; }
      displayFloor(floor, true);
      // Callback
      if ( typeof onChangeFloor === 'function' ) {
        // Callback when run the code
        onChangeFloor.call(this, floor);
      }
    }
    var uuid = typeof floor !== 'undefined' && floor.uuid ? floor.uuid : 'all';
    // Remove active class from all elements
    $parent.find('.infrastructure-floor-level').removeClass('active');
    // Add active class based on uuid value
    if (uuid == 'all') {
        $parent.find('.infrastructure-floor-level[data-index="all"]').addClass('active');
    } else {
        $selector = $parent.find('.floor-selector');
        $selector.find('.infrastructure-floor-level[data-uuid="' + uuid + '"]').addClass('active');
    }
  }

  /* -------------------
   *  Domains
   * ------------------- */

  function addDomains(domainList){
    domainList.forEach(function (d) {
      addDomain(d);
    });
  }

  function addDomain(domain){
    for(var j = 0; j < domain["domain_zones"].length; j++){
      var polygon_tmp = [],
          polygon_bounds = new google.maps.LatLngBounds();
      // Iterate over zone's bounds
      for(var z = 0; z < domain["domain_zones"][j].length; z++){
        var lat = domain["domain_zones"][j][z][0],
            lng = domain["domain_zones"][j][z][1],
            latlng = new google.maps.LatLng(lat, lng);
        polygon_tmp.push({
          lat: lat,
          lng: lng
        });
        getRbMap().extendBounds(lat, lng);
        polygon_bounds.extend(latlng);
      }

      var contentString = ' \
        <div class="map-icon"> \
          <i class="'+ domain["icon"] +'"></i> \
        </div> \
        <div class="map-marker-info"> \
          <p>' + domain['title'] + '</p> \
        </div> \
        <div class="map-options"> \
          <a class="block-ui-but" href="/dashboard/sensors/' + domain['id'] + '/overview"> \
            <i class="icon fa fa-eye"></i> \
          </a> \
        </div>';

      if (!opts.showDomainsAsMarkers) {
        if(opts.bindEvents){
          var polygon = getRbMap().addPolygon(polygon_tmp, {
            clickable: true,
            fillColor: opts.domainFillColor,
            fillOpacity: 0.5,
            strokeWeight: 0.5,
            strokeColor: opts.domainStrokeColor,
            strokeOpacity: 0.7,
            onClick: (function (id) {
              return function (polygon, map, event) {
                var onClickDomain = opts.onClickDomain;
                if ( typeof onClickDomain === 'function' ) {
                  // Callback when run the code
                  onClickDomain.call(this, id);
                }
              }
            })(domain.id)
          });
        } else {
          var polygon = getRbMap().addPolygon(polygon_tmp, {
            clickable: true,
            fillColor: opts.domainFillColor,
            fillOpacity: 0.5,
            strokeWeight: 0.5,
            strokeColor: opts.domainStrokeColor,
            strokeOpacity: 0.7,
            content: contentString
          });
        }
      }

      // Add the marker
      getRbMap().addMarker(
        polygon_bounds.getCenter().lat(),
        polygon_bounds.getCenter().lng(),
        {
          content: contentString
        },
        true
      );
    }
  }

  /*
   * Remove all domains (polygons) from the map
   */
  function removeDomains(){
    getRbMap().removePolygons();
    getRbMap().removeMarkers();
  }

  /* -------------------
   *  Utils
   * ------------------- */

  /*
   * Center the map in building
   */
  function centerMapInBuilding() {
    if (opts.parentDomains === undefined || opts.parentDomains.length === 0) { return; }
    //Fit polygons
    if (opts.parentDomains && opts.parentDomains["domain_zones"].length > 0) {
      var map = getRbMap().getMap();
      google.maps.event.trigger(map, 'resize');
      map.fitBounds(getBuildingBounds());
    }
  }

  /*
   * Calculate building bounds
   */
  function getBuildingBounds() {
    if (buildingBounds === undefined) {
      buildingBounds =
        redBorder.utils.map.getPolygonsBounds(opts.parentDomains["domain_zones"]);
    }
    // Return it
    return buildingBounds;
  }

  /*
   * Return the icon to show in a floor level based on the access point status
   */
  function levelIcon(floor) {
    var color;
    switch (checkApsUp(floor.access_points)) {
      case true:
        color = 'green';
        break;
      default:
        color = 'red';
    }
    // Return the icon
    return '<i class="fa fa-circle ' + color + '"></i>';
  }

  /*
   * Check if all access points are up in the given collection.
   */
  function checkApsUp(access_points) {
    var l = access_points.length,
        up = true;
    // Iterate
    for(var i = 0; i < l; i++){
      if (access_points[i].status !== 1) {
        up = false;
        break;
      }
    }
    return up;
  }

  /*
   * Check if the map was initialized
   */
  function wasInitialize() {
    return (getRbMap() !== undefined);
  }

  /*
   * Get redBorder.map
   */
  function getRbMap() {
    if (opts.heatmap){
      return map.getRbMap();
    } else {
      return map;
    }
  }

  /*
   * Get the Heatmap
   */
  function getRbHeatmap(){
    if (opts.heatmap){
      return map;
    } else {
      return undefined;
    }
  }

  /*
   * Resize the current map
   */
  function resize() {
    if (!wasInitialize()) { return; }
    if (opts.heatmap){
      getRbHeatmap().resize();
    } else {
      getRbMap().reload();
      getRbMap().center();
    }
  }

  /*
   * Sort floors by floor level and name
   */
  function sortFloors(a, b) {
    // Order by floor_level first
    if (a.floor_level < b.floor_level) {
      return -1;
    } else if (a.floor_level > b.floor_level) {
      return 1;
    } else {    // Floor levels are equal, compare name
      if (a.name < b.name) {
        return -1;
      } else if (a.name > b.name) {
        return 1;
      }
    }
    // If no return has been executed return equality
    return 0;
  }

  return {
    initialize: initialize,
    wasInitialize: wasInitialize,
    resize: resize,
    addDomains: addDomains,
    addDomain: addDomain,
    removeDomains: removeDomains,
    updateAccessPoints: updateAccessPoints,
    addAccessPoints: addAccessPoints,
    addAccessPoint: addAccessPoint,
    removeAccessPoints: removeAccessPoints,
    unbindEvents: unbindEvents,
    getRbMap: getRbMap,
    getRbHeatmap: getRbHeatmap,
    changeFloor: changeFloor
  };
}
