/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

//
// This module renders a Google map and lets you add markers on it
// opts = {
//   container: The ID of the container of the map
//   searchBox: Specifies if the map should have a search box (default true)
//   scrollwheel: Specifies if the map should attend scroll wheel events (default true)
//   type: satellite or roadmap
//   data: initial marker data for the map. Will be passed to the addMarker method.
//   dataOpts: opts to apply to the markers added with the data property.
//   bounds: Array of objects with lat and lng properties on it
//   clustered: Specifies if the markers should form a cluster (default false)
//   autoinit: Call initialize automatically
//   onClick: Event fired when clicking on any point of the map
//   onLoad: Event fired when the map is loaded
//   onDrag: Event fired when the user draggs (on drag end)
// }
//

// Store all the maps of redBorder. It's store by section
redBorder.maps = {}

redBorder.map = function (opts) {
  var infos = [],
      markers = [],
      customMarkers = [],
      polygons = [],
      polylines = [],
      markersToAdd = [],
      polygonsToAdd = [],
      $container,
      mapOptions,
      map,
      bounds,
      oldMarker,
      oldHighlight,
      markerClusterer,
      markerSpiderfier;

  var palette = [
    "#638bdf", "#f26e57", "#89a54e", "#e89c4f", "#946acc",
    "#1758b5", "#c74531", "#3e812d", "#d87e34", "#734ca2",
    "#63a0e7", "#f4988a", "#a2bb77", "#edbc6c", "#be90fd",
    "#0f3742", "#783234", "#55633f", "#fdc52b", "#371c39",
    "#38606b", "#996062", "#6f795f", "#878787", "#8a3b8f",
    "#5d8d9a", "#ae9596", "#a8b298", "#bcbcbc", "#eb80f1"
  ]

  // Defaults values
  opts = opts || {};
  opts.container    = (opts.container || 'map_container');
  opts.searchBox    = (opts.searchBox !== false);
  opts.scrollwheel  = (opts.scrollwheel !== false);
  opts.draggable    = (opts.draggable !== false);
  opts.clustered    = (opts.clustered == true);
  opts.dataOpts     = (opts.dataOpts || {});
  opts.autoinit     = (opts.autoinit !== false);
  opts.fullScreen   = (opts.fullScreen === true);
  opts.disableDoubleClickZoom = (opts.disableDoubleClickZoom === true);
  opts.mirrorControls = (opts.mirrorControls === true);
  opts.hideUI       = (opts.hideUI === true);

  if (opts.autoinit === true) {
    initialize();
  }

  function initialize() {
    // Check if google maps is defined. If not, then we return and do nothing
    if (google.maps.LatLng === undefined) {
      throw new Error(I18n.t('javascript.google_maps'));
    }

    // Map default settings
    mapOptions = {};
    mapOptions.zoom = 3;
    mapOptions.center = new google.maps.LatLng(25.6586, -80.3568)
    mapOptions.scrollwheel = opts.scrollwheel;
    mapOptions.draggableCursor = opts.draggableCursor;
    mapOptions.draggingCursor = opts.draggingCursor;
    mapOptions.draggable = opts.draggable;
    mapOptions.disableDoubleClickZoom = opts.disableDoubleClickZoom;
    mapOptions.zoomControlOptions = { position: google.maps.ControlPosition.RIGHT_TOP }
    mapOptions.streetViewControlOptions = { position: google.maps.ControlPosition.RIGHT_TOP }
    mapOptions.tilt = 0;
    mapOptions.mapTypeControlOptions = {
      style: google.maps.MapTypeControlStyle.DROPDOWN_MENU
    };

    if (opts.styles !== undefined && opts.styles.constructor === Array) {
      mapOptions.styles = opts.styles;
    } else {
      mapOptions.styles = redBorder.mapStyles.brown;
    }

    if (opts.mirrorControls){
      mapOptions.zoomControlOptions = {position: google.maps.ControlPosition.LEFT_BOTTOM }
      mapOptions.streetViewControlOptions = {position: google.maps.ControlPosition.LEFT_BOTTOM }
      mapOptions.panControlOptions = {position: google.maps.ControlPosition.LEFT_BOTTOM }
      mapOptions.mapTypeControlOptions = {position: google.maps.ControlPosition.RIGHT_BOTTOM }
    }

    if (opts.hideUI) {
      mapOptions.disableDefaultUI = true;
    }

    if (opts.type === 'satellite') {
      mapOptions.mapTypeId = google.maps.MapTypeId.SATELLITE;
    } else {
      mapOptions.mapTypeId = google.maps.MapTypeId.ROADMAP;
    }

    bounds = new google.maps.LatLngBounds();

    // Set the bounds from opts
    if (opts.bounds !== undefined && opts.bounds.length > 0) {
      opts.bounds.forEach(function (bound) {
        extendBounds(bound.lat, bound.lng);
      });

      center();
    }

    $container = $('#' + opts.container);
    map = new google.maps.Map($container[0], mapOptions);

    // Creates the marker cluster if specified in opts
    if (opts.clustered && opts.calculator) {
      markerClusterer = new MarkerClusterer(map, markers, {
        maxZoom: 15,
        gridSize: 50,
        calculator: opts.calculator
      });
    } else if (opts.clustered) {
      markerClusterer = new MarkerClusterer(map, markers, {
        maxZoom: 15,
        gridSize: 50
      });
    }

    markerSpiderfier = new OverlappingMarkerSpiderfier(map, {
        keepSpiderfied: true,
        event: 'mouseover'
      });

    // If there is data to be appended, parse it
    if (opts.data !== undefined && opts.data.length > 0) {
      addMarkers(opts.data, opts.dataOpts);
    }

    // If we have markers to add, lets add it
    markersToAdd.forEach(function (d) {
      addMarker(d.lat, d.lng, d.opts);
    });
    markersToAdd = [];

    // If we have markers to add, lets add it
    polygonsToAdd.forEach(function (d) {
      addPolygon(d.vertexes, d.opts);
    });
    polygonsToAdd = [];

    // Creates the search box if specified in opts
    if (opts.searchBox) createSearchBox();
    if (opts.contextMenu) createContextMenu();

    // Creates listeners for these events
    if (opts.onClick !== undefined) onClick(opts.onClick);
    if (opts.onLoad !== undefined) onLoad(opts.onLoad);
    if (opts.onDrag !== undefined) onDrag(opts.onDrag);

    if(opts.fullScreen === true){
      expand_b = "<i class='expand-map fa fa-expand'></i>"
      $container.prepend(expand_b);

      $('.expand-map', $container).on('click', function() {
        fullScreen();
      });
    }
  }

  //
  // Creates the context menu for the map
  //
  function createContextMenu() {
    var contextMenuOptions = {};
    contextMenuOptions.classNames = {menu: 'context_menu', menuSeparator: 'context_menu_separator'};

    //  create an array of ContextMenuItem objects
    var menuItems = [];
    menuItems.push({ className: 'context_menu_item', eventName: 'zoom_in_click', label: I18n.t('javascript.zoom_in') });
    menuItems.push({ className: 'context_menu_item', eventName: 'zoom_out_click', label: I18n.t('javascript.zoom_out') });
    //  a menuItem with no properties will be rendered as a separator
    menuItems.push({});
    menuItems.push({ className: 'context_menu_item', eventName: 'center_map_click', label: I18n.t('javascript.center_map') });
    if (window.redBorder.admin_user === true) {
      menuItems.push({});
      menuItems.push({ className: 'context_menu_item', eventName: 'create_ap_gmap', label: I18n.t('javascript.create_ap') });
      menuItems.push({ className: 'context_menu_item', eventName: 'locate_ap_gmap', label: I18n.t('javascript.locate_ap') });
    }

    contextMenuOptions.menuItems = menuItems;

    //  create the ContextMenu object
    var contextMenu = new ContextMenu(map, contextMenuOptions);

    //  display the ContextMenu on a Map right click
    google.maps.event.addListener(map, 'rightclick', function(mouseEvent){
      contextMenu.show(mouseEvent.latLng);
    });

    google.maps.event.addListener(contextMenu, 'menu_item_selected', function(latLng, eventName){
      //  latLng is the position of the ContextMenu
      //  eventName is the eventName defined for the clicked ContextMenuItem in the ContextMenuOptions
      switch(eventName){
        case 'create_ap_gmap':
          create_ap_gmap(latLng);
          break;
        case 'locate_ap_gmap':
          locate_ap_gmap(latLng);
          break;
        case 'zoom_in_click':
          map.setZoom(map.getZoom()+1);
          break;
        case 'zoom_out_click':
          map.setZoom(map.getZoom()-1);
          break;
        case 'center_map_click':
          map.panTo(latLng);
          break;
      }
    });
  }

  function create_ap_gmap(latLng) {
    var form_url = $container.data('rb-aps-create-ap-url');
    $('#new_ap_modal').modal('show');

    $.get(form_url, { point: [latLng.lng(), latLng.lat()], latLngFormat: true }, function (data) {
      $('#new_ap_modal .modal-content').html(data);
    });
  }

  function locate_ap_gmap(latLng) {
    var form_url = $container.data('rb-aps-locate-ap-url');
    $('#new_locate_ap_modal').modal('show');

    $.get(form_url, { point: [latLng.lng(), latLng.lat()], latLngFormat: true }, function (data) {
      $('#new_locate_ap_modal .modal-content').html(data);
    });
  }

  //
  // Creates the search box for the map
  // Creates a input element for the search, and assigns it to the map
  //
  function createSearchBox() {
    var input,
        searchBox;

    // Create the input node
    $input = $('#gmaps-input', $container);

    if ($input.length === 0) {
      $container.append('<input id="gmaps-input" class="gmaps-controls" type="text" placeholder="' + I18n.t('javascript.search_box') + '">');
      $input = $('#gmaps-input', $container);
    }

    // Create a keypress listener to avoid submiting forms
    $input.keypress(function (event) {
      if (event.which == 13) {
        event.preventDefault();
        return false;
      }
    });

    // Creates the searchBox object
    searchBox = new google.maps.places.SearchBox($input[0]);
    // Link it to the UI element.
    map.controls[google.maps.ControlPosition.LEFT_TOP].push($input[0]);

    // Bias the SearchBox results towards current map's viewport.
    map.addListener('bounds_changed', function() {
      searchBox.setBounds(map.getBounds());
    });

    var markers = [];
    // Listen for the event fired when the user selects a prediction and retrieve
    // more details for that place.
    searchBox.addListener('places_changed', function() {
      var places = searchBox.getPlaces();

      if (places.length == 0) {
        return;
      }

      // For each place, get the icon, name and location.
      var bounds = new google.maps.LatLngBounds();
      places.forEach(function(place) {

        if (place.geometry.viewport) {
          // Only geocodes have viewport.
          bounds.union(place.geometry.viewport);
        } else {
          bounds.extend(place.geometry.location);
        }
      });
      map.fitBounds(bounds);
    });
  }

  //
  // Add a class to the map to change the opacity
  // of custom markers and clusters
  //
  function setMarkersOpacity(opacity){
    // TODO don't ignore the opacity
    if(opacity == 1){
      $container.removeClass('opacity-markers');
    } else {
      $container.addClass('opacity-markers');
    }
  }

  // Replace marker icon when select it in table
  function selectMarker(id){

      if (oldMarker != undefined){
        for(var i = 0; i < markers.length; i++){
          if (markers[i].id == oldMarker.id){
            markers[i].setIcon(oldMarker.icon);
            center();
            i = markers.length;
          }
        }
        if(oldMarker.id == id) {
          oldMarker = undefined;
          return;
        }
      }

      for (var i = 0; i < markers.length; i++){
        if (markers[i].id == id){
          oldMarker = $.extend(true, {}, markers[i]);
          var icon = markers[i].icon;
          icon.strokeWeight = 2.5;
          icon.strokeColor = "yellow";

          markers[i].setIcon(icon);
          map.setCenter(markers[i].getPosition());
          map.setZoom(20);
          i = markers.length;
        }
      }

  }

  function fitMapOnGpsMarkers() {
    if (markers.length > 0 || redBorder.gps_markers_json === undefined ||
        redBorder.gps_markers_json.length < 3) {
      return;
    }
    var gpsMarkers = redBorder.gps_markers_json,
        gpsMarkersBounds = new google.maps.LatLngBounds(),
        gpsLatLng;
    for (var i = 0; i < gpsMarkers.length; i++) {
      gpsLatLng = new google.maps
        .LatLng(gpsMarkers[i].latitude, gpsMarkers[i].longitude, false);
      gpsMarkersBounds.extend(gpsLatLng);
    }
    map.fitBounds(gpsMarkersBounds);
  }

  function highlightMarker(id){
    if($('a.center_map_on_marker.btn-success').length > 0) {
      var id_selected = $('a.center_map_on_marker.btn-success').parents('tr').data('access-point-id');
    }

    if (id === id_selected) { return; }

    if (oldHighlight !== undefined){
      storeMarker(id_selected);
    }

    if (id === null) { return; }

    restoreMarker(id, id_selected);
  }

  function storeMarker(id_selected) {
    for(var i = 0; i < markers.length; i++){
      if ((markers[i].id === oldHighlight.id) && (markers[i].id !== id_selected)){
        markers[i].setIcon(oldHighlight.icon);
        i = markers.length;
      }
    }
  }

  function restoreMarker(id, id_selected) {
    for (var i = 0; i < markers.length; i++){
      if ((markers[i].id === id) && (markers[i].id !== id_selected)){
        oldHighlight = $.extend(true, {}, markers[i]);
        var icon = markers[i].icon;
        icon.strokeWeight = 2.5;
        icon.strokeColor = "yellow";

        markers[i].setIcon(icon);
        i = markers.length;
      }
    }
  }

  //
  // Adds a new marker to the map at the position specified
  // marker_opts = {
  //   content: The tooltip shown when clicking on a marker
  //   draggable: Specifies if the marker should be draggable or not
  //   onMove: Event fired when the marker is moved
  // }
  //

  function addMarker(lat, lng, markerOpts, nocenter) {
    var marker, info, latLng;

    // If the map is not initialized, we save the marker to add it
    // later on, when the map is initialized
    if (!wasInitialize()) {
      markersToAdd = markersToAdd.concat([{
        lat: lat,
        lng: lng,
        opts: markerOpts
      }]);

      return;
    }

    markerOpts = markerOpts || {};
    markerOpts.draggable = markerOpts.draggable || false;
    markerOpts.filterWhenClick = markerOpts.filterWhenClick || false;
    markerOpts.content = markerOpts.content || false;
    markerOpts.infoBox = markerOpts.infoBox || false;
    markerOpts.infoBoxCloseURL = markerOpts.infoBoxCloseURL || '/assets/icons/map-close.png';
    markerOpts.pixelOffset = markerOpts.pixelOffset || false;
    markerOpts.onMove = markerOpts.onMove || false;

    latLng = new google.maps.LatLng(lat, lng, false);

    marker = new google.maps.Marker({
      map: map,
      draggable: markerOpts.draggable,
      raiseOnDrag: true,
      icon: markerOpts.icon,
      position: latLng
    });

    if (markerOpts.markerId != undefined){
      marker.set("id", markerOpts.markerId);
    }

    if (opts.clustered) {
      markerClusterer.addMarker(marker, true);
    }

    markerSpiderfier.addMarker(marker);
    bounds.extend(latLng);
    if (nocenter === undefined || nocenter === false){
      center();
    }

    if (markerOpts.onMove !== false) {
      google.maps.event.addListener(marker, 'dragend', function(mouseEvent) {
        markerOpts.onMove(marker, mouseEvent);
      });
    }

    if (markerOpts.infoBox !== false && markerOpts.pixelOffset !== false){

        info = new InfoBox({
            content: markerOpts.infoBox,
            map: map,
            closeBoxMargin: "0",
            pixelOffset: markerOpts.pixelOffset,
            closeBoxURL: markerOpts.infoBoxCloseURL
        });

        infos.push(info);

        google.maps.event.addListener(marker, 'click', function (e) {
            info.open(map, marker);
            map.panTo(latLng);
        });
    }

    if (markerOpts.content !== false && (markerOpts.infoBox === false && markerOpts.pixelOffset === false)) {
      info = new google.maps.InfoWindow({
        content: markerOpts.content
      });

      infos.push(info);

      google.maps.event.addListener(marker, 'click', function (e) {
        if (markerOpts.filterWhenClick === true){
          $("#search").val(markerOpts.content.split(I18n.t('javascript.mac_address') + "</b> ")[1].split("<br />")[0]);
          apply_ap_filter($("#search"));
          $("#search").val('');
          $('.center_map_on_marker').show();
        }
        infos.forEach(function (i) { i.close(); });
        info.open(map, this);
      });

      if ($("#search").length > 0 && $("#access_points table").length > 0) {
        google.maps.event.addListener(info, 'closeclick', function (e) {
          $("#search").val('');
          apply_ap_filter($("#search"));
        });
      }
    }

    if (markerOpts.onClick !== undefined) {
      var fun = markerOpts.onClick;
      google.maps.event.addListener(marker, 'click', function(mouseEvent) {
        fun(marker, map, mouseEvent);
      });
    }

    if (opts.draggable === true ){
      if (markerOpts.icon !== undefined){
        google.maps.event.addListener(marker,'dragstart',function(event) {
          marker.icon.fillOpacity = 0.3;
          marker.icon.strokeWeight = 0.3;
          marker.setIcon(marker.icon);
        });
      }
      google.maps.event.addListener(marker,'dragend',function(event) {
        if (markerOpts.icon !== undefined) marker.setIcon(markerOpts.icon);
        if (markerOpts.dragEndCallback !== undefined) markerOpts.dragEndCallback(marker, event);
      });
    }
    // Add current marker to markers container
    markers.push(marker);
  }

  //
  // Adds a list of markers to the map.
  // list: Array of objects with the data
  // markerOpts = {
  //   lat: The objects' property where the latitude is located
  //   lng: The objects' property where the longitude is located
  // }
  //
  function addMarkers(list, markersOpts, nocenter) {
    if (list === undefined || list.length === 0) return;
    if (markersOpts === undefined) markersOpts = {};

    var lngPty, latPty, contentString;
    latPty = markersOpts.lat || 'lat';
    lngPty = markersOpts.lng || 'lng';

    list.forEach(function (d) {
      contentString = d.description;
      if (d.description === undefined) {
        contentString = ' \
          <div class="map-icon map-sensor"><img src="'+ d.icon +'"></div> \
          <div class="map-marker-info"> \
            <p>' + d.title + '</p> \
            <p class="map-marker-ip">' + d.ip + '</p> \
          </div> \
          <div class="map-options"> \
            <a class="block-ui-but" href="/sensors/' + d.id + '/edit"> \
              <i class="icon fa fa-wrench"></i> \
            </a> \
          </div>';
      }
      addMarker(d[latPty], d[lngPty], {
        content: contentString
      });
    });
  }

  function addCustomMarker(lat, lng, markerOpts, nocenter) {
    var marker, info, latLng;

    // If the map is not initialized, we save the marker to add it
    // later on, when the map is initialized
    if (!wasInitialize()) {
      markersToAdd = markersToAdd.concat([{
        lat: lat,
        lng: lng,
        opts: markerOpts
      }]);

      return;
    }

    markerOpts = markerOpts || {};
    markerOpts.draggable = markerOpts.draggable || false;
    markerOpts.content = markerOpts.content || false;
    markerOpts.onMove = markerOpts.onMove || false;
    markerOpts.text = (markerOpts.text === undefined) ? '' : markerOpts.text;
    markerOpts.backgroundColor = markerOpts.backgroundColor || 'blue';
    markerOpts.fontColor = markerOpts.fontColor || 'white';

    latLng = new google.maps.LatLng(lat, lng, false);

    customMarker = new CustomMarker(latLng, map, markerOpts);

    if (opts.clustered) {
      markerClusterer.addMarker(customMarker, true);
    }

    customMarkers.push(customMarker);
    bounds.extend(latLng);
    if (nocenter === undefined || nocenter === false){
      center();
    }

  }

  function addCustomMarkers(list, markersOpts) {
    if (list === undefined || list.length === 0) return;
    if (markersOpts === undefined) markersOpts = {};

    var lngPty, latPty;
    latPty = markersOpts.lat || 'lat';
    lngPty = markersOpts.lng || 'lng';

    list.forEach(function (d) {
      addCustomMarker(d[latPty], d[lngPty], {
        content: d.description
      });
    });
  }

  /*
   * If the cluster option is active, repaint clusters
   */
  function repaintCluster(){
    if(opts.clustered && markerClusterer !== undefined){
      markerClusterer.repaint();
    }
  }

  function getMarkers() {
    return markers;
  }

  function deleteMarker(id){
    var i = markers.map(function(d){return d.id}).indexOf(id),
        marker;
    if (markers[i] != undefined){
      markers[i].setMap(null);
      marker = markers.splice(i, 1)[0];
      if (markerClusterer !== undefined){
        markerClusterer.removeMarker(marker);
      }
    }
  }

  function removeMarkers() {
    for (var i = 0; i < markers.length; i++){
      markers[i].setMap(null);
    }
    markers = [];
    if (markerClusterer !== undefined){
      markerClusterer.clearMarkers();
    }
    resetBounds();
  }

  function markersByIds(ids) {
    if (markerClusterer !== undefined) markerClusterer.clearMarkers();
    for (var i = 0; i < markers.length; i++){
      if (jQuery.inArray(markers[i].id, ids) === -1) {
        markers[i].setMap(null)
      } else {
        markers[i].setMap(redBorder.access_points_map.getMap())
        if (markerClusterer !== undefined){
          markerClusterer.addMarker(markers[i]);
        }
      }
    }
  }

  function showMarkers() {
    if (markerClusterer !== undefined) markerClusterer.clearMarkers();
    for (var i = 0; i < markers.length; i++) {
      markers[i].setMap(redBorder.access_points_map.getMap())
      if (markerClusterer !== undefined){
        markerClusterer.addMarker(markers[i]);
      }
    }
  }

  function getCustomMarkers() {
    return customMarkers;
  }

  function removeCustomMarkers() {
    for (var i = 0; i < customMarkers.length; i++){
      customMarkers[i].setMap(null);
    }
    customMarkers = [];
    if (markerClusterer !== undefined){
      markerClusterer.clearMarkers();
    }
    resetBounds();
  }

  function fullScreen(){
    var self = $container.parent();
    if (self.hasClass('full_screen_map')){
      map.setOptions({scrollwheel: false});
    }
    else{
      map.setOptions({scrollwheel: true});
    }

    self.toggleClass('full_screen_map');
    $('.expand-map').toggleClass("fa-expand");
    $('.expand-map').toggleClass("fa-compress");
    $container.toggleClass('full_screen_map_child');
    google.maps.event.trigger(map,'resize');

  }

  function addPolyline(positions, opts, identifier) {
    // Force the map
    opts['map'] = map;

    // Get the color
    if(opts['strokeColor'] === undefined){
      opts['strokeColor'] = getPolylineColor();
    }

    var polyline = new google.maps.Polyline(opts);
    var markers = [];
    var bounds = new google.maps.LatLngBounds();

    // Set the markers in the line
    var z = 0;
    var path = [];
    path[z] = polyline.getPath();

    for (var i = 0; i < positions.length; i++) //LOOP TO DISPLAY THE MARKERS
    {
      var pos = positions[i];

      // We only show the first and last marker
      if (i === 0 || i === (positions.length - 1)){
        var marker = new google.maps.Marker({
            position: pos,
            map: map
        });
        markers.push(marker);
      }

      bounds.extend(pos);
      path[z].push(pos); //PUSH THE NEWLY CREATED MARKER'S POSITION TO THE PATH ARRAY
    }

    // Store the polyline with an identifier
    polylines.push({
      id: identifier,
      line: polyline,
      markers: markers
    })

    // Center the map in the polyline
    map.fitBounds(bounds);

    // Return the color used
    return opts['strokeColor'];
  }

  function getPolylines(){
    return polylines;
  }

  function removePolyline(identifier) {
    var index = -1;

    for (var i = 0; i < polylines.length; i++){
      if (polylines[i].id === identifier){
        polylines[i].line.setMap(null);
        index = i;
        // Delete all markers
        while(polylines[i].markers.length > 0){
          polylines[i].markers.pop().setMap(null);
        }
      }
    }

    if (index >= 0) {
      polylines.splice(index, 1);
    }
  }

  function getPolylineColor(){
    var color;

    if (palette.length >= polylines.length) {
      color = palette[polylines.length - 1];
    } else {
      color = getRandomColor();
    }

    return color;
  }

  function getRandomColor() {
    var letters = '0123456789ABCDEF'.split('');
    var color = '#';

    for (var i = 0; i < 6; i++ ) {
        color += letters[Math.floor(Math.random() * 16)];
    }

    return color;
  }

  function addPolygon(vertexes, opts) {
    var polygon, latLngs = [];

    // If the map is not initialized, we save the marker to add it
    // later on, when the map is initialized
    if (!wasInitialize()) {
      polygonsToAdd = polygonsToAdd.concat({
        vertexes: vertexes,
        opts: opts
      });

      return;
    }

    vertexes.forEach(function (v) {
      latLngs.push(new google.maps.LatLng(v.lat, v.lng));
    });

    polygon = new google.maps.Polygon({
      paths: latLngs,
      clickable: opts.clickable,
      fillColor: opts.fillColor,
      fillOpacity: opts.fillOpacity,
      strokeWeight: opts.strokeWeight,
      strokeOpacity: opts.strokeOpacity
    });

    // Create a listener on click if there is a content opt
    if (opts.content !== undefined) {
      var polygon_bounds = new google.maps.LatLngBounds(),
          infoWindowPosition;

      latLngs.forEach(function(xy,i) {
        bounds.extend(xy);
        polygon_bounds.extend(xy);
      });

      infoWindowPosition = polygon_bounds.getCenter();

      var info = new google.maps.InfoWindow({
        content: opts.content,
        position: infoWindowPosition
      });

      infos.push(info);

      google.maps.event.addListener(polygon, 'rightclick', function (e) {
        infos.forEach(function (i) { i.close(); });
        info.open(map);
      });
    }

    if (opts.onClick !== undefined) {
      var fun = opts.onClick;
      google.maps.event.addListener(polygon, 'click', function(mouseEvent) {
        fun(polygon, map, mouseEvent);
      });
    }

    if (opts.onDblClick !== undefined) {
      var func = opts.onDblClick;
      google.maps.event.addListener(polygon, 'dblclick', function(mouseEvent) {
        func(polygon, map, mouseEvent);
      });
    }

    polygon.setMap(map);
    polygons.push(polygon);

    return polygon;
  }

  function getPolygons() {
    return polygons;
  }

  function removePolygons(){
    polygons.forEach( function(d) {
      d.setMap(null);
    });

    polygons = [];
  }

  /*
   * Reset bounds in map
   */
  function resetBounds() {
    var lmarkers = markers.length,
        lcmarkers = customMarkers.length;

    // Clear bounds
    bounds = new google.maps.LatLngBounds();

    // Add bounds from markers
    for(var i = 0; i < lmarkers; i++) {
      bounds.extend(markers[i].getPosition());
    }

    // Add bounds from CustomMarkers
    for(var i = 0; i < lcmarkers; i++) {
      bounds.extend(CustomMarkers[i].getPosition());
    }
  }

  //
  // Creates a listener on the idle event
  // Executes the passed fun passing it the current google map
  // instance and the mouseEvent that was fired
  //
  function onLoad(fun) {
    google.maps.event.addListenerOnce(map, 'idle', function(mouseEvent) {
      fun(map, mouseEvent);
    });
  }

  //
  // Creates a listener on the click event
  // Executes the passed fun passing it the current google map
  // instance and the mouseEvent that was fired
  //
  function onClick(fun) {
    google.maps.event.addListener(map, 'click', function(mouseEvent) {
      fun(map, mouseEvent);
    });
  }

  //
  // Creates a listener on the dragend event
  // Executes the passed fun passing it the current google map
  // instance and the mouseEvent that was fired
  //
  function onDrag(fun) {
    google.maps.event.addListener(map, 'dragend', function(mouseEvent) {
      fun(map, mouseEvent);
    });
  }

  //
  // Reloads the map using a resize
  //
  function reload() {
    google.maps.event.trigger(map, 'resize');
    /*if (markerClusterer !== undefined) {
      markerClusterer.repaint();
    }*/
    $container.css('opacity', '1');
  }

  //
  // Centers the map automatically based on the markers
  //
  function center() {
    if (!wasInitialize() || bounds.isEmpty()) return;
    map.fitBounds(bounds);
  }

  //
  // Extends the bounds adding a LatLng point
  //
  function extendBounds(lat, lng) {
    bounds.extend(new google.maps.LatLng(lat, lng));
  }

  //
  // Returns the raw google map instance
  //
  function getMap() {
    return map;
  }

  function wasInitialize() {
    return (map !== undefined);
  }

  function setOptions(opts) {
    map.setOptions(opts);
  }

  return {
    initialize: initialize,
    wasInitialize: wasInitialize,
    addMarker: addMarker,
    addMarkers: addMarkers,
    removeMarkers: removeMarkers,
    markersByIds: markersByIds,
    showMarkers: showMarkers,
    addCustomMarker: addCustomMarker,
    addCustomMarkers: addCustomMarkers,
    repaintCluster: repaintCluster,
    removeCustomMarkers: removeCustomMarkers,
    getMarkers: getMarkers,
    fitMapOnGpsMarkers: fitMapOnGpsMarkers,
    selectMarker: selectMarker,
    highlightMarker: highlightMarker,
    getCustomMarkers: getCustomMarkers,
    addPolyline: addPolyline,
    getPolylines: getPolylines,
    removePolyline: removePolyline,
    addPolygon: addPolygon,
    getPolygons: getPolygons,
    removePolygons: removePolygons,
    setMarkersOpacity: setMarkersOpacity,
    reload: reload,
    resize: reload,
    deleteMarker: deleteMarker,
    center: center,
    extendBounds: extendBounds,
    getMap: getMap,
    setOptions: setOptions,
    onLoad: onLoad,
    onClick: onClick,
    onDrag: onDrag
  }
};

function CustomMarker(latlng, map, markerOpts) {
  this.latlng_ = latlng;
  this.markerOpts_ = markerOpts;
  this.infoWindow;

  // Once the LatLng and text are set, add the overlay to the map.  This will
  // trigger a call to panes_changed which should in turn call draw.
  this.setMap(map);
}

CustomMarker.prototype = new google.maps.OverlayView();

CustomMarker.prototype.draw = function() {
  var me = this;

  // Check if the div has been created.
  var div = this.div_;
  if (!div) {
    // Create a overlay text DIV
    div = this.div_ = document.createElement('DIV');
    // Create the DIV representing our CustomMarker
    div.style.border = "3px solid white";
    div.style.position = "absolute";
    div.style.paddingLeft = "0px";
    div.style.cursor = 'pointer';
    div.innerHTML = this.markerOpts_.text;
    div.style.padding = "7px";
    div.style.borderRadius = "50%";
    div.style.backgroundColor = this.markerOpts_.backgroundColor;
    div.style.fontWeight = "bold";
    div.style.color = this.markerOpts_.fontColor;
    div.style.width = "32px";
    div.style.height = "32px";
    div.style.textAlign = "center";
    div.classList.add('rb-maps-custom-marker');

    if(this.markerOpts_.content != false) {
      google.maps.event.addListener(me, "click", function() {
        if(me.infoWindow === undefined || me.infoWindow === null || me.infoWindow.map === undefined || me.infoWindow.map === null){
          me.infoWindow = new google.maps.InfoWindow(
            {content: me.markerOpts_.content,
             pixelOffset: new google.maps.Size(5,0)}
          );
          me.infoWindow.open(this.getMap(), me);
        } else {
          me.infoWindow.close();
          me.infoWindow = undefined;
        }
      });
    }

    google.maps.event.addDomListener(div, "click", function(event) {
      google.maps.event.trigger(me, "click");
      event.preventDefault();
    });

    // Then add the overlay to the DOM
    var panes = this.getPanes();
    panes.overlayImage.appendChild(div);
  }

  // Position the overlay
  var point = this.getProjection().fromLatLngToDivPixel(this.latlng_);
  if (point) {
    div.style.left = (point.x - 21)  + 'px';
    div.style.top = (point.y - 20) + 'px';
  }
};

CustomMarker.prototype.remove = function() {
  // Check if the overlay was on the map and needs to be removed.
  if (this.div_) {
    this.div_.parentNode.removeChild(this.div_);
    this.div_ = null;
  }
};

CustomMarker.prototype.getDraggable = function() {
  // TODO
  return false;
};

CustomMarker.prototype.setVisible = function(visible) {
  var div = this.div_;

  if(visible == true){
    div.style.display = 'block';
  } else {
    div.style.display = 'hidden';
  }

};

CustomMarker.prototype.getPosition = function() {
 return this.latlng_;
};

// Function that is used when clustering customMarkers that represent
// APs and the value inside are the number of clients.
function sum_clients(markers) {
  var total = 0;
  var index = 0;

  for (var i = 0; i < markers.length; i++) {
    value = parseInt(markers[i].markerOpts_.text);

    if(!isNaN(value)){
      total = total + value;
    }

  }

  var dv = total;
  while (dv !== 0) {
    dv = parseInt(dv / 10, 10);
    index++;
  }

  return {
    text: total,
    index: index
  };
}
