/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

// The general object for widgets related stuff
redBorder.widgets = {};

// Saves graphs types for widgets
redBorder.widgets.graphs = {};

// Saves widgets types for widgets
redBorder.widgets.types = {};

// The currently loaded widgets
redBorder.widgets.active = {};
redBorder.widgets.active_index = 0;
redBorder.widgets.active_ids   = [];
redBorder.widgets.initialized  = false;

// Util methods to use on widgets
redBorder.widgets.util = {};

// This function returns the unit of a given widget. Modules can
// write their own unitsHandler function in order to override
// the custom behaviour, which is the default function
redBorder.widgets.unitsHandlers = {}
redBorder.widgets.unitsHandlers.default = function (widget) {
  var product = widget.property.product_type,
      aggregation = widget.property.aggregation;

  if (widget.property.widget_type === 'unique' ||
      widget.property.widget_type === 'single_unique' ||
      widget.property.widget_type === 'grouped_unique' ||
      widget.property.widget_type === 'combination') return '';
  if(product == 'none') return ''
  return redBorder.modules[product].aggregations[aggregation].unit;
}

redBorder.widgets.unitsHandler = function (widget) {
  var product = widget.property.product_type;

  if (redBorder.widgets.unitsHandlers[product] !== undefined) {
    return redBorder.widgets.unitsHandlers[product](widget);
  } else {
    return redBorder.widgets.unitsHandlers.default(widget);
  }
}

redBorder.widgets.build = function (opts, data) {
  var graph,
      wtype,
      widget_type,
      container;

  // Update redBorder time
  redBorder.timeNow = redBorder.timeMachine || new Date();

  // The container where the widget will be rendered
  container = 'widget_placeholder_' + opts.id;

  // Set the graph type
  if (opts.property.graph_type !== 'none') {
    graph = redBorder.widgets.graphs[opts.property.graph_type](container, opts);
  }

  // Set the widget type
  if (opts.property.widget_type !== 'none') {
    widget_type = opts.property.widget_type;

    if (widget_type === 'single_unique' ||
        widget_type === 'grouped_unique') {
      widget_type = 'unique';
    }

    wtypeModule = redBorder.widgets.types[widget_type]

    if (wtypeModule !== undefined) {
      wtype = wtypeModule(opts, data);
    }
  }

  // We init the graph and pass it the widget type, so the graph
  // can query data when needed
  graph.init(wtype);

  // Save a reference to access the widget later
  redBorder.widgets.active['widget_' + opts.id] = graph;
  if (data != undefined || redBorder.widgets.initialized) {
    graph.render();
  }
  // Store the IDs to render after
  redBorder.widgets.active_ids.push(opts.id);
}

// Trigger the reload method of a widget to delete the
// reloading class
//
redBorder.widgets.util.reloadDone = function($widget) {
  $widget.removeClass('reloading');
  // Continue loading widgets.
  redBorder.widgets.util.loadNextWidget();
}

// Trigger the reload fail method of a widget to delete the
// reloading class and add the fail one
//
redBorder.widgets.util.reloadFail = function($widget) {
  $widget.removeClass('reloading');
  $widget.addClass('fail');
  // Continue loading widgets.
  redBorder.widgets.util.loadNextWidget();
}

// Render next widget
redBorder.widgets.util.loadNextWidget = function() {
  next_widget = redBorder.widgets.getActiveWidgetAtIndex(++redBorder.widgets.active_index)
  if ( redBorder.widgets.active_index < Object.keys(redBorder.widgets.active).length &&
       next_widget !== undefined ) {
    // There are more widgets to render
    next_widget.render();
  } else {
    // Dashboard its initialized
    redBorder.widgets.initialized = true;
  }
}

redBorder.widgets.getActiveWidgetAtIndex = function(index){
  return redBorder.widgets.active['widget_' + redBorder.widgets.active_ids[index] ];
}

// This function outputs an object that gives you the multiplier
// and prefix (K, M, G, T) that you need to use in order to
// use humanizable nums
redBorder.widgets.util.getMultiplier = function (num) {
  var multiplier = {};

  if (num < 1e3) {
    multiplier.num = 1;
    multiplier.prefix = '';
  } else if (num >= 1e3 && num < 1e6) {
    multiplier.num = 1e-3;
    multiplier.prefix = 'K';
  } else if (num >= 1e6 && num < 1e9) {
    multiplier.num = 1e-6;
    multiplier.prefix = 'M';
  } else if (num >= 1e9 && num < 1e12) {
    multiplier.num = 1e-9;
    multiplier.prefix = 'G';
  } else if (num >= 1e12) {
    multiplier.num = 1e-12;
    multiplier.prefix = 'T';
  }

  return multiplier;
}

// This function is called to process the data of tops queries
// and is used to return a understandable input to graphs types
redBorder.widgets.util.parseTopData = function (data, aggregation, column, topNum, column_name) {
  var datos = data.map(function (d, i) {
    if (topNum !== undefined && topNum <= i) return;
    var events = d[aggregation];
    if (column_name){
      return {
        name: d[column_name],
        events: events
      };
    }else if(d['secondary_name'] === undefined || column === 'client_rssi' ||
       (column === 'wireless_station' && d['secondary_name'] === 'N/A') ||
       column === 'client_mac'){
      return {
        name: d[column],
        events: events
      };
    }else {
      return {
        name: d['secondary_name'],
        events: events,
        value: d[column]
      };
    }
  });

  datos = datos.filter(function (d, i) {
    if (topNum === undefined || i < topNum) return true;
    else return false;
  });

  return datos;
}

redBorder.widgets.util.checkPerSecondAggregationTops = function(data, aggregation, seconds) {
  if (aggregation === 'bps' || aggregation === 'bits_per_sec_per_client' || aggregation === 'bits_per_sec_per_station'){
    for(var i = 0; i < data.length; i++){
      data[i][aggregation] = (data[i][aggregation]*8)/seconds;
    }
  } else if(aggregation === 'fps' || aggregation === 'pps') {
    for(var i = 0; i < data.length; i++){
      data[i][aggregation] = (data[i][aggregation])/seconds;
    }
  }

  return data;
}

redBorder.widgets.util.checkPerSecondAggregationTotal = function(data, aggregation, seconds) {
  // if (aggregation === 'bps' || aggregation === 'bits_per_sec_per_client' || aggregation === 'bits_per_sec_per_station'){
  //   data[0].event.events = (data[0].event.events*8)/seconds;
  // } else if(aggregation === 'fps' || aggregation === 'pps') {
  //   data[0].event.events = (data[0].event.events)/seconds;
  // }

  return data;
}

// Does a query to the website. It provides standard params and
// merges graph_opts with the default params.
// This method always gives priority to low levels.
// Graph opts > Widget type opts > Default opts
redBorder.widgets.util.query = function (params) {
  var widget = params.widget,
      url = params.url,
      typeOpts = params.typeOpts,
      graphOpts = params.graphOpts || {},
      diffSecs = widget.wtype.end_time_secs - widget.wtype.start_time_secs,
      ignoreFilters = params.ignoreFilters || false,
      data, time_filters = {};

  var start_time = graphOpts.start_time || parseInt((redBorder.timeNow.getTime() / 1000) - diffSecs);
  var end_time = graphOpts.end_time || parseInt(redBorder.timeNow.getTime() / 1000)

  if (widget.property.range == 'custom') {
    $.each(widget.property.filters, function (index, value) {
      if (value['column'] == 'start_time' || value['column'] == 'end_time')
        time_filters[value['column']] = value['value']
    })
    if (Object.keys(time_filters).length == 2) {
      start_time = time_filters['start_time']
      end_time = time_filters['end_time']
    }
  }

  data = {
    start_time: start_time,
    end_time: end_time,
    w: 'true' // Set call from a widget
  };

  // By default we will add filters to the queries
  if (!ignoreFilters) {
    data = $.extend(data, { filters: JSON.stringify(widget.property.filters) });
  }

  if (widget.graph.step !== undefined && widget.graph.step <= 60 * 5 * 1000) {
    data.end_time = parseInt(data.end_time - widget.graph.step / 1000);
  }

  typeOpts = typeOpts || {};
  graphOpts = graphOpts || {};
  data = $.extend(data, typeOpts);
  data = $.extend(data, graphOpts);

  return $.ajax({
    url: url,
    type: 'GET',
    dataType: 'json',
    data: data
  });
}

// Returns the given range as seconds from the day before
redBorder.widgets.util.lastDay = function (range, nowSecs) {
  var rangeLastDayMillis = [],
      rangeLastDay = [],
      nowMillis;

  if (nowSecs !== undefined) {
    nowMillis = nowSecs * 1000;
  } else {
    nowMillis = redBorder.timeNow.getTime();
  }

  // Get the total data based on the range
  switch (range) {
    case "last_1_minute":
      rangeLastDayMillis[0] = nowMillis - (1).days() - (1).minutes();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case "last_2_minute":
      rangeLastDayMillis[0] = nowMillis - (1).days() - (2).minutes();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case "last_5_minute":
      rangeLastDayMillis[0] = nowMillis - (1).days() - (5).minutes();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_15_minute':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (15).minutes();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_30_minute':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (30).minutes();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_1_hour':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (1).hours();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_2_hour':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (2).hours();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_3_hour':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (3).hours();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_6_hour':
    rangeLastDayMillis[0] = nowMillis - (1).days() - (6).hours();
    rangeLastDayMillis[1] = nowMillis - (1).days();
    break;
    case 'last_12_hour':
      rangeLastDayMillis[0] = nowMillis - (1).days() - (12).hours();
      rangeLastDayMillis[1] = nowMillis - (1).days();
      break;
    case 'last_24':
      rangeLastDayMillis[0] = nowMillis - (1).weeks() - (1).hours();
      rangeLastDayMillis[1] = nowMillis - (1).weeks();
      break;
    case "last_week":
      rangeLastDayMillis[0] = nowMillis - (28).days() - (1).week();
      rangeLastDayMillis[1] = nowMillis - (28).days()
      break;
  }

  // Convert to seconds
  rangeLastDay[0] = parseInt(rangeLastDayMillis[0] / 1000);
  rangeLastDay[1] = parseInt(rangeLastDayMillis[1] / 1000);

  return rangeLastDay;
}

// Returns the given range as seconds from the last period
redBorder.widgets.util.lastRange = function (range, nowSecs) {
  var lastRangeMillis = [],
      lastRange = [],
      nowMillis;

  if (nowSecs !== undefined) {
    nowMillis = nowSecs * 1000;
  } else {
    nowMillis = redBorder.timeNow.getTime();
  }

  // Get the total data based on the range
  switch (range) {
    case "last_1_minute":
      lastRangeMillis[0] = nowMillis - (2).minutes();
      lastRangeMillis[1] = nowMillis - (1).minutes();
      break;
    case "last_2_minute":
      lastRangeMillis[0] = nowMillis - (4).minutes();
      lastRangeMillis[1] = nowMillis - (2).minutes();
      break;
    case "last_5_minute":
      lastRangeMillis[0] = nowMillis - (10).minutes();
      lastRangeMillis[1] = nowMillis - (5).minutes();
      break;
    case "last_15_minute":
      lastRangeMillis[0] = nowMillis - (30).minutes()
      lastRangeMillis[1] = nowMillis - (15).minutes()
      break;
    case "last_30_minute":
      lastRangeMillis[0] = nowMillis - (1).hour()
      lastRangeMillis[1] = nowMillis - (30).minutes()
      break;
    case "last_1_hour":
      lastRangeMillis[0] = nowMillis - (2).hours()
      lastRangeMillis[1] = nowMillis - (1).hour()
      break;
    case "last_2_hour":
      lastRangeMillis[0] = nowMillis - (4).hours()
      lastRangeMillis[1] = nowMillis - (2).hour()
      break;
    case "last_3_hour":
      lastRangeMillis[0] = nowMillis - (6).hours()
      lastRangeMillis[1] = nowMillis - (3).hour()
      break;
    case "last_6_hour":
      lastRangeMillis[0] = nowMillis - (12).hours()
      lastRangeMillis[1] = nowMillis - (6).hour()
      break;
    case "last_12_hour":
      lastRangeMillis[0] = nowMillis - (1).day()
      lastRangeMillis[1] = nowMillis - (12).hours()
      break;
    case "last_24":
      lastRangeMillis[0] = nowMillis - (2).days()
      lastRangeMillis[1] = nowMillis - (1).day()
      break;
    case "last_week":
      lastRangeMillis[0] = nowMillis - (2).weeks()
      lastRangeMillis[1] = nowMillis - (1).weeks()
      break;
  }

  // Convert to seconds
  lastRange[0] = parseInt(lastRangeMillis[0] / 1000);
  lastRange[1] = parseInt(lastRangeMillis[1] / 1000);

  return lastRange;
}
