/* 
 * movementsMap-gmaps.js GMaps overlay
 *
 * Copyright (c) 2024, Miguel Negrón
 * based on Patrick Wied (http://www.patrick-wied.at) heatmap.js
 * Dual-licensed under the MIT (http://www.opensource.org/licenses/mit-license.php)
 * and the Beerware (http://en.wikipedia.org/wiki/Beerware) license.
 */ 

function MovementsMapOverlay(map, cfg){
    var me = this;

    me.heatmap = null;
    me.conf = cfg;
    me.latlngs = [];
    me.bounds = null;
    me.routes = [];
    me.drawn_polylines = [];
    me.setMap(map);
    me.dataToAddLater = undefined;
  
  //google.maps.event.addListener(map, 'idle', function() { me.draw() });
}


function formatUpdate(startIndex, endIndex, endTime, startTime) {
  if (startIndex === endIndex) {
      return 'Update #' + startIndex + ': ' + endTime + '<br/>';
  } else {
      var startText = startTime ? ' from ' + startTime : '';
      return 'Update #' + startIndex + '-' + endIndex + startText + ' to ' + endTime + '<br/>';
  }
}

MovementsMapOverlay.prototype = new google.maps.OverlayView();

// Función para añadir una ruta al overlay.
// route debe ser un array de objetos google.maps.LatLng o coordenadas literales.
MovementsMapOverlay.prototype.addRoute = function(route) {
  this.routes.push(route);
  this.draw(); // Redibujar después de añadir una nueva ruta.
};

MovementsMapOverlay.prototype.removeRoute = function(routeName) {
  var index = this.routes.findIndex(function(route) {
    return route.name === routeName;
  });

  if (index == -1) return ;
  
  if (this.routes[index].polyline != undefined)  this.routes[index].polyline.setMap(null);
  
  this.routes.splice(index, 1);
};

MovementsMapOverlay.prototype.unhideRoute = function(routeName) {
  var index = this.routes.findIndex(function(route) {
    return route.name === routeName;
  });

  if (index == -1) return ;
  
  if (this.routes[index].polyline != undefined)  this.routes[index].polyline.setVisible(true);

  if (this.routes[index].markers != undefined) {
    this.routes[index].markers.forEach(function(marker) {
      marker.setVisible(true);
    });
  } 

  // if (this.routes[index].infowindows != undefined) {
  //   this.routes[index].infowindows.forEach(function(infowindow) {
  //     infowindow.setVisible(true);
  //   });
  // } 

};

MovementsMapOverlay.prototype.hideRoute = function(routeName) {
  var index = this.routes.findIndex(function(route) {
    return route.name === routeName;
  });

  if (index == -1) return ;
  
  if (this.routes[index].polyline != undefined)  this.routes[index].polyline.setVisible(false);

  if (this.routes[index].markers != undefined) {
    this.routes[index].markers.forEach(function(marker) {
      marker.setVisible(false);
    });
  } 

  if (this.routes[index].infowindows != undefined) {
    this.routes[index].infowindows.forEach(function(infowindow) {
      infowindow.close();
    });
  } 
};


MovementsMapOverlay.prototype.onAdd = function(){
  var panes = this.getPanes(),
      w = this.getMap().getDiv().clientWidth,
      h = this.getMap().getDiv().clientHeight,
      el = document.createElement("div");
    
  el.style.cssText = "position:absolute;top:0;left:0;width:"+w+"px;height:"+h+"px;";
      
  this.conf.element = el;
  panes.overlayLayer.appendChild(el);

  this.heatmap = h337.create(this.conf);
}

MovementsMapOverlay.prototype.onRemove = function(){
    // Empty for now.
}

// Implementación del método draw.
MovementsMapOverlay.prototype.draw = function() {
  var self = this;
  var routes = this.routes;

  // Asegurarse de borrar rutas previas antes de dibujar.
  this.clearRoutes();
  
  // Define a symbol using a predefined path (an arrow)
  // supplied by the Google Maps JavaScript API.
  const lineSymbol = {
    path: google.maps.SymbolPath.FORWARD_CLOSED_ARROW,
  };

  const dashlineSymbol = {
    path: "M 0,-1 0,1",
    strokeOpacity: 1,
    scale: 4,
  };


  this.routes.forEach(function(route, index) {
    if (route.latlong == undefined) return;
    if (route.latlong.length == 0) return;
    if (route.polyline != undefined) return;

    var map = self.map;
    var markers = [];
    var infowindows = [];

    // Agrupar por latitud y longitud
    var groupedByLatLong = {};
    route.latlong.forEach(function(step, index) {
        var key = step.lat + ',' + step.lng;
        if (!groupedByLatLong[key]) {
            groupedByLatLong[key] = [];
        }
        groupedByLatLong[key].push({
            index: index + 1, // Índice basado en 1 para los updates
            timestamp: step.timestamp
        });
    });

    Object.keys(groupedByLatLong).forEach(function(key, groupIndex) {
      var steps = groupedByLatLong[key];
      var contentString = '<div id="content">' +
                          '<div id="siteNotice"></div>' +
                          '<div id="bodyContent">' +
                          '<b>Client MAC:</b> ' + route.name + '<br/>';
  
      var lastProcessedIndex = 0; // Para mantener el seguimiento del último índice procesado
      var rangeStart = null;
  
      steps.forEach(function(step, stepIndex) {
          var isLastStep = stepIndex === steps.length - 1;
          var isConsecutiveStep = stepIndex === 0 || step.index === steps[stepIndex - 1].index + 1;
          var previousStepTimestamp = stepIndex > 0 ? steps[stepIndex - 1].timestamp : undefined;
  
          if (isConsecutiveStep) {
              if (rangeStart === null) {
                  rangeStart = step.index; // Iniciar un nuevo rango
              }
              if (isLastStep) {
                  // Último paso del grupo, cerrar el rango o punto
                  contentString += formatUpdate(rangeStart, step.index, step.timestamp, previousStepTimestamp);
              }
          } else {
              // No es consecutivo, cerrar el rango anterior y empezar uno nuevo
              if (stepIndex > 1) {
                  var prevPrevStepTimestamp = steps[stepIndex - 2].timestamp;
                  contentString += formatUpdate(rangeStart, steps[stepIndex - 1].index, previousStepTimestamp, prevPrevStepTimestamp);
              }
              rangeStart = step.index; // Iniciar un nuevo rango con el paso actual
              if (isLastStep) {
                  // Si es el último paso, cerrarlo adecuadamente
                  contentString += formatUpdate(rangeStart, step.index, step.timestamp);
              }
          }
      });
  
      contentString += '</div></div>';
  
      var latLng = key.split(',').map(Number);
      var marker = new google.maps.Marker({
          position: { lat: latLng[0], lng: latLng[1] },
          map: map,
          title: route.name,
          icon: {
            path: google.maps.SymbolPath.CIRCLE,
            scale: 10,
            fillColor: route.color,
            fillOpacity: 1,
            strokeWeight: 2,
            strokeColor: "#FFFFFF"
          }
      });
  
      var infowindow = new google.maps.InfoWindow({
          content: contentString
      });
  
      marker.addListener('click', function() {
          if (!infowindow.getMap()) {
              infowindow.open(map, marker);
          } else {
              infowindow.close();
          }
      });
  
      infowindows.push(infowindow);
      markers.push(marker);
    });
  
  
    // Dibujar cada ruta.
    var polyline = new google.maps.Polyline({
        path: route.latlong,
        geodesic: true, // Para que la ruta siga la curvatura de la Tierra.
        strokeColor: route.color,
        strokeOpacity: 0,
        icons: [
          {
            icon: lineSymbol,
            offset: "100%",
          },
          {
            icon: dashlineSymbol,
            offset: "0",
            repeat: "20px",
          },
    
        ]      
    });
    polyline.setMap(self.map);

    routes.splice(index, 1);
    routes.push({name: route.name, polyline: polyline, markers: markers, infowindows: infowindows});

  });
  
};

// Función para limpiar (borrar) todas las rutas del mapa.
MovementsMapOverlay.prototype.clearRoutes = function() {
  // Implementar la lógica para remover las rutas del mapa.
  // Esto puede requerir almacenar referencias a los objetos Polyline al crearlos.
};


MovementsMapOverlay.prototype.pixelTransform = function(p){
  var w = this.heatmap.get("width"),
      h = this.heatmap.get("height");

  while(p.x < 0){
    p.x += w;
  }
  
  while(p.x > w){
    p.x -= w;
  }
      
  while(p.y < 0){
    p.y += h;
  }

  while(p.y > h){
    p.y -= h;
  }

  p.x = (p.x >> 0);
  p.y = (p.y >> 0);
  
  return p;
}

MovementsMapOverlay.prototype.setDataSet = function(data){
  var me = this,
      currentBounds = me.map.getBounds(),
      projection = me.getProjection(),
      mapdata, d, dlen, latlng, point, width, height;

  me.latlngs = [];

  if (data === undefined ||
      data.data === undefined ||
      data.data.length === 0) {
    return;
  }

  d = data.data;
  dlen = d.length;
  mapdata = {
    max: data.max,
    data: []
  };

  if (me.heatmap !== undefined && me.heatmap !== null) {
    width = this.heatmap.get('width');
    height = this.heatmap.get('height');

    if (width === 0 || height === 0) {
      this.dataToAddLater = data;

      google.maps.event.addListenerOnce(me.map, 'idle', function() {
        me.setDataSet(this.dataToAddLater);
      });

      return;
    }
  } else {
    this.dataToAddLater = data;

    google.maps.event.addListenerOnce(me.map, 'idle', function() {
      me.setDataSet(this.dataToAddLater);
    });

    return;
  }

  this.dataToAddLater = undefined;

  while (dlen--) {    
    latlng = new google.maps.LatLng(d[dlen].lat, d[dlen].lng);
    
    if (!currentBounds.contains(latlng)) { 
        continue; 
    }

    me.latlngs.push({latlng: latlng, c: d[dlen].count});
    point = me.pixelTransform(projection.fromLatLngToDivPixel(latlng));
    mapdata.data.push({x: point.x, y: point.y, count: d[dlen].count});
  }

  me.heatmap.clear();
  me.heatmap.store.setDataSet(mapdata);
  google.maps.event.addListenerOnce(me.map, 'idle', function() { me.draw() });
}

MovementsMapOverlay.prototype.addDataPoint = function(lat, lng, count){
  var projection = this.getProjection(),
      latlng = new google.maps.LatLng(lat, lng),
      point = this.pixelTransform(projection.fromLatLngToDivPixel(latlng));
  
  this.heatmap.store.addDataPoint(point.x, point.y, count);
  this.latlngs.push({ latlng: latlng, c: count });
}

MovementsMapOverlay.prototype.toggle = function(){
  this.heatmap.toggleDisplay();
}
