(function () {
  function isDarkModeEnabled() {
    return document.documentElement.classList.contains('dark-mode');
  }

  function applyTheme(theme) {
    theme = theme || 'light';
    if (theme === 'dark-mode') {
      document.documentElement.classList.add('dark-mode');
    } else {
      document.documentElement.classList.remove('dark-mode');
    }
    updateThemeUI();
  }

  function updateThemeUI() {
    var themeIcon = document.querySelector('.theme-icon');
    var themeSwitch = document.querySelector('.theme-switch');

    if (themeIcon) {
      themeIcon.classList.remove('fa-moon-o', 'fa-sun-o');
      if (isDarkModeEnabled()) {
        themeIcon.classList.add('fa-sun-o');
      } else {
        themeIcon.classList.add('fa-moon-o');
      }
    }

    if (themeSwitch) {
      themeSwitch.setAttribute(
        'aria-label',
        'Switch to ' + (isDarkModeEnabled() ? 'light' : 'dark') + ' mode'
      );
    }
  }

  function toggleTheme() {
    var newTheme = isDarkModeEnabled() ? 'light' : 'dark-mode';
    applyTheme(newTheme);
  
    saveThemeLocally(newTheme);

    if (document.body.dataset.userId) {
      saveThemeToServer(newTheme);
    }
  }  

  function saveThemeLocally(theme) {
    localStorage.setItem('theme', theme);
    setCookie('theme', theme, 365);
  }

  function saveThemeToServer(theme) {
    var userId = document.body.dataset.userId;
    var xhr = new XMLHttpRequest();
    xhr.open('PATCH', '/users/' + userId + '/update_theme', true);
    xhr.setRequestHeader('Content-Type', 'application/json');
    xhr.setRequestHeader('X-CSRF-Token', document.querySelector("meta[name='csrf-token']").content);
    xhr.send(JSON.stringify({ theme: theme }));

    xhr.onerror = function () {
      console.error('Error saving theme preference to the server.');
    };
  }

  function fetchTheme(callback) {
    if (document.body.dataset.userId) {
      var xhr = new XMLHttpRequest();
      xhr.open('GET', '/users/theme', true);
      xhr.setRequestHeader('X-CSRF-Token', document.querySelector("meta[name='csrf-token']").content);
      xhr.setRequestHeader('Accept', 'application/json');
      xhr.withCredentials = true;

      xhr.onload = function () {
        if (xhr.status >= 200 && xhr.status < 300) {
          var data = JSON.parse(xhr.responseText);
          callback(data.theme || 'light');
        } else {
          console.warn('Failed to fetch theme from the server. Using default theme.');
          callback('light');
        }
      };

      xhr.onerror = function () {
        console.warn('Failed to fetch theme from the server. Using default theme.');
        callback('light');
      };

      xhr.send();
    } else {
      var theme = localStorage.getItem('theme') || getCookie('theme') || 'light';
      callback(theme);
    }
  }

  function setCookie(name, value, days) {
    var expires = days
      ? '; expires=' + new Date(Date.now() + days * 864e5).toUTCString()
      : '';
    document.cookie = name + '=' + (value || '') + expires + '; path=/';
  }

  function getCookie(name) {
    var cookies = document.cookie.split('; ');
    for (var i = 0; i < cookies.length; i++) {
      var parts = cookies[i].split('=');
      if (parts[0] === name) {
        return parts[1];
      }
    }
    return null;
  }

  function eraseCookie(name) {
    document.cookie = name + '=; Max-Age=-1; path=/';
  }

  function updateThemeImages() {
    var images = document.querySelectorAll('img.tab_loading');
    for (var i = 0; i < images.length; i++) {
      images[i].src = isDarkModeEnabled()
        ? '/assets/horizontal_loading_dark.gif'
        : '/assets/horizontal_loading_light.gif';
    }
  }

  function init() {
    fetchTheme(function (theme) {
      applyTheme(theme);

      document.addEventListener('click', function (event) {
        var element = event.target;
        while (element) {
          if (element.classList && element.classList.contains('theme-switch')) {
            toggleTheme();
            break;
          }
          element = element.parentNode;
        }
      });

      var logoutLink = document.querySelector('[data-logout]');
    });
  }

  document.addEventListener('DOMContentLoaded', init);
})();
