/*!
 * Copyright (c) 2014 ENEO Tecnología S.L.
 * This file is part of redBorder.
 * redBorder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * redBorder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with redBorder. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Utils functions for images. This library calculate the aspect ratio of an
 * image and apply some styles and classes to the container to keep the
 * aspect ratio of the image using the max size.
 *
 * @param $image [JQuery selector] Image
 * @param $container [JQuery selector] Container of the image
 * @param opts [Object] Options for the library. These fields are optional.
 *          - widthFeet: Width of the image
 *          - heightFeet: Height of the image
 *
 */
redBorder.utils.imageAspectRatio = function ($image, $container, opts) {
  var margin_top = 0,
      margin_left = 0,
      // Container size
      widthContainerPx, heightContainerPx,
      // Image size
      widthPx, heightPx,
      originalWidthPx, originalHeightPx,
      widthFeet, heightFeet,
      // Ratios
      ratio, rr, rc,
      // Load callback
      deferred,
      calcDeferred,
      initialized = false,
      // Final result. It will be cached.
      dimensions,
      calculated = false;

  // Initialize options
  opts = opts || {};
  widthFeet = opts.widthFeet || 0;
  heightFeet = opts.heightFeet || 0;

  // Hide the container to prevent blink when update the size of the image
  $image.css('opacity', 0);

  /*
  * Return the calculated dimensions
  */
  function load() {
    if (!calculated) {
      return _load();
    } else {
      return $.Deferred.resolve(dimensions);
    }
  }

  /*
  * Check if image is loaded
  */
  function loaded() {
    return initialized;
  }

  /*
   * Delete cache and reload aspect ratio
   */
  function reload() {
    calculated = false;
    initialized = false;
    return load();
  }

  /*
  * Return a deferred Object that is resolved when the image is loaded. Then
  * you can execute the different methods on the library
  */
  function _load() {
    // Initialize Deferred object
    return $.Deferred(_loadImage).promise();
  }

  /*
  * Use the max size of the image but keep aspect ratio of it. This method
  * initialize the different variables of the image and the container.
  */
  function _calculateAspectRatio() {
    // Initialization
    if (!_initialized()) { return; }
    // Reset
    _reset();
    _loadSize();

    if (rr > rc) {
      // The limit is on height
      heightPx = Math.min(heightPx, heightContainerPx);
      widthPx = parseInt(heightPx / rr);
      ratio = rr;
      $container.addClass('keep_aspect_height');
      $container.css('width', widthPx);
      $container.css('height', heightPx);
    } else if (rr < rc) {
      // The limit is on width
      widthPx = Math.min(widthPx, widthContainerPx);
      heightPx = parseInt(widthPx * rr);
      ratio = rc;
      $container.addClass('keep_aspect_width');
      $container.css('height', 'auto');
      $container.css('width', widthPx);
    } else {
      ratio = 1;
      // Same aspect ratio
      heightPx = $container.height();
      widthPx = $container.width();
    }

    // Calc margins
    _calcMargin();

    // Return result
    calculated = true;

    // Show container
    $image.css('opacity', 1);

    // Set the result and return
    dimensions = {
      width: widthPx,
      widthFeet: widthFeet,
      originalWidth: originalWidthPx,
      height: heightPx,
      heightFeet: heightFeet,
      originalHeight: originalHeightPx,
      ratio: ratio
    }

    return dimensions;
  }

  /*
   * Check if the image was initialized
   */
  function _initialized() {
    if (!initialized) {
      console.warn("Error when load the image");
    }
    // Return status
    return initialized;
  }

  /*
   * Load sizes of image and container
   */
  function _loadSize() {
    // Container
    heightContainerPx = $container.height();
    widthContainerPx = $container.width();
    // Image
    heightPx = $image[0].naturalHeight;
    widthPx = $image[0].naturalWidth;
    // Ratios
    rr = heightPx / widthPx;
    rc = heightContainerPx / widthContainerPx;
    // Original sise
    originalWidthPx = widthPx;
    originalHeightPx = heightPx;
    // Set image feet
    if (widthFeet === 0 || heightFeet === 0) {
      widthFeet = widthPx;
      heightFeet = heightPx;
    }
  }

  /*
   * Calculate margin to apply
   */
  function _calcMargin() {
    if (widthContainerPx > widthPx) {
      marginLeft = (widthContainerPx - widthPx) / 2;
      $container.css('margin-left', marginLeft + 'px');
    }
    // We must center the img
    if (heightContainerPx > heightPx) {
      marginTop = (heightContainerPx - heightPx) / 2;
      $container.css('margin-top', marginTop + 'px' );
    }
  }

  /*
   * Reset the container
   */
  function _reset() {
    $container.removeClass('keep_aspect_height keep_aspect_width');
    $container.removeAttr('style');
  }

  /*
   * Load the image and return and resolver or reject a deferred object based
   * on events.
   */
  function _loadImage(_deferred) {
    // Initialize callbacks
    deferred = _deferred;
    $image.on('load', _successOnLoad);
    $image.on('error', _errorOnLoad);
  }

  /*
   * Resolve the deferred with success
   */
  function _successOnLoad() {
    // Return a success status
    _finishLoad();
    initialized = true;
    deferred.resolve(_calculateAspectRatio());
  }

  /*
   * Resolve the deferred with error
   */
  function _errorOnLoad() {
    // Return an error status
    _finishLoad();
    deferred.reject(false);
  }

  /*
   * Remove callback on the image
   */
  function _finishLoad() {
    $image.off('load');
    $image.off('error');
  }

  // Public
  return {
    load: load,
    loaded: loaded,
    reload: reload
  }
}
