#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'csv'

class AlarmsController < ApplicationController
  helper Redborder::ModulesHelper
  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_alarm, only: [:edit, :update, :destroy, :reactivate]
  before_action :convert_severity, only: [:create, :update]
  before_action :check_tabs, only: [:new]
  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all alarms in a table
  #
  def index
    @alarms = []

    all_alarms = if current_user.super_admin?
                   Alarm.all
                 else
                   current_user.alarms
                 end

    all_alarms.each do |a|
      @alarms << a if Redborder::Module.keys_with_interface
                      .include? a.product_type.to_sym
    end

    if request.xhr?
      @alarms =
        @alarms.select(&:fired).sort! do |a, b|
          b.last_fired <=> a.last_fired
        end
    end

    render layout: !request.xhr?
  end

  #
  # => Check for custom tabs and added
  #
  def check_tabs
    CustomTab.all.each do |custom_tab|
      if Redborder::Module.registered_modules.key? :monitor and
         !Redborder::Module.registered_modules[:monitor][:tabs].key? custom_tab.tab_name.to_sym
          Redborder::Module.registered_modules[:monitor][:tabs][custom_tab.tab_name.to_sym] = {  :name=>custom_tab.name,
                                                                                      :enabled=>custom_tab.enabled,
                                                                                      :percentage=>custom_tab.percentage,
                                                                                      :category=>custom_tab.category,
                                                                                      :units=>custom_tab.units,
                                                                                      :instance => custom_tab.tab_name.to_s + "_per_instance"
                                                                                   }

      end
    end

  end

  #
  # => Render the form to create a new alarm
  #
  def new
    @alarm = Alarm.new(severity_id: 0)

    return unless params[:from]
    @alarm.product_type = params.fetch(:product_type) do
      Redborder::Module.keys_with_interface.first.to_s
    end

    section =
    if Redborder::Module.keys_with_interface
       .include?(@alarm.product_type.to_sym)
      Redborder::Module
      .registered_modules[@alarm.product_type.to_sym][:session_id].to_sym
    else
      :event
    end

    # Get the source to get filters
    source = Redborder::DruidSource.get(source: @alarm.product_type)
    f = Redborder::Filter.new(source, load_filters)
    # Get the filters as json and exclude time filters
    @alarm.condition = f.serialize(true, true)
  end

  #
  # => Receive the data to create a new alarm. Try to create a new alarm
  #    and redirect to alarms_path if it's ok
  #
  def create
    @alarm = current_user.alarms.build(alarm_params)
    sensor = Sensor.find_by(id: alarm_params[:sensor_id])
    if sensor
      @alarm.sensor_id = sensor.id
    end
    @alarm.enabled = true

    @alarm.enabled = true

    respond_to do |format|
      if @alarm.save
        format.html do
          redirect_to alarms_path,
                      notice: "#{I18n.t('alarm_success')}"
        end
        format.js { flash[:notice] = "#{I18n.t('alarm_success')}" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @alarm.errors.full_messages.map do
            |msg| "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @alarm.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Render the form to edit an alarm
  #
  def edit
  end

  #
  # => Receive the data to update a new alarm. Try to update a new alarm
  #    and redirect to alarms_path if it's ok
  #
  def update
    sensor = Sensor.find_by(id: alarm_params[:sensor_id])
    if sensor
      @alarm.sensor_id = sensor.id
    end
    @alarm.enabled = true

    respond_to do |format|
      if @alarm.update(alarm_params)
        format.html do
          redirect_to alarms_path, notice: "#{I18n.t('alarm_update')}"
        end
        format.js { flash[:notice] = "#{I18n.t('alarm_update')}" }
        format.json { render json: { status: 200 } }
      else
        format.html { render action: 'edit' }
        format.js do
          flash[:error] = @alarm.errors.full_messages.map do
            |msg| "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @alarm.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Destroy an alarm
  #
  def destroy
    @alarm.destroy
    respond_to do |format|
      format.html { redirect_to alarms_path }
      format.js
      format.json { head :no_content }
    end
  end

  #
  # => Reset the interval to reactivate the alarm. When an alarm is fired
  #    it's wait interval time before trigger it again.
  #    This method reset this time.
  #
  def reactivate
    @alarm.update(fired: nil)
    unless DelayedJob.alarm?
      Delayed::Job.enqueue(
        Redborder::Jobs::CheckAlarmsJob.new(true),
        priority: 10,
        run_at: Time.now + 2.seconds
      )
    end
  end

  def search
  
    @alarms = Alarm.where("name ILIKE ?", "%#{params[:search]}%").order(created_at: :desc)
  
    respond_to do |format|
      format.js
    end
  end

  def export
    export_alarms_csv
  end

  #
  # Show a modal to upload a file to import alarms.
  #
  def import
    render layout: false
  end

  def import_file
    content_file = params[:alarm][:csv_file]
    
    unless content_file.content_type == "text/csv"
      redirect_to alarms_path, alert: t("alarms.import.csv_error")
      return
    end
  
    begin
      created_count, updated_count = Alarm.import_from_csv(content_file, current_user)
      flash[:notice] = t("alarms.import.import_success")
    rescue => e
      flash[:alert] = t("alarms.import.csv_error", error: e.message)
    end
  
    redirect_to alarms_path
  end
  
  
  private

  #
  # => Return a hash with the aprams to create/update an alarm
  #
  def alarm_params
    params
      .require(:alarm)
      .permit(:name, :product_type, :time_window, :interval,
              :upper_limit_enabled, :upper_limit, :bottom_limit_enabled,
              :bottom_limit, :condition, :notify_to_syslog, :notify_to_slack, :notify_to_email,
              :direction, :aggregation, :email_list, :slack_webhook,
              :multiplier_upper, :multiplier_bottom, :monitor_tab, :enabled,
              :severity_id, :timezone, :outliers_enabled, :outliers_granularity, :incident_enabled, :sensor_id)
  end

  #
  # => Take the severity of the form and convert it to real values
  #
  def convert_severity
    return if params.require(:alarm)[:severity_id].nil?

    case params.require(:alarm)[:severity_id].to_i
    when 0
      # Very low
      return params[:alarm][:severity_id] =
        Severity.find_by_name('very low').id
    when 1
      # Low
      return params[:alarm][:severity_id] =
        Severity.find_by_name('low').id
    when 2
      # Medium
      return params[:alarm][:severity_id] =
        Severity.find_by_name('medium').id
    when 3
      # High
      return params[:alarm][:severity_id] =
        Severity.find_by_name('high').id
    when 4
      # Very High
      return params[:alarm][:severity_id] =
        Severity.find_by_name('very high').id
    when 5
      # Critical
      return params[:alarm][:severity_id] =
        Severity.find_by_name('critical').id
    else
      # Very low by default
      return params[:alarm][:severity_id] =
        Severity.find_by_name('very low').id
    end
  end

  #
  # => Init the alarm if the user has permissions
  #
  def set_alarm
    @alarm = Alarm.find(params[:id])
    return if current_user.super_admin? || current_user.id == @alarm.user_id
    error_404('')
  end

  

  def alarms_to_csv_format
    columns = Alarm.column_names
  
    CSV.generate(col_sep: ";") do |csv|
      csv << columns
      Alarm.all.order(created_at: :desc).each do |alarm|
        csv << columns.map { |col| alarm.send(col) }
      end
    end
  end
  
  def export_alarms_csv
    filename = "#{Time.now.strftime("%Y%m%d%H%M%S")}-Alarms.csv"
    content = alarms_to_csv_format
    send_data content, filename: filename, type: 'text/csv'
  end
end
