#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::BaseController < ApplicationController
  # Skip CRSF authentication
  protect_from_forgery with: :null_session
  # Skip base authentication token. Login is defined on application contorller
  skip_before_filter :verify_authenticity_token, raise: false
  # Skip standard user authentication and setup
  skip_before_action :authenticate_user!, :user_setup, raise: false

  # Init the user for API
  # Destroy session vars on request and responses
  before_action :force_format, :authenticate_user_from_token!, :user_setup_api,
                :destroy_session

  #
  # APIs on HTTP are stateless so we must ignore cookies and sessions (always
  # check auth_token)
  #
  def destroy_session
    request.session_options[:skip] = true
  end

  #
  # Init the user for API
  #
  def user_setup_api
    return unless user_signed_in?

    if current_user.enabled
      User.current_user = current_user
    else
      # The account is disabled
      sign_out current_user
      error_401('Your account has been disabled.'\
                'Please contact the administrator.')
    end
  end

  #
  # Login the user based on auth token
  #
  def authenticate_user_from_token!
    auth_token = params[:auth_token].presence
    user = User.find_by(authentication_token: auth_token.to_s)

    if user && auth_token
      sign_in user, store: false, bypass: true
      # TODO: We musn't override current_user, but there aren't any way
      # to kkep session user and auth_token user separated
      @current_user = user
    elsif auth_token
      # The user doesn't exist
      error_401('There aren\'t any user with this authentication token')
    else
      # No auth token sent
      error_401('You must specify the auth_token to use API')
    end
  end

  #
  # Check the format of request. If format isn't CSV or JSON, we will force it
  # to JSON
  #
  def force_format
    request.format = :json unless [:json, :csv].include?(request.format.symbol)
  end

  # -------------------------------------------------------------
  # Helper functions for permissions
  # -------------------------------------------------------------

  #
  # Check super admin permissions
  #
  def need_superadmin
    return error_403('You don\'t have enought privileges to do that.') unless
      current_user.super_admin?
  end

  # -------------------------------------------------------------
  # Helper functions for parameters
  # -------------------------------------------------------------
  
  #
  # Check existence of a mandatory parameter.
  #
  # == Parameters:
  # mandatory::
  #   Array of symbols or symbol.
  #
  # == Returns:
  # If some parameter is not present, this method return a 422 status
  #
  def mandatory_parameters(mandatory)
    Array(mandatory).each do |p|
      next if params[p].present?
      return error_422("#{p} is a mandatory parameter and it's not present")
    end
  end
end
