
#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::ChefController < Api::V1::BaseController
  ## ======================================================================
  # => Global values
  ## ======================================================================

  DATABAGS = begin
    YAML.load_file("#{Rails.root}/config/databags.yml") || { 'default' =>  { 'passwords' => [], 'rBglobal' => [] } }
  rescue Errno::ENOENT
    { 'default' =>  { 'passwords' => [], 'rBglobal' => [] } }
  end

  READ_MANAGERS =
    YAML.load_file("#{Rails.root}/config/chef_config.yml")['read_manager_nodes'] || {}

  ## ======================================================================
  # => Before actions
  ## ======================================================================

  skip_before_action :authenticate_user_from_token!, :user_setup_api
  before_action :validate_cert
  before_action :set_sensor
  before_action :get_managers, only: [:nodes, :node, :update_node]
  before_action :get_databags_passwords, only: :data_passwords
  before_action :get_databags_rBglobal, only: :data_rBglobal
  after_action :validate_ip, only: :client, if: -> { @sensor.is_client_proxy? || @sensor.is_exporter? }

  # GET /nodes
  def nodes
    result_json = { @username => "#{Chef::Config[:chef_server_url]}/nodes/#{@username}" }

    # Add managers to json response if read_managers is enabled in the config file
    if @managers.present?
      @managers.each do |manager|
        result_json.merge!(manager => "#{Chef::Config[:chef_server_url]}/nodes/#{manager}")
      end
    end
    render json: result_json
  end

  # GET /roles
  def roles
    # Get all its ancestors' roles
    ids = @sensor.self_and_ancestors_with_root.map(&:id)
    json_response = {}
    ids.each do |id|
      json_response["rBsensor-#{id}"] = "#{Chef::Config[:chef_server_url]}/nodes/rBsensor-#{id}"
    end

    run_list_text = run_list_by_type(@sensor)

    # Default role per type
    json_response[run_list_text] = "#{Chef::Config[:chef_server_url]}/nodes/#{run_list_text}"
    # Default role for all sensors
    json_response['sensor'] = "#{Chef::Config[:chef_server_url]}/nodes/sensor"

    render json: json_response.to_json
  end

  # GET /nodes/:id
  def node
    node_name = params[:id]
    node_id = get_sensor_id_by_chef(node_name)
    if @sensor.id == node_id || (@managers.present? &&
                                @managers.include?(node_name))
      rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
      result = rest.get_rest("nodes/#{node_name}", false)
      render json: result.for_json
    else
      return error_401('You are not allowed to access this node.')
    end
  end

  # PUT /nodes/:id
  def update_node
    node_name = params[:id]
    node_id = get_sensor_id_by_chef(node_name)
    if @sensor.id == node_id || (@managers.present? &&
                                @managers.include?(node_name))
      rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
      result = rest.put("/nodes/#{node_name}", params.except(:action, :id, :chef, :controller))
      render json: result.as_json
    else
      return error_401('You are not allowed to access this node.')
    end
  end

  # GET /clients/:id
  def client
    client_name = params[:id]
    client_id = get_sensor_id_by_chef(client_name)

    if @sensor.id == client_id
      rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
      result = rest.get_rest("clients/#{client_name}", false)
      render json: result.as_json
    else
      return error_401('')
    end
  end

  # GET /roles/:id
  def role
    role_name = params[:id]
    role_id = get_sensor_id_by_chef(role_name)
    run_list_text = ['sensor', run_list_by_type(@sensor)]

    if (role_id && @sensor.self_and_ancestors_with_root.map(&:id).include?(role_id)) || run_list_text.include?(role_name)
      render json: Chef::Role.load(role_name).as_json
    else
      return error_401('')
    end
  end

  # GET /environments
  def environments
    environment_name = @sensor.chef_node.chef_environment
    render json: { environment_name => "#{Chef::Config[:chef_server_url]}/environments/#{environment_name}" }
  end

  # GET /environments/:id
  def environment
    environment_name = params[:id]
    if environment_name != @sensor.chef_node.chef_environment
      return error_401('')
    else
      render json: Chef::Environment.load(environment_name).as_json
    end
  end

  # POST /environments/:id/cookbook_versions
  def cookbook_versions
    environment_name = params[:id]
    json_response = {}
    if environment_name != @sensor.chef_node.chef_environment
      return error_401('')
    else
      rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
      result = rest.post("/environments/#{environment_name}/cookbook_versions", params)

      params["run_list"].each do |key|
        json_response[key] = result[key].manifest
      end

      render json: json_response.as_json, status: :ok
    end
  end

  # GET /data/passwords/:id
  def data_passwords
    return error_401('') unless @passwords.include?(params[:id])
    rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
    result = rest.get_rest("/data/passwords/#{params[:id]}", false)
    render json: result.as_json
  end

  # GET /data/rBglobal/:id
  def data_rBglobal
    return error_401('') unless @rBglobal.include?(params[:id])
    rest = Chef::REST.new(Chef::Config[:chef_server_url], 'rb-chef-webui', Chef::Config[:client_key])
    result = rest.get_rest("/data/rBglobal/#{params[:id]}", false)
    render json: result.as_json
  end

  private

  #
  # => Validate the certificate of the client that make the request.
  #
  def validate_cert
    begin
      authenticator = Mixlib::Authentication::SignatureVerification.new(request)
      @username = authenticator.user_id
      user = Chef::ApiClient.load(@username)
      user_key = OpenSSL::PKey::RSA.new(user.public_key)
      authenticator.authenticate_request(user_key)
    rescue Mixlib::Authentication::MissingAuthenticationHeader => e
      return error_401(e.message)
    rescue StandardError => se
      return error_401(se.message)
    end

    if !authenticator.valid_request? && authenticator.valid_timestamp?
      return error_401('Failed to authenticate. Ensure that your client key is valid.')
    elsif !authenticator.valid_request?
      return error_401('Failed to authenticate. Please synchronize the clock on your client.')
    end
  end

  # Returns databags passwords for the given username
  def get_databags_passwords
    name = @username.match(/^(?<name>\w+)-\d+$/)[:name]
    @passwords = (DATABAGS[name].try(:[], 'passwords') || []) | (DATABAGS['default'].try(:[], 'passwords') || [])
  end

  # Returns databags rBglobal for the given username
  def get_databags_rBglobal
    name = @username.match(/^(?<name>\w+)-\d+$/)[:name]
    @rBglobal = (DATABAGS[name].try(:[], 'rBglobal') || []) | (DATABAGS['default'].try(:[], 'rBglobal') || [])
  end

  # Returns managers names if read_managers is enabled in the config file
  def get_managers
    name = @username.match(/^(?<name>\w+)-\d+$/)[:name]
    can_read_manager = READ_MANAGERS[name]
    @managers = Manager.all.map(&:name) if can_read_manager.to_s == "true"
  end

  def set_sensor
    username_match = @username.match(/^\w+-(?<id>\d+)$/)
    @sensor = Sensor.find(username_match[:id])
  end

  def run_list_by_type(sensor)
    role_name = sensor.send("run_list_text_#{sensor.type}")
    role_name.match(/^role\[(?<role_name>\w+\-\w+)\]$/)[:role_name]
  end

  #
  # => Get the ID from the role/node name.
  #    i.e From xxxxxxx-35, the ID is 35
  #
  def get_sensor_id_by_chef(chef_name)
    chef_name_attributes = chef_name.match(/^\w+-(?<id>\d+)$/)
    return chef_name_attributes[:id].to_i if chef_name_attributes
  end

  def validate_ip
    client_ip = (request.headers['X-Forwarded-For'] ||
                 request.headers['X-Real-IP'] ||
                 request.remote_ip).split(',').first
    @sensor.update(ip: client_ip)
  end
end
