#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::DataController < Api::V1::BaseController
  # Init the data source based on product
  before_action :check_content_type, :check_allowed_data_sources,
                :set_data_source, :set_time_out

  #
  # Call direct to Druid. The user who send the query defines the data sources
  # he can access
  #
  def index
    # Get the client based on product
    client = Redborder::Rbdruid.get_clients[@product] ||
             Redborder::Rbdruid.druid_reconnect(@product)
    # Try to send the query
    send_query(client)

  rescue StandardError => e
    Rails.logger.error e
    Rails.logger.error e.backtrace.join("\n")
    return error_500('')
  end

  private

  # Check mime type of the query. If it's different from JSON, the API will
  # return a 400 error
  #
  def check_content_type
    return if request.content_type == 'application/json'
    error_400('Content-Type must be application/json')
  end

  #
  # Send the query over the RubyDruid gem
  #
  # == Parameters:
  # client::
  #   Client object to send the data.
  #
  def send_query(client)
    return error_503('Zookeeper is not available') unless client

    begin
      # Show the query and try to get the data
      # TODO: We must protect Druid from unpermitted params
      # query = druid_params.merge(dataSource: @datasource)
      query = params['datum'].merge(dataSource: @datasource) { |_, _, v2| v2 }
      # Set timeout
      query['context'] ||= {}
      query['context'].merge!('timeout' => @timeout)

      Rails.logger.info "Druid Query: #{query.to_json}"
      response = get_response(client.broker_uri, query.to_json)
    rescue StandardError => e
      # Return error and log
      Rails.logger.error e
      Rails.logger.error e.backtrace.join("\n")
      return error_500("Bad request. Druid can\'t process this query.")
    end

    # Check response
    if response.is_a? Net::HTTPOK
      render json: JSON.parse(response.body), status: :ok
    elsif response.is_a? Net::HTTPInternalServerError
      error_message = JSON.parse(response.body)['error']
      Rails.logger.error error_message
      Rails.logger.error query.to_json
      error_500(error_message)
    else
      Rails.logger.error response
      Rails.logger.error query.to_json
      error_500(response)
    end
  end

  #
  # Get the response
  #
  def get_response(uri, request_body)
    http = Net::HTTP.new(uri.host, uri.port)
    http.use_ssl = false
    http.read_timeout = @timeout / 1_000

    # TODO: prevent use VERIFY_NONE
    http.verify_mode = OpenSSL::SSL::VERIFY_NONE
    request = Net::HTTP::Post.new(uri.path)
    request.add_field('Content-Type', 'application/json')
    request.body = request_body
    http.request(request)
  rescue Net::ReadTimeout => e
    e.message
  end

  #
  # Define query params available to send to Druid.
  #
  def druid_params
    params
      .permit(:queryType, :dimension, :metric, :threshold,
              granularity_fields,
              intervals: [],
              filter: filter_fields,
              context: context_fields,
              aggregations: aggregations_fields,
              postAggregations: post_aggregations_fields
             )
  end

  #
  # Granularity fields.
  #
  # == Returns:
  # Array or symbol with available params for granularity
  #
  def granularity_fields
    return :granularity if params[:granularity].is_a?(String)
    { granularity: [:type, :period, :origin] }
  end

  #
  # Aggregation fields.
  #
  # == Returns:
  # Array with available params for aggregations
  #
  def aggregations_fields
    [:type, :name, :fieldName, :fn]
  end

  #
  # PostAggregation fields.
  #
  # == Returns:
  # Array with available params for postAggregations
  #
  def post_aggregations_fields
    [:type, :fn, :name, { fields: [:value, :type, :fieldName] }]
  end

  #
  # Filter fields.
  #
  # == Returns:
  # Array with available params for filter
  #
  def filter_fields
    [:type,
     { fields: [:type, :dimension, :value,
                { fields: [:type, :dimension, :value] }
               ]
     }
    ]
  end

  #
  # Context fields
  #
  # == Returns:
  # Array with available params for context
  #
  def context_fields
    [:timeout, :skipEmptyBuckets]
  end

  #
  # Iterate over available data sources
  #
  def check_allowed_data_sources
    return error_400('dataSource must be specified') unless params[:dataSource]
    # Get the product from druid
    RedBorder::DRUID_CONFIG.each do |key, value|
      if value.is_a?(Hash) && value['datasource'].present? &&
         value['datasource'].to_sym == params[:dataSource].to_sym
        @product = key
      end
    end
    # Check if dataSource exist
    return error_404('The data source doesn\'t exist') unless @product
  end

  #
  # Init Data Source based on product. It fails if the product isn't specified.
  #
  def set_data_source
    # If multidatasource is present and value is false, dont set data_source
    # depend on namespaces
    if RedBorder::DRUID_CONFIG[@product]['multidatasource'] && RedBorder::DRUID_CONFIG[@product]['multidatasource'] == 'false'
      @datasource = params[:dataSource]
      return
    end

    # Init namespaces
    if params[:namespace_uuids]
      @namespaces = namespaces_to_i
      # Check permissions on namespaces ids.
      if (@namespaces - current_user.data_sources_ids).any?
        forbidden_namespaces = (@namespaces - current_user.data_sources_ids)
        return error_403("You don't have enought privileges on namespace_uuids #{forbidden_namespaces.join(', ')}.")
      end
    else
      @namespaces = [current_user.data_sources_ids.first]
    end
    # Init the data source
    @datasource = datasource(@namespaces)
  end

  #
  # Init datasource based on a source
  #
  # == Parameters:
  # namespaces::
  #   Array of integers with the namespace ids
  #
  # == Returns:
  # String or Hash with dataSource to send to druid
  #
  def datasource(namespaces)
    ds = params[:dataSource]
    data_sources = []

    namespaces.each do |namespace|
      data_sources << (namespace == Sensor.root.uuid ? ds : "#{ds}_#{namespace}")
    end

    return data_sources.first if data_sources.size == 1
    # Return an union
    { type: 'union',
      dataSources: data_sources }
  end

  #
  # Convert namespaces to integer.
  #
  # == Returns:
  # Array of integers
  #
  def namespaces_to_i
    did = params[:namespace_uuids]
    did.is_a?(Array) ? did.map(&:to_s) : [did.to_s]
  end

  #
  # Init context timeout if the query has no implicit indicated
  #
  #
  def set_time_out
    @timeout =
      if params['datum']['context'] && params['datum']['context']['timeout']
        params['datum']['context']['timeout'].to_i
      else
        (RedBorder::DRUID_CONFIG[:timeout_api] || 20).to_i * 1_000
      end
  end
end
