#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Call to dataSources to get data.
#
class Api::V1::ModulesController < Api::V1::BaseController
  # All responses are on JSON
  respond_to :json

  # Before actions
  before_action :init_data
  before_action :set_interval, only: [:tops, :top_serie]
  before_action :set_time_by_granularity, only: [:index, :raw_serie]
  before_action :set_granularity_by_range, only: :top_serie

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Returns raw info from module.
  #
  def index
    @offset = (params[:offset] || 0).to_i

    # Default query hash
    @events = @source.where(
      query_type: :group_by,
      range: :custom,
      start_time: @start_time,
      end_time: @end_time,
      granularity: @granularity,
      group_by: @source.default_columns_raw,
      aggregation: @aggregation,
      limit: current_user.per_page_count,
      offset: @offset
    )

    past_events if @events.size.zero? &&
                   @granularity.to_sym != :all &&
                   range == :raw_column

    render json: {
      query: true,
      offset: @offset + @events.size,
      end_time: @end_time.to_i,
      aggregation: @aggregation,
      events: @events
    }
  end

  #
  # Returns top info from module.
  #
  def tops
    # Tabs and offset
    @tab = params[:id].to_sym
    @offset = (params[:offset] or 0).to_i

    tops = @source.where(
      query_type: :group_by,
      group_by: @source::COLUMNS[@tab.to_sym][:columns].keys,
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      sort: [@aggregation, @source::COLUMNS[@tab.to_sym][:columns].keys.first],
      offset: @offset,
      limit: current_user.per_page_count,
      aggregation: @aggregation
    )

    total = @source.total(
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      aggregation: @aggregation
    ) || 0

    # Render the response
    render json: {
      query: true,
      aggregation: @aggregation,
      offset: @offset + tops.size,
      total: total,
      tops: tops
    }
  end

  #
  # Returns the info of an event.
  #
  def info
    @filter = Redborder::Filter.new(@source, {})
    @filter.parse(params[:search], true) if params[:search]

    # Create a search hash to find the element
    @source.default_columns_raw.each do |column|
      @filter.add(column.to_sym, [params[column]])
    end

    @start_time = Time.at(params[:timestamp].to_i)
    @end_time = @start_time +
      Redborder::Rbdruid::GRANULARITY_TYPE[@granularity.to_sym][:step]/1000

    # Set the columns to search
    @source::COLUMNS.each{|k,v| (@keys ||= []) << k unless v[:hide_raw]}

    # Get the event
    events = @source.where(
      query_type: :group_by,
      range: :custom,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      group_by: @keys,
      aggregation: @aggregation,
      search: @filter.get
    )

    render json: {
      query: true,
      events: events
    }
  end

  #
  # Get events as raw serie to display the data in a graph
  #
  def raw_serie
    # Parse the filters
    @filter = Redborder::Filter.new(@source, {})
    @filter.parse(params[:search], true) if params[:search]

    serie = @source.timeserie(
      range: params[:range] || :custom,
      start_time: @start_time,
      end_time: @end_time,
      granularity: @granularity,
      search: @filter.get,
      aggregation: @aggregation
    )

    # Render the data
    render json: {
      query: true,
      start_time: @start_time,
      end_time: @end_time,
      aggregation: @aggregation,
      serie: serie
    }
  end

  #
  # Get events as top serie to display the data in a graph
  #
  def top_serie
    # Init filters
    @filter = Redborder::Filter.new(@source, {})
    @filter.parse(params[:search], true) if params[:search]
    # Init the data
    tab = params[:id].to_sym
    series = []

    if params[:values] && params[:values].is_a?(Array)
      params[:values].each do |value|
        @filter.remove(tab)
        series << get_serie(value, @filter.add(tab, [value]))
      end
    elsif params[:values]
      series << get_serie(params[:values], @filter.add(tab, [params[:values]]))
    else
      series << get_serie('all')
    end

    # Render series
    render json: {
      query: true,
      start_time: @start_time,
      end_time: @end_time,
      aggregation: @aggregation,
      series: series
    }
  end

  # -------------------------------------------------------------
  # Helpers
  # -------------------------------------------------------------

  private

  #
  # Return a serie and create a has based on data, name and filters.
  #
  # == Parameters:
  # name::
  #   Name of the serie
  # filters::
  #   Hash with filters
  #
  # == Returns:
  # A hash with the name and serie.
  #
  def get_serie(name, filters = {})
    {
      name: name,
      serie: @source.timeserie(
               start_time: @start_time,
               end_time: @end_time,
               granularity: @granularity,
               aggregation: @aggregation,
               search: filters
             )
    }
  end

  #
  # Init times based on ranges.
  #
  def set_interval
    @range = (params[:range] || 'last_1_hour')
    @end_time = in_time_machine { Time.now }

    case @range.to_sym
    when :last_1_hour
      @start_time = @end_time - 1.hour
    when :last_2_hour
      @start_time = @end_time - 2.hour
    when :last_3_hour
      @start_time = @end_time - 3.hour
    when :last_12_hour
      @start_time = @end_time - 12.hour
    when :last_24
      @start_time = @end_time - 24.hour
    when :last_week
      @start_time = @end_time - 1.week
    end
  end

  #
  # Init the times based on the granularity or start and end times.
  #
  def set_time_by_granularity
    @end_time =
      if params[:end_time].present?
        Time.at(params[:end_time].to_i)
      else
        in_time_machine { Time.now }
      end

    @start_time =
      if params[:start_time].present?
        Time.at(params[:start_time].to_i)
      else
        case @granularity.to_sym
        when :minute
          @end_time - 3.hours
        when :pt5m
          @end_time - 24.hours
        when :fifteen_minute
          @end_time - 3.days
        when :thirty_minute
          @end_time - 5.days
        when :hour
          @end_time - 1.week
        else
          Time.at(0)
        end
      end
  end

  #
  # Init the granularity based on the range.
  #
  def set_granularity_by_range
    @range ||= (params[:range] || 'last_1_hour')
    @granularity =
      case @range.to_sym
      when :last_1_hour, :last_3_hour
        :minute
      when :last_24
        :pt5m
      when :last_week
        :hour
      end
  end

  #
  # Set the source from url ID. If module doesn't exist, a 404 is returned
  #
  def init_data
    begin
      @source = Redborder::DruidSource.get(source: params[:module])
      # Set another elements
      @granularity  = (params[:granularity] || :minute)
      set_aggregation
    rescue NotImplementedError
      # Render an error. The module doesn't exist or source name hasn't been
      # sent
      render json: {
        status: false,
        error: 'The module doesn\'t exist or it isn\'t specified'
      }, status: 404
    end
  end

  #
  # Set the aggregation to use on queries. We check if the dataSource
  # has that aggregation defined. If it isn't defined, we take default
  # aggregation.
  #
  def set_aggregation
    @aggregation =
      if @source::AGGREGATIONS.include?(params[:aggregation].try('to_sym'))
        params[:aggregation].to_sym
      else
        @source.default_aggregation
      end
  end

  #
  # Search past events if the result is empty.
  #
  def past_events
    # Change the end_time and check if we have events.
    @end_time -= GRANULARITY_TYPE[@granularity.to_sym][:step] / 1000
    events = @source.where(
      query_type: :timeseries,
      range: :custom,
      start_time: @start_time,
      end_time: @end_time,
      granularity: @granularity,
      aggregation: @aggregation
    )

    return unless events.present?

    # We have events, return them
    @offset = 0
    @end_time = events.first.timestamp +
                GRANULARITY_TYPE[@granularity.to_sym][:step] / 1000

    @events = @source.where(
      query_type: :group_by,
      range: :raw_column,
      start_time: @start_time,
      end_time: @end_time,
      granularity: @granularity,
      limit: current_user.per_page_count,
      offset: @offset,
      group_by: @source.default_columns_raw,
      aggregation: @aggregation
    )
  end
end
