#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::MonitorController < Api::V1::BaseController
  include Redborder::Rbdruid

  # Respond to json
  respond_to :json

  # Before...
  before_action :set_interval, :set_source, :set_sensor, only: [:info]

  # Get the list of devices
  def index
    devices = {}
    # Get the IPS and Flow sensors
    devices[:sensors] = current_user
                        .sensor.self_and_descendants
                        .where(is_deleted: false)
                        .order(type: :asc)
    # Add managers if user is superadmin
    devices[:managers] = Manager.all.map(&:name) if current_user.super_admin?

    render json: { query: true }.merge(devices)
  end

  # Get the general information. It's include name of devices,
  # and mean of CPU, memory and Load
  # in the last 3 hours
  def info
    # Get the sensor
    @filter = Redborder::Filter.new(@source, {})
    @filter.add(:sensor_name, [@sensor.name])

    # Tabs to get data
    categories_to_search = [
      :cpu,
      :memory,
      :memory_cache,
      :memory_buffer,
      :load_1,
      :fan,
      :system_temp]

    # Result
    result = []

    # Get the data
    categories_to_search.each do |category|
      data = @source.where(search_hash(category.to_sym))
      result << data.first if data.first.present?
    end

    render json: {
      query: true,
      id: @sensor.id,
      sensor_name: @sensor.name,
      end_time: @end_time.to_i,
      monitor: result
    }
  end

  private

  #
  # Set dataSource
  #
  def set_source
    @source = Redborder::DruidSource.get(source: 'monitor')
  end

  #
  # Set the sensor to search
  #
  def set_sensor
    # Use find_by_id to display an error and don't show a 500
    @sensor = Sensor.find_by_id(params[:id].to_i)

    # Return errors iff sensor doesn't exist or user can't read it
    return error_404('The sensor doesn\'t exist') if @sensor.nil?
    return error_401('You don\'t have enought privileges to do that.') unless
      current_user.can?(:read, @sensor)
    return error_400('Domain sensors haven\'t got monitors') if
      @sensor.is_domain?
  end

  #
  # Set the hash to search
  #
  # == Parameters:
  # tab::
  #   Symbol with hash to search
  #
  # == Returns:
  # A hash with options to search.
  #
  def search_hash(tab)
    {
      query_type: :group_by,
      monitor_tab: tab,
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      sort: [:max, :sensor_name],
      limit: current_user.per_page_count,
      group_by: @source::COLUMNS.keys,
      aggregation: @source.default_aggregation,
      search: @filter.get
    }
  end

  # Set the interval
  def set_interval
    @range = (params[:range] || 'last_10_minutes')
    @end_time = in_time_machine { Time.now }

    case @range.to_sym
    when :last_10_minutes
      @start_time = @end_time - 10.minute
    when :last_1_hour
      @start_time = @end_time - 1.hour
    when :last_2_hour
      @start_time = @end_time - 2.hour
    when :last_3_hour
      @start_time = @end_time - 3.hour
    when :last_12_hour
      @start_time = @end_time - 12.hour
    when :last_24
      @start_time = @end_time - 24.hour
    when :last_week
      @start_time = @end_time - 1.week
    end
  end
end
