#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Provide an api to work with sensors
#
class Api::V1::SensorsController < Api::V1::BaseSensorsController
  # ----------------------------------------------
  # Before and after actions
  # ----------------------------------------------
  cattr_accessor :skip_methods
  @skip_methods = [:create]
  # There's no need to authenticate user when creating a sensor
  skip_before_action :authenticate_user!, raise: false, only: @skip_methods
  skip_before_action :authenticate_user_from_token!, only: @skip_methods
  # Set the parent
  before_action :set_parent,
                only: [:index, :create_domain, :move]
  before_action :set_domain,
                only: [:override_all, :destroy, :tree, :override_aps,
                       :delete_aps, :move]
  before_action :set_domain_only_param,
                only: [:update]

  before_action :destroying_own_top_domain, only: [:destroy]

  before_action :manage_parent, only: [:create_domain, :move]
  before_action :manage_domain,
                only: [:override_all, :destroy, :override_aps, :delete_aps,
                       :update, :move]
  before_action :check_params_format, only: [:override_all]
  before_action :filter_mac_address, only: [:create]

  # ----------------------------------------------
  # Actions
  # ----------------------------------------------

  #
  # Return the list of sensors.
  #
  def index
    render json: {
      query: true,
      sensors: @parent.all_children(true)
    }
  end

  #
  # Return the sensors as a tree
  #
  def tree
    render json: {
      query: true,
      sensor: @domain.self_and_children_as_json
    }
  end

  def override_aps
    @domain.self_and_descendants.each do |s|
      s.delete_all_access_points
    end
    # override with given ones
    @domain.create_access_points_from_json params[:access_points], @domain
    respond_to do |format|
      format.json do
        render json: { query: true }
      end
    end
  end

  #
  # Update current sensor and his children. If append param is true, keep old
  # sensors, else, destroy them. The sensor param must be an object with
  # current sensor and his children. If you don't need to modify the parent
  # only set values on children:
  #
  #   sensor: { children: [...] }
  #
  # @see redBorder API documentation
  #
  def override_all
    # override with given ones. el has the uuid of elements updated or created
    el = []
    @err = nil

    # Perform in a transaction to rollback changes
    ActiveRecord::Base.transaction do
      begin
        el = @domain.override_sensor_with_children(params[:sensor],
                                                   User.current_user)
      rescue ActiveRecord::StatementInvalid => e
        # A sensor is invalid, we must rollback changes
        @err = e.to_s
        raise ActiveRecord::Rollback
      end
    end

    if @err.nil?
      # No error
      # If the param to append is true, don't delete old elements
      if (params[:append].nil? || params[:append] == false) &&
         @domain.descendants
        @domain.descendants.each do |s|
          # Skip updated/created elements
          next if el.include?(s.uuid)

          # Assign the users to the parent of sensor (if exist) or to Main
          # domain (first override). The users will be disabled
          id = s.id
          dst = if s.parent && el.include?(s.parent.uuid)
                  s.parent.id
                else
                  @domain.id
                end

          # We need to define id and dst to prevent PSQL deadlocks on UPDATE
          # and DESTROY
          User
            .unscoped # If the user is disable, we need to update the sensor too
            .where(sensor_id: id)
            .update_all(sensor_id: dst, enabled: false)
          s.destroy
        end
      end

      respond_to do |format|
        format.json do
          render json: { query: true }
        end
      end
    else
      return error_400("Any changes have been saved. #{@err}")
    end
  end

  #
  # Create a new domain sensor.
  #
  def create_domain
    # Request is valid, proceed to sensor creation
    sensor_params = if params[:domain_type].to_s == Sensor::ORGANIZATION.to_s
                      organization_params
                    else
                      domain_params
                    end
    # Create the sensor
    sensor = Sensor.new(sensor_params)
    sensor.parent = @parent
    sensor.type = Sensor::DOMAIN
    sensor.domain_type = @parent.children_type if sensor.domain_type.nil?

    # Try to save the sensor.
    if sensor.save
      respond_to do |format|
        format.json do
          render json: {
            query: true,
            sensor: sensor.to_api
          }
        end
        format.csv { render text: sensor.to_csv }
      end
    else
      error_422("The sensor can\'t be created due: "\
                "#{sensor.errors.full_messages.join(', ')}")
    end
  end

  #
  # Update a domain sensor
  #
  def update
    # params to update
    sensor_params = @domain.organization? ? organization_params : domain_params
    if @domain.update(sensor_params)
      respond_to do |format|
        format.json do
          render json: {
            query: true,
            sensor: @domain.to_api
          }
        end
        format.csv { render text: @domain.to_csv }
      end
    else
      error_422("The sensor can\'t be updated due: "\
                "#{@domain.errors.full_messages.join(', ')}")
    end
  end

  #
  # Destroy a sensor if the user has permissions
  #
  def destroy
    # Change the users top domain (and below) to the parent of destroyed.
    # These users will be disabled
    destroyed_ids = @domain.self_and_descendants.map(&:id)
    User.where(sensor_id: destroyed_ids)
      .update_all(sensor_id: @domain.parent.id, enabled: false)

    # Destroy it
    @domain.destroy

    render json: { query: true }
  end

  #
  # POST /api/v1/sensors
  # params:
  #       order: [register, verify]
  #       mac
  #       uuid
  #
  def create
    if params[:order] == 'register'
      register_sensor(params[:type], params[:hash])
    elsif params[:order] == 'verify'
      verify_sensor(params[:hash], params[:uuid])
    else
      # no allowed request
      render json: {}, status: :forbidden
    end
  end

  #
  # Move domains/sensors to another parent
  #
  def move
    if @domain.update(parent: @parent)
      render json: {
        query: true,
        sensor: @domain.to_api
      }
    else
      error_422("The sensor can\'t be moved due: "\
                "#{@domain.errors.full_messages.join(', ')}")
    end
  end

  def claim
    parent_sensor = User.current_user.sensor
    @sensor =
      Sensor.where(unclaimed: true)
      .where('property like ?', "%\"mac_address\":\"#{params['uuid']}\"%")
      .first

    if @sensor && current_user.can?(:manage, parent_sensor)
      @sensor.update(
        name: params[:name],
        unclaimed: false,
        parent_id: parent_sensor.id
      )
      render json: { status: 'claimed', uuid: @sensor.uuid }
    elsif @sensor
      error_422("You don\'t have privileges to do that.")
    else
      error_422("The UUID doesn\'t correspond to any unclaimed device.")
    end
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    #
    # GET /api/v1/sensors
    # params:
    #       sensor_uuid
    #
    def nmsp_info
      sensor = Sensor.find_by_uuid(params[:sensor_uuid])

      return error_404('There is no sensor with that uuid.') unless
        sensor.present?
      return error_422("You don\'t have privileges to do that.") unless
        current_user.can?(:manage, sensor)
      return error_404('This sensor hasn\'t got any NMSP information.') unless
        sensor.chef_node['redborder']['nmsp']

      render json: sensor.chef_node['redborder']['nmsp'].to_json
    rescue StandardError
      error_500('There is an error trying to obtain this sensor information.')
    end
  end

  #
  # Return the list of licenses.
  #
  # GET /api/v1/sensors/:sensor_uuid/licenses
  # params:
  #       available_by_type (optional)
  #       auth_token
  #
  def licenses
    return error_422("You don\'t have privileges to do that.") unless
      current_user.admin?
    return error_400('The licensing mode is not by organization.') unless
      RedBorder::LICENSING_MODE == 'organization'

    sensor = Sensor.find_by_uuid(params[:sensor_uuid])

    return error_404('There is no sensor with that uuid.') unless
      sensor.present?
    return error_422("You don\'t have privileges to do that.") unless
      current_user.can?(:manage, sensor)

    organization_licenses = License.where(sensor_id: sensor.id)
    licenses = if params[:available_by_type].blank?
                 organization_licenses
               else
                 organization_licenses.available_by_type(params[:available_by_type], sensor)
               end
    if licenses.count.zero?
      render json: { message: 'There are no licenses available for this type of sensor.' }
    else
      render json: licenses.as_json(except: [:encoded_info, :signature,
                                             :created_at, :updated_at])
    end
  rescue StandardError
    error_500('There is an error trying to obtain this sensor information.')
  end

  #
  # Return the information of a license.
  #
  # GET /api/v1/sensors/:sensor_uuid/licenses/:id
  # params:
  #       auth_token
  #
  def license
    return error_422("You don\'t have privileges to do that.") unless
      current_user.admin?
    return error_400('The licensing mode is not by organization.') unless
      RedBorder::LICENSING_MODE == 'organization'

    sensor = Sensor.find_by_uuid(params[:sensor_uuid])

    return error_404('There is no sensor with that uuid.') unless
      sensor.present?
    return error_422("You don\'t have privileges to do that.") unless
      current_user.can?(:manage, sensor)

    license = License.find_by(sensor_id: sensor.id, id: params[:id])
    if license
      render json: license.as_json(except: [:encoded_info, :signature,
                                            :created_at, :updated_at])
    else
      render json: { message: 'There is no license with that id.' }
    end
  rescue StandardError
    error_500('There is an error trying to obtain this information.')
  end

  private

  def filter_mac_address
    return error_400('You must specify a hash parameter') if params[:hash].blank?
    return error_400('mac_address contains forbidden characters') unless
      /^[0-9A-Z\-:]+$/i.match(params[:hash])
  end

  #
  # Check if the user are trying to destroy his own top domain
  #
  def destroying_own_top_domain
    return if @domain != User.current_user.sensor
    error_401('An Admin can\'t delete his own Top domain')
  end

  #
  # Check if the payload is a hash.
  # Note: ActionController::Parameters is new in Rails 5
  #
  def check_params_format
    return error_422('Sensor param must be a Hash') unless
      params[:sensor].is_a?(ActionController::Parameters) || params[:sensor].is_a?(Hash)
  end 

  #
  # Define the params to create a domain.
  #
  def domain_params
    params.permit(:name, :domain_type, :uuid, :mac_hashing_salt,
                  :general_description, :company, :email)
  end

  # Domain params for organizations
  #
  def organization_params
    return domain_params unless current_user.super_admin?
    domain_params.merge(params.permit(:visualization_range, :megabytes_limit))
  end

  
  def register_sensor(type, hash)
    @sensor = Sensor.where(
      type: type
    ).where(
      'property like ?', "%\"mac_address\":\"#{hash}\"%"
    ).first

    if type.blank? || hash.blank?
      render json: { status: '' }, status: :forbidden
    elsif @sensor.nil?
      unclaimed_domain =
        Sensor.find_or_create_by(
          name: 'Unclaimed',
          parent_id: Sensor.root.id,
          type: Sensor::DOMAIN,
          domain_type: Sensor::SERVICE_PROVIDER
        )

      @sensor = Sensor.create(
        name: hash,
        type: type,
        ip: (request.headers['X-Forwarded-For'] ||
             request.headers['X-Real-IP'] ||
             request.remote_ip).split(',').first,
        mac_address: hash,
        unclaimed: true,
        parent_id: unclaimed_domain.id
      )

      render json: { status: 'registered', uuid: @sensor.uuid }
    else
      render json: { status: '' }, status: :forbidden
    end
  end

  # TODO: Catch possible errros
  def verify_sensor(hash, uuid)
    @sensor = Sensor.where(
      uuid: uuid
    ).where(
      'property like ?', "%\"mac_address\":\"#{hash}\"%"
    ).first

    if hash.blank? || uuid.blank?
      render json: { status: '' }, status: :forbidden
    elsif @sensor && @sensor.unclaimed
      @sensor.update(updated_at: Time.now)
      render json: { status: 'registered' }
    elsif @sensor && @sensor.chef_client
      render json: { status: 'claimed' }
    elsif @sensor
      send("verify_sensor_#{@sensor.type}")
    else
      render json: { status: '' }, status: :forbidden
    end
  rescue
    render json: { status: '' }, status: :forbidden
  end

  define_method("verify_sensor_#{Sensor::CLIENT_PROXY}") do
    node_json = {
      'name' => "rbproxy-#{@sensor.id}",
      'run_list' => [@sensor.send("run_list_text_#{@sensor.type}"), "role[#{@sensor.chef_name}]"],
      'attributes' => {
        'rbname' => @sensor.name,
        'ipaddress' => @sensor.ip,
        'redborder' => {
          'monitors' => [],
          'sensor_id' => @sensor.id,
          'sensor_uuid' => @sensor.uuid
        }
      }
    }

    service_provider = @sensor.ancestors.find_by(domain_type: Sensor::SERVICE_PROVIDER)
    deployment       = @sensor.ancestors.find_by(domain_type: Sensor::DEPLOYMENT)
    namespace        = @sensor.ancestors.find_by(domain_type: Sensor::NAMESPACE)
    organization     = @sensor.ancestors.find_by(domain_type: Sensor::ORGANIZATION)
    market           = @sensor.ancestors.find_by(domain_type: Sensor::MARKET)

    %w(service_provider deployment namespace organization market).each do |v|
      node_json['attributes']['redborder']["#{v}"] = eval(v).try('name')
      node_json['attributes']['redborder']["#{v}_uuid"] = eval(v).try('uuid')
    end

    node = Chef::Node.from_hash(node_json)
    node.create

    # create client
    node_name = "rbproxy-#{@sensor.id}"
    client_json = { 'name' => node_name, 'validator' => true }
    @client = Chef::ApiClient.from_hash(client_json)
    @client = @client.save

    # Give to client permissions over the node
    rest = Chef::ServerAPI.new(Chef::Config[:chef_server_url])
    # possible options: %w(read update delete grant)
    %w(update).each do |perm|
      ace = rest.get_rest("nodes/#{node_name}/_acl")[perm]
      ace['actors'] << node_name unless ace['actors'].include?(node_name)
      rest.put_rest("nodes/#{node_name}/_acl/update", perm => ace)
    end

    render json: { status: 'claimed', cert: @client['private_key'].to_json, nodename: node_name }
  end

  define_method("verify_sensor_#{Sensor::EXPORTER}") do
    node_json = {
      'name' => "rbexporter-#{@sensor.id}",
      'run_list' => [@sensor.send("run_list_text_#{@sensor.type}"), "role[#{@sensor.chef_name}]"],
      'attributes' => {
        'rbname' => @sensor.name,
        'ipaddress' => @sensor.ip,
        'redborder' => {
          'monitors' => [],
          'sensor_id' => @sensor.id,
          'sensor_uuid' => @sensor.uuid
        }
      }
    }

    service_provider = @sensor.ancestors.find_by(domain_type: Sensor::SERVICE_PROVIDER)
    deployment       = @sensor.ancestors.find_by(domain_type: Sensor::DEPLOYMENT)
    namespace        = @sensor.ancestors.find_by(domain_type: Sensor::NAMESPACE)
    organization     = @sensor.ancestors.find_by(domain_type: Sensor::ORGANIZATION)
    market           = @sensor.ancestors.find_by(domain_type: Sensor::MARKET)

    %w(service_provider deployment namespace organization market).each do |v|
      node_json['attributes']['redborder']["#{v}"] = eval(v).try('name')
      node_json['attributes']['redborder']["#{v}_uuid"] = eval(v).try('uuid')
    end

    node = Chef::Node.from_hash(node_json)
    node.create

    # create client
    node_name = "rbexporter-#{@sensor.id}"
    client_json = { 'name' => node_name, 'validator' => true }
    @client = Chef::ApiClient.from_hash(client_json)
    @client = @client.save
     
    # Give to client permissions over the node
    rest = Chef::ServerAPI.new(Chef::Config[:chef_server_url])
    # possible options: %w(read update delete grant)
    %w(update).each do |perm|
      ace = rest.get_rest("nodes/#{node_name}/_acl")[perm]
      ace['actors'] << node_name unless ace['actors'].include?(node_name)
      rest.put_rest("nodes/#{node_name}/_acl/update", perm => ace)
    end

    ## Give to client permissions over the node
    #rest = Chef::REST.new(Chef::Config[:chef_server_url])
    ## possible options: %w(read update delete grant)
    #%w(update).each do |perm|
    #  ace = rest.get_rest("nodes/#{node_name}/_acl", false)[perm]
    #  ace['actors'] << node_name unless ace['actors'].include?(node_name)
    #  rest.put_rest("nodes/#{node_name}/_acl/update", perm => ace)
    #end
    
    render json: { status: 'claimed', cert: @client['private_key'].to_json, nodename: node_name }
  end
end
