#
# Implemets redirections for success or error in controllers.
#
module Redirections
  extend ActiveSupport::Concern

  #
  # Redirect on an user to OK and show a notice.
  #
  # == Parameters:
  # notice::
  #   String with notice to show.
  #
  # == Returns:
  # Redirect user to path with a notice message
  #
  def redirect_and_notice(notice, options = {})
    path = options.fetch(:redirect_to, root_path)

    respond_to do |f|
      f.html { redirect_to path, notice: build_for_alert(notice) }
      f.json { render json: { status: 200 } }
    end
  end

  #
  # Redirect the user to a page with an error based on given error.
  #
  # == Parameter:
  # error::
  #   The error to show. It can be a model or String. If this not satisfied,
  #   We will use a default string
  # options::
  #   Hash with options. Available options:
  #     - redirect_to: path to redirect the user. Root_path by default
  #
  # == Returns:
  # Redirect user to path with an error message
  #
  def redirect_and_error(error, options = {})
    path = options.fetch(:redirect_to, root_path)
    messages = fetch_error(error)

    respond_to do |f|
      f.html { redirect_to path, alert: build_for_alert(messages) }
      f.json { render json: { errors: messages } }
    end
  end

  #
  # Redirect user to a path and show a message about privileges.
  #
  # == Parameters:
  # options::
  #   Hash with options. Available options:
  #     - redirect_to: path to redirect the user. Root_path by default
  #
  # == Returns:
  # Redirect user to a path
  #
  def no_enough_privileges(options = {})
    path = options.fetch(:redirect_to, root_path)
    respond_to do |f|
      f.html { redirect_to path, alert: privileges_error }
      f.json { render json: { error: privileges_error }, status: :unauthorized }
      f.js { render json: { error: privileges_error }, status: :unauthorized }
    end
  end

  private

  #
  # Format an String/Array of Strings to show in alerts
  #
  # == Parameters:
  # messages::
  #   String or Array of Strings
  #
  # == Returns:
  # String with HTML code to show
  #
  def build_for_alert(messages)
    messages = Array(messages)
    messages.map { |msg| "<li>#{msg}</li>" }.join
  end

  #
  # Fetch an error message based on given object.
  #
  # == Parameters:
  # error::
  #   Object to get the error
  #
  # == Returns:
  # Array of Strings
  #
  def fetch_error(error)
    return Array(error) if error.is_a? String
    return Array(error.errors.full_messages) if error.respond_to? :errors
    ['Your request has an error. Please check it.']
  end

  #
  # Return the not enough privileges string error.
  #
  # == Returns:
  # String with the value of the error.
  #
  def privileges_error
    "You don't have enough privileges to do that."
  end
end
