#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class DashboardsController < ApplicationController
  ## ======================================================================
  # => Controller properties
  ## ======================================================================

  respond_to :html
  respond_to :js, :json, only: [
    :create, :destroy, :save_composition
  ]

  layout false, only: [:clone, :import, :to_report, :new_user, :to_report,
                       :new_sensor]

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  DittoCode::Exec.if 'ENTERPRISE' do
    before_action :set_dashboard, only: [
      :edit, :destroy, :update, :image, :remove_image,
      :set_background_position, :set_locked_status,
      :clone, :create_clone, :export, :to_report,
      :convert_to_report
    ]
  end

  DittoCode::Exec.if 'COMMUNITY' do
    before_action :set_dashboard, only: [
      :edit, :destroy, :update, :image, :remove_image,
      :set_background_position, :set_locked_status,
      :export, :to_report, :convert_to_report
    ]
  end

  # Checks if the user has permissions to write on these actions
  before_action :check_write_permissions, only: [
    :edit, :update, :set_background_position, :set_locked_status
  ]

  # Checks if the user has permissions over these actions
  before_action :check_read_permissions, only: [
    :convert_to_report, :show, :clone, :create_clone
  ]

  # Checks if the user has at least one dashboard
  before_action :check_dashboards, only: [:index, :show, :overview]

  # Set the sensor for overview
  before_action :set_sensor, :set_domain_zones, :set_overview_range,
                :overview_times, :set_sensor_to_filter, only: :overview

  before_action :dashboard_to_display, only: [:show]

  before_action :dashboards_list, only: [:overview, :show]

  before_action :redirect_to_dashboard, only: :index

  before_action :add_global_filters, only: [:show]

  before_action :check_collaborators, only: [:create, :update]

  # Update dashboard params after creation
  after_action :update_params, only: :create

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    # Set the dashboard and varaibles needed to work with widgets.
    # This is an optimization,
    # @see #4248
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def new
      @dashboard = current_user.dashboards.build
    end

    def create
      @dashboard = Dashboard.new(dashboard_params)
      @dashboard.owner = current_user

      if @dashboard.save
        rb_flash "#{I18n.t('controllers.dashboard_created')}"
        session[:dashboard_id] = @dashboard.id
      end

      respond_with @dashboard, location: @dashboard
    end

    def clone
      @new_dashboard = Dashboard.new
    end

    def create_clone
      @new_dashboard = @dashboard.dup
      @new_dashboard.name = params[:dashboard][:name]
      @new_dashboard.owner = current_user

      if @new_dashboard.save
        rb_flash "#{I18n.t('controllers.dashboard_created')}"
        session[:dashboard_id] = @new_dashboard.id
      end

      redirect_to dashboard_path(@new_dashboard)
    end
  end

  def show
    session[:dashboard_id] = @parent.id
    @manage = @writable_dashboards.include?(@parent)
  end

  def edit
  end

  # An overview of the status of the environment. Here you can see a map with
  # access points and zones defined by user. You will see some metrics based
  # on installed modules.
  #
  def overview
    # Get the time in milliseconds
    @reload =
      (Setting.find_by(name: 'autoreload_time').try(:value) || 180).to_i * 1000
  end

  #
  # POST /dashboard/overview/store_search_type
  #
  # Store overview search type in session
  #
  def overview_store_search_type
    session[:overview_search_type] =
      params[:search_type] if %w(domains places).include?(params[:search_type])
    render nothing: true
  end

  def update
    if update_params
      rb_flash "#{I18n.t('controllers.dashboard_updated')}"
      redirect_to @dashboard
    else
      render action: :edit
    end
  end

  def update_params
    @dashboard.update dashboard_params
  end

  def destroy
    if @dashboard.owner == current_user
      @dashboard.destroy
    elsif current_user.readable_dashboards.include? @dashboard
      dashboard_user = DashboardUser.where(dashboard_id: @dashboard).where(user_id: current_user)
      unless dashboard_user.empty?
        dashboard_user.first.delete
      else
        rb_flash "#{I18n.t('controllers.cant_leave_dashboard')}", type: :error
        redirect_to @dashboard
        return
      end
    end
    redirect_to dashboards_path
  end

  def image
    if @dashboard.layout == 'columns' || @dashboard.image.nil?
      head 200
      return
    end

    Timecop.return do
      uri = @dashboard.image_uri
      img = open(uri, ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE).read
      render text: img, content_type: @dashboard.image.content_type
    end
  end

  def remove_image
    @dashboard.update(image: nil)
  end

  def set_background_position
    %i(x y width height).each do |a|
      @dashboard.img_position[a] = params[a] unless params[a].nil?
    end

    @dashboard.save!
    head 200
  end

  def set_locked_status
    @dashboard.locked = (params[:locked] == 'true')
    @dashboard.save!
    head 200
  end

  def new_user
    @users = current_user.users_to_share
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def new_sensor
      @sensors = if current_user.super_admin?
                   current_user.sensor.children.where(type: Sensor::DOMAIN)
                   .reject { |s| (s.unassigned? && s.children.count.zero?) ||
                                  s.is_root? }
                 else
                   [current_user.sensor]
                 end
    end
  end

  #
  # => Show an advice to convert the report
  #
  def to_report
  end

  #
  # => Convert the dashboard to a report if the user has read permissions.
  # => When it's converted, redirect the user to the dashboard.
  #
  def convert_to_report
    report = @dashboard.to_report
    report.owner = current_user

    if report.save
      redirect_to edit_report_path(report, converted: '')
    else
      rb_flash report.errors, type: :error
      redirect_to @dashboard
    end
  end

  def export
    return error_404('') unless current_user.writable_dashboards.include?(@dashboard)
    send_file @dashboard.export, filename: "#{@dashboard.name}.tar.gz"
  end

  def import
    @dashboard = current_user.dashboards.build
  end

  def save_import
    return unless file_params = import_params
    begin
      @dashboard = Dashboard.import(file_params)
      @dashboard.owner = current_user

      DittoCode::Exec.if 'COMMUNITY' do
      current_user.dashboards.first.destroy
      end

      if @dashboard.save
        session[:dashboard_id] = @dashboard.id
        rb_flash "#{I18n.t('controllers.dashboard_created')}"
      else
        rb_flash @dashboard.errors, type: :error
      end

      redirect_to dashboard_path(@dashboard)
    rescue
      rb_flash "Could not import this dashboard", type: :error
      redirect_to dashboards_path
    end
  end

  private

  def import_params
    redirect_and_error "#{I18n.t('controllers.specify_file')}",
                       redirect_to: dashboards_path if params[:dashboard].nil?
    params[:dashboard].try(:[], :file)
  end

  # Set the sensor to use in overview method
  #
  def set_sensor
    @sensor = Sensor.find_by(uuid: @global_filters[:sensor_uuid]) if @global_filters &&
                                                                     @global_filters[:sensor_uuid]

    if @sensor
      return error_404('') unless current_user.can?(:read, @sensor)
      return
    end

    if params[:id]
      @sensor = Sensor.find(params[:id])
      return error_404('') unless current_user.can?(:read, @sensor)
    else
      @sensor = current_user.sensor
    end
  end

  # Set domain zones based on the sensor
  #
  def set_domain_zones
    if @sensor.building?
      @floors = @sensor.floors.map(&:to_floor_zone)
      @domain_zones = @sensor.to_domain_zone
    else
      @domain_zones = @sensor.children
                             .where(type: Sensor::DOMAIN)
                             .map(&:to_domain_zone)
                             .compact
      next_level_sensors unless @domain_zones.present?
    end
  end

  # Set domain zone of next level sensors and display it as markers.
  # See #7508
  #
  def next_level_sensors
    @domain_zones = []
    @display_as_markers = true
    # Cache parents to prevent it the database
    parents = {}
    Sensor.each_with_level(@sensor.descendants) do |sensor, level|
      # level 2 are the children of children
      next unless level <= 2
      if level == 1
        # Store parent names
        parents[sensor.id] = sensor.name
      elsif (domain = sensor.to_domain_zone)
        # We need to check if domain is not nil because the domain zones of sensor
        # can be nil.
        # Concatename the parent name
        domain[:title] = "#{parents[sensor.parent_id]} / #{sensor.name}"
        @domain_zones << domain
      end
    end
  end

  # Set the sensor to filter in overview
  #
  def set_sensor_to_filter
    @filter_sensor = if @sensor.building? && @floors.present?
                       Sensor.find_by(id: @floors.first[:id]) || @sensor
                     else
                       @sensor
                     end
  end

  # Set the range for overview
  #
  def set_overview_range
    ranges = %w(last_1_hour last_36 last_12)
    @range = if ranges.include?(params[:range])
               params[:range]
             elsif session[:overview_range] &&
                   ranges.include?(session[:overview_range])
               session[:overview_range]
             else
               'last_1_hour'
             end
    # Store in session
    session[:overview_range] = @range
  end

  # Initialize start and end_time for overview
  #
  def overview_times
    @end_time = in_time_machine { Time.now }
    diff = case @range
           when 'last_1_hour' then 1.hour
           when 'last_12' then 12.hours
           when 'last_36' then 36.hours
           end
    # Start_time
    @start_time = @end_time - diff
  end

  def dashboard_to_display
    readable_ids = current_user.readable_dashboards.map(&:id)

    dashboard_id = params[:id].to_i

    @parent = if readable_ids.include? dashboard_id
                Dashboard.find_by(id: dashboard_id)
              else
                current_user.dashboards.find_by(default: true) || current_user.dashboards.first
              end
  end

  #
  # Set the list of dashboards to show, sorted by different criteria
  #
  def dashboards_list
    @user_dashboards = current_user.dashboards
    @dashboards_from_others = current_user.dashboards_from_others_filtered
    @dashboards_from_sensors = current_user.dashboards_from_sensors_filtered
    @writable_dashboards = current_user.writable_dashboards
  end

  def redirect_to_dashboard
    default_dashboard = current_user.dashboards.find_by(default: true)

    if default_dashboard
      redirect_to dashboard_path(default_dashboard)
    else
      redirect_to overview_dashboards_path
    end
  end

  def check_dashboards
    return unless current_user.dashboards.blank?
    current_user.dashboards.create(name: 'Default')
    current_user.reload
  end

  def check_write_permissions
    dashboard = Dashboard.find(params[:id])
    return if current_user.writable_dashboards.include? dashboard
    no_enough_privileges redirect_to: dashboards_path
  end

  def check_read_permissions
    dashboard = Dashboard.find(params[:id])
    return if current_user.readable_dashboards.include? dashboard
    no_enough_privileges redirect_to: dashboards_path
  end

  def set_dashboard
    @dashboard = Dashboard.find(params[:id])
  end

  def dashboard_params
    params.require(:dashboard)
      .permit(:name, :description, :layout,
              :default, :image, :background_color,
              collaborators: [:user_id, :can_edit],
              sensors_collaborators: [:sensor_id, :can_edit])
  end

  def add_global_filters
    return unless params[:globals]
    AddGlobalFilters.call(session, params[:globals])
    redirect_to dashboard_path(params[:id])
  end

  def check_collaborators
    user_collaborators = User.where(id: (params[:dashboard][:collaborators] || []).map(&:values).flatten)
    domain_collaborators = Sensor.where(type: Sensor::DOMAIN,
                                        id: (params[:dashboard][:sensors_collaborators] || {}).values.map { |s| s['sensor_id'] })
    return error_403('') if domain_collaborators.select { |d| current_user.cannot?(:read, d) }.any?
    return error_403('') if (user_collaborators - current_user.users_to_share - [current_user]).any?
  end
end
