#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Works with filters, indepedent of the module. This controller add and delete
# filters.
#
class FiltersController < ApplicationController
  # Storage of the filters
  include Redborder::Filters::Storage::Session

  #
  # Don't render anything, only redirect.
  #
  layout false

  # -------------------------------------------------------------
  # Before and after methods
  # -------------------------------------------------------------

  before_action :set_filters, :time_filters, :set_tab, except: [:multi, :globals, :expanded]

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Add a new filter to current.
  #
  def add
    #
    # A filters can return a hash in the values. This is used to add multiple
    # related filters like sensor monitor (sensor_name, group_name)
    #
    @values = @values.permit!.to_h rescue @values
    if @values.is_a?(Hash)
      # Parse a hash of filters
      @values.each do |_, f|
        f.keys.each do |k|
          @filter.add(k, f[k])
        end
      end
    else
      @filter.add(params[:id], @values)
    end
    save_and_redirect
  end

  #
  # Delete all filters and add a new filter.
  #
  def delete_and_add
    @filter.remove_all
    time_filters
    add
  end

  #
  # Add a multi filter. This action is used to parse fast search filters with
  # multiple products type...
  #
  def multi
    # Iterate over filters
    params[:columns].each_with_index do |column, i|
      # Load filter library
      @source = Redborder::DruidSource.get(source: params[:products][i])
      @filter = Redborder::Filter.new(@source, load_filters)
      @filter.add(column, [params[:values][i]])
      save_filters(@filter.filters)
    end

    # Add time filters
    st = params[:start_times].map(&:to_i).min
    et = [Time.zone.now.to_i, Time.zone.at(st).end_of_day.to_i].min
    # Use last filter to set the time
    @filter.add(:start_time, [st])
    @filter.add(:end_time, [et])

    # Return last tab and redirect.
    save_tab(@source.fetch_tab(params[:columns].last))
    save_and_redirect
  end

  #
  # Add a new contains filter to current.
  #
  def contains
    #
    # A filters can return a hash in the values. This is used to add multiple
    # related filters like sensor monitor (sensor_name, group_name)
    #
    if @values.is_a?(Hash)
      @values.each do |_, f|
        key = f.keys.first
        @filter.contains(key, f[key])
      end
    else
      @filter.contains(params[:id], @values)
    end
    save_and_redirect
  end

  #
  # Add an exclude filter
  #
  def exclude
    #
    # A filters can return a hash in the values. This is used to add multiple
    # related filters like sensor monitor (sensor_name, group_name)
    #
    @values = @values.permit!.to_h rescue @values
    if @values.is_a?(Hash)
      @values.each do |_, f|
        f.keys.each do |k|
          @filter.exclude(k, f[k])
        end
      end
    else
      @filter.exclude(params[:id], @values)
    end
    save_and_redirect
  end

  #
  # Delete a filter from current.
  #
  def delete
    if params[:id] == 'all'
      @filter.remove_all
      save_and_redirect(:minute, :last_1_hour)
    else
      #
      # A filters can return a hash in the values. This is used to add multiple
      # related filters like sensor monitor (sensor_name, group_name)
      #
      if @values.is_a?(Hash)
        @values.each do |_, f|
          key = f.keys.first
          @filter.remove(key, f[key])
        end
      else
        @filter.remove(params[:id], @values)
      end

      save_and_redirect
    end
  end

  #
  # Add a time filter between two dates. This method is called
  # from the date range custom slider.
  #
  def domain
    # Save and returns an empty JS, the client must reload the page
    save_filters(@filter.filters)
    render js: ''
  end

  #
  # Get data from advanced search and store it on filters
  #
  def advanced
    search = filter_allowed(params[:search])
    # Remove all filters and parse it
    @filter.parse(search, true) if params[:search]
    save_and_redirect
  end

  def globals
    AddGlobalFilters.call(session, params[:globals])
    redirect_to request.referer || root_path
  end

  def expanded
    session[:show_simple]   = (params[:simple] == 'true')
    session[:show_advanced] = (params[:advanced] == 'true')
    render nothing: true
  end

  private

  #
  # Set the tab to add and select on redirected view
  #
  def set_tab
    return if params[:tab].nil?
    # Convert params
    save_tab(params[:tab].to_sym)
  end

  #
  # Store the tab in the user session.
  #
  # == Parameters:
  # tab::
  #   Symbol of the tab.
  #
  def save_tab(tab)
    unless session[@source.session][:active_tabs].include?(tab)
      session[@source.session][:active_tabs] << tab
    end
    session[@source.session][:tab] = tab
  end

  #
  # Add time filters if it's exist
  #
  def time_filters
    return if params[:start_time].nil? || params[:end_time].nil?
    # Convert params
    st = params[:start_time].to_i
    et = [Time.zone.now.to_i, params[:end_time].to_i].min

    # Delete previous time_filters and add current
    @filter.remove(:start_time)
    @filter.remove(:end_time)
    @filter.add(:start_time, [st])
    @filter.add(:end_time, [et])
  end

  #
  # Redirect the user to the last action
  #
  # == Parameters:
  # granularity::
  #   Symbol with granularity, if it's nil params[:granularity] is taken
  # range::
  #   Symbol with the range, if it's nil params[:range] is taken
  #
  def save_and_redirect(granularity = nil, range = nil)
    # Save current filters
    save_filters(@filter.filters)
    # Redirect
    action = session[@source.session][:default_action] || @source.redirect
    # When redirect to cardinality we must show the tab and group_by.
    if params[:cardinality_tab]
      redirect_to_cardinality(action)
    else
      redirect_to action.merge(
        granularity: granularity || params[:granularity],
        range: range || params[:range]
      )
    end
  end

  #
  # Redirect the user to cardinality view.
  #
  # == Parameters:
  # action::
  #   Default action to redirect
  #
  def redirect_to_cardinality(action)
    redirect_to action.merge(
      action: 'cardinality',
      granularity: params[:granularity],
      range: params[:range],
      group_by: params[:id],
      tab: params[:cardinality_tab]
    )
  end

  #
  # Load filters an initialize filter and product variable.
  #
  def set_filters
    product = params[:product]
    # Set the values
    @values = params[:values]
    @source = Redborder::DruidSource.get(source: product)
    # TODO: If source is nil, Filter will be throw an exception
    @filter = Redborder::Filter.new(@source, load_filters)
  end

  def filter_allowed(search)
    JSON.parse(search).select do |_k, v|
      v['column'] != 'domain' || (v['column'] == 'domain' && current_user.can?(:read, Sensor.find(v['value'])))
    end.to_json
  end
end
