#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class IncidentsController < ApplicationController
  helper Redborder::ModulesHelper

  include ApplicationHelper

  include IncidentsHelper

  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_incident, only: [:edit, :edit_description, :update, :destroy, :show,
                                      :next_phase, :previous_phase, :close, :edit_report_modal,
                                      :update_report_section, :create_note, :create_response_action_note,
                                      :create_response_action_execute, :update_note, :destroy_note,
                                      :generate_llm_reports, :create_incident_report, :new_user, :search_user,
                                      :assign_user, :unassign_user, :sidebar_content, :new_note, :new_response_action_note,
                                      :new_response_action_execute, :edit_note, :destroy_note, :worklog_search, :detection_search
                                    ]
  before_action :set_sensor_names, only: [:index, :bulk_delete, :bulk_update_status, :update, :destroy, :close,
                                          :bulk_create_links, :search, :show, :create_note, :create_response_action_note,
                                          :create_response_action_execute, :update_note, :destroy_note,:generate_llm_reports, :update_report_section, :statistics,
                                          :assets_sidebar, :get_incidents_widget_data]
  before_action :check_incident_sensor_privileges, only: [:show, :update, :destroy, :close,
                                                          :create_note, :create_response_action_note,
                                                          :create_response_action_execute, :update_note, :destroy_note,
                                                          :generate_llm_reports, :update_report_section]
  before_action :check_admin_privileges, only: [:update, :destroy, :close, :bulk_delete, :bulk_update_status,
                                                :bulk_create_links, :unlink, :assign_user, :unassign_user,
                                                :create_note, :create_response_action_note,
                                                :create_response_action_execute, :update_note,
                                                :destroy_note, :generate_llm_reports, :update_report_section]
  before_action :set_report_incident, only: [:show, :create_incident_report, :edit_report_modal, :update_report_section]
  before_action :set_priorities, only: [:index, :show, :search, :update]
  before_action :set_elements, only: [:detection_search, :show, :create_incident_report]
  before_action :calculate_nodes, only: [:show, :create_incident_report]
  before_action :load_actions_and_worklog_data, only: [:show, :create_incident_report]
  before_action :set_statuses, only: [:index, :show, :search, :update, :create_incident_report]
  before_action :assign_first_user_and_open!, only: [:show]
  before_action :set_users, only: [:show]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all incidents in a table
  #
  def index
    @page = params[:page] || 1

    @total_incidents_count = Incident.where(visible: true, sensor_id: @sensor_names.keys).count

    # Get only the incidents where the sensor_id is among those allowed
    @incidents = Incident.where(visible: true, sensor_id: @sensor_names.keys).order(created_at: :desc)

    # Apply cached filters by reusing search method if it is present
    @cached_filters = Incident.read_filters(current_user.id)
    if @cached_filters.present?
      params[:search] = @cached_filters[:search]
      params[:name] = @cached_filters[:name]
      @incidents = search
    end

    # Apply time machine
    if current_user.time_machine_now
      time_machine = Time.zone.at(current_user.time_machine_now).strftime('%Y-%m-%d %H:%M')
      @incidents = @incidents.where('created_at <= ?', time_machine)
    end

    # To show statistics on index view
    incident_counters

    # Paginate results
    @incidents = @incidents.paginate(page: @page, per_page: current_user.per_page_count)

    @priorities = Priority.all.order(score: :desc)
    @incident_statuses = IncidentStatus.all
  end

  #
  # => Incident statistics
  #
  def statistics
    created_at_condition = if current_user.time_machine_now
      time_machine = Time.zone.at(current_user.time_machine_now)
      ["created_at <= ?", time_machine]
    else
      nil
    end

    # Incindents by priority
    @incident_counts_by_priority = Priority.all.map do |priority|
      {
        name: priority.name.capitalize,
        events: Incident.where(
          visible: true,
          priority_id: priority.id,
          sensor_id: @sensor_names.keys
        ).where(created_at_condition).count
      }
    end.select { |priority_data| priority_data[:events] > 0 }    

    # Incindents by creation date (2 weeks only)
    @start_time = 2.weeks.ago.beginning_of_day
    @end_time = Date.today.end_of_day

    @incident_counts_by_period = (@start_time.to_date..@end_time.to_date).map do |date|
      {
        event: { events: Incident.where(visible: true, created_at: date.beginning_of_day..date.end_of_day, sensor_id: @sensor_names.keys).count },
        timestamp: date.beginning_of_day.utc.strftime("%Y-%m-%d %H:%M:%S UTC")
      }
    end
  
    # Incindents by status
    @incidents_counts_by_status = IncidentStatus.all.map do |statu|
      {
        name: statu.name.capitalize,
        events: Incident.where(visible: true, incident_status_id: statu, sensor_id: @sensor_names.keys)
        .where(created_at_condition).count
      }
    end.select { |status_data| status_data[:events] > 0 }

    # Incindents by source
    @incidents_counts_by_source = Incident.where(visible: true, sensor_id: @sensor_names.keys)
                                      .where(created_at_condition)
                                      .group(:source)
                                      .count
                                      .map do |source, count|
      {
        name: source,
        events: count
      }
    end

    # Incindents by domain
    @incidents_counts_by_domain = Incident.where(visible: true, sensor_id: @sensor_names.keys)
                                          .where(created_at_condition)
                                          .group(:sensor_id)
                                          .count
                                          .map do |sensor_id, count|
      {
        name: @sensor_names[sensor_id] == 'root' ? '/' : @sensor_names[sensor_id],
        events: count
      }
    end

    # Count incidents by assigned user
    @incidents_counts_by_user = Incident.joins(:incident_users)
                                        .where(visible: true, sensor_id: @sensor_names.keys)
                                        .where(created_at_condition)
                                        .where(incident_users: { user_id: User.pluck(:id) })
                                        .group("incident_users.user_id")
                                        .count
                                        .map do |user_id, count|
      {
        name: User.find(user_id).name, 
        events: count
      }
    end
  end

  #
  # => Render the form to create a new incident
  #
  def new
    @incident = Incident.new(priority_id: Priority.first.id)
    @incident.incident_status = IncidentStatus.find_by(name: IncidentStatus::NEW)
  end

  #
  # => Receive the data to create a new incident. Try to create a new incident
  #    and redirect to incidents_path if it's ok
  #
  def create
    @incident = Incident.new(incident_params)

    respond_to do |format|
      if @incident.save
        format.html do
          redirect_to incidents_path,
                      notice: "#{I18n.t('controllers.incident_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_created')}" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do
            |msg| "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @incident.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit
    respond_to do |format|
      format.js
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit_description
    respond_to do |format|
      format.js
    end
  end

  #
  # => Receive the data to update a new incident. Try to update a new incident
  #    and redirect to incident_path if it's ok
  #
  def update
    @page = params[:page] || 1
    original_name = @incident.name
    original_description = @incident.description

    respond_to do |format|
      if @incident.update(incident_params)

        # Delete LLM Generation Record for name if it exists
        if incident_params[:name] && incident_params[:name] != original_name
          LlmGenerationRecords.where(table_name: 'incidents', field_name: 'name', record_id: @incident.id).destroy_all
        end

        # Delete LLM Generation Record for description if it exists
        if incident_params[:description] && incident_params[:description] != original_description
          LlmGenerationRecords.where(table_name: 'incidents', field_name: 'description', record_id: @incident.id).destroy_all
        end

        format.html do
          notice_message = "#{I18n.t('controllers.incident_updated')}"

          if request.referer && request.referer.include?(incident_path(@incident))
            redirect_to incident_path(@incident), notice: notice_message
          else
            redirect_to incidents_path(page: @page), notice: notice_message
          end
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_updated')}" }
        format.json { render json: @incident }
      else
        format.html { render action: 'show' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Destroy an incident
  #
  def destroy
    @incident.destroy
    respond_to do |format|
      format.html { redirect_to incidents_path }
      format.js
      format.json { head :no_content }
    end
  end

  def next_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @next_phase = @current_phase.next_phase
      @next_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def previous_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @previous_phase = @current_phase.previous_phase
      @previous_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def close
    respond_to do |format|
      if @incident.update(incident_status: IncidentStatus.find_by(name: IncidentStatus::CLOSED))
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_status_update')}"
        end
        format.js { flash.now[:notice] = "#{I18n.t('controllers.incident_status_update')}" }
        format.json { render json: @incident }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), alert: "#{I18n.t('controllers.error_incident_status_update')}"
        end
        format.js { flash.now[:error] = @incident.errors.full_messages.join(", ") }
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  def bulk_delete
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)

    if incident_ids.any?
      Incident.where(id: incident_ids).destroy_all
      flash[:notice] = "#{I18n.t('controllers.incident_delete')}"
    else
      flash[:alert] = "#{I18n.t('controllers.incident_no_incidents_delete')}"
    end

    redirect_to incidents_path(page: page)
  end

  def bulk_update_status
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
    status = IncidentStatus.find_by(name: params[:status])
    status_id = status.id if status
    incidents = Incident.where(id: incident_ids)

    if status && incidents.update_all(incident_status_id: status_id)
      flash[:notice] = "#{I18n.t('controllers.incidents_updated')}"
    else
      flash[:alert] = "#{I18n.t('controllers.error_incidents_updated')}"
    end

    redirect_to incidents_path(page: page)
  end

  def bulk_create_links
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)

    if incident_ids.size > 1
      incidents = Incident.where(id: incident_ids)
      successfully_linked = []
      failed_links = []

      incidents.each do |incident|
        remaining_incidents = incidents.where.not(id: incident.id)  # Exclude the current incident

        remaining_incidents.each do |child_incident|
          begin
            unless IncidentLink.link_exists?(incident, child_incident)
              # Create the bidirectional link
              incident.linked_incidents << child_incident
              create_note_link_incident(incident, child_incident, true)
              successfully_linked << "#{incident.id} - #{child_incident.id}"
            end
          rescue ActiveRecord::RecordInvalid => e
            failed_links << "#{incident.id} - #{child_incident.id} (#{e.message})"
          end
        end
      end

      if successfully_linked.any?
        flash[:notice] = "#{I18n.t('controllers.incident_create_link')}: #{successfully_linked.join(", ")}"
      end

      if failed_links.any?
        flash[:alert] = "#{I18n.t('controllers.incident_create_link_fail')}: #{failed_links.join(", ")}"
      end
    else
      flash[:alert] = I18n.t('controllers.incident_create_link_warning')
    end

    redirect_to incidents_path(page: page)
  end

  def unlink
    incident = Incident.find(params[:incident_id]) 
    linked_incident = Incident.find(params[:linked_incident_id])

    # Destroy the original link
    incident.linked_incidents.delete(linked_incident)
    # Destroy the reverse link
    linked_incident.linked_incidents.delete(incident)

    # Create notification and register in worklog
    flash[:notice] = "#{linked_incident.name} #{I18n.t('controllers.incident_unlink')}"
    create_note_link_incident(incident, linked_incident, false)

    redirect_to incident_path(incident)
  end

  #
  # Search on incidents
  #
  def search paginate=true, widget_mode=false
    @page = params[:page] || 1
    @incidents = Incident.where(visible: true)

    # If params[:search] or params[:name] exists, save the current search filters to Memcached
    unless widget_mode
      if params[:search].present? || params[:name].present?
        Incident.create_filters(current_user.id, params[:search], params[:name])
      else
        # Retrieve filters from Memcached if no search params are provided
        cached_filters = Incident.read_filters(current_user.id)
        if cached_filters.present?
          params[:search] = cached_filters[:search]
          params[:name] = cached_filters[:name]
        end
      end
    end

    # Filter by name
    if params[:name].present?
      @incidents = @incidents.where("incidents.name ILIKE ? OR incidents.uuid ILIKE ?", "%#{params[:name]}%", "%#{params[:name]}%")
    end

    if params[:search].present?
      # Filter by date range
      if params[:search][:range].present?
        end_time = start_time = Time.now

        case params[:search][:range].to_sym
        when :last_5_minute then start_time = end_time - 5.minutes
        when :last_15_minute then start_time = end_time - 15.minutes
        when :last_30_minute then start_time = end_time - 30.minutes
        when :last_1_hour then start_time = end_time - 1.hour
        when :last_2_hour then start_time = end_time - 2.hour
        when :last_3_hour then start_time = end_time - 3.hour
        when :last_12_hour then start_time = (end_time - 12.hour).beginning_of_hour
        when :last_24 then start_time = (end_time - 24.hour).beginning_of_hour
        when :last_week then start_time = (end_time - 1.week).beginning_of_hour
        end

        if start_time && end_time
          @incidents = @incidents.where('incidents.created_at BETWEEN ? AND ?', start_time, end_time)

        end
      end

      # Filter by priority
      if params[:search][:priority].present?
        @incidents = @incidents.where(priority_id: params[:search][:priority])
      end

      # Filter by domain
      if params[:search][:domain].present?
        @incidents = @incidents.where(sensor_id: params[:search][:domain])
      end

      # Filter by minimum priority
      if params[:search][:minimum_priority].present?
        selected_priority_id = params[:search][:minimum_priority].to_i
        selected_priority_score = Priority.where(id: selected_priority_id).pluck(:score)
        higher_priority_ids = Priority.where('score > ?', selected_priority_score).pluck(:id)
        @incidents = @incidents.where(priority_id: higher_priority_ids)
      end

      # Filter by source
      if params[:search][:source].present?
        value_to_search = nil
        if params[:search][:source].kind_of?(Array)
          value_to_search = params[:search][:source].map(&:downcase)
        elsif params[:search][:source].kind_of?(String)
          value_to_search = [params[:search][:source]].map(&:downcase)
        end
        @incidents = @incidents.where('LOWER(source) IN (?)', value_to_search)
      end

      # Filter by assigned user
      if params[:search][:assigned].present?
        user_ids = User.where(id: params[:search][:assigned]).pluck(:id)
        @incidents = @incidents.joins(:incident_users)
                               .where(incident_users: { user_id: user_ids })
      end

      # Filter by date
      if params[:search][:start_time].present? && params[:search][:end_time].present?
        start_time = DateTime.strptime(params[:search][:start_time], "%m/%d/%Y %H:%M") rescue nil
        end_time = DateTime.strptime(params[:search][:end_time], "%m/%d/%Y %H:%M") rescue nil

        if start_time && end_time
          @incidents = @incidents.where(created_at: start_time..end_time)
        end
      end

      # Filter by status
      if params[:search][:status].present?
        @incidents = @incidents.where(incident_status_id: params[:search][:status])
      end

      # Exclude incidents by title (not like)
      if params[:search][:exclude].present?
        exclude_query = params[:search][:exclude].split(',').map { |term| "incidents.name NOT ILIKE ?" }.join(' AND ')
        exclude_values = params[:search][:exclude].split(',').map { |term| "%#{term.strip}%" }
        @incidents = @incidents.where(exclude_query, *exclude_values)
      end

      # Handle sorting
      order_by = params[:search][:order_by]
      direction = params[:search][:order_direction] == 'desc' ? 'desc' : 'asc'

    end
    if params[:search][:order_by].present?
      if order_by == 'assigned_user'
        # Join the incident_users and users tables to sort by the users name
        @incidents = @incidents.joins(:incident_users)
                               .joins('LEFT JOIN users ON users.id = incident_users.user_id')
                               .order("users.name #{direction}")
      else
        # For other fields, apply safe sorting
        @incidents = @incidents.order(Incident.arel_table[order_by].send(direction))
      end
    else
      # Fallback to a default order if the column is not allowed
      @incidents = @incidents.order("incidents.created_at DESC")
    end

    if current_user.time_machine_now
      time_machine = Time.zone.at(current_user.time_machine_now).strftime('%Y-%m-%d %H:%M')
      @incidents = @incidents.where('incidents.created_at <= ?', time_machine)
    end

    if paginate
      # Paginate results
      @incidents = @incidents.paginate(page: @page, per_page: current_user.per_page_count)
    else
      @incidents
    end
  end

  def reset_filters
    Incident.clear_filters(current_user.id)
    redirect_to incidents_path
  end

  def search_user
    query = params[:q]

    users = User.where("name LIKE ?", "%#{query}%")

    # Filter users that are not already in the incident
    filtered_users = users
      .where(id: current_user.users_to_assign.select{ |user| user.domains_with_permission.pluck(:id).include?(@incident.sensor_id) }.pluck(:id))
      .where.not(id: @incident.users.pluck(:id))

    render json: filtered_users.map { |user| { id: user.id, name: user.name } }
  end

  #
  # => Show Incident details
  #    (load data on before actions private methods)
  #
  def show
    @incident_links_bidirectional = []

    @detection_unique_sources = @elements.map { |log| log[:incident_source_name] }.uniq
    @detection_unique_priorities = @elements.map { |log| log[:priority] }.uniq

    @incident_nodes = @incident_nodes.uniq
    
    # Create relations
    @incident_nodes.each do |node|
      related_objects = []
      @elements.each do |element|
        next if element[:outlier_event]

        #objects = (element[:assets] + element[:observables]).map(&:keys).flatten
        objects = ((element[:assets]).map {|e| e[:name] } + (element[:observables]).map(&:keys)).flatten
        if objects.include? node[:name]
          related_objects += objects
        end
      end

      # Make the link
      related_objects.uniq.each do |object|
        related_node = @incident_nodes.find { |hash| hash[:name] == object }

        next unless related_node

        @incident_links_bidirectional.push({source: @incident_nodes.find_index(node), target: @incident_nodes.find_index(related_node) ,value: 1})
      end
    end

    @incident_links = []

    @incident_links_bidirectional.each do |link|
      unless @incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        @incident_links << link
      end
    end

    @indicators = load_indicators(@elements)

    @tactics = []
    @techniques = []
    @subtechniques = []
    load_mitre_data(@indicators, @tactics, @techniques, @subtechniques)

    # Define current phase
    unless @incident.playbook.phases.current
      @incident.playbook.phases.order(:id).first.set_as_current
    end

    @current_phase = @incident.playbook.phases.current

    # Paginate results
    @page = params[:page] || 1
    @elements = @elements.paginate(page: @page, per_page: current_user.per_page_count)
  end

  def new_user
    @users = current_user.users_to_assign
    respond_to do |format|
      format.js
    end
  end

  # Assign a user to the incident
  def assign_user
    user = User.find(params[:user_id])
    @incident.users << user

    gravatar_uri = get_gravatar_uri(user.email)

    if @incident.save
      notification = Notification.create(
        user_id: user.id,
        comment: I18n.t('notifications.incident.assigned', incident_id: @incident.id),
        incident_id: @incident.id
      )
      notification.save
      if SyslogService.syslog_integration_configured?
        syslog_message = Rfc5424SyslogMessage.new
        syslog_message.msg_id = 'incident-assignment'
        syslog_message.structured_data.push(@incident.basic_info_structured_data)
        syslog_message.structured_data.push(user.basic_info_structured_data)
        syslog_message.msg = "The user #{user.name} was added to the incident"

        SyslogService.send_to_syslog(syslog_message.to_s)
      end
    end

    render json: { id: @incident.id, user_id: user.id, user_name: user.name, user_email: user.email, gravatar_uri: gravatar_uri }
  end

  def unassign_user
    user = User.find(params[:user_id])

    if @incident.users.delete(user)
      respond_to do |format|
        format.html { redirect_to incident_path(@incident), notice: "#{I18n.t('controllers.incident_remove_user')}" }
        format.js
      end
      if @incident.save
        notification = Notification.create(
          user_id: user.id,
          comment: I18n.t('notifications.incident.unassigned', incident_id: @incident.id),
          incident_id: @incident.id
        )
        notification.save
        if SyslogService.syslog_integration_configured?
          syslog_message = Rfc5424SyslogMessage.new
          syslog_message.msg_id = 'incident-assignment'
          syslog_message.structured_data.push(@incident.basic_info_structured_data)
          syslog_message.structured_data.push(user.basic_info_structured_data)
          syslog_message.msg = "The user \"#{user.name}\" was remove from the incident"

          SyslogService.send_to_syslog(syslog_message.to_s)
        end
      end
    else
      respond_to do |format|
        format.html { redirect_to incident_path(@incident), alert: "#{I18n.t('controllers.error_incident_remove_user')}" }
        format.js
      end
    end
  end

  def sidebar_linked_incidents_content
    @incident = Incident.find(params[:incident_id])

    render partial: 'sidebar_linked_incidents_content', locals: { incident: @incident }
  end

  def sidebar_observable_content
    @observable = params[:observable]
    @observable_links, @incident_nodes = related_assets_observables(params[:incident_id], params[:observable][:name])
    @events = related_events(params[:incident_id], params[:observable][:name], false)

    @assets = RedborderObject.where("user_id = ? OR user_id IS NULL", current_user.id)
                         .where("value = ? OR ip_addresses LIKE ?", @observable[:name], "%#{@observable[:name]}%")

    render partial: 'sidebar_observable_content', locals: { observable: @observable, observable_links: @observable_links, incident_nodes: @incident_nodes, events: @events, assets: @assets }
  end

  def sidebar_content
    @elements = @incident.get_data({})
    @assets = []

    @elements.each do |data|
      @assets += data[:assets] if data[:assets]
    end

    # Sort elements by count
    @assets = @assets.tally.sort_by { |asset, count| -count }

    render partial: 'sidebar_content', locals: { incident: @incident, assets: @assets }
  end

  def assets_sidebar
    asset_id = params[:asset]
    @asset = RedborderObject.find(asset_id)
    @observable_links, @incident_nodes = related_assets_observables(params[:incident_id], @asset.value)
    @events = related_events(params[:incident_id], @asset.value, true)

    render partial: 'asset_details_sidebar', locals: { asset: @asset, observable_links: @observable_links, events: @events, incident_nodes: @incident_nodes}
  end

  def new_note
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_note
    @task = Task.find(params[:task_id])
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_execute
    @task = Task.find(params[:task_id])

    Rails.logger.info "Task is #{@task.id}"

    @action = nil

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue nil
    end

    Rails.logger.info "Action is #{@action}"

    respond_to do |format|
      format.js
    end
  end

  def create_note
    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.incident = @incident

    respond_to do |format|
      if @note.save

        @worklog_note = sort_worklog_data(gather_notes_worklog_data(@incident), :desc).first

        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('controllers.incident_note_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_note_created')}" }
        format.json { head :no_content }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{@note.errors.full_messages.join(", ")}"
        end
        format.js { flash[:error] = @note.errors.full_messages.join(", ") }
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_note
    @task = Task.find(note_params[:task_id])
    @phase = @task.phase

    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.response_action =  @response_action

    respond_to do |format|
      if @note.save
        # TODO: make this atomic
        @response_action = ResponseAction.new
        @response_action.task = @task
        @response_action.user = current_user
        @response_action.notes.append(@note)
        @response_action.save

        @worklog_response_action_note = sort_worklog_data(gather_response_action_notes_worklog_data(@incident), :desc).first
        @action_taken_data = sort_worklog_data(fetch_actions_taken_data(@incident), :desc).first

        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_note_created')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.incident_note_created')}" }
        format.json { head :no_content }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), alert: "#{@note.errors.full_messages.join(", ")}"
        end
        format.js { flash[:error] = @note.errors.full_messages.join(", ") }
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_execute
    @task = Task.find(params[:task_id])

    @action = {}

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue {}
    end

    action_params = {}
    @action.keys.each { |key| action_params[key] = params[key] unless params[key].empty? || params[key].nil? }

    # TODO: make this atomic
    @response_action = ResponseAction.new
    @response_action.task = @task
    @response_action.user = current_user
    @response_action.save

    Delayed::Job.enqueue(
      Redborder::Jobs::ResponseActionJob.new(true, current_user.id, @response_action.id, @task.integration_service, @task.action, action_params),
      priority: 10,
      run_at: Time.now + 2.seconds
    )

     #stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{@response_action.id}%").first

    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "response"), notice: "#{I18n.t('controllers.incident_action_enqueued')}"
      end
      format.js { flash[:notice] = "#{I18n.t('controllers.incident_action_enqueued')}" }
      format.json { head :no_content }
    end
  end

  def edit_note
    @note = Note.find(params[:note_id])

    respond_to do |format|
      format.js
    end
  end

  def update_note
    @note = Note.find(params[:note_id])

    if current_user.id != @note.user_id
      redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{I18n.t('controllers.incident_note_credentials')}"
    end

    respond_to do |format|
      if @note.update(content: note_params[:content])
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('controllers.incident_note_updated')}"
        end
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{@note.errors.full_messages.join(", ")}"
        end
        format.js { flash[:error] = @note.errors.full_messages.join(", ") }
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def destroy_note
    @note = Note.find(params[:note_id])

    if current_user.id != @note.user_id
      redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{I18n.t('controllers.incident_note_credentials')}"
    end

    if @note.destroy
      respond_to do |format|
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "#{I18n.t('controllers.incident_note_destroyed')}"
        end
      end
    else
      format.html do
        redirect_to incident_path(@incident, anchor: "worklog"), alert: "#{@note.errors.full_messages.join(", ")}"
      end
      format.js { flash[:error] = @note.errors.full_messages.join(", ") }
      format.json { render json: @note.errors, status: :unprocessable_entity }
    end
  end

  def worklog_search
    search_params = params[:worklog_search] || {}

    type = search_params[:type].presence || 'All'
    user_id = search_params[:user]
    order = search_params[:order] == 'Newest' ? :desc : :asc

    @worklog_data = fetch_worklog_data(type, @incident)
    @worklog_data = filter_worklog_data_by_user(@worklog_data, user_id) if user_id.present?
    @worklog_data = sort_worklog_data(@worklog_data, order)

    @worklog_data
  end

  def detection_search
    @page = params[:page] || 1
    @elements = @elements.paginate(page: @page, per_page: current_user.per_page_count)
  end

  ## ======================================================================
  # => LLM Generation Content Methods
  ## ======================================================================
  def check_llm_job_status
    section = params[:section]
    incident_id = params[:incident_id]
    running = is_llm_running?(section, incident_id)

    render json: { running: running }
  end

  def generate_llm_reports
    section = params[:section]
    events = @incident.get_data({})

    # Define the function name and content based on the section
    case section
    when 'executive_summary'
      function_name = 'generate_incident_executive_summary'
      content = "\n- Incident title: #{@incident.name}\n- Incident date: #{@incident.created_at}\n Status: #{@incident.incident_status.name}\n- Assigned users: #{@incident.users.map(&:name).join(', ')}\n- Priority: #{@incident.priority.name}\n- Number of events: #{events.count}\n- Reported by: #{@incident.source}"
      functions_params = { content: content }
    when 'event_summary'
      function_name = 'generate_incident_event_summary'
      content = ""
      events.each_with_index do |event, index|
        observables = event[:observables].map { |obs| obs.keys.first }.join(", ")
        content += "Event #{index + 1}:\n"
        content += "Message: #{event[:msg]}\n" if event[:msg]
        content += "Observables: #{observables}\n" if observables.present?
        content += "Source Address: #{event[:url_filters][:src]}\n" if event.dig(:url_filters, :src)
        content += "Destination Address: #{event[:url_filters][:dst]}\n" if event.dig(:url_filters, :dst)
      end
      functions_params = { content: content }
    when 'impact'
      function_name = 'generate_incident_impact'
      content = ""
      indicators = load_indicators(events)
      indicators.each_with_index do |indicator, index|
        content += "Indicator #{index + 1}:\n"
        content += "Message: #{indicator[:msg]}\n" if indicator[:msg]
      end
      functions_params = { content: content }
    when 'mitigation'
      function_name = 'generate_incident_mitigation'
      content = ""
      indicators = load_indicators(events)
      indicators.each_with_index do |indicator, index|
        content += "Indicator #{index + 1}:\n"
        content += "Message: #{indicator[:msg]}\n" if indicator[:msg]
      end
      functions_params = { content: content }
    else
      redirect_to incident_path(@incident, anchor: "report"), alert: "Invalid section"
      return
    end

    # Define the model params
    model_params = {
      model_name: 'IncidentReports',
      name_field: section,
      record_id: @incident.incident_reports.id
    }
    # Enqueue the LLM job
    Delayed::Job.enqueue(
      Redborder::Jobs::LLMGenerateContentJob.new(true, current_user.id, function_name, functions_params, model_params),
      priority: 10,
      run_at: Time.now
    )

    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "report"), notice: "#{I18n.t('controllers.incident_generate_llm_' + section)}"
      end
    end
  end

  ## ======================================================================
  # => Report methods
  ## ======================================================================

  #
  # => Create PDF Report
  #
  def create_incident_report
    @time = Time.now
    @indicators = load_indicators(@elements)

    respond_to do |format|
      format.pdf do
        grover = Grover.new(
          render_to_string(
            template: "incidents/incident_report.pdf.erb",
            locals: { incident_report: @incident_report }
          ),
          orientation: "Portrait",
          margin: { top: 20, bottom: 20, left: 20, right: 20 },
          format: "A4",
          scale: 0.8,
          print_backgroud: true
        )

        # Generate the PDF data
        pdf_data = grover.to_pdf

        Tempfile.open(['temp', '.pdf']) do |file|
          file.binmode
          file.write(pdf_data)
          file.rewind
          create_incident_historical(@incident_report.id, file)
        end

        send_data pdf_data, filename: "#{@incident.name}.pdf",
        type: 'application/pdf',
        disposition: 'inline'
      end
    end
  end

  def create_incident_historical(incident_id , file)
    @incident_reports_historical = IncidentReportsHistorical.new
    @incident_reports_historical.incident_reports_id = incident_id
    @incident_reports_historical.generated_by = current_user.name
    @incident_reports_historical.user_id = current_user.id
    @incident_reports_historical.pdf_file = file
    if @incident_reports_historical.save
      Rails.logger.info "Incident Report: '#{incident_id}' generated."
    else
      Rails.logger.error "Incident Report '#{incident_id}' couldn't be generated."
    end
  end

  #
  # => Render edit report modal
  #
  def edit_report_modal
    @section = params[:report_section]

    # Fetch the current content of the section from @incident_report
    @section_content = @incident_report[@section]

    respond_to do |format|
      format.js
    end
  end

  #
  # => Update report section (post)
  #
  def update_report_section
    section = params[:section]
    content = params[:content]
    original_content = @incident_report[section]

    # Update the corresponding section in the incident report
    respond_to do |format|
      if @incident_report.update(section => content)
        # Delete LLM Generation Record for the section if the content has changed
        if content && content != original_content
          LlmGenerationRecords.where(table_name: 'incident_reports', field_name: section, record_id: @incident_report.id).destroy_all
        end

        format.html do
          redirect_to incident_path(@incident, anchor: "report"), notice: "#{I18n.t('controllers.edit_report')}"
        end
      else
        format.js { render status: :unprocessable_entity }
      end
    end
  end

  def get_incidents_widget_data
    params[:search] ||= {}

    param_keys = {
      range: :range,
      priority: :priorities,
      status: :status,
      domain: :domains,
      source: :sources,
      assigned: :users,
    }

    param_keys.each do |search_key, param_key|
      params[:search][search_key] = params[param_key] unless params[param_key].nil?
    end

    params[:search] = params[:search].permit!
    search(false, true)

    filtered_incidents = Incident.where(
      visible: true, sensor_id: @sensor_names.keys).order(created_at: :desc)
    filtered_incidents = filtered_incidents.where(
      sensor_id: params[:search][:domain]) if params[:search][:domain].present?

    render json: get_widget_data(filtered_incidents)
  end

  private

  # Set @sensor_names for all actions
  def set_sensor_names
    @sensor_names = current_user.domains_with_permission.map { |d| [d.id, d.path_name] }.to_h
  end

  #
  # => Init the incident if the user has permissions
  #
  def set_incident
    @incident = Incident.find(params[:id])
  end

  def set_priorities
    @priorities = Priority.all.order(score: :desc)
  end

  def set_statuses
    @incident_statuses = IncidentStatus.all
  end

  # Those are the users the current_user can manage to add into the incident
  def set_users
    @users = current_user.users_to_share
  end

  def check_admin_privileges
    current_user.admin?
  end

  def set_elements
    search_params = params[:detection_search] || {}
    search_params = search_params.reject { |key, value| value.blank? }
    @elements = @incident.get_data(search_params)

    if !@incident.list_linked_incidents.empty?
      @incident.list_linked_incidents.each do |linked_incident|
        @elements.concat(linked_incident.get_data(search_params))
      end
    end
  end

  def calculate_nodes
    @incident_nodes = []
    @timestamps = []

    process_assets(@elements)
    process_observables(@elements)
    process_timestamps(@elements)

    @assets = group_and_count(@incident_nodes, :asset)
    @observables = group_and_count(@incident_nodes, :observable)
  end

  def process_assets(elements)
    elements.each do |element|
      next if element[:assets].blank?
      timestamp = element[:timestamp]

      element[:assets].each do |asset_element|
        existing_node = find_existing_node(asset_element[:name], :asset)

        if existing_node
          add_timestamp(existing_node, timestamp)
        else
          @incident_nodes.push({ 
            name: asset_element[:name],
            type: asset_element[:type],
            value: asset_element[:value],
            id: asset_element[:id],
            image: get_asset_image(asset_element[:type], view_context),
            asset: true,
            timestamps: [timestamp]
          })
        end
      end
    end
  end

  def process_observables(elements)
    elements.each do |element|
      next if element[:observables].blank?
      timestamp = element[:timestamp]

      element[:observables].each do |observable_element|
        observable_element.each do |key, value|
          existing_node = find_existing_node(key, :observable)

          if existing_node
            add_timestamp(existing_node, timestamp)
          else
            @incident_nodes.push({
              name: key,
              type: value,
              image: get_observable_image(value, view_context),
              observable: true,
              timestamps: [timestamp]
            })
          end
        end
      end
    end
  end

  def find_existing_node(name, type_key)
    @incident_nodes.find { |node| node[:name] == name && node[type_key] }
  end

  def add_timestamp(node, timestamp)
    node[:timestamps] << timestamp unless node[:timestamps].include?(timestamp)
  end

  def process_timestamps(elements)
    elements.each do |element|
      timestamp = element[:timestamp]
      @timestamps << timestamp
    end

    @timestamps = @timestamps.sort
  end
  
  def group_and_count(nodes, type)
    grouped_nodes = nodes.select { |node| node[type] == true }.group_by { |e| e }
    grouped_nodes.map { |key, value| { key: key, count: key[:timestamps].size } }.sort_by { |e| -e[:count] }
  end

  def load_indicators(elements)
    @indicators = Hash.new(0)

    grouped_elements = elements.group_by { |event| [event[:msg], event[:incident_source_name], event[:sig_id], event[:classification]] }

    @indicators = grouped_elements.map do |(msg, incident_source_name, sig_id, classification), entries|
      indicator = {
        incident_source_name: incident_source_name,
        msg: msg,
        classification: classification,
        count: entries.size,
        sig_id: sig_id
      }
      indicator
    end.compact

    @indicators = @indicators.sort_by { |entry| -entry[:count] }
  end

  def load_actions_and_worklog_data
    @actions = gather_all_actions(@incident)
    @worklog_data = sort_worklog_data(gather_all_worklog_data(@incident), :desc)
    @actions_taken_data = fetch_actions_taken_data(@incident)
    @actions_taken_data = sort_worklog_data(@actions_taken_data, :desc)
  end

  def load_mitre_data(indicators, tactics, techniques, subtechniques)

    indicators.each do |indicator|
      next if indicator[:sig_id] == "N/A"

      sid = indicator[:sig_id]
      tactics.concat(find_mitre_data(sid, :tactic)).uniq! { |t| t.tactic_id }
      techniques.concat(find_mitre_data(sid, :technique)).uniq! { |t| t.technique_id }
      subtechniques.concat(find_mitre_data(sid, :subtechnique)).uniq! { |t| t.subtechnique_id }
    end
  end

  def find_mitre_data(sid, type)
    case type
    when :tactic
      Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_tactic = ?", [sid], true)
    when :technique
      Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_technique = ?", [sid], true)
    when :subtechnique
      Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_subtechnique = ?", [sid], true)
    else
      []
    end
  end

  def set_report_incident 
    @incident_report = @incident.incident_reports
  end
  
  def get_widget_data incidents
    data = []

    case params[:type]
    when 'status'
      incident_status_ids = incidents.pluck(:incident_status_id)
      names_by_id = IncidentStatus.where(id: incident_status_ids).pluck(:id, :name).to_h
      data = incident_status_ids.map { |id| names_by_id[id] }
    when 'priority'
      incident_priorities_ids = incidents.pluck(:priority_id)
      names_by_id = Priority.where(id: incident_priorities_ids).pluck(:id, :name).to_h
      data = incident_priorities_ids.map { |id| names_by_id[id] }
    when 'assigned'
      assigned_users_ids = IncidentUser.where(incident_id: incidents.pluck(:id)).pluck(:user_id)
      names_by_id = User.where(id: assigned_users_ids).pluck(:id, :name).to_h
      data = assigned_users_ids.map { |id| names_by_id[id] }
    when 'source'
      data = incidents.pluck(:source)
    when 'domain'
      sensors = incidents.pluck(:sensor_id)
      sensors_by_id = Sensor.where(id: sensors, type: '1').pluck(:id, :name).to_h
      data = sensors.map { |id| sensors_by_id[id] }
    when 'playbooks'
      incidents = incidents.includes(:playbook)
      data = incidents.map { |incident| incident.playbook.name }
    when 'table' #list
      incident_priorities_ids = incidents.pluck(:priority_id)
      names_by_id = Priority.where(id: incident_priorities_ids).pluck(:id, :name).to_h
      data = incidents.map do |incident|
        # Asignar el nombre de la prioridad según el priority_id
        incident.attributes.merge(
          'priority_name' => names_by_id[incident.priority_id],
          'priority_color' => priority_color(names_by_id[incident.priority_id]),
          'priority_bgcolor' => priority_bgcolor(names_by_id[incident.priority_id])
        )
      end

    else
      Rails.logger.info "Unknown type: #{params[:type]} trying to create a widget"
    end

    combined_data = data.each_with_object({}) do |entry, hash|
      name = entry
      if hash.key?(name)
        hash[name][:events] += 1
      else
        hash[name] = {name: entry, events: 1}
      end
    end

    final_data = combined_data.values
    final_data.sort_by! { |entry| -entry[:events] }
    final_data
  end

  # Assign firs user that visits a new incident and open it
  def assign_first_user_and_open!
    if @incident.users.empty? && current_user && @incident.new?
      @incident.users.push(current_user)
      if @incident.save
        notification = Notification.create(
          user_id: current_user.id,
          comment: I18n.t('notifications.incident.assigned', incident_id: @incident.id),
          incident_id: @incident.id
        )
        notification.save
        unless @incident.open!
          Rails.logger.error "Failed to open the incident: #{@incident.errors.full_messages.join(', ')}"
        end
      else
        Rails.logger.error "Failed to save the incident: #{@incident.errors.full_messages.join(', ')}"
      end
    end
  end

  def gather_all_actions(incident)
    response_actions = incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions)
    all_actions = (response_actions + response_actions).sort_by(&:created_at)
    all_actions
  end

  def gather_all_worklog_data(incident)
    incident_changes = gather_incident_audit_worklog_data(incident)
    notes = gather_response_action_notes_worklog_data(incident) + gather_notes_worklog_data(incident)
    executions = gather_response_action_executions_worklog_data(incident)
    all_worklog_data = (incident_changes + notes + executions)
    all_worklog_data
  end

  def gather_incident_audit_worklog_data(incident)
    incident.audits.map do |audit|
      formatted_content = audit.comment

      if audit.audited_changes['incident_status_id']
        # Obtener los valores antiguos y nuevos del estado
        old_status_id, new_status_id = audit.audited_changes['incident_status_id']

        # Buscar los nombres de los estados antiguos y nuevos
        old_status = IncidentStatus.find_by(id: old_status_id).try(:name) ? IncidentStatus.find_by(id: old_status_id).name : old_status_id
        new_status = IncidentStatus.find_by(id: new_status_id).try(:name) ? IncidentStatus.find_by(id: new_status_id).name : new_status_id

        # Reemplazar el texto en el comentario
        # TODO: refactor this to cover all languages..
        # English
        formatted_content = formatted_content.gsub(/incident_status_id from '(\d+)' to '(\d+)'/) do
          "status from '#{old_status}' to '#{new_status}'"
        end
        # Spanish
        formatted_content = formatted_content.gsub(/incident_status_id desde '(\d+)' a '(\d+)'/) do
          "estado desde '#{old_status}' a '#{new_status}'"
        end
      end

      {
        action: audit.action,
        user: audit.user,
        created_at: audit.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: formatted_content
      }
    end
  end

  def gather_response_action_executions_worklog_data(incident)
    incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map do |response_action|
      if response_action.notes.any?
        []
      else
        stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{response_action.id}%").first
        stored_job_id = stored_job.present? ? stored_job.id : nil
        [{
          action: "Executed Action",
          user: response_action.user,
          created_at: response_action.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
          source: "[RESPONSE TASK] #{response_action.task.name}",
          stored_job_id: stored_job_id
        }]
      end
    end.compact
  end

  def gather_response_action_notes_worklog_data(incident)
    return incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map(&:notes).map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        source: "[RESPONSE TASK] #{note.response_action.task.name}",
        content: note.content,
        type: "note",
        note_id: note.id
      }
    end
  end

  def gather_notes_worklog_data(incident)
    return incident.notes.map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: note.content,
        type: "note",
        note_id: note.id
      }
    end
  end

  def fetch_worklog_data(type, incident)
    case type
    when 'Notes'
      (gather_response_action_notes_worklog_data(incident) +  gather_notes_worklog_data(incident))
    when 'Incident Changes'
      gather_incident_audit_worklog_data(incident)
    when 'Response Actions'
      gather_response_action_notes_worklog_data(incident) + gather_response_action_executions_worklog_data(incident)
    else
      gather_all_worklog_data(incident)
    end
  end

  def format_actions_taken_note(note)
    action = note[:action]
    if note[:action] == "Create"
      action = "Added a note for"
    end
    user = note[:user]
    source = note[:source].sub("[RESPONSE TASK] ", "")
    content = note[:content]
    created_at = note[:created_at]

    { action: action, user: user, created_at: created_at, source: source, content: content }
  end

  def format_actions_taken_execution(execution)
    action = execution[:action]
    if execution[:action] == "Executed Action"
      action = "Executed Action for "
    end
    user = execution[:user]
    source = execution[:source].sub("[RESPONSE TASK] ", "")
    # content = execution[:content]
    created_at = execution[:created_at]
    stored_job_id = execution[:stored_job_id]

    { action: action, user: user, created_at: created_at, source: source, stored_job_id: stored_job_id }
  end

  def fetch_actions_taken_data(incident)
    data = []
    notes = gather_response_action_notes_worklog_data(incident)
    notes.each do |note|
      data.push(format_actions_taken_note(note))
    end

    executions = gather_response_action_executions_worklog_data(incident)

    executions.each do |execution|
      data.push(format_actions_taken_execution(execution))
    end

    data
  end

  def filter_worklog_data_by_user(worklog_data, user_id)
    user = User.find_by(id: user_id)
    return worklog_data unless user

    worklog_data.select { |item| item[:user]['id'] == user.id rescue false }
  end

  def sort_worklog_data(worklog_data, order)
    date_format = "%B %d, %Y %I:%M:%S %p"

    converted_data = worklog_data.map do |item|
      item[:created_at_datetime] = DateTime.strptime(item[:created_at], date_format)
      item
    end

    sorted_data = converted_data.sort_by { |item| item[:created_at_datetime] }

    sorted_data.reverse! if order == :desc

    sorted_data.each { |item| item.delete(:created_at_datetime) }

    sorted_data
  end

  # Dynamic mapping method
  def map_dynamic_fields(data, field_map)
    field_map.each do |key, possible_fields|
      possible_fields.each do |possible_field|
        if data.methods.include?(possible_field)
          value = data.send(possible_field)
          return value unless value.nil?
        end
      end
    end
    'N/A'
  end

  # TODO: each module should define this
  def get_module_source_name(m)
    case m
    when :ips
      "redBorder Intrusion"
    when :vault
      "redBorder Vault"
    else
      m
    end
  end

  #
  # => Return a hash with the params to create/update an incident
  #
  def incident_params
    params
      .require(:incident)
      .permit(:name, :source, :description, :status, :user,
              :priority_id, :incident_status_id, :page, :search, :fast_search)
  end

  def note_params
    params.require(:note).permit(:id, :content, :task_id)
  end

  # Build a GRAFO that represents the relation's diagram in table form
  def related_assets_observables(incident_id, observable_name)
    incident = Incident.find(incident_id)
    elements = incident.get_data({})
    incident_nodes = []
    incident_links_bidirectional = []
    # Extract observable nodes from the incident data
    elements.each do |element|
      if element[:observables] && !element[:observables].empty?
        element[:observables].each do |observable_element|
          observable_element.each do |key, value|
            incident_nodes.push({ name: key, type: value, image: get_observable_image(value), observable: true })
          end
        end
      end
      if element[:assets] && !element[:assets].empty?
        element[:assets].each do |asset_element|
          incident_nodes.push({ name: asset_element[:mac_address], obj_name: asset_element[:name], type: "MacObject", image: get_asset_image("MacObject"), asset: true })
        end
      end
    end

    # Remove duplicate nodes
    incident_nodes = incident_nodes.uniq

    # Build bidirectional links between nodes
    incident_nodes.each do |node|
      related_objects = []

      elements.each do |element|
        
        asset_objects = element[:assets].map { |asset| asset[:mac_address] }.compact
        observable_objects = element[:observables].map(&:keys).flatten
      
        objects = (asset_objects + observable_objects)
      
        if objects.include?(node[:name])
          related_objects += objects
        end
      end

      related_objects.uniq.each do |object|
        related_node = incident_nodes.find { |hash| hash[:name] == object }
        next unless related_node

        incident_links_bidirectional.push({
          source: incident_nodes.find_index(node),
          target: incident_nodes.find_index(related_node),
          value: 1
        })
      end
    end

    # Ensure that each link is unique and does not duplicate the reverse direction
    incident_links = []

    incident_links_bidirectional.each do |link|
      unless incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        incident_links << link
      end
    end

    # Filter links involving the specified observable
    observable = incident_nodes.find { |node| node[:name] == observable_name }
    observable_links = incident_links.select do |link|
      link[:source] == incident_nodes.find_index(observable) || link[:target] == incident_nodes.find_index(observable)
    end

    [observable_links, incident_nodes]
  end

  def related_events(incident_id, observable_name, is_asset)
    incident = Incident.find(incident_id)
    elements = incident.get_data({})

    unless is_asset
      filtered_elements = elements.select do |element|
        element[:observables].any? do |observable|
          observable.include?(observable_name)
        end
      end
    else
      filtered_elements = elements.select do |element|
        element[:assets].any? do |asset|
          asset[:mac_address] == observable_name
        end
      end
    end

    filtered_elements
  end

  ## ======================================================================
  # => Incident links methods
  ## ======================================================================

  def create_note_link_incident(incident, linked_incident, is_create)
    if is_create
      # Create note for the original link
      Note.create(
        incident_id: incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_create_link_note')}:
      <a href='/incidents/#{linked_incident.id}'>#{linked_incident.id}</a>.
      </div>",
      )

      # Create note for the reverse link
      Note.create(
        incident_id: linked_incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_create_link_note')}:
      <a href='/incidents/#{incident.id}'>#{incident.id}</a>.
      </div>",
      )
    else
      Note.create(
        incident_id: incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_unlink_note')}:
      <a href='/incidents/#{linked_incident.id}'>#{linked_incident.id}</a>.
      </div>",
      )

      Note.create(
        incident_id: linked_incident.id,
        user_id: User.current_user.id,
        content: "<div>#{I18n.t('controllers.incident_unlink_note')}:
      <a href='/incidents/#{incident.id}'>#{incident.id}</a>.
      </div>",
      )
    end
  end

  def check_incident_sensor_privileges
    if !@sensor_names.include?(@incident.sensor.id)
      respond_to do |format|
        format.html do
          redirect_to root_path,
                      alert: "#{I18n.t('controllers.no_privileges')}"
        end
        format.json do
          render json: {error: "#{I18n.t('controllers.no_privileges')}"},
                 status: :unauthorized
        end
        format.js { head 403 }
      end
    end

    true
  end

  def incident_counters
    created_at_condition = if current_user.time_machine_now
      time_machine = Time.zone.at(current_user.time_machine_now)
      ["created_at <= ?", time_machine]
    else
      nil
    end

    @all_visible_incidents_count = Incident.where(visible: true, sensor_id: @sensor_names.keys).where(created_at_condition).count

    @incidents_lh_count = Incident.where(visible: true, created_at: 1.hour.ago..Time.current, sensor_id: @sensor_names.keys).where(created_at_condition).count

    @new_incidents_count = Incident.where(visible: true, incident_status_id: 5, sensor_id: @sensor_names.keys).where(created_at_condition).count

    @open_incidents_count = Incident.left_joins(:incident_users)
                                .where(visible: true, incident_users: { incident_id: nil }, sensor_id: @sensor_names.keys)
                                .where(created_at_condition ? [created_at_condition[0].gsub('created_at', 'incidents.created_at'), created_at_condition[1]] : nil).count
  end
end
