#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class LicensesController < ApplicationController
  before_action :require_administrative_privileges
  before_action :set_sensor
  before_action :check_privileges

  def index
    @start_time = params[:month] ? Time.parse(params[:month]) : Time.now.utc.beginning_of_month
    @end_time = @start_time.next_month

    @resume = License.resume(@sensor)
    @licenses = @sensor.licenses
    @consumption = SensorMonitor.consumption_data(sensor: @sensor, start_time: @start_time, end_time: @end_time)
  end

  def new
    render layout: false
  end

  def create
    file_data = params[:license][:file]
    license = License.new_or_update(encoded_license: file_data.read)

    if license.sensor != @sensor
      redirect_to sensor_licenses_path(@sensor), alert: t('licenses.errors.wrong_organization')
    elsif license.save
      redirect_to sensor_licenses_path(@sensor)
    else
      redirect_and_error license, redirect_to: sensor_licenses_path(@sensor)
    end
  rescue License::InvalidFormat
    redirect_to sensor_licenses_path(@sensor), alert: t('licenses.errors.invalid_format')
  end

  def destroy
    license = License.find(params[:id])

    if license.sensor != @sensor
      redirect_to sensor_licenses_path(@sensor), alert: t('licenses.errors.wrong_organization')
    else
      license.destroy
      redirect_to sensor_licenses_path(@sensor)
    end
  end

  def create_request
    name = params[:license][:name]
    email = params[:license][:email]
    megabytes_limit = params[:license][:megabytes_limit]
    sensors = params[:license][:type]

    LicenseMailer.request_email(Setting.cluster_uuid, name, email, megabytes_limit, sensors, @sensor).deliver_now
    redirect_to sensor_licenses_path(@sensor), notice: t('licenses.request_sent')
  end

  def new_renew
    @license = License.find(params[:id])
    render layout: false
  end

  def create_renew
    @license = License.find(params[:license][:id])

    name = params[:license][:name]
    email = params[:license][:email]
    megabytes_limit = params[:license][:megabytes_limit]
    sensors = params[:license][:type]

    LicenseMailer.renew_email(@license.uuid, Setting.cluster_uuid, name, email, megabytes_limit, sensors, @sensor).deliver_later
    redirect_to sensor_licenses_path(@sensor), notice: t('licenses.renew_sent')
  end

  def export_consumption_to_file
    @start_time = Time.parse(params[:month])
    @end_time = @start_time.next_month

    csv_info = Hash.new { |h, k| h[k] = { bytes: 0 } }

    SensorMonitor.consumption_data(
      sensor: @sensor, start_time: @start_time, end_time: @end_time
    ).each do |c|
      csv_info[c[:timestamp].to_time.to_i][:bytes] = c[:event][:events]
    end

    namespace_uuids = @sensor.namespaces.map(&:uuid)

    modules = Redborder::Module.keys_with_interface.map(&:to_s)
    modules.delete('monitor')
    modules.delete('location')

    modules.each do |m|
      source = Redborder::DruidSource.get(source: m.to_s)
      source_filter = Redborder::Filter.new(source)

      source_filter.add(:organization_uuid, @sensor.uuid) if RedBorder::LICENSING_MODE == 'organization'

      source.where(
        query_type: :cardinality,
        namespaces: namespace_uuids,
        ignore_user_namespaces: true,
        cardinality_column: :sensor_uuid,
        range: :custom,
        start_time: @start_time,
        end_time: @end_time,
        granularity: :pt24h,
        search: source_filter.get
      ).each do |e|
        csv_info[e.timestamp.to_i][m.to_sym] = e.result
      end
    end

    filename = "Consumption_#{@start_time}.csv"
    content = to_csv(csv_info, modules)
    send_data content, filename: filename
  end

  private

  def set_sensor
    @sensor = if params[:sensor_id]
                Sensor.find(params[:sensor_id])
              else
                current_user.sensor
              end
    not_found unless (RedBorder::LICENSING_MODE == 'global' && @sensor.is_root?) || (RedBorder::LICENSING_MODE == 'organization' && @sensor.organization?)
  end

  def check_privileges
    if @sensor.is_root? && !current_user.super_admin?
      no_enough_privileges
    elsif @sensor.organization? && current_user.cannot?(:manage, @sensor)
      no_enough_privileges
    end
  end

  def to_csv(content, modules)
    csv = %w(Date Bytes)
    modules.each do |m|
      csv << "#{m} sensors"
    end

    csv = csv.join(',') + "\n"

    content.each do |timestamp, values|
      csv += "#{Time.at(timestamp).utc},#{values[:bytes]}"
      modules.each do |m|
        csv += ",#{values[m.to_sym]}"
      end
      csv += "\n"
    end

    csv
  end
end
