#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'zk'
require 'getopt/std'

#
# => Show the info and configuration of managers
#
class ManagersController < ApplicationController
  # Include ManagerHelper
  include ManagersHelper
  include Redborder::Namespaces

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :require_super_reader_privileges_manager

  before_action :require_super_administrative_privileges_manager,
                only: [
                  :save_cluster_synchronization,
                  :save_virtual_ip_addresses, :save_table_data,
                  :save_rules, :reassign_partitions,
                  :up_pcp, :recovery_pcp, :update_namespace_rules
                ]
  before_action :historical_nodes, only: [:save_table_data]
  DittoCode::Exec.if 'ENTERPRISE' do
    before_action :synchronization_options, only: :save_cluster_synchronization
  end

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all managers in a table.
  #    It's showed inside a tab
  #
  def index
    @managers = Manager.all.sort_by(&:name)
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def historical_nodes
      @chef_nodes = Tier.chef_historical_tiers_names
      @tiers = Tier.all.order(:name)
      # Add historicals based on chef values.
      Redborder::Cluster.instance.historicals.each do |historical|
        # Get the tier from Chef
        ctier = @chef_nodes[gsub_cdomain(historical.host)]
        ctier = ctier.nil? ? historical.tier : ctier
        tier = @tiers.find { |t| t.name == ctier }
        next if tier.nil?
        tier.historicals << historical
      end
    rescue StandardError => e
      @info = nil
      Rails.logger.error e
    end

    #
    # Return the rules of a namespace.
    #
    def namespaces
      sensor_uuids = Sensor.where(type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE).limit(20).map(&:uuid)
      @namespaces = ['Default', 'Unassigned'] + sensor_uuids
      @offset = sensor_uuids.count
    end

    def load_more_namespaces
      @offset = params[:offset].to_i
      @more_children = Sensor.where(type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE)
                        .offset(@offset).limit(20).map(&:uuid)
    end

    #
    # Return the rules of a given namespace. The way to return rules of this
    # method is:
    #
    # 1. Return all datasources from /datasources endpoint and select the
    #    elements that coincide with the namespace. This endpoint return
    #    the datasource WITH DATA.
    # 2. Check if there are a datasource present. If it's true, return the
    #    rules of that datasource
    # 3. If datasources array is empty, get the first topic defined in
    #    rbdruid_config.yml (we need it to create a DataSource entity),
    #    append _NAMESPACE and try to these rules. In this case, the datasource
    #    is empty.
    #
    def namespace_rules
      @namespace = params[:id]
      cluster = Redborder::Cluster.instance
      @tiers = cluster.tiers.map { |tier| Tier.default_tier_name(tier.name) }
      if @namespace == 'Default'
        @rules = cluster.default_datasource.rules
      else
        real_namespace = @namespace == 'Unassigned' ? '' : "_#{@namespace}"
        # 1. Get datasources with data
        datasources =
          cluster.datasources.select do |el|
            namespace_name(el.name) == @namespace
          end
        if datasources.present?
          # 2. Return rules
          @rules = datasources.first.rules
        else
          # 3. Get a topic from rbdruid config and try to get rules
          topic =  RedBorder::DRUID_CONFIG.map { |_, v| v['datasource'] rescue nil }.compact.first
          datasource = DruidConfig::Entities::DataSource.new(
            { 'name' => "#{topic}#{real_namespace}", 'properties' => '{}' }, {})
          @rules = datasource.rules
        end
      end
    rescue DruidConfig::Exceptions::NotAvailableNodes
      @not_available_nodes = true
    end

    #
    # Update Rules for all namespace. A namespace can include multiple
    # datasources.
    #
    # This method applies rules to all datasources defined in rbdruid_config.
    #
    def update_namespace_rules
      @namespace = params[:id]
      # Init rules
      @rules = DruidConfig::Entities::RuleCollection.new
      if params['rules']
        params['rules'].each do |_, rule|
          @rules << DruidConfig::Entities::Rule.parse(rule)
        end
      end
      # Check consistency
      if @rules.consistent?
        @error = []
        if @namespace == 'Default'
          Redborder::Cluster.instance.default_datasource.update_rules(@rules)
        else
          real_namespace = @namespace == 'Unassigned' ? '' : "_#{@namespace}"
          # Create rules for a non existent datasource
          RedBorder::DRUID_CONFIG.map { |_, v| v['datasource'] rescue nil }.compact.each do |topic|
            datasource = DruidConfig::Entities::DataSource.new(
              { 'name' => "#{topic}#{real_namespace}", 'properties' => '{}' }, {})
            @error.push("error updating #{datasource.name}") unless
              datasource.update_rules(@rules)
          end
        end
      else
        @error = "#{I18n.t('controllers.consistent')}"
      end
      # Check errors
      if @error && !@error.empty?
        render json: { query: false, error: @error }
      else
        render json: { query: true, namespace: @namespace }
      end
    rescue DruidConfig::Exceptions::NotAvailableNodes
      render json: { query: false,
                     error: "#{I18n.t('controllers.no_nodes')}" }
    end

    def cluster_synchronization
      @role = Chef::Role.manager
      @role.override_attributes['redborder']['camus-sync'] ||= {}
    end

    def save_cluster_synchronization
      return unless @errors.empty?
      # Update the role
      @role = Chef::Role.manager

      @role.override_attributes['redborder']['camus-sync'] ||= {}
      @role.override_attributes['redborder']['camus-sync']['start'] =
        @sync['start']
      @role.override_attributes['redborder']['camus-sync']['interval'] =
        @sync['interval']
      @role.override_attributes['redborder']['camus-sync']['mode'] =
        @sync['mode']

      # Save data
      @role.save
    end
  end

  #
  # => Show the kafka nodes
  #    It's showed inside a tab
  #

# def kafka_nodes
#   begin
#     info = Redborder::Node.get_kafka_info
#     @kafka_info = info[0]
#     @reasigning_partitions = info[1]

#     unless (@kafka_info.fetch(:rb_flow, {}).map do |_key, value|
#       value[:storm].nil? || value[:storm].eql?('?')
#     end).include? false
#       @kafka_info.delete(:rb_flow_post)
#     end

#     unless (@kafka_info.fetch(:rb_event, {}).map do |_key, value|
#       value[:storm].nil? || value[:storm].eql?('?')
#     end).include? false
#       @kafka_info.delete(:rb_event_post)
#     end

#   rescue StandardError => e
#     @kafka_info = nil
#     Rails.logger.error e
#   end
# end

DittoCode::Exec.if 'ENTERPRISE' do
  #
  # => Show virtual_ip_addresses.
  #    It's showed inside a tab
  #
  def virtual_ip_addresses
    @vips = Setting.virtual_ips
  end

  #
  # => Save a virtual_ip_addresses
  #
  def save_virtual_ip_addresses
    vips = params[:vips]
    vips_mode = params[:vips_mode]

    @errors = save_virtual_ips vips, vips_mode
  end
end

  #
  # => Show the cluster status.
  #    It's showed inside a tab
  #
  def cluster_status
    @managers = Manager.all
    render layout: false, partial: 'cluster_status'
  end

  #
  # => Show the small version of cluster status.
  #    It's showed inside a tab
  #
  def cluster_status_small
    @managers = Manager.all
    render layout: false, partial: 'cluster_status_small'
  end

  def save_table_data
    begin
      # These are hashes where the key is the manager, and the value the data
      data = JSON.parse params[:managers]
      errors = []

      # There must be at least one of these services running in the cluster
      services = Manager::MANDATORY_SERVICES.select do |s|
        Manager::DISPLAYABLE_SERVICES.include? s
      end

      services.each do |service|
        values = data.map { |element| element['services'][service] }
        unless values.any? { |val| val }
          errors << "#{I18n.t('controllers.one_instance')} #{service} #{I18n.t('controllers.running')}"
        end
      end

      # The number of Zookeeper instances must be odd
      # unless data.select{|t| t['services']['zookeeper']}.length.odd?
      # errors << "The number of zookeeper instances must be odd."
      # end

      if errors.empty?
        # Save services and modes
        data.each do |node|
          m = Manager.new node['name']
          m.set_mode node['mode']
          node['services'].each do |service, status|
            m.set_service service, status
          end
          errors.concat m.save
        end
      end

      # Redraw the table
      @managers = Manager.all
      @info = Redborder::Node.get_occupation
      @errors = errors.map { |msg| "<li>#{msg}</li>" }.join
    rescue
      @errors = "<li>Unknow error: Cluster Status could not be saved</li>"
    end
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def save_rules
      rules = JSON.parse params[:data]
      DruidRules.rules = rules
      @druid_rules = DruidRules.rules
    end
  end

    def reassign_partitions
      unless DelayedJob.reassign_partitions? ||
             !User.current_user.super_admin?
        Delayed::Job.enqueue(
          Redborder::Jobs::ReassignPartitionsJob.new(true, current_user.id)
        )
      end
      redirect_to :back
    end

    def up_pcp
      unless DelayedJob.up_pcp? || !User.current_user.super_admin?
        Delayed::Job.enqueue(
          Redborder::Jobs::UpPCPJob.new(true, current_user.id, params[:id])
        )
      end
      redirect_to :back
    end

    def recovery_pcp
      unless DelayedJob.recovery_pcp? || !User.current_user.super_admin?
        Delayed::Job.enqueue(
          Redborder::Jobs::RecoveryPCPJob.new(
            true,
            current_user.id,
            params[:id]
          )
        )
      end
      redirect_to :back
    end

    private

    #
    # Init and check syncrhonization options
    #
    def synchronization_options
      @sync = params[:sync].permit(:nodes, :interval, :start, :mode)
      @errors = []
      # Check options
      @errors << [:interval, "#{I18n.t('controllers.correct_interval')}"] unless
        %w(1 2 4 6 8 12 24).include? @sync['interval']
      @errors << [:start, "#{I18n.t('controllers.select_hour')}"] unless
        (0..23).include? @sync['start'].to_i
      @errors << [:mode, "#{I18n.t('controllers.select_mode')}"] unless
        %w(deduplicate synchronize).include? @sync['mode']
    end

    def save_virtual_ips(data, mode)
      errors = []
      allowed_vips = ['external-kafka', 'external-webui',
                      'external-sfacctd', 'external-f2k'
                    ]
      current_vips = Setting.virtual_ips
      # Use global vips or not
      if mode == 'basic'
        allowed_vips.each do |vip|
          if vip[0..1] == 'in'
            data[vip] = data['internal-global']
          else
            data[vip] = data['external-global']
          end
        end
      end

      data.delete('internal-global')
      data.delete('external-global')

      if Rails.env.development?
        node = Manager.all.sample.node
      else
        node = (Manager.new `hostname -s`.strip).node
      end

      # Clean invalid IPs
      data.delete_if do |hname, hip|
        if allowed_vips.include? hname
          if hip.empty?
            false
          else
            res = hip =~ /^([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])$/
            if res == 0
              flag = false
      
              if node['network']['interfaces'].values.any? { |iface| iface['addresses'].present? }
                node['network']['interfaces'].each_value do |iface|
                  iface['addresses'].each do |ip, ip_prop|
                    if ip_prop['family'] == 'inet'
                      sip = ip
                      mask = ip_prop['prefixlen']
      
                      net_iface = NetAddr::CIDR.create("#{sip}/#{mask}").to_s
                      net_virtual = NetAddr::CIDR.create("#{hip}/#{mask}").to_s
      
                      if net_virtual == net_iface
                        flag = true
                        break
                      end
                    end
                  end
                  break if flag
                end
      
                unless flag
                  errors << "#{I18n.t('controllers.couldnt_save')} #{hname} #{I18n.t('controllers.because')}
                            #{hip} #{I18n.t('controllers.is_not')}"
                end
                !flag
              else
                errors << "#{I18n.t('controllers.couldnt_save')} #{hname} #{I18n.t('controllers.because2')}"
                true
              end
            else
              if (current_vips[hname.to_sym] || '') != hip
                errors << "#{I18n.t('controllers.couldnt_save')} #{hname} #{I18n.t('controllers.because3')}
                          #{hip} #{I18n.t('controllers.not_valid')}"
                true
              end
            end
          end
        else
          errors << "#{hname} #{I18n.t('controllers.not_allowed')}"
          true
        end
      end

      # Save the data
      data.each do |name, ip|
        data = Chef::DataBagItem.new
        data.data_bag Setting::DBNAME
        data['id'] = "ipvirtual-#{name}"

        data['ip'] = ip if !ip.empty?

        errors << "#{I18n.t('controllers.upload_databag')}#{name}" unless data.save
      end # data.each do |name, ip|

      errors
    end

  private

  #
  # Check administrative privileges. If the request comes from a widget,
  # render the template of no privileges
  #
  def require_super_administrative_privileges_manager
    return true if user_signed_in? && current_user.super_admin?
    # Check if request comes from a widget
    if params['w'] == 'true'
      render template: 'managers/no_privileges', layout: false, status: 403
    else
      no_enough_privileges
    end
  end

  #
  # Check root reader privileges. If the request comes from a widget,
  # render the template of no privileges
  #
  def require_super_reader_privileges_manager
    return true if user_signed_in? && current_user.super_reader?
    # Check if request comes from a widget
    if params['w'] == 'true'
      render template: 'managers/no_privileges', layout: false, status: 403
    else
      no_enough_privileges
    end
  end
end
