#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class MitreController < ApplicationController
  ## ======================================================================
  # => Helpers
  ## ======================================================================

  include MitreHelper

  ## ======================================================================
  # => Constants
  ## ======================================================================

  # Maximum number of results per page for Mitre Navigator
  MITRE_NAVIGATOR_MAX_RESULTS = 10
  # Default page number for pagination
  MITRE_NAVIGATOR_DEF_PAGE    = 1
  # Page name for editing Mitre records
  MITRE_NAVIGATOR_EDIT_PAGE   = 'edit'
  MITRE_NAVIGATOR_MATRIX_PAGE = 'view_matrix'
  # Constants to identify Mitre types
  MITRE_IS_TECHNIQUE          = MITRE_TECHNIQUE
  MITRE_IS_SUBTECHNIQUE       = MITRE_SUBTECHNIQUE
  # No Mitre rules constant
  NO_MITRE_RULES              = 0
  # Max druid groupby results
  MITRE_GROUPBY_LIMIT         = 1e3
  # Default delayed job enqueue priority
  MITRE_DJOB_PRIORITY         = 10
  # Default module that druid will query to
  MITRE_DRUID_QUERY_MODULE    = 'ips'

  ## ======================================================================
  # => Before actions
  ## ======================================================================

  # Set session data for interacting with Mitre
  before_action :set_session_data
  # Set time variables for matrix view
  before_action :set_time, only: [:view_mitre_matrix]
  # Retrieve rules for index, update, and edit actions
  before_action :get_rules, only: %i[update edit]
  # Load active rules for index action
  before_action :load_active_rules, only: [:index]

  # GET /mitre
  # Renders the Mitre index page, showing the status of Mitre jobs
  def index
    @assigning = Delayed::Job.where('handler LIKE ?', '%Redborder::Jobs::AssignMitreJob%').exists?
    @downloading = Delayed::Job.where('handler LIKE ?', '%Redborder::Jobs::UpdateMitreJob%').exists?
  end

  # GET /mitre/view_mitre_matrix
  # Displays the Mitre matrix view with intrusion data and rule counters
  def view_mitre_matrix
    @tactics = Mitre.where(is_tactic: true)
    @techniques = Mitre.where(is_technique: true)
    @subtechniques = Mitre.where(is_subtechnique: true)

    source = Redborder::DruidSource.get(source: MITRE_DRUID_QUERY_MODULE)
    @data = fetch_intrusion_data(source, @start_time, @end_time)

    rules = extract_rule_ids(@data)

    Rails.logger.info rules
    @counter = populate_counter(@tactics, @techniques, rules)

    render MITRE_NAVIGATOR_MATRIX_PAGE
  end

  # GET /mitre/search
  # Searches for rules based on a query parameter and paginates results
  def search
    page = params[:page] || MITRE_NAVIGATOR_DEF_PAGE
    per_page = MITRE_NAVIGATOR_MAX_RESULTS
    query = params[:query]

    cte_query = <<-SQL
      WITH ranked_rules AS (
        SELECT *,
               row_number() OVER (PARTITION BY rule_sid ORDER BY id ASC) AS rn
        FROM rules
        WHERE msg LIKE '%#{query}%'
      )
      SELECT * FROM ranked_rules WHERE rn = 1
    SQL

    @rules = Rule.find_by_sql(cte_query)
                 .paginate(page: page, per_page: per_page)

    render MITRE_NAVIGATOR_EDIT_PAGE
  end

  # GET /mitre/:id/edit
  # Displays the edit page for a specific Mitre tactic
  def edit
    tactic = Mitre.find_by(mitre_tactic_id: params[:id], is_tactic: true)
    save_session_data(tactic)
    set_session_data
    render MITRE_NAVIGATOR_EDIT_PAGE
  end

  # POST /mitre/:id/edit/assign_rule_to_mitre
  # Assigns a rule to a Mitre technique or subtechnique
  def assign_rule_to_mitre
    mitre_id = params[:mitre_id]
    rule_id = params[:rule_id].to_i
    mitre_type = is_technique_or_subtechnique(mitre_id)
    message = ''
    if mitre_type == MITRE_IS_TECHNIQUE
      message = assign_rule_to_mitre_matrix(mitre_id, rule_id, is_technique: true,
                                                               is_subtechnique: false)
    elsif mitre_type == MITRE_IS_SUBTECHNIQUE
      message = assign_rule_to_mitre_matrix(mitre_id, rule_id, is_technique: false,
                                                               is_subtechnique: true)
    end

    render json: { status: 'success', message: message }
  rescue StandardError
    render json: { status: 'error', message: message }, status: :unprocessable_entity
  end

  # POST /mitre/id/:edit/assign_tactic_to_rule
  # Assigns a tactic to a rule
  def assign_tactic_to_rule
    mitre_tactic_id = params[:mitre_tactic_id]
    rule_id = params[:rule_id].to_i
    tactic = Mitre.find_by(is_tactic: true, mitre_tactic_id: mitre_tactic_id)

    unless tactic.rule_ids.include?(rule_id)
      tactic.rule_ids << rule_id
      tactic.save!
    end

    save_session_data(tactic)

    render json: { status: 'success', message: 'Mitre tactic has been assigned' }
  rescue StandardError => e
    render json: { status: 'error', message: e.message }, status: :unprocessable_entity
  end

  # POST /mitre/id/:edit/deassign_tactic_from_rule
  # Deassigns a tactic from a rule
  def deassign_tactic_from_rule
    mitre_tactic_id = params[:mitre_tactic_id]
    rule_id = params[:rule_id].to_i
    tactic = Mitre.find_by(is_tactic: true, mitre_tactic_id: mitre_tactic_id)

    if tactic.rule_ids.include?(rule_id)
      tactic.rule_ids.reject! { |rule| rule == rule_id }
      tactic.save!
    end

    save_session_data(tactic)

    render json: { status: 'success', message: 'Mitre tactic has been deassigned' }
  rescue StandardError => e
    render json: { status: 'error', message: e.message }, status: :unprocessable_entity
  end

  # POST /mitre/:edit/assign_rule_to_mitre
  # Deassigns a rule from a Mitre technique or subtechnique
  def deassign_rule_from_mitre
    mitre_id = params[:mitre_id]
    rule_id = params[:rule_id].to_i
    mitre_type = is_technique_or_subtechnique(mitre_id)

    if mitre_type == MITRE_IS_TECHNIQUE
      deassign_rule_from_mitre_matrix(mitre_id, rule_id, is_technique: true, is_subtechnique: false)
    elsif mitre_type == MITRE_IS_SUBTECHNIQUE
      deassign_rule_from_mitre_matrix(mitre_id, rule_id, is_technique: false, is_subtechnique: true)
    end

    render json: { status: 'success', message: 'Mitre rule has been deassigned' }
  rescue StandardError => e
    render json: { status: 'error', message: e.message }, status: :unprocessable_entity
  end

  # GET /mitre/:id/json_mitre_techniques
  # Returns a JSON representation of Mitre techniques and subtechniques for a given tactic
  def json_mitre_techniques
    tactic_name = @tactic_name || session['mitre_tactic_name']

    if tactic_name.blank?
      render json: { error: 'Tactic name is required' }, status: :bad_request
      return
    end

    tactics = Mitre.where(name: tactic_name, is_tactic: true)

    data = tactics.map do |tactic|
      techniques = Mitre.where(name: tactic.name, is_technique: true)
      subtechniques = Mitre.where(name: tactic.name, is_subtechnique: true)

      techniques_data = techniques.map do |technique|
        {
          'technique_name' => technique.technique_name,
          'technique_id' => technique.technique_id,
          'rule_ids' => technique.rule_ids,
          'subtechniques' => subtechniques.select do |st|
                               st.technique_id == technique.technique_id
                             end.map do |subtechnique|
                               {
                                 'subtechnique_name' => subtechnique.subtechnique_name,
                                 'subtechnique_id' => subtechnique.subtechnique_id,
                                 'rule_ids' => subtechnique.rule_ids
                               }
                             end
        }
      end

      {
        'tactic' => tactic.name,
        'techniques' => techniques_data
      }
    end

    render json: data
  end

  # GET /mitre/download
  # Enqueues a job to update Mitre data if no update job is already enqueued
  def download
    if Delayed::Job.where('handler LIKE ?', '%Redborder::Jobs::UpdateMitreJob%').exists?
      flash[:alert] = 'Job already enqueued.'
      alert_type = 'danger'
    else
      Delayed::Job.enqueue(Redborder::Jobs::UpdateMitreJob.new(true, current_user.id),
                           priority: MITRE_DJOB_PRIORITY, run_at: Time.now)
      flash[:success] = 'Job successfully enqueued.'
      alert_type = 'success'
    end

    session[:response_flash] = flash[:success] || flash[:alert]
    session[:alert_type] = alert_type

    flash.keep

    redirect_to mitre_index_path
  end

  # GET /mitre/auto_assign_job
  # Automatically assigns a Mitre job if no such job is currently running and if there are Mitre rules
  def auto_assign_job
    assign_mitre_job_exists = Delayed::Job.where('handler LIKE ?',
                                                 '%Redborder::Jobs::AssignMitreJob%').exists?
    update_mitre_job_exists = Delayed::Job.where('handler LIKE ?',
                                                 '%Redborder::Jobs::UpdateMitreJob%').exists?
    if assign_mitre_job_exists
      flash[:alert] = 'Mitre rules are assigning, please wait.'
      alert_type = 'danger'
    elsif update_mitre_job_exists
      flash[:alert] = 'Mitre rules are creating, please wait.'
      alert_type = 'danger'
    elsif Mitre.count == NO_MITRE_RULES
      flash[:alert] = "You don't have any Mitre rules! Please download them."
      alert_type = 'danger'
    else
      Delayed::Job.enqueue(Redborder::Jobs::AssignMitreJob.new(true, current_user.id),
                           priority: MITRE_DJOB_PRIORITY, run_at: Time.now)
      flash[:success] = 'Job successfully enqueued.'
      alert_type = 'success'
    end

    session[:response_flash] = flash[alert_type.to_sym]
    session[:alert_type] = alert_type

    flash.keep

    redirect_to mitre_index_path
  end

  # GET /mitre/set_matrix_custom_time
  # Sets custom range for searching in the mitre matrix
  def set_matrix_custom_time
    filter = get_default_filters
    filter["filters"]["time"] = MITRE_CUSTOM_FILTER
    filter["filters"]["start_time"] = params[:start_time]
    filter["filters"]["end_time"] = params[:end_time]
    save_filter_to_session(filter)

    render json: { redirect_url: view_mitre_matrix_mitre_index_path(range: MITRE_CUSTOM_FILTER) }
  end

  def get_mitre_widget_data
    @tactics = Mitre.where(is_tactic: true)
    @techniques = Mitre.where(is_technique: true)
    @subtechniques = Mitre.where(is_subtechnique: true)
    start_time = Time.at params[:start_time].to_i
    end_time = Time.at params[:end_time].to_i
    source = Redborder::DruidSource.get(source: MITRE_DRUID_QUERY_MODULE)
    @data = fetch_intrusion_data(source, start_time, end_time)
    rules = extract_rule_ids(@data)
    render json: get_widget_tactics(@tactics, @techniques, @subtechniques, rules)
  end

  private

  #
  # => Retrieve all Snort rules for the index page
  #
  # We use a Common Table Expression (CTE) to efficiently find the minimum id for each rule_sid.
  # The CTE (ranked_rules) applies a window function (row_number) to partition by rule_sid and
  # assign a row number ordered by id within each partition. By filtering the CTE to keep only
  # rows where rn = 1, we effectively select the minimum id for each rule_sid. This approach is
  # both more efficient and more readable than using GROUP BY with a subquery.
  #
  def get_rules
    page = params[:page] || MITRE_NAVIGATOR_DEF_PAGE
    per_page = MITRE_NAVIGATOR_MAX_RESULTS

    cte_query = <<-SQL
      WITH ranked_rules AS (
        SELECT id, row_number() OVER (PARTITION BY rule_sid ORDER BY id ASC) AS rn
        FROM rules
      )
      SELECT id FROM ranked_rules WHERE rn = 1
    SQL

    subquery = Rule.find_by_sql(cte_query).map(&:id)
    @rules = Rule.where(id: subquery)
                 .paginate(page: page, per_page: per_page)
  end

  #
  # => Deassign a rule from Mitre techniques based on technique ID
  #
  def deassign_rule_from_mitre_matrix(mitre_id, rule_id, is_technique: false,
                                      is_subtechnique: false)
    if is_technique && is_subtechnique
      raise ArgumentError,
            'Specify either is_technique or is_subtechnique, not both'
    end
    unless is_technique || is_subtechnique
      raise ArgumentError,
            'Specify either is_technique or is_subtechnique'
    end

    if is_technique
      techniques = Mitre.where(is_technique: true, technique_id: mitre_id)
      techniques.each do |technique|
        subtechniques_with_rule = Mitre.where(is_subtechnique: true, technique_id: technique.technique_id).where(
          '? = ANY(rule_ids)', rule_id
        )
        unless subtechniques_with_rule.empty?
          raise ArgumentError,
                'Cannot deassign the rule from the technique because it is associated with one or more subtechniques'
        end

        unless technique.rule_ids.include?(rule_id)
          raise ArgumentError, 'The rule is not associated with the technique'
        end

        technique.rule_ids.reject! { |rule| rule == rule_id }
        technique.save!
      end
    elsif is_subtechnique
      subtechniques = Mitre.where(is_subtechnique: true, subtechnique_id: mitre_id)
      subtechniques.each do |subtechnique|
        unless subtechnique.rule_ids.include?(rule_id)
          raise ArgumentError, 'The rule is not associated with the subtechnique'
        end

        subtechnique.rule_ids.reject! { |rule| rule == rule_id }
        subtechnique.save!
      end
    end
  end

  #
  # => Assign a rule to Mitre techniques based on technique ID
  #
  def assign_rule_to_mitre_matrix(mitre_id, rule_id, is_technique: false, is_subtechnique: false)
    if is_technique && is_subtechnique
      raise ArgumentError,
            'Specify either is_technique or is_subtechnique, not both'
    end
    unless is_technique || is_subtechnique
      raise ArgumentError,
            'Specify either is_technique or is_subtechnique'
    end

    if is_subtechnique
      subtechniques = Mitre.where(is_subtechnique: true, subtechnique_id: mitre_id)
      subtechniques.each do |subtechnique|
        technique = Mitre.find_by(is_technique: true, technique_id: subtechnique.technique_id)
        raise StandardError, 'Technique not found for the given subtechnique' unless technique

        tactics = Mitre.where(is_tactic: true, tactic_id: technique.tactic_id)
        raise StandardError, 'No tactics found for the given technique' if tactics.empty?

        unless technique.rule_ids.include?(rule_id)
          raise StandardError,
                'You cannot assign a subtechnique if the rule is not associated with the technique'
        end

        unless subtechnique.rule_ids.include?(rule_id)
          subtechnique.rule_ids << rule_id
          subtechnique.save!
        end
      end
    else
      techniques = Mitre.where(is_technique: true, technique_id: mitre_id)
      techniques.each do |technique|
        tactics = Mitre.where(is_tactic: true, tactic_id: technique.tactic_id)
        tactics.each do |tactic|
          unless tactic.rule_ids.include?(rule_id)
            tactic.rule_ids << rule_id
            tactic.save!
          end
        end

        unless technique.rule_ids.include?(rule_id)
          technique.rule_ids << rule_id
          technique.save!
        end
      end
    end
  end

  #
  # => Load active Mitre rules and their assigned rules count
  #
  def load_active_rules
    @matrix = Mitre.where(is_tactic: true)
    @assigned_rules = {}
    @matrix.each do |tactic|
      name = tactic.name
      rules = Mitre.where('name = ? AND array_length(rule_ids, 1) > 0', name)
      unique_rule_ids = Set.new
      rules.each do |rule|
        unique_rule_ids.merge(rule.rule_ids)
      end
      @assigned_rules[name] = unique_rule_ids.size
    end
  end

  #
  # => Fetch intrusion data from a source within a time range
  #
  def fetch_intrusion_data(source, start_time, end_time)
    source.where({
                   query_type: :group_by,
                   range: :custom,
                   start_time: start_time,
                   end_time: end_time,
                   granularity: :minute,
                   aggregation: source.default_aggregation,
                   group_by: source.default_columns_mitre,
                   limit: MITRE_GROUPBY_LIMIT
                 })
  end

  #
  # => Extract rule IDs from intrusion data
  #
  def extract_rule_ids(data)
    data.each_with_object([]) do |entry, rules|
      next if entry.sig_id.nil?

      rules << {
        priority: entry.priority,
        sig_id: entry.sig_id.to_i
      }
    end
  end

  #
  # => Populate a counter with the number of matching rules for each tactic and technique
  #
  def populate_counter(tactics, techniques, rules)
    counter = {}
    tactics.each do |tactic|
      matching_rules_count = NO_MITRE_RULES

      rules.each do |rule|
        matching_rules_count += 1 if tactic.rule_ids.include?(rule[:sig_id])
      end

      counter[tactic.tactic_id] = if matching_rules_count.positive?
                                    matching_rules_count
                                  else
                                    NO_MITRE_RULES
                                  end
    end

    rules.map { |rule| rule[:sig_id].to_i }

    techniques.each do |technique|
      counter[technique.technique_id] = { rules: [], color: nil }
      matching_rules = rules.select do |rule|
        technique.rule_ids.map(&:to_i).include?(rule[:sig_id].to_i)
      end

      if matching_rules.any?
        counter[technique.technique_id][:rules].concat(matching_rules)
        priorities = matching_rules.map { |rule| rule[:priority] }.uniq
        highest_priority = priorities.max_by { |priority| %w[low medium high].index(priority) }
        counter[technique.technique_id][:color] = determine_matrix_color(highest_priority)
      else
        counter[technique.technique_id][:rules] = NO_MITRE_RULES
        counter[technique.technique_id][:color] = MITRE_UNKNOWN_COLOR
      end
    end

    counter
  end

  #
  # => Process entith wfor widget data calculation
  #
  def process_entity(entity, rules, counter, data, entity_name_key)
    matching_rules_count = calculate_matching_rules_count(entity, rules)
    counter[entity.tactic_id] = matching_rules_count.positive? ? matching_rules_count : NO_MITRE_RULES

    if counter[entity.tactic_id] != NO_MITRE_RULES
      matching_rules = rules.select { |rule| entity.rule_ids.map(&:to_i).include?(rule[:sig_id].to_i) }
      highest_priority = find_highest_priority(matching_rules)

      data << {
        events: counter[entity.tactic_id],
        name: entity.send(entity_name_key),
        priority: highest_priority
      }
    end
  end

  #
  # => Get last tactics triggered by intrusion rules
  #
  def get_widget_tactics(tactics, techniques, subtechniques, rules)
    data = []
    counter = {}

    case params[:type]
    when 'tactics'
      tactics.each { |tactic| process_entity(tactic, rules, counter, data, :name) }
    when 'techniques'
      techniques.each { |technique| process_entity(technique, rules, counter, data, :technique_name) }
    when 'subtechniques'
      subtechniques.each { |subtechnique| process_entity(subtechnique, rules, counter, data, :subtechnique_name) }
    end

    combined_data = data.each_with_object({}) do |entry, hash|
      name = entry[:name]
      if hash.key?(name)
        hash[name][:events] += entry[:events]
        if %w[high medium low].index(entry[:priority]) > %w[high medium low].index(hash[name][:priority])
          hash[name][:priority] = entry[:priority]
        end
      else
        hash[name] = entry
      end
    end

    final_data = combined_data.values
    final_data.sort_by! { |entry| -entry[:events] }
    final_data
  end
end
