#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ObjectTypesController < ApplicationController
  before_action :set_object_type, only: [:update, :destroy]

  def index
    session.delete(:search)
    @object_types = ObjectType.order(created_at: :desc).paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  def new
    @object_type = ObjectType.new
    render layout: false
  end

  def create
    @object_type = ObjectType.new(object_type_params)
    respond_to do |format|
      if @object_type.save

        if session[:search].present?
          filter = ['( LOWER(name) LIKE :search or ' \
                  'LOWER(comment) LIKE :search)',
                  search: "%#{session[:search].try(:downcase)}%"]
          @object_types = ObjectType.where(filter)
        else
          @object_types = ObjectType.all
        end

        new_page = params[:page] ? params[:page].to_i : 1

        @object_types = @object_types.order(created_at: :desc)
                        .paginate(page: new_page, per_page: current_user.per_page_count)

        flash.now[:notice] = I18n.t('object_types.create.success')
        format.html { redirect_to object_types_path, notice: I18n.t('object_types.create.success') }
      else
        flash.now[:alert] = I18n.t('object_types.create.error')
        format.html { render :new, status: :unprocessable_entity }
      end
      format.js
    end
  end


  def edit
    @object_type = ObjectType.find(params[:id])
    render layout: false
  end

  def update
    old_name = @object_type.name_was
    respond_to do |format|
      if @object_type.update(object_type_params)
        update_associated_objects(old_name, @object_type.name)
        flash[:notice] = I18n.t('object_types.update.success')
        format.html { redirect_to object_types_path, notice: I18n.t('object_types.update.success') }
      else
        flash.now[:alert] = I18n.t('object_types.update.error')
        format.html { render :edit, status: :unprocessable_entity }
      end
      format.js
    end
  end

  def destroy
    respond_to do |format|
      if @object_type.destroy
        destroy_associated_objects

        if session[:search].present?
          filter = ['( LOWER(name) LIKE :search or ' \
                  'LOWER(comment) LIKE :search)',
                  search: "%#{session[:search].try(:downcase)}%"]
          @object_types = ObjectType.where(filter)
        else
          @object_types = ObjectType.all
        end

        current_page = [params[:page].to_i, 1].max
        max_page = (@object_types.any? ? (@object_types.count.to_f / current_user.per_page_count).ceil : 1)
        new_page = [current_page, max_page].min

        @object_types = @object_types.order(created_at: :desc)
                        .paginate(page: new_page, per_page: current_user.per_page_count)

        flash[:notice] = I18n.t('object_types.destroy.success')
        format.html { redirect_to object_types_path, notice: I18n.t('object_types.destroy.success') }
      else
        flash[:alert] = I18n.t('object_types.destroy.error')
        format.html { redirect_to object_types_path, alert: I18n.t('object_types.destroy.error') }
      end
      format.js
    end
  end

  def search
    if params[:search].blank?
      session.delete(:search)
      @object_types = ObjectType.all.order(created_at: :desc)
                    .paginate(page: params[:page], per_page: current_user.per_page_count)
    else
      session[:search] = params[:search] if params[:search].present?
      filter = ['( LOWER(name) LIKE :search or ' \
              'LOWER(comment) LIKE :search)',
              search: "%#{session[:search].try(:downcase)}%"]
      @object_types = ObjectType.where(filter).order(created_at: :desc)
              .paginate(page: params[:page], per_page: current_user.per_page_count)
    end
  end

  #
  # Export all objects as CSS
  #
  def export
    export_objects_type_csv
  end

  #
  # Show a modal to upload a file to import objects.
  #
  def import
    render layout: false
  end

  def import_file
    content_file = params[:object_type][:csv_file]
    
    unless content_file.content_type == "text/csv"
      redirect_to object_types_path, alert: t(".csv_error")
      return
    end
  
    begin
      csv_text = content_file.read
      csv = CSV.parse(csv_text, headers: false, col_sep: ';')

      
      if csv.any? { |row| row.length != 2 }
        redirect_to object_types_path, alert: t("object_types.import_file.wrong_column_format")
        return
      end
  
      
      if params[:add_vars] == "true"
        result = ObjectType.add_vars_from_file(content_file)
        msg = "Processed with add_vars_from_file: #{result.inspect}"
      else
        created_count, updated_count = ObjectType.replace_vars_from_file(content_file)
        msg = "#{created_count} records created, #{updated_count} records updated."
      end
    rescue StandardError => e
      redirect_to object_types_path, alert: t("object_types.import_file.import_error", error: e.message)
      return
    end
  
    redirect_to object_types_path, notice: msg
  end

  private

  def set_object_type
    @object_type = ObjectType.find(params[:id])
  rescue ActiveRecord::RecordNotFound
    flash[:alert] = I18n.t("object_types.not_found")
    redirect_to object_types_path
  end

  def update_associated_objects(old_name, new_name)
    RedborderObject.where(object_type: old_name).update_all(object_type: new_name)
  end

  def destroy_associated_objects
    RedborderObject.where(object_type: @object_type.name).update_all(object_type: nil)
  end

  def object_type_params
    params.require(:object_type).permit(:name, :comment)
  end

  def export_objects_type_csv
    filename = in_time_machine { "#{Time.now.strftime("%Y%m%d%H%M%S")}-Object_types.csv" }
    content = objects_to_csv_format
    send_data content, filename: filename
  end

  def objects_to_csv_format
    if session[:search].present?
      filter = ['( LOWER(name) LIKE :search or ' \
                'LOWER(comment) LIKE :search)',
                search: "%#{session[:search].try(:downcase)}%"]
      @object_types = ObjectType.where(filter).order(created_at: :desc).map{ |o| "#{o.name};#{o.comment}" }.join("\n")
    else
      @object_types = ObjectType.all.order(created_at: :desc).map{ |o| "#{o.name};#{o.comment}" }.join("\n")
    end
  end
end
