#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class PlaybookAssignmentRulesController < ApplicationController
  helper Redborder::ModulesHelper
  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_assignment_rule, only: [:edit, :update, :destroy, :show, :create_clone, :set_enabled]
  before_action :set_properties_and_subproperties, only: [:new, :edit, :update]
  before_action :set_dropdown_subproperties, only: [:new, :edit, :update, :show]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all assignment rules in a table
  #
  def index
    @assignment_rules = PlaybookAssignmentRule.order(priority: :asc)
    @playbooks = PlaybookTemplate.order(:name)

    # Paginate results
    @page = params[:page] || 1
    @assignment_rules = @assignment_rules.paginate(page: params[:page],
                                                   per_page: current_user.per_page_count)
  end

  #
  # => Render the form to create a new playbook template
  #
  def new
    @assignment_rule = PlaybookAssignmentRule.new
    @playbooks = PlaybookTemplate.order(:name)
    @assignment_rule.playbook_assignment_rules_conditions.new
  end

  #
  # => Receive the data to create a new playbook template. Try to create a new playbook template
  #    and redirect to playbook_assignment_rules_path if it's ok
  #
  def create
    begin
      @assignment_rule = current_user.playbook_assignment_rules.build(assignment_rule_params)

      if @assignment_rule.save
        Rails.logger.info 'Created assignment rule successfully'
      else
        Rails.logger.error "Error in create assignment rule: #{@assignment_rule.errors.full_messages.join(', ')}"
      end

      respond_to do |format|
        format.html do
          redirect_to playbook_assignment_rules_path,
                      notice: "#{I18n.t('controllers.playbook_assignment_rules.created_ok')}"
        end
        format.js { flash[:notice] = "#{I18n.t('controllers.playbook_assignment_rules.created_ok')}" }
        format.json { head :no_content }
      end
    rescue => e
      Rails.logger.error "Error in create assignment rule: #{e.message}"
      respond_to do |format|
        format.html { redirect_to playbook_assignment_rules_path, alert: "#{I18n.t('controllers.playbook_assignment_rules.created_error')}" }
        format.js do
          flash[:error] = @assignment_rule.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @assignment_rule.errors, status: :unprocessable_entity
        end
      end
    end
  end

  def create_clone
    begin
      new_assignment_rule = nil
      ActiveRecord::Base.transaction do
        new_assignment_rule = @assignment_rule.dup
        new_assignment_rule.name = "#{@assignment_rule.name} (#{I18n.t('controllers.playbook_assignment_rules.clone')})"
        new_assignment_rule.priority += 1
        new_assignment_rule.save!

        @assignment_rule.playbook_assignment_rules_conditions.each do |condition|
          new_condition = condition.dup
          new_condition.playbook_assignment_rule = new_assignment_rule
          new_condition.save!
        end
        check_priority(new_assignment_rule)
      end

      rb_flash "#{I18n.t('controllers.playbook_assignment_rules.clone_ok')}"
      redirect_to playbook_assignment_rules_path
    rescue => e
      Rails.logger.error "Error cloning assignment rule: #{e.message}"
      respond_to do |format|
        format.html { redirect_to playbook_assignment_rules_path, alert: "#{I18n.t('controllers.playbook_assignment_rules.clone_error')}" }
        format.js do
          flash[:error] = "#{e.message}"
        end
        format.json do
          render json: @assignment_rule.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Render the form to edit an assignment rule
  #
  def edit
    @playbooks = PlaybookTemplate.order(:name)
  end

  def update
    @page = params[:page] || 1

    respond_to do |format|
      if @assignment_rule.update(assignment_rule_params)
        format.html do
          notice_message = "#{I18n.t('controllers.playbook_assignment_rules.update_ok')}"

          if request.referer && request.referer.include?(playbook_assignment_rules_path(@assignment_rule))
            redirect_to playbook_assignment_rules_path(@assignment_rule), notice: notice_message
          else
            redirect_to playbook_assignment_rules_path(page: @page), notice: notice_message
          end
        end

        format.js { flash[:notice] = "#{I18n.t('controllers.playbook_assignment_rules.update_ok')}" }
        format.json { render json: @assignment_rule }
      else
        Rails.logger.error "Error in updating assignment rule: #{e.message}"
        format.html { redirect_to playbook_assignment_rules_path, alert: "#{I18n.t('controllers.playbook_assignment_rules.update_error')}" }
        format.js do
          flash[:error] = @assignment_rule.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @assignment_rule.errors, status: :unprocessable_entity
        end
      end
    end
  end

  def show
    @playbook = PlaybookTemplate.find_by(id: @assignment_rule.playbook_templates_id)
    @playbooks = PlaybookTemplate.order(:name)
    @operation = { '=' => "#{I18n.t('controllers.playbook_assignment_rules.operations.eq')}",
                   '>' => "#{I18n.t('controllers.playbook_assignment_rules.operations.greater')}",
                   '<' => "#{I18n.t('controllers.playbook_assignment_rules.operations.less')}",
                   '!=' => "#{I18n.t('controllers.playbook_assignment_rules.operations.not_eq')}" }
    @format_condition = @assignment_rule.playbook_assignment_rules_conditions.map do |cond|
      if cond.sub_property == 'name'
        "#{cond.property} > #{I18n.t('controllers.playbook_assignment_rules.tactic_name')}"
      else
        "#{cond.property} > #{cond.sub_property.humanize}"
      end
    end

    @format_value = @assignment_rule.playbook_assignment_rules_conditions.map do |cond|
      if @dropdown_sub_properties[cond.sub_property]
        @dropdown_sub_properties[cond.sub_property].find { |sub_array| sub_array[0].to_s == cond.value }[1]
      else
        cond.value
      end
    end
  end

  #
  # Search on playbook assignment rules
  #
  def search
    @playbooks = PlaybookTemplate.order(:name)
    @page = params[:page] || 1

    filter = ['( LOWER(name) LIKE :search or ' \
      'LOWER(description) LIKE :search )',
      search: "%#{params[:fast_search].try(:downcase)}%"]

    @assignment_rules = PlaybookAssignmentRule.where(filter).order(priority: :asc)
                                              .paginate(page: @page, per_page: current_user.per_page_count)
  end

  #
  # => Destroy an assignment_rule
  #
  def destroy
    down_one_priority_to_another(@assignment_rule)
    @assignment_rule.destroy
    respond_to do |format|
      format.html { redirect_to playbook_assignment_rules_path }
      format.js
      format.json { head :no_content }
    end
  end

  def set_enabled
    respond_to do |format|
      begin
        @assignment_rule.set_enabled
        @assignment_rule.save!
        format.html { redirect_to playbook_assignment_rules_path }
        format.js { render js: "window.location.href='#{playbook_assignment_rules_path}'" }
        format.json { render json: @assignment_rule }
      rescue => e
        format.html { render action: 'show', alert: e.message }
        format.js { flash[:error] = e.message }
        format.json { render json: { error: e.message }, status: :unprocessable_entity }
      end
    end
  end

  def update_priority
    params[:priority].each do |assignment_rule|
      PlaybookAssignmentRule.find(assignment_rule[:id]).update(priority: assignment_rule[:priority])
    end
    head :ok
  end

  private

  #
  # => Init the playbook template if the user has permissions
  #
  def set_assignment_rule
    @assignment_rule = PlaybookAssignmentRule.find(params[:id])
  end

  def assignment_rule_params
    params
      .require(:playbook_assignment_rule)
      .permit(:name, :description, :enabled, :page, :search, :fast_search,
              :condition_type, :playbook_templates_id,
              playbook_assignment_rules_conditions_attributes: [:id, :property, :sub_property, :operation, :value, :_destroy])
  end

  def check_priority(new_assignment_rule)
    assignment_rules = PlaybookAssignmentRule.where('priority >= ?', new_assignment_rule.priority)
                                             .where.not(id: new_assignment_rule.id)
                                             .order(priority: :asc)
    previous = new_assignment_rule
    assignment_rules.each do |assignment_rule|
      next if assignment_rule.priority > previous.priority

      assignment_rule.priority += 1
      assignment_rule.save!
      previous = assignment_rule
    end
  end

  def down_one_priority_to_another(specific_assignment_rule)
    assignment_rules = PlaybookAssignmentRule.where('priority >= ?', specific_assignment_rule.priority)
                                             .where.not(id: specific_assignment_rule.id)
                                             .order(priority: :asc)
    previous = specific_assignment_rule
    assignment_rules.each do |assignment_rule|
      assignment_rule.priority -= 1
      assignment_rule.save!
      previous = assignment_rule
    end
  end

  def set_properties_and_subproperties
    incidents_excluded_attributes = ['id', 'uuid', 'description', 'visible', 'incident_status_id',
                                     'created_at', 'updated_at', 'sent', 'failure_count', 'sent_timestamp']

    mitre_excluded_attributes = ['id', 'description', 'mitre_tactic_id', 'created_at', 'updated_at',
                                 'rule_ids', 'is_tactic', 'is_technique', 'is_subtechnique']

    @properties = [[I18n.t('controllers.playbook_assignment_rules.incidents'), 'Incidents'], ['Mitre', 'Mitre']]
    @sub_properties = {
      'Incidents' => Incident.column_names
                             .reject { |col| incidents_excluded_attributes.include?(col) }
                             .map { |col| [I18n.t("controllers.playbook_assignment_rules.incidents_cols.#{col}"), col] },
      'Mitre' => Mitre.column_names
                      .reject { |col| mitre_excluded_attributes.include?(col) }
                      .map { |col| [I18n.t("controllers.playbook_assignment_rules.mitre_cols.#{col}"), col] }
    }
  end

  def set_dropdown_subproperties
    @dropdown_sub_properties = {
      'priority_id' => Priority.all.order(score: :desc).map { |priority| [priority.score, I18n.t("incidents.priorities.#{priority.name}")] },
      'incident_status_id' => IncidentStatus.all.map { |status| [status.id, I18n.t("incidents.status.#{status.name.downcase}")] },
      'sensor_id' => Sensor.where(type: 1).map { |sensor| [sensor.id, sensor.name] },
      'source' => Redborder::Module.registered_modules
                                   .select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
                                   .map { |_, v| [v[:incident_source_name], v[:incident_source_name]] }
    }
  end
end
