#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class RedborderObjectsController < ApplicationController
  # -------------------------------------------------------------
  # Before actions
  # -------------------------------------------------------------

  before_action :set_object, only: [:edit, :update, :destroy]
  before_action :set_domain_names, only: [:index, :edit, :update, :create, :search]

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Render all the objects inside tabs.
  #
  def index
    session.delete(:search)
    
    sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

    @net_objects = fetch_objects(NetObject, sensor_ids)
                    .order(created_at: :desc)
                    .paginate(page: 1, per_page: current_user.per_page_count)
  
    @mac_objects = fetch_objects(MacObject, sensor_ids)
                    .order(created_at: :desc)
                    .paginate(page: 1, per_page: current_user.per_page_count)
  
    @vlan_objects       = current_user
                          .vlan_objects
                          .order(created_at: :desc)
                          .paginate(page: 1, per_page: current_user.per_page_count)
    @service_objects    = current_user
                          .service_objects
                          .order(created_at: :desc)
                          .paginate(page: 1, per_page: current_user.per_page_count)
    @application_objects = ApplicationObject.all
                            .order(created_at: :desc)
                            .paginate(page: 1, per_page: current_user.per_page_count)
  end

  #
  # Render the form to create a new object
  #
  def new
    return unless RedborderObject::OBJECT_CLASSES.include? params[:class]
    @object = params[:class].constantize.new(value: params[:value])
    render layout: false
  end

  #
  # Render the form to edit an object
  #
  def edit
    # Define a whitelist of allowed object types
    allowed_classes = {
      'NetObject' => NetObject,
      'MacObject' => MacObject,
      'ProtocolObject' => ProtocolObject,
      'ServiceObject' => ServiceObject,
      'VlanObject' => VlanObject
    }

    object_type_param = params[:object][:type]

    # Validate that the object type is in the whitelist
    unless allowed_classes.key?(object_type_param)
      rb_flash "#{I18n.t('controllers.object_not_found')}", type: :error
      redirect_to objects_path and return
    end

    # Find the object based on the class and ID
    object_type = allowed_classes[object_type_param]
    @object = object_type.find_by(id: params[:id])

    unless @object
      rb_flash "#{I18n.t('controllers.object_not_found')}", type: :error
      redirect_to objects_path and return
    end

    render layout: false
  end

  #
  # Receive the POST data of an object and try to save it
  #
  def create
    # Create the object with the received parameters
    @object = RedborderObject.new(object_params)

    # Assign user_id and sensor_id
    assign_user_and_sensor

    # Get all objects to calculate new pagination
    if @object.save
      sensor_ids = current_user.domains_with_manage_permission.pluck(:id)
      
      if session[:search].present?
        filter = ['( LOWER(name) LIKE :search or ' \
                'LOWER(comment) LIKE :search or LOWER(value) LIKE :search)',
                search: "%#{session[:search].try(:downcase)}%"]
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids).where(filter)
      else
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids)
      end

      new_page = params[:page] ? params[:page].to_i : 1
      @objects = @objects.order(created_at: :desc).paginate(page: new_page, per_page: current_user.per_page_count)
      flash.now[:notice] = I18n.t('redborder_objects.create.success')
    else
      flash[:alert] = I18n.t('redborder_objects.create.error')
    end
  end

  #
  # Receive the POST data of an object and try to update it
  #
  def update
    # Get all domain IDs that the current user can manage
    assign_user_and_sensor

    # Attempt to update the object with the received parameters (excluding the sensor)
    if @object.update(object_params.except(:sensor))
      flash.now[:notice] = I18n.t('redborder_objects.update.success')
      render layout: false
    else
      flash[:alert] = I18n.t('redborder_objects.update.error')
      render :edit
    end
  end

  #
  # Destroy an object
  #
  def destroy
    if @object && @object.destroy
      sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

      # Keep filter when deleting
      if session[:search].present?
        filter = ['( LOWER(name) LIKE :search or ' \
                'LOWER(comment) LIKE :search or LOWER(value) LIKE :search)',
                search: "%#{session[:search].try(:downcase)}%"]
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids).where(filter)
      else
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids)
      end

      # Keep page in range and deleting last page if it is empty
      current_page = [params[:page].to_i, 1].max
      max_page = (@objects.any? ? (@objects.count.to_f / current_user.per_page_count).ceil : 1)
      new_page = [current_page, max_page].min

      # Show table fixed (pagination)
      @objects = @objects.order(created_at: :desc).paginate(page: new_page, per_page: current_user.per_page_count)
      flash.now[:notice] = I18n.t('redborder_objects.destroy.success')
    else
      flash[:alert] = I18n.t('redborder_objects.destroy.error')
    end
  end

  #
  # Store session tab of the user
  #
  def change_session_tab
    session[:object] = params[:id].to_sym
    render js: ''
  end

  # TODO: THIS NEED TO BE DELETED
  # def apply
  #   @sensor.apply_objects
  #   Manager.apply_objects
  #   respond_to do |format|
  #     format.html { redirect_to :back, notice: 'Objects has been applied.' }
  #     format.json { head :no_content }
  #   end
  # end

  #
  # Search in objects
  #
  def search
    begin
      @object_type = RedborderObject.from_string(params[:tab])
    rescue RedborderObject::Unknown
      error_400("#{I18n.t('controllers.no_object')}") && return
    end
    sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

    if params[:search].blank?
      session.delete(:search)
      @objects = fetch_objects(@object_type.classify.constantize, sensor_ids)
              .order(created_at: :desc)
              .paginate(page: params[:page], per_page: current_user.per_page_count)
    else
      session[:search] = params[:search] if params[:search].present?
      filter = ['( LOWER(name) LIKE :search or ' \
              'LOWER(comment) LIKE :search or LOWER(value) LIKE :search or LOWER(ip_addresses) LIKE :search )',
              search: "%#{session[:search].try(:downcase)}%"]
      @objects = fetch_objects(@object_type.classify.constantize, sensor_ids)
              .where(filter)
              .order(created_at: :desc)
              .paginate(page: params[:page], per_page: current_user.per_page_count)
    end
  end

  #
  # Export all objects as CSS
  #
  def export
    tab = params[:tab]
    export_objects_csv tab
  end

  #
  # Show a modal to upload a file to import objects.
  #
  def import
    @type = params[:type]
    render layout: false
  end

  #
  # Import objects from a given CSV file.
  #
  def import_file
    tab = params[:tab]
    
    if params[:redborder_object][:csv_file].blank?
      msg = "#{vars[0]} #{I18n.t('controllers.unknow_content_csv')}"
      redirect_to objects_path, alert: msg
    else
      content_file = params[:redborder_object][:csv_file]
    end

    required_headers = RedborderObject.validators.select { |v| v.kind == :presence }.map { |v| v.attributes }.flatten.map { |attr| attr.to_s }
    normalized_required_headers = required_headers.map{|h| h.strip.downcase}
    csv_headers = CSV.read(content_file.path, headers: true).headers.flat_map do |header|
      header.strip.split(/[\s;,]+/).map{|h| h.strip.downcase}
    end
    missing_headers = required_headers - csv_headers

    if missing_headers.empty? && params[:add_vars] == 'true'
      vars_count = RedborderObject.add_vars_from_file(content_file, tab)
      msg = "#{vars_count[0]} #{I18n.t('controllers.objects_added')} #{vars_count[1]} #{I18n.t('controllers.objects_updated')}"
      redirect_to objects_path, notice: msg
    elsif missing_headers.empty? && params[:add_vars] == 'false'
      vars_count = RedborderObject.replace_vars_from_file(content_file, tab)
      msg = "#{vars_count} #{I18n.t('controllers.objects_added')}"
      redirect_to objects_path, notice: msg
    else
      msg = "#{I18n.t('controllers.objects_added_no_valid')}"
      redirect_to objects_path, alert: msg
    end
  end

  private

  # This function retrieves objects of a given class based on the following conditions:
  # 1. Returns objects belonging to the current user (user_id = current_user.id).
  # 2. If sensor IDs are provided, includes objects where user_id is NULL and sensor_id is in the list of allowed IDs.
  # 3. If the current user is a super admin, also includes objects where sensor_id is NULL.
  # It returns an ActiveRecord::Relation with all objects that meet these conditions.
  def fetch_objects(object_class, sensor_ids = [])

    conditions = object_class.where(user_id: current_user.id)
  
    if sensor_ids.present?
      conditions = conditions.or(object_class.where(user_id: nil, sensor_id: sensor_ids))
    end
  
    if current_user.super_admin?
      conditions = conditions.or(object_class.where(sensor_id: nil))
    end
  
    conditions
  end  

  # Set @domain_names
  def set_domain_names
    @domain_names = current_user.domains_with_manage_permission.map { |d| [d.id, d.path_name] }.to_h
  end

  # Method to assign user_id and sensor_id to the object based on the provided parameters.
  def assign_user_and_sensor
    # Handle the assignment of user and sensor
    if params[:check] == "1"
      @object.user_id = nil
      @object.sensor_id = params[:sensor_id].present? ? params[:sensor_id] : current_user.sensor_id
    else
      @object.user_id = current_user.id
      @object.sensor_id = current_user.sensor_id
    end
  end

  #
  # Set the object.
  #
  def set_object
    # Convert the type name to a class format
    type = RedborderObject.from_string(params[:object][:type])
    model_class = type.classify.constantize  # Convert the name to CamelCase format and then get the class

    @object = model_class
                .where(id: params[:id])
                .where("user_id = ? OR user_id IS NULL", current_user.id)
                .first

    return if @object

    unless @object
      rb_flash "#{I18n.t('controllers.object_not_exist')}", type: :error
      return
    end
  end

  #
  # Export objects as CSV
  #
  def export_objects_csv(tab)
    filename = in_time_machine { Time.now.strftime('%Y%m%d%H%M%S') +"-#{tab.capitalize}-Objects.csv" }
    content = objects_to_csv_format(tab)
    send_data content, filename: filename
  end

  #
  # Convert current Objects to CSV
  #
  def objects_to_csv_format(tab)
    search_term = params[:search].presence || ""
    search_term = "" if search_term == "undefined"
  
    # Define the filter for the search term
    filter = ['(LOWER(name) LIKE :search OR LOWER(comment) LIKE :search OR LOWER(value) LIKE :search)',
              { search: "%#{search_term.downcase}%" }]
  
    # Fetch objects based on user permissions
    objects = if current_user.super_admin?
                RedborderObject.where(user: [current_user, nil], type: "#{tab.capitalize}Object")
              else
                RedborderObject.where(user: current_user, type: "#{tab.capitalize}Object")
              end
  
    # Load domains into a hash for quick access
    domain_map = current_user.domains_with_manage_permission.each_with_object({}) do |domain, hash|
      hash[domain.id] = domain.path_name
    end
  
    # Generate CSV content
    objects.where(filter).order(:name).map do |o|
      # Get the domain name using the id; fallback to id if not found
      sensor_name = domain_map[o.sensor_id] || o.sensor_id
  
      if o.is_a?(MacObject)
        object_type = o.object_type.presence || "N/A"
        "#{o.name};#{o.value};#{o.comment};#{object_type};#{sensor_name};#{o.operating_system};#{o.ip_addresses}"
      else
        "#{o.name};#{o.value};#{o.comment}"
      end
    end.join("\n")
  end

  #
  # Allowed params when create an object
  #
  def object_params
    params.require(:object).permit(:page, :search, :name, :value, :comment, :type, :user_id, :object_type, :sensor_id, :operating_system, ip_addresses: [])
  end
end
