#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class ReportBlocksController < ApplicationController
  include Redborder::Rbdruid
  include WidgetsHelper
  helper Redborder::ModulesHelper

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_report_block, only: [:edit, :update, :destroy,
                                          :common_edit, :image, :clone,
                                          :create_clone]
  before_action :set_report
  # Check the permission to write in the report
  before_action :check_write_permissions, only: [:edit, :update, :destroy,
                                                 :common_edit, :create_clone]
  before_action :init_writable_reports, only: [:new, :edit, :clone]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  # GET /report_blocks/new
  def new
    @widget = ReportBlock.new(
      # Default
      top_num: 5,
      width: ReportBlock::WIDTH_MEDIUM,
      timezone: 'auto',
      reloadeable: true,
      transparent: false,
      show_total: true,
      border: true,
      show_name: true,
      ball_max: 100,
      keep_aspect_ratio: true,
      square_opacity: 0.5,
      square_fill: 'FFFFFF',
      square_border: '000000',
      square_border_enabled: true,
      map_type: :roads,
      limit: 100
    )

    if params[:from]
      modules = Redborder::Module.keys_with_interface
      @widget.product_type = (params[:product_type] || modules
        .map(&:to_s).first)

      section = begin
        Redborder::Module
        .registered_modules[@widget.product_type.to_sym][:session_id].to_sym
      rescue
        Redborder::Module.registered_modules[modules.first][:session_id].to_sym
      end

      @widget.filters = filters_items(session[section][:filter]
        .except(:start_time, :end_time)).to_json
      @widget.graph_type = session[section][:chart_type][params[:from]]

      @widget.stream_columns = session[section][:active_columns].map(&:to_s) if
        @widget.graph_type.to_sym == :streams

      @widget.report_block_type =
        case params[:from].to_sym
        when :tops then 'Top'
        when :cardinality then 'Cardinality'
        when :comparative, :compare then 'Compare'
        else 'Raw'
        end

      @widget.column = @widget.category = session[section][:tab]
      @widget.column = session[section][:cardinality_tab] if
        params[:from].to_sym == :cardinality
      @widget.aggregation = params[:units] unless params[:units].nil?
      @widget.cardinality_group_by = params[:group_by] if
        params[:from].to_sym == :cardinality
      @widget.show_total = session[section][:hide_total] == 'false'
    end

    @widget.granularity = params[:granularity] if params[:granularity]
    @widget.range = params[:range] if params[:range]
  end

  #
  # => Permit the edition of a block
  #
  def edit
  end

  #
  # => Return the form to edit a common block
  #
  def common_new
    # Get the thype of the common block
    type = params.require(:id).to_i

    return unless [ReportBlock::TYPE_CUSTOM, ReportBlock::TYPE_TITLE,
                   ReportBlock::TYPE_TEXT, ReportBlock::TYPE_SEPARATOR,
                   ReportBlock::TYPE_IMAGE].include? type

    # Random name
    name = (0...10).map do
      (('a'..'z').to_a + (0..9).to_a + ('A'..'Z').to_a)[rand(62)]
    end.join

    @widget =
      ReportBlock.new(report_type: type,
                      name: name,
                      width: ReportBlock::WIDTH_MEDIUM
      )

    if @widget.is_title?
      @widget.text_alignment = ReportBlock::ALIGNMENT_LEFT
      @widget.width = ReportBlock::WIDTH_EXTRA_LARGE
    end

    @widget.text_alignment = ReportBlock::ALIGNMENT_JUSTIFY if @widget.is_text?

    # Stop the loading of the layout
    render layout: false
  end

  #
  # => Render the edit form for a common widget
  #
  #    @route     /reports/:report_id/blocks/:id/common/edit
  #    @helper    common_edit_report_report_block_path
  #
  def common_edit
    render layout: false
  end

  #
  # => Create a new custom block in the database. If the current block is custom
  #    the user will be redirected to the edition page of the report
  #
  #    @route   POST /reports/:report_id/blocks
  #    @helper  report_report_blocks_path
  #
  def create
    if params['report_block']['stream_columns']
      params['report_block']['stream_columns'] =
        params['report_block']['stream_columns'][0].split(',')
    end
    @widget = ReportBlock.new(report_block_params)
    @widget.report = @report

    # Check the permissions when the widget is created
    check_write_permissions

    respond_to do |format|
      if @widget.save

        if @widget.is_custom?
          format.html do
            redirect_to composition_report_path(@report),
                        notice: 'Report block was successfully created.'
          end
          format.json do
            render action: 'show', status: :created, location: @widget
          end
        else
          format.js do
            render file: 'report_blocks/created_common.js.erb', layout: false
          end
        end

      else
        format.html { render action: 'new' }
        format.js do
          render file: 'report_blocks/error_created_common.js.erb',
                 layout: false
        end
        format.json do
          render json: @widget.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Return the image of a report block
  #
  def image
    return unless @widget.image.present?
    img = ''
    Timecop.return do
      img = open(@widget.image_uri,
                 ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE).read
    end
    render text: img, content_type: @widget.image.content_type
  end

  #
  # => Update the current widget
  #
  #    @route   PATCH   /reports/:report_id/blocks/:id
  #    @helper  report_report_block_path
  #
  def update
    if params['report_block']['stream_columns']
      params['report_block']['stream_columns'] =
        params['report_block']['stream_columns'][0].split(',')
    end
    respond_to do |format|
      if @widget.report_id != report_block_params[:report_id]
        # If we move the report block to another report, we need to
        # reset the page and position
        moved = true
      end

      if @widget.update(report_block_params)

        if moved
          # reset the page and position
          @widget.page = nil
          @widget.position = 0
        end

        if @widget.is_custom?
          format.html do
            redirect_to composition_report_path(@report),
                        notice: 'Report block was successfully created.'
          end
          format.json do
            render action: 'show', status: :created, location: @widget
          end
        else
          format.js do
            render file: 'report_blocks/created_common.js.erb', layout: false
          end
        end

      else
        format.html { render action: 'edit' }
        format.js do
          render file: 'report_blocks/error_created_common.js.erb',
                 layout: false
        end
        format.json do
          render json: @widget.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Destroy the current block. We must check if the user has the
  #    permissions to delete this item.
  #
  #    @route   DELETE  /reports/:report_id/blocks/:id
  #    @helper  report_report_block_path
  #
  def destroy
    # Save the block id
    @block_id = @widget.id

    # Destroy the block
    @widget.destroy

    # Render the response
    respond_to do |format|
      format.html { redirect_to edit_report_path(@report) }
      format.js do
        render file: 'report_blocks/delete_block.js.erb', layout: false
      end
    end
  end

  def clone
    @new_report_block = ReportBlock.new

    render layout: false
  end

  def create_clone
    @new_report_block = @widget.clone(params[:report_block][:report_id],
                                      params[:report_block][:name])

    # Check problems
    if @new_report_block.id.nil?

      init_writable_reports

    end

    # Render the response
    respond_to do |format|
      format.html { redirect_to report_path(@report) }
      format.js do
        render file: 'report_blocks/clone_block.js.erb',
               layout: false,
               notice: 'Report block was successfully cloned.'
      end
    end
  end

  private

  #
  # => Use callbacks to share common setup or constraints between actions.
  #
  def set_report_block
    @widget = ReportBlock.find(params[:id])
  end

  #
  # => Initiate the report
  #
  def set_report
    @report = Report.find(params[:report_id])
  end

  #
  # => Initiate the writable reports of a user
  #
  def init_writable_reports
    @reports = []
    @selected = 0
    current_user.writable_reports.select do |item|
      if item.id == @report.id
        @reports << ["#{item.name} (Current)", item.id]
        @selected = item.id
      else
        @reports << [item.name, item.id]
      end
    end
  end

  #
  # => Never trust parameters from the scary internet, only allow
  #    the white list through.
  #
  def report_block_params
    params.require(:report_block)
      .permit(:name, :description, :report_type, :width, :report_id, :column,
              :product_type, :position, :transparent, :hide_axises,
              :reference, :graph_type, :direction, :granularity, :aggregation,
              :range, :category, :url, :height, :show_total, :show_name,
              :ball_max, :compare_range, :color, :report_block_type, :top_num,
              :filters, :timezone, :reloadeable, :dashboard_id,
              :cardinality_group_by, :sensor_table_type, :map_column,
              :click_action, :click_dashboard, :image, :keep_aspect_ratio,
              :square_fill, :square_border_enabled, :border, :square_border,
              :square_opacity, :map_type, :loop_video, :limit, :title, :text,
              :text_alignment, stream_columns: [])
  end

  #
  # => It's check the permissions of the user
  #    If the user hasn't got permissions, he will
  #    be redirected to the index
  #
  def check_write_permissions
    # Reports the user can read
    available_reports = current_user.writable_reports

    return if available_reports.include?(@report) &&
              available_reports.include?(@widget.report)

    respond_to do |format|
      format.html do
        redirect_to reports_path,
                    alert: "#{I18n.t('controllers.no_privileges')}"
      end
      format.json do
        render json: { error: "#{I18n.t('controllers.no_privileges')}" },
               status: :unauthorized
      end
      format.js do
        render file: 'report_blocks/unauthorized.js.erb',
               alert: "#{I18n.t('controllers.no_privileges')}",
               status: :unauthorized
      end
    end
  end
end
