#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'aws-sdk-s3'

#
# => Controls the creation and different data visualization
#    of sensors.
#
class SensorsController < ApplicationController
  # Includes
  include Redborder::Namespaces

  ## ======================================================================
  # => Controller properties
  ## ======================================================================

  respond_to :html
  respond_to :js, :json, only: [
    :list, :destroy
  ]

  layout false, only: [
    :user_root_domains, :new, :new_domain, :domains_children
  ]

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_sensor, only: [
    :edit, :update, :change_parent, :destroy,
    :new_route, :create_route, :delete_route,
    :new_location, :add_location, :new_interface, :add_interface, :new_trap_server,
    :bypass,:segment_enabled, :create_trap_server, :delete_trap_server,
    :set_chef_status, :remove_location, :delete_observation_id_filter,
    :add_image, :delete_image, :new_social_filter, :new_observation_id_filter,
    :create_social_filter, :delete_social_filter, :create_observation_id_filter,
    :indexed_data, :indexed_data_last_month, :indexed_data_per_day,
    :save_exporter_data
  ]

  before_action :manage_privileges, only: [
    :edit, :update, :change_parent, :new_route,
    :create_route, :delete_route, :new_location, :new_interface, :add_interface,
    :add_location, :new_trap_server, :create_trap_server, :create_observation_id_filter,
    :delete_trap_server, :destroy, :delete_observation_id_filter,
    :new_social_filter, :create_social_filter, :new_observation_id_filter,
    :delete_social_filter
  ]

  before_action :check_domain_type, only: [:update]

  skip_before_action :authenticate_user!, raise: false, if: :skip_authenticate_user

  before_action :set_default_sensor_view, only: [:map, :view, :tree, :client_proxy, :exporter]

  before_action :set_sensors_per_page, only: [:client_proxy_load_more, :client_proxy, :exporter, :exporter_load_more]

  before_action :require_target_manage_privileges, only: [:change_parent]

  before_action :privileges_on_top_domain, only: [:create]

  before_action :destroying_own_top_domain, only: [:destroy]

  before_action :require_super_administrative_privileges,
                only: [:reregister, :indexed_data, :indexed_data_last_month, :indexed_data_per_day]

  after_action :store_location, only: :edit

  after_action :set_last_sensor_view, only: [:map, :view, :edit]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # Change the current user namespace. This method always will return a JSON
  #
  def namespaces
    # Check if params exist.
    return render json: { status: false } if
      params[:namespaces].nil? || params[:namespaces].size == 0

    # Load namespaces
    namespaces = params[:namespaces]
    # Check selected namespaces
    namespaces &= current_user.data_sources.map(&:uuid)
    # Apply it to the user
    current_user.active_data_sources = namespaces
    current_user.save

    render json: { status: true }
  end

  # GET /sensors
  # GET /sensors.json
  def index
    redirect_to action: session[:sensor_view]
  end

  #
  # This method is called from sensor map widget. We must return only lat
  # and lng attributes.
  #
  def list
    respond_with current_user
                 .sensors_with_permission
                 .reject { |s| s.latitude.nil? || s.longitude.nil? }
                 .map { |s| { lat: s.latitude,
                              lng: s.longitude,
                              title: "#{s.name}",
                              ip: "#{s.ip}",
                              id: "#{s.id}",
                              icon: ActionController::Base.helpers.asset_path(s.icon) } }
  end

  #
  # Save sensor tree state in the user model.
  #
  def save_sensor_tree_state
    if current_user.nil?
      return render json: { status: false, message: 'User not authenticated' }, status: 401
    end

    sensor_tree_state = params[:sensor_tree_state]
    if sensor_tree_state.nil? || sensor_tree_state.empty? || !sensor_tree_state.is_a?(String)
      return render json: { status: false, message: 'Invalid sensor_tree_state parameter' }, status: 400
    end

    begin
      parsed_tree = JSON.parse(sensor_tree_state)
      current_user.sensor_tree_state = parsed_tree
      if current_user.save
        render json: { status: true, message: 'Sensor tree state saved successfully' }, status: 200
      else
        render json: { status: false, message: 'Failed to save sensor tree state' }, status: 500
      end
    rescue JSON::ParserError => e
      render json: { status: false, message: 'Invalid JSON format' }, status: 400
    end
  end

  #
  # Render the sensors as tree.
  #
  def tree
    @sensors = if request.format == :json
                 current_user
                 .sensor
                 .self_and_descendants
                 .where(is_deleted: false)
               else
                 s = []
                 s + Sensor
                 .where(is_deleted: false, parent_id: current_user.sensor.id)
                 .reject { |s| (s.unassigned? && s.children.count.zero?) ||
                                s.is_root? }
               end

    # To add it to the route on add domain menu
    @sensor = if current_user.super_admin?
                Sensor.root
              else
                current_user.sensor
              end

    @search = params[:search]

    if @search.blank? && session[:selected_sensor].present?
      begin
        @selected_path = Sensor
                       .find(session[:selected_sensor]).ancestors
                       .select{ |s| current_user.can?(:read, s) &&
                                    current_user.sensor != s }
                       .map(&:id)
      rescue Exception => e
        Rails.logger.error "Error with the selected_path"
      end
      session.delete(:selected_sensor)
    end

    @sensor_tree_state = current_user.sensor_tree_state
    @sensor_tree_state = {} if @sensor_tree_state.nil?

    render layout: false if request.xhr?
  end

  #
  # GET /sensors/:id/search
  #
  # Search at Domains Tree
  #
  def search
    sensor_ids = params[:sensor_ids]
    
    @view_name = params[:view_name]
    @search = params[:search]
    @sensors = Sensor.where(id: sensor_ids).where('lower(name) LIKE ?', "%#{params[:search].downcase}%")
  end

  #
  # GET /sensors/search
  #
  # Search at Domain Tree and return them with the absolute path depending on
  # the user logged.
  #
  def global_search
    sensors =
      if params[:search].present?
        current_user.sensor.descendants.where(is_deleted: false)
        .where('lower(name) LIKE ?',
               "%#{params[:search].downcase}%").order(name: :asc)
        .map { |s| { path: s.path_name, uuid: s.uuid } }
      else
        []
      end

    respond_to do |format|
      format.json { render json: sensors }
    end
  end

  #
  # Manage searching by name at Sensor's tree
  #
  def tree_search
    @sensors =
      current_user.sensor.descendants.where(is_deleted: false)
      .where('lower(name) LIKE ?',
             "%#{params[:search].downcase}%") if params[:search].present?
    return unless @sensors && @sensors.count == 1
    @path = @sensors.first.ancestors
            .select { |s| current_user.can?(:read, s) && current_user.sensor != s }
            .map(&:id)

    respond_to do |format|
      format.js {}
      format.json { render json: @sensors }
    end
  end

  #
  # Expand all Sensor's tree
  #
  def expand_all_tree
    @sensors =
      current_user.sensor.descendants.where(is_deleted: false)
    return unless @sensors && @sensors.count > 1
    @paths = []
    @sensors.each{|sensor|
      # Check if any of the sensor's ancestors is deleted
      all_ancestors_active = sensor.ancestors.all?{|ancestor| !ancestor.is_deleted }
      if !sensor.is_deleted && all_ancestors_active
        expandable_nodes = sensor.ancestors.select{ |s| current_user.can?(:read, s) && current_user.sensor != s}.sort_by(&:level).map(&:id)
        if expandable_nodes.count > 0
          @paths.push(expandable_nodes);
        end
      end
    }

    @paths = @paths.uniq.reject{|p1| (@paths.uniq.reject{|p2| p1 == p2}).select{|p3| p3 & p1 == p1}.count > 0 }

    respond_to do |format|
      format.js {}
      format.json { render json: @sensors }
    end
  end

  #
  # Show sensors in a map with it's domains defined.
  #
  def map
    if params[:id]
      sensors = Sensor.find(params[:id])
                .children
                .where(is_deleted: false)
                .where.not(latitude: nil)
                .where.not(longitude: nil)
                .select { |s| current_user.can?(:read, s) }
      @this_domain_zones = Sensor.find(params[:id]).domain_zones
      domain_sensors = Sensor.find(params[:id]).children
                       .where(type: Sensor::DOMAIN)
                       .select { |s| s.domain_zones.present? && current_user.can?(:read, s) }
    else
      sensors = Sensor.where(is_deleted: false, parent_id: Sensor.root.id)
                .where.not(latitude: nil)
                .where.not(longitude: nil)
                .select { |s| current_user.can?(:read, s) }
      domain_sensors =
        Sensor.where(type: Sensor::DOMAIN, parent_id: Sensor.root.id)
        .select { |s| s.domain_zones.present? && current_user.can?(:read, s) }
    end

    @locations =  sensors
                  .map do |m|
                    {
                      lng: m.longitude,
                      lat: m.latitude,
                      title: "#{m.name}",
                      ip: "#{m.ip}",
                      id: "#{m.id}",
                      type: m.type,
                      icon: ActionController::Base.helpers.asset_path(m.icon)
                    } unless m.latitude.nil?
                  end.compact

    @domain_zones =
      domain_sensors.collect do |s|
        {
          domain_zones: JSON.parse(s.domain_zones),
          title: s.name,
          id: s.id,
          icon: Sensor::DOMAIN_TYPES[s.domain_type][:icon],
          type_icon: Sensor::DOMAIN_TYPES[s.domain_type][:type_icon],
          color: Sensor::DOMAIN_TYPES[s.domain_type][:color]
        }
      end
  end

  #
  # Returns all readable client proxy sensors
  #
  def client_proxy
    # The current page
    @page = 0

    # The current letter
    @current_letter = ''

    if params[:id].nil?
      sensor_ids = current_user.sensor.all_children(true).where(type: Sensor::CLIENT_PROXY, is_deleted: false).map(&:id)
    else
      @sensor = Sensor.find(params[:id])
      sensor_ids = @sensor.all_children(true)
                   .where(type: Sensor::CLIENT_PROXY, is_deleted: false)
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
    end

    # Use unscoped to remove the ordering setted in sensor.rb
    @sensors = Sensor.unscoped
                   .where(id: sensor_ids)
                   .order('LOWER(name) ASC')
                   .take(@sensors_page)

    @sensors = [@sensors] unless @sensors.is_a? Array
  end

  #
  # Returns all readable exporter sensors
  #
  def exporter
    # The current page
    @page = 0

    # The current letter
    @current_letter = ''

    if params[:id].nil?
      sensor_ids = current_user.sensor.all_children(true).where(type: Sensor::EXPORTER, is_deleted: false).map(&:id)
    else
      @sensor = Sensor.find(params[:id])
      sensor_ids = @sensor.all_children(true)
                   .where(type: Sensor::EXPORTER, is_deleted: false)
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
    end

    # Use unscoped to remove the ordering setted in sensor.rb
    @sensors = Sensor.unscoped
                   .where(id: sensor_ids)
                   .order('LOWER(name) ASC')
                   .take(@sensors_page)

    @sensors = [@sensors] unless @sensors.is_a? Array
  end

  #
  # Load more client proxy sensors
  #
  def client_proxy_load_more

    # Get the page
    @page = params[:page].to_i

    # Get the sensors
    if params[:sensor_id].nil?
      sensor_ids = current_user.sensor.all_children(true).where(type: Sensor::CLIENT_PROXY, is_deleted: false).map(&:id)
    else
      @sensor = Sensor.find(params[:sensor_id])
      sensor_ids = @sensor.all_children(true)
                   .where(type: Sensor::CLIENT_PROXY, is_deleted: false)
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
    end

    @sensors = Sensor.unscoped.where(id: sensor_ids).order('LOWER(name) ASC').offset(@page * @sensors_page).take(@sensors_page)

    # Set the current letter empty
    before_sensor = Sensor.unscoped.where(id: sensor_ids).order(:name).offset((@page * @sensors_page) - 1).first

    if before_sensor.nil?
      @current_letter = ''
    else
      @current_letter = before_sensor.secondary_name[0,1].upcase
    end

    respond_to do |format|
      format.js {}
      format.json { render json: @sensors }
    end
  end

  #
  # Load more exporter sensors
  #
  def exporter_load_more

    # Get the page
    @page = params[:page].to_i

    # Get the sensors
    if params[:sensor_id].nil?
      sensor_ids = current_user.sensor.all_children(true).where(type: Sensor::EXPORTER, is_deleted: false).map(&:id)
    else
      @sensor = Sensor.find(params[:sensor_id])
      sensor_ids = @sensor.all_children(true)
                   .where(type: Sensor::EXPORTER, is_deleted: false)
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
    end

    @sensors = Sensor.unscoped.where(id: sensor_ids).order('LOWER(name) ASC').offset(@page * @sensors_page).take(@sensors_page)

    # Set the current letter empty
    before_sensor = Sensor.unscoped.where(id: sensor_ids).order(:name).offset((@page * @sensors_page) - 1).first

    if before_sensor.nil?
      @current_letter = ''
    else
      @current_letter = before_sensor.secondary_name[0,1].upcase
    end

    respond_to do |format|
      format.js {}
      format.json { render json: @sensors }
    end
  end

  #
  # Client Proxy: search by initial letter
  #
  def client_proxy_letter_search
    # Get the letter
    @letter = params[:id]

    # Check the letter
    if [*"A".."Z"].include? @letter
      sensor_ids = if params[:sensor_id].nil?
                     current_user.sensor.all_children(true).where(type: Sensor::CLIENT_PROXY, is_deleted: false).where("LOWER(name) LIKE ?", "#{@letter.downcase}%").map(&:id)
                   else
                     @sensor = Sensor.find(params[:sensor_id])
                     @sensor.all_children(true)
                     .where(type: Sensor::CLIENT_PROXY, is_deleted: false)
                     .where("LOWER(name) LIKE ?", "#{@letter.downcase}%")
                     .select{|s| current_user.can?(:read, s)}.map(&:id)
                   end
      @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
      render :client_proxy_search
    else
      # Something went wrong
      redirect_to client_proxy_sensors_path
    end
  end

  #
  # Exporter: search by initial letter
  #
  def exporter_letter_search
    # Get the letter
    @letter = params[:id]

    # Check the letter
    if [*"A".."Z"].include? @letter
      sensor_ids = if params[:sensor_id].nil?
                     current_user.sensor.all_children(true).where(type: Sensor::EXPORTER, is_deleted: false).where("LOWER(name) LIKE ?", "#{@letter.downcase}%").map(&:id)
                   else
                     @sensor = Sensor.find(params[:sensor_id])
                     @sensor.all_children(true)
                     .where(type: Sensor::EXPORTER, is_deleted: false)
                     .where("LOWER(name) LIKE ?", "#{@letter.downcase}%")
                     .select{|s| current_user.can?(:read, s)}.map(&:id)
                   end
      @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
      render :exporter_search
    else
      # Something went wrong
      redirect_to exporter_sensors_path
    end
  end

  def client_proxy_search
    @search = params[:search]
    sensor_ids = if params[:sensor_id].nil?
                   current_user.sensor.all_children(true).where(type: Sensor::CLIENT_PROXY, is_deleted: false)
                       .where("LOWER(name) LIKE ?", "%#{@search.downcase.split.join('%')}%").map(&:id)
                 else
                   @sensor = Sensor.find(params[:sensor_id])
                   @sensor.all_children.where(type: Sensor::CLIENT_PROXY, is_deleted: false)
                   .where("LOWER(name) LIKE ?", "%#{@search.downcase.split.join('%')}%")
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
                 end
    @sensors = Sensor.where(id: sensor_ids).order("LOWER(name) ASC")
  end

  def exporter_search
    @search = params[:search]
    sensor_ids = if params[:sensor_id].nil?
                   current_user.sensor.all_children(true).where(type: Sensor::EXPORTER, is_deleted: false)
                       .where("LOWER(name) LIKE ?", "%#{@search.downcase.split.join('%')}%").map(&:id)
                 else
                   @sensor = Sensor.find(params[:sensor_id])
                   @sensor.all_children.where(type: Sensor::EXPORTER, is_deleted: false)
                   .where("LOWER(name) LIKE ?", "%#{@search.downcase.split.join('%')}%")
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
                 end
    @sensors = Sensor.where(id: sensor_ids).order("LOWER(name) ASC")
  end


  #
  # Show statistics about indexed data from Namespaces and data sources
  #
  def indexed_data
    @month = params['month'] ? params['month'].to_time : Time.now
    @now = Time.now
    @segments = DruidSegments.new(@month, @sensor)
    # Resume table
    @total = @segments.total
    max_namespace = @segments.by_namespaces.values.max
    max_topic = @segments.by_topics.values.max
    @biggest = {
      namespace:
        @segments.by_namespaces.select { |_, v| v == max_namespace }.keys.first,
      topic: @segments.by_topics.select { |_, v| v == max_topic }.keys.first
    }
    # Graphs
    @start_time = @month.beginning_of_month
    @end_time = @start_time + 1.month
    @timeserie_accumulated = @segments.timeserie_accumulated
    @timeserie_per_day = @segments.timeserie_per_day
  end

  #
  # Return the data of Druid consumption in last month to represent it in
  # current month (Last month in graph).
  #
  # The returned format is JSON.
  #
  def indexed_data_last_month
    render action: indexed_data if request.format == :html
    @month = params['month'] ? params['month'].to_time : Time.now
    @month -= 1.month
    @segments = DruidSegments.new(@month, @sensor)
    @timeserie_accumulated = []
    last = 0
    last_time = false
    @segments.timeserie_accumulated.each do |event|
      time = event[:timestamp].to_time + 1.month
      if last_time
        #
        # I need to perform this iteration to fill all empty values. How we can
        # get an empty value? For example, if I try to assign the data from
        # september to october domain I have a problem because october has 31
        # days and september only 30. Same problem appears on timezone change
        # (Spain)
        # - Angel
        while (time + 1.hour) < last_time
          last_time -= 1.hour
          @timeserie_accumulated << {
            event: { events: last },
            timestamp: last_time.to_s
          }
        end
      end
      last_time = time
      last = event[:event][:events]
      event[:timestamp] = (event[:timestamp].to_time + 1.month).to_s
      @timeserie_accumulated << event
    end
  end

  #
  # Return the data of Druid consumption in last month (per day) to represent it in
  # current month (Last month in graph).
  #
  # The returned format is JSON.
  #
  def indexed_data_per_day
    render action: indexed_data if request.format == :html
    @month = params['month'] ? params['month'].to_time : Time.now
    @month -= 1.month
    @segments = DruidSegments.new(@month, @sensor)
    @timeserie_per_day = []
    @segments.timeserie_per_day.each do |event|
      event[:timestamp] = (event[:timestamp].to_time + 1.month).to_s
      @timeserie_per_day << event
    end
  end

  #
  # Create a new empty sensor.
  #
  def new
    @sensor = Sensor.new
  end

  def new_domain
    @parent_id = params[:id] if params[:id]
    @parent = @parent_id ? Sensor.find(@parent_id) : Sensor.root
    @sensor = Sensor.new(parent: @parent,
                         type: Sensor::DOMAIN,
                         domain_type: params[:domain_type] || @parent.children_type)

    render :new
  end

  # GET /sensors/1/edit
  def edit
    # @overlay_map = OverlayMap.find_or_initialize_by(sensor: @sensor)
    render "edit_#{@sensor.type}"
  rescue ActionView::MissingTemplate, NoMethodError
    DittoCode::Exec.if 'ENTERPRISE' do
      # Load rules
      initialize_namespace_rules if current_user.super_admin?
      @consumption = consumption_data if @sensor.organization?
    end
    render :edit
  end

  # POST /sensors
  # POST /sensors.json
  def create
    @sensor = Sensor.new(sensor_params)
    @sensor.floor_level = 0 if @sensor.floor? && @sensor.floor_level.nil?
    if params[:sensor][:parent_id].blank?
      @sensor.parent_id =
        (current_user.super_admin? ? Sensor.root.id : current_user.sensor.id)
    else
      @parent_id = params[:sensor][:parent_id]
    end
    @permission = true
    # Fetch zones in model
    fetch_zones(@sensor)

    respond_to do |format|
      if current_user.cannot?(:manage, @sensor.parent)
        @permission = false
        # No permissions on parent.
        format.html do
          redirect_to root_path,
                      alert: "#{I18n.t('controllers.no_privileges')}"
        end
        format.json do
          render json: { error: "#{I18n.t('controllers.no_privileges')}" },
                 status: :unauthorized
        end
        format.js do
          flash[:error] = "#{I18n.t('controllers.no_privileges')}"
        end
      elsif @sensor.save
        # You need to reload the user sensor rgt value before rendering menus
        # in order to set manage privileges for the current user
        current_user.reload

        format.html do
          redirect_to sensors_path,
                      notice: "#{@sensor.is_domain? ? "#{I18n.t('controllers.domain')}" : 'Sensor'} #{I18n.t('controllers.successfully')}"
        end
        format.json { head :no_content }
        format.js
      else
        format.html { render action: 'new' }
        format.json do
          render json: @sensor.errors, status: :unprocessable_entity
        end
        format.js
      end
    end
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    # GET /sensors/:id/claim
    def new_claim
      @sensor = Sensor.new(parent_id: params[:id])
      render layout: false
    end

    def claim
      parent_sensor = Sensor.find(params[:sensor][:parent_id])
      @sensor =
        Sensor.where(unclaimed: true)
        .where('property like ?', "%\"mac_address\":\"#{params['sensor']['mac_address']}\"%")
        .first

      if @sensor && current_user.can?(:manage, parent_sensor)
        @sensor.update(
          name: params[:sensor][:name],
          unclaimed: false,
          parent_id: parent_sensor.id
        )
      elsif @sensor
        @sensor.errors.add(:parent_id, 'is not allowed')
        flash[:error] = "#{I18n.t('controllers.no_privileges')}"
      else
        flash[:error] = "#{I18n.t('controllers.uuid')}"
      end
    end
  end

  # PATCH/PUT /sensors/1
  # PATCH/PUT /sensors/1.json
  def update
    # Update the parent by drag and drop
    respond_to do |format|
      # Store zones
      fetch_zones(@sensor)

      if @sensor.update(sensor_params)
        # Try to save namespace rules
        if params['namespace_rules'] && current_user.super_admin?
          data = JSON.parse(params['namespace_rules']).map { |el| JSON.parse(el) }
          rules = DruidConfig::Entities::RuleCollection.new
          data.each do |el|
            rules << DruidConfig::Entities::Rule.parse(el)
          end
          ns = @sensor.namespaces(current_user).first
          real_namespace = (ns == Sensor.root) ? '' : "_#{ns.uuid}"
          # Initialize the DruidConfig gem
          Redborder::Cluster.instance
          # Create rules for a non existent datasource
          RedBorder::DRUID_CONFIG.map { |_, v| v['datasource'] rescue nil }.compact.each do |topic|
            datasource = DruidConfig::Entities::DataSource.new(
              { 'name' => "#{topic}#{real_namespace}", 'properties' => '{}' }, {})
            datasource.update_rules(rules)
          end
        end

        # save or update the overlay
        if params[:overlay_map]
          overlay_map =
            OverlayMap.where(sensor: @sensor)
            .first_or_create(name: @sensor.name,
                             ne: params[:overlay_map][:ne],
                             sw: params[:overlay_map][:sw],
                             image: params[:overlay_map][:image])

          overlay_map_params = {
            ne: params[:overlay_map][:ne],
            sw: params[:overlay_map][:sw],
            enabled: params[:overlay_map][:enabled]
          }

          image = params[:overlay_map][:image]
          overlay_map_params[:image] = image unless image.nil?

          overlay_map.update(overlay_map_params)
        end

        # DEPRECATED ----------------
        # We must use has_chef_role_property in models and update the views
        if params[:role]
          @sensor.update_chef_role(params[:role])
          @sensor.wakeup_chef
        end

        format.html do
          redirect_to sensors_path,
                      notice: "#{@sensor.is_domain? ? "#{I18n.t('controllers.domain')}" : 'Sensor'} #{I18n.t('controllers.successfully_updated')}"
        end
        format.js do
          flash[:notice] = "#{@sensor.is_domain? ? "#{I18n.t('controllers.domain')}" : 'Sensor'} #{I18n.t('controllers.successfully_updated')}"
        end
        format.json { head :no_content }
      else
        format.html do
          redirect_to edit_sensor_path(@sensor),
                      alert: @sensor.errors
                        .full_messages
                        .map { |msg| "<li>#{msg}</li>" }.join
        end
        format.json do
          render json: @sensor.errors, status: :unprocessable_entity
        end
      end
    end
  end

  def change_parent
    respond_to do |format|
      if @sensor.update(parent_id: params[:parent_id])
        format.html do
          redirect_to edit_sensor_path(@sensor),
                      notice: "#{@sensor.is_domain? ? "#{I18n.t('controllers.domain')}" : 'Sensor'} #{I18n.t('controllers.successfully_updated')}"
        end
        format.js do
          flash[:notice] = "#{@sensor.is_domain? ? "#{I18n.t('controllers.domain')}" : 'Sensor'} #{I18n.t('controllers.successfully_updated')}"
        end
        format.json { head :no_content }
      else
        format.html do
          redirect_to edit_sensor_path(@sensor),
                      alert: @sensor.errors
                        .full_messages
                        .map { |msg| "<li>#{msg}</li>" }.join
        end
        format.js { head 422 }
        format.json do
          render json: @sensor.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # Set a sensor as destroyed (hidden on the UI).
  # It's set a Job to delete the current sensor.
  # If it's a domain, the users with top domain equals
  # to the sensor are assigned to the parent.
  #
  def destroy
    case @sensor.type
    when Sensor::DOMAIN
      @sensor.update_column(:is_deleted, true)
      Zone.destroy_all(sensor_id: @sensor.id)
      Delayed::Job.enqueue(Redborder::Jobs::DeleteSensorJob
                            .new(@sensor.id, current_user.id, true, I18n.locale),
                           priority: 10, run_at: Time.now)

      respond_to do |format|
        format.html { redirect_to sensors_url }
        format.json { render json: { query: true, id: @sensor.id } }
      end
    else
      send("destroy_#{@sensor.type}")
    end
  end

  # TODO: Must send info to proxy?
  define_method("destroy_#{Sensor::CLIENT_PROXY}") do
    @sensor.update_column(:is_deleted, true)

    Delayed::Job.enqueue(Redborder::Jobs::DeleteSensorJob
                          .new(@sensor.id, current_user.id, true, I18n.locale),
                         priority: 10, run_at: Time.now)

    respond_to do |format|
      format.html { redirect_to sensors_url }
      format.json { render json: { query: true, id: @sensor.id } }
    end
  end

  # TODO: Must send info to proxy?
  define_method("destroy_#{Sensor::EXPORTER}") do
    @sensor.update_column(:is_deleted, true)

    Delayed::Job.enqueue(Redborder::Jobs::DeleteSensorJob
                          .new(@sensor.id, current_user.id, true, I18n.locale),
                         priority: 10, run_at: Time.now)

    respond_to do |format|
      format.html { redirect_to sensors_url }
      format.json { render json: { query: true, id: @sensor.id } }
    end
  end

  #
  # Reregister a sensor will update the private key o physical node.
  # Only superadmins can perform this action.
  #
  def reregister
    @sensor = Sensor.find(params[:id])
    if @sensor.physical?
      @private_key = @sensor.reregister
      if @private_key.present?
        @result = "#{I18n.t('controllers.the_sensor')} #{@sensor.full_name} #{I18n.t('controllers.registered')} \n"\
                  "#{I18n.t('controllers.successfully2_html')}"\
                  "'/etc/chef/client.pem' #{I18n.t('controllers.file')}"
      else
        @result = "#{I18n.t('controllers.sensor_error')}"\
                  "#{@sensor.full_name} #{I18n.t('controllers.not_registered')}"
      end
    else
      @result = "#{I18n.t('controllers.the_sensor')} #{@sensor.full_name} #{I18n.t('controllers.physical_node')}"
    end

    render layout: false
  end

  def children
    @sensors =
      if current_user.super_admin?
        Sensor.find(params[:id].to_i)
        .children.where(is_deleted: false)
        .order(:name)
        .reject { |s| s.unassigned? && s.children.count.zero? }
      else
        sensor = Sensor.find(params[:id].to_i)
        return error_404("#{I18n.t('controllers.domain_dont_exist')}") if current_user.cannot?(:read, sensor)
        sensor.children
        .where(is_deleted: false)
        .order(:name)
        .reject { |s| s.unassigned? && s.children.count.zero? }
      end
    render layout: false
  end

  def new_route
    render layout: false
  end

  def new_location
    render layout: false
  end

  def add_location
    @sensor.update_attributes!(latitude: params[:sensor_location][:latitude],
                               longitude: params[:sensor_location][:longitude])
  end

  def remove_location
    @sensor.update_attributes!(latitude: nil, longitude: nil)
  end

  def new_interface
    @iface = params[:iface]
    role = @sensor.chef_role

    @management_iface = @sensor.default_iface.to_s
    @iface_data = {}
    @iface_data['arp'] = 'Disabled'
    @iface_data['iface_type'] = 'DHCP'
    @iface_data['iface_ip'] = ''
    @iface_data['iface_netmask'] = ''
    @iface_data['iface_gateway'] = ''
    @iface_data['dstAddress'] = ''
    @iface_data['observationId'] = ''
    @iface_data['protocol_type'] = 'Netflow v5'
    if role.override_attributes['redborder']['interfaces'] && role.override_attributes['redborder']['interfaces'][@iface]
      @iface_data = @iface_data.merge(role.override_attributes['redborder']['interfaces'][@iface])
    end
    @iface_data['iface_observation_id_filters'] ||= []

    render layout: false
  end

  def add_interface
    @error = false
    iface = params[:iface]
    iface_config = {}

    unless iface
      @error = true
      flash[:error] = "#{I18n.t('controllers.exporter_update2')}"
    else
      # set default config values
      iface_config['iface_type'] = params[:iface_type]
      iface_config['iface_ip'] = params[:iface_ip] || ''
      iface_config['iface_netmask'] = params[:iface_netmask] || ''
      iface_config['iface_gateway'] = params[:iface_gateway] || ''
      iface_config['dstAddress'] = params[:dstAddress] || ''
      iface_config['protocol_type'] = params[:protocol_type]
      iface_config['arp'] = params[:arp] || 'Disable'
      iface_config['observationId'] = params[:observationId] || ''
      iface_config['sampling_rate'] = params['sampling_rate'] || ''

      # Validations...
      # Check that the dstAddress is correct and the protocol type is correct
      if iface_config['dstAddress'] != ''
        @error = iface_config['dstAddress'].match(/\A(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9]):[0-9]+\z/).nil?
      end
      @error = !['NetFlow v5', 'NetFlow v9', 'NetFlow v10', 'sFlow'].include?(iface_config['protocol_type'])
      # Check that sFlow contains sampling rate
      @error = iface_config['protocol_type'] == 'sFlow' && (iface_config['sampling_rate'].nil? || iface_config['sampling_rate'].to_i < 1)
      # Check the iface_ip, iface_netmask, iface_gateway
      ['iface_ip', 'iface_netmask', 'iface_gateway'].each do |key|
        if (iface_config.key? key) &&
           (iface_config[key].match(/^((25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\.){3}(25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)$/x).nil? &&
            iface_config[key] != '')
          @error = true
        end
      end
      # Check observationID
      @error = iface_config['observationId'] && iface_config['observationId'].to_i < 0
      # Check that iface_span_port is true or false
      @error = iface_config['iface_type'] && iface_config['iface_type'] == 'Static' && iface_config['iface_ip'] == ''
    end

    if @error
      flash[:error] = "#{I18n.t('controllers.exporter_update')}"
    else
      role = @sensor.chef_role
      role.override_attributes['redborder']['interfaces'] ||= {}

      if role.override_attributes['redborder']['interfaces'][iface] &&
         role.override_attributes['redborder']['interfaces'][iface]['observation_id_filters']
           iface_config['observation_id_filters'] = role.override_attributes['redborder']['interfaces'][iface]['observation_id_filters']
      else
        iface_config['observation_id_filters'] = []
      end

      if role.override_attributes['redborder']['interfaces'][iface]
        role.override_attributes['redborder']['interfaces'][iface].each { |key, value| iface_config[key] ||= value }
      end

      role.override_attributes['redborder']['interfaces'][iface] = iface_config
      role.save

      flash[:notice] = "#{I18n.t('controllers.sensor_updated')}"
    end

    respond_to :js
  end

  def new_observation_id_filter
    @iface = params[:iface]
    render layout: false
  end


  def create_observation_id_filter
    @error = false
    @iface = params[:iface]

    subnet = params[:observation_id_filter][:subnet]
    @error = subnet.empty? ||
             !subnet.match(/^((25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\.){3}(25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\/(3[0-2]|[1-2]?\d)$/x)

    if @error
      flash[:error] = "#{I18n.t('sensors.new_observation_id_filter.subnet_not_valid')}"
    else
      role = @sensor.chef_role
      # set default values
      role.override_attributes['redborder']['interfaces'] ||= {}
      role.override_attributes['redborder']['interfaces'][@iface] ||= {}

      observation_id_filters = role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters']

      @error = observation_id_filters &&
               observation_id_filters.any? { |filter| filter['subnet'] == params[:observation_id_filter][:subnet] }

      if @error
        flash[:error] = "#{I18n.t('sensors.new_observation_id_filter.subnet_already_defined')}"
      else
        observation_id_filter = {
                                  'subnet' => params[:observation_id_filter][:subnet],
                                  'observation_id' => (params[:observation_id_filter][:observation_id]).to_i
                                }
        filters = []
        if role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters']
          filters = role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters'].dup
        end
        filters << observation_id_filter

        role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters'] = filters
        role.save
      end
    end

    respond_to :js
  end

  def delete_observation_id_filter
    @iface = params[:iface]
    role = @sensor.chef_role

    observation_id_filters = role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters'].dup
    observation_id_filters.delete_at(params[:observation_id_filter_id].to_i)

    role.override_attributes['redborder']['interfaces'][@iface]['observation_id_filters'] = observation_id_filters
    role.save

    respond_to :js
  end

  def create_route
    @error = []
    role = @sensor.chef_role

    # Is the route valid?
    begin
      network = IPAddr.new(params[:sensor_route][:network], Socket::AF_INET)
      network = params[:sensor_route][:network]
    rescue
      @error << "#{I18n.t('controllers.invalid_network')}"
    end

    begin
      gateway = IPAddr.new(params[:sensor_route][:gateway], Socket::AF_INET)
    rescue
      @error << "#{I18n.t('controllers.invalid_gateway')}"
    end

    # Create network hash if it dont exist
    if role.override_attributes['redborder']['network'].nil?
      role.override_attributes['redborder']['network'] = {}
    end

    # Create routes hash if it dont exist
    if role.override_attributes['redborder']['network']['routes'].nil?
      role.override_attributes['redborder']['network']['routes'] = {}
    end

    # Save the route
    if @error.empty?
      role.override_attributes['redborder']['network']['routes'][network.to_s] = gateway.to_s
      role.override_attributes['redborder']['network']['configure_routes'] = true
    end

    role.save

    respond_to :js
  end

  def delete_route
    role = @sensor.chef_role

    aux = role.override_attributes['redborder']['network']['routes'].to_a
    aux.delete_at(params[:what].to_i)
    role.override_attributes['redborder']['network']['routes'] = Hash[aux]

    role.save
    respond_to :js
  end

  def new_trap_server
    render layout: false
  end

  def create_trap_server
    role = @sensor.chef_role
    trap_server = { 'ip' => params[:trap][:ip],
                    'community' => params[:trap][:community] }

    @found = false

    role.override_attributes['redborder']['snmp']['trap_servers']
      .to_a.each do |server|
      @found = (server['ip'] == params[:trap][:ip])
      break if @found
    end

    unless @found
      role.override_attributes['redborder']['snmp']['trap_servers'] ||= []
      role.override_attributes['redborder']['snmp']['trap_servers'] << trap_server
      role.save
    end

    respond_to :js
  end

  def delete_trap_server
    role = @sensor.chef_role
    role.override_attributes['redborder']['snmp']['trap_servers']
      .delete_at(params[:trap_id].to_i)
    role.save
    respond_to :js
  end

  def restart_chef_client
    unless DelayedJob.restart_chef_client?
      Delayed::Job.enqueue(Redborder::Jobs::RestartChefClientJob
                           .new(true, current_user.id, params[:id]))
    end
    redirect_to :back
  end

  def set_chef_status
    must_activate = (params[:status] == '1')

    @sensor.chef_enabled_set must_activate

    status = (@sensor.chef_enabled? == must_activate) ? 200 : 500

    render json: {}, status: status
  end

  def user_root_domains
    domains_roots = Sensor.root_children.where(type: Sensor::DOMAIN)

    # Select all roots domains that the user can see
    @sensors = domains_roots.select do |domain|
      current_user.domains_with_permission.include? domain
    end
  end

  def domains_children
    sensor_id = params[:sensor_id].to_i
    @sensors = Sensor.find(sensor_id).children.where(type: Sensor::DOMAIN)
  end

DittoCode::Exec.if 'ENTERPRISE' do

  def import_tree
    @sensor = Sensor.find(params[:id])
    return error_401("#{I18n.t('controllers.no_privileges')}") unless
      current_user.can?(:manage, @sensor)
    if request.get?
      render layout: false
    elsif request.post?
      s3_credentials = YAML.load_file("#{Rails.root}/config/aws.yml")[Rails.env]
      s3_credentials["bucket"].chomp!('/')
      s3_credentials.symbolize_keys!

      s3_client = Aws::S3::Client.new(
        access_key_id: s3_credentials[:access_key_id],
        secret_access_key: s3_credentials[:secret_access_key],
        region: 'us-east-1',
        endpoint: "https://" + s3_credentials[:s3_host_name],
        force_path_style: true,
        ssl_verify_peer: false
      )

      filename = params['imported_file']['tgz_file'].original_filename.gsub(' ', '_')
      s3_resource = Aws::S3::Resource.new(client: s3_client)
      bucket = s3_resource.bucket(s3_credentials[:bucket])
      object = bucket.object("imports/#{filename}")
      object.upload_file(params['imported_file']['tgz_file'].tempfile.path)

      Delayed::Job.enqueue(
        Redborder::Jobs::ImportTreeJob.new(
          true, @sensor.id,
          filename,
          params[:override],
          current_user.id
        ),
        priority: 10, run_at: Time.now
      )
      redirect_to sensors_path, notice: "#{I18n.t('controllers.job_queu')}"
    end
  end

  def export_tree
    if params[:id]
      sensor = Sensor.find(params[:id])
      return error_404("#{I18n.t('controllers.domain_dont_exist')}") if current_user.cannot?(:read, sensor)
    else
      sensor = current_user.sensor
    end

    # When the result is an array (ActiveRecord::Relation)
    if !(sensor.is_a? Sensor)
      sensors_json = sensor.flat_map do |s|
        s.as_json.merge!(children: s.children_json)
      end.to_json
      sensor_ids = sensor.flat_map do |s|
        s.all_children_ids(true)
      end
    elsif sensor.is_root?
      sensors_json = sensor.children_json.to_json
      sensor_ids = sensor.all_children_ids(true)
    else
      sensors_json =
        [sensor.as_json.merge!(children: sensor.children_json)].to_json
      sensor_ids = sensor.all_children_ids(true)
    end

    # Create structure
    dir = "/tmp/#{Time.now.to_i}"
    Dir.mkdir dir

    FileUtils.mkdir_p "#{dir}/images"
    FileUtils.mkdir_p "#{dir}/images/rotated"
    FileUtils.mkdir_p "#{dir}/otherCsv"
    FileUtils.mkdir_p "#{dir}/jsonFiles"

    File.open("#{dir}/jsonFiles/sensors.json", 'w') do |f|
      f.write(sensors_json)
    end

    # Create maps.json
    maps_json = {}

    OverlayMap.where(sensor_id: sensor_ids)
      .where.not(original_image_file_name: nil).each do |overlay_map|
        maps_json[overlay_map.sensor_id] = overlay_map.as_json

        # Include GPS Points
        maps_json[overlay_map.sensor_id].merge!(
          gps_markers: overlay_map.gps_markers.as_json
        )
        begin
          # Download original image
          Redborder::RulesUpdate.download_file(
            overlay_map.original_image_uri.to_s,
            "#{dir}/images/#{overlay_map.sensor_id}"
          )

          # Download rotated image
          Redborder::RulesUpdate.download_file(
            overlay_map.image_uri.to_s,
            "#{dir}/images/rotated/#{overlay_map.sensor_id}"
          )
        rescue OpenURI::HTTPError => e
          Rails.logger.error "Missing image #{overlay_map.image_uri.to_s} during export sensor tree"
        end

      end

    File.open("#{dir}/jsonFiles/maps.json", 'w') do |f|
      f.write(maps_json.to_json)
    end

    # Compress directory
    tar_file = "#{Rails.root}/tmp/Export_#{Time.now.to_i}_#{Random.rand(1000)}.tar.gz"
    system('tar', 'cvzf', tar_file, '-C', dir, 'images', 'otherCsv', 'jsonFiles')

    send_file tar_file
  end
end

  def add_image
    overlay_map =
      OverlayMap.create_with(
        name: params[:image].original_filename,
        ne: '0.000000,0.000000',
        sw: '0.000000,0.000000',
        image: params[:image].tempfile,
        original_image: params[:image].tempfile,
        sensor_id: @sensor.id
      ).find_or_create_by(sensor_id: @sensor.id)

    tempfile = params[:image].tempfile
    dimensions = Dimensions.dimensions tempfile
    @sensor.property[:dimensions] ||= {}
    @sensor.property[:dimensions][:width] = dimensions[0]
    @sensor.property[:dimensions][:height] = dimensions[1]
    @sensor.save
    render json: { status: true }
  end

  def delete_image
    overlay_map = OverlayMap.find_by(sensor_id: @sensor.id)
    overlay_map.destroy if overlay_map

    redirect_to :back
  end

  private

  def set_sensors_per_page
    @sensors_page = Sensor::SENSORS_PAGE
  end

  def skip_authenticate_user
    false
  end
  #
  # => Use callbacks to share common setup or constraints between actions.
  #
  def set_sensor
    @sensor = Sensor.find(params[:id])
  end

  def set_type_sensors
    :sensors
  end

  def set_default_sensor_view
    session[:sensor_view] = params[:action].to_sym
  end

  def set_last_sensor_view
    # Set sensor id in session
    return unless @sensor
    session[:selected_sensor] = @sensor.id
  end

  #
  # => Never trust parameters from the scary internet, only allow
  #    the white list through.
  #
  def sensor_params
    allowed_parameters = [:parent_id, :name, :domain_type, :general_description,
                          :company, :email, :floor_level, :mac_hashing_salt,
                          :index_partitions, :index_replicas, :zone]

    if current_user.super_admin?
      allowed_parameters << :visualization_range
      allowed_parameters << :megabytes_limit
      allowed_parameters << :active_modules
    end

    if @sensor
      respond_to?("sensor_params_#{@sensor.type}", :include_private) ?
        send("sensor_params_#{@sensor.type}") :
        params.require(:sensor)
          .permit(*allowed_parameters)
    else
      respond_to?("new_sensor_params_#{params[:sensor][:type].to_i}", :include_private) ?
        send("new_sensor_params_#{params[:sensor][:type].to_i}") :
        params.require(:sensor).permit(:type, :name, :domain_type, :parent_id, :floor_level, :zone)
    end
  end

  def store_location
    session[:previous_url] = request.env['HTTP_REFERER']
  end

  # Load consumption data from Monitor. This functionality requires MONITOR
  # module active
  #
  def consumption_data
    # Check if the module is active
    return [] unless Redborder::Module.registered_modules.include? :monitor
    filter = Redborder::Filter.new(SensorMonitor, {})
    filter.add(:organization_uuid, @sensor.uuid)
    @start_time = Time.now.utc.beginning_of_day
    @end_time = Time.now.utc.tomorrow.beginning_of_day

    data = SensorMonitor.where(
      query_type: :timeseries,
      range: :custom,
      start_time: @start_time,
      # We can't use here @end_time because is the future and R will be executed
      end_time: Time.now.utc,
      granularity: :hour,
      search: filter.get,
      monitor_tab: :organization_received_bytes,
      timezone: 'UTC'
    ).map { |e| { event: { events: e.sum_value }, timestamp: e.timestamp.utc } }

    # Accumulate data
    iterate = @start_time.to_i
    grouped = []
    value = 0

    data.reverse!.each do |e|
      while iterate < e[:timestamp].to_i
        grouped << { event: { events: value }, timestamp: Time.at(iterate).utc.to_s }
        iterate += 1.hour
      end
      value += e[:event][:events]
      grouped << { event: { events: value }, timestamp: Time.at(iterate).utc.to_s }
      # Increase hour
      iterate = e[:timestamp].to_i + 1.hour
    end
    # Return grouped
    grouped
  end

  #
  # Set the parent id. This method don't cehck any privileges, we must do this
  # on another methods
  #
  def set_parent_id
    @parent_id = if params[:id]
                   params[:id].to_i
                 else
                   current_user.sensor_id
                 end
  end

  # Save the zones of a sensor
  #
  def fetch_zones(sensor)
    return unless params[:domain_zones]
    sensor.property =
      sensor.property
            .merge!({ domain_zones: params[:domain_zones] }
                      .delete_if { |k, v| k.eql?(:zones) && v.blank? })
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def initialize_namespace_rules
      @namespace = @sensor.namespaces(current_user).first
      @namespace_name = (@namespace == Sensor.root) ? 'default' : @namespace.name
      @namespace = (@namespace == Sensor.root) ? 'default' : @namespace.uuid
      cluster = Redborder::Cluster.instance
      @tiers = cluster.tiers.map { |tier| Tier.default_tier_name(tier.name) }
      datasource =
        cluster.datasources.find do |el|
          namespace_name(el.name) == @namespace
        end
      if datasource
        @rules = datasource.rules
      else
        # Try to get from global rules
        global = cluster.rules.map(&:datasource).find do |el|
          namespace_name(el) == @namespace
        end

        if global
          global = global.split('_')
          global.pop
          global = global.join('_')
          @rules = DruidConfig::Entities::RuleCollection.new(
            cluster.rules.select { |el| el.datasource == "#{global}_#{@namespace}" })
        else
          @rules = []
        end
      end
    rescue DruidConfig::Exceptions::NotAvailableNodes
      @not_available_nodes = true
    end
  end

  #
  # Check if the user has manage privileges on a sensor
  #
  def manage_privileges
    return if current_user.admin? && current_user.can?(:manage, @sensor)
    error_404('')
  end

  #
  # If the user is not super administrator, the domain type can't be
  # changed to NAMESPACE
  #
  def check_domain_type
    return if current_user.super_admin?
    if params[:sensor][:domain_type] &&
       params[:sensor][:domain_type].to_i == Sensor::NAMESPACE &&
       @sensor.domain_type != Sensor::NAMESPACE

      respond_to do |format|
        format.html do
          redirect_to sensors_path,
                      alert: "#{I18n.t('controllers.no_privileges')}"
        end
        format.json do
          render json: { query: false,
                         error: "#{I18n.t('controllers.no_privileges')}" },
                 status: :unauthorized
        end
        format.js { head 403 }
      end

    end
  end

  def require_target_manage_privileges
    if params[:parent_id]
      sensor_parent = Sensor.find(params[:parent_id])

      if current_user.cannot? :manage, sensor_parent
        respond_to do |format|
          format.html do
            redirect_to root_path,
                        alert: "#{I18n.t('controllers.no_privileges')}"
          end
          format.json do
            render json: {error: "#{I18n.t('controllers.no_privileges')}"},
                   status: :unauthorized
          end
          format.js { head 403 }
        end
      end

      if @sensor.all_children(true).include?(sensor_parent) ||
         (sensor_parent.type == Sensor::DOMAIN &&
         sensor_parent.domain_type == Sensor::FLOOR) ||
         (sensor_parent.type == Sensor::CLIENT_PROXY &&
         @sensor.type == Sensor::DOMAIN) ||
         (sensor_parent.type == Sensor::EXPORTER &&
         @sensor.type == Sensor::DOMAIN)
        respond_to do |format|
          format.html do
            redirect_to root_path, alert: "#{I18n.t('controllers.cant_move')}"
          end
          format.json do
            render json: {error: "#{I18n.t('controllers.cant_move')}"},
                   status: :unauthorized
          end
          format.js { head 422 }
        end
      end
    end

    return unless @sensor.all_children(true).include? sensor_parent
    respond_to do |format|
      format.html do
        redirect_to root_path,
                    alert: "#{I18n.t('controllers.move_itself')}"
      end
      format.json do
        render json: { error: "#{I18n.t('controllers.move_itself')}" },
               status: :unauthorized
      end
      format.js { head 422 }
    end
  end

  #
  # Check privileges manage privileges or show an error.
  #
  def privileges_on_top_domain
    return if current_user.admin?
    respond_to do |format|
      format.html do
        render alert: "#{I18n.t('controllers.no_privileges')}"
      end
      format.json do
        render json: { error: "#{I18n.t('controllers.no_privileges')}" },
               status: :unauthorized
      end
      format.js { head 403 }
    end
  end

  #
  # Check if the user are trying to destroy his own top domain
  #
  def destroying_own_top_domain
    return if @sensor != current_user.sensor
    respond_to do |format|
      format.html do
        render alert: "#{I18n.t('controllers.top_domain')}"
      end
      format.json do
        render json: { error: "#{I18n.t('controllers.top_domain')}" },
               status: :unauthorized
      end
      format.js { head 403 }
    end
  end
end
