DittoCode::HoldFile.if 'ENTERPRISE'
#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Support integration with Zendesk
#
class SupportController < ApplicationController
  # Cluster UUID identifier
  CLUSTER_UUID_ZENDESK_ID = 28275421
  TYPE_OF_DEPLOYMENT_ZENDESK_ID = 24545879

  # -------------------------------------------------------------
  # Before actions
  # -------------------------------------------------------------

  before_action :check_zendesk_authentication,
                only: [:index, :new, :create]
  before_action :init_parameters_on_login, only: [:login_to_zendesk]
  before_action :init_colors, only: [:index]
  before_action :check_empty, only: [:create]

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  def index
    @zendesk = current_user.zendesk_client
  end

  def login
    redirect_to root_path
  end

  def login_to_zendesk
    redirect_to root_path
    # Init OAuth client
    oauthclient = OAuth2::Client.new(
      SUPPORT_IDENTIFIER,
      SUPPORT_SECRET,
      site: 'https://redborder.zendesk.com',
      token_url: '/oauth/tokens',
      authorize_url: '/oauth/authorizations/new')
    # Try to login
    begin
      token =
        oauthclient.password.get_token(@email, @password, scope: 'read write')
      current_user.update(ztoken: token.token)
      redirect_to support_index_path
    rescue OAuth2::Error => e
      Rails.logger.info 'Invalid OAuth Authentication:'
      Rails.logger.info e.description
      @invalid_credentials = true
      #render 'login'
    end
  end

  def logout
    redirect_to root_path
    current_user.update(ztoken: nil)
    #redirect_to login_support_index_path
  end

  def new; end

  def create
    request_data =
      params.require(:request).permit(:subject, :priority, :description, :type,
                                      :type_of_deployment)
    params[:cluster_info] = params[:cluster_info] == '1' ? true : false
    cluster_uuid = get_cluster_uuid
    request = ZendeskAPI::Request.new(
      current_user.zendesk_client,
      subject: request_data[:subject],
      comment: { value: request_data[:description] },
      priority: request_data[:priority],
      type: request_data[:type],
      custom_fields: [
        { id: CLUSTER_UUID_ZENDESK_ID, value: cluster_uuid },
        { id: TYPE_OF_DEPLOYMENT_ZENDESK_ID, value: request_data[:type_of_deployment] }
      ])
    # Add attachments
    tmp = attachments(request)
    # Save the request
    request.save
    FileUtils.rm_rf(tmp) if tmp
    redirect_to support_index_path
  end

  private

  #
  # Check selectors are filled
  #
  def check_empty
    fields = %i(priority type subject description type_of_deployment)
    return unless fields.any? { |f| params[:request][f].empty? }
    fields.each do |f|
      if params[:request][f].present?
        instance_variable_set("@#{f}", params[:request][f])
      else
        instance_variable_set("@#{f}_error", true)
      end
    end
    render :new
  end

  #
  # Initialize a hash with the spriority and the color associated
  #
  def init_colors
    @colors = Hash.new('#aaaaaa')
    Severity.all.each do |severity|
      @colors[severity.name] = severity.bg_color
    end
    @colors['normal'] = @colors['medium']
  end

  #
  # Add Cluster UUID to the request as a custom field
  #
  def get_cluster_uuid
    data = Redborder::Databag.load('global', 'cluster', false, false)
    return '' unless data
    data['uuid']
  end

  #
  # Attach files in the request to the Zendesk::Request
  #
  # == Returns:
  # String with TMP directory
  #
  def attachments(request)
    return nil unless params[:attachments].present? ||
                      (current_user.super_user? && params[:cluster_info])
    # Create tmp folder
    tmp = Rails.root.join 'tmp', (0...50).map { ('a'..'z').to_a[rand(26)] }.join
    FileUtils.mkdir(tmp)
    # Add attachments
    attach_files(request, tmp)
    # Add cluster info
    attach_cluster_info(request, tmp)
    # Return path
    tmp
  end

  #
  # Attach files to the request based on params
  #
  # == Parameters:
  # request::
  #   Zendesk Request
  # tmp::
  #   String with the path of TMP folder
  #
  def attach_files(request, tmp)
    return unless params[:attachments].present?
    params[:attachments].each do |file|
      new_path = "#{tmp}/#{file.original_filename}"
      FileUtils.mv(file.path, new_path)
      request.comment.uploads << new_path
    end
  end

  #
  # Attach cluster info. This option is only available for super users
  #
  # == Parameters:
  # request::
  #   Zendesk Request
  # tmp::
  #   String with the path of TMP folder
  #
  def attach_cluster_info(request, tmp)
    return unless current_user.super_user? && params[:cluster_info]
    s3 = Redborder::S3.new
    path = "#{tmp}/cluster_info"
    begin
      File.write(path, open(
        s3.object_url(:read, 'rbcheck/data-last'),
        ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE
      ).read)
      # Attach the path
      request.comment.uploads << path
    rescue OpenURI::HTTPError
      # The object is not found set an error
      Rails.logger.error 'rbcheck/data-last is NOT AVAILABLE in S3!'
    end
  end

  #
  # Check if the user is logged on Zendesk or redirect to login page
  #
  def check_zendesk_authentication
    redirect_to login_support_index_path unless current_user.zendesk_client
  end

  #
  # Check email and password are present
  #
  def init_parameters_on_login
    @email = params[:email]
    @password = params[:password]
    return if @email && @password
    # Redirect to login with error
    @email_error, @password_error = !@email, !@password
    render 'login'
  end
end
