#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class WidgetsController < ApplicationController
  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Controller properties
  ## ======================================================================

  helper DashboardHelper
  helper WidgetsHelper

  respond_to :html
  respond_to :js, :json, only: [
    :create, :destroy, :update
  ]

  layout false, only: [
    :clone, :reload, :list, :new,
    :graph, :edit, :types_graphs, :new_module
  ]

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_widget,
                only: [:destroy, :toggle, :reload,
                       :edit, :update, :set_absolute_position,
                       :image, :clone, :create_clone, :move, :remove_image]

  before_action :init_filters_and_settings, :load_parsed_columns, only: [:edit]

  before_action :set_clone_parent, only: [:create_clone]

  before_action :set_parent,
                only: [:create, :update, :list, :new, :set_position, :destroy,
                       :graph, :move, :edit, :types_graphs, :clone,
                       :set_absolute_position, :new_module]

  before_action :edit_permission,
                only: [:destroy, :toggle, :edit, :update,
                       :set_absolute_position, :clone, :create_clone,
                       :move, :types_graphs]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def new
    widget_type = params.fetch(:type, 'none')
    graph_type = params.fetch(
      :graph,
      WidgetType.all[widget_type.to_sym][:graphs].try(:first) || widget_type)
    product_type = params.fetch(:product, 'none')

    # If we come from report, we must save the number of page
    page = params.fetch(:page, 1)

    @widget = Widget.new(
      graph_type: graph_type,
      widget_type: widget_type,
      product_type: product_type,
      page: page
    )

    init_filters_and_settings if @widget.filtrable?
    load_parsed_columns
  end

  #
  # Create a widget from a external module. Receive all the params to build
  # the form and show a hidden form with a dashboard selector.
  #
  def new_module
    widget_type = params[:widget_type]
    graph_type = params[:graph_type]
    product_type = params[:id]

    @widget = Widget.new(
      graph_type: graph_type,
      widget_type: widget_type,
      product_type: product_type,
      parent: @parent
    )

    @widget.range = params[:range]
    @widget.granularity = params[:granularity]
    @widget.aggregation = params[:aggregation]
    @widget.client_profile = params[:client_profile]
    @widget.serie_type = params[:serie_type]
    @widget.column = params[:column]
    @widget.stream_columns = params[:stream_columns]
    @widget.group_by = params[:group_by]
    @widget.elements = params[:elements]

    # For monitor
    @widget.category = params[:category] if product_type == 'monitor'

    # Fetch filters
    filter = Redborder::Filter.new(Redborder::DruidSource.get(source: product_type),
                        load_filters)
    @widget.filters = filter.serialize(true, false, false)
    init_filters_and_settings
    load_parsed_columns if @widget.filtrable?
  end

  def create
    # We are going to check first what's going to be the default name for the widget. Name is going to be Sensor Name + Tab Name
    widget_params_aux = widget_params.clone
    default_name = ''

    json = ''
    if widget_params.key?"filters"
      json = JSON.parse(widget_params["filters"])
    end

    if widget_params.has_key?("name")
        widget_params_aux[:name] = widget_params[:name] if !widget_params[:name].empty?
    else #if we didnt specified any name, build one with Sensor Name + Tab Name
        # We check if we applied any filter
        if !json.empty?
          json.each_with_index do |filter, index|
            # We check only for the sensor name filter, so we put the sensor name in the widget title.
            if filter[1]['column'] == 'sensor_name'
              default_name = filter[1]["value"] + ' '
            end
          end
        end

        # We check if there is any column or category so we know what title we can give to the widget (Tab Name)
        if widget_params[:category]
          default_name << SensorMonitor::TABS[widget_params[:category].to_sym][:name] if widget_params[:product_type] == "monitor"
          widget_params_aux[:name] = default_name
        end

        if widget_params[:column]
          default_name << Social::COLUMNS[widget_params[:column].to_sym][:columns][widget_params[:column].to_sym] if  widget_params[:product_type] == "social"
          default_name << Vault::COLUMNS[widget_params[:column].to_sym][:columns][widget_params[:column].to_sym] if  widget_params[:product_type] == "vault"
          default_name << Traffic::COLUMNS[widget_params[:column].to_sym][:columns][widget_params[:column].to_sym] if  widget_params[:product_type] == "flow"
          default_name << Incident::COLUMNS[widget_params[:column].to_sym][:columns][widget_params[:column].to_sym] if  widget_params[:product_type] == "incidents"
          widget_params_aux[:name] = default_name
        end
    end

    @widget = Widget.new widget_params_aux
    @widget.parent = @parent
    @manage = manage_parent?(@parent)
    no_enough_privileges redirect_to: @parent.main_path and return unless @manage

    init_filters_and_settings
    #When implementing multilanguage, the Status parameter changes depending on the language, since it corresponds to the name of the button.
    if ["Save", "Guardar"].include? params['status']
      if @widget.save
        redirect_to :back # redirect to the same page
      end
    else
      if @widget.save
        rb_flash "#{I18n.t('controllers.widget_created')}"
        if @widget.dashboard?
          respond_with @widget, location: dashboards_path
        else
          respond_with @widget, location: composition_report_path(@widget.parent)
        end
      else
        load_parsed_columns
        rb_flash "#{I18n.t('controllers.invalid_image')}", type: :error
        if @widget.dashboard?
          redirect_to dashboards_path
        else
          redirect_to composition_report_path(@widget.parent)
        end
      end
    end
  end

  def edit
  end

  #
  # => Render the list of available widgets
  #
  def list
    @module = params[:module]
  end

  #
  # => Render the graph selector based on a product
  #
  def graph
    # TODO: Check the product
    @product = params[:product]
    @type    = params[:type]
  end

  #
  # => Render the graph selector based on a widget type
  #
  def types_graphs
    widget_type_name = WidgetType.all[params[:widget_type].to_sym][:name]
    product = Widget.find_by_name(widget_type_name)
    
    @product_type = nil
    unless product.nil?
      @product_type = product.product_type == 'none' ? nil : product.product_type
    end
    @module = params[:module]
    @widget_type = params[:widget_type].to_sym
    @page = params[:page]
  end

  def update
    if (params[:widget][:hidden_check] == "true")
      @widget.update(media_file_name: nil, media_content_type: nil, media_file_size: nil, media_updated_at: Time.now)
    end
    update_params = widget_params
    @manage = manage_parent?(@parent)
    # If we dont received the array properties on params, we
    # suppose they are empty, so we include them in order to remove
    # the Widget value associated to that property from the database
    Widget.array_properties.each do |prop|
      next if update_params.include? prop
      update_params[prop] = []
    end

    if @widget.update(update_params)
      rb_flash "#{I18n.t('controllers.widget_updated')}"

      if @widget.dashboard?
        init_filters_and_settings if @widget.filtrable?
        respond_with @widget, location: dashboards_path
      else
        respond_with @widget, location: composition_report_path(@widget.parent)
      end
    else
      rb_flash "#{I18n.t('controllers.invalid_image')}", type: :error
      if @widget.dashboard?
          redirect_to dashboards_path
      else
          redirect_to composition_report_path(@widget.parent)
      end
    end
  end

  def destroy
    @widget.destroy
    rb_flash "#{I18n.t('controllers.widget_removed')}"
  end

  def clone
    @new_widget = Widget.new
  end

  def create_clone
    @new_widget = @widget.dup
    @new_widget.parent = @parent
    @new_widget.name = params[:widget][:name]
    @new_widget.displace(20, 20) if @new_widget.dashboard? &&
                                    @new_widget.parent.absolute?
    @new_widget.page = params[:widget][:page] if @new_widget.report?
    @new_widget.save

    @manage = manage_parent?(@parent)
    rb_flash "#{I18n.t('controllers.widget_cloned')}" if @new_widget.save
  end

  def reload
  end

  #
  # Move a report widget to another page
  #
  def move
    head 404 if @widget.dashboard? || params[:page].nil?

    @widget.page = params[:page]
    @widget.save

    head 200
  end

  def image
    unless %w(hardwarestatus image video).include? @widget.graph_type
      head 200
      return
    end

    Timecop.return do
      render text: @widget.export_media, content_type: @widget.media.content_type
    end
  end

  def set_position
    data = JSON.parse params[:data]

    if data.present?
      widget = Widget.find(data.first['id'])
      @parent = widget.parent
      manage = manage_parent?(@parent)
      return head 403 unless manage
    end

    data.each do |w|
      widget = Widget.find w['id']
      next if widget.parent.dashboard? && (widget.parent.locked ||
              widget.parent.layout == 'absolute')

      widget.position_x = w['position_x'].to_i
      widget.position_y = w['position_y'].to_i
      widget.position_width = w['position_width'].to_i
      widget.position_height = w['position_height'].to_i
      widget.save
    end

    head 200
  end

  def set_absolute_position
    %i(x y width height).each do |a|
      @widget.send("position_#{a}=", params[a]) unless params[a].nil?
    end

    @widget.save!
    head 200
  end

  private

  def set_widget
      @widget = Widget.find(params[:id])
  end

  #
  # Check if the current user can manage the parent
  #
  def manage_parent?(parent)
    if parent.dashboard?
      current_user.writable_dashboards.include?(parent)
    else
      current_user.writable_reports.include?(parent)
    end
  end

  #
  # Init inactive settings and filters
  #
  def init_filters_and_settings
    return unless @widget.filtrable?
    product_type = @widget.product_type.eql?('all') ? Redborder::Module.keys_with_interface.first : @widget.product_type
    @source = Redborder::DruidSource.get(source: product_type)
    set_module_settings
  end

  #
  # Load enabled columns. @Settings must be initialized
  #
  # == Returns:
  # An array of symbols with the enabled columns
  #
  def load_columns
    return @source::COLUMNS if @inactive_settings.nil?
    @source::COLUMNS.select do |k, v|
      !v[:hide_raw] && !@inactive_settings.key?("#{@source.id}_#{k}") &&
      (!v[:admin] || (v[:admin] && User.current_user.super_admin?))
    end
  end

  def load_parsed_columns
    return unless %w(raw top single_unique grouped_unique).include?(@widget.widget_type)
    @columns = load_columns.except(
      :conversation, :scatterplot, :src_map, :dst_map, :public_ip_map, :coordinates_map
    ).to_a.map { |c| [c[1][:columns].to_a[0][1], c[0]] }.sort
  end

  def set_parent
    klass = [Report, Dashboard].detect { |c| params["#{c.name.underscore}_id"] }
    @parent = klass.find(params["#{klass.name.underscore}_id"])

    # Check permissions
    return if @parent.read_permission?
    no_enough_privileges redirect_to: @parent.main_path
  end

  #
  # We must set the clone parent based on the parent destination
  #
  def set_clone_parent
    # Check if we are on report or dashboard
    klass = [Report, Dashboard].detect { |c| params["#{c.name.underscore}_id"] }
    @parent = klass.find(params[:widget][:parent_id])

    # Check permissions
    return if @parent.read_permission?
    no_enough_privileges redirect_to: @parent.main_path
  end

  def edit_permission
    return if manage_parent?(@parent)
    no_enough_privileges redirect_to: @parent.main_path
  end

  def widget_params
    array_props = Widget.array_properties.each_with_object([]) do |prop, memo|
      memo << { prop => [] }
    end

    if params.require(:widget)[:stream_columns_ordered].present?
      params.require(:widget)[:stream_columns] = params.require(:widget)[:stream_columns_ordered].split(',')
    end
    params.require(:widget)
      .permit(Widget.properties + array_props +
              [:name, :dashboard_id, :width, :media, :parent_id, :parent_type])
  end
end

