#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Methods to help the creation of UI.
#
module ApplicationHelper
  # Includes
  include BreadcrumbHelper
  include CommonHelper
  include ErrorHelper
  include GraphHelper
  include TimeHelper
  include PermissionsHelper

  def filters_for_sensor(sensor)
    filters = {}
    return filters.to_json if sensor.is_root?

    filters[0] = {
      column: 'sensor_name',
      value: sensor.name,
      operator: 'is'
    }

    sensor.ancestors.where(domain_type: Sensor::DOMAIN_TYPES.keys).each_with_index do |ancestor, i|
      filters[i+1] = {
        column: "#{Sensor::DOMAIN_TYPES[ancestor.domain_type][:name].downcase.gsub(' ', '_')}_uuid",
        value: ancestor.uuid,
        operator: 'is'
      }
    end
    filters.to_json
  end

  def global_filters_for(key)
    return nil unless @global_filters.is_a? Hash
    @global_filters[key]
  end

  def advanced_global_filter_by(index: )
    return nil unless global_filters_for(:advanced)
    global_filters_for(:advanced)[index.to_s]
  end

  def dump_advanced_globals
    return [] unless @global_filters && @global_filters[:advanced]

    filters = []
    @global_filters[:advanced].each do |i, filter|
      filters << filter.merge({ 'operator' => 'is' })
    end

    @custom_global_filters.keys.each do |k|
      next unless @global_filters[k]
      filters << { 'column' => k, 'value' => @global_filters[k], 'operator' => 'is' }
    end

    filters
  end

  def simple_filters_expanded?
    session[:show_simple] == true
  end

  def advanced_filters_expanded?
    session[:show_advanced] == true
  end

  def options_for_columns
    columns = []
    Redborder::Module.columns_with_modules(hide_global: true).each do |col, info|
      next if @custom_global_filters && @custom_global_filters.keys.include?(col)
      columns << [info[:name], col, :'data-modules' => info[:modules].join(',')]
    end
    columns.sort!
  end

  def conversation_values(serie)
    columns = (Redborder::Module.get @source.id, :columns).keys
    if (columns & [:src, :dst]).any?
      [serie.send(:src), serie.send(:dst)]
    elsif (columns & [:wan_ip, :lan_ip]).any?
      [serie.send(:wan_ip), serie.send(:lan_ip)]
    end
  end

  def conversation_map(e)
    columns = (Redborder::Module.get @source.id, :columns).keys
    if (columns & [:src, :dst]).any?
      {src: e.src, dst: e.dst, timestamp: e.timestamp, value: e.events}
    elsif (columns & [:wan_ip, :lan_ip]).any?
      {lan_ip: e.lan_ip, wan_ip: e.wan_ip, timestamp: e.timestamp, value: e.send(@aggregation)}
    end
  end

  #
  # => Create a title for a section
  #
  def title(header, back=nil, title=nil, menu_content=true, editable=false, &block)
    show_title(title ? title : header)

    title_header = link_to header, '#', class: 'navbar-brand'

    unless back.nil?
      title_header = "<a href='#{back}' class='navbar-back'><i class='fa fa-level-up'></i></a>".html_safe + title_header
    end

    toggle_navbar =
      content_tag :div, class: 'navbar-header' do
        html_content = content_tag :button, class: "navbar-toggle collapsed", :"data-toggle" => "collapse", :"data-target" => ".collapse-title" do
          content_tag(:span, "#{I18n.t('application_helper.toggle')}", class: "sr-only") + content_tag(:span, nil, class: "icon-bar") + content_tag(:span, nil, class: "icon-bar") + content_tag(:span, nil, class: "icon-bar")
        end

        html_content + title_header
      end

    if block_given?
      data = (capture(&block) || '').gsub("\n|\t", '')
      if menu_content
        menu_holder = content_tag(:div, content_tag(:ul, "#{data}".html_safe, class: 'nav navbar-right navbar-nav'), class: "navbar-collapse collapse collapse-title")
        html = toggle_navbar + menu_holder
      else
        html = toggle_navbar + data.html_safe
      end
    else
      html = title_header
    end

    html += content_tag(:span, Time.zone.now.strftime('%Y-%m-%d %H:%M:%S'), class: "visible-print")
    html += content_tag(:h2, header, class: "visible-print text-center")

    container = if editable
                  content_tag(:div, html, class: "container-fluid", id: 'title-editable')
                else
                  content_tag(:div, html, class: "container-fluid")
                end
    content_tag(:nav, container, class: "navbar navbar-default")
  end

  def enable_forecast
    Time.now.utc.to_i > @end_time.utc.to_i + Redborder::Rbdruid::GRANULARITY_TYPE[@granularity.to_sym][:step]
    # Setting.forecast? || !Setting.forecast? &&
    #   Time.now.utc.to_i > @end_time.utc.to_i + Redborder::Rbdruid::GRANULARITY_TYPE[@granularity.to_sym][:step]
  end

  #
  # => Change the title of the page
  #
  def show_title(title)
    title = (title.to_s.start_with? "tr.") ? (I18n.t title.to_s[3..-1]) : (title)
    content_for(:title) { "#{title} | redborder Manager" } if title
  end

  def css_chart(percentage)
    begin
      percentage = 100.0 if percentage.to_f > 100.0
    rescue FloatDomainError
      percentage = 100.0
    end
    html = content_tag(:div, "<span>#{percentage}%</span>".html_safe, class: "bar", style: "width: #{percentage}%")
    content_tag(:div, html, :class => "progress progress-info")
  end

  #
  # => Return the number of filters
  #
  def filter_count(filter)
    count = 0

    filter.each do |key, value|
      count += value.size
    end

    count
  end

  # Get the number of loaded modules. License module is ignored
  #
  # @return [Integer] number of modules
  #
  def count_modules
    Redborder::Module.registered_modules.keys.size - 1
  end

  #
  # Return a hash with inputs to use on filters views like alarms,
  # widgets... This method is used to check columns on products.
  #
  # == Parameters:
  # source::
  #   DataSource object to filter
  # inactive_settings::
  #   Inactive columns for this module.
  #
  # == Returns:
  # Hash with filters
  #
  def ui_filters(source, inactive_settings)
    tmp = source.filter_to_json(current_user, false)
    tmp["#{source.id}_columns".to_sym] =
      tmp["#{source.id}_columns".to_sym].map do |e|
        e if valid_ui_column?(e, source, inactive_settings)
      end.compact
    tmp.to_json.html_safe
  end

  #
  # Check if a column to show in UI filters is valid. This means, the column
  # isn't disabled and the value exists.
  #
  # == Parameters:
  # column::
  #   Hash with the column to filter
  # source::
  #   DataSource object to filter
  # inactive_settings::
  #   Inactive columns for this module.
  #
  # == Returns:
  # Boolean
  #
  def valid_ui_column?(column, source, inactive_settings)
    column[:value] && ([:start_time, :end_time, :group_id, :domain, :organization]
                       .include?(column[:id]) ||
                       !inactive_settings.key?("#{source.id}_#{column[:id]}"))
  end

  def objects_types_options(value, column, divider = true)
    return unless value.present? && [:dst, :src, :dst_port, :src_port,
      :ethdst, :ethsrc, :vlan, :client_mac, :lan_ip, :wan_ip,
      :lan_l4_port, :wan_l4_port, :lan_vlan, :wan_vlan,
      :l4_proto, :application_id_name].include?(column.to_sym)

    return if column.to_sym.eql?(:application_id_name) &&
      (!current_user.super_admin? || value.match(/\A\d+:\d+\z/).nil?)

    list = ''
    list += content_tag(:li, nil, class: 'divider') if divider

    if column.to_sym.eql?(:application_id_name)
      object = ApplicationObject.find_by(value: value)
      url = object.blank? ? new_application_object_path(value: value) : edit_application_object_path(object)
    else
      type = object_type_from_column(column.to_sym)
      object = RedborderObject.find_by(value: value, type: type, user: current_user)
      url = object.blank? ? new_object_path(:class => type, value: value) : edit_object_path(object, object: { type: object.type })
    end

    list += content_tag(:li) do
      link_to (object.blank? ? "#{I18n.t('application_helper.create')}" : "#{I18n.t('application_helper.update')}"), url,
        'data-toggle' => "modal", 'data-target' => "#default_modal", 'value' => "#{value}"
    end
    list.html_safe
  end

  #
  # => Humanize the name of a product
  #
  def object_type_from_column(column)
    case column
    when :dst, :src, :lan_ip, :wan_ip
      'NetObject'
    when :ethdst, :ethsrc, :client_mac
      'MacObject'
    when :vlan, :src_vlan, :dst_vlan
      'VlanObject'
    when :src_port, :dst_port, :lan_l4_port, :wan_l4_port
      'ServiceObject'
    when :l4_proto
      'ProtocolObject'
    end
  end

  #
  # => Humanize the name of a product
  #
  def humanize_product(product)
    if product == 'ips'
      product.upcase
    else
      product.titleize
    end
  end

  #
  # Helper to render FontAwesome icon
  #
  def fa_icon(icon, args = {})
    content_tag(:i, nil, class: "fa fa-#{icon} #{args[:class]}")
  end

  #
  # Return redborder version as String. If version file doesn't exist, it
  # returns a generic version (3.1)
  #
  def redborder_version
    return '3.1' unless File.exist?("#{Rails.root}/version")
    File.read("#{Rails.root}/version")
  end

  def webui_version
    return '' unless File.exist?("#{Rails.root}/webui_version")
    File.read("#{Rails.root}/webui_version")
  end

  def hostname
    return '' unless File.exist?("#{Rails.root}/hostname")
    File.read("#{Rails.root}/hostname")
  end

  #
  # => Return an array with the object types
  #
  def object_types_list
    %w(net_object mac_object vlan_object service_object protocol_object)
  end

  def show_lookups(ip, column)
    return unless ip
    if [:dst, :src, :lan_ip, :wan_ip].include?(column.to_sym)
      lookups = Lookup.all
      list = ''

      list += content_tag(:li, nil, class: 'divider')

      lookups.each do |lookup|
        list += content_tag(:li) do
          link_to "#{lookup.title}", lookup.value.sub('${ip}', ip), target: '_blank', class: 'view_all'
        end
      end

      list += content_tag(:li) do
        link_to "#{I18n.t('application_helper.basic_source')}", "/#{@source.id}/lookup?address=#{ip}", 'data-toggle' => "modal", 'data-target' => '#lookup_modal'
      end

      list.html_safe
    end
  end

  # def show_scan(ip, column)
  #   list = ''
  #   return unless ip
  #   if [:dst, :src, :lan_ip, :wan_ip].include?(column.to_sym)
  #     list += content_tag(:li, nil, class: 'divider')
  #     list += content_tag(:li) do
  #       ip_aux = ip.gsub(".", "_")
  #       link_to "Vulnerabilities Scan", quick_scan_scanner_index_path(:target => ip_aux)

  #       #  <%=  link_to content_tag(:i, scan.scan_type_humanize), :controller => :scan_history, :action => :scan_result,
  #       #       :scan_profile_id => scan.scan_profile_id, :scan_history_id => scan.id, :scan_type_id => scan.scan_type_id %>
  #     end
  #     list.html_safe
  #   end
  # end

  #
  # => Return a list with new menu options for the tables.
  #
  def more_options(column, value)
    list = ""
    if column.to_sym.eql?(:sensor_name)
      list += content_tag(:li) do
        link_to content_tag(:i, nil, class: "fa fa-external-link") + "#{I18n.t('application_helper.explore')}", tree_sensors_path(search: (value or "N/A")), method: :get
      end
    end
    list.html_safe
  end

  #
  # => Traduce the operator
  #
  def operator(operator)

    val = case operator
    when "not", "is_not"
      "!="
    when "gt"
      ">"
    when "gte"
      ">="
    when "lt"
      "<"
    when "lte"
      "<="
    when "isnull"
      "is null"
    when "notnull"
      "is not null"
    when "contains"
      "LIKE"
    when "contains_not"
      "NOT LIKE"
    when "starts_with"
      "^="
    when "ends_with"
      "=$"
    else
      ""
    end

    val

  end

  #
  # Return the array of users the current user can view.
  # This is based on the permissions of the current user.
  #
  # == Returns
  # Array of arrays with user names and ids
  #
  def user_array
    ret = [["#{I18n.t 'application_helper.me'} (#{current_user.name})", current_user.email]]
    sensors = current_user.domains_with_permission.map(&:id)
    users = User.unscoped
            .where(sensor_id: sensors)
            .order('LOWER(name) ASC')

    users.each do |user|
      # Skip current user
      next if user.id == current_user.id
      ret.push [user.name, user.email]
    end
    ret
  end

  #
  # Generate the menu to select the granularity
  #
  # == Returns:
  # Safe string with the HTML code of the menu.
  #
  def granularity
    html = link_to '#', class: 'dropdown-toggle', 'data-toggle' => 'dropdown' do
      content_tag(:i, nil, class: 'fa fa-clock-o') +
      " #{Redborder::Rbdruid::GRANULARITY_TYPE[@granularity.to_sym][:long]} " +
      content_tag(:span, nil, class: 'caret')
    end

    html += content_tag :ul, class: 'dropdown-menu' do
      list = ''.html_safe

      Redborder::Rbdruid::slow_granularity_list({start_time: @start_time, end_time: @end_time}).each do |k, v|
        list += content_tag :li do
          link_to url_for(granularity: k, range: @range), class: "block-ui-but" do
            "#{Redborder::Rbdruid::GRANULARITY_TYPE[k][:long]} #{content_tag(:i, nil, class: "fa fa-exclamation-triangle", :"data-toggle" => "tooltip", title: "#{I18n.t('application_helper.slow')}")}".html_safe
          end
        end
      end

      Redborder::Rbdruid::granularity_list({start_time: @start_time, end_time: @end_time}).each do |k, v|
        list += content_tag :li do
          link_to Redborder::Rbdruid::GRANULARITY_TYPE[k][:long], url_for(granularity: k, range: @range), class: 'block-ui-but'
        end
      end

      list
    end

    html
  end

  def raw_field(value, column)
    case column.to_sym
    when :priority, :darklist_score_name
      color = Severity.find_by(name: value).try('bg_color') || '#4572A7'
      content_tag(:span, value, style: "background-color: #{color}; color: white; padding: 3px 7px; border-radius: 4px; font-weight: bold;")
    when :action
      content_tag(:span, value, class: "action #{value}")
    when :src_country_code, :dst_country_code, :wan_ip_country_code
      value == 'N/A' ? 'N/A' : content_tag(:span, nil, class: "flag #{value.downcase}") + value
    when :wireless_station, :client_mac
      if Traffic::COLUMNS[column.to_sym][:complementary_columns].try('keys')
        value.send(Traffic::COLUMNS[column.to_sym][:complementary_columns].keys.first) || 'N/A'
      else
        value.send(column.to_sym) || 'N/A'
      end
    else
      value
    end
  end

  #
  # => Return secondary names for some specific columns in the graph.
  #
  def secondary_columns_graph(columns)
    columns.map do |c|
      [:wireless_station, :client_mac].include?(c) ?
        @source::COLUMNS[c][:complementary_columns].keys.first : c
    end
  end

  def flag_country(e, column)
    value = case column
            when :src_country
              e.send(:src_country_code)
            when :dst_country
              e.send(:dst_country_code)
            when :ip_country
              e.send(:ip_country_code)
            end

    (value == 'N/A' ? 'N/A' : content_tag(:span, nil, class: "flag #{value.downcase}", :'data-toggle' => 'tooltip', title: value)) if value
  end

  #
  # => Return the color for IPS actions
  #
  def get_color_for_action(action)
    case action.to_sym
    when :drop, :sdrop
      '#FF0000'
    when :pass
      '#00FF00'
    when :alert
      '#ffab2e'
    end
  end

  #
  # Check if we are on first login after installation. If this is the
  # first login, we show a message with the default user and password
  #
  # == Returns:
  # A string with the code to show the message
  def sign_in_count_zero
    return unless User.sum(:sign_in_count).zero?
    content_tag(:div,
      "<a class='close' data-dismiss='alert'>&#215;</a><span> #{I18n.t('application_helper.info')} \
      #{I18n.t('application_helper.login_password')} - <b>#{RedBorder::USER_CONFIG['login']}</b> / \
      <b>#{RedBorder::USER_CONFIG['password']}</b></span>"
      .html_safe, class: 'alert alert-info')
  end

  def number_with_delimiter(number, delimiter = ',', separator = '.')
    parts = number.to_s.split('.')
    parts[0].gsub!(/(\d)(?=(\d\d\d)+(?!\d))/, "\\1#{delimiter}")
    parts.join separator
  rescue
    number
  end

  def show_category(categories, category, inactive_settings)
    name = @source.id
    categories[category].each do |tab, value|
      # if Setting.send("#{controller.to_sym}_#{tab}".to_sym) == "1"
      unless inactive_settings.key?("#{name}_#{tab}")
        return true unless value[:admin] && !User.current_user.super_admin?
        return true unless value[:superuser] && !User.current_user.super_user?
      end
    end
    false
  end

  def now_to_time_machine
    in_time_machine do
      Time.zone.now
    end
  end

  def ohai_last_seen(time)
    return_text = ""
    time = Time.at(time)
    if time < Time.now - 30.minutes
      return_text += content_tag(:i, nil, class: "icon-circle icon-red")
    end
    return_text += " " + distance_of_time_in_words(Time.now, time)

    return_text.html_safe
  end

  #
  # => Show the increase or decrease of the second argument relative to the
  #    first. It shows the value in green if it has been increased,
  #    or red if it has been drecrease. Also, show an up or down arrow.
  #
  def difference_percent(first_value, second_value)
    if first_value.zero? && second_value.zero?
      percent = 0
    elsif second_value.zero?
      percent = -100
    elsif first_value.zero?
      # infinite increase
      percent = Float::INFINITY
    else
      percent = ((second_value - first_value).to_f / first_value * 100).round(2)
    end

    content = content_tag :span do
      if percent == Float::INFINITY
        '&infin;%'.html_safe
      else
        percent.abs.to_s + '%'
      end
    end

    content += if percent < 0
      content_tag(:i, nil, class: 'fa fa-arrow-down')
    elsif percent == 0
      content_tag(:i, nil, class: 'fa fa-minus')
    else
      content_tag(:i, nil, class: 'fa fa-arrow-up')
    end

    content.html_safe
  end

  # ------------------------------------------------------------
  #  Tabs methods
  # ------------------------------------------------------------

  #
  # Create HTML structure for tabs navigation
  #
  # == Parameters:
  # tabs::
  #   Array of strings with tabs
  # active::
  #   (Optional) Active element. If it's nil, the first tab will be the active
  #
  # == Returns:
  # HTML safe string
  #
  def create_tabs(tabs, active = nil)
    html = ''
    tabs.each_with_index do |tab, i|
      css_class = (tab == active || (active.nil? && i == 0)) ? 'active' : ''
      html +=
        content_tag(:li, class: css_class, id: tab) do
          link_to(tab.capitalize, "##{tab}", data: { toggle: 'tab' })
        end
    end
    # Return HTML safe string
    html.html_safe
  end

  def categories_from_columns(columns, user = User.current_user)
    categories = {}
    columns.each do |k, v|
      next if (v[:superuser] && !user.super_user?) || (v[:admin] && !user.super_admin?)
      (categories[v[:category]] ||= {})[k] = v
    end
    categories
  end

  def asset_path_from_controller(asset)
    asset_path(asset)
  end

  # Check if the record exists or not
  # Example usage: llm_generated_message = llm_generated?(model_name: 'Incident', field_name: 'name', record_id: 1)
  def llm_generated?(model_name:, field_name:, record_id:)
    record = LlmGenerationRecords.find_by(
      table_name: model_name.tableize,
      field_name: field_name,
      record_id: record_id,
      record_type: model_name
    )
    
    if record
      "#{I18n.t 'application_helper.llm_generated_at'} #{record.updated_at.strftime('%Y-%m-%d %H:%M:%S')}"
    else
      nil
    end
  end

  def theme_loading_image
    # Check if the user has dark mode enabled
    dark_mode = current_user.dark_mode
    
    # Return the appropriate image tag
    image_tag(
      dark_mode ? 'horizontal_loading_dark.gif' : 'horizontal_loading.gif',
      class: 'tab_loading',
      alt: 'Loading...'
    )
  end  
  
  def theme_class
    if user_signed_in?
      current_user.dark_mode ? 'dark-mode' : 'light'
    else
      cookies[:theme] || 'light'
    end
  end
  
  def theme_error_image
    # Check if the user has dark mode enabled
    dark_mode = current_user.dark_mode
    
    # Return the appropriate image tag
    image_tag(
      dark_mode ? 'logowhite.png' : 'logo.png',
    )
  end
  

  private

  def in_time_machine
    unless User.current_user.nil? or User.current_user.time_machine_now.nil?
      Timecop.travel(Time.at(User.current_user.time_machine_now)) do
        yield
      end
    else
      yield
    end
  end

  def set_ticket_desc
    result = "#{I18n.t('application_helper.fill_in')}\n\n\n\n\n" +
      "---------------------------------------------------------\n"

    result
  end

  def sortable(column, page, title = nil, controller = nil, action = nil)
    title ||= column.titleize
    css_class = (column == sort_column) ? "current #{sort_direction}" : 'not-current'
    direction = (column == sort_column && sort_direction == 'asc') ? 'desc' : 'asc'
    opts = { sort: column, direction: direction, page: page }
    opts[:controller] = controller unless controller.nil?
    opts[:action] = action unless action.nil?
    link_to title, opts, class: css_class, remote: true
  end

  def reportSortable(prefix, column, page, title = nil, controller = nil, action = nil)
    title ||= column.titleize
    sort_param = "#{prefix}_sort"
    direction_param = "#{prefix}_direction"
    current_sort_column = params[sort_param]
    current_sort_direction = params[direction_param]

    css_class = (column == report_sort_column(current_sort_column)) ? "current #{report_sort_direction(current_sort_direction)}" : 'not-current'
    direction = (column == report_sort_column(current_sort_column) && report_sort_direction(current_sort_direction) == 'asc') ? 'desc' : 'asc'

    opts = {
      "#{prefix}_sort" => column,
      "#{prefix}_direction" => direction,
      "#{prefix}_page" => page
    }

    other_prefixes = %w[user_reports others_reports sensor_reports] - [prefix]

    other_prefixes.each do |other_prefix|
      opts["#{other_prefix}_sort"] = params["#{other_prefix}_sort"]
      opts["#{other_prefix}_direction"] = params["#{other_prefix}_direction"]
    end

    opts[:controller] = controller unless controller.nil?
    opts[:action] = action unless action.nil?

    link_to title, opts, class: css_class, remote: true
  end

  def reportHistoricalsSortable(prefix, column, page, title = nil, controller = nil, action = nil)
    title ||= column.titleize
    sort_param = "#{prefix}_sort"
    direction_param = "#{prefix}_direction"
    current_sort_column = params[sort_param]
    current_sort_direction = params[direction_param]

    css_class = (column == report_historical_sort_column(current_sort_column)) ? "current #{report_sort_direction(current_sort_direction)}" : 'not-current'
    direction = (column == report_historical_sort_column(current_sort_column) && report_sort_direction(current_sort_direction) == 'asc') ? 'desc' : 'asc'

    opts = {
      "#{prefix}_sort" => column,
      "#{prefix}_direction" => direction,
      "#{prefix}_page" => page
    }

    other_prefixes = %w[user_reports others_reports sensor_reports] - [prefix]

    other_prefixes.each do |other_prefix|
      opts["#{other_prefix}_sort"] = params["#{other_prefix}_sort"]
      opts["#{other_prefix}_direction"] = params["#{other_prefix}_direction"]
    end

    opts[:controller] = controller unless controller.nil?
    opts[:action] = action unless action.nil?

    link_to title, opts, class: css_class, remote: true
  end

  def incidentReportHistoricalsSortable(prefix, column, page, title = nil, controller = nil, action = nil)
    title ||= column.titleize
    sort_param = "#{prefix}_sort"
    direction_param = "#{prefix}_direction"
    current_sort_column = params[sort_param]
    current_sort_direction = params[direction_param]

    css_class = (column == incident_historical_sort_column(current_sort_column)) ? "current #{report_sort_direction(current_sort_direction)}" : 'not-current'
    direction = (column == incident_historical_sort_column(current_sort_column) && report_sort_direction(current_sort_direction) == 'asc') ? 'desc' : 'asc'

    opts = {
      "#{prefix}_sort" => column,
      "#{prefix}_direction" => direction,
      "#{prefix}_page" => page
    }

    other_prefixes = %w[user_reports others_reports sensor_reports] - [prefix]

    other_prefixes.each do |other_prefix|
      opts["#{other_prefix}_sort"] = params["#{other_prefix}_sort"]
      opts["#{other_prefix}_direction"] = params["#{other_prefix}_direction"]
    end

    opts[:controller] = controller unless controller.nil?
    opts[:action] = action unless action.nil?

    link_to title, opts, class: css_class, remote: true
  end

  # ============================================================
  # => Namespace methods
  # ============================================================

  #
  # Create the namespace menu if the user as multiple data sources.
  #
  def namespace_menu
    # Hide on:
    hide_on = %w(sensors alarms audits settings managers licenses
                 redborder_objects overlay_maps users jobs rule_versions
                 registrations)
    # Check number of namespaces
    return true if !hide_on.include?(params[:controller]) &&
                   current_user.data_sources.size > 1
    false
  end

  #
  # Get active namespaces of an user
  #
  def active_namespaces
    return 'All' if current_user.data_sources.map(&:uuid) == active_namespaces_ids
    # Load namespaces
    namespaces = []
    active_namespaces_ids.each do |ds|
      s = Sensor.find_by(uuid: ds, type: Sensor::DOMAIN)
      namespaces << namespace_name(s) if s && current_user.can?(:read, s)
    end
    namespaces.join(', ').truncate(18)
  end

  #
  # Get ids of active namespaces for an user
  #
  def active_namespaces_ids
    # Load namespaces
    current_user.active_data_sources
  end

  #
  # If namespace is root, the name is default
  #
  # == Parameters:
  # sensor::
  #   A sensor object
  #
  # == Returns:
  # String with the name of the namespace
  #
  def namespace_name(sensor)
    return 'Default' if sensor.is_root?
    sensor.name
  end

  # ============================================================
  # => Gravatar methods
  # ============================================================

  #
  # Return Gravatar URL
  #
  # == Parameters:
  # email::
  #   String with user email
  # size::
  #   Integer with size. By default, this value is 100
  #
  # == Returns:
  # String with the URL
  #
  def get_gravatar_uri(email, size = 100)
    # Default image when gravatar isn't available
    default = 'https://static.redborder.com/redborder/logo.png'
    # create the md5 hash
    hash = Digest::MD5.hexdigest(email.downcase)
    # compile URL which can be used in <img src="RIGHT_HERE"...
    "https://www.gravatar.com/avatar/#{hash}?s=#{size}&d=#{default}"
  end

  #
  # Set HTML to display a gravatar image as img
  #
  # == Parameters:
  # email::
  #   String with user email
  # size::
  #   Integer with size. By default, this value is 100
  # class_attr::
  #   String CSS classes for the img
  #
  # == Returns:
  # Safe string with HTML code
  #
  def get_gravatar_image(email, size = 100, class_attr = '')
    content_tag(:image,
                nil,
                src: get_gravatar_uri(email, size),
                class: class_attr,
                width: "#{size}px",
                height: "#{size}px"
               )
  end

  # ============================================================
  # => Widgets methods
  # ============================================================

  #
  # Change the name of of the action to a symbol understand by a widget
  #
  # == Returns:
  # An string identify the widget_type
  #
  def widget_type_from_action
    case params[:action]
    when 'tops'
      'top'
    when 'cardinality'
      'unique'
    when 'comparative'
      'compare'
    else
      'raw'
    end
  end

  # ============================================================
  # => Partial methods
  # ============================================================

  #
  # => Try to render a partial or rescue and return an empty string
  #
  def try_partial(path, variables = {})
    begin
      render partial: path, locals: variables
    rescue
      ''
    end
  end

  # ============================================================
  # => Modal methods
  # ============================================================

  #
  # Create the HTML of default modal. It will be a normal size modal
  #
  # == Returns
  # HTML code of a medium size modal with the id default_modal
  #
  def default_modal
    create_modals(['default_modal'])
  end

  #
  # Create the HTML of defaults modals. It will be a normal size modal
  #
  # == Returns
  # HTML code of a small, medium, large, xlarge size modals with the ids
  # default_small, default_modal, default_large, default_xlarge
  #
  def default_modals
    create_modals([['default_small', 'small'],
                   'default_modal',
                   ['default_large', 'large'],
                   ['default_xlarge', 'xlarge']])
  end

  #
  # => Create multiples modals for bootstrap based on ids passed in an array.
  #    The array must have the format:
  #
  #    [['id', 'size'], ['id', 'size'], 'id', 'id']
  #
  #    Available sizes are: small, large
  #
  def create_modals(modal_ids)
    # Base code
    html = ''

    modal_ids.each do |modal|
      if modal.kind_of?(Array)
        size = case modal[1]
               when 'small'
                 'modal-sm'
               when 'medium'
                 'modal-md'
               when 'large'
                 'modal-lg'
               when 'xlarge'
                 'modal-xlg'
               end

        html += content_tag(:div,
                  content_tag(:div,
                    content_tag(:div, '', class: 'modal-content'),
                  class: "modal-dialog #{size}"),
                class: 'modal fade', id: modal[0], :"aria-labelledby" => modal[0],
                role: 'dialog', tabindex: '-1', :"aria-hidden" => 'true')

      else
        html += content_tag(:div,
                  content_tag(:div,
                    content_tag(:div, '', class: 'modal-content'),
                  class: 'modal-dialog'),
                class: 'modal fade', id: modal, :"aria-labelledby" => modal,
                role: 'dialog', tabindex: '-1', :"aria-hidden" => 'true')
      end
    end

    html.html_safe
  end

  #
  # => Create a modal to create or edit an object.
  #    This function take the param and call create_modals
  #    with an array of new_param_modal and edit_param_modal
  #
  #    @param [array|string] ids -> Ids for the modals
  #
  def create_edit_modals(ids)
    ids_real = []

    if ids.kind_of?(Array)
      ids.each do |modal_id|
        ids_real << "new_#{modal_id}_modal"
        ids_real << "edit_#{modal_id}_modal"
      end
    else
      ids_real << "new_#{ids}_modal"
      ids_real << "edit_#{ids}_modal"
    end

    create_modals(ids_real)
  end

  #
  # => Render a progress bar in a modal
  #
  def render_progress_bar(from, to)
    # Render the html
    html = content_tag :div, class: 'modal-progress', id: 'modal-progress' do
             content_tag :div, '', style: "width: #{from}%"
           end
    html += javascript_tag "$('#modal-progress > div').animate({
                           width: '+=#{(to - from).to_s}%'}, 500);"

    html
  end

  # ============================================================
  # => Pagination methods
  # ============================================================

  # QUESTION: We need this method?

  #
  # => Change the HTML renderer of will_paginate to
  #    adapt the UI to the boostrap navigation
  #
  def will_paginate(collection_or_options = nil, options = {})
    if collection_or_options.is_a? Hash
      options, collection_or_options = collection_or_options, nil
    end
    unless options[:renderer]
      options = options.merge renderer: Redborder::LinkRenderer
    end
    super *[collection_or_options, options].compact
  end

  # ============================================================
  # => Aggregation methods
  # ============================================================

  #
  # => Scale a number based K,M,G,T and add the unit.
  #
  def cast_num(count, aggregation='')
    case count
    when 1e12..Float::INFINITY
      "#{(count.to_f / 1e12).round(2)} T#{aggregation}"
    when 1e9..1e12
      "#{(count.to_f / 1e9).round(2)} G#{aggregation}"
    when 1e6..1e9
      "#{(count.to_f / 1e6).round(2)} M#{aggregation}"
    when 1e3..1e6
      "#{(count.to_f / 1e3).round(2)} K#{aggregation}"
    else
      "#{count} #{aggregation}"
    end
  end

  def aggregation_name(model, aggregation)
    aggregation_data = aggregation_data model, aggregation
    return '' if aggregation_data.nil?
    if aggregation_data[:name].start_with? "tr." then I18n.t aggregation_data[:name][3..-1] else aggregation_data[:name] end
  end

  def aggregation_unit(model, aggregation)
    aggregation_data = aggregation_data model, aggregation
    return '' if aggregation_data.nil?
    aggregation_data[:unit]
  end

  def aggregation_top_name(model, aggregation)
    aggregation_data = aggregation_data model, aggregation
    return if aggregation_data.nil?
    aggregation_top = aggregation_data.fetch(:top, {})
    aggregation_top[:name] || aggregation_name(model, aggregation)
  end

  def aggregation_top_unit(model, aggregation)
    aggregation_data = aggregation_data model, aggregation
    return if aggregation_data.nil?
    aggregation_top = aggregation_data.fetch(:top, {})
    aggregation_top[:unit] || aggregation_unit(model, aggregation)
  end

  def aggregation_data(model, aggregation)
    return if aggregation.nil?
    model::AGGREGATIONS[aggregation.to_sym]
  end

  def data_header(model, value, column = nil)
    if model::COLUMNS.keys.include? value.to_sym
      if column.nil?
        model::COLUMNS[value.to_sym][:columns].values.first
      else
        model::COLUMNS[value.to_sym][:complementary_columns][column.to_sym]
      end
    else
      value.capitalize
    end
  end

  def model_cast_num(model, count, aggregation)
    return 'N/A' unless count
    aggregation_unit = aggregation_top_unit(model, aggregation) || ''
    cast_num count, aggregation_unit
  end

  def aggregations_collection(model)
    model::AGGREGATIONS.each_with_object([]) do |(_k, v), memo|
      memo << [if v[:name].start_with? "tr." then "#{I18n.t v[:name][3..-1]}" else v[:name] end, v[:value].to_s]
    end
  end

  def all_aggregations_collection
    all_aggregations = Redborder::Module.get_from_all :aggregations

    collection = all_aggregations.each_with_object([]) do |(mod, aggregations), memo|
      next if aggregations.nil?

      memo << aggregations.each_with_object([]) do |(name, data), memo|
        memo << [data[:name], data[:value].to_s]
      end
    end

    collection.flatten(1)
  end

  def concat_json(object, name, data)
    result = object.as_json
    result[name.to_s] = data.as_json
    result.to_json
  end

  def module_controller_name(controller)
    eval('module_controller_name_' + controller)
  end

  # ============================================================
  # => Chosen
  # ============================================================

  #
  # Helper to implement a JS Chosen
  # Parameters:
  # key: key submited in form
  # collection: Array of Elements to select
  # opts: Hash with extra options:
  #   :label(string), :required(boolean),
  #   :selected (default element), :multiple(boolean)
  #
  def select_chosen(key, collection, opts = {})
    content_tag(:div, class: 'form-group select', id: 'select-chosen') do
      if opts[:label]
        concat(content_tag(:label) do
          concat(content_tag(:abbr, '*', title: 'required')) if opts[:required]
          concat(opts[:required] ? " #{opts[:label]}" : opts[:label])
        end)
      end
      concat(select_tag(key, options_for_select(collection, opts[:selected]),
                        multiple: opts[:multiple] || false, class: 'chosen-select form-control'))
      concat(javascript_tag('initChosen();'))
    end
  end

  # ============================================================
  # => Panels
  # ============================================================

  def panel(title)
    content_tag(:div, class: 'panel panel-default') do
      concat(content_tag(:div, class: 'panel-heading') do
        concat(content_tag(:div, class: 'panel-title') do
          title
        end)
      end)
      concat(content_tag(:div, class: 'panel-body') do
        yield
      end)
    end
  end

  # ============================================================
  # => ALERT BOXES
  # ============================================================

  def alert_box(type)
    content_tag(:div, class: "alert alert-#{type}") do
      concat fa_icon(type)
      yield
    end
  end

  # ============================================================
  # => Head layout
  # ============================================================

  def gmaps_key
    type = Setting.find_by(name: :gmaps_type).try(:value)
    value = Setting.find_by(name: :gmaps_key).try(:value)
    type.eql?('api_key') ? "key=#{value}".html_safe : "client=#{value}".html_safe
  end

  # ============================================================
  # => Unit conversion
  # ============================================================

  def unit_aware_scale(value, unit)
    case unit.to_s.strip.downcase
    when '%', '°', 'ms'
      [value, unit]
    else
      [scale_si(value), unit]
    end
  end

  def scale_si(count)
    scales = [
      [1e12, 'T'],
      [1e9, 'G'],
      [1e6, 'M'],
      [1e3, 'K']
    ]

    count = count.to_f

    scales.each do |threshold, prefix|
      next unless count >= threshold

      scaled = (count / threshold).round(3)
      return [scaled.to_s, prefix]
    end

    scaled_count = count.round(3)
    [scaled_count.to_s, '']
  end
end
