#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Methods to help the creation of breadcrumbs
#
module BreadcrumbHelper
  #
  # Create a breadcrumb for sensors. Based on top domain
  #
  # == Parameters:
  # sensor::
  #   Sensor object the user wants to view. We stop to render nodes at this
  #   sensor.
  # url::
  #   (Optional) String of the url helper to set links on elements
  # - top: User top domain.
  # manage::
  #   (Optional) The user need manage permissions to show the url.
  # options::
  #   Hash with options. Available options are:
  #     - home: Hash that add a first element to the list. The format is:
  #            { name: '', url: '' }
  #     - root: String to show as caption if the sensor is root
  #     - html_options: Hash with HTML options for links
  #     - route_id: Set the key for the ID on GET route
  #     - include_root: By default, root isn't show
  #
  # == Returns:
  # Safe string with HTML code
  #
  def sensor_breadcrumb(sensor, url = nil, manage = false, options = {})
    return '' if sensor.nil?
    # Iterate over nodes
    list = render_breadcrumb_nodes(sensor, url, manage, options)
    # Add first element and return
    content_tag(:ul, list.join.html_safe, class: 'breadcrumb')
  end

  #
  # Iterate over sensors and return a list of nodes.
  #
  # == Parameters:
  # sensor::
  #   Sensor object the user wants to view. We stop to render nodes at this
  #   sensor.
  # url::
  #   (Optional) String of the url helper to set links on elements
  # - top: User top domain.
  # manage::
  #   (Optional) The user need manage permissions to show the url.
  # options::
  #   Hash with options. Available options are:
  #     - home: Hash that add a first element to the list. The format is:
  #            { name: '', url: '' }
  #     - root: String to show as caption if the sensor is root
  #     - html_options: Hash with HTML options for links
  #     - route_id: Set the key for the ID on GET route
  #
  # == Returns:
  # Safe string with HTML code
  #
  def render_breadcrumb_nodes(sensor, url = nil, manage = false, options = {})
    top = User.current_user.sensor
    list = []
    # Iterate from top domain
    sensor.self_and_ancestors_with_root.reverse.each do |s|
      # Ignore root by default
      next if options[:include_root].nil? && s.is_root?
      # Stop if the current sensor is the sensor to view
      break if s == top
      # Render node
      list.unshift(
        render_breadcrumb_node(s, url, s == sensor, manage, options))
    end
    list.unshift(render_first_breadcrumb_node(options[:home], options))
  end

  #
  # Render the first element of the list
  #
  # == Parameters:
  # home::
  #   Hash that add a first element to the list. The format is:
  #   { name: '', url: '' }
  #
  # == Returns:
  # Safe String with the HTML code for a li element or empty
  #
  def render_first_breadcrumb_node(home, options)
    return '' if home.nil?
    el = link_to(home[:name], home[:url],
                 breadcrumb_link_options(options[:html_options]))
    content_tag(:li, el)
  end

  #
  # Render a node of breadcrumb (li element)
  #
  # == Parameters:
  # sensor::
  #   Sensor object to render
  # url::
  #   Url to render
  # active::
  #   Boolean to set this element as active
  # manage::
  #   The user need manage permissions to show the url.
  # options::
  #   Hash with options. Used options:
  #     - root: String to show as caption if the sensor is root
  #     - route_id: String if you need to specify a different
  #                 id to send over GET request
  #     - html_options: HTML options for links
  #
  # == Returns:
  # Safe String with the HTML code for a li element
  #
  def render_breadcrumb_node(sensor, url, active, manage, options = {})
    # Get the link
    el = breadcrumb_link(sensor, url, active, manage, options)
    # Return the html element
    content_tag(:li, el, class: (active ? 'active' : ''))
  end

  #
  # Create a the link for a node
  #
  # == Parameters:
  # sensor::
  #   Sensor object to render
  # url::
  #   Url to render
  # active::
  #   Boolean to set this link as active
  # manage::
  #    The user need manage permissions to show the url.
  # options::
  #   Hash with options for links
  #
  # == Returns:
  # String with the caption
  #
  def breadcrumb_link(sensor, url, active, manage, options)
    # Init the caption
    caption = breadcrumb_caption(sensor, options[:root])
    # Params for the link
    p = options[:route_id].nil? ? sensor : { options[:route_id] => sensor.id }
    # Create the link or string only
    if url.nil? || active || (manage && !User.current_user.admin?)
      caption
    else
      # We must use send because by default, rails take globals values
      # to build helpers path
      link_to caption,
              send(url, p),
              breadcrumb_link_options(options[:html_options])
    end
  end

  #
  # Set the caption for the element
  #
  # == Parameters:
  # sensor::
  #   Sensor object to render
  # root::
  #   String to show as caption if the sensor is root
  #
  # == Returns:
  # String with the caption
  #
  def breadcrumb_caption(sensor, root)
    (root && sensor.is_root?) ? root : sensor.secondary_name
  end

  #
  # Return the options for links
  #
  # == Parameters:
  # html_options::
  #   Hash with HTML options
  #
  # == Returns:
  # Hash with options
  #
  def breadcrumb_link_options(html_options)
    opt = html_options || {}
    opt.merge(class: 'block-ui-but')
  end
end
