#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module IncidentsHelper
  include ActionView::Helpers::AssetUrlHelper

  def get_asset_image(type, view_context = nil)
    default_path = 'icons/assets/default_asset.svg'
    file_path = (type && type.is_a?(String) && type.present?) ? "icons/assets/#{type.downcase}.svg" : default_path

    view_context ||= self
    asset_exists?(file_path) ? view_context.asset_path(file_path) : view_context.asset_path(default_path)
  end

  def get_observable_image(type, view_context = nil)
    default_path = 'icons/observables/default_observable.svg'
    file_path = (type && type.is_a?(String) && type.present?) ? "icons/observables/#{type.downcase}.svg" : default_path

    view_context ||= self
    asset_exists?(file_path) ? view_context.asset_path(file_path) : view_context.asset_path(default_path)
  end

  def asset_exists?(file_path)
    (Rails.application.assets && Rails.application.assets.find_asset(file_path)) || 
      File.exist?(Rails.root.join("app/assets/images", file_path))
  end  

  def incidents_uniq_users
    User.joins(:incident_users).distinct
  end

  def incidents_uniq_sources
    Incident.pluck(:source).uniq
  end

  def incidents_order_by
    [['ID', 'id'], [I18n.t('incidents_helper.name'), 'name'], [I18n.t('incidents_helper.source'), 'source'], [I18n.t('incidents_helper.created_at'), 'created_at'], [I18n.t('incidents_helper.assigned'), 'assigned_user'], [I18n.t('incidents_helper.domain'), 'sensor_id']]
  end

  def incidents_order_direction
    [[I18n.t('incidents_helper.ascending'), 'asc'], [I18n.t('incidents_helper.descending'), 'desc']]
  end

  def incident_detection_fields(incident)
    elements = incident.get_data({})

    detection_unique_sources = elements.map { |log| log[:incident_source_name] }.uniq
    detection_unique_priorities = elements.map { |log| log[:priority] }.uniq

    [detection_unique_sources, detection_unique_priorities]
  end

  def incidents_domains
    incident_sensor_ids = Incident.distinct.pluck(:sensor_id)

    current_user.domains_with_permission
                              .where(id: incident_sensor_ids) # Filter by incident sensor_ids
                              .map { |d| [d.id, d.path_name] }
                              .to_h
  end

  def worklog_uniq_types(user = current_user)
    [['Notes','Notes'], ['Incident Changes', 'Incident Changes'], ['Response Actions', 'Response Actions']]
  end

  def worklog_uniq_users(worklog_data, user = current_user)
    filtered_worklog_data = worklog_data.reject { |entry| entry[:user].nil? || entry[:user].is_a?(String)}
    filtered_worklog_data.pluck(:user).uniq || []
  end

  def worklog_uniq_orders(user = current_user)
    [['Newest','Newest'], ['Oldest', 'Oldest']]
  end

  def priority_color(priority_name)
    case priority_name.downcase
    when "low"
      "#ffffff"
    when "medium"
      "#ffffff"
    when "high"
      "#ffffff"
    when "critical"
      "#ffffff"
    when "none"
      "#ffffff"
    when "unknow"
      "#ffffff"
    when "info"
      "#ffffff"
    when "warning"
      "#ffffff"
    when "notice"
      "#ffffff"
    when "error"
      "#ffffff"
    when 'emergency'
      "#ffffff"
    when 'alert'
      "#ffffff"
    when 'debug'
      "#ffffff"
    else
      "#ffffff"
    end
  end

  def priority_bgcolor(priority_name)
    case priority_name.downcase
    when "low"
      "#89a54e"
    when "medium"
      "#c69958"
    when "high"
      "#aa4643"
    when "critical"
      "#ff4643"
    when "none"
      "#30bf93"
    when "unknow"
      "#919191"
    when "info"
      "#5eca5e"
    when "warning"
      "#ffa500"
    when "notice"
      "#3399ff"
    when "error"
      "#ff0000"
    when 'emergency'
      "#b71c1c"
    when 'alert'
      "#f44336"
    when 'debug'
      "#b2ebf2"
    else
      "#000"
    end
  end

  def show_incident_lookups(ip)
    return unless ip
    
    lookups = Lookup.all
    list = ''

    list += content_tag(:li, nil, class: 'divider')

    lookups.each do |lookup|
      list += content_tag(:li) do
        link_to "#{lookup.title}", lookup.value.sub('${ip}', ip), target: '_blank', class: 'view_all'
      end
    end

    list.html_safe
  end

  def get_module_name_by_source(source)
    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:incident_source_name] && v[:incident_source_name] == source }

    incident_modules.first
  end

  def get_incident_filters_outliers(incident, sensor_name, sensor_uuid, lan_ip_pattern, wan_ip_pattern)
    start_time, end_time = Redborder::IncidentsOutliersUtils::RangeUtils.calculate_incident_outlier_range(incident, incident.outliers.first.range || Setting.outliers_incidents_range)

    filter_hash = {
      "0" => {
        "column" => "sensor_name",
        "operator" => "is",
        "value" => sensor_name,
        "enabled" => true
      },
      "1" => {
        "column" => "lan_ip",
        "operator" => "regex",
        "value" => lan_ip_pattern,
        "enabled" => true
      },
      "2" => {
        "column" => "wan_ip",
        "operator" => "regex",
        "value" => wan_ip_pattern,
        "enabled" => true
      },
      "3" => {
        "column" => "start_time",
        "operator" => "is",
        "value" => start_time,
        "enabled" => true
      },
      "4" => {
        "column" => "end_time",
        "operator" => "is",
        "value" => end_time,
        "enabled" => true
      },
      "5" => {
        "column" => "sensor_uuid",
        "operator" => "is",
        "value" => sensor_uuid,
        "enabled" => true
      }
    }
  
    filter_json = filter_hash.to_json
    
    encoded_filter = URI.encode_www_form_component(filter_json)
  
    filter_url = "/flow/tops?search=#{encoded_filter}&outliers=#{sensor_name}&tab=sensor_name"
  
    filter_url
  end

  def get_incident_filters_link(incident)
    return nil unless incident.source
  
    module_name = get_module_name_by_source(incident.source).try(:first)
    return nil unless module_name
  
    filters = {
      0 => {
        "column" => "incident_uuid",
        "operator" => "is",
        "value" => incident.uuid,
        "enabled" => true
      },
      1 => {
        "column" => "start_time",
        "operator" => "is",
        "value" => incident.created_at - 1.hour,
        "enabled" => true
      },
      2 => {
        "column" => "end_time",
        "operator" => "is",
        "value" => incident.created_at + 1.day,
        "enabled" => true
      }
    }
  
    if module_name == :flow
      start_time, end_time = Redborder::IncidentsOutliersUtils::RangeUtils.calculate_incident_outlier_range(incident, incident.outliers.first.range || Setting.outliers_incidents_range)
      filters[1]["value"] = start_time
      filters[2]["value"] = end_time
      filters[3] = {
        "column" => "sensor_name",
        "operator" => "is",
        "value" => incident.outliers.first["sensor_name"] || "",
        "enabled" => true
      }
      filters[4] = {
        "column" => "wan_ip",
        "operator" => "regex",
        "value" => incident.outliers.first["wan_ip_regex"] || "",
        "enabled" => true
      }
      filters[5] = {
        "column" => "lan_ip",
        "operator" => "regex",
        "value" => incident.outliers.first["lan_ip_regex"] || "",
        "enabled" => true
      }
    end
  
    json_string = filters.to_json
    encoded_json = CGI.escape(json_string)
    
    link = "/#{module_name}/tops?search=#{encoded_json}"
    link += "&outliers=#{incident.outliers.first["sensor_name"]}&tab=sensor_name" if module_name == :flow && incident.outliers
    link
  end

  def get_detection_element_filters_link(element)
    hash = {}
    element[:url_filters] = {} if element[:url_filters].nil?
    element[:url_filters].each_with_index do |e, index|
      field = e.first
      value = e.last

      hash[index] = {
        "column" => field,
        "operator" => "is",
        "value" => value,
        "enabled" => true
      }
    end

    json_string = hash.to_json
    encoded_json = CGI.escape(json_string)

    "/#{element[:module_name]}/tops?search=#{encoded_json}"
  end

  # Dynamic mapping method
  def map_dynamic_fields(data, field_map)
    field_map.each do |key, possible_fields|
      possible_fields.each do |possible_field|
        if data.methods.include?(possible_field)
          value = data.send(possible_field)
          return value unless value.nil?
        end
      end
    end
    'N/A'
  end
  
  def is_llm_running?(section, incident_id)
    incident_report = IncidentReports.find_by(incident_id: incident_id)
    return false if incident_report.nil?

    Delayed::Job.where('handler LIKE ? AND handler LIKE ? AND handler LIKE ?',
                       '%Redborder::Jobs::LLMGenerateContentJob%',
                       "%#{section}%",
                       "%#{incident_report.id}%")
                .exists?
  end
end