# frozen_string_literal: true

#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# This is the mitre helper for the views and controllers related to mitre
# Its main task its to allow developers to reuse common functions in the front
# or in the back
#
module MitreHelper
  # Constants representing MITRE technique and subtechnique types
  MITRE_TECHNIQUE     = 'technique'
  MITRE_SUBTECHNIQUE  = 'subtechnique'
  # Colors
  MITRE_UNKNOWN_COLOR = 'unknown'
  MITRE_LOW_COLOR     = 'low'
  MITRE_MEDIUM_COLOR  = 'medium'
  MITRE_HIGH_COLOR    = 'high'
  # Filters
  MITRE_CUSTOM_FILTER = 'custom'

  # Determine if the provided technique identifier is a technique or a subtechnique
  #
  # @param [String] technique The technique identifier
  # @return [String] Returns "technique" if the identifier does not include a dot, otherwise "subtechnique"
  def is_technique_or_subtechnique(technique)
    return MITRE_SUBTECHNIQUE if technique.include?('.')

    MITRE_TECHNIQUE
  end

  # Set the time range based on the parameters or session filters
  #
  # This method updates instance variables for start time and end time, which define the time range for filters.
  # It also updates the session data with the selected time range.
  def set_time
    if params[:range] != MITRE_CUSTOM_FILTER
      @end_time = Time.now
      if params[:range]
        time = params[:range].to_sym
      else
        filters = session['mitre_filters'] && session['mitre_filters']["filters"]
        time = filters && filters["time"] ? filters["time"].to_sym : nil
        time ||= :last_15_minutes
        time = :last_15_minutes if time == MITRE_CUSTOM_FILTER.to_sym
      end
      @range = time

      # Set the start time based on the selected range
      case time
      when :last_15_minutes
        @start_time = @end_time - 15.minutes
      when :last_30_minutes
        @start_time = @end_time - 30.minutes
      when :last_1_hour
        @start_time = @end_time - 1.hour
      when :last_2_hour
        @start_time = @end_time - 2.hours
      when :last_3_hour
        @start_time = @end_time - 3.hours
      when :last_12_hour
        @start_time = @end_time - 12.hours
      when :last_24
        @start_time = @end_time - 24.hours
      end

      # Save the selected time range to session filters
      filter_session = get_default_filters
      filter_session["filters"]["time"] = @range.to_s
      save_filter_to_session(filter_session)

    # Update slider range for UI display
    else
      @start_time = Time.at session['mitre_filters']["filters"]["start_time"]
      @end_time = Time.at session['mitre_filters']["filters"]["end_time"]
      @range = :custom
    end
    @start_time_for_slider = @start_time
    @end_time_for_slider = @end_time
  end

  # Return default filter values
  #
  # @return [Hash] A hash with default filter values
  def get_default_filters
    {
      'filters' => {
        'time' => 'last_15_minutes'
      }
    }
  end

  # Generate a custom tooltip for the specified time range
  #
  # @param [Symbol] range The time range identifier
  # @param [Time, nil] start_time The start time of the range
  # @param [Time, nil] end_time The end time of the range
  # @return [String, nil] A formatted string showing the start and end times if range is 'custom'; otherwise, nil
  def custom_tooltip(range, start_time, end_time)
    return unless (start_time && end_time) && range.to_s == 'custom'

    "From: #{@start_time_for_slider.strftime('%Y/%m/%d %H:%M:%S')} <br>" \
    "To: #{@end_time_for_slider.strftime('%Y/%m/%d %H:%M:%S')}"
  end

  # Determine the matrix color based on priority level
  #
  # @param [String] priority The priority level
  # @return [String] The corresponding color string
  def determine_matrix_color(priority)
    color_map = {
      'low' => MITRE_LOW_COLOR,
      'medium' => MITRE_MEDIUM_COLOR,
      'high' => MITRE_HIGH_COLOR
    }

    color_map.fetch(priority, MITRE_UNKNOWN_COLOR)
  end

  # Set session data related to MITRE tactics
  #
  # This method initializes instance variables from the session, which may include tactic ID, name, rule IDs, and filters.
  def set_session_data
    @tactic_id = session['mitre_tactic_id']
    @tactic_name = session['mitre_tactic_name']
    @rule_ids = session['mitre_rule_ids']
    @filters = session['mitre_filters'] || get_default_filters
  end

  # Save the given filter settings to the session
  #
  # @param [Hash] filter The filter settings to be saved
  def save_filter_to_session(filter)
    session['mitre_filters'] = filter
  end

  # Save MITRE tactic data to the session
  #
  # @param [Object] tactic An object containing tactic data
  def save_session_data(tactic)
    session['mitre_tactic_id'] = tactic.mitre_tactic_id
    session['mitre_tactic_name'] = tactic.name
    session['mitre_rule_ids'] = tactic.rule_ids
  end

  # Calculate the count of matching rules for a given entity
  #
  # @param [Object] entity An object with a rule_ids attribute, which is a list of rule IDs
  # @param [Array<Hash>] rules An array of hashes, each representing a rule with a sig_id key
  # @return [Integer] The count of rules that match the entity's rule IDs
  def calculate_matching_rules_count(entity, rules)
    rules.count { |rule| entity.rule_ids.include?(rule[:sig_id]) }
  end

  # Determine the highest priority among a list of matching rules
  #
  # @param [Array<Hash>] matching_rules An array of hashes, each representing a rule with a priority key
  # @return [String] The highest priority found among the rules, or 'low' if no priorities are found
  def find_highest_priority(matching_rules)
    priorities = matching_rules.map { |rule| rule[:priority] }.uniq
    priorities.max_by { |priority| %w[low medium high].index(priority) } || 'low'
  end

  # Get the color code for an unknown priority level
  #
  # @return [String] The color code for unknown priority, as defined by the constant MITRE_UNKNOWN_COLOR
  def get_unknown_color
    MITRE_UNKNOWN_COLOR
  end
end
