#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Methods to help the renderization at sensor UI.
#
module SensorsHelper
  #
  # Render a button to access to access points view.
  #
  # == Parameters:
  # sensor::
  #   Sensor to render access points
  #
  # == Returns:
  # String with HTML code of button.
  #
  def sensor_view_button(sensor)
    return unless Redborder::Module.registered_modules.include?(:flow) &&
                  respond_to?(:access_points_button)
    # Set the link
    access_points_button(sensor)
  end

  #
  # Render the sensor menu
  #
  # == Parameters:
  # menu::
  #   Symbol indentifies the menu to render
  # sensor::
  #   Sensor object to render the menu
  #
  # == Returns:
  # The HTML code to render the list sensor menu
  #
  def render_sensor_menu(menu, sensor)
    # Render menu and link
    list = render_menu(menu, sensor) || ''
    # Skip if the menu is empty
    return '' if list.empty?

    link = link_to '#', class: 'btn dropdown-toggle',
                        data: { toggle: 'dropdown' } do
                          content_tag(:i, nil, class: 'fa fa-list')
                        end
    ul_tag = content_tag :ul, list.html_safe, class: 'dropdown-menu'

    # Return the button and the list
    link + ul_tag
  end

  #
  # Return HTML code to show last_check_in
  #
  # == Parameters:
  # sensor::
  #   The sensor object
  #
  # == Returns:
  # A safe string with the HTML code of last-check in
  #
  def sensor_last_check_in(sensor)
    # Domains doesn't have last-check in
    return '' if sensor.is_domain?
    return color_icon(nil, sensor.unclaimed) unless sensor.last_check_in

    return_text = ''
    return_text += color_icon(sensor)
    return_text += distance_of_time_in_words(Time.now, sensor.last_check_in)

    return_text.html_safe
  end

  # Return the megabytes limit to include in the graph. It's returned as
  # javascript string, you don't need to parse with JSON.
  #
  # @param sensor [Sensor] The sensor to get the limit
  # @return [String] javascript array with the limit or an empty array;
  #
  def sensor_limit(sensor)
    return '[]' unless sensor.organization? && sensor.megabytes_limit
    "[#{sensor.megabytes_limit.to_f * 1_000_000}]"
  end

  #
  # Generate the red circle icon HTML code
  #
  # == Returns:
  # A safe String with the HTML code
  #
  def color_icon(sensor = nil, unclaimed = nil)
    if sensor.nil? && unclaimed
      content_tag(:i, nil, class: 'fa fa-circle gray') + 'unclaimed'
    elsif sensor.nil?
      content_tag(:i, nil, class: 'fa fa-circle blue') + 'configuring'
    elsif sensor.last_check_in >= Time.now - 20.minutes
      content_tag(:i, nil, class: 'fa fa-circle green')
    else
      content_tag(:i, nil, class: 'fa fa-circle red')
    end
  end

  #
  # String with the options for the role selector
  #
  # == Parameters:
  # sensor::
  #   The sensor to get the role
  # chef_value_path::
  #   Path of chef to get the data
  # options::
  #   The options available. It can be an array or hash.
  # attrib::
  #   The selector attributes like include_blank...
  #
  # == Returns:
  # The HTML code of options.
  #
  def select_role_options(sensor, chef_value_path, options, attrib = {})
    select_options = attrib[:include_blank] ? '<option value=''></option>' : ''

    # If the options parameter is an array, it must be converted to hash
    options = options.to_h if options.is_a?(Array)

    options.each do |k, v|
      name = extend_role_name(k, sensor.parent, options,
                              chef_value_path, attrib[:inherited_name])
      if attrib[:selected].to_s == v.to_s
        content = content_tag(:option, name, value: v, selected: 'selected')
      else
        content = content_tag(:option, name, value: v)
      end
      select_options += content
    end
    select_options.html_safe
  end

  #
  # Add the inherited role to the option name.
  #
  # == Parameters:
  # value::
  #   Value of the option
  # parent::
  #   The parent of current sensor
  # chef_path::
  #   Chef path to get the value
  # opt::
  #   Hash with the options available for this select
  # inherited::
  #   String value for inherited option on this select
  #
  # == Returns:
  # The name
  #
  def extend_role_name(name, parent, opt, chef_path, inherited)
    # By default, inherited string is Inherited
    inherited = 'Inherited' if inherited.nil?
    return name if (parent && !parent.role_value(chef_path).nil? &&
                    name == '') || inherited != name
    name + " (#{opt.key(parent.role_value(chef_path))})"
  end

  #
  # Return the Value for the snort type and sensor role.
  #
  # == Parameters:
  # sensor::
  #   The sensor to get the role
  # type::
  #   Type of variable
  # variable::
  #   Variable to get the value
  #
  # == Returns:
  # The value of the variable or nil if it's undefined
  #
  def selected_variable_value(sensor, type, variable)
    sensor.chef_role
      .override_attributes['redborder']['snort']['default'][type][variable]
  end

  #
  # Generate the HTML code forthe row of segments
  #
  # == Parameters:
  # sensor::
  #   The sensor to get the role
  # status::
  #   String with the current status. It can be
  #   on or bypass
  #
  # == Returns:
  # The HTML code of the TD
  #
  def segment_status_td(sensor, status)
    segments_by_status = sensor.segments_by_status[status] || []
    content_tag :td, class: 'text-center',
      data: { label: label_segment(status) } do
      content_tag :span, class: span_class_segment(sensor, status, segments_by_status.size),
        data: { toggle: 'tooltip', html: 'true' },
        title: segments_by_status.sort.join('<br>') do
          "#{segments_by_status.size}"
      end
    end
  end

  #
  # Return the classes for the text of a segment based on the status
  # and the size of segments
  #
  # == Parameters:
  # sensor::
  #   The sensor object
  # status::
  #   String with the current status. It can be
  #   on or bypass
  # segment_size::
  #   Size of segments by status
  #
  # == Returns:
  # String with classes to apply in the span
  def span_class_segment(sensor, status, segment_size)
    return '' unless sensor.segments
    equal_size = segment_size.size == sensor.segments.size

    if equal_size && status == 'on'
      'segments green'
    elsif equal_size && status == 'bypass'
      'segments red'
    else
      ''
    end
  end

  #
  # Return the classes for the label of a segment based on the status
  #
  # == Parameters:
  # status::
  #   String with the current status. It can be
  #   on or bypass
  #
  # == Returns:
  # String with classes to apply in the label
  #
  def label_segment(status)
    if status == 'on'
      'Segments ON'
    else
      'Segments Bypass'
    end
  end

  #
  # Domain types of a sensor
  #
  # == Returns:
  # Array to set in a select with the domain types
  #
  def domain_type_collection(selected = Sensor::DEPLOYMENT)
    domains = [
                ['Service Provider', Sensor::SERVICE_PROVIDER],
                ['Organization', Sensor::ORGANIZATION],
                ['Market', Sensor::MARKET],
                ['Deployment', Sensor::DEPLOYMENT]
              ]
    if current_user.super_admin? || selected == Sensor::NAMESPACE
      domains << ['Namespace', Sensor::NAMESPACE]
    end

    domains += [
                 ['Campus', Sensor::CAMPUS],
                 ['Building', Sensor::BUILDING],
                 ['Floor', Sensor::FLOOR]
               ]

    domains
  end

  #
  # Rnder the link of sensors. If the user has manage permissions,
  # the url redirect to edit view, else i will redirect to view.
  #
  # == Parameters:
  # sensor::
  #   Sensor wants to create the link
  #
  # == Returns:
  # String with the url.
  #
  def sensor_url(sensor)
    return edit_sensor_path(sensor) if current_user.admin?
    # Check if we have view action
    if current_user.modules.include?('flow')
      view_sensor_path(sensor)
    else
      map_sensor_path(sensor)
    end
  end

  #
  # Get the parent id of a sensor
  #
  # == Returns:
  # An string with the ID of the parent if it exists.
  #
  def parent_id(sensor, user = current_user)
    # We don't need to check permissions because sensors tree is render
    return '' if sensor.parent == current_user.sensor
    sensor.parent_id
  end

  #
  # Get HTML icon of a sensor
  #
  # == Returns:
  # An HTML string with sensor's icon depending on sensor type
  #
  def html_icon(sensor)
    if sensor.icon_type == 'image'
      image_tag asset_path(sensor.icon), class: "custom-icon".html_safe
    else
      content_tag(:i, nil, class: sensor.icon ).html_safe
    end
  end

  #
  # Get HTML for claim sensor link depending on the top domain's user
  #
  # == Returns:
  # An HTML string with an a tag for new claim sensor link
  #
  def claim_sensor_link(sensor_id)
    link_to new_claim_sensor_path(sensor_id),
            'data-backdrop' => 'static', 'data-toggle' => 'modal',
            'data-target' => '#claim_sensor_modal', class: 'menu-divider' do
      content_tag(:i, nil, class: 'sensor-menu-icon fa fa-plus') +
        "#{I18n.t 'claim'}"
    end
  end

  #
  # License's select input. If no licenses are available, return an alert.
  #
  def license_selector(form, sensor)
    licenses = available_licenses(sensor)
    if licenses.count.zero?
      html = content_tag(:label, "#{t 'license'}", class: 'select', for: 'sensor_license_id')
      html = content_tag(:div, class: 'alert alert-danger', role: 'alert') do
        content = "#{I18n.t('sensors_helper.no_licenses')}"
        content += content_tag(:a, "#{I18n.t('sensors_helper.add_license')}",
                               href: licenses_path, target: '_blank') if current_user.super_admin?
        content.html_safe
      end
      html.html_safe
    else
      form.input(:license_id, label: "#{t 'sensors_helper.license'}", collection: licenses, include_blank: false)
    end
  end

  #
  # Return a collection to be used in a select input that represent the available licenses
  # for a certain sensor.
  #
  def available_licenses(sensor)
    licenses = License.available_by_type(sensor.type_to_s, sensor.ancestor_organization)
    licenses << sensor.license if sensor.license && !sensor.license.expired?

    licenses.uniq.map { |l| ["#{l.uuid} - #{l.expire_at.strftime('%Y-%m-%d')} (#{l.available_slots(sensor.type_to_s)})", l.id] }
  end

  #
  # Return a list of domains that can be created above the sensor passed by parameter
  #
  def allowed_domains(sensor)
    return unless sensor.is_domain?
    content_tag(:ul, class: 'dropdown-menu add-domain-menu') do
      if sensor.is_root?
        content_tag(:li) do
          link_to icon_and_text_by_type(Sensor::SERVICE_PROVIDER), domain_new_path(domain_type: Sensor::SERVICE_PROVIDER, id: sensor.id), 'data-toggle' => "modal", 'data-target' => "#default_large", "data-backdrop" => "static"
        end
      elsif sensor.service_provider?
        content_tag(:li) do
          link_to icon_and_text_by_type(Sensor::NAMESPACE), domain_new_path(domain_type: Sensor::NAMESPACE, id: sensor.id), 'data-toggle' => "modal", 'data-target' => "#default_large", "data-backdrop" => "static"
        end
      elsif sensor.namespace?
        content_tag(:li) do
          link_to icon_and_text_by_type(Sensor::ORGANIZATION), domain_new_path(domain_type: Sensor::ORGANIZATION, id: sensor.id), 'data-toggle' => "modal", 'data-target' => "#default_large", "data-backdrop" => "static"
        end
      elsif sensor.organization?
        [Sensor::MARKET, Sensor::DEPLOYMENT, Sensor::CAMPUS, Sensor::BUILDING, Sensor::FLOOR].map do |domain_type|
          content_tag(:li) do
            link_to icon_and_text_by_type(domain_type).html_safe, domain_new_path(domain_type: domain_type, id: sensor.id), 'data-toggle' => "modal", 'data-target' => "#default_large", "data-backdrop" => "static"
          end
        end.join.html_safe
      else
        ([Sensor::MARKET, Sensor::DEPLOYMENT, Sensor::CAMPUS, Sensor::BUILDING, Sensor::FLOOR] - [sensor.domain_type]).map do |domain_type|
          content_tag(:li) do
            link_to icon_and_text_by_type(domain_type).html_safe, domain_new_path(domain_type: domain_type, id: sensor.id), 'data-toggle' => "modal", 'data-target' => "#default_large", "data-backdrop" => "static"
          end
        end.join.html_safe
      end
    end
  end

  #
  # Return the icon associated to the domain type
  #
  def icon_and_text_by_type(domain_type)
    domain_type_info = Sensor::DOMAIN_TYPES[domain_type]
    if domain_type_info[:type_icon] == 'font'
      (content_tag(:i, nil, class: domain_type_info[:icon]) + domain_type_info[:name]).html_safe
    else
      (image_tag(asset_path(domain_type_info[:icon]), class: 'custom-icon') + domain_type_info[:name]).html_safe
    end
  end
end
