#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module WidgetsHelper
  #
  # Render all the options (css classes, transparency, id...) for the
  # main container of a widget
  #
  # == Parameters:
  # widget::
  #   Widget to render
  # parent::
  #   Dashboard or Report
  # style::
  #   String with CSS attributes
  #
  # == Returns:
  # Safe String of HTML
  #
  def widget_options(widget, parent, style = '')
    html = "class='widget reloading'" \
           "data-widget-transparent='#{widget.transparent}'"
    html += "data-widget-id='#{widget.id}' \
            data-widget-type='#{widget.widget_type}'
            data-widget-graph-type='#{widget.graph_type}'"

    if parent.dashboard? && parent.absolute?
      html += "style='position: absolute; top: #{widget.position_y}; \
              left: #{widget.position_x}; width: #{widget.position_width}; \
              height: #{widget.position_height}; #{style}'"
    else
      html += "data-sizex='#{widget.position_width}' \
              data-sizey='#{widget.position_height}' \
              data-col='#{widget.position_x}' data-row='#{widget.position_y}' \
              style='#{style}'"
    end
    # Set click options
    html += widget_click_action(widget) if parent.dashboard?
    html.html_safe
  end

  #
  # Render the toolbar for current widget
  #
  # == Parameters:
  # widget::
  #   Widget to render
  # parent::
  #   Dashboard or Report
  #
  # == Returns:
  # Safe String with the HTML code
  #
  def widget_toolbar(widget, parent, manage)
    html = ''
    # Reloading icon
    html += content_tag :div, class: 'widget-reload-icon' do
              content_tag(:i, '', class: 'fa fa-spinner fa-pulse')
            end
    # Reloading icon fail
    html += content_tag :div, class: 'widget-fail-icon' do
              content_tag(:i, '', class: 'fa fa-times light-red')
            end
    # Toolbar
    html +=
      content_tag :div, class: 'widget-toolbar hidden-print' do
        toolbar = link_to '<i class="fa fa-ellipsis-v"></i>'.html_safe, '#', class: 'dropdown-toggle', 'data-toggle' => 'dropdown'
        toolbar += content_tag :ul, class: 'dropdown-menu' do
          if manage
            menu = ''.html_safe
            menu += content_tag(:li, link_to(('<i class="fa fa-copy"></i>' + I18n.t('widgets_helper.clone')).html_safe, create_url('clone_widget_path', parent, widget), 'data-toggle' => 'modal', 'data-target' => '#default_modal', 'data-backdrop' => 'static'))
            menu += content_tag(:li, link_to(('<i class="fa fa-wrench"></i>' + I18n.t('widgets_helper.edit')).html_safe, create_url('edit_widget_path', parent, widget), 'data-toggle' => 'modal', 'data-target' => '#default_xlarge', 'data-backdrop' => 'static'))
            menu += content_tag(:li, link_to(('<i class="fa fa-download"></i>' + I18n.t('widgets_helper.export')).html_safe, '#', onclick: "export_element_as_image(event, #{widget.id}, '#{widget.name}');"))
            if widget.widget_type != 'shape'
              menu += content_tag(:li, link_to(('<i class="fa fa-refresh"></i>' + I18n.t('widgets_helper.reload')).html_safe, "#", class: 'widget-reload'))
            end
            if parent.report?
              menu += content_tag :li, class: 'dropdown-submenu move-menu' do
                        code = link_to(('<i class="fa fa-arrows"></i>' + I18n.t('widgets_helper.move_to')).html_safe, '#')
                        code += content_tag :ul, class: 'dropdown-menu' do
                          code = ''.html_safe
                          for i in 1..parent.number_pages do
                            next if widget.page.to_i == i
                            code += content_tag :li, ("<a href='#' class='move-widget' data-move='#{i}'>" + I18n.t('widgets_helper.page') + " #{i}</a>").html_safe
                          end
                          code
                        end
                      end
            end
            menu += content_tag(:li, link_to(('<i class="fa fa-remove"></i>' + I18n.t('widgets_helper.delete')).html_safe, create_url('widget_path', parent, widget), method: :delete, remote: true,  data: { confirm: "#{I18n.t 'widgets_helper.confirm'}" }))
          else
            if widget.widget_type != 'shape'
              content_tag(:li, link_to(('<i class="fa fa-refresh"></i>' + I18n.t('widgets_helper.reload')).html_safe, "#", class: 'widget-reload'))
            else
              content_tag :li, link_to("#{I18n.t('widgets_helper.no_actions')}", '#', onclick: 'return false;')
            end
          end
        end
      end

    html.html_safe
  end

  #
  # => Render breadcrumb
  #
  def widget_breadcrumb(widget)
    html = ''
    # Comes from edit?
    edit = params[:action] == 'edit'
    css_class = edit ? 'disabled' : ''
    css_class += ' reload-modal'
    html = content_tag :ol, class: 'breadcrumb' do
      list = ''
      unless widget.product_type == 'none'
        list += content_tag :li,
                link_to(humanize_product(@widget.product_type),
                        edit ? '#' : create_url('list_widgets_path', @parent),
                        class: css_class,
                        'data-target' => '#select_widget_type')
      else
        unless params[:module].nil?
          module_link = create_url('list_widgets_path', @parent, @page)
          list += content_tag :li,
                link_to(humanize_product(params[:module].to_s.humanize),
                        module_link,
                        class: css_class,
                        'data-target' => '#select_widget_type')
        end
      end
      unless widget.widget_type.nil?
        # Check the link (shape hasn't need to select the graph type)
        widget_type_link = if edit
                             '#'
                           elsif widget.widget_type == 'shape' || widget.widget_type == widget.graph_type
                             create_url('list_widgets_path', @parent)
                           else
                             opts = if @parent.report?
                                      [@widget.product_type, @widget.page, { type: @widget.widget_type }]
                                    else
                                      [@widget.product_type, { type: @widget.widget_type }]
                                    end
                            unless params[:module]
                             create_url('graph_widgets_path', @parent, opts)
                            else
                              create_url('types_graphs_widgets_path', @parent, [@widget.widget_type, params[:page], {module: params[:module]}])
                            end
                           end
        list += content_tag :li,
                link_to(humanize_type(@widget.widget_type),
                        widget_type_link,
                        class: css_class,
                        'data-target' => '#select_widget_type')
      end
      if !widget.graph_type.nil? && widget.widget_type != widget.graph_type
        list += content_tag :li, humanize_graph(@widget.graph_type)
      end
      if edit
        list += content_tag :li,
                link_to(@widget.name,
                        '#',
                        class: css_class,
                        'data-target' => '#select_widget_type')
      end
      # Return the list of events to html
      list.html_safe
    end

    html.html_safe
  end

  #
  # => Set the style in the element to show the alignment and the size
  #
  def widget_text_settings(widget)
    style = 'style="'
    case widget.text_alignment.to_i
    when WidgetGraph::Text::ALIGNMENT[:center]
      style += 'text-align: center;'
    when WidgetGraph::Text::ALIGNMENT[:right]
      style += 'text-align: right;'
    when WidgetGraph::Text::ALIGNMENT[:justify]
      style += 'text-align: justify;'
    end

    if widget.report? || widget.parent.columns?
      height = "#{(widget.position_height * widget.grid_height) + ((widget.position_height - 1) * 20)}px"
    else
      height = widget.position_height
    end

    (style + " font-size: #{widget.font_size}; line-height: #{height};\"").html_safe
  end

  #
  # Return the height of the graph container. If the header is true
  # the height will be decreased on 43px.
  #
  # == Parameters:
  # widget::
  #   Widget to render
  # parent::
  #   A dashboard o Report object who render the widget
  # header::
  #   Boolean to add the 43px of header to the height
  # secondary::
  #   Integer with extra pixels (like secondary data...) to add to the height
  #
  # == Returns:
  # String with the CSS values

  def widget_height_graph(widget, parent, header = true, secondary = 0)
    head = (header && widget.name.size > 0) ? 43 : 0
    head += secondary
    if parent.report? && widget.widget_type != "hardwareinfo"
      width = "#{widget.position_width * 132 - 20}px"
      height = "#{(widget.position_height * widget.grid_height(parent)) + ((widget.position_height - 1) * 20) - head}px"
      "style='height: #{height}; width: #{width}'".html_safe
    elsif parent.dashboard? && parent.columns?
      height = "#{(widget.position_height * widget.grid_height(parent)) + ((widget.position_height - 1) * 20) - head}px"
      "style='height: #{height}'".html_safe
    elsif widget.widget_type == "hardwareinfo"
      height = "#{(widget.position_height * widget.grid_height(parent)) + ((widget.position_height - 1) * 20) - head}px"
      "style='height: #{height}'".html_safe
    else
      "style='height: #{widget.position_height.to_i - head}px'".html_safe
    end
  end

  #
  # Given a string that is either a font awesome icon name, or a URI, this
  # helper renders either an i tag or an image tag, based on the icon name
  #
  def widget_icon_tag(icon)
    if icon =~ /^fa-/  # If it is an font awesome icon
      "<i class='fa #{icon}'></i>".html_safe
    else               # If it is an image
      image_tag asset_path(icon), class: 'widget-selector-icon'
    end
  end

  #
  # => Humanize type
  #
  def humanize_type(type)
    if type == 'raw'
      type.upcase
    else
      type.titleize
    end
  end

  #
  # => Humanize the graph_type
  #
  def humanize_graph(graph)
    if graph == 'multiserie'
      'Serie'
    else
      graph.titleize
    end
  end

  #
  # => Return granularities for the form
  #
  def granularity_collection
    WidgetGraph::Base::GRANULARITY_TYPE.each_with_object([]) do |(_name, hash), memo|
      memo << [hash[:long], _name]
    end
  end

  #
  # => Return the serie type of a widget
  #
  def serie_type_collection
    WidgetGraph::Serie::SERIE_TYPES.each_with_object([]) do |(_id, name), memo|
      name = I18n.t name[3..-1] if name.start_with? "tr."
      memo << [name, _id]
    end
  end

  #
  # => Return the aggregations based on a product
  #
  def aggregation_collection(product)
    if product.to_sym.eql?(:monitor)
      # Redborder::Module.get(product.to_sym, :tabs)
      #   .each_with_object([]) do |(name, hash), memo|
      #     hash[:name] = t hash[:name][3..-1] if hash[:name].start_with? "tr"
      #     memo << [hash[:name], name]
      #   end
        Redborder::Module.get(product.to_sym, :aggregations)
        .each_with_object([]) do |(_name, hash), memo|
          hash[:name] = t hash[:name][3..-1] if hash[:name].start_with? "tr"
          memo << [hash[:name], hash[:value]] if !hash[:name].eql?"Sum"
        end

    else
      Redborder::Module.get(product.to_sym, :aggregations)
        .each_with_object([]) do |(_name, hash), memo|
          hash[:name] = t hash[:name][3..-1] if hash[:name].start_with? "tr"
          memo << [hash[:name], hash[:value]]
        end
    end
  end

  def client_profile_collection
    if Module.const_defined?('Location') && Location.const_defined?('CLIENT_PROFILE')
      Location::CLIENT_PROFILE.map do |key, value|
        translation_key = "location.options_client_profile.#{key}"
        translated_label = I18n.t(translation_key)
        [translated_label, key]
      end
    else
      []
    end
  end
  #
  # => Return a list of device sensors
  #
  def device_sensor_collection
    devices = []
    sensors = Sensor.where(:type => Sensor::DEVICE)
    sensors.each do |sensor|
      devices.push([sensor.name, sensor.id])
    end
    devices
  end

  #
  # => Return a list of widgets of type Monitor
  #
  def widget_monitor_collection
    monitorwidgets = []
    widgets = Widget.all
    widgets.each do |widget|
      monitorwidgets.push([widget.name, widget.id]) if widget.property["product_type"] == "monitor"
    end
    monitorwidgets
  end

  #
  # => Return the modules collection
  #
  def module_collection
    Redborder::Module.registered_modules.each_with_object([]) do |(k, v), memo|
      v[:name] = t v[:name][3..-1] if v[:name].to_s.start_with? "tr"
      memo << [v[:name], k] if v.fetch(:selectable_widget, false)
    end
  end

  #
  # => Return all ranges
  #
  def ranges_collection(include_custom: true)
    ranges = [
      [I18n.t('ranges.last_5_minutes'), :last_5_minute],
      [I18n.t('ranges.last_15_minutes'), :last_15_minute],
      [I18n.t('ranges.last_30_minutes'), :last_30_minute],
      [I18n.t('ranges.last_1_hour'), :last_1_hour],
      [I18n.t('ranges.last_2_hours'), :last_2_hour],
      [I18n.t('ranges.last_3_hours'), :last_3_hour],
      [I18n.t('ranges.last_12_hours'), :last_12_hour],
      [I18n.t('ranges.last_24_hours'), :last_24],
      [I18n.t('ranges.last_week'), :last_week],
    ]
    ranges.append([I18n.t('ranges.custom'), :custom]) if include_custom
    ranges
  end

  #
  # => Return a reduced version of ranges collection
  #
  def reduced_ranges_collection
    [
      [I18n.t('ranges.last_15_minutes'), :last_15_minute],
      [I18n.t('ranges.last_30_minutes'), :last_30_minute],
      [I18n.t('ranges.last_1_hour'), :last_1_hour],
      [I18n.t('ranges.last_2_hours'), :last_2_hour],
      [I18n.t('ranges.last_3_hours'), :last_3_hour],
      [I18n.t('ranges.last_12_hours'), :last_12_hour],
      [I18n.t('ranges.last_24_hours'), :last_24],
      [I18n.t('ranges.custom'), :custom]
    ]
  end

  #
  # => Return the compare ranges
  #
  def compare_ranges_collection
    [
      ["#{I18n.t 'report_blocks.form.hours'}", :pt24h],
      ["#{I18n.t 'report_blocks.form.days'}", :pt168h],
      ["#{I18n.t 'report_blocks.form.weeks'}", :pt720h],
      ["#{I18n.t 'report_blocks.form.months'}", :pt4320h]
    ]
  end

  #
  # => Return the font size range
  #
  def font_size_collection
    [
      ['12px', '12px'],
      ['16px', '16px'],
      ['24px', '24px'],
      ['36px', '36px'],
      ['48px', '48px'],
      ['72px', '72px'],
      ['96px', '96px']
    ]
  end

  #
  # => Return the color array
  #
  def color_collection
    [
      ['Blue', '#4572a7'],
      ['Red', '#aa4643'],
      ['Green', '#89a54e'],
      ['Violet', '#80699b'],
      ['Light blue', '#3d96ae'],
      ['Orange', '#db843d'],
      ['Grey blue', '#92a8cd'],
      ['Brownish red', '#a47d7c'],
      ['Light green', '#b5ca92'],
      ['Dark grey', '#666666'],
      ['Light grey', '#ededed'],
      ['White', '#ffffff']
    ]
  end

  #
  # => Return the array of available maps
  #
  def map_type_collection
    [
      ["#{t 'widgets_helper.roads'}", "roads"],
      ["#{t 'widgets_helper.satellite'}", "satellite"]
    ]
  end

  #
  # => Return a select input for map columns. If only one column is available,
  #    return a disabled input.
  #

  def map_columns_select(f, product_type)
    if product_type.to_sym == :flow
      f.input(:column, disabled: true, input_html: { value: 'WAN' })
    else
      columns = [%w(SRC src_map), %w(DST dst_map)]
      f.input(:column, collection: columns, include_blank: false)
    end
  end

  #
  # => Return the timezones
  #
  def timezone_collection
    [[I18n.t('auto_detect'), 'auto']] +
      ActiveSupport::TimeZone.all.map { |e| [e, e.name] }
  end

  #
  # => Return the array of available sizes of
  #    cluster status tables
  #
  def cluster_table_sizes_collection
    [
      ["#{t 'widgets_helper.big'}", 'big'],
      ["#{t 'widgets_helper.small'}", 'small']
    ]
  end

  #
  # => Return the array of available priorities of
  #    incident tables
  #
  def incident_priorities_collection
    Priority.all.pluck(:name, :id)
            .map { |priority| [I18n.t('incidents.priorities.' + priority[0].downcase), priority[1]] }
  end
  
  #
  # => Return the array of available STATUS of
  #    incident tables
  #
  def incident_status_collection
    IncidentStatus.all.pluck(:name, :id)
                  .map { |status| [I18n.t('incidents.status.' + status[0].downcase), status[1]] }
  end

  #
  # => Return the array of available assigned users of
  #    incident tables
  #
  def incident_users_collection
    User.joins(:incident_users).distinct
  end

  #
  # => Return the array of available incident sources of
  #    incident tables
  #
  def incident_sources_collection
    Incident.pluck(:source).uniq
            .map { |domain| [domain, domain] }
  end

  #
  # => Return the array of available domains of
  #    incident tables
  #
  def incident_domains_collection
    incident_sensor_ids = Incident.distinct.pluck(:sensor_id)
    current_user.domains_with_permission
                                  .where(id: incident_sensor_ids) # Filter by incident sensor_ids
                                  .map { |domain| [domain.name, domain.id] }
  end

  #
  # => Return all the actions can do when click
  #    on a widget
  #
  def click_actions_collection(widget)
    if widget.widget_type == 'shape'
      click_actions_without_details_collection
    elsif widget.graph_type == 'map' ||
          widget.graph_type == 'heatmap' ||
          widget.graph_type == 'sensor_tree' ||
          widget.graph_type == 'clients_map' ||
          widget.graph_type == 'hardwareinfo' ||
          widget.graph_type == 'incident_table'
      disabled_click_actions
    else
      [[I18n.t('go_details'), :details],
       [I18n.t('go_dashboard'), :dashboard],
       [I18n.t('none'), :none]]
    end
  end

  #
  # => Return all the actions less details can do when click
  #    on a widget. For example, this is used by text widgets.
  #
  def click_actions_without_details_collection
    [["#{t 'widgets_helper.go_dashboard'}", :dashboard],
     ["#{t 'widgets_helper.none'}", :none]]
  end

  #
  # => Return all the actions less details can do when click
  #    on a widget. For example, this is used by text widgets.
  #
  def disabled_click_actions
    [["#{t 'widgets_helper.none'}", :none]]
  end

  #
  # => Return all the usable raw widget analytics
  #
  def serie_analytics_collection
    [
      ["#{t 'widgets_helper.moving'}", :moving_average],
      ["#{t 'widgets_helper.previous'}", :previous_period],
      ["#{t 'widgets_helper.average'}", :average],
      ["#{t 'widgets_helper.day_over'}", :day_over_day]
    ]
  end

  #
  # => Render the widget progress bar based on the
  #    progress bar method helper
  #
  def widget_progress_bar(from_step_2)
    html = ''
    # Do not render on edit
    unless params[:action] == 'edit'
      from = from_step_2 ? 50 : 0
      html = render_progress_bar(from, 100)
    end
    html
  end

  #
  # => Renders the appropiate partial based on the graph
  #    type to append it to the widget's form.
  #
  def render_graph_form(widget, form)
    graph_type = widget.graph_type
    locals = { widget: widget, f: form }
    partial_file = "widgets/forms/graphs/#{graph_type}"
    html = render(partial: partial_file, locals: locals) rescue ''
    html.html_safe
  end

  #
  # => Renders the appropiate partial based on the widget
  #    type to append it to the widget's form.
  #
  def render_type_form(widget, form)
    widget_type = widget.widget_type
    locals = { widget: widget, f: form }
    partial_file = "widgets/forms/types/#{widget_type}"
    html = render(partial: partial_file, locals: locals) rescue ''
    html.html_safe
  end

  #
  # => Returns a string with a list of graphs allowed to a
  #    given widget type joined by commas
  #
  def graphs_list(widget_type)
    WidgetType.all[widget_type][:graphs].join(',')
  end

  #
  # Generate the correct url based on the parent
  #
  def create_url(destination, parent, options = [])
    # Split the url helper
    splitted = destination.split('_')

    # Force to array
    options = [options] unless options.is_a?(Array)

    # Calc the position to add the element
    position = splitted.size - 2

    # Clean options
    options.reject! do |m|
      (m.is_a?(Object) && m.nil?) || (!m.is_a?(Object) && m.empty?)
    end

    if request.path.include? 'reports'
      # Reports url
      splitted.insert(position, 'report')
      send("#{splitted.join('_')}", parent.id, *options)
    else
      # dashboards url
      splitted.insert(position, 'dashboard')
      send("#{splitted.join('_')}", parent.id, *options)
    end
  end

  # Render the graph selector for the second step of widget generation process
  #
  # @param type [Symbol] Widget type to render graph selector form
  # @return [String] HTML safe code
  #
  def graph_selector(type)
    # Get available graphs
    graphs = WidgetGraph.all_sorted.select do |id, _|
      WidgetType.all[type.to_sym][:graphs].include? id
    end
    # Build HTML
    html = content_tag(:div, class: 'widget-graph-wrap') do
      render_graphs_selectors graphs
    end
    # Return the current HTML + example images
    html + content_tag(:div, class: 'widget-graph-example') do
      render_graphs_images graphs
    end
  end

  # HTML code for the graphs selectors
  #
  # @param graphs [Hash] Available graph types
  # @return [String] HTML safe code
  #
  def render_graphs_selectors(graphs)
    # Initialize html
    html = ''
    graphs.each do |id, graph|
      classes = 'widget-graph-selector'
      classes += (id == graphs.keys.first ? ' selected' : '')
      html += content_tag :div, class: classes, 'data-widget-graph' => id do
        I18n.t("widgets.forms.graphs.type.#{graph[:name].downcase}")
      end
    end
    html.html_safe
  end

  # HTML code for the graphs images
  #
  # @param graphs [Hash] Available graph types
  # @return [String] HTML safe code
  def render_graphs_images(graphs)
    # Initilize HTML
    html = ''
    graphs.each do |id, graph|
      style = id == graphs.keys.first ? 'display: block;' : ''
      html += content_tag(:div, class: 'widget-graph-selector',
                                style: style, 'data-widget-graph-example' => id) do
        image_tag graph[:image]
      end
    end
    html.html_safe
  end

  #
  # => Return the first graph for a certain type
  #
  def first_graph_for(widget_type)
    WidgetType.all[widget_type.to_sym][:graphs].first
  end

  #
  # Add the correct class to the widget placeholder (div to show graphs)
  #
  # == Parameters:
  # widget::
  #   Widget to render
  #
  # == Returns:
  # String with the class
  #
  def widget_class(widget)
    ret = ''

    if Redborder::Module.keys_with_interface.include? widget.product_type.to_sym
      case widget.graph_type.to_sym
      when :pie then ret += 'widget_pie'
      when :area, :stacked, :bars, :sbars, :line
        ret += 'widget_chart'
      when :table then ret += 'widget_table'
      when :value then ret += 'widget_value'
      when :map then ret += 'widget_map'
      when :heatmap then ret += 'widget_map'
      when :streams then ret += 'widget_streams'
      when :clients_map then ret += 'widget_clients_map'
      when :hardwareinfo then ret += 'widget_vertical_scroll' if widget.graph_type.to_sym == :hardwareinfo
      else ret += ''
      end if widget.graph_type
    elsif widget.product_type.to_sym == :sensors
      ret += 'widget_table'
    elsif widget.product_type.to_sym == :cluster
      ret += 'widget_cluster'
    elsif widget.product_type.to_sym == :custom
      ret += 'widget_custom'
    elsif widget.graph_type.to_sym == :incident_table
      ret += 'widget_vertical_scroll'
    end
    ret
  end

  #
  # Render the data to perform the click event in the widget. It can be go
  # to details, none...
  #
  # == Parameters:
  # widget::
  #   Widget to render
  #
  # == Returns:
  # String with the attributes.
  #
  def widget_click_action(widget)
    return 'data-widget-click=\'none\'' if widget.click_action == 'none'
    html = ''

    if widget.click_action.to_sym == :details
      html += 'data-widget-click=\'details\' '
      if widget.graph_type.to_sym == :custom
        html += "data-widget-url='#{widget.url}' "
      elsif widget.widget_type.to_sym == :sensor
        html += "data-widget-url='#{sensors_path}'"
      elsif widget.widget_type.to_sym == :cluster
        html += "data-widget-url='#{managers_path}'"
      elsif widget.widget_type.to_sym == :alarm
        html += "data-widget-url='#{alarms_path}'"
      elsif widget.widget_type.to_sym == :priority || widget.widget_type.to_sym == :source || widget.widget_type.to_sym == :status
        html += "data-widget-url='#{incidents_path}'"
      elsif widget.widget_type.to_sym == :assigned || widget.widget_type.to_sym == :domain
        html += "data-widget-url='#{incidents_path}'"
      elsif widget.widget_type.to_sym == :playbooks
        html += "data-widget-url='#{playbook_templates_path}'"
      elsif widget.widget_type.to_sym == :combination
        html += "data-widget-url='#'"
      else
        begin
          url = if WidgetType.all[widget.widget_type.to_sym][:url]
                  WidgetType.all[widget.widget_type.to_sym][:url]
                else
                  send("more_info_url_#{widget.product_type}", widget)
                end
          html += "data-widget-url='#{url}'"
        rescue Exception => e
          Rails.logger.error e
          html += 'data-widget-url=\'\''
        end
      end
    elsif widget.click_action.to_sym == :dashboard
      html += 'data-widget-click=\'dashboard\' '
      html += "data-widget-url='#{dashboard_path(widget.click_dashboard)}'"
    end

    html
  end

  def widget_sensor_table_type_collection
    list = []
    menu_items_for(:widget_sensor_table_type_menu) do |node|
      list << [node.caption, node.name]
    end
    list
  end

  # Clean the widget collection to not show a widget of a non installed module
  def check_widgets_modules(widgets, user)
    widgets_active_modules = []

    widgets.each do |widget|

      # If the widget isn't inside the modules... delete
      if (Redborder::Module.keys_with_interface + [:custom, :sensors, :cluster, :text, :date, :square, :image, :video, :webcam, :alarms]).include? widget.product_type.to_sym
        next if widget.product_type.to_sym == :cluster && !user.super_admin?
        widgets_active_modules << widget
      end

    end

    return widgets_active_modules
  end

  # Convert hex color and opacity to rgba
  def hex_opacity_to_rgba(hex, opacity)
    hex = hex[1..-1] if hex.start_with? '#'
    "rgba(#{hex[0,2].hex},#{hex[2,2].hex},#{hex[4,2].hex},#{opacity})"
  end
end
