#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'csv'
#
# => Store conditions to trigger alarms. We can send the alarm
#    throught email, logs and mobile. A delayed Job check the alarms
#
class Alarm < ApplicationRecord
  include Redborder::Audit

  # ==========================================================
  # Constants
  # ==========================================================
  MULTIPLIER_RANGE = [0, 3, 6, 9]
  INTERVAL_RANGE = [300, 900, 1800, 3600, 21_600, 43_200, 86_400]
  TIME_WINDOW_RANGE = [ 300, 600, 900, 1800, 3600, 7200, 28800, 86400, 172800, 259200, 345600, 432000, 518400, 604800]

  # ==========================================================
  # Custom attributes
  # ==========================================================

  # ==========================================================
  # Relations
  # ==========================================================

  belongs_to :user
  belongs_to :severity

  belongs_to :incident, optional: true
  belongs_to :sensor, optional: true

  # ================================================================
  # Callbacks
  # ================================================================

  before_save :remove_disabled_filters

  # ================================================================
  # Validations
  # ================================================================

  validates :name, presence: true,  length: { in: 1..160 },
            format: {
              with: /\A[\w\s_\.\-,@+=()!\/:#>&\u007C[\u00C0-\u0178][\u0400-\u04FF]]{2,150}\z/,
              message: "#{I18n.t('models.allowed_characters')}"
            }
  validates :interval, :time_window, :product_type,
            :severity, presence: true
  validates :upper_limit, presence: true, if: :upper_limit_enabled?
  validates :upper_limit, numericality: { only_integer: true },
                          if: :upper_limit_enabled?
  validates :bottom_limit, presence: true, if: :bottom_limit_enabled?
  validates :bottom_limit, numericality: { only_integer: true },
                           if: :bottom_limit_enabled?
  validates :product_type, inclusion: { in: :product_types_registered }
  validates :multiplier_upper, inclusion: { in: MULTIPLIER_RANGE },
                               allow_blank: true
  validates :multiplier_bottom, inclusion: { in: MULTIPLIER_RANGE },
                                allow_blank: true
  validates :time_window, inclusion: { in: TIME_WINDOW_RANGE }
  validates :interval, inclusion: { in: INTERVAL_RANGE }
  validate :upper_or_bottom_limit
  validate :check_email_list
  validate :check_slack_webhook

  audited on: [:create, :update, :destroy, :launch]

  # ================================================================
  # Public methods
  # ================================================================

  #
  # => Returns the unit of the upper or bottom limit
  #    limit = [:multiplier_upper, :multiplier_bottom]
  #
  def units(limit)
    send("units_#{product_type}", limit, aggregation)
  end

  #
  # Update last_fired, fired and save the model
  #
  def launch
    # For some reason  when is nil its not generating an audit
    # so when is nill we do it manually.
    send_audit = fired

    self.fired = Time.now
    self.last_fired = Time.now

    if save && !send_audit
      register_alarm_changes(self, "Alarm (#{name}) was triggered at #{Time.now}")
    end
  end

  def outliers_enabled?
    outliers_enabled
  end

  #
  # Disable current alarm
  #
  def disable
    self.enabled = false
    save
  end

  #
  # Enable current alarm
  #
  def enable
    self.enabled = true
    save
  end

  #
  # Return if alarm was fired
  #
  def fired?
    fired.present?
  end

  #
  # Return true if the alarm is enabled and it isn't
  # in the "waiting period". This period is the range between
  # last fired and the interval selected.
  #
  def must_be_checked?
    enabled && (fired.nil? || (fired + interval) <= Time.now)
  end

  def upper_limit_enabled?
    upper_limit_enabled
  end

  def bottom_limit_enabled?
    bottom_limit_enabled
  end

  #
  # => Returns the upper limit multiplied
  #    with the proper factor.
  #    18 Kbps = 18 * (10**3)
  #    18 Mbps = 18 * (10**6)
  #
  def upper_mul_limit
    return upper_limit if multiplier_upper.nil?
    upper_limit * (10**multiplier_upper)
  end

  #
  # => Returns the bottom limit multiplied
  #    with the proper factor.
  #    18 Kbps = 18 * (10**3)
  #    18 Mbps = 18 * (10**6)
  #
  def bottom_mul_limit
    return bottom_limit if multiplier_bottom.nil?
    bottom_limit * (10**multiplier_bottom)
  end

  #
  # => Return products registered in redBorder
  #
  def product_types_registered
    Redborder::Module.keys_with_interface.map(&:to_s)
  end

  #
  # => Returns the name of the product
  #
  def product_type_humanize
    Redborder::Module.registered_modules[product_type.to_sym][:name]
  rescue NoMethodError
    'Default'
  end
  
  def self.import_from_csv(content_file, current_user)
    created_count = 0
    updated_count = 0
  
    csv_data = CSV.read(content_file.path, col_sep: ';', headers: true, encoding: 'UTF-8')

    columns = csv_data.headers.map(&:to_sym)
    
    CSV.foreach(content_file.path, col_sep: ';', headers: false, encoding: 'UTF-8') do |row|
      next if row.empty? || row[0].nil?
  
      attrs = {}
      columns.each_with_index do |col, index|
        attrs[col] = row[index].try(:strip)
      end

      [:enabled, :upper_limit_enabled, :bottom_limit_enabled, :notify_to_syslog,
       :notify_to_email, :notify_to_mobile, :notify_to_slack, :outliers_enabled,
       :incident_enabled].each do |bool_field|
        attrs[bool_field] = (attrs[bool_field].to_s.downcase == "true")
      end
  
      [:interval, :upper_limit, :bottom_limit, :time_window, :multiplier_upper,
       :multiplier_bottom, :severity_id, :user_id, :sensor_id, :incident_id].each do |num_field|
        attrs[num_field] = attrs[num_field].to_i if attrs[num_field].present?
      end
  
      
      [:fired, :last_fired].each do |dt_field|
        begin
          attrs[dt_field] = Time.parse(attrs[dt_field]) if attrs[dt_field].present?
        rescue
          attrs[dt_field] = nil
        end
      end
  
      attrs[:created_at] = Time.now
  
      attrs[:user_id] = current_user.id

      alarm = Alarm.find_or_initialize_by(id: attrs[:id])
      if alarm.new_record?
        alarm.assign_attributes(attrs)
        if alarm.save
          created_count += 1
        end
      else
        # Actualiza los atributos (excluyendo created_at)
        attrs.delete(:created_at)
        if alarm.update(attrs)
          updated_count += 1
        end
      end
    end
  
    [created_count, updated_count]
  end
  

  # ================================================================
  # Private methods
  # ================================================================

  private

  def upper_or_bottom_limit
    return if upper_limit_enabled? || bottom_limit_enabled?
    errors.add(:upper_limit, "#{I18n.t('models.one_limit')}")
  end

  #
  # => Check if emails are well-formed
  #
  def check_email_list
    return unless notify_to_email.present?

    if email_list.present?
      incorrect = email_list.split(',').any? do |email|
        check_email(email)
      end
      errors.add(:email_list, "#{I18n.t('models.not_formatted')}") if incorrect
    else
      errors.add(:email_list, "#{I18n.t('models.present')}")
    end
  end

  #
  # Receive an email and return if it's valid
  #
  # == Parameters:
  # email::
  #   String with email to check
  #
  # == Returns:
  # true or false
  def check_email(email)
    reg = /^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,4}$/i
    email.match(reg).nil?
  end

  # Receive a Slack Webhook and checks that
  #
  # is correctly formatted

  def check_slack_webhook
    return unless notify_to_slack.present?
    reg = /^https\:\/\/hooks\.slack\.com\/services\/.*$/
    if slack_webhook.present?
      is_match = slack_webhook.match(reg)
      if !is_match
        errors.add(:slack_webhook, "#{I18n.t('models.not_formatted')}")
      end
    else
      errors.add(:slack_webhook, "#{I18n.t('models.present')}")
    end
  end
  #
  # => If there are disabled filters, remove them.
  #
  def remove_disabled_filters
    return nil if condition.nil?
    self.condition = (JSON.parse(condition).select do |_, value|
      value['enabled']
    end).to_json
  end

end
