#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ApplicationObject < ApplicationRecord
  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------

  validates :name, presence: true,
                   format: { with: /\A[\w\s\/\-\.\+]+\z/,
                             message: "#{I18n.t('models.only')}" }
  validates :value, presence: true,
                    format: { with: /\A\d+:\d+\z/,
                              message: "#{I18n.t('models.only_x_y')}" }
  validates_uniqueness_of :value

  after_save :update_app_databag
  after_destroy :update_app_databag

  # -------------------------------------------------------------
  # Class methods
  # -------------------------------------------------------------

  #
  # Create applications from a file. This method open the file and process all
  # elements. If exist an application with same name, it will be updated.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.add_apps_from_file(content_file)
    file = File.open(content_file.tempfile)
    apps = ApplicationObject.all
    created_count = 0
    updated_count = 0

    file.each do |line|
      columns = line.sub(/[\n]+/, '').split(';')
      name, value, description = columns

      app = apps.where(name: name).first

      # update
      if app.present? && app.value != value && app.description != description
        app.update_attributes(value: value, description: description)
        next unless app.save
        updated_count += 1
      # create
      elsif app.nil?
        app = ApplicationObject.new(name: name, value: value,
                                    description: description)
        next unless app.save
        created_count += 1
      end
    end

    file.close
    [created_count, updated_count]
  end

  #
  # Replace applications from a file. This method open the file and process all
  # elements.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.replace_apps_from_file(content_file)
    file = File.open(content_file.tempfile)
    apps = ApplicationObject.all

    created_count = 0

    file.each do |line|
      columns = line.sub(/[\n]+/, '').split(';')
      name, value, description = columns

      app = apps.where(name: name).first

      # update
      if app.present?
        app.update_attributes(value: value, description: description)
      # create
      else
        app = ApplicationObject.new(name: name, value: value,
                                    description: description)
      end

      next unless app.save
      created_count += 1
    end

    file.close
    created_count
  end

  private

  def map_value(value)
    x, y = value.split(':').map(&:to_i)
    (x << 24) + y
  end

  def update_app_databag
    UpdateApplicationDatabag.call(
      { map_value(value).to_s => { name: name,
                                   value: value,
                                   value_int: map_value(value) } },
      destroyed?
    )
  end
end
