#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Implements getters and setters for Chef properties safely
# Raises an ChefPropertyUnavailable if the node is not available
# Raises an ChefPropertyNotFound if the path dont exist
#
module ChefProperty
  extend ActiveSupport::Concern

  class ChefPropertyUnavailable < StandardError; end;
  class ChefPropertyIncorrectPath < StandardError; end;

  included do
    # The problem of unsaved properties only appears in sensor creation. The
    # problem is chef node is created in a after_create and I can't order the
    # callbacks. So I put it after all after_create (in after_save). The
    # method will be executed in update too, but unsaved_chef_properties will
    # be nil so the callback will be skipped.
    #
    # - Angel
    after_save :apply_unsaved_chef_properties
  end

  #
  # Store non saved chef properties. We need this funciton because the setter
  # can try to call chef_node and it only will be availabe after save the
  # element. The problem only affect to creation.
  #
  def apply_unsaved_chef_properties
    return if @unsaved_chef_properties.blank?
    # Iterate over stored values
    @unsaved_chef_properties.keys.each do |property|
      send("#{property}=", @unsaved_chef_properties[property])
    end
  end

  #
  # Create a method to define a setter and getter for chef properties
  #
  module ClassMethods
    def has_chef_property(property_name, path)
      define_chef_property(property_name, path, :node)
    end

    def has_chef_role_property(property_name, path)
      define_chef_property(property_name, path, :role)
    end

    def define_chef_property(property_name, path, type)
      define_method(property_name) do
        if new_record?
          @unsaved_chef_properties ||= {}
          return @unsaved_chef_properties[property_name]
        end
        store = send("chef_#{type}")
        path_cpy = path.dup
        fail ChefPropertyUnavailable if store.nil? || path_cpy.empty?

        current_store = type == :role ? store.override_attributes : store
        last_leaf = path_cpy.pop

        path_cpy.each do |element|
          if type == :node
            current_store = current_store[element]
            fail ChefPropertyIncorrectPath if current_store.nil?
          else
            current_store[element] ||= {}
            current_store = current_store[element]
          end
        end

        # TODO: This functions returns a copy of the chef value
        # That means that you if you use methods like <<, push, pop
        # that is not going to work and WONT BE SAVED until you use the
        # setter below
        value = current_store[last_leaf]
        value = value.to_s if value.is_a?(Fixnum)
        value = value.dup unless value.nil?
        value
      end

      # Defines the setter
      define_method("#{property_name}=") do |param|
        # If it's a new record, we must store variables to apply it to store
        # after save it
        if new_record?
          @unsaved_chef_properties ||= {}
          @unsaved_chef_properties[property_name] = param
          return
        end
        store = send("chef_#{type}")
        path_cpy = path.dup
        fail ChefPropertyUnavailable if store.nil? || path_cpy.empty?

        current_store = type == :role ? store.override_attributes : store.normal
        last_leaf = path_cpy.pop

        path_cpy.each do |element|
          if type == :node
            current_store = current_store[element]
            fail ChefPropertyIncorrectPath if current_store.nil?
          else
            current_store[element] ||= {}
            current_store = current_store[element]
          end
        end

        current_store[last_leaf] = param
        store.save

        value = current_store[last_leaf]
        value = value.to_s if value.is_a?(Fixnum)
        value = value.dup unless value.nil?
        value
      end
    end
  end
end
