#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Implements exportable and importable behaviour for rails models,
#    including importing and exporting their relations if needed in the
#    same compressed file
#
module Exportable
  extend ActiveSupport::Concern

  included do
    @exportable_relations = []
  end

  module ClassMethods
    attr_reader :exportable_relations

    #
    # => Methods that the model should call to implement importable
    #    and exportable behaviour
    #
    def exportable_opts(options = {})
      @exportable_relations = options[:relations]
    end

    #
    # => Creates a element from an imported file
    #
    def import(file_params)

      importing = new
      extracted_folder = Redborder::ExportableUtils.extract_files file_params
      main_file = extracted_folder + '/' + importing.class.name.underscore + '.json'
      media_file = File.open(extracted_folder + '/media/' + importing.class.name.underscore, 'r') rescue nil

      # Process main file
      unless File.exist? main_file
        fail "#{I18n.t 'models_concern.wrong_format'}"
      end

      File.open(main_file, 'r') do |file|
        json = JSON.parse file.read

        if importing.respond_to? :import_callback
          json = importing.import_callback json, media_file
        end

        importing.attributes = json
      end

      # Process relations
      @exportable_relations.each do |relation|
        relation_file = extracted_folder + '/' + relation.to_s + '.json'
        next unless File.exist? relation_file

        File.open(relation_file, 'r') do |file|
          data = JSON.parse file.read
          data.each do |element|
            element = JSON.parse element if element.is_a? String
            # Don't create a new widget if the module, widget type or graph type don't exist.
            next unless WidgetType.all.keys.include?(element['property']['widget_type'].to_sym) &&
                        WidgetType.all[element['property']['widget_type'].to_sym][:graphs]
                        .include?(element['property']['graph_type'].to_sym)
            instance = importing.send(relation).build
            media = nil
            if element['id'].is_a? Integer
              media = File.open("#{extracted_folder}/media/#{element['id']}", 'r') rescue nil
            end
            element.delete 'id'

            if instance.respond_to? :import_callback
              element = instance.import_callback element, media
            end
            # Don't assign attributes if they aren't defined for the new widget.
            instance.attributes = element.delete_if { |k, _v| !instance.respond_to?(k) }
          end
        end
      end

      importing
    end
  end

  #
  # => Exports the element as a tar file
  #
  def export
    folder = Redborder::ExportableUtils.generate_export_folder
    class_name = self.class.name.underscore

    File.open("#{folder}/#{class_name}.json", 'w') do |file|
      if self.respond_to? :export_json
        file.write export_json
      else
        file.write to_json
      end
    end

    if self.respond_to? :export_media
      media = export_media
      unless media.nil?
        open("#{folder}/media/#{class_name}", 'wb') do |file|
          file.write export_media
        end
      end
    end

    self.class.exportable_relations.each do |relation_name|
      relation = send(relation_name)
      # We need it to avoid erros to get the name
      next if relation.empty?

      File.open("#{folder}/#{relation_name.to_s}.json", 'w') do |file|
        mapped = relation.map do |element|
          if element.respond_to? :export
            element.export
          else
            element.attributes
          end
        end

        # mapped is a string with JSON
        # calling mapped.to_json directly will cause
        # the scape characters to multiply
        file.write mapped.to_json
      end

      relation.each do |element|
        next unless element.respond_to? :export_media
        element_media = element.export_media
        next if element_media.nil?
        open("#{folder}/media/#{element.id}", 'wb') do |file|
          file.write element_media
        end
      end
    end

    Redborder::ExportableUtils.compress_folder folder, name
  end
end
