#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Implements a attached file that can be stored
#    on S3 or on the local disk, depending if the
#    file aws is present or not
#
module StorageFile
  extend ActiveSupport::Concern

  AWS_FILE_PATH = "#{Rails.root}/config/aws.yml"
  LOCAL_STORAGE = "#{Rails.root}/data"

  included do
  end

  def self.s3_enabled?
    @s3_enabled ||= File.exist? AWS_FILE_PATH
  end

  def self.get_uri(path)
    return path unless s3_enabled?
  
    bucket = Aws::S3::Resource.new(client: s3_get_client).bucket(s3_get_credentials[:bucket])
  
    # Remove the first slash
    path = path[1..-1] if path.starts_with?('/') && s3_get_credentials[:s3_host_name] =~ /\.amazonaws\.com$/
  
    # Get the URL to the S3 object
    object = bucket.object(path)
    URI.parse(object.presigned_url(:get, expires_in: 3600)) # Adjust expiration as needed
  end

  def self.delete_file(path)
    if s3_enabled?
     
      bucket = Aws::S3::Resource.new(client: s3_get_client).bucket(s3_get_credentials[:bucket])
      
      # Delete object
      begin
        s3_get_client.delete_object(bucket: bucket.name, key: path)
        return true # Success
      rescue StandardError => e
        Rails.logger.error "Error deleting object: #{e.message}"
        return false
      end
    else 
      if File.exist? (LOCAL_STORAGE + path)
        FileUtils.rm(LOCAL_STORAGE + path)
        Rails.logger.info "File: '#{path}' succesfully deleted."
      else
        Rails.logger.error "File: '#{path}' doesn't exist."
      end
    end
  end

  private

  def self.s3_get_credentials
    s3_credentials = YAML.load(File.read(AWS_FILE_PATH))[Rails.env]
    s3_credentials["bucket"].chomp!('/')
    s3_credentials.symbolize_keys!
    s3_credentials
  end

  def self.s3_get_client
    s3_client = Aws::S3::Client.new(
      access_key_id: s3_get_credentials[:access_key_id],
      secret_access_key: s3_get_credentials[:secret_access_key],
      region: 'us-east-1',
      endpoint: "https://" + s3_get_credentials[:s3_host_name],
      ssl_verify_peer: false, 
      force_path_style: true # Required for Minio
    )
    s3_client
  end

  module ClassMethods

    def storage_file(name, options = {})
      if StorageFile.s3_enabled?
        s3_credentials = StorageFile.s3_get_credentials
        has_attached_file name,
                          storage: :s3,
                          s3_protocol: :s3,
                          s3_region: 'us-east-1',
                          path: options[:path],
                          styles: options.fetch(:styles, {}),
                          convert_options: options[:convert_options],
                          s3_credentials: s3_credentials,
                          s3_options: {
                            endpoint: "https://" + s3_credentials[:s3_host_name],
                            force_path_style: true,
                            ssl_verify_peer: false
                          },
                          s3_host_name: s3_credentials[:s3_host_name],
                          url: ":s3_path_url",
                          s3_permissions: :private
      else
        has_attached_file name,
                          path: "#{Rails.root}/data/#{options[:path]}",
                          styles: options.fetch(:styles, {}),
                          convert_options: options[:convert_options]
      end
      do_not_validate_attachment_file_type name
    end
  end

  def image_uri
    return nil if image.path.nil?
    StorageFile.get_uri image.path
  end

  def media_uri
    return nil if media.path.nil?
    StorageFile.get_uri media.path
  end

  def pdf_uri
    return nil if pdf_file.options[:path].nil?
    StorageFile.get_uri pdf_file_path
  end
  
end
