#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Dashboard < ApplicationRecord
  include StorageFile
  include Exportable

  # ==========================================================
  # Custom attributes
  # ==========================================================

  attr_accessor :collaborators
  attr_accessor :sensors_collaborators
  storage_file :image, path: '/rb-webui/:class/:id'
  exportable_opts relations: [:all_widgets]
  serialize :img_position, Hash

  # ==========================================================
  # Relations
  # ==========================================================

  has_many :dashboard_users, dependent: :destroy
  has_many :dashboard_sensors, dependent: :destroy
  has_many :users, through: :dashboard_users
  has_many :sensors, through: :dashboard_sensors
  has_many :all_widgets, dependent: :destroy, class_name: 'Widget', as: :parent

  belongs_to :owner, class_name: 'User', foreign_key: :user_id,
                     inverse_of: :dashboards

  # ================================================================
  # Callbacks
  # ================================================================

  before_save :set_image_dimensions
  before_save :reset_default_dashboard
  before_save :check_layout, if: proc { |d| d.layout_changed? }
  before_save :save_collaborators, unless: proc { |d| d.locked_changed? }
  before_save :save_sensors_collaborators, unless: proc { |d| d.locked_changed? }

  after_initialize :defaults

  # ================================================================
  # Validations
  # ================================================================

  validates :name, presence: true
  validates :layout, inclusion: { in: %w(columns absolute) }
  validates :background_color, format: {
    with: /\A#([0-9a-fA-F]{3}|[0-9a-fA-F]{6})\z/,
    message: 'only allows colors such as #123456'
  }

  validates_attachment :image, content_type: {
    content_type: %w(image/jpg image/jpeg image/png)
  }, size: { in: 0..10.megabytes }

  # Only allow one dashboard per user on community version
  DittoCode::Exec.if 'COMMUNITY' do
    validates_uniqueness_of :user_id
  end

  # ================================================================
  # URL helpers
  # ================================================================

  delegate :url_helpers, to: 'Rails.application.routes'

  # ================================================================
  # Public methods
  # ================================================================

  #
  # => Helpers to check for the dashboard layout type
  #
  def absolute?
    layout.to_sym == :absolute
  end

  def columns?
    layout.to_sym == :columns
  end

  #
  # => Only returns widgets that can be shown based on the current modules
  #
  def widgets
    allowed_products = Redborder::Module.keys_with_interface + [:none, :all]

    all_widgets.select do |widget|
      allowed_products.include? widget.product_type.to_sym
    end
  end

  #
  # =>  Duplicate the dashboard and related widgets
  #
  def dup
    new_dashboard = super
    new_dashboard.image = image

    all_widgets.each do |widget|
      new_widget = Widget.new(widget
                              .attributes
                              .except('id', 'dashboard_id', 'created_at',
                                      'updated_at', 'media_content_type',
                                      'media_file_size', 'media_file_name',
                                      'media_updated_at'))
      # Check if we have an image
      if %(image video).include?(widget.graph_type) && !widget.media.nil?
        new_widget.media = widget.media
      end

      # Store new widget
      new_dashboard.all_widgets << new_widget
    end

    new_dashboard
  end

  #
  # Convert the current dashboard to a report. If the dashboard is
  # column based the report will use the position of the widget and
  # break page every 8 widgets.
  # If the dashboard is  absolute, the report will save all the widgets
  # in the custom blocks box.
  #
  def to_report
    # Create a new report
    report = User.current_user.reports.build(name: name, orientation: Report::ORIENTATION_LANDSCAPE, number_pages: 0)
    # Save to get the ID
    report.save

    # Init the pages
    page = 1
    position_y = 1
    position_y_offset = 0
    # Only for aboslute
    position_x = 1
    per_page = 0

    # Iterate the widgets ordered by position
    widgets.sort_by { |w| [w.position_y, w.position_x] }.each do |widget|
      # Check if we can add this widget to a report
      next unless widget.can_report?

      # Clone the widget and assign the report to the parent.
      rw = widget.deep_dup
      report.widgets << rw

      # Check the layout
      if columns?
        # if a page change is needed, an offset must be calculated
        # and the rest of widgets must be pushed down
        page_init = ((rw.position_y + position_y_offset) / 8.to_f).ceil
        page_end = ((rw.position_y + position_y_offset + rw.position_height - 1) / 8.to_f).ceil

        if page_init != page_end
          position_y_offset += (rw.position_y + position_y_offset + rw.position_height - 1) % 8
        end

        # calculate the page of the current widget
        rw.page = ((rw.position_y + position_y_offset + rw.position_height - 1) / 8.to_f).ceil
        # calculate the position of the widget with the offset applied
        rw.position_y = if (rw.position_y + position_y_offset) <= 8
                          (rw.position_y + position_y_offset) % 9
                        else
                          (rw.position_y + position_y_offset) % 9 + 1
                        end
      else
        # Check if the page is completed.
        if per_page >= 18
          # Reset all params, we are on a new page
          per_page = 0
          position_y = 1
          position_x = 1
          page += 1
        end
        # Change the line
        if position_x > 5
          position_y += 2
          position_x = 1
        end
        # Absolute, asign the position and the page
        rw.position_y = position_y
        rw.position_x = position_x
        rw.position_width = 2
        rw.position_height = 2
        rw.page = page
        # Increase vars
        per_page += 1
        position_x += 2
      end

      # Save the widget
      rw.save
    end

    # Add the pages and save
    report.number_pages = report.widgets.map(&:page).max
    report.save
    report
  end

  # ================================================================
  # Define the type for widgets
  # ================================================================

  #
  # Check if the class is a dashbaord or a report
  #
  # == Returns:
  # True, this is a dashboard
  #
  def dashboard?
    true
  end

  #
  # Check if the class is a dashbaord or a report
  #
  # == Returns:
  # False, this is a dashboard
  #
  def report?
    false
  end

  # ================================================================
  # Permissions for users. We MUST implements this methods.
  # They are required for widgets
  # ================================================================

  #
  # Check if the user has permissions on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def permission?(user = self.owner)
    return false if user.nil?
    user.writable_dashboards.include? self
  end

  #
  # Check if the user has read permission on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def read_permission?(user = self.owner)
    return false if user.nil?
    user.readable_dashboards.include? self
  end

  #
  # Main URL to redirect when the permissions are wrong
  #
  # == Returns:
  # An url to redirect
  #
  def main_path
    url_helpers.dashboards_path
  end

  # ================================================================
  # Import/export methods for exportable concern
  # ================================================================

  #
  # => The attributes of this object that will be exported
  #
  def export_json
    to_json(except: [
      :id, :user_id, :created_at, :updated_at, :image_content_type,
      :image_file_size, :image_updated_at, :image_file_name,
      :content_type, :time
    ])
  end

  #
  # => Called when the object will be exported to get a media file
  #    (img, video, audio, etc) that will be exported as well
  #
  def export_media
    return unless absolute? && image?
    open(image_uri, ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE).read
  end

  #
  # => Called when an object of this class is imported, so it
  #    can be customized and save correctly into the DB.
  #
  def import_callback(params, media)
    params['img_position'] =
      Hash[params['img_position'].map { |k, v| [k.to_sym, v] }]
    self.image = media unless media.nil?
    params
  end

  #
  # => Return the maximum number of rows that the columnar dashboard can use. At least, it will
  #    use 15 rows.
  #
  def max_rows
    return 0 if absolute?
    rows = widgets.sort_by { |w| [w.position_y, w.position_x] }.last.try('position_y') || 15
    [rows, 15].max
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  # Calculate the scale to convert 8 col dashboards to 6 col reports. I use:
  # http://snipplr.com/view/67224/scale-a-number-into-a-defined-range/
  #
  # @param value [Integer] Value to convert
  # @return [Integer] Scaled value
  #
  def scale_width_to_report(value)
    ((6 - 1) * (value - 1)) / (8 - 1) + 1
  end

  # Calculate the scale to convert 8 col dashboards to 6 col reports. I use:
  # http://snipplr.com/view/67224/scale-a-number-into-a-defined-range/
  #
  # @param value [Integer] Value to convert
  # @return [Integer] Scaled value
  #
  def scale_x_to_report(value)
    (((6 - 1) * (value - 1)).to_f / (8 - 1) + 1).ceil
  end

  def set_image_dimensions
    return if image.queued_for_write.empty? || !absolute?
    img = image.queued_for_write[:original]
    dimensions = Dimensions.dimensions(img.path)
    img_position[:width] = "#{dimensions[0]}px"
    img_position[:height] = "#{dimensions[1]}px"
  end

  def reset_default_dashboard
    return unless default
    owner.dashboards.where(default: true).each do |d|
      d.default = false
      d.save
    end
  end

  def save_collaborators
    # Remove all relations, except those with the owner and the current user
    dashboard_users.destroy_all
    return if collaborators.nil?

    # Dont save relations with the owner or the current user
    collaborators.each do |user|
      dashboard_users.build(
        user_id: user['user_id'],
        can_edit: (user['can_edit'] == 'on')
      )
    end
  end

  def save_sensors_collaborators
    # Remove all relations
    dashboard_sensors.destroy_all
    return if sensors_collaborators.nil?

    # Save all relations
    sensors_collaborators.each do |_, sensor|
      dashboard_sensors.build(
        sensor_id: sensor['sensor_id'],
        can_edit: sensor['can_edit']
      )
    end
  end

  #
  # => Change the position of the widgets if it's needed
  #
  def check_layout
    widgets.each do |widget|
      Widget::POSITIONS.each do |position, defaults|
        if layout.to_sym == :columns
          if widget.send(position).to_s[-2..-1] == 'px'
            measure = widget.send(position)[0..-3].to_i

            # Get the value for width-height widget value.
            value = measure / defaults[:standard]
            # Get the x-y column position, one after this measure.
            # This measure indicates the margin right-top position, so we add 1.
            value += 1 if %w(position_x position_y).include?(position.to_s)

            widget.send(position.to_s + '=', value)
            widget.save
          end
        else
          if widget.send(position).to_s[-2..-1] != 'px'
            measure = widget.send(position)

            if %w(position_x position_y).include?(position.to_s)
              margin = defaults[:absolute][0..-3].to_i * measure
              # Get the margin right-top position.
              # The limit is the x-y positions, so we subtract 1 to the x-y value.
              value = (margin + (measure - 1) * defaults[:standard])
            else
              # Get the width-height widget value.
              value = measure * defaults[:standard]
            end

            widget.send(position.to_s + '=', value.to_s + 'px')
            widget.save
          end
        end
      end
    end
  end

  def defaults
    self.locked               ||= false
    self.layout               ||= 'columns'
    self.background_color     ||= '#EDEDED'
  end
end
