DittoCode::HoldFile.if 'ENTERPRISE'
#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class DruidManager < Manager
  # ==========================================================
  # Custom attributes
  # ==========================================================

  DEAFULT_TIER = 'default'.freeze

  attr_reader :name

  # ==========================================================
  # Public methods
  # ==========================================================

  def initialize(node_name)
    @errors = []
    @name = node_name

    if node_name.nil?
      @node = nil
      @errors << "#{I18n.t('models.node_name')}"
      return
    end

    begin
      @node = Chef::Node.load node_name
      @role = Chef::Role.load node_name
      @role.override_attributes['redborder'] ||= {}
      @role.override_attributes['redborder']['druid'] ||= {}
      @role.override_attributes['redborder']['druid']['historical'] ||= {}
    rescue => e
      @errors << e.inspect
      Rails.logger.error e.inspect
      Rails.logger.error e.backtrace.join('\n')
      @node = nil
      @role = nil
    end

    return unless @node['redborder']['druid'].nil?

    @node = nil
    @errors << "#{I18n.t('models.this_node')}"
  end

  # Return a list of managers that have any druid service enabled.
  #
  def self.all
    by_type(%w(coordinator realtime middlemanager overlord
               historical broker))
  end

  # Aliases based on type:
  #
  %w(coordinator realtime middlemanager overlord
     historical broker).each do |druid_type|
    self.class.send(:define_method, druid_type) do
      by_type(druid_type)
    end
  end

  # Return Chef nodes based on type given
  #
  def self.by_type(types)
    types = Array(types)
    # Get all managers
    list = Manager.all
    ret = []
    # List managers
    list.each do |manager|
      services = Hash[*manager.services.flat_map { |k, v| [k, v[:status]] }]
      overwrites = manager.overwrites
      druid_node = false

      types.each do |service|
        next unless services["druid-#{service}"] ||
                    overwrites["druid-#{service}"]
        druid_node = true
      end
      # Check if this is a druid node
      ret << DruidManager.new(manager.name) if druid_node
    end
    # Return the hash
    ret
  end

  #
  # Return a list of available tiers
  #
  def self.available_tiers
    Tier.all.map(&:name) + [DEAFULT_TIER]
  end

  def available_tiers
    self.class.available_tiers
  end

  #
  # Return the Tier of the historical service. We check Role first because
  # when save the tier, the change doesn't appear in the node until chef update
  # values
  #
  # == Returns:
  # String with the tier
  #
  def tier
    if @role.override_attributes['redborder']['druid']['historical'].empty?
      @node['redborder']['druid'].fetch('historical', {}).fetch('tier', nil)
    else
      @role.override_attributes['redborder']['druid']['historical']['tier']
    end
  end

  #
  # Assign the value to tier on the historical service.
  #
  # == Parameters:
  # tier_to_save::
  #   String with the nmae of the tier
  # memory:
  #   Boolean indicates if this tier store data in memory
  #
  def update_tier(tier_to_save, memory)
    return unless valid_tier?(tier_to_save) &&
                  @node['redborder']['druid'].present? &&
                  @node['redborder']['druid']['historical'].is_a?(Hash)
    # Check memory value
    memory = clean_memory(memory)
    # Store values
    @node.normal['redborder']['druid']['historical']['tier'] = tier_to_save
    @node.normal['redborder']['druid']['historical']['memory_only'] = memory
    @role.override_attributes['redborder']['druid']['historical']['tier'] =
      tier_to_save
    @role.override_attributes['redborder']['druid']['historical']['memory_only'] =
      memory
    self.pending_changes = 1
  end

  private

  #
  # Check if the new Tier is a valid Tier
  #
  # == Parameters:
  # new_tier::
  #   String with the new tier
  #
  # == Returns:
  # True or false
  #
  def valid_tier?(new_tier)
    return true if (available_tiers).include? new_tier
    @errors << "Tier #{new_tier} #{I18n.t('models.is_not')}"
    false
  end

  #
  # Check if the memory value is valid. If it's not, return 0 value
  #
  # == Paramters:
  # memory_only::
  #   String with the value for memory
  #
  # == Returns:
  # '1' or '0'
  #
  def clean_memory(memory_only)
    return memory_only if memory_only.is_a?(TrueClass) ||
                          memory_only.is_a?(FalseClass)
    false
  end
end
