#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Get data from Druid Segments and format it to represent
#
class DruidSegments
  include Redborder::Namespaces

  # Time format of Druid
  TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%LZ'
  TIME_MONTH_SEARCH = '%Y-%m'

  #
  # Set the time
  #
  def initialize(time, sensor)
    self.month = init_time(time)
    @sensor = sensor
    @namespaces = sensor.namespaces.map(&:uuid)
    @namespaces << 'default' if @namespaces.include?(Sensor.root.uuid)
  end

  #
  # Load data of a month and group the results based on a namespace name
  #
  # == Returns:
  # Hash with the name of namespaces as keys and size of them as value
  #
  def by_namespaces_names
    return @cached_namespaces_names if @cached_namespaces_names
    namespaces = Hash.new(0)
    month_data_struct.each do |segment|
      uuid = namespace_name(segment.data_source)
      name = (uuid=='default' ? 'None' : Sensor.find_by_uuid(uuid).name rescue uuid)
      namespaces[name] += segment.size
    end
    @cached_namespaces_names = namespaces
  end

  #
  # Fetch the data from the month.
  #
  # == Returns:
  # Array of elements
  #
  def month_data
    return @cached_month_data if @cached_month_data
    # Cache data
    @cached_month_data =
      Segment
      .where('start LIKE ?', "#{format_time(month, true)}%")
      .select { |seg| @namespaces.include? seg.namespace }
  end

  #
  # Return Payload of the data.
  #
  # == Returns:
  # Array of Structs with the payload defined in Segment#SegmentStruct
  #
  def month_data_struct
    return @cached_month_data_struct if @cached_month_data_struct
    @cached_month_data_struct = month_data.map(&:payload_struct)
  end

  #
  # Get the selected month, or set it to now
  #
  def month
    @month || Time.now.utc.beginning_of_month
  end

  #
  # Change the month of the data. If the month is different than last
  # selected, this method will delete instance variable
  #
  def month=(time)
    tmp = init_time(time)
    # Delete cached data
    clear_cache unless @month == tmp
    @month = tmp
  end

  # ----------------------------------------------------
  # Format and filter output
  # ----------------------------------------------------

  #
  # Return current total for all namespaces and all dataSources
  #
  # == Returns:
  # Integer with the total
  #
  def total
    month_data.map { |el| el.payload_struct.size }.sum
  end

  #
  # Get the data prepared to show in a timeserie
  #
  # == Returns:
  # Array of elements formatted to time series
  #
  def timeserie
    data = size_by_time
    # Return timeseries
    timeserie = []
    data.each do |k, v|
      timeserie << { event: { events: v },
                     timestamp: k.to_time.to_s }
    end
    # Return formatted data
    timeserie.reverse
  end

  #
  # Get the data prepared to show in a timeserie
  #
  # == Parameters:
  # time::
  #   Time in the range of selected month
  #
  # == Returns:
  # Array of elements formatted to time series
  #
  def timeserie_accumulated
    data = size_by_time(true)
    # Return timeseries
    timeserie = []
    acc = 0
    (month.to_i..(month + 1.month).to_i).step(1.hour) do |timestamp|
      acc += data[timestamp]
      timeserie << { event: { events: acc },
                     timestamp: Time.at(timestamp).to_time.to_s }
    end
    # Return formatted data
    timeserie.reverse
  end

  #
  # Get the data prepared to show in a timeserie (per day)
  #
  # == Returns:
  # Array of elements formatted to time series
  #
  def timeserie_per_day
    data = size_per_day
    # Return timeseries
    timeserie = []
    (month.to_i..(month + 1.month).to_i).step(1.day) do |timestamp|
      timeserie << { event: { events: data[timestamp] },
                     timestamp: Time.at(timestamp).to_time.utc.to_s }
    end
    # Return formatted data
    timeserie.reverse
  end

  #
  # Load data of a month and group the results based on a namespace
  #
  # == Returns:
  # Hash with the name of namespaces as keys and size of them as value
  #
  def by_namespaces
    return @cached_namespaces if @cached_namespaces
    namespaces = Hash.new(0)
    month_data_struct.each do |segment|
      name = namespace_name(segment.data_source)
      namespaces[name] += segment.size
    end
    @cached_namespaces = namespaces
  end

  #
  # Load the data of given namespace consumed in a month
  #
  # == Returns:
  # Integer with the number of bytes
  #
  def by_namespace(namespace)
    by_namespaces[namespace]
  end

  #
  # Load data of a month and group the results based on a namespace
  #
  # == Returns:
  # Hash with the name of namespaces as keys and size of them as value
  #
  def by_topics
    return @cached_topics if @cached_topics
    topics = Hash.new(0)
    month_data_struct.each do |segment|
      name = topic_name(segment.data_source)
      topics[name] += segment.size
    end
    @cached_topics = topics
  end

  #
  # Load the data of given topic consumed in a month
  #
  # == Returns:
  # Integer with the number of bytes
  #
  def by_topic(topic)
    by_topics[topic]
  end

  private

  #
  # Get the name of the topic based on the data source
  #
  # == Parameters:
  # data_source::
  #   String with the name of the data source
  #
  # == Returns:
  # Name of the topic
  #
  def topic_name(name)
    match = /rb_([a-zA-Z]+)[.]*/.match(name)
    return 'N/A' unless match
    name = match[1]
    if name == 'ips'
      name.upcase
    elsif name == 'loc'
      'Location'
    else
      name.capitalize
    end
  end

  #
  # Clear all cached values
  #
  def clear_cache
    @cached_month_data, @cached_namespaces,
      @cached_topics, @cached_month_data_struct = [nil] * 4
  end

  #
  # Initialize Time to use in queries
  #
  def init_time(time)
    (time || Time.now).utc.beginning_of_month
  end

  #
  # Load month data and created a Hash with end time as key and size
  # of segments as values
  #
  # == Parameters:
  # time::
  #   Time in the range of selected month
  #
  # == Returns:
  # Hash with format
  #
  def size_by_time(numeric_keys = false)
    size_by_time = Hash.new(0)
    month_data.reverse.each do |data|
      key = numeric_keys ? data.end.to_time.to_i : data.end
      size_by_time[key] += data.payload_struct.size
    end
    # Return it!
    size_by_time
  end

  #
  # Load month data and created a Hash with end time per day as key,
  # and size of segments as values
  #
  # == Returns:
  # Hash with format
  #
  def size_per_day
    size_per_day = Hash.new(0)
    month_data.reverse.each do |data|
      key = data.end.to_time.utc.beginning_of_day.to_i
      size_per_day[key] += data.payload_struct.size
    end
    # Return it!
    size_per_day
  end

  #
  # Format the time to correct string for Druid
  #
  # == Parameters:
  # time::
  #   Time object to format
  # search::
  #   Boolean indicating if we must apply search format
  #
  # == Returns:
  # String with the time
  #
  def format_time(time, search = false)
    return time.utc.strftime(TIME_MONTH_SEARCH) if search
    # Always convert it to UTC
    time.utc.strftime(TIME_FORMAT)
  end
end
