#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'time'

#
# => Store conditions to trigger incidents.
#
class Incident < ApplicationRecord
  include Redborder::Audit

  # ==========================================================
  # Constants
  # ==========================================================
  MULTIPLIER_RANGE = [0, 3, 6, 9]
  INTERVAL_RANGE = [300, 900, 1800, 3600, 21_600, 43_200, 86_400]
  TIME_WINDOW_RANGE = [ 300, 600, 900, 1800, 3600, 7200, 28800, 86400, 172800, 259200, 345600, 432000, 518400, 604800]

  # ==========================================================
  # Custom attributes
  # ==========================================================

  # ==========================================================
  # Relations
  # ==========================================================

  has_many :incident_users, dependent: :destroy  
  has_many :users, through: :incident_users
  has_many :notes, dependent: :destroy
  has_many :outliers, dependent: :destroy

  # Link / unlink incidents
  has_many :incident_links, dependent: :destroy
  has_many :linked_incidents, through: :incident_links, source: :linked_incident

  # Reverse relationship
  has_many :inverse_incident_links, class_name: 'IncidentLink', foreign_key: 'linked_incident_id', dependent: :destroy
  has_many :inverse_linked_incidents, through: :inverse_incident_links, source: :incident

  belongs_to :priority
  belongs_to :incident_status

  belongs_to :sensor

  has_one :playbook, dependent: :destroy
  has_one :incident_reports, dependent: :destroy

  has_one :alarm

  # ================================================================
  # Callbacks
  # ================================================================

  before_destroy :remove_alarm_association

  after_create :assign_playbook
  after_create :create_incident_report
  after_update :send_to_syslog_incident_changes

  # ================================================================
  # Validations
  # ================================================================

  validates :uuid, presence: true, uniqueness: true

  validates :name, presence: true,  length: { in: 1..255 }

  validates :priority, presence: true

  validates :incident_status, presence: true

  audited on: [:create, :update, :destroy]

  # ================================================================
  # Public methods
  # ================================================================

  def closed?
    incident_status.name == IncidentStatus::CLOSED
  end

  def closed!
    assign_status_by_name!(IncidentStatus::CLOSED)
  end

  def containment_achieved?
    incident_status.name == IncidentStatus::CONTAINMENT_ACHIEVED
  end

  def containment_achieved!
    assign_status_by_name!(IncidentStatus::CONTAINMENT_ACHIEVED)
  end

  def incident_reported?
    incident_status.name == IncidentStatus::INCIDENT_REPORTED
  end

  def incident_reported!
    assign_status_by_name!(IncidentStatus::INCIDENT_REPORTED)
  end

  def new?
    incident_status.name == IncidentStatus::NEW
  end

  def new!
    assign_status_by_name!(IncidentStatus::NEW)
  end

  def open?
    incident_status.name == IncidentStatus::OPEN
  end

  def open!
    assign_status_by_name!(IncidentStatus::OPEN)
  end

  def rejected?
    incident_status.name == IncidentStatus::REJECTED
  end

  def rejected!
    assign_status_by_name!(IncidentStatus::REJECTED)
  end

  def restoration_achieved?
    incident_status.name == IncidentStatus::RESTORATION_ACHIEVED
  end

  def restoration_achieved!
    assign_status_by_name!(IncidentStatus::RESTORATION_ACHIEVED)
  end

  def stalled?
    incident_status.name == IncidentStatus::STALLED
  end

  def stalled!
    assign_status_by_name!(IncidentStatus::STALLED)
  end

  # Combined view of all linked incidents
  def list_linked_incidents
    (linked_incidents + inverse_linked_incidents).uniq
  end

  # Get incident data
  def get_data(options = {}, user = User.current_user)
    user = User.first if user.nil?
    data = Incidents::IncidentDruidRetriever.get_data(self, options, user)
    data += Incidents::IncidentOutlierRetriever.get_data(self, user)
    data
  end

  def get_incident_mitre_data
    elements = get_data({})
    tactics = []
    techniques = []
    subtechniques = []

    grouped_elements = elements.group_by { |event| [event[:msg], event[:incident_source_name], event[:sig_id]] }
    indicators = grouped_elements.map do |(msg, incident_source_name, sig_id), entries|
      {
        incident_source_name: incident_source_name,
        msg: msg,
        count: entries.size,
        sig_id: sig_id
      }
    end

    indicators = indicators.sort_by { |entry| -entry[:count] }
    indicators.each do |indicator|
      next if indicator[:sig_id] == 'N/A'

      sid = indicator[:sig_id]
      tactics_ = Mitre.where('rule_ids && ARRAY[?]::integer[] AND is_tactic = ?', [sid], true)
      tactics.concat(tactics_).uniq! { |obj| obj.is_a?(Hash) ? obj[:tactic_id] : obj.tactic_id }
      tactics = tactics.map { |obj| obj.attributes.except('rule_ids', 'description') } unless tactics.nil?

      techniques_ = Mitre.where('rule_ids && ARRAY[?]::integer[] AND is_technique = ?', [sid], true)
      techniques.concat(techniques_).uniq! { |obj| obj.is_a?(Hash) ? obj[:tactic_id] : obj.tactic_id }
      techniques = techniques.map { |obj| obj.attributes.except('rule_ids') } unless techniques.nil?

      subtechniques_ = Mitre.where('rule_ids && ARRAY[?]::integer[] AND is_subtechnique = ?', [sid], true)
      subtechniques.concat(subtechniques_).uniq! { |obj| obj.is_a?(Hash) ? obj[:tactic_id] : obj.tactic_id }
      subtechniques = subtechniques.map { |obj| obj.attributes.except('rule_ids') } unless subtechniques.nil?
    end

    { tactics: tactics, techniques: techniques, subtechniques: subtechniques }
  end


  # ================================================================
  # Public cache methods
  # ================================================================
  def self.cache_key_for_user(user_id)
    "rb-webui:incidents:#{user_id}"
  end
  
  # Save the current filters to Memcached
  def self.create_filters(user_id, search_params, name_param)
    filters = { search: search_params, name: name_param }
    Rails.cache.write(cache_key_for_user(user_id), filters, expires_in: 3.hours)
  end
  
  # Read the filters from Memcached
  def self.read_filters(user_id)
    Rails.cache.read(cache_key_for_user(user_id))
  end
  
  # Clear the filters from Memcached
  def self.clear_filters(user_id)
    Rails.cache.delete(cache_key_for_user(user_id))
  end

  def basic_info_structured_data
    sd_id = "incident-basic-info@#{RedBorder::PEN}"
    fields = {
      "incident-id" => self.id,
      "incident-uuid" => self.uuid,
      "incident-uri" => self.uri,
      "incident-name" => remove_special_chars(self.name.gsub('"', '\"')),
      "incident-domain" => remove_special_chars(self.domain),
      "incident-priority" => self.priority.name
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end

  def details_structured_data(assets = [], observables = [])
    sd_id = "incident-details@#{RedBorder::PEN}"
    fields ={
      "incident-id" => self.id,
      "incident-uuid" => self.uuid,
      "incident-uri" => self.uri,
      "incident-domain" => remove_special_chars(self.domain),
      "incident-name" => remove_special_chars(self.name.gsub('"', '\"')),
      "incident-priority" => self.priority.name,
      "incident-source" => self.source,
      "incident-status" => self.incident_status.name,
      "incident-description" => self.description.nil? ? "" : remove_special_chars(self.description.truncate(200, omission: '...').gsub('"', '\"')),
      "incident-assets" => assets,
      "incident-observables" => observables
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end

  def send_pending_syslog_messages
    pending = pending_syslog_messages
    pending.each do |syslog_message|
      SyslogService.send_to_syslog(syslog_message.to_s)
      self.sent_timestamp = Time.now
      self.save!
    end

    pending.count
  end

  def pending_syslog_messages
    syslog_messages = []
    incident_data = self.get_data({})

    if self.sent_timestamp
      incident_data = incident_data.select do |data|
        data[:timestamp] > self.sent_timestamp
      end
    end

    return [] if incident_data.length == 0

    sources = incident_data.map { |data| data[:incident_source_name] }.uniq.join(', ')

    # get assets && observables from data
    assets = []
    observables = []
    incident_data.each do |data|
      assets += data[:assets] if data[:assets]
      observables += data[:observables] if data[:observables]
    end      
    assets.uniq!
    observables.uniq!

    # Incident Creation message
    unless self.sent_timestamp
      syslog_message = Rfc5424SyslogMessage.new
      syslog_message.msg_id = 'incident-create'
      syslog_message.structured_data.push(self.details_structured_data(hash_to_escaped_json(assets), hash_to_escaped_json(observables, true)))
      syslog_message.msg = "New incident: '#{self.name}\' (Priority: #{self.priority.name}) (Domain: #{self.domain})."
      
      syslog_messages.push(syslog_message)
    end

    # Incident Detection Event messages
    incident_data.each do |event|
      syslog_message = Rfc5424SyslogMessage.new
      if Redborder::Module.registered_modules.include?(:ips) && event[:incident_source_name] == Redborder::Module.registered_modules[:ips][:incident_source_name]
        syslog_message.msg_id = 'incident-intrusion-event'
        syslog_message.structured_data.push(self.intrusion_event_structured_data(event))
      end

      if Redborder::Module.registered_modules.include?(:flow) && event[:incident_source_name] == Redborder::Module.registered_modules[:flow][:incident_source_name]
        syslog_message.msg_id = 'incident-network-anomaly-event'
        syslog_message.structured_data.push(self.network_anomaly_event_structured_data(event))
      end

      if Redborder::Module.registered_modules.include?(:vault) && event[:incident_source_name] == Redborder::Module.registered_modules[:vault][:incident_source_name]
        syslog_message.msg_id = 'incident-vault-event'
        syslog_message.structured_data.push(self.vault_event_structured_data(event))
      end

      next unless syslog_message.msg_id != '-'

      syslog_message.msg = "New #{syslog_message.msg_id.gsub('incident', '').gsub('-', ' ')}: '#{event[:msg]}' (Priority: #{event[:priority]}). Associated incident: '#{self.uuid}' (Domain: #{self.domain})."
      syslog_messages.push(syslog_message)
    end

    syslog_messages
  end

  def uri
    "https://#{Setting.webpage_url.sub('http://','').sub('https://','')}/incidents/#{self.id}"
  end

  def domain
    self.sensor.name
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  def assign_status_by_name!(status_name)
    status = IncidentStatus.find_by(name: status_name)
    if status
      self.incident_status = status
      if save
        true
      else
        errors.full_messages.join(", ")
      end
    else
      false
    end
  end

  def assign_playbook
    assignment_rules = PlaybookAssignmentRule.where(enabled: true).order(:priority)
    playbook_assigned = false

    assignment_rules.each do |rule|
      next if playbook_assigned

      playbook_to_assign = PlaybookTemplate.find_by(id: rule.playbook_templates_id)

      if rule_conditions_met?(rule)
        playbook = create_playbook_from_template(playbook_to_assign)
        playbook_assigned = true
      end
    end
    assign_default_playbook unless playbook_assigned
  end

  def assign_default_playbook
    default_template = PlaybookTemplate.find_by(default: true)
    if default_template
      playbook = create_playbook_from_template(default_template)
      self.playbook = playbook
    end
  end

  def create_playbook_from_template(template)
    playbook = Playbook.create(name: template.name, description: template.description, incident: self)

    template.phase_templates.each do |phase_template|
      phase = playbook.phases.create(name: phase_template.name)

      phase_template.task_templates.each do |task_template|
        phase.tasks.create(name: task_template.name, summary: task_template.summary, description: task_template.description, execution: task_template.execution, integration_service: task_template.integration_service, action: task_template.action, position: task_template.position)
      end
    end

    playbook
  end

  def create_incident_report
    IncidentReports.create(incident: self)
  end

  def rule_conditions_met?(assignment_rule)
    return true if assignment_rule.playbook_assignment_rules_conditions.empty?

    execute_assignment_rule_conditions(assignment_rule)
  end

  def execute_assignment_rule_conditions(assignment_rule)
    if assignment_rule.condition_type == 'AND'
      assignment_rule.playbook_assignment_rules_conditions.all? { |condition| execute_condition(condition) }
    else
      assignment_rule.playbook_assignment_rules_conditions.any? { |condition| execute_condition(condition) }
    end
  end

  def execute_condition(condition)
    sub_property_value = if condition.property == 'Incidents'
                           send(condition.sub_property).to_s
                         elsif condition.property == 'Mitre'
                           mitre_data = get_incident_mitre_data
                           if condition.sub_property == 'name' || condition.sub_property.include?('tactic')
                             mitre_data[:tactics]
                           elsif condition.sub_property.include?('subtechnique')
                             mitre_data[:subtechniques]
                           else
                             mitre_data[:techniques]
                           end
                         end
    if condition.sub_property == 'priority_id'
      sub_property_value = Priority.where(id: sub_property_value).pluck(:score)[0]
    end
    condition.execute(sub_property_value)
  end

  def msg_changes(changes)
    message_changes = ''

    if changes.any?
      changes.each do |attribute, (old_value, new_value)|
        next if attribute == "sent_timestamp" || attribute == "visible" || attribute == "updated_at" || attribute == "failure_count"

        value = send(attribute)
        next unless value

        if attribute == "incident_status_id"
          status = IncidentStatus.find_by(id: value)
          next unless status
          message_changes += "status=\"#{status.name}\", "
        elsif attribute == "description"
          truncated_description = value.nil? ? '' : value.truncate(200, omission: '...') 
          message_changes += "description=\"#{truncated_description.gsub('"', '\"')}\", "
        else
          message_changes += "#{attribute}=\"#{value.gsub('"', '\"')}\", "
        end
      end
    end

    return nil if message_changes.empty?    
   
    "Incident Attributes Updated: #{message_changes.sub(/, $/, '')}"
  end

  def send_to_syslog_incident_changes
    return unless visible

    return if self.changes.empty? 

    return unless SyslogService.syslog_integration_configured?

    msg = msg_changes(changed_attributes)
    return unless msg

    syslog_message = Rfc5424SyslogMessage.new
    syslog_message.msg_id = 'incident-update'
    syslog_message.structured_data.push(self.basic_info_structured_data)
    syslog_message.msg = msg

    SyslogService.send_to_syslog(syslog_message.to_s)
  end

  def intrusion_event_structured_data(event)
    sd_id = "intrusion-event@#{RedBorder::PEN}"

    timestamp = event[:timestamp].utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ') if event[:timestamp].is_a?(ActiveSupport::TimeWithZone)
    timestamp ||= Time.parse(event[:timestamp]).utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ')

    fields ={
      "timestamp" => timestamp,
      "incident-uuid" => self.uuid,
      "incident-uri" => self.uri,
      "incident-domain" => remove_special_chars(self.domain),
      "sensor-name" => remove_special_chars(event[:sensor_name]),
      "sensor-uuid" => event[:sensor_uuid],
      "priority" => event[:priority],
      "message" => remove_special_chars(event[:msg]),
      "assets" => hash_to_escaped_json(event[:assets]),
      "observables" => hash_to_escaped_json(event[:observables], true),
      "classification" => remove_special_chars(event[:classification]),
      "src" => event[:src],
      "dst" => event[:dst],
      "src-port" => event[:src_port],
      "dst-port" => event[:dst_port],
      "src-mac-address" => event[:ethsrc],
      "dst-mac-address" => event[:ethdst]
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end

  def vault_event_structured_data(event)
    sd_id = "vault-event@#{RedBorder::PEN}"

    timestamp = event[:timestamp].utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ') if event[:timestamp].is_a?(ActiveSupport::TimeWithZone)
    timestamp ||= Time.parse(event[:timestamp]).utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ')

    fields ={
      "timestamp" => timestamp,
      "incident-uuid" => self.uuid,
      "incident-uri" => self.uri,
      "incident-domain" => remove_special_chars(self.domain),
      "sensor-name" => remove_special_chars(event[:sensor_name]),
      "sensor-uuid" => event[:sensor_uuid],
      "priority" => event[:priority],
      "message" => remove_special_chars(event[:msg]),
      "assets" => hash_to_escaped_json(event[:assets]),
      "observables" => hash_to_escaped_json(event[:observables], true)
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end
 
  def network_anomaly_event_structured_data(event)
    sd_id = "network-anomaly-event@#{RedBorder::PEN}"

    timestamp = event[:timestamp].utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ') if event[:timestamp].is_a?(ActiveSupport::TimeWithZone)
    timestamp ||= Time.parse(event[:timestamp]).utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ')

    fields ={
      "timestamp" => timestamp,
      "incident-uuid" => self.uuid,
      "incident-uri" => self.uri,
      "incident-domain" => remove_special_chars(self.domain),
      "sensor-name" => remove_special_chars(event[:sensor_name]),
      "sensor-uuid" => event[:sensor_uuid],
      "priority" => event[:priority],
      "message" => remove_special_chars(event[:msg]),
      "assets" => hash_to_escaped_json(event[:assets]),
      "observables" => hash_to_escaped_json(event[:observables], true)
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end

  def remove_special_chars(text)
    return "" if text.nil?

    text.to_s.gsub('[','').gsub(']','')
  end

  def hash_to_escaped_json(event_hash, invert_key_value = false)
    return "" if event_hash.empty?
  
    json = "["
  
    event_hash.each do |element|
      element_json = "{"
  
      element_json += element.map do |key, value|
        if invert_key_value
          "\"#{value}\":\"#{key}\""
        else
          "\"#{key}\":\"#{value}\""
        end
      end.join(", ")
  
      element_json += "}"
      json += element_json + ", "
    end
  
    json = json.strip.chomp(',') + "]"
  
    json.gsub('"', '\"')
  end

  def remove_alarm_association
    alarm.update(incident_id: nil) if alarm.present?
  end
end
