DittoCode::HoldFile.if 'ENTERPRISE'
#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# This class manage the licenses needed for controling
# the data ingested into the cluster.
class License < ApplicationRecord
  class InvalidFormat < StandardError; end
  class InvalidOrganization < StandardError; end

  cattr_accessor :available_sensor_types

  self.available_sensor_types = {}

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------
  has_many :licensed_sensors, class_name: 'Sensor'
  belongs_to :sensor

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------
  validate :check_cluster_uuid
  validate :check_signature
  validates_uniqueness_of :uuid

  # -------------------------------------------------------------
  # Callbacks
  # -------------------------------------------------------------
  after_save :store_in_chef
  before_destroy :disassociate_sensors, :destroy_in_chef

  scope :valid, -> { where('expire_at > ?', Time.now) }

  def encoded_license=(encoded_license)
    return unless encoded_license
    license = JSON.parse(Base64.urlsafe_decode64(encoded_license))
    self.encoded_info = license['encoded_info']
    self.signature = license['signature']
  rescue InvalidOrganization # force to raise InvalidOrganization and not be catched and raised InvalidFormat
    raise InvalidOrganization
  rescue
    raise InvalidFormat
  end

  # Decode info
  def encoded_info=(encoded_info)
    # self.encoded_info = encoded_info
    self.info = JSON.parse(Base64.urlsafe_decode64(encoded_info))
    super
  end

  def info
    {
      uuid: uuid,
      expire_at: expire_at.to_i,
      limit_bytes: limit_bytes,
      cluster_uuid: cluster_uuid,
      sensors: sensors
    }
  end

  # Info intialize another fields.
  def info=(info)
    self.uuid = info['uuid']
    self.organization_uuid = info['organization_uuid']
    self.expire_at = Time.at(info['expire_at'].to_i)
    self.limit_bytes = info['limit_bytes']
    self.cluster_uuid = info['cluster_uuid']
    self.sensors = info['sensors']
  end

  def organization_uuid=(organization_uuid)
    if organization_uuid
      organization = Sensor.find_by(type: Sensor::DOMAIN, domain_type: Sensor::ORGANIZATION,
                                    uuid: organization_uuid)
      raise InvalidOrganization unless organization
      self.sensor = organization
    else
      self.sensor = Sensor.root
    end
  end

  # Represent the limit in MB
  def limit_megabytes
    limit_bytes / (1024**2)
  end

  # Get an existing license if uuid already exists in the system or
  # initialize a new one
  def self.new_or_update(attributes)
    new_license = License.new(attributes)

    if license = License.find_by(uuid: new_license.uuid, cluster_uuid: new_license.cluster_uuid)
      license.assign_attributes new_license.attributes.slice('encoded_info', 'signature', 'limit_bytes', 'sensors')
      license
    else
      new_license
    end
  end

  # Create and store the license in the license data bag.
  def self.create(encoded_license)
    license = JSON.parse(Base64.urlsafe_decode64(encoded_license))
    License.new(encoded_info: license['encoded_info'], signature: license['signature']).save
  end

  # Return if a license is expired (its expire_at
  # is greater than actual timestamp) or not.
  def status
    if expire_at <= Time.now
      :expired
    elsif expire_at < Time.now + 1.month
      :warning
    else
      :ok
    end
  end

  def expired?
    status == :expired
  end

  # Return a hash representing the resume of all licenses that belongs to
  # the domain passed.
  def self.resume(sensor = Sensor.root)
    resume = { limit_bytes: 0, active_licenses: 0, expire_at: Time.at(0),
               sensors: {}, cluster_uuid: Setting.cluster_uuid }

    sensor.licenses.each do |license|
      if license.status != :expired
        resume[:limit_bytes] += license.limit_bytes
        resume[:active_licenses] += 1
        available_sensor_types.each do |k, v|
          resume[:sensors][k] ||= { name: v[:name], limit: 0, in_use: 0 }
          resume[:sensors][k][:limit] += license.send(v[:limit])
          resume[:sensors][k][:in_use] += license.send(v[:in_use])
        end
      end
      resume[:expire_at] = [resume[:expire_at], license.expire_at].max
    end

    resume
  end

  def self.available_by_type(type, root = Sensor.root)
    valid.where(sensor: root).select do |license|
      license.send(available_sensor_types[type.to_sym][:limit]) - license.send(available_sensor_types[type.to_sym][:in_use]) > 0
    end
  end

  def available_slots(type)
    slots = send(available_sensor_types[type.to_sym][:limit]) - send(available_sensor_types[type.to_sym][:in_use])
    return 0 if slots < 0
    slots
  end

  private

  def store_in_chef
    data = Chef::DataBagItem.load('rBglobal', 'licenses')
    data['licenses'][uuid] = { info: info, encoded_info: encoded_info,
                               signature: signature, created_at: created_at }

    if sensor.organization?
      data['licenses'][uuid][:organization_uuid] = sensor.uuid
    end

    data.save
  end

  def destroy_in_chef
    data = Chef::DataBagItem.load('rBglobal', 'licenses')
    data['licenses'].delete(uuid)
    data.save
  end

  def disassociate_sensors
    licensed_sensors.delete_all
  end

  # Validation that checks if the cluster uuid referenced in the license
  # is equal than the current cluster uuid.
  def check_cluster_uuid
    errors.add(:cluster_uuid, "#{I18n.t('models.is_incorrect')}") if Setting.cluster_uuid != cluster_uuid
  end

  # Check the consistency of the license's signature.
  def check_signature
    key = OpenSSL::PKey::RSA.new(RedBorder::LICENSE_PUBLIC_KEY)
    digest = OpenSSL::Digest::SHA256.new
    return if key.verify digest, Base64.urlsafe_decode64(signature), encoded_info
    errors.add(:signature, "#{I18n.t('models.is_incorrect')}")
  end
end
