#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Manager

  attr_reader :node
  attr_reader :role
  attr_reader :errors

  def self.after_save(*args)
  end

  def self.after_destroy(*args)
  end

  include Redborder::Audit

  DISPLAYABLE_SERVICES = %w(zookeeper consul kafka webui rb-workers memcached mongod
                            druid-coordinator druid-realtime druid-historical
                            druid-broker opscode-erchef druid-middlemanager postgresql redborder-postgresql nginx sfacctd redborder-llm)

  DittoCode::Exec.if 'ENTERPRISE' do
    #DISPLAYABLE_SERVICES.unshift 'keepalived'
    DISPLAYABLE_SERVICES << 'minio'
  end

  MODIFICABLE_SERVICES = %w(zookeeper kafka webui memcached mongod
                            druid-coordinator druid-realtime druid-historical
                            druid-broker opscode-erchef postgresql redborder-postgresql
                            nginx minio sfacctd redborder-llm)

  DittoCode::Exec.if 'COMMUNITY' do
    MODIFICABLE_SERVICES.clear
  end

  MANDATORY_SERVICES = %w(opscode-erchef nginx webui minio postgresql redborder-postgresql)

  AVAILABLE_MODES = %w(full custom core chef kafka historical middlemanager
                       broker druid http2k webui)
  #
  # Fetch all the cluster nodes from Chef.
  # @note  This method may late several seconds
  #
  # == Returns:
  # An array with the nodes of the cluster
  #
  def self.all
    ret = []
    Chef::Node.list.keys.each do |element|
      node = Chef::Node.load element
      if node['roles'].present? && node['roles'].include?('manager')
        manager = Manager.new(element)
        ret << manager if manager.errors.blank?
      end
    end
    ret.sort { |x, y| x.name <=> y.name }
  end

  def self.modes
    manager = all.first
    # TODO: Change attribute to ['redborder']['services_group'][mode]
    # TODO: this functionality must be implemented and reconfigurated.
    if manager
      modes = Manager::AVAILABLE_MODES.flat_map do |mode|
        [mode, manager.node['redborder']['manager']['services'][mode]] rescue [mode, {}]
      end
      Hash[*modes]
    else
      {}
    end
  end

  def self.nmap_installed?
    begin 
      return Manager.all.all? do |manager|
        manager.node['redborder-scanner']['is_nmap_installed']
      end
    rescue StandardError => e
      return false
    end
  end

  def initialize(node_name)
    @errors = []

    begin
      @node = Chef::Node.load node_name
      @role = Chef::Role.load node_name
      @role.override_attributes['redborder'] ||= {}
      @role.override_attributes['redborder']['manager'] ||= {}
      @role.override_attributes['redborder']['manager']['services'] ||= {}
      @role.override_attributes['redborder']['manager']['services']['overwrite'] ||= {}
      @role.override_attributes['redborder']['manager']['rsyslog'] ||= {}
      @role.override_attributes['redborder']['manager']['rsyslog']['loggly'] ||= {}
    rescue StandardError => e
      @errors << e.inspect
      Rails.logger.error e.inspect
      Rails.logger.error e.backtrace.join("\n")
      @node = nil
      @role = nil
    end
  end

  def name
    @node.name
  end

  def mode
    @node['redborder']['mode']
  end

  def license
    @node['redborder']['license']
  end

  def pending_changes
    @node['redborder']['pending_changes'].to_i
  end

  def pending_changes=(num)
    @node.normal['redborder']['pending_changes'] = num
  end

  # TODO: Change attribute to...
  # TODO: this functionality must be implemented and reconfigurated.
  def services
    s = @node['redborder']['cluster']['services'] rescue {}
    Hash[
      s.map { |x| [x['name'], { status: x['status'], ok: x['ok'] }] }
    ]
  end

  def managers_per_services(service)
    @node['redborder']['managers_per_services']["#{service}"] rescue []
  end

  def has_service_enabled?(service)
    overwrites[service.to_sym]
  end

  # TODO: Change attribute to...
  # TODO: this functionality must be implemented and reconfigurated.
  def overwrites
    @node['redborder']['manager']['services']['overwrite'] rescue {}
  end

  def version
    @node['redborder']['rpms']['manager']
  end

  def last_seen
    @node['ohai_time']
  end

  def set_mode (mode_to_save)
    return if mode_to_save == 'master' || mode_to_save.nil?

    unless AVAILABLE_MODES.include? mode_to_save
      @errors << "#{I18n.t('models.mode')} #{mode_to_save} #{I18n.t('models.is_not')}"
      return
    end

    return unless mode != mode_to_save
    @node.normal['redborder']['manager']['mode'] = mode_to_save
    @node.normal['redborder']['manager']['services']['overwrite'] = {}
    @role.override_attributes['redborder']['manager']['mode'] = mode_to_save
    @role.override_attributes['redborder']['manager']['services']['overwrite'] = {}
    self.pending_changes = 2
  end

  def set_service (name, status)
    unless MODIFICABLE_SERVICES.include? name
      @errors << "#{I18n.t('models.cant_modify')} #{name} #{I18n.t('models.service')}"
      return
    end

    # If overwrite returns nil, create it
    overwrite = @node.normal['redborder']['manager']['services']['overwrite']
    if overwrite.nil?
      overwrite = @node.normal['redborder']['manager']['services']['overwrite'] = {}
    end

    # Dependencies
    services_to_save = [[name, status]]
    services_to_save << ['riak-cs', status]              if name == 'riak'
    services_to_save << ['rb-workers', status]           if name == 'rb-webui'
    services_to_save << ['ospcode-solr4', status]        if name == 'opscode-erchef'
    services_to_save << ['opscode-expander', status]     if name == 'opscode-erchef'
    services_to_save << ['opscode-chef-mover', status]   if name == 'opscode-erchef'
    services_to_save << ['oc_bifrost', status]           if name == 'opscode-erchef'
    services_to_save << ['redis_lb', status]             if name == 'opscode-erchef'
    services_to_save << ['rabbitmq', status]             if name == 'opscode-erchef'

    # Save them
    services_to_save.each do |services|
      service_name = services[0]
      service_status = services[1]

      if service_status != @node['redborder']['manager']['services']['overwrite'][service_name]
        @node.normal['redborder']['manager']['services']['overwrite'][service_name] = service_status
        @role.override_attributes['redborder']['manager']['services']['overwrite'][service_name] = service_status
        self.pending_changes = 2
      end
    end
  end

  # Upload a file to the remote_file path of the manager
  def upload(local_file, remote_file)
    Net::SSH.start(
      @node[:ipaddress],
      'redborder',
      auth_methods: ['publickey'], # auth method
      paranoid: false,
      timeout: 5,
      keys: ["#{Rails.root}/config/#{RedBorder::CONFIG[:rsa_filename]}"],
      config: false
    ) do |ssh|
      File.open(local_file, 'r+') do |f|
        ssh.scp.upload!(f.path, remote_file, verbose: true)
      end
    end

    register_license_changes self, local_file, remote_file
  end

  def save
    if @errors.empty?
      @node.save
      @role.save
    end
    @errors
  end

  def audit_rsa_change
    register_rsa_changes self
  end

  def audit_setting_change(message)
    register_setting_changes(self, message)
  end

  def audit_alarm_change(message)
    register_alarm_changes(self, message)
  end

  def audit_fail_login(message)
    register_fail_login_changes(self, message)
  end

end
