#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# This class represents a Mitre tactic and its associated techniques
# and subtechniques within the redBorder system.
class Mitre < ApplicationRecord
  # Validations
  # -------------------------------------------------------------
  # Ensure the presence of a name for each Mitre tactic.
  validates :name, presence: true

  # Associations
  # -------------------------------------------------------------
  # A tactic can have many techniques. Techniques are destroyed
  # when the tactic is deleted.
  has_many :techniques, foreign_key: 'mitre_tactic_id', dependent: :destroy

  # Class Methods
  # -------------------------------------------------------------
  # Import Mitre tactics and associated techniques from a JSON matrix.
  def self.import_from_json(matrix)
    ActiveRecord::Base.transaction do
      matrix.each_with_index do |entry, index|
        import_tactic(entry, index + 1)
      end
    end
  rescue ActiveRecord::RecordInvalid => e
    Rails.logger.error("Import failed: #{e.message}")
    false
  end

  private

  # Import a single Mitre tactic and its associated techniques.
  def self.import_tactic(entry, tactic_counter)
    tactic = find_or_create_by!(
      mitre_tactic_id: tactic_counter,
      tactic_id: entry["ID"],
      name: entry["Tactic"],
      description: entry["Description"],
      is_tactic: true
    )

    entry["Techniques"].each do |technique|
      import_technique(technique, tactic_counter, tactic)
    end
  end

  # Import a technique and its associated subtechniques.
  def self.import_technique(technique, tactic_counter, tactic)
    technique_record = find_or_create_by!(
      mitre_tactic_id: tactic_counter,
      name: tactic.name,
      is_technique: true,
      technique_id: technique["ID"],
      description: technique["Description"],
      technique_name: technique["Technique"],
      tactic_id: tactic.tactic_id
    )

    if technique["Subtechniques"]
      technique["Subtechniques"].each do |subtechnique|
        import_subtechnique(subtechnique, tactic_counter, technique_record)
      end
    end
  end

  # Import a subtechnique related to a specific technique.
  def self.import_subtechnique(subtechnique, tactic_counter, technique_record)
    find_or_create_by!(
      mitre_tactic_id: tactic_counter,
      name: technique_record.name,
      subtechnique_name: subtechnique["Subtechnique"],
      technique_name: technique_record.technique_name,
      is_subtechnique: true,
      technique_id: technique_record.technique_id,
      description: subtechnique["Description"],
      subtechnique_id: subtechnique["ID"],
      tactic_id: technique_record.tactic_id
    )
  end
end