#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Store the overlay maps showed in the app.
#
class OverlayMap < ApplicationRecord
  include StorageFile

  # ==========================================================
  # Relations
  # ==========================================================

  belongs_to :sensor
  has_many :gps_markers, dependent: :destroy

  before_save :check_original_image

  # ================================================================
  # Files
  # ================================================================

  # Storage the last image with two sizes: normal and small.
  storage_file :image,
               path: 'rb-webui/:class/:id_:style',
               styles: {
                 large: {
                   geometry: '999999x999999>',
                   format: :jpg
                 },
                 small: {
                   geometry: '300x200^',
                   format: :jpg
                 }
               },
               convert_options: {
                 small: '-gravity center -extent 300x200'
               }

  # Storage the original image with two sizes: normal and small.
  storage_file :original_image,
               path: 'rb-webui/:class/:id_:style.original',
               styles: {
                 large: {
                   geometry: '999999x999999>',
                   format: :jpg
                 },
                 small: {
                   geometry: '300x200^',
                   format: :jpg
                 }
               },
               convert_options: {
                 small: '-gravity center -extent 300x200'
               }

  # ================================================================
  # Validations
  # ================================================================

  validates_attachment :image, :original_image,
                       content_type: {
                         content_type: ['image/jpg', 'image/jpeg', 'image/gif',
                                        'image/png', 'image/svg+xml']
                       },
                       size: { in: 0..10.megabytes }
  validates_presence_of :name, :ne, :sw
  validates_uniqueness_of :name, scope: :sensor_id
  validates_format_of :ne, :sw,
                      with: /\A[+|-]?\d{1,3}.\d{,6},\s*[+|-]?\d{1,3}.\d{,6}\z/
  validates_attachment_presence :image

  after_initialize :defaults

  # ================================================================
  # Public methods
  # ================================================================

  #
  # => Coordinates of the map
  #
  def lat_ne
    ne.split(',')[0].to_f
  end

  def lng_ne
    ne.split(',')[1].to_f
  end

  def lat_sw
    sw.split(',')[0].to_f
  end

  def lng_sw
    sw.split(',')[1].to_f
  end

  # Methods to return URI for the images
  #
  %w(image original_image).each do |image|
    # Return the (original) image uri
    #
    # @param controller [Boolean] Return S3 or overlay_map controller URL
    # @return [String] URL of the image
    #
    define_method("#{image}_uri") do |controller: false|
      return nil unless send(image).path
      if controller
        overlay_map_controller_url(image)
      else
        StorageFile.get_uri send(image).path(:large)
      end
    end

    # Return the (original) small image uri
    #
    # @param controller [Boolean] Return S3 or overlay_map controller URL
    # @return [String] URL of the image
    #
    define_method("#{image}_uri_small") do |controller: false|
      return nil unless send(image).path
      if controller
        overlay_map_controller_url("#{image}_small")
      else
        StorageFile.get_uri send(image).path(:small)
      end
    end
  end

  private

  # Return the URL of a overlay_map controller action. The suffix of the controller
  # will be added to the given path.
  #
  def overlay_map_controller_url(path)
    ::Rails.application
           .routes
           .url_helpers
           .send("#{path}_overlay_map_path", self)
  end

  #
  # => Default values
  #
  def defaults
    rotation ||= 0
  end

  #
  # => Check if we have saved an original image
  #
  def check_original_image
    original_image = image if original_image.nil?
  end
end
