#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Redborder Objecst to define aliases.
#
class RedborderObject < ApplicationRecord
  class Unknown < StandardError; end
  # -------------------------------------------------------------
  # Constants
  # -------------------------------------------------------------
  OBJECT_TYPES = { net: :networks,
                   protocol: :protocols,
                   mac: :macs,
                   service: :services,
                   vlan: :vlans }
  OBJECT_CLASSES = %w(MacObject NetObject ProtocolObject ServiceObject
                      VlanObject)

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------
  
  validates :name,
            presence: true,
            format: { with: /\A[A-Za-z0-9_\-.:;!@#$%^&*()+=~`<>? ]+\z/, 
                      message: "#{I18n.t('models.alphanumerical')}" }
  validates :value, presence: true
  # validates_uniqueness_of :name, scope: [:type, :user_id]

  after_save :update_app_databag
  after_destroy :update_app_databag

  serialize :ip_addresses, Array
  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------

  belongs_to :user

  belongs_to :sensor

  # -------------------------------------------------------------
  # Class methods
  # -------------------------------------------------------------

  #
  # Create objects from a file. This method open the file and process all
  # elements. If exist an object with same name, it will be updated.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  # tab::
  #   String with current tab to set the correct type of the objects.
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.add_vars_from_file(content_file, tab)
    file = File.open(content_file.tempfile)
    objects = RedborderObject.where(type: tab.capitalize + "#{I18n.t('models.nice')}",
                                    user: User.current_user)
    created_count = 0
    updated_count = 0

    file.each do |line|
      columns = line.sub(/[\n]+/, '').split(';')
      name, value, comment, object_type, operating_system, ip_addresses = columns
      ip_addresses = ip_addresses.split(',').map(&:strip) if ip_addresses 

      object = objects.where(name: name, user_id: User.current_user).first

      # update
      if object.present? && object.value != value && object.comment != comment
        object.update_attributes(value: value, comment: comment, object_type: object_type, 
                                operating_system: operating_system, ip_addresses: ip_addresses)
        updated_count += 1
      # create
      elsif object.nil?
        object = RedborderObject.new(type: tab.capitalize + "#{I18n.t('models.object')}",
                                     user: User.current_user,
                                     name: name, value: value,
                                     comment: comment, object_type: object_type,
                                     operating_system: operating_system, ip_addresses:ip_addresses)
        created_count += 1
      end
      object.save
    end

    file.close
    [created_count, updated_count]
  end

  #
  # Replace objects from a file. This method open the file and process all
  # elements.
  #
  # == Parameters:
  # content_file::
  #   String with the path of the file
  # tab::
  #   String with current tab to set the correct type of the objects.
  #
  # == Returns:
  # Array with number of created elements and updated.
  #
  def self.replace_vars_from_file(content_file, tab)
    file = File.open(content_file.tempfile)
    objects = RedborderObject.where(type: tab.capitalize + "#{I18n.t('models.object')}",
                                    user: User.current_user)
    created_count = 0

    file.each do |line|
      columns = line.sub(/[\n]+/, '').split(';')
      name, value, comment, object_type, operating_system, ip_addresses = columns
      ip_addresses = ip_addresses.split(',').map(&:strip) if ip_addresses 

      object = objects.where(name: name, user_id: User.current_user).first

      # update
      if object.present?
        object.update_attributes(value: value, comment: comment, object_type: object_type,
                                 operating_system: operating_system, ip_addresses: ip_addresses)
      # create
      else
        object = RedborderObject.new(type: tab.capitalize + "#{I18n.t('models.object')}",
                                     user: User.current_user,
                                     name: name, value: value,
                                     comment: comment, object_type: object_type,
                                     operating_system: operating_system, ip_addresses: ip_addresses)
      end
      object.save
      created_count += 1
    end
    file.close

  end

  #
  # Find elements based on a query parameter in name and value fields
  #
  # == Parameters:
  # query::
  #   String to search
  # user::
  #   User needed in isolated caches
  #
  # == Returns:
  # ActiveRelation
  #
  def self.search(query, user = nil)
    return RedborderObject.none if isolated? && user.nil?
    term = "%#{query.downcase.split.join('%')}%"

    if isolated?
      where('(LOWER(name) like ? OR LOWER(value) LIKE ?) AND user_id = ?',
            term, term, user.id)
    else
      where('LOWER(name) like ? OR LOWER(value) LIKE ?', term, term)
    end
  end

  #
  #
  #
  def self.from_string(object)
    if RedborderObject::OBJECT_CLASSES.include?(object)
      object
    else
      fail Unknown
    end
  end

  #
  # Save in chef
  #
  def update_app_databag
    UpdateObjectDatabag.call(
      { name.to_s => { name: name,
                       value: value
                     } },
       self.class,
       self.class.to_s.gsub("Object","s").downcase,
      destroyed?
    )
  end


end
