#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Store the reports blocks of a report. A report block is associated to
#    to a report. It can be in a page with a position, or in not assigned box.
#
class ReportBlock < ApplicationRecord
  # ================================================================
  # Includes
  # ================================================================

  include Redborder::Rbdruid
  include StorageFile

  # ================================================================
  # Relations and filters
  # ================================================================

  belongs_to :report

  serialize :property, Hash

  before_save :remove_disabled_filters

  storage_file :image, path: 'rb-webui/:class/:id'

  # ================================================================
  # Constants
  # ================================================================

  # Type of block
  TYPE_TITLE = 1
  TYPE_TEXT = 2
  TYPE_SEPARATOR = 3
  TYPE_IMAGE = 4
  TYPE_CUSTOM = 10

  # Sizes of the blocks
  WIDTH_SMALL         = 1
  WIDTH_MEDIUM        = 2
  WIDTH_LARGE         = 3
  WIDTH_EXTRA_LARGE   = 4

  # Alignment of the titles
  ALIGNMENT_LEFT = 1
  ALIGNMENT_CENTER = 2
  ALIGNMENT_RIGHT = 3
  ALIGNMENT_JUSTIFY = 4

  # Allowed methods
  ALLOWED_METHODS = [:color, :top_num, :report_block_type, :hide_axises, :map_column,
                     :height, :url, :product_type, :category, :range, :show_name, :compare_range,
                     :aggregation, :granularity, :graph_type, :show_total, :transparent, :reference,
                     :column, :filters, :timezone, :reloadeable, :cardinality_group_by, :ball_max,
                     :position_x, :position_y, :position_width, :position_height, :sensor_table_type, :limit,
                     :keep_aspect_ratio, :square_fill, :square_border_enabled, :square_border,
                     :square_opacity, :map_type, :stream_columns, :title, :text, :text_alignment, :border,
                     :cluster_type]

  # ================================================================
  # Validation
  # ================================================================

  # Validate the width
  validates :width, inclusion: { in: [WIDTH_SMALL, WIDTH_MEDIUM, WIDTH_LARGE, WIDTH_EXTRA_LARGE] }

  # Validate the type
  validates :report_type, inclusion: { in: [TYPE_CUSTOM, TYPE_TITLE, TYPE_TEXT, TYPE_SEPARATOR, TYPE_IMAGE] }

  # Validates the number of items will be displayed
  validates :top_num, numericality: { only_integer: true }, allow_blank: true

  # Validate the timezone
  validates :timezone, inclusion: { in: ['auto'] + ActiveSupport::TimeZone.all.map(&:name) }, allow_blank: true

  # The name must be unique
  validates :name, uniqueness: { case_sensitive: false, scope: :report }

  # This must be present
  validates :name, :width, :report_type, presence: true
  validates_presence_of :product_type, if: proc { |w| w.report_type == TYPE_CUSTOM }
  validates_presence_of :image, if: proc { |w| w.report_type == TYPE_IMAGE }

  # Onlye HEX colors
  validates :color, format: { with: /\A#[a-fA-F0-9]{6}\z/,
                              message: "#{I18n.t('models.hexadecimal')}" }, allow_blank: true

  # Validate the position and the page
  validates :position, inclusion: { in: 0..99 }, allow_blank: true
  validates :page, inclusion: { in: 0..99 }, allow_blank: true

  # Another validations
  validate :product_type_inclusion, if: proc { |w| w.report_type == TYPE_CUSTOM }
  validate :required_params_others, if: proc { |w| %w(custom sensors cluster).include? w.product_type }
  validate :cardinality_horbars_has_group_by
  validate :content_type_media, if: proc { |w| 'image' == w.product_type }
  validate :min_height

  validates_attachment :image,
                       content_type: { content_type: ['image/jpg', 'image/jpeg', 'image/gif', 'image/png', 'image/svg+xml'] },
                       size: { in: 0..10.megabytes }

  # ================================================================
  # METHODS
  # ================================================================

  # Method definitions to ask for the type of the report block
  %w(title text image separator custom).each do |type|
    define_method "is_#{type}?" do
      report_type == ReportBlock.const_get("TYPE_#{type.upcase}")
    end
  end

  #
  # Parse the methods to allow the use of ALLOWED_METHODS
  #
  def method_missing(name, *args)
    method = name.to_s.chomp('=').to_sym
    if ALLOWED_METHODS.include? method
      if name.to_s.last == '='
        property[method] = args[0]
      else
        property[method]
      end
    else
      super
    end
  end

  #
  # => Called when an object of this class is imported, so it
  #    can be customized and save correctly into the DB.
  #
  def import_callback(params, media)
    image = media if is_image? && !image_file_name.nil?
    params.merge(params['property']).except('property', 'id')
  end

  #
  # => The attributes of this object that will be exported
  #
  def export_json
    to_json(
      no_methods: true,
      except: [:id, :report_id, :created_at, :updated_at,
               :image_content_type, :image_file_size, :image_updated_at]
    )
  end

  # Return the report blocks as a JSON
  #
  # @param options  ->
  # @return [Hash]    -> Return the block as a hash
  #
  def as_json(options = nil)
    hash = super
    hash['cleaned_name'] = cleaned_name unless options[:no_methods] == true
    if Redborder::Module.keys_with_interface.map(&:to_s).include? product_type
      hash['property'][:filters] = JSON.parse(filters || '{}')
      unless options[:no_methods] == true
        hash['start_time'] = start_time unless start_time.nil?
        hash['start_time_secs'] = start_time.to_i unless start_time.nil?
        hash['end_time'] = end_time unless end_time.nil?
        hash['end_time_secs'] = end_time_secs.to_i unless end_time_secs.nil?
        hash['tzoffset'] = tzoffset
        hash['step'] = step
        hash['aggregation'] = if aggregation
                                aggregation
                              elsif respond_to?("default_aggregation_#{product_type}", :include_private)
                                send("default_aggregation_#{product_type}")
                              end
      end
    end

    hash
  end

  def print_filters
    respond_to?("print_filters_#{product_type}", :include_private) ?
      send("print_filters_#{product_type}") : ""
  end

  # Get the style of the text-align based on the text_alignment property.
  # If this property doesn't exist, the function returns an empty string
  def get_style_text_alignment

    if self.text_alignment.nil?
      ""
    elsif self.text_alignment.to_i == ALIGNMENT_CENTER
      "text-align: center;"
    elsif self.text_alignment.to_i == ALIGNMENT_RIGHT
      "text-align: right;"
    elsif self.text_alignment.to_i == ALIGNMENT_JUSTIFY
      "text-align: justify;"
    else
      "text-align: left;"
    end

  end

  #
  # Return the start_time based on the time
  #
  def get_regular_start_time
    case range.to_sym
    when :last_1_minute
      end_time - 1.minute
    when :last_2_minutes
      end_time - 2.minutes
    when :last_5_minute
      end_time - 5.minute
    when :last_15_minute
      end_time - 15.minute
    when :last_30_minute
      end_time - 30.minute
    when :last_1_hour
      end_time - 1.hour
    when :last_2_hour
      end_time - 2.hour
    when :last_3_hour
      end_time - 3.hour
    when :last_12_hour
      (end_time - 12.hour).beginning_of_hour
    when :last_24
      (end_time - 24.hour).beginning_of_hour
    when :last_week
      (end_time - 1.week).beginning_of_hour
    when :last_month
      (end_time - 1.month).beginning_of_hour
    end
  end

  # Clone the report block in the selected report
  def clone(report_id, name=nil)
    new_report_block = self.dup
    new_report_block.report_id = report_id
    new_report_block.name = name unless name.nil?

    # Clean page and position
    new_report_block.page = nil
    new_report_block.position = nil

    if self.image_file_name
      DittoCode::Exec.if 'ENTERPRISE' do
        http = Net::HTTP.new(image_uri.host, image_uri.port)
        http.use_ssl = true
        # TODO: prevent use VERIFY_NONE
        http.verify_mode = OpenSSL::SSL::VERIFY_NONE
        resp = http.get(image_uri.request_uri)

        open("#{Rails.root}/tmp/temp_widget_file", "wb") do |file|
          file.write(resp.body)
        end

        new_report_block.image = File.open("#{Rails.root}/tmp/temp_widget_file")
      end

      DittoCode::Exec.if 'COMMUNITY' do
        new_report_block.image = File.open(image_uri)
      end
    end

    new_report_block.save

    # Return the new block
    new_report_block

  end

  def get_comparative_start_time
    date = end_time

    case compare_range.to_sym
      when :pt24h
        date.beginning_of_day
      when :pt168h
        date.beginning_of_day
      when :pt720h
        date.beginning_of_week
      when :pt4320h
        date.beginning_of_month
    end
  end

  def start_time
    if report_block_type == "Compare"
      res = get_comparative_start_time
    else
      res = get_regular_start_time
    end

    if granularity
      roundto = step / 1000
      Time.at((res.to_f / roundto).floor * roundto)
    else
      res
    end
  end

  def end_time
    (User.current_user.nil? or User.current_user.time_machine_now.nil?) ? Time.now : Timecop.travel(Time.at(User.current_user.time_machine_now)) { Time.now }
  end

  def end_time_secs
    unless granularity.nil?
      roundto = step / 1000
      Time.at((end_time.to_f / roundto).floor * roundto)
    else
      end_time.to_i
    end
  end

  def tzoffset
    if timezone.nil?
      0
    else
      if timezone == "auto"
        Time.zone.now.utc_offset * 1000
      else
        ActiveSupport::TimeZone[timezone].now.utc_offset * 1000
      end
    end
  end

  def step
    if report_block_type == 'Compare'
      GRANULARITY_TYPE[COMPARATIVE_COLUMNS[compare_range.to_sym][:granularity][:graph]][:step]
    elsif report_block_type == 'Cardinality'
      respond_to?("step_#{product_type}", :include_private) ?
        send("step_#{product_type}") :
        (granularity ? GRANULARITY_TYPE[granularity.to_sym][:step] : 60_000)
    elsif granularity.present?
      GRANULARITY_TYPE[granularity.to_sym][:step]
    else
      60_000
    end
  end

  def search
    send("search_#{product_type}")
  rescue NoMethodError => e
    Rails.logger.error e
    Rails.logger.error e.backtrace.join('\n')
  end

  def cleaned_name
    ret = name.downcase
    ret = ret.gsub(/\s+/, ' ')
    ret = ret.gsub(/\.|\,/, '-')
    ret = ret.gsub(/[^a-zA-Z0-9_-]/, '')
    ret
  end

  #
  # Return the data for the report block.
  #
  # == Parameters:
  # user::
  #   The user who ask for  the data. We need it to set the permissions
  # options::
  #   A Hash with the secondary options. The options are:
  #     - start_time: Time object
  #     - end_time: Time object
  #
  # == Returns:
  # Return the data for the report block.
  #
  def get_data(user, options = {})
    return nil if report_type != TYPE_CUSTOM
    user = user || User.current_user

    if product_type == 'sensors'
      # getSensors
      nil
    elsif product_type == 'flow' && report_block_type == 'Aps'
      getAps(user)
    elsif product_type == 'cluster'
      if cluster_type == 'status'
        getClusterStatus
      elsif cluster_type == 'diagram'
        getClusterDiagram
      end
    elsif product_type == 'monitor'
      getTops(user, options)
    elsif ['area', 'stacked', 'line', 'bars', 'sbars'].include? graph_type
      case report_block_type
        when 'Top' then getTops(user, options)
        when 'Raw' then getRaw(user, options)
        when 'Cardinality' then getCardinality(user, options)
        else getCompare(user, options)
      end
    elsif graph_type == 'pie'
      getPie(user, options)
    elsif ['horbars', 'verbars'].include? graph_type && report_block_type != 'Cardinality'
      getBars(user, options)
    elsif ['horbars', 'verbars'].include? graph_type && report_block_type == 'Cardinality'
      getCardinalityBars(user, options)
    elsif graph_type == 'table'
      getTable(user, options)
    elsif ['ball', 'value'].include?(graph_type) && report_block_type == 'Cardinality'
      getCardinalityBall(user)
    elsif ['ball', 'value'].include?(graph_type) && report_block_type == 'Raw'
      getRawBall(user, options)
    elsif graph_type == 'map' && map_column == 'coordinates_map'
      getCoordinatesMap(user, options)
    elsif graph_type == 'map'
      getMap(user, options)
    elsif graph_type == 'streams'
      getStreams(user, options)
    end
  end

  private

    def required_params_others
      makeValidation = Hash.new

      if product_type == "sensors"
        makeValidation[:sensor_table_type] = sensor_table_type
      elsif product_type == "custom"
        makeValidation[:url] = url
        makeValidation[:height] = height
      end

      makeValidation.each do |key, value|
        errors.add(key, "#{I18n.t('models.present2')}") if value.nil?
      end
    end

    def content_type_media
      if product_type == "image" and !["image/jpg", "image/jpeg", "image/gif", "image/png", "image/svg+xml"].include? image_content_type
        errors.add(:image, "#{I18n.t('models.only_jpg')}")
      elsif product_type == "video" and image_content_type != "video/mp4"
        errors.add(:image, "#{I18n.t('models.only_mp4')}")
      end
    end

    def remove_disabled_filters
      if self.filters && !self.filters.empty?
        temp_filters = JSON.parse(self.filters) rescue self.filters
        self.filters = (temp_filters.select do |key, value|
          value["enabled"]
        end).to_json
      else
        self.filters = "{}";
      end
    end

    def cardinality_horbars_has_group_by
      if report_block_type == "Cardinality" and graph_type == "horbars" and cardinality_group_by == "none"
        errors.add(:cardinality_group_by, "#{I18n.t('models.specify')}")
      end
    end

    def product_type_inclusion
      errors.add(:product_type, "#{I18n.t('models.not_included')}") unless (Redborder::Module.keys_with_interface.map(&:to_s) + ["custom", "sensors", "cluster", "date", "text", "square", "image", "video"]).include? self.product_type
    end

    def getRaw(user, options = {})
      # done
    end

    def getTops(user, options = {})
      # done
    end

    def getPie(user, options = {})
      #
      # In pie.js a groupby and total is called, how can implements that on top?
      # A widget with show_total true with granularity all?

      # data = {series: []}

      # # Override the report block time if it's defined in the option
      # ovrr_start_time = options[:start_time] || start_time
      # ovrr_end_time = options[:end_time] || end_time

      # data[:total] =
      #   case product_type
      #   when 'ips'
      #     Event.total(
      #       start_time: ovrr_start_time,
      #       user: user,
      #       end_time: ovrr_end_time,
      #       aggregation: aggregation,
      #       search: search
      #     )
      #   when 'flow'
      #     Traffic.total(
      #       start_time: ovrr_start_time,
      #       user: user,
      #       end_time: ovrr_end_time,
      #       aggregation: aggregation,
      #       search: search
      #     )
      #   when 'social'
      #     Social.total(
      #       start_time: ovrr_start_time,
      #       user: user,
      #       end_time: ovrr_end_time,
      #       aggregation: aggregation,
      #       search: search
      #     )
      #   end

      # case product_type
      # when 'ips'
      #   series = Event.where({
      #     query_type: :group_by,
      #     range: range,
      #     user: user,
      #     start_time: ovrr_start_time,
      #     end_time: ovrr_end_time,
      #     granularity: :all,
      #     sort: :events,
      #     aggregation: aggregation,
      #     group_by: Event::COLUMNS[column.to_sym][:columns].keys,
      #     limit: top_num,
      #     search: search
      #   })

      #   series.each do |serie|
      #     data_serie = Event.timeserie({range: range, start_time: ovrr_start_time, end_time: ovrr_end_time,
      #                                   granularity: :all,
      #                                   aggregation: aggregation,
      #                                   user: user,
      #                                   search: search.merge({column.to_sym => {serie.send(column.to_sym) => {operator: :is}}})})
      #     data[:series] << [{name: serie.send(column.to_sym), data: data_serie, secondary_name: (Event::COLUMNS[column.to_sym][:complementary_columns] and column.to_sym != :sensor_name) ?
      #                                                                                            serie.send(Event::COLUMNS[column.to_sym][:complementary_columns].keys.first.to_s) : serie.send(column)}]
      #   end
      # when 'flow'
      #   series = Traffic.where({
      #     query_type: :group_by,
      #     user: user,
      #     range: range,
      #     start_time: ovrr_start_time,
      #     end_time: ovrr_end_time,
      #     granularity: :all,
      #     sort: aggregation,
      #     limit: top_num,
      #     group_by: [Traffic::COLUMNS[column.to_sym][:columns].keys[0]],
      #     aggregation: aggregation,
      #     search: search
      #   })

      #   series.each do |serie|
      #     data_serie = Traffic.timeserie({range: range, start_time: ovrr_start_time, end_time: ovrr_end_time,
      #                                     granularity: :all,
      #                                     aggregation: aggregation,
      #                                     user: user,
      #                                     search: search.merge({column.to_sym => {serie.send(column.to_sym) => {operator: :is}}})})
      #     data[:series] << [{name: serie.send(column.to_sym), data: data_serie, secondary_name: (Traffic::COLUMNS[column.to_sym][:complementary_columns] and column.to_sym != :sensor_name) ?
      #                                                                                            serie.send(Traffic::COLUMNS[column.to_sym][:complementary_columns].keys.first.to_s) : serie.send(column)}]
      #   end
      # when 'social'
      #   series = Social.where(
      #     query_type: :group_by,
      #     user: user,
      #     range: range,
      #     start_time: ovrr_start_time,
      #     end_time: ovrr_end_time,
      #     granularity: :all,
      #     sort: aggregation,
      #     limit: top_num,
      #     group_by: [Social::COLUMNS[column.to_sym][:columns].keys[0]],
      #     aggregation: aggregation,
      #     search: search
      #   )

      #   series.each do |serie|
      #     data_serie = Social.timeserie({range: range, start_time: ovrr_start_time, end_time: ovrr_end_time,
      #                                     granularity: :all,
      #                                     aggregation: aggregation,
      #                                     user: user,
      #                                     search: search.merge({column.to_sym => {serie.send(column.to_sym) => {operator: :is}}})})
      #     data[:series] << [{name: serie.send(column.to_sym), data: data_serie, secondary_name: (Social::COLUMNS[column.to_sym][:complementary_columns] and column.to_sym != :sensor_name) ?
      #                                                                                            serie.send(Social::COLUMNS[column.to_sym][:complementary_columns].keys.first.to_s) : serie.send(column)}]
      #   end

      # end

      # data
    end

    def getBars(user, options = {})
      # Same problem than pie
    end

    def getCompare(user, options = {})
      # Override the report block time if it's defined in the option

      # We must think the method to get data

      # ovrr_start_time = options[:start_time] || start_time
      # ovrr_end_time = options[:end_time] || end_time

      # data = {series: []}

      # if compare_range == "pt24h"
      #   end_time_base = ovrr_end_time.beginning_of_hour
      #   start_time_base = end_time_base - COMPARATIVE_COLUMNS[compare_range.to_sym][:step]
      # else
      #   end_time_base = (ovrr_end_time + 1.day).beginning_of_day
      #   start_time_base = end_time_base - COMPARATIVE_COLUMNS[compare_range.to_sym][:step]
      # end

      # for i in 0..(top_num.to_i - 1)

      #   end_time_graph = end_time_base - COMPARATIVE_COLUMNS[compare_range.to_sym][:step] * i
      #   start_time_graph = start_time_base - COMPARATIVE_COLUMNS[compare_range.to_sym][:step] * i

      #   serie =
      #     case product_type
      #     when 'ips'
      #       Event.where(
      #         user: user,
      #         range: :custom,
      #         start_time: start_time_graph,
      #         end_time: end_time_graph,
      #         granularity:  COMPARATIVE_COLUMNS[compare_range.to_sym][:granularity][:graph],
      #         aggregation: aggregation,
      #         search: search
      #       )
      #     when 'flow'
      #       Traffic.where({
      #         user: user,
      #         query_type: :timeseries,
      #         range: :custom,
      #         start_time: start_time_graph,
      #         end_time: end_time_graph,
      #         granularity:  COMPARATIVE_COLUMNS[compare_range.to_sym][:granularity][:graph],
      #         aggregation: aggregation,
      #         search: search
      #       })
      #     when 'social'
      #       Social.where({
      #         user: user,
      #         query_type: :timeseries,
      #         range: :custom,
      #         start_time: start_time_graph,
      #         end_time: end_time_graph,
      #         granularity:  COMPARATIVE_COLUMNS[compare_range.to_sym][:granularity][:graph],
      #         aggregation: aggregation,
      #         search: search
      #       })
      #     end

      #   serie = serie.map{|t| {event: {events: t.send(aggregation)},
      #                         timestamp: compare_range == "pt24h" ? (t.timestamp - t.timestamp.hour.hours).change({day: Time.now.day}).to_s :
      #                                                              (t.timestamp + (i * COMPARATIVE_COLUMNS[compare_range.to_sym][:step]).to_i).to_s}}

      #   data[:series] << [{ name: start_time_graph.to_s, data: serie }]

      # end

      # data
    end

    def getClusterStatus
      Manager.all
    end

    def getClusterDiagram
      nodes = []
      nprobes = []
      s3_nodes = []
      web_nodes = []
      druid_historicals = []
      druid_brokers = []
      kafka_nodes = []

      begin
        nodes = Redborder::Node.get_kafka_info.first
        kafka_nodes = nodes.select{|k,v| [:rb_event, :rb_flow].include? k}.values.flat_map(&:values)

        managers = Manager.all
        managers.each do |manager|
          s3_nodes << manager.name if manager.services["riak"][:status]
          web_nodes << manager.name if manager.services["rb-webui"][:status]
          druid_historicals << manager.name if manager.services["druid_historical"][:status]
          druid_brokers << manager.name if manager.services["druid_broker"][:status]
          nprobes << manager.name if manager.services["nprobe"][:status]
        end

      rescue Exception => e
        nodes = []
        s3_nodes = []
        web_nodes = []
        druid_historicals = []
        druid_brokers = []
        nprobes = []
        kafka_nodes = []
        Rails.logger.error e
      end

      { nodes: nodes, nprobes: nprobes, s3_nodes: s3_nodes,
        web_nodes: web_nodes, druid_historicals: druid_historicals,
        druid_brokers: druid_brokers, kafka_nodes: kafka_nodes }
    end

    def min_height
      self.height = 180 if ["stacked", "area", "line", "bars", "sbars", "verbars", "pie"].include?(self.graph_type) and self.height.to_i < 180
      self.height = 400 if self.graph_type == "streams" and self.height.to_i < 400
    end

end
