#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Segment < ApplicationRecord
  # Change the database connection
  establish_connection :druid
  self.table_name = 'druid_segments'

  # Create an struct for the payload
  SegmentStruct =
    Struct.new(:data_source, :interval, :version, :loadSpec, :dimensions,
               :metrics, :shardSpec, :binaryVersion, :size, :identifier)

  #
  # Get the namespace of this data segment
  #
  # == Returns:
  # String with the UUID of data source
  #
  def namespace
    match = /rb_(.*_post|[^_]*)_(.*)/.match(datasource)
    return 'default' unless match && match[match.size - 1] != 'post'
    match[match.size - 1]
  end

  #
  # Load the payload of the segment
  #
  # == Returns::
  # Hash with data
  #
  def payload_hash
    begin
      JSON.parse([payload[2..-1]].pack('H*'))
    rescue JSON::ParserError
      begin
        JSON.parse([payload[1..-1]].pack('H*'))
      rescue JSON::ParserError
        JSON.parse(payload)
      end
    end
  end

  #
  # Load the payload of the segment and initialize structs
  #
  # == Returns::
  # Struct with data
  #
  def payload_struct
    SegmentStruct.new(*payload_hash.values)
  end

  #
  # Destroy the segmnet in s3
  #
  def destroy
    # If S3 is enabled, destroy the segments from it
    if StorageFile.s3_enabled?
      segment_s3_dir = self.payload_hash["loadSpec"]["key"][0..-38]
      s3_config = YAML.load_file("#{Rails.root}/config/aws.yml")
      # Get S3 connection
      s3 = AWS::S3.new(access_key_id: s3_config[Rails.env]["access_key_id"],
        secret_access_key: s3_config[Rails.env]["secret_access_key"],
        s3_endpoint: s3_config[Rails.env]["s3_host_name"])
      bucket = s3.buckets[s3_config[Rails.env]["bucket"]]
      # Get all files in that segment directory and delete them all
      files_to_delete = bucket.objects.with_prefix(segment_s3_dir).collect(&:key)
      files_to_delete.each do |file|
        bucket.objects[file].delete
      end
      # Delete the directory
      bucket.objects[segment_s3_dir].delete
    end
    
    # Call the parent destroy method to remove the segment from DB
    super
  end

  # Default Scope is made by function because of the complexity of it.
  def self.default_scope
    datasources = RedBorder::DRUID_CONFIG.map { |_, v| v['datasource'] if v.is_a?(Hash) }
    datasources_or = datasources.map { |_| 'datasource like ?' }.join(' OR ')
    where(datasources_or, *datasources.map { |d| "#{d}%" })
      .order("to_timestamp(start, 'YYYY-MM-DD\"T\"HH24:MI:SS\".000Z\"') DESC")
  end
end
