#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Sensors are machines or end point APIs that retrieve and store the data
# used in the platform. There are many type of sensors, defined in modules.
# The base sensor is the domain.
#
class Sensor < ApplicationRecord
  include Redborder::Audit
  include ChefProperty
  include PhysicalSensor

  # -------------------------------------------------------------
  # Constants
  # -------------------------------------------------------------

  # Kind of sensor
  DOMAIN = 1

  ROOT              = 0
  ORGANIZATION      = 2
  MARKET            = 3
  CAMPUS            = 4
  BUILDING          = 5
  SERVICE_PROVIDER  = 6
  DEPLOYMENT        = 7
  NAMESPACE         = 8

  # To order the domain_types leaf or tree,
  # we assign the 10* to leaf nodes.
  FLOOR         = 101

  # Recommended hierarchy
  HIERARCHY = [SERVICE_PROVIDER, NAMESPACE, ORGANIZATION, MARKET, DEPLOYMENT,
               CAMPUS, BUILDING, FLOOR].freeze

  DOMAIN_TYPES = {
    ROOT => {
      icon: 'fa fa-hierarchy', color: '#7ecefd',
      type_icon: 'font', name: 'Root'
    },
    ORGANIZATION => {
      icon: 'fa fa-group', color: '#7ecefd',
      type_icon: 'font', name: "Organization",
      parent_type: [NAMESPACE]
    },
    MARKET => {
      icon: 'fa fa-globe', color: '#c7b29b',
      type_icon: 'font', name: "Market",
      parent_type: [ORGANIZATION, DEPLOYMENT, CAMPUS, BUILDING]
    },
    CAMPUS => {
      icon: 'icons/campus.svg', color: '#c7b29b',
      type_icon: 'image', name: 'Campus',
      parent_type: [ORGANIZATION, MARKET, DEPLOYMENT, BUILDING]
    },
    BUILDING => {
      icon: 'fa fa-building-o', color: '#c7b29b',
      type_icon: 'font', name: "Building",
      parent_type: [ORGANIZATION, MARKET, DEPLOYMENT, CAMPUS]
    },
    SERVICE_PROVIDER => {
      icon: 'icons/sp.svg', color: '#c7b29b',
      type_icon: 'image', name: "Service Provider",
      parent_type: [ROOT]
    },
    DEPLOYMENT => {
      icon: 'fa fa-cube', color: '#c7b29b',
      type_icon: 'font', name: "Deployment",
      parent_type: [ORGANIZATION, MARKET, CAMPUS, BUILDING]
    },
    NAMESPACE => {
      icon: 'fa fa-database', color: '#c7b29b',
      type_icon: 'font', name: "Namespace",
      parent_type: [SERVICE_PROVIDER]
    },
    FLOOR => {
      icon: 'icons/floor.svg', color: '#c7b29b',
      type_icon: 'image', name: "Floor",
      parent_type: [ORGANIZATION, MARKET, DEPLOYMENT, CAMPUS, BUILDING]
    }
  }

  CLIENT_PROXY = 31

  EXPORTER = 41

  # Available SNMP versions
  SNMP_VERSIONS = %w(1 2c).freeze

  #
  # Symbols of keys to import a sensor.
  #
  cattr_accessor :import_columns_for
  self.import_columns_for = {}
  self.import_columns_for[:base] = %i(name uuid type)
  self.import_columns_for[DOMAIN] =
    %i(domain_type floor_level email gps_points general_description
       company domain_zones mac_hashing_salt)

  #
  # => Generate a helper to compare the type of the sensor.
  #    The methods created are: .domain?...
  #
  %w(organization service_provider market campus building floor deployment
     namespace).each do |dtype|
    define_method "#{dtype}?" do
      is_domain? && domain_type == Sensor.const_get(dtype.upcase.to_s)
    end
  end

  # Sensors per page
  SENSORS_PAGE = 20

  # Childrens Tiles per page
  CHILDREN_PAGE = 6

  audited except: [
    :latitude, :longitude, :is_deleted, :applied_at, :real_parent_id,
    :group_id, :binding_id, :need_apply, :applying, :property, :properties,
    :domain_type, :unassigned, :lft, :rgt, :depth, :children_count
  ], on: [:create, :update, :destroy]

  cattr_accessor :available_sensors
  self.available_sensors = [DOMAIN, CLIENT_PROXY, EXPORTER]

  default_scope { where(type: Sensor.available_sensors).order('type DESC, domain_type DESC, name') }

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------

  validates :name,
            presence: true,
            length: { in: 1..63 },
            uniqueness: { scope: [:parent_id, :type, :domain_type] },
            format: {
              with: /\A[\w\s_\.\-,@+=()!\/:#>&\u007C[\u00C0-\u0178][\u0400-\u04FF]]{2,50}\z/,
              message: "#{I18n.t('models.allowed_characters')}"
            }

  validates :uuid, uniqueness: true
  validates :name, uniqueness: true, if: proc { |sensor| sensor.name == 'root' }
  validates :domain_type, uniqueness: true, if: proc { |sensor| sensor.domain_type == ROOT }
  validates :domain_type, inclusion: {
    in: [ROOT, MARKET, ORGANIZATION, CAMPUS, BUILDING, FLOOR,
         SERVICE_PROVIDER, DEPLOYMENT, NAMESPACE]
  }, if: proc { |sensor| sensor.type == DOMAIN }

  # Check if the SNMP version is allowed
  validate :check_snmp_version, if: proc { |sensor| sensor.type != DOMAIN }
  validate :check_parent

  # Validate the domain type is not changed
  validate :domain_type_not_change

  DittoCode::Exec.if 'ENTERPRISE' do
    validate :valid_license
  end

  validate :parent_is_correct

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------

  has_many :permissions
  has_many :users_with_permission, through: :permissions, source: :user

  has_many :dashboard_sensors, dependent: :destroy
  has_many :dashboards, through: :dashboard_sensors
  has_many :report_sensors, dependent: :destroy
  has_many :reports, through: :report_sensors
  has_many :zones, dependent: :destroy

  has_one :overlay_map, dependent: :destroy
  has_many :gps_markers, through: :overlay_map

  has_many :incidents, dependent: :destroy

  belongs_to :real_parent, class_name: 'Sensor', foreign_key: 'real_parent_id'

  has_many :real_childrens, -> { order(name: :asc) },
           class_name: 'Sensor', foreign_key: 'real_parent_id',
           dependent: :destroy

  has_many :redborder_objects, dependent: :destroy

  DittoCode::Exec.if 'ENTERPRISE' do
  belongs_to :license # for sensors, the license is using.
  has_many :licenses, dependent: :destroy # for domains, the licenses that has been loaded in it.
  end

  has_many :alarms

  # We nede to set acts_as_nested_set at this position to destroy
  # real childrens (from relation) and later, destroy another children.
  #
  # - Angel
  #
  # Use nested sets to fast tree queries
  # When sensor is destroyed, destroy all its descendants
  acts_as_nested_set dependent: :destroy

  # -------------------------------------------------------------
  # Store
  # -------------------------------------------------------------
  store :property, coder: JSON, accessors: [
    :gps_points, :domain_zones, :domain_type_options, :general_description,
    :floor_level, :active_modules, :company, :email, :path, :mac_hashing_salt,
    :mac_address, :check_in_time, :visualization_range, :last_organization_notification
  ]

  # -------------------------------------------------------------
  # Before and after calls
  # -------------------------------------------------------------

  before_create :default_values

  before_destroy :disable_users_in_domain
  before_destroy :destroy_sensor
  before_destroy :destroy_chef_sensor
  before_destroy :handle_alarms_on_destroy

  after_create :create_chef_role
  after_update :update_run_list_chef_role,
               if: proc { |sensor| !sensor.is_root? }
  after_destroy :destroy_datasources, if: proc { |s| s.namespace? }

  around_move :update_chef_values_on_children

  # Fill organizations with license modules
  before_create :default_modules, if: proc { organization? }

  # We must reset active_data_sources of users
  after_move :reset_user_data_sources

  # Update path_names on children!
  after_move :update_path_names
  after_update :update_path_names,
               if: proc { |s| s.is_domain? && s.name_changed? }
  before_create :update_path_name

  # Update parent_id in chef sensor node
  after_update :update_chef_parent_id


  after_validation :geolocate_ip, on: [:create, :update]

  DittoCode::Exec.if 'ENTERPRISE' do
  # License callbacks
  after_save :update_license_data_bag
  after_destroy :remove_from_license_data_bag
  end

  # -------------------------------------------------------------
  # Chef
  # -------------------------------------------------------------

  has_chef_property :snmp_community, %w(redborder snmp_community)
  has_chef_property :snmp_version, %w(redborder snmp_version)
  has_chef_property :homenets, %w(redborder homenets)
  has_chef_role_property :index_partitions, %w(redborder index_partitions)
  has_chef_role_property :index_replicas, %w(redborder index_replicas)
  has_chef_role_property :megabytes_limit, %w(redborder megabytes_limit)

  # -------------------------------------------------------------
  # Serialization
  # -------------------------------------------------------------

  def as_json(options = {})
    options[:except] ||= [:created_at, :updated_at,
                          :lft, :rgt, :depth, :children_count]
    super
  end

  #
  # Set an array of properties a sensor must return when create it.
  #
  # == Parameters:
  # sensor::
  #   Sensor that will be returned
  #
  # == Returns:
  # An array of symbol
  #
  def to_api
    base = [:id, :name, :domain_type, :type, :uuid, :general_description,
            :company, :email]
    case
    when service_provider?
      as_json(only: base).merge('mac_hashing_salt' => mac_hashing_salt)
    when organization? && User.current_user && User.current_user.super_user?
      as_json(only: base).merge(
        'visualization_range' => visualization_range,
        'megabytes_limit' => megabytes_limit
      )
    else
      as_json(only: base)
    end
  end

  def check_in_time
    time = self[:property]["check_in_time"]
    time.is_a?(String) ? Time.parse(time).to_i : time
  rescue ArgumentError => e
    nil
  end

  # -------------------------------------------------------------
  # Path methods
  # -------------------------------------------------------------

  #
  # Update path names of current sensors and children if is a domain.
  #
  def update_path_names
    # Update
    self_and_descendants.each do |s|
      s.update_path_name(true)
    end
  end

  # Returns a collection including itself and all of its nested available children
  def self_and_descendants
    # using _left_ for both sides here lets us benefit from an index on that column if one exists
    nested_set_scope.right_of(left).left_of(right).where(type: Sensor.available_sensors)
  end

  #
  # Update the path_name of the sensor.
  #
  # Parameters::
  # save_model::
  #   Boolean to save the model. False by default
  #
  def update_path_name(save_model = false)
    self.path = (ancestors_with_root - [Sensor.root]).map(&:name).join(' / ')
    save if save_model
  end

  #
  # Get the path name of the user
  #
  # == Parameters:
  # include_myself::
  #   Boolean indicate include the current sensor on path. True by default
  # user::
  #   User to check permissions of the route
  #
  # == Returns:
  # String with the current path name
  #
  def path_name(include_myself = true, user = User.current_user)
    # Check if the user has root as top domain. In this case we can use cached
    # routes
    return cached_path_name(include_myself) if user && user.sensor.is_root?
    if include_myself
      path = (ancestors.select { |s| user.can?(:read, s) }
               .map(&:name) + [name])
      path.count == 1 ? 'root' : path.drop(1).join(' / ')
    else
      path = ancestors.select { |s| user.can?(:read, s) }.map(&:name)
             .join(' / ')
      path.blank? ? 'root' : path
    end
  end

  #
  # Return cached string with apth name because user can see all sensors.
  #
  # == Parameters:
  # include_myself::
  #   Boolean indicate include the current sensor on path. True by default
  #
  # == Returns:
  # String with the current path name
  #
  def cached_path_name(include_myself)
    if include_myself
      path.blank? ? name : path + " / #{name}"
    else
      path.blank? ? 'root' : path
    end
  end

  #
  # Rebuild all path_name of sensors. This action take all sensors and
  # recalculate path names. Be careful, this will take some time.
  #
  def self.reload_path_names!
    Sensor.where.not(name: 'root').each do |s|
      s.update_path_name
      s.save
    end
  end

  # -------------------------------------------------------------
  # Users
  # -------------------------------------------------------------

  #
  # Reset the data sources of affected users
  #
  def reset_user_data_sources
    User.where(sensor_id: self_and_descendants.map(&:id)).each do |user|
      user.update(active_data_sources: [user.data_sources.first.uuid])
    end
  end

  # -------------------------------------------------------------
  # Domains
  # -------------------------------------------------------------

  #
  # Return the name of the sensor based on the domain type
  #
  # @param to_key [Boolean] Return the domain type as name or key to filter/Hash
  # @return [String] Domain type name
  #
  def domain_type_name(to_key: false)
    return '' unless type == DOMAIN && domain_type
    name = DOMAIN_TYPES[domain_type][:name]
    # Check conversion
    to_key ? name.tr(' ', '_').downcase! : name
  end

  # -------------------------------------------------------------
  # Chef methods
  # -------------------------------------------------------------

  define_method("update_chef_domain_values_#{}") do
    update_chef_domain_values_in_role
  end

  # Store hierarchy information on the nodes of a sensor. This is used by
  # sensors that haven't got chef clients. Sensors with chef clients can
  # inherit these values from parent roles.
  #
  def update_chef_domain_values_in_node
    node = chef_node
    return unless node
    # For each element in the hierarchy, try to get the values.
    (HIERARCHY - [FLOOR]).each do |domain_type|
      s = ancestors.find_by(domain_type: domain_type)
      if s
        dname = s.domain_type_name(to_key: true)
        node.normal['redborder'][dname] = s.try(:name)
        node.normal['redborder']["#{dname}_uuid"] = s.try(:uuid)
      else
        # We have to delete the references
        dname = DOMAIN_TYPES[domain_type][:name].tr(' ', '_').downcase
        node.normal['redborder'].delete(dname)
        node.normal['redborder'].delete("#{dname}_uuid")
      end
    end
    # Save the Chef node
    node.save
  end

  # Store hierarchy information on the role of a sensor.
  #
  def update_chef_domain_values_in_role
    role = chef_role
    return unless role
    # For each element in the hierarchy, try to get the values.
    (HIERARCHY - [FLOOR]).each do |domain_type|
      s = ancestors.find_by(domain_type: domain_type)
      if s
        dname = s.domain_type_name(to_key: true)
        role.override_attributes['redborder'][dname] = s.try(:name)
        role.override_attributes['redborder']["#{dname}_uuid"] = s.try(:uuid)
      else
        # We have to delete the references
        dname = DOMAIN_TYPES[domain_type][:name].tr(' ', '_').downcase
        role.override_attributes['redborder'].delete(dname)
        role.override_attributes['redborder'].delete("#{dname}_uuid")
      end
    end
    # Save the Chef node
    role.save
  end

  # -------------------------------------------------------------
  # Other methods
  # -------------------------------------------------------------

  # Returns an nested hash with sensor hierarchy.
  #
  # @return [Hash] Hierarchy of this sensor.
  #
  def self_and_children_as_json
    as_json.merge!(children: children_json)
  end

  # Check if the sensor has children
  #
  # @return [Boolean] Current sensor has children
  #
  def has_children?
    leaf?
  end

  # Return recommended type for children of this sensor. It's based in
  # HIERARCHY.
  #
  # @return [Fixnum] Recommended domain type of children
  #
  def children_type
    return HIERARCHY.first if is_root?
    HIERARCHY[HIERARCHY.index(domain_type) + 1] || FLOOR
  end

  #
  # Return the columns of this sensor can import. It's based on type of sensor
  # and import_columns_for class attribute. See definition on line 73
  #
  # == Returns:
  # Array of symbols
  #
  def import_columns
    import_columns_for[:base] + (import_columns_for[type] || [])
  end

  #
  # Static method to get import columns of a given sensor type. It uses
  # import_columns_for class attribute. See definition on line 73
  #
  # == Parameters:
  # type::
  #   Type of sensor
  #
  # == Returns:
  # Array of symbols
  #
  def self.import_columns(type)
    import_columns_for[:base] + (import_columns_for[type] || [])
  end

  #
  # Return permitted params for a given

  #
  # This methods UPDATE the current sensor and his tree. If there are any
  # children define, this method will call @see create_sensor_with_children
  # to create the children tree.
  #
  # == Parameters:
  # sensor_attr::
  #   A hash with the attributes to create the current sensor. This can include
  #   childrens
  # user::
  #   The user who perform the query
  #
  # == Returns:
  # Array with the UUID of sensors.
  #
  def override_sensor_with_children(sensor_attr, user)
    el = []
    # Update values.
    type = sensor_attr[:type] || DOMAIN
    attributes = sensor_attr.permit(*Sensor.import_columns(type))

    if update(attributes)
      # Store uuid and create childrens
      el << uuid

      # Add access points
      move_access_points(sensor_attr[:access_points], user)

      # Iterate over children
      (sensor_attr[:children] || []).each do |child|
        el += create_sensor_with_children(child, self, user)
      end
    else
      # Error on update
      err = "#{I18n.t('models.the_sensor')} #{name} #{I18n.t('models.cant_update')} "\
            "#{errors.full_messages.join(', ')}"
      fail(ActiveRecord::StatementInvalid, err)
    end
    # Return elements
    el
  end

  #
  # Full sync method for RIB.
  # Update the given hierarchy from the given nested JSON. If the sensor
  # exist, we must update the values, else, create it.
  #
  # == Parameters:
  # sensor_attr::
  #   Params hash with the attributes to create the current sensor. This can
  #   include childrens
  # parent::
  #   A sensor object with the parent.
  # user::
  #   The user who perform the query
  #
  # == Returns:
  # Array with the UUID of sensors.
  #
  def create_sensor_with_children(sensor_attr, parent, user)
    # Initialize elements to empty
    el = []
    # Create or update values.
    type = sensor_attr[:type] || DOMAIN
    attributes = sensor_attr.permit(*Sensor.import_columns(type))
    # Force domain type and type to generic
    attributes[:domain_type] ||= parent.children_type
    attributes[:type] ||= Sensor::DOMAIN

    # Find from parent
    if sensor_attr[:uuid]
      sensor = parent.children.where(uuid: sensor_attr[:uuid])
    else
      sensor = parent.children.where(name: attributes[:name],
                                     domain_type: attributes[:domain_type],
                                     type: attributes[:type])
    end

    # Check if sensor exist to update or create
    if sensor.present?
      sensor = sensor.first
      sensor.update(attributes)
    else
      sensor = Sensor.create(attributes.merge(parent: parent))
    end

    # Fail on not valid sensors
    unless sensor.valid?
      err = "#{I18n.t('models.the_sensor')} #{sensor.name} #{I18n.t('models.cant_create')} "\
            "#{sensor.errors.full_messages.join(', ')}"
      fail(ActiveRecord::StatementInvalid, err)
    end

    # Store the sensor uuid
    el << sensor.uuid

    # Add access points
    sensor.move_access_points(sensor_attr[:access_points], user)

    # Iterate over children
    (sensor_attr[:children] || []).each do |child|
      el += create_sensor_with_children(child, sensor, user)
    end

    el
  end

  def to_csv
    header = self.as_json.keys
    csv = header.join(',') + "\n"

    aux_array = []
    header.each do |head|
      aux_array << self[head]
    end
    csv += aux_array.join(',') + "\n"
  end

  def last_uptime
    begin
      node = chef_node

      return '' unless node && node['uptime']

      uptime = node['uptime']
  
      days = uptime[/(\d+)\s*days?/, 1].to_i
      hours = uptime[/(\d+)\s*hours?/, 1].to_i
      minutes = uptime[/(\d+)\s*minutes?/, 1].to_i
      seconds = uptime[/(\d+)\s*seconds?/, 1].to_i
  
      [].tap do |result|
        result << "#{days}d" if days > 0
        result << "#{hours}h" if hours > 0
        result << "#{minutes}m" if minutes > 0
        result << "#{seconds}s" if seconds > 0
      end.join(" ")
    rescue
      ''
    end
  end

  def self.to_csv(content)
    header = content.as_json.keys
    csv = header.join(',') + "\n"

    aux_array = []
    header.each do |head|
      aux_array << content[head]
    end
    csv += aux_array.join(',') + "\n"
  end

  def to_partial_path
    if is_domain?
      Sensor.send("to_partial_path_#{type}_#{domain_type}")
    else
      Sensor.send("to_partial_path_#{type}")
    end
  rescue
    Sensor.send("to_partial_path_#{type}")
  end

  def icon
    if is_domain?
      DOMAIN_TYPES[domain_type][:icon]
    else
      send("icon_#{type}")
    end
  end

  def icon_type
    if is_domain?
      DOMAIN_TYPES[domain_type][:type_icon]
    else
      'image'
    end
  end

  def self.to_partial_path_1
    'sensors/domain_sensor'
  end

  def self.to_partial_path_31
    'sensors/client_proxy_sensor'
  end

  def self.to_partial_path_41
    'sensors/exporter_sensor'
  end

  def full_name
    if type == DOMAIN
      name
    else
      send("full_name_#{type}")
    end
  end

  def secondary_name
    if type == DOMAIN || real_parent == parent
      name
    else
      send("secondary_name_#{type}")
    end
  end

  define_method("secondary_name_#{CLIENT_PROXY}") do
    name
  end

  define_method("secondary_name_#{EXPORTER}") do
    name
  end

  define_method("full_name_#{CLIENT_PROXY}") do
    name
  end

  define_method("full_name_#{EXPORTER}") do
    name
  end

  define_method("version_#{CLIENT_PROXY}") do
    chef_node['redborder']['rpms']['proxy'] rescue nil
  end

  define_method("version_#{EXPORTER}") do
    chef_node['redborder']['rpms']['exporter'] rescue nil
  end

  def complete_name
    if type == DOMAIN
      name
    else
      send("complete_name_#{type}")
    end
  end

  define_method("complete_name_#{CLIENT_PROXY}") do
    name
  end

  define_method("complete_name_#{EXPORTER}") do
    name
  end

  define_method("icon_#{CLIENT_PROXY}") do
    'icons/client_proxy.png'
  end

  define_method("icon_#{EXPORTER}") do
    'icons/exporter.svg'
  end

  def self.inheritance_column
    nil
  end

  def real_sensor
    respond_to?("real_sensor_#{type}") ? send("real_sensor_#{type}") : self
  end

  def real_children
    Sensor.where(real_parent_id: id, is_deleted: false)
  end

  def is_client_proxy?
    type == CLIENT_PROXY
  end

  def is_exporter?
    type == EXPORTER
  end

  def is_domain?
    type == DOMAIN
  end

  #
  # Check if a LatLng is contained in a geoposicionated floor.
  #
  # == Parameters:
  # point::
  #   A hash with format:
  #     { latitude: float, longitude: float }
  #
  # == Returns:
  # True or false
  #
  def contains?(point)
    return unless geoposicionated?

    m_trans = self[:property][:matrix_trans]
    m_offset = self[:property][:matrix_offset]
    w = self[:property][:dimensions][:width]
    h = self[:property][:dimensions][:height]
    nw = PrimeInfrastructure::Functions
       .point_to_lnglat(m_trans, m_offset, [0, 0])
    se = PrimeInfrastructure::Functions
       .point_to_lnglat(m_trans, m_offset, [w, h])
    ( point[:longitude] > nw[0,0] && point[:longitude] < se[0,0] && point[:latitude] < nw[1,0] && point[:latitude] > se[1,0] )
  end

  #
  # Check if the current sensor is geoposicionated
  #
  # == Returns:
  # True or false
  #
  def geoposicionated?
    self[:property][:matrix_trans] &&
      self[:property][:matrix_offset] &&
      self[:property][:dimensions][:width] &&
      self[:property][:dimensions][:height]
  end

  def type_to_s
    case type
    when DOMAIN
      return 'domain'
    when CLIENT_PROXY
      return 'client_proxy'
    when EXPORTER
      return 'exporter'
    else
      send("type_to_s_#{type}")
    end
  end

  def type_humanize
    case type
    when DOMAIN
      return 'domain'
    else
      send("type_humanize_#{type}")
    end
  end

  define_method("type_humanize_#{CLIENT_PROXY}") do
    "proxy"
  end

  define_method("type_humanize_#{EXPORTER}") do
    "exporter"
  end

  #
  # How to reference sensor's type in the licenses
  #
  def license_type
    send("license_type_#{type}")
  rescue NoMethodError
    type_to_s
  end

  # -------------------------------------------------------------
  # Relation with sensors
  # -------------------------------------------------------------

  #
  # Return the namespaces based on current sensor. If the sensor has an
  # ancestor namespace, it will be returned. Else, all descendants
  # namespaces + root will be returned.
  #
  # == Parameters:
  # user::
  #   (Optional) To check if the user has permission on current namespaces.
  #
  # == Returns:
  # Array of Sensor (namespaces)
  #
  def namespaces(user = nil)
    # Check permissions for the user
    return Sensor.none if user && user.cannot?(:read, self)
    namespaces = ancestors_namespaces
    # Check if there are a nampesce up of the sensor
    return [namespaces.first] if namespaces.present?
    # There aren't ancestors, try to search children
    namespaces = children_namespaces
    # Root means default
    [Sensor.root] + namespaces
  end

  #
  # Ancestors namespaces based on user top domain
  #
  # == Returns:
  # ActiveRecord relation, we must check permissions
  #
  def ancestors_namespaces
    self_and_ancestors_with_root.where(domain_type: Sensor::NAMESPACE)
  end

  #
  # Children namespaces based on user top domain
  #
  # == Returns:
  # ActiveRecord relation, we must check permissions
  #
  def children_namespaces
    self_and_descendants.where(domain_type: Sensor::NAMESPACE)
  end

  #
  # Ancestors organizations based on user top domain
  #
  # == Returns:
  # ActiveRecord relation, we must check permissions
  #
  def ancestors_organizations
    self_and_ancestors_with_root.where(domain_type: Sensor::ORGANIZATION)
  end

  #
  # Get all IDs of sensor children.
  #
  # == Parameters:
  # include_itself::
  #   Boolean indicating if includes current sensor.
  #
  # == Returns:
  # Array of integers (IDs)
  #
  def all_children_ids(include_itself = false)
    all_children(include_itself).map(&:id)
  end

  # Get all sensor children.
  #
  # @param include_itself [Boolean] includes current sensor
  # @return [ActiveRecord::Relation] Sensors
  #
  def all_children(include_itself = false)
    if include_itself
      self_and_descendants
    else
      descendants
    end
  end

  # Return only children that are floors
  #
  # @return [ActiveRecord::Relation] Floors
  #
  def floors
    children.where(type: DOMAIN, domain_type: FLOOR)
  end

  # Return the children of top domain of an user.
  #
  # @param user [User] to check permissions
  # @return [ActiveRecord::Relation] Sensors
  #
  def self.root_children(user = User.current_user)
    user.sensor.children
  end

  # Return the Root domain
  #
  # @return [Sensor] Root Sensor
  #
  def self.root
    Sensor.find_by name: 'root'
  end

  # Return the children of this sensor that can send data to Druid.
  #
  # @return [ActiveRecord::Relation] Sensors
  #
  def self_and_descendants_with_data
    self_and_descendants.where.not(type: DOMAIN)
  end

  #
  # => Return the tree for this sensor, this means
  #    the current sensor + ancestors + children
  #
  def tree(include_itself = false)
    if include_itself
      self_and_ancestors_with_root + all_children
    else
      ancestors_with_root + all_children
    end
  end

  #
  # => Return Array of hashes containing sensor
  #    details (ID, namespace, and name).
  #
  def self.get_all_flow_sensors
    if Sensor.where(type: RedBorder::SensorFlow::FLOW).present?
      sensors = Sensor.where(type: RedBorder::SensorFlow::FLOW).map do |sensor|
        {
          sensor_id: sensor.id,
          namespace_uuid: find_namespace_uuid(sensor),
          domain_uuid: find_domain_uuid(sensor),
          sensor_uuid: sensor.uuid,
          sensor_name: sensor.name
        }
      end
    else
      sensors = []
    end
    sensors
  end

  #
  # => Return namespace uuid for the sensor
  #
  def self.find_namespace_uuid(sensor)
    namespace = sensor.namespaces.first
    namespace.root? ? '' : namespace.uuid
  end

  #
  # => Return first domain parent like uuid for the sensor
  #
  def self.find_domain_uuid(sensor)
    loop do
      parent = sensor.parent
      return parent.uuid if parent.type == DOMAIN

      sensor = sensor.parent
      
      break unless sensor

      break if sensor && sensor.root?
    end
    ""
  end

  #
  # => Return the tree for this sensor, this means
  #    the current sensor + ancestors + children
  #
  def tree_domains(include_itself = false)
    if include_itself
      self_and_ancestors_with_root + all_children.where(type: DOMAIN)
    else
      ancestors_with_root + all_children.where(type: DOMAIN)
    end
  end

  # Returns list of ancestors, starting from parent until root (included).
  #
  # root
  #   - child1
  #     - subchild1
  #
  # subchild1.ancestors # => [child1, root]
  alias_method :ancestors_with_root, :ancestors

  # Returns list of ancestors, starting from parent until root (not included).
  #
  # Root
  #   - child1
  #     - subchild1
  #
  # subchild1.ancestors # => [child1]
  def ancestors
    nodes = ancestors_with_root - [Sensor.root]
    Sensor.unscoped.where(id: nodes.map(&:id)).order(:depth)
  end

  # Returns self and the list of ancestors, starting from parent
  # until root (included).
  #
  # root
  #   - child1
  #     - subchild1
  #
  # subchild1.ancestors # => [subchild1, child1, root]
  alias_method :self_and_ancestors_with_root, :self_and_ancestors

  # Returns self and the list of ancestors, starting from parent
  # until root (not included).
  #
  # root
  #   - child1
  #     - subchild1
  #
  # subchild1.ancestors # => [subchild1, child1]
  def self_and_ancestors
    nodes = self_and_ancestors_with_root - [Sensor.root]
    Sensor.unscoped.where(id: nodes.map(&:id)).order(:depth)
  end

  def is_root?
    name == 'root'
  end

  # Returns the first domain of its ancestors, i.e. the closest domain
  def first_domain
    if is_domain?
      self
    elsif parent_id == 1
      Sensor.root
    else
      parent.first_domain
    end
  end

  def objects
    if self.is_root?
      super
    else
      main_objects('objects', RedborderObject)
    end
  end

  def net_objects
    if self.is_root?
      super
    else
      main_objects('net_objects', NetObject)
    end
  end

  def icmp_objects
    if self.is_root?
      super
    else
      main_objects('icmp_objects', IcmpObject)
    end
  end

  def protocol_objects
    if self.is_root?
      super
    else
      main_objects('protocol_objects', ProtocolObject)
    end
  end

  def mac_objects
    if self.is_root?
      super
    else
      main_objects('mac_objects', MacObject)
    end
  end

  def interface_objects
    if self.is_root?
      super
    else
      main_objects('interface_objects', InterfaceObject)
    end
  end

  def service_objects
    if self.is_root?
      super
    else
      main_objects('service_objects', ServiceObject)
    end
  end

  def vlan_objects
    if self.is_root?
      super
    else
      main_objects('vlan_objects', VlanObject)
    end
  end

  # Check if current sensor is physical. This method is used now to reregister
  # the sensor and get a new private key. When we change the model of sensors to
  # inheritance, we can delete this method and uncomment physical? defined un
  # PhysicalSensor module
  #
  # @return [Boolean] Indicating if sensor is physical
  #
  def physical?
    respond_to?("physical_#{type}?") ? send("physical_#{type}?") : false
  end

  #
  # Return Chef node of this sensor.
  #
  # == Returns:
  # Chef::Node initializated.
  #
  def chef_node
    return nil if type == DOMAIN
    @node ||= send("chef_node_#{type}")
  end

  #
  # Get cheg node name of current sensor. It's based on the type of the sensor
  #
  # == Returns:
  # String with Chef node name
  #
  def chef_node_name
    respond_to?("chef_node_name_#{type}") ? send("chef_node_name_#{type}") :
                                            "rBsensor-#{id}"
  end

  # -------------------------------------------------------------
  # Client Proxy
  # -------------------------------------------------------------

  #
  # Client Proxy is a physical sensor.
  #
  # == Returns:
  # true
  #
  define_method("physical_#{CLIENT_PROXY}?") do
    true
  end

  # -------------------------------------------------------------
  # Exporter
  # -------------------------------------------------------------

  #
  # Exporter is a physical sensor.
  #
  # == Returns:
  # true
  #
  define_method("physical_#{EXPORTER}?") do
    true
  end

  #
  # Define Chef name for NODE and APICLIENT!
  #
  # == Returns:
  # String with the name
  #
  define_method("chef_node_name_#{CLIENT_PROXY}") do
    "rbproxy-#{id}"
  end

  #
  # Define Chef name for NODE and APICLIENT!
  #
  # == Returns:
  # String with the name
  #
  define_method("chef_node_name_#{EXPORTER}") do
    "rbexporter-#{id}"
  end

  define_method("chef_node_#{CLIENT_PROXY}") do
    Chef::Node.load("rbproxy-#{id}") rescue nil
  end

  define_method("chef_node_#{EXPORTER}") do
    Chef::Node.load("rbexporter-#{id}") rescue nil
  end

  define_method("chef_client_#{CLIENT_PROXY}") do
    Chef::ApiClient.load("rbproxy-#{id}") rescue nil
  end

  define_method("chef_client_#{EXPORTER}") do
    Chef::ApiClient.load("rbexporter-#{id}") rescue nil
  end

  def chef_role
    @role ||= if is_root?
                Chef::Role.load('sensor')
              else
                Chef::Role.load("rBsensor-#{id}")
              end
  end

  def chef_enabled_set(activated)
    chef_role.override_attributes['redborder']['chef_enabled'] = activated
    chef_role.save
  end

  def chef_enabled?
    chef_enabled = chef_role.override_attributes['redborder']['chef_enabled']
    chef_enabled.nil? || chef_enabled
  end

  def role_value(attributes)
    value = chef_role.override_attributes

    attributes.each do |k|
      value.has_key? k ? value = value[k] : value = nil
      if value.nil?
        break
      end
    end

    if (value.nil? or ([Array, Hash].include?(value.class) and value.blank?))
      if parent
        value = parent.role_value(attributes)
      else
        value = Sensor.root.chef_role.override_attributes
        attributes.each do |k|
          value.has_key? k ? value = value[k] : value = nil
          if (value.nil? or value.blank?)
            break
          end
        end
      end
    end

    value
  end

  # Return all ipvars/portvars values, even the inherited.
  def role_value_for_vars(type)
    Sensor.root.role_value(['redborder', 'snort', 'default', type.to_s])
      .each_with_object({}) do |(k, _), h|
      h[k] = role_value(['redborder', 'snort', 'default',
                         type.to_s, k.to_s])
    end
  end

  def update_chef_node(values)
    node = chef_node
    node.normal["redborder"]["snmp_community"] = values["redborder"]["snmp_community"]
    node.normal["redborder"]["snmp_version"] = values["redborder"]["snmp_version"]
    node.normal["redborder"]["telnet_user"] = values["redborder"]["telnet_user"]
    node.normal["redborder"]["telnet_password"] = values["redborder"]["telnet_password"]
    node.save
  end

  def execute_rb_get_bulkstats_columns
    if is_client_proxy?
      node = chef_node
      node.normal["redborder"]["pending_bulkstats_changes"] = 1
      node.save
    end
  end

  def update_chef_role(values)
    if respond_to? :update_chef_role_IPS
      update_chef_role_IPS values
      return
    end

    if values["redborder"]["proxy"]["enabled"].blank?
      values["redborder"]["proxy"].delete("enabled")
      values["redborder"]["proxy"].delete("address") if values["redborder"]["proxy"]["address"].blank?
      values["redborder"]["proxy"].delete("port") if values["redborder"]["proxy"]["port"].blank?
      values["redborder"]["proxy"].delete("user") if values["redborder"]["proxy"]["user"].blank?
      values["redborder"]["proxy"].delete("password") if values["redborder"]["proxy"]["password"].blank?
    elsif values["redborder"]["proxy"]["enabled"] == "true"
      values["redborder"]["proxy"]["enabled"] = true
    else
      values["redborder"]["proxy"] = {"enabled" => false}
    end

    values["redborder"]["rsyslog"]["servers"] = values["redborder"]["rsyslog"]["servers"].split(",")
    values["redborder"]["rsyslog"].delete("protocol") if values["redborder"]["rsyslog"]["protocol"].blank?
    values["redborder"]["rsyslog"].delete("mode") if values["redborder"]["rsyslog"]["mode"].blank?

    role = chef_role
    role.override_attributes["redborder"]["rsyslog"] = values["redborder"]["rsyslog"]
    role.save
  end

  #
  # => It represents the sensor that responds to wake up chef call
  #
  define_method("wakeupable_#{DOMAIN}") do
    self
  end

  def wakeup_chef
    case type
    when Sensor::DOMAIN
      sensors = all_children.map do |s|
        s.send("wakeupable_#{s.type}") rescue nil
      end

      sensors.compact.uniq.each{|s| s.wakeup_chef}

    else
      execute('sudo -E /usr/lib/redborder/bin/rb_wakeup_chef.sh')
    end
  rescue
    Rails.logger.info $!.inspect
    Rails.logger.info $@.join("\n")
  end

  def chef_client
    respond_to?("chef_client_#{type}") ? self.send("chef_client_#{type}") : nil
  end

  def chef_name
    "rBsensor-#{self.id}"
  end

  # TODO: THIS NEED TO BE DELETED
  # def apply_objects
  #   role = self.chef_role

  #   role.override_attributes["redborder"]["objects"] ||= {}

  #   RedborderObject::OBJECT_TYPES.each do |key, value|

  #     if key == :net

  #       hosts = {}
  #       networks = {}

  #       RedborderObject.where(sensor_id: id, type: "#{key.to_s.capitalize}Object").select(:name, :value).each do |object|
  #         if object.value =~ /^((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))$/
  #           hosts[object.name] = object.value
  #         elsif object.value =~ /^((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))\/32$/
  #           host = object.value.match(/^((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))\/32$/)[1]
  #           object.assign_attributes(value: host)
  #           hosts[object.name] = object.value
  #         else
  #           networks[object.name] = object.value
  #         end
  #       end

  #       role.override_attributes["redborder"]["objects"]["hosts"] = hosts
  #       role.override_attributes["redborder"]["objects"]["networks"] = networks

  #     else

  #       role.override_attributes["redborder"]["objects"][value.to_s] ||= {}

  #       RedborderObject.where(sensor_id: id, type: "#{key.to_s.capitalize}Object").select(:name, :value).each do |object|
  #         role.override_attributes["redborder"]["objects"][value.to_s][object.name] = object.value
  #       end

  #     end

  #   end

  #   role.save
  # end

  def last_check_in
    case type
    when DOMAIN
      false
    else
      self.send("last_check_in_#{type}")
    end
  rescue NameError
    false
  end

  define_method("last_check_in_#{CLIENT_PROXY}") do
    begin
      Time.at(chef_node[:ohai_time])
    rescue
      return false
    end
  end

  define_method("last_check_in_#{EXPORTER}") do
    begin
      Time.at(chef_node[:ohai_time])
    rescue
      return false
    end
  end

  def segments
    respond_to?("segments_#{type}") ? self.send("segments_#{type}") : nil
  end

  def ifaces
    return nil unless (nil != chef_node and chef_node.key? "network" and chef_node[:network].key? "interfaces")
    chef_node[:network][:interfaces].reject { |k| k == "lo" }
  end

  def default_iface
    return nil unless (nil != chef_node and chef_node.key? "network" and chef_node[:network].key? "default_interface")
    chef_node[:network][:default_interface]
  end

  def cpus_number
    respond_to?("cpu_number_#{type}") ? self.send("cpu_number_#{type}") : nil
  end

  def hardware_status
    return nil unless chef_node and chef_node[:redborder][:ipmi]
    sensors = chef_node[:redborder][:ipmi][:sensors]
    if sensors.blank?
      nil
    else
      sensors.select{|s| !["ok", "nc", "ns"].include?s["status"]}
    end
  end

  #
  # Return status of segments.
  #
  # == Returns:
  # Hash with values
  #
  def segments_by_status
    status = {}
    return status unless chef_node && chef_node[:redborder][:segments]
    chef_node[:redborder][:segments].each do |k, v|
      (status[v['status']] ||= []) << k
    end
    status
  end

  # Return the Sensor to show as data in map tooltips.
  #
  # @return [Hash] Data to show the sensor in a tooltip.
  #
  def to_map
    description = "<html>
                      <b>" + I18n.t('models.name') +"</b> #{name}<br/>
                      <b>IP:</b> #{ip}<br/>
                      <b>" + I18n.t('models.latitude') +"</b> #{latitude}<br/>
                      <b>" + I18n.t('models.longitude') +"</b> #{longitude}<br/>
                  </html>"
    { lat: latitude, lng: longitude, description: description }
  end

  # Return the zones of a sensor to represent in a map using infrastructureMap
  # javascript library.
  #
  # @return [Hash] Data to represent the zone
  #
  def to_domain_zone
    parsed_zones = parsed_domain_zones
    return unless parsed_zones.present?
    {
      domain_zones: parsed_zones,
      title: name,
      id: id,
      icon: DOMAIN_TYPES[domain_type][:icon],
      type_icon: DOMAIN_TYPES[domain_type][:type_icon],
      color: DOMAIN_TYPES[domain_type][:color]
    }
  end

  # Return parsed domain zones from JSON
  #
  # @return [Array] Domain zones of the sensor
  #
  def parsed_domain_zones
    return [] unless domain_zones.present?
    JSON.parse(domain_zones)
  end

  # Execute a SSH command in a sensor.
  #
  # @param commands [Array, String] Commands to execute in the sensor.
  # @return [String] Output of the execution
  #
  def execute_output(commands)
    return '' if type == DOMAIN
    output = ''
    Net::SSH.start(
      real_sensor.ip, 'redborder',
      auth_methods: ['publickey'], paranoid: false, timeout: 5,
      keys: ["#{Rails.root}/config/#{RedBorder::CONFIG[:rsa_filename]}"],
      config: false) do |ssh|
      # Execute the command
      if commands.respond_to?('each')
        commands.each { |cmd| output += ssh.exec!(cmd) }
      else
        output += ssh.exec!(commands)
      end
    end
    # Return output
    output
  end

  # Return users that exist from this level (domain/sensor)
  def users
    sensor_ids = ancestors.map(&:id)
    # also include the root sensor and self
    User.where(sensor_id: sensor_ids + [Sensor.root.id, id])
  end

  def children_json
    children.reorder(:type).map do |sensor|
      respond_to?("as_json_with_children_#{sensor.type}") ?
        sensor.send("as_json_with_children_#{sensor.type}") :
        sensor.as_json.merge!(children: sensor.children_json)
    end.compact
  end

  #
  # Return if a sensor/domain can contain others sensors/domains
  # By default it will be false
  #
  def can_have_children?
    if is_domain?
      domain_type != Sensor::FLOOR
    elsif is_client_proxy?
      true
    else
      self.send("can_have_children_#{type}?")
    end
  rescue NoMethodError
    false
  end

  #
  # Return if a domain can contain domains
  #
  def domain_children?
    is_domain? && children.where(type: Sensor::DOMAIN).present?
  end

  def update_chef_domain_values
    respond_to?("update_chef_domain_values_#{type}", :include_private) ?
      self.send("update_chef_domain_values_#{type}") : nil
  end

  # -------------------------------------------------------------
  # Methods for organizations
  # -------------------------------------------------------------

  #
  # Check if an organization has an active mod
  #
  # == Parameters:
  # mod:
  #   An String with the name of module
  #
  # == Returns:
  # True if the module is active
  #
  def active_module?(mod)
    # Set an error if the domain isn't an organization
    fail "#{I18n.t('models.only_org')}" unless
      domain_type == ORGANIZATION

    # Load modules
    @modules ||= active_modules.split(',')
    # Return the result
    @modules.include?(mod.to_s)
  end

  def version
    self.send("version_#{type}")
  rescue NoMethodError
  end

  def reduced_version
    self.send("reduced_version_#{type}")
  rescue NoMethodError
  end

  def complete_version
    if chef_node && chef_node['redborder']['snort'] &&
       chef_node['redborder']['snort']['version']
      "Snort Version: #{chef_node['redborder']['snort']['version']}".html_safe
    end
  end

  #
  # Method user to get the organization that belongs a sensor/domain.
  # If no organization exists in it ancestors, return root
  #
  def ancestor_organization
    if RedBorder::LICENSING_MODE == 'global'
      Sensor.root
    elsif RedBorder::LICENSING_MODE == 'organization'
      parent.ancestors_organizations.first
    end
  end

  protected

  def main_objects(object_assoc, object_class)
    object_ids = Sensor.root.send(object_assoc).collect{|o| self.find_object(object_class, o.id)}.map(&:id)
    object_class.send("where", {id: object_ids})
  end

  def find_object(object_class, object_id)
    if self.is_root?
      object_class.send("find", object_id)
    else
      object_class.send("where", {sensor_id: real_sensor.id, parent_id: object_id}).first or real_sensor.parent.find_object(object_class, object_id)
    end
  end

  def execute(commands)
    return if type == DOMAIN
    Net::SSH.start(self.real_sensor.ip, "redborder", auth_methods: ["publickey"], paranoid: false,
                  timeout: 5, keys: ["#{Rails.root}/config/#{RedBorder::CONFIG[:rsa_filename]}"],
                  config: false) do |ssh|

      commands.respond_to?('each') ? commands.each{|cmd| ssh.exec!(cmd)} : ssh.exec!(commands)
    end
  end

  #
  # Customize the comment depending on the type of sensor
  #
  def audit_comment_on_create(_audit)
    case type
    when Sensor::DOMAIN
      "#{I18n.t('models.created_domain')} #{Sensor::DOMAIN_TYPES[domain_type][:name]}: #{name}"
    else
      "#{I18n.t('models.created_sensor')} #{type_humanize}: #{name}"
    end
  end

  def audit_comment_on_update(audit)
    comment = "#{I18n.t('models.updated')} #{name}:"
    comment += (audit.audited_changes || {}).collect do |key, values|
      case key.to_sym
      when :parent_id
        " #{I18n.t('models.parent_to')} '#{parent.try(:name)}'"
      else
        " #{key} #{I18n.t('models.from')} '#{values[0]}' #{I18n.t('models.to')} '#{values[1]}'"
      end
    end.join(', ')
    comment
  end

  private

  def update_license_data_bag
    data = Chef::DataBagItem.load('rBglobal', 'licenses')

    if license_id == 0
      data['sensors'][uuid] = { license: 'auto', type: type_to_s }
    elsif license.nil?
      data['sensors'].delete(uuid)
    else
      data['sensors'][uuid] = { license: license.uuid, type: license_type }
    end
    data.save
  end

  def remove_from_license_data_bag
    data = Chef::DataBagItem.load('rBglobal', 'licenses')
    data['sensors'].delete(uuid)
    data.save
  end

  def disable_users_in_domain
    ancestor = self.ancestors
                   .where(type: Sensor::DOMAIN).reverse.first || Sensor.root
    destroyed_ids = self.self_and_descendants.map(&:id)
    User.unscoped.where(sensor_id: destroyed_ids)
        .update_all(sensor_id: ancestor, enabled: false)
  end

  def default_values
    # TODO: Set policy by default
    self.uuid ||= SecureRandom.uuid
  end

  #
  # Set default modules when create a new organization
  #
  def default_modules
    return unless organization? && active_modules.nil?
    # Save available modules
    self.active_modules = Redborder::Module.registered_modules.keys.join(',')
  end

  # Check if the domain_type of the sensor is changed
  #
  def domain_type_not_change
    return unless domain_type_changed? && persisted?
    errors.add(:domain_type, "#{I18n.t('models.cant_change')}")
  end

  # Check that the associated license is valid
  #
  def valid_license
    return unless license
    errors.add(:license_id, 'The license is expired') if license.expired?
    if license.sensor != ancestor_organization
      errors.add(:base, "#{I18n.t('models.license_incorrect')}")
    end
  end

  def parent_is_correct
    return unless is_domain? && domain_type && DOMAIN_TYPES[domain_type]
    return if (parent_type = DOMAIN_TYPES.try(:[], domain_type).try(:[], :parent_type)).nil?
    return if parent_type.include?(parent.domain_type)
    errors.add(:parent_id, "#{I18n.t('models.wrong_parent')}")
  end

  def geolocate_ip
    return if latitude && longitude
    return unless ip && IPAddress.valid?(ip) && !IPAddress(ip).private?
    geolocation = Redborder::Geoip.geolocate(ip)
    self.latitude = geolocation[:latitude]
    self.longitude = geolocation[:longitude]
  end

  # Create the Chef Role of the sensor. It will send the method to the correct
  # based on the type
  #
  def create_chef_role
    send("create_chef_role_#{type}")
  end

  # Create Chef Role of domain.
  #
  define_method("create_chef_role_#{DOMAIN}") do
    run_list = is_root? ? 'role[sensor]' : "role[#{parent.chef_name}]"
    builder = Redborder::Chef::Role::Builder.new(self, run_list, domain: true)
    builder.create
  end

  define_method("create_chef_role_#{CLIENT_PROXY}") do
    run_list = ['role[proxy-sensor]', "role[#{parent.chef_name}]"]
    override_attributes = {
      'redborder' => {
        'role' => chef_name,
        'sensors_mapping' => {}
      }
    }
    builder = Redborder::Chef::Role::Builder.new(
      self, run_list, override_attributes: override_attributes)
    builder.create
  end

  define_method("create_chef_role_#{EXPORTER}") do
    run_list = ['role[exporter-sensor]', "role[#{parent.chef_name}]"]
    override_attributes = {
      'redborder' => {
        'role' => chef_name,
        'sensors_mapping' => {}
      }
    }
    builder = Redborder::Chef::Role::Builder.new(
      self, run_list, override_attributes: override_attributes)
    builder.create
  end

  define_method("run_list_text_#{CLIENT_PROXY}") do
    'role[proxy-sensor]'
  end

  define_method("run_list_text_#{EXPORTER}") do
    'role[exporter-sensor]'
  end

  def update_run_list_chef_role
    role = chef_role
    run_list_text = if is_root?
                      ['role[sensor]']
                    else
                      ["role[#{parent.chef_name}]"]
                    end

    begin
      run_list_text << send("run_list_text_#{type}") unless is_domain?
    rescue NoMethodError
    end

    role.run_list(run_list_text)
    role.save
  end

  # Destroy datasources associated to this sensor
  #
  def destroy_datasources
    datasources = RedBorder::DRUID_CONFIG
                  .values
                  .map { |d| d.try(:[], 'datasource') rescue nil }
                  .compact
    # Initialize cluster
    Redborder::Cluster.instance
    datasources.each do |prefix|
      ds_name = "#{prefix}_#{uuid}"
      Rails.logger.info "Destroying datasource -> #{ds_name} ..."
      # Destroy the datasource
      res = DruidConfig::Entities::DataSource.new(
        { 'name' => ds_name }, {}
      ).destroy
      # Log it
      if res.nil?
        Rails.logger.info "Destroyed datasource -> #{ds_name}"
      else
        Rails.logger.error "There was a problem destroying -> #{ds_name}. "\
                           "Response from Druid: #{res}"
      end
    end
  end

  #
  # Use this callback to use alias_method_chain later
  #
  #
  def destroy_sensor
  end

  def destroy_chef_sensor
    begin
      chef_client.try("destroy")
    rescue
      #client not found. Ignore it
    end

    begin
      self.send("destroy_chef_node_#{type}")
    rescue
      #node not found. Ignore it
    end

    begin
      chef_role.destroy
    rescue
      #role not found. Ignore it
    end
  end

  define_method "destroy_chef_node_#{CLIENT_PROXY}" do
    chef_node.destroy
  end

  define_method "destroy_chef_node_#{EXPORTER}" do
    chef_node.destroy
  end

  # ----------------------------------------------------
  # Validations
  # ----------------------------------------------------

  #
  # Check the format of IPs. This method adds an error if the IP has an
  # unrecognized format.
  #
  def format_ip
    regex256 =
      /0
      |1(?:[0-9][0-9]?)?
      |2(?:[0-4][0-9]?|5[0-5]?|[6-9])?
      |[3-9][0-9]?/x

    errors.add(:ip, "#{I18n.t('models.valid_format')}") if
      !(ip =~ /\A(#{regex256})\.(#{regex256})\.(#{regex256})\.(#{regex256})\z/)
  end

  #
  # Check that the parent can have children
  #
  def check_parent
    return if parent_id.nil?
    parent = Sensor.find(parent_id)
    errors.add(:parent_id, "#{I18n.t('models.not_container')}") unless parent.can_have_children?
  end

  #
  # Check if the SNMP version is allowed
  #
  def check_snmp_version
    return if chef_node.nil? || snmp_version.nil? ||
              SNMP_VERSIONS.include?(snmp_version)
    errors.add(:snmp_version, "#{I18n.t('models.snmp_version_html')}" +
                              "versions: #{SNMP_VERSIONS.join(', ')}")
  end

  def update_chef_values_on_children
    return unless parent_id_changed?
    yield
    if [Sensor::DOMAIN, Sensor::CLIENT_PROXY, Sensor::EXPORTER].include? type
      all_children(true).where.not(type: Sensor::DOMAIN).each(&:update_chef_domain_values)
    else
      update_chef_domain_values
    end
  end

  def update_chef_parent_id
    return unless parent_id_changed?

    node = chef_node
    return unless node

    node.normal['redborder']['parent_id'] = parent_id
    node.save
  end

  def handle_alarms_on_destroy
    alarms.each do |alarm|
      # Actualiza la alarma antes de que se destruya el sensor
      alarm.update(incidents_enabled: false, sensor_id: nil)
    end
  end
end
