#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Store the settings of a user. It has a mutex to avoid writing
#    two threads at the same time.
#
class Setting < ApplicationRecord
  validates_uniqueness_of :name

  # ================================================================
  # Callbacks
  # ================================================================

  after_save :save_into_chef, if: proc { |setting| setting.name == 'webpage_url' }

  # ================================================================
  # Constants
  # ================================================================
  DBNAME = 'rBglobal'
  WEBPAGE_URL_REGEX = '^(http(s?):\/\/)?((?:[\w\-]+\.)+[a-z]{2,6}|(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))(\/)?$'

  # This is created to disable writing on two threads at same time
  @mutex = Mutex.new

  def self.method_missing(name, *args)
    if name.to_sym != :find_by_name
      if name.to_s.last == '='
        method = name.to_s.chomp('=').to_sym
        self[method] = args[0]
      elsif name.to_s.last == '?'
        method = name.to_s.chomp('?').to_sym
        self[method].to_i > 0
      else
        self[name.to_sym]
      end
    else
      super
    end
  end

  def self.[](name)
    setting = find_by_name(name)
    setting ? setting.value : '1'
  end

  def self.[]=(name, v)
    setting = find_or_create_by(name: name)
    setting.value = (v ? v : '')
    setting.save
    setting.value
  end

  def self.geoip_country_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'country')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoIP.dat")
    Setting.geoip_country_date = Time.now.to_i
    Setting.geoip_country_custom = 1
    Setting.geoip_country_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  def self.save_sp_sso_cert_tmp(file)
    path_tmp = "#{Rails.root}/tmp/sp-sso-tmp.crt"
    File.open(path_tmp, 'wb') { |f| f.write(file.read) }
  end

  def self.save_sp_sso_priv_key_tmp(file)
    path_tmp = "#{Rails.root}/tmp/sp-sso-tmp.key"
    File.open(path_tmp, 'wb') { |f| f.write(file.read) }
  end

DittoCode::Exec.if 'ENTERPRISE' do
  def self.save_http_key_tmp(file)
    path_tmp = "#{Rails.root}/tmp/HTTPKey-tmp.key"
    File.open(path_tmp, 'wb') { |f| f.write(file.read) }
  end

  def self.save_http_pem_tmp(file)
    path_tmp = "#{Rails.root}/tmp/HTTPPem-tmp.crt"
    File.open(path_tmp, 'wb') { |f| f.write(file.read) }
  end

  def self.save_http_files(crt_filename, key_filename)
    return false unless crt_matches_key
    Setting.http_key_file = "#{Rails.root}/tmp/HTTPKey-tmp.key"
    Setting.http_key = key_filename
    Setting.http_pem_file = "#{Rails.root}/tmp/HTTPPem-tmp.crt"
    Setting.http_pem = crt_filename
  end

  def self.crt_matches_key
    key = `openssl rsa -noout -modulus -in tmp/HTTPKey-tmp.key`
    crt = `openssl x509 -noout -modulus -in tmp/HTTPPem-tmp.crt`
    return false unless key.present? && crt.present?
    key.eql?(crt)
  end

  def self.http_key_content
    IO.read(File.expand_path("tmp/HTTPKey-tmp.key"))
  end

  def self.http_pem_content
    IO.read(File.expand_path("tmp/HTTPPem-tmp.crt"))
  end
end

  def self.geoip_countryv6_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'countryv6')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoIPv6.dat")
    Setting.geoip_countryv6_date = Time.now.to_i
    Setting.geoip_countryv6_custom = 1
    Setting.geoip_countryv6_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  def self.geoip_city_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'city')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoLiteCity.dat")
    Setting.geoip_city_date = Time.now.to_i
    Setting.geoip_city_custom = 1
    Setting.geoip_city_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  def self.geoip_cityv6_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'cityv6')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoLiteCityv6.dat")
    Setting.geoip_cityv6_date = Time.now.to_i
    Setting.geoip_cityv6_custom = 1
    Setting.geoip_cityv6_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  def self.geoip_asn_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'asn')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoIPASNum.dat")
    Setting.geoip_asn_date = Time.now.to_i
    Setting.geoip_asn_custom = 1
    Setting.geoip_asn_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  def self.geoip_asnv6_file=(file)
    return false unless Redborder::Geoip.check_database(file, 'asnv6')
    FileUtils.cp(file, "#{Rails.root}/tmp/geodb/GeoIPASNumv6.dat")
    Setting.geoip_asnv6_date = Time.now.to_i
    Setting.geoip_asnv6_custom = 1
    Setting.geoip_asnv6_md5 = Digest::MD5.hexdigest(File.read(file))
    system('sudo /opt/rb/bin/rb_update_geodb_cookbook.sh')
  end

  DittoCode::Exec.if 'ENTERPRISE' do
    def self.virtual_ips
      {
        external_kafka: data_bag_item_attribute(DBNAME, 'ipvirtual-external-kafka', 'ip'),
        external_webui: data_bag_item_attribute(DBNAME, 'ipvirtual-external-webui', 'ip'),
        external_sfacctd: data_bag_item_attribute(DBNAME, 'ipvirtual-external-sfacctd', 'ip'),
        external_f2k: data_bag_item_attribute(DBNAME, 'ipvirtual-external-f2k', 'ip')
      }
    end

    def self.vips_mode
      vips = virtual_ips
      # internal = vips.select do |k, _v|
      #   k.to_s.include? 'internal'
      # end.values
      external = vips.select do |k, _v|
        k.to_s.include? 'external'
      end.values

      # if internal.uniq.length == 1 && external.uniq.length == 1
      if external.uniq.length == 1
        'basic'
      else
        'advanced'
      end
    end
  end

  def self.cluster_uuid
    Redborder::Databag.load('rBglobal', 'cluster')['uuid']
  end

  private

  def save_into_chef
    role = Chef::Role.load('manager')
    role.override_attributes['redborder'][name] = value
    role.save
  end

  def self.data_bag_item_attribute(data_bag, name, element)
    Chef::DataBagItem.load(data_bag, name)[element]
  rescue Net::HTTPServerException
    nil
  end
end
