DittoCode::HoldFile.if 'ENTERPRISE'
#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# A tier is a group of Historicals in Druid infrastructure.
#
class Tier < ApplicationRecord
  # Default Tier name
  DEFAULT_TIER_NAME = 'default'
  DEFAULT_HOT_TIER_NAME = 'hot'
  REAL_DEFAULT_TIER_NAME = '_default_tier'

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------

  validates :name, :user_id, presence: true
  validates :name,
            uniqueness: true,
            format: {
              with: /\A\w*\z/,
              message: "#{I18n.t('models.alphanumeric2')}"
            }
  validates :name, exclusion: { in: [DEFAULT_TIER_NAME] },
                   if: proc { Tier.exists?(name: DEFAULT_TIER_NAME) }
  validate :user_permissions

  # -------------------------------------------------------------
  # Callbacks
  # -------------------------------------------------------------

  # Default tier can't be destroyed
  before_destroy :check_protected, prepend: true
  before_update :check_protected

  after_update :update_nodes

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------

  # It has a creator
  belongs_to :user

  # Check if current tier is default
  #
  def default?
    name == DEFAULT_TIER_NAME
  end

  # Check if the current tier is the default hot
  #
  def default_hot?
    name == DEFAULT_HOT_TIER_NAME
  end

  # Check if the Tier is protected. This means you can't update or delete it
  #
  def protected?
    default? || default_hot?
  end

  #
  # Return current historicals
  #
  def historicals
    @historicals ||= []
  end

  #
  # Convert the tier name to the real name
  #
  def default_tier_name
    return name unless name == REAL_DEFAULT_TIER_NAME
    DEFAULT_TIER_NAME
  end

  #
  # Like #real_tier_name, but receive the name as parameter
  #
  def self.default_tier_name(name)
    return name unless name == REAL_DEFAULT_TIER_NAME
    DEFAULT_TIER_NAME
  end

  #
  # Load Tiers of historicals based on Chef.
  #
  # == Returns:
  # Hash with format:
  #   { 'node' => 'tier', 'node2' => 'tier', ... }
  #
  def self.chef_historical_tiers_names
    # Return the hash
    historical_tiers(name: true)
  end

  #
  # Load Tiers of historicals based on Chef.
  #
  # == Returns:
  # Hash with format:
  #   { 'node' => TierObject, 'node2' =>  TierObject, ... }
  #
  def self.chef_historical_tiers
    # Return the hash
    historical_tiers
  end

  #
  # Load Tier's nodes based on Chef.
  #
  # == Returns:
  # DruidManager Array
  #
  def chef_tier_nodes
    nodes = []
    Tier.chef_historical_tiers.each do |_, node|
      nodes << node if node.node.redborder.druid.historical.tier == name
    end
    nodes
  end

  private

  def self.historical_tiers(name: false)
    tiers = {}
    DruidManager.historical.each do |chef_node|
      next unless chef_node.node['redborder']['druid'].present? &&
                  chef_node.node['redborder']['druid']['historical'].present?
      tiers[chef_node.name] = (name ? default_tier_name(chef_node.tier) : chef_node)
    end
    tiers
  end

  #
  # Prevent to destroy Default tier
  #
  def check_protected
    return true unless protected?
    errors[:base] << "#{I18n.t('models.default_tiers')}"
    throw(:abort)
  end

  #
  # Apply changes in Chef role/node data.
  #
  def update_nodes
    chef_tier_nodes.each do |node|
      node.update_tier(name, memory)
      node.save
    end
  end

  #
  # Check if the user is a super admin
  #
  def user_permissions
    return true if user && user.super_admin?
    errors[:user] << "#{I18n.t('models.user_admin')}"
    false
  end
end
