#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Define an user in redBorder app
#
class User < ApplicationRecord
  # Include libraries
  include Redborder::Audit
  include Redborder::MenuManager::MenuHelper

  # Include default devise modules. Availables are:
  # :confirmable,
  # :lockable, :timeoutable and :omniauthable,
  # :database_authenticatable, :registerable,
  # :recoverable, :rememberable, :trackable, :validatable

  # Lockable for Users limit login attempts
  devise :lockable, :timeoutable

  # -------------------------------------------------------------
  # Stores
  # -------------------------------------------------------------

  # Store properties of a user to general config
  store :properties, accessors: [:active_data_sources, :ztoken]

  # Store tutorials in the platform
  store :tutorials

  # -------------------------------------------------------------
  # Log and login
  # -------------------------------------------------------------

  audited only: [:name, :email, :admin, :login, :sensor_id, :enabled, :current_sign_in_at, :last_sign_out_at, :encrypted_password, :authentication_token, :failed_attempts, :locked_at],
          on:   [:create, :update, :destroy]

  if RedBorder::CONFIG[:authentication_mode] == 'ldap'
    devise :ldap_authenticatable, :rememberable, :trackable, :registerable, authentication_keys: [:username]
  elsif RedBorder::CONFIG[:authentication_mode] == 'saml'
    devise :saml_authenticatable, :database_authenticatable, :registerable, :rememberable, :trackable, authentication_keys: [:username]
  else
    devise :database_authenticatable, :registerable, :recoverable,
           :rememberable, :trackable, :validatable, :timeoutable,
           authentication_keys: [:username]
  end

  # -------------------------------------------------------------
  # Validationss
  # -------------------------------------------------------------
  validates :name, :login, presence: true
  validates :login, uniqueness: true
  validates :root_page, inclusion: { in: :root_page_collection }
  validates :per_page_count, inclusion: { in: [25, 45, 65, 100] }
  validate :password_complexity
  # -------------------------------------------------------------
  # Callbacks
  # -------------------------------------------------------------
  before_update :change_authentication_token,
                if: proc { |user| user.encrypted_password_changed? }

  before_save :ensure_authentication_token
  after_initialize :initialize_properties
  after_create :load_default_dashboard

  before_update :reset_user_data_sources,
                if: proc { |u| u.sensor_id_changed? }

  DittoCode::Exec.if 'COMMUNITY' do
    before_validation(on: :create) do
      self.root_page = '/sensors'
    end
  end

  cattr_accessor :current_user
  cattr_accessor :user_modules
  attr_accessor :username

  # -------------------------------------------------------------
  # Scope
  # -------------------------------------------------------------

  default_scope { where(enabled: true) }

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------

  # This is the sensor from which the user exists.
  belongs_to :sensor

  has_many :widgets, through: :dashboards
  has_many :notifications, dependent: :destroy
  has_many :alarms, dependent: :destroy
  has_many :saved_filters, dependent: :destroy
  has_many :dashboard_users, dependent: :destroy
  has_many :dashboards_from_others, -> { order(:name) },
           through: :dashboard_users, source: :dashboard
  has_many :editable_dashboards_from_others,
           -> { where(locked: false)
                .where(dashboard_users: { can_edit: true })
                .order(:name) },
           through: :dashboard_users, source: :dashboard
  has_many :dashboards, -> { order :name },
           class_name: 'Dashboard', dependent: :destroy, inverse_of: :owner
  has_many :report_users, dependent: :destroy
  has_many :reports_from_others, -> { order :name },
           through: :report_users, source: :report
  has_many :editable_reports_from_others,
           -> { where(report_users: { can_edit: true }).order(:name) },
           through: :report_users, source: :report
  has_many :reports, -> { order :name },
           class_name: 'Report', dependent: :destroy, inverse_of: :owner
  has_many :correlation_engine_rules

  # Objects
  has_many :objects, class_name: 'RedborderObject', dependent: :destroy
  has_many :net_objects
  has_many :mac_objects
  has_many :vlan_objects

  # Incidents
  has_many :incident_users, dependent: :destroy
  has_many :incidents, through: :incident_users

  has_many :response_actions, dependent: :nullify
  has_many :notes, dependent: :nullify

  # Playbook Templates
  has_many :playbook_templates, dependent: :destroy

  # Playbook Assignment Rules
  has_many :playbook_assignment_rules, dependent: :destroy

  # Return all service / protocol global objects for users
  def protocol_objects
    RedborderObject.where(type: 'ProtocolObject')
  end

  def service_objects
    super_admin? ? ServiceObject.where(user_id: [self, nil]) : ServiceObject.where(user_id: self)
  end

  # -------------------------------------------------------------
  # Methods
  # -------------------------------------------------------------
  # checks when changing a password if it matches the regular expression
  def password_complexity
    minium_password_length = (Setting.find_by(name: 'minimum_password_length').try(:value) || 12).to_i
    return if password.blank? || password =~ /^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.*[ !"#\$%&'*()+,-.\/:;<=>?@\[\]\\^_`{}|~])[a-zA-Z0-9 !"#\$%&'*()+,-.\/.;<=>?@\[\]\\^_`{}|~]{#{minium_password_length},128}$/
    errors.add(:password, "#{I18n.t('models.password_format_error')}, min char: #{minium_password_length} max char: 128")
  end

  # To not lock root admin accounts after too many failed attempts
  def lock_strategy_enabled?(strategy)
    #return false if self.super_admin? and id == 1
    true
  end

  ## Send email to admin that user is locked
  def send_user_locked(user)
    @administrator = User.first
    @user = user
    MyMailer.user_locked(@administrator, @user ).deliver_later
  end

  def unauthenticated_message
    # If set to paranoid mode, do not show the locked message because it
    # leaks the existence of an account.
    if Devise.paranoid
      super
    elsif access_locked? || (lock_strategy_enabled?(:failed_attempts) && attempts_exceeded?)
      send_user_locked(self)
      :locked
    elsif lock_strategy_enabled?(:failed_attempts) && last_attempt? && self.class.last_attempt_warning
      :last_attempt
    else
      super
    end
  end

  def attempts_exceeded?
    return false if self.super_admin? and id == 1
    self.failed_attempts >= self.class.maximum_attempts
  end

  #
  # Return the root page collection available for the user.
  #
  # == Parameters:
  # for_form::
  #   True or false to render the collection as array of strings or an array
  #   to set in a Select.
  #
  # == Returns:
  # Array of elements
  #
  def root_page_collection(for_form = false)
    list = []
    menu_items_for(:user_home_page_collection_menu) do |node|
      list << (for_form ? [node.name, node.caption] : node.caption)
    end
    list
  end

  def self.to_csv(content)
    header = %w(Name Email ID Login Admin
                Enabled Authentication_token Sensor_ID)
    csv = header.join(',') + "\n"

    content.each do |user|
      aux_array = []
      header.each do |head|
        aux_array << user[head.downcase.to_sym]
      end
      csv += aux_array.join(',') + "\n"
    end

    csv
  end

  def self.find_first_by_auth_conditions(warden_conditions)
    conditions = warden_conditions.dup
    username = conditions.delete(:username)
    if username
      unscoped.where(conditions)
        .where(['lower(login) = :value OR lower(email) = :value',
               { value: username.downcase }]).first
    else
      conditions = ActionController::Parameters.new(conditions) unless
        conditions.is_a?(ActionController::Parameters)
      unscoped.where(conditions.permit(:email, :login,
                                       :reset_password_token)).first
    end
  end

  def ensure_authentication_token
    self.authentication_token = recreate_token if authentication_token.blank?
  end

  def reset_authentication_token!
    self.authentication_token = recreate_token
    update(authentication_token: recreate_token)
  end

  if RedBorder::CONFIG[:authentication_mode] == 'ldap'
    def authenticatable_salt
      encrypted_password[0, 29] if encrypted_password
    end
  end

  def as_json(options = {})
    result = super(options)
    result['sensor_name'] = sensor.name if sensor
    result['sensor_uuid'] = sensor.uuid if sensor
    result
  end

  # -------------------------------------------------------------
  # Zendesk
  # -------------------------------------------------------------

  #
  # Initialize and return a zendesk Token
  #
  def zendesk_client
    return nil unless ztoken.present?
    client = ZendeskAPI::Client.new do |c|
      c.access_token = ztoken
      c.retry = true
      c.url = 'https://redborder.zendesk.com/api/v2'
      #c.logger = Logger.new(STDOUT)
    end
    # Check if the token is valid for the current user
    return client if client.current_user
    # Not valid, we must reset it
    update(ztoken: nil)
    nil
  end

  # -------------------------------------------------------------
  # Permissions
  # -------------------------------------------------------------

  #
  # Check if the user is a super admin or super reader
  #
  # == Returns:
  # Boolean indicating if the user is a super user
  #
  def super_user?
    super_admin? || super_reader?
  end

  #
  # A super admin is an user with admin permissions on Root sensor
  #
  # == Returns:
  # True if the user is a super admin
  #
  def super_admin?
    # Cache this method
    return @super_admin unless @super_admin.nil?

    @super_admin = (sensor == Sensor.root && admin)
  end

  #
  # Check if an user is admin, independent of sensor.
  #
  # == Returns:
  # True if the user is admin
  #
  def admin?
    admin
  end

  #
  # A super_reader is an user without admin permissions on Root sensor
  #
  # == Returns:
  # True if the user is a super_reader
  #
  def super_reader?
    # Cache this method
    return @super_reader unless @super_reader.nil?

    @super_reader = (sensor == Sensor.root)
  end

  #
  # Check if an user is admin on a sensor
  #
  # == Parameters:
  # s::
  #   Sensor to check permission
  #
  # == Returns:
  # True if the user is admin
  #
  def admin_on?(s)
    return false unless admin
    # Check if the sensor is descendant of user top domain
    s.is_or_is_descendant_of?(sensor)
  end

  #
  # Check if an user can see a sensor
  #
  # == Parameters:
  # s::
  #   Sensor to check permission
  #
  # == Returns:
  # True if the user is admin
  #
  def read_on?(s)
    s.is_or_is_descendant_of?(sensor)
  end

  #
  # Check permission with CanCan style. Available permissions are :manage
  # and :read
  #
  # == Parameters:
  # permission::
  #   Symbol with the type of permission
  # s::
  #   Sensor to check permission
  #
  # == Returns:
  # True if the user has this permission on the sensor.
  #
  def can?(permission, s)
    return false unless s

    if permission == :manage
      admin_on? s
    else
      read_on? s
    end
  end

  #
  # Negate can?. This method check is an user isn't authorized to read or
  # manage a sensor
  #
  # == Parameters:
  # permission::
  #   Symbol with the type of permission
  # s::
  #   Sensor to check permission
  #
  # == Returns:
  # True if the user has this permission on the sensor.
  #
  def cannot?(permission, s)
    !can?(permission, s)
  end

  # Checks if the user is the first user created
  def first_user?
    self == User.first
  end

  # -------------------------------------------------------------
  # Namespaces
  # -------------------------------------------------------------

  #
  # Return an array with available modules of user. We find this modules based
  # on namespaces in the structure.
  #
  # == Returns:
  # An array of String with the name of modules
  #
  def modules
    # Cache
    return @modules if @modules

    organization = sensor.ancestors_organizations
    # If there isn't any organization, we are on "admin zone", so active modules
    # are determined by license
    if organization.nil? || organization.empty?
      @modules = Redborder::Module.registered_modules.keys.map(&:to_s)
    else
      # Return active modules on organization
      @modules = organization.first.active_modules.split(',')
    end
  end

  #
  # Children namespaces based on user top domain
  #
  # == Returns:
  # ActiveRecord relation, we must check permissions
  #
  def children_namespaces
    sensor.children_namespaces
  end

  #
  # Fetch available namespaces to search.
  #
  # == Returns:
  # Array of Sensors.
  #
  def data_sources
    # Cache
    return @namespaces if @namespaces
    # Get from the sensor
    @namespaces = sensor.namespaces
  end

  #
  # Fetch available namespace ids to search.
  #
  # == Returns:
  # Array of Sensors.
  #
  def data_sources_ids
    data_sources.map(&:uuid)
  end

  #
  # Get active data sources for this user. This method override the store getter
  # to always return an array instead of active_data_sources be nil
  #
  # == Returns:
  # An array of IDs (integers)
  #
  def active_data_sources
    properties[:active_data_sources] || [data_sources.first.uuid]
  end

  #
  # Reset the active data sources of affected user
  #
  def reset_user_data_sources
    properties[:active_data_sources] = data_sources.map(&:uuid)
  end

  # -------------------------------------------------------------
  # Sensors, Domains and Zones
  # -------------------------------------------------------------

  def sensors_with_permission
    @sensors_with_permission ||=
      sensor.self_and_descendants
  end

  def sensors_with_manage_permission
    # None is method that return an empty relation
    return Sensor.none unless admin
    sensors_with_permission
  end

  def floor_with_permission
    @floor_with_permission ||=
      sensor.self_and_descendants.where(domain_type: Sensor::FLOOR)
  end

  def floor_with_manage_permission
    # None is method that return an empty relation
    return Sensor.none unless admin
    floor_with_permission
  end

  def domains_with_permission
    @domains_with_permission ||=
      sensor.self_and_descendants.where(type: Sensor::DOMAIN)
  end

  def domains_with_manage_permission
    # None is method that return an empty relation
    return Sensor.none unless admin
    domains_with_permission
  end

  def locations_with_permission
    @locations_with_permission ||=
      sensor.self_and_descendants.where(type: Sensor::DOMAIN,
                                        domain_type: [Sensor::BUILDING, Sensor::CAMPUS,
                                                      Sensor::FLOOR])
  end

  #
  # Return zones in top domain branch
  #
  def zones_with_permission
    Zone.where(sensor_id: sensor.tree_domains(true).map(&:id))
  end

  #
  # => User root sensor with restricted permissions to create sensors if
  #    the licensing mode is by organization.
  #
  def restricted_organization_root?
    RedBorder::LICENSING_MODE == 'organization' &&
      (sensor.is_root? || sensor.is_domain? &&
       [Sensor::NAMESPACE, Sensor::SERVICE_PROVIDER].include?(sensor.domain_type))
  end

  # -------------------------------------------------------------
  # Reports
  # -------------------------------------------------------------

  def reports_from_sensors
    Report.includes(:report_sensors).where(
      report_sensors: {
        sensor_id: domains_with_permission.map(&:id)
      }
    ).where.not(owner: self).order(:name)
  end

  def audit_sign_in
    self.current_sign_in_at = Time.now
    save(:validate => false)
  end

  def audit_last_sign_out_at
    self.last_sign_out_at = Time.now
    save(:validate => false)
  end

  def editable_reports_from_sensors
    all = Report.includes(:report_sensors).where(
      report_sensors: {
        sensor_id: domains_with_permission.map(&:id), can_edit: :all
      }
    )

    admin = Report.includes(:report_sensors).where(
      report_sensors: {
        sensor_id: domains_with_manage_permission.map(&:id), can_edit: :admin
      }
    )

    (all + admin).uniq.sort_by(&:name)
  end

  #
  # => Return all the reports the user can read
  #
  def readable_reports
    (reports + reports_from_others + reports_from_sensors).uniq.sort_by(&:name)
  end

  #
  # => Return all the reports the user can manage
  #
  def writable_reports
    (reports + editable_reports_from_others + editable_reports_from_sensors).uniq.sort_by(&:name)
  end

  # -------------------------------------------------------------
  # Dashboards
  # -------------------------------------------------------------

  def dashboards_from_sensors
    return @dashboards_from_sensor if @dashboards_from_sensor

    @dashboards_from_sensor = Dashboard.includes(:dashboard_sensors).where(
      dashboard_sensors: {
        sensor_id: domains_with_permission.map(&:id)
      }
    ).where.not(owner: self).order(:name)
  end

  def editable_dashboards_from_sensors
    all = Dashboard.includes(:dashboard_sensors).where(
      dashboard_sensors: {
        sensor_id: domains_with_permission.map(&:id), can_edit: :all
      }
    )

    admin = Dashboard.includes(:dashboard_sensors).where(
      dashboard_sensors: {
        sensor_id: domains_with_manage_permission.map(&:id), can_edit: :admin
      }
    )

    (all + admin).uniq.sort_by(&:name)
  end

  def readable_dashboards
    (dashboards + dashboards_from_others + dashboards_from_sensors).uniq.sort_by(&:name)
  end

  def writable_dashboards
    (dashboards + editable_dashboards_from_others + editable_dashboards_from_sensors).uniq.sort_by(&:name)
  end

  #
  # => Return all the dashboards shared by others unless the ones which
  #    have been shared by sensors with highest privileges.
  #
  def dashboards_from_others_filtered
    dashboards_from_others.reject { |d| editable_dashboards_from_sensors.include?(d) &&
                                        !editable_dashboards_from_others.include?(d) }
  end

  #
  # => Return all the dashboards shared by domains unless the ones which
  #    have been shared by others with equal or highest privileges.
  #
  def dashboards_from_sensors_filtered
    dashboards_from_sensors
      .reject { |d| editable_dashboards_from_others.include?(d) ||
                    dashboards_from_others.include?(d) &&
                    !editable_dashboards_from_sensors.include?(d)
              }
  end

  #
  # => Get the users which name starts with the letter
  #
  # => @param letter the letter
  #
  def self.find_by_first_letter(letter)
    User.where('LOWER(name) LIKE ?', "#{letter.downcase}%").order('name ASC')
  end

  #
  # Return users current user can share dashboard with. IN cloud, we must
  # return this users based on OrganizationUser.
  #
  # == Returns:
  # ActiveRelation with users.
  #
  def users_to_share
    return User.all.order('LOWER(name) ASC') - [self] if super_admin?
    users = User.where(sensor_id: domains_with_permission.map(&:id))
            .order('LOWER(name) ASC')
    users - [current_user]
  end

  #
  # Return users current user can assign to an incident. IN cloud, we must
  # return this users based on OrganizationUser.
  #
  # == Returns:
  # ActiveRelation with users.
  #
  def users_to_assign
    return User.all.order('LOWER(name) ASC') if super_admin?
    users = User.where(sensor_id: domains_with_permission.map(&:id))
            .order('LOWER(name) ASC')
    users
  end

  def send_creation_instructions(password = nil)
    MyMailer.creation_instructions(self, password).deliver_later
  end

  # Overriding method from trackable module
  def update_tracked_fields!(request)
    if self.persisted?
      old_current = current_sign_in_at
      new_current = Time.now.utc
      self.last_sign_in_at = old_current || new_current
      self.current_sign_in_at = new_current

      old_current = current_sign_in_ip
      new_current = (request.headers['X-Forwarded-For'] ||
                     request.headers['X-Real-IP'] ||
                     request.remote_ip).split(',').first
      self.last_sign_in_ip    = old_current || new_current
      self.current_sign_in_ip = new_current

      self.sign_in_count ||= 0
      self.sign_in_count += 1
      #self.remember_created_at = DateTime.parse(@user.created_at).to_time
      save(:validate => false) or raise "#{I18n.t('models.devise')} #{inspect}." \
        "#{I18n.t('models.make_sure')}"
    end
  end

  def audit_comment_on_update(audit)
    object = audit.auditable_type.constantize.where(id: audit.auditable_id)
    name = object.first.try(:name)

    comment = "#{I18n.t 'lib_redborder.updated'} #{name}:"
    comment += audit.audited_changes.collect do |key, values|

      case key.to_sym
      when :current_sign_in_at
        " Sign In"
      when :last_sign_out_at
        " Sign Out"
      when :encrypted_password
        " Encrypted password updated"
      when :authentication_token
        " Authentication Token updated"
      else
        " #{key} #{I18n.t 'lib_redborder.from'} '#{values[0]}' #{I18n.t 'lib_redborder.to'} '#{values[1]}'"
      end

    end.join(', ')
    comment
  end

  # TODO: Is this method used?
  def permission_for_sensor(sensor)
    sensor.permissions_from_parent.where(user: self).first
  end

  #
  #
  #
  def oldest_timestamp_allowed
    return 0 if super_user?

    organization = sensor.self_and_ancestors
                   .find_by(type: Sensor::DOMAIN, domain_type: Sensor::ORGANIZATION)

    return 0 if organization.nil? || organization.visualization_range.nil? ||
                organization.visualization_range.to_i.zero?

    (Time.now - organization.visualization_range.to_i.days).to_i
  end

  def valid_password?(password)
    if RedBorder::CONFIG[:authentication_mode] == 'ldap'
      valid_ldap_authentication?(password)
    else
      super(password)
    end
  end

  def basic_info_structured_data
    sd_id = "user-basic-info@#{RedBorder::PEN}"
    fields = {
      "user-login" => self.login,
      "user-name" => self.name,
      "user-email" => self.email
    }

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end
  
  protected

  def recreate_token
    loop do
      token = SecureRandom.urlsafe_base64(30).tr('lIO0', 'sxyz')
      break token unless User.where(authentication_token: token).first
    end
  end

  def change_authentication_token
    self.authentication_token = recreate_token
  end

  def initialize_properties
    self.properties ||= {}
  end

  def load_default_dashboard
    Delayed::Job.enqueue(
      Redborder::Jobs::DefaultDashboardsJob.new(true, self),
      priority: 5, run_at: Time.now
    )
  end

  def timeout_in
    (Setting.find_by(name: 'session_timeout_time').try(:value) || 1800).to_i.seconds
  end
end
