#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class WidgetGraph::Base
  include ActiveModel::Serializers::JSON

  GRANULARITY_TYPE = {
    minute:         { long: '1 Minute',   short: '1m',  step: 60_000        },
    pt2m:           { long: '2 Minutes',  short: '2m',  step: 120_000       },
    pt5m:           { long: '5 Minutes',  short: '5m',  step: 300_000       },
    pt15m:          { long: '15 Minutes', short: '15m', step: 900_000       },
    fifteen_minute: { long: '15 Minutes', short: '15m', step: 900_000       },
    pt30m:          { long: '30 Minutes', short: '30m', step: 1_800_000     },
    thirty_minute:  { long: '30 Minutes', short: '30m', step: 1_800_000     },
    hour:           { long: '1 Hour',     short: '1h',  step: 3_600_000     },
    pt2h:           { long: '2 Hours',    short: '2h',  step: 7_200_000     },
    pt8h:           { long: '8 Hours',    short: '8h',  step: 28_800_000    },
    pt24h:          { long: '1 Day',      short: '1d',  step: 86_400_000    },
    pt120h:         { long: '5 Days',     short: '5d',  step: 432_000_000   },
    pt168h:         { long: '1 Week',     short: '1w',  step: 604_800_000   },
    pt720h:         { long: '30 Days',    short: '30d', step: 2_592_000_000 },
    p1m:            { long: '30 Days',    short: '30d', step: 2_592_000_000 }
  }

  DEFAULT_GRANULARITY = {
    last_1_minute: :minute,
    last_2_minutes: :minute,
    last_5_minute: :minute,
    last_15_minute: :minute,
    last_30_minute: :pt2m,
    last_1_hour: :pt2m,
    last_2_hour: :pt5m,
    last_3_hour: :pt5m,
    last_12_hour: :pt2m,
    last_24: :fifteen_minute,
    last_week: :pt2h
  }

  mattr_accessor :types
  private_class_method :types=
  self.types = {}

  cattr_accessor :display_name, :image

  def initialize(widget)
    @widget = widget
  end

  def attributes=(hash)
    hash.each do |key, value|
      send("#{key}=", value)
    end
  end

  def attributes
    instance_values
  end

  def as_json(opts = {})
    obj_opts = opts.dup
    obj_opts[:except] ||= []
    obj_opts[:except] = Array[obj_opts[:except]].flatten(1)
    obj_opts[:except] << :widget
    super(obj_opts)
  end

  def data
    types[self.class.to_s.demodulize.underscore.to_sym][:data]
  end

  def self.inherited(subclass)
    subclass_name_underscored = subclass.name.demodulize.underscore.to_sym

    types[subclass_name_underscored] = {
      class: subclass,
      name: subclass_name_underscored.to_s.titleize,
      image: 'area.png',
      order: 999
    }
  end

  def self.name_underscored
    name.demodulize.underscore.to_sym
  end

  def self.display_name(name)
    types[name_underscored][:name] = name
  end

  def self.image(image)
    types[name_underscored][:image] = image
  end

  def self.order(order)
    types[name_underscored][:order] = order
  end

  def self.data(data)
    types[name_underscored][:data] = data
  end

  def self.widget_inject_into(types)
    types.each do |type|
      eval("WidgetType::#{type.classify}").graphs [name_underscored]
    end
  end

  def self.widget_validate(name, &block)
    type_name_underscored = name_underscored
    Widget.class_eval do
      validate name.to_sym

      define_method(name.to_sym) do
        return unless type_name_underscored.to_s == graph_type.to_s
        instance_eval(&block)
      end
    end
  end

  def self.widget_validates(*name, opts)
    type_name_underscored = name_underscored
    mopts = opts.merge(if: proc { |w| w.graph_type.to_s == type_name_underscored.to_s })

    Widget.class_eval do
      validates(*name, mopts)
    end
  end

  def self.widget_define(name, &block)
    Widget.class_eval do
      define_method(name.to_sym) do |*params|
        instance_exec(*params, &block)
      end
    end
  end

  def self.widget_eval(&block)
    Widget.class_eval(&block)
  end

  def self.widget_fields(names = [])
    delegate(*names, to: :@widget)

    Widget.class_eval do
      add_properties names

      names.each do |method_name|
        define_method(method_name) do
          property[method_name]
        end

        define_method("#{method_name}=") do |value|
          property[method_name] = value
        end
      end
    end
  end

  def self.widget_array_fields(names = [])
    delegate(*names, to: :@widget)

    Widget.class_eval do
      add_array_properties names

      names.each do |method_name|
        define_method(method_name) do
          property[method_name]
        end

        define_method("#{method_name}=") do |value|
          property[method_name] = value
        end
      end
    end
  end

  def self.widget_after_initialize(&block)
    method_name = (name_underscored.to_s + '_graph_after_initializer').to_sym
    type_name_underscored = name_underscored

    Widget.class_eval do
      after_initialize method_name

      define_method(method_name) do
        return unless graph_type.to_sym == type_name_underscored.to_sym
        instance_eval(&block)
      end
    end
  end

  def self.widget_before_save(&block)
    method_name = (name_underscored.to_s + '_graph_before_save').to_sym
    type_name_underscored = name_underscored

    Widget.class_eval do
      before_save method_name

      define_method(method_name) do
        return unless graph_type.to_sym == type_name_underscored.to_sym
        instance_eval(&block)
      end
    end
  end
end
