#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class WidgetType::Base
  include ActiveModel::Serializers::JSON

  TIME_RANGES = {
    last_1_minute: 1.minute,
    last_2_minutes: 2.minutes,
    last_5_minute: 5.minute,
    last_15_minute: 15.minute,
    last_30_minute: 30.minute,
    last_1_hour: 1.hour,
    last_2_hour: 2.hour,
    last_3_hour: 3.hour,
    last_12_hour: 12.hour,
    last_24: 24.hour,
    last_week: 1.week
  }

  attr_writer :now
  mattr_accessor :types
  private_class_method :types=
  self.types = {}

  def initialize(widget)
    @widget = widget
  end

  def attributes=(hash)
    hash.each do |key, value|
      send("#{key}=", value)
    end
  end

  def in_time_machine(&block)
    if User.current_user.nil? || User.current_user.time_machine_now.nil?
      yield
    else
      Timecop.travel(Time.at(User.current_user.time_machine_now), &block)
    end
  end

  def attributes
    instance_values
  end

  def as_json(opts = {})
    obj_opts = opts.dup
    obj_opts[:except] ||= []
    obj_opts[:except] = Array[obj_opts[:except]].flatten(1)
    obj_opts[:except] << :widget
    obj_opts[:methods] ||= []
    obj_opts[:methods] = Array[obj_opts[:methods]].flatten(1)
    obj_opts[:methods] += [
      :start_time, :end_time,
      :start_time_secs, :end_time_secs
    ]
    super(obj_opts)
  end

  def start_time
    if @widget.filtrable? && @widget.fetch_filters.key?(:start_time)
      Time.at(@widget.fetch_filters[:start_time].keys.first)
    else
      range = @widget.range || 'last_1_hour'
      time_range = TIME_RANGES.fetch(range.to_sym) { 1.hour }
      result = end_time - time_range
      result = result.beginning_of_hour if time_range >= 12.hours
      round_time result
    end
  end

  def end_time
    if @widget.filtrable? && @widget.fetch_filters.key?(:end_time)
      Time.at(@widget.fetch_filters[:end_time].keys.first)
    else
      round_time(Timecop.travel(now) { Time.now })
    end
  end

  def start_time_secs
    start_time.to_i
  end

  def end_time_secs
    end_time.to_i
  end

  def now
    @now || Time.now
  end

  def self.name_underscored
    name.demodulize.underscore.to_sym
  end

  def self.inherited(subclass)
    subclass_name_underscored = subclass.name.demodulize.underscore.to_sym

    types[subclass_name_underscored] = {
      class: subclass,
      filters: true,
      name: subclass_name_underscored.to_s.titleize,
      graphs: []
    }
  end

  def self.graphs(graphs)
    types[name_underscored][:graphs] += graphs
  end

  def self.filters(state)
    types[name_underscored][:filters] = state
  end

  def self.create_default(create = false)
    types[name_underscored][:create_default] = create
  end

  def self.url(url)
    types[name_underscored][:url] = url
  end

  def self.widget_validate(name, &block)
    type_name_underscored = name_underscored
    Widget.class_eval do
      validate name.to_sym, if: proc { |w| w.widget_type.to_sym == type_name_underscored }

      define_method(name.to_sym) do
        instance_eval(&block)
      end
    end
  end

  def self.widget_validates(*name, opts)
    type_name_underscored = name_underscored
    mopts = opts.merge(if: proc { |w| w.graph_type == type_name_underscored })

    Widget.class_eval do
      validates(*name, mopts)
    end
  end

  def self.widget_define(name, &block)
    Widget.class_eval do
      define_method(name.to_sym) do |*params|
        instance_exec(*params, &block)
      end
    end
  end

  def self.widget_eval(&block)
    Widget.class_eval(&block)
  end

  def self.widget_fields(names = [])
    delegate(*names, to: :@widget)

    Widget.class_eval do
      add_properties names

      names.each do |method_name|
        define_method(method_name) do
          property[method_name]
        end

        define_method("#{method_name}=") do |value|
          property[method_name] = value
        end
      end
    end
  end

  def self.widget_array_fields(names = [])
    delegate(*names, to: :@widget)

    Widget.class_eval do
      add_array_properties names

      names.each do |method_name|
        define_method(method_name) do
          property[method_name]
        end

        define_method("#{method_name}=") do |value|
          property[method_name] = value
        end
      end
    end
  end

  def self.widget_after_initialize(&block)
    method_name = (name_underscored.to_s + '_type_after_initializer').to_sym
    type_name_underscored = name_underscored

    Widget.class_eval do
      after_initialize method_name

      define_method(method_name) do
        return unless widget_type.to_sym == type_name_underscored.to_sym
        instance_eval(&block)
      end
    end
  end

  def self.widget_before_save(&block)
    method_name = (name_underscored.to_s + '_type_before_save').to_sym
    type_name_underscored = name_underscored

    Widget.class_eval do
      before_save method_name

      define_method(method_name) do
        return unless widget_type.to_sym == type_name_underscored.to_sym
        instance_eval(&block)
      end
    end
  end

  protected

  #
  # Create an Array with a Hash inside based on the data returned of a serie
  #
  def format_serie(name, data, s_name)
    [
      {
        name: name,
        data: data,
        secondary_name: s_name
      }
    ]
  end

  #
  # Get the secondary name based on a source
  #
  def find_secondary_name(source, column)
    if source::COLUMNS[column.to_sym][:complementary_columns] &&
        column.to_sym != :sensor_name
      source::COLUMNS[column.to_sym][:complementary_columns].keys.first.to_s
    else
      column
    end
  end

  #
  # Get the secondary name for cardinality based on a source
  #
  def find_secondary_unique_name(source, group_by)
    if source::COLUMNS[group_by.to_sym][:complementary_columns] &&
        group_by.to_sym != :sensor_name
      serie.send(source::COLUMNS[group_by.to_sym][:complementary_columns].keys.first.to_s)
    else
      serie.send(group_by)
    end
  end

  private

  def round_time(time)
    return time unless @widget.granularity
    return time if @widget.graph.nil? || !@widget.graph.respond_to?(:step)

    roundto = @widget.graph.step / 1000
    Time.at((time.to_f / roundto).floor * roundto)
  end

end
