#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class WidgetType::Compare < WidgetType::Base
  graphs [:multiserie]

  COMPARATIVE_COLUMNS = {
    pt24h:   { table: :hour,   graph: :minute },
    pt168h:  { table: :pt24h,  graph: :pt5m   },
    pt720h:  { table: :pt168h, graph: :hour   },
    pt4320h: { table: :p1m,    graph: :pt8h   }
  }

  widget_fields [:range, :elements, :aggregation]

  widget_after_initialize do
    self.elements ||= 5
  end

  widget_before_save do
    self.granularity = COMPARATIVE_COLUMNS[range.to_sym][:graph]
  end

  def as_json(opts = {})
    obj_opts = { methods: [:domain_time] }
    super(obj_opts.merge(opts))
  end

  def start_time
    case range.to_sym
    when :pt24h
      now.beginning_of_day
    when :pt168h
      now.beginning_of_day
    when :pt720h
      now.beginning_of_week
    when :pt4320h
      now.beginning_of_month
    end
  end

  def end_time
    in_time_machine { start_time } + domain_time
  end

  def domain_time
    case range.to_sym
    when :pt24h
      1.hour
    when :pt168h
      1.day
    when :pt720h
      1.week
    when :pt4320h
      1.month
    end
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def list(user = @widget.parent.owner, options = {})
    # Required by multiserie but not used on comparative
    {}
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def total_timeserie(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_start_time = options[:start_time] || start_time
    ovrr_end_time = options[:end_time] || end_time
    ovrr_granularity = options[:granularity] || @widget.granularity

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)

    serie = source.timeserie(
      user: user,
      include_all_namespaces: options[:include_all_namespaces] || false,
      range: @widget.range,
      start_time: ovrr_start_time,
      end_time: ovrr_end_time,
      granularity: ovrr_granularity,
      aggregation: @widget.aggregation
    )

    { total_timeserie: serie }
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def timeseries(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_granularity = options[:granularity] || @widget.granularity

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)

    # Init timeseries
    timeseries = {}
    search = @widget.fetch_filters
    # Hours range
    if @widget.range.to_sym == :pt24h
      # The intervals are different for pt24h
      ovrr_start_time = options[:start_time] || start_time
      ovrr_end_time = options[:end_time] || ovrr_start_time + domain_time

      elements.to_i.times do |i|
        # Get the start and end times
        end_t_comp = ovrr_end_time - i * 1.hour
        start_t_comp = ovrr_start_time - i * 1.hour

        serie = source.where(
          query_type: :timeseries,
          user: user,
          include_all_namespaces: options[:include_all_namespaces] || false,
          range: :custom,
          start_time: start_t_comp,
          end_time: end_t_comp,
          granularity: ovrr_granularity,
          aggregation: @widget.aggregation,
          search: search
        ).map do |t|
          timestamp = (t.timestamp - t.timestamp.hour.hours)
                      .change(
                        day: in_time_machine { Time.zone.now.day },
                        month: in_time_machine { Time.zone.now.month },
                        year: in_time_machine { Time.zone.now.year }
                      ).to_s
          { event: { events: t.send(@widget.aggregation) }, timestamp: timestamp }
        end

        timeseries.merge!(start_t_comp.to_i => serie)
      end
    else
      elements.to_i.times do |i|
        # Get the start and end times
        ovrr_start_time = options[:start_time] || start_time
        ovrr_end_time = options[:end_time] || end_time

        end_t_comp = ovrr_end_time - i * domain_time
        start_t_comp = ovrr_start_time - i * domain_time

        serie = source.where(
          query_type: :timeseries,
          user: user,
          include_all_namespaces: options[:include_all_namespaces] || false,
          range: :custom,
          start_time: start_t_comp,
          end_time: end_t_comp,
          granularity: ovrr_granularity,
          aggregation: @widget.aggregation,
          search: search
        )

        case @widget.range.to_sym
        when :pt168h
          serie = serie.map { |t| {event: {events: t.send(@widget.aggregation)},
                                   timestamp: (t.timestamp)
                                   .change(
                                      day: in_time_machine { Time.zone.now.day },
                                      month: in_time_machine { Time.zone.now.month },
                                      year: in_time_machine { Time.zone.now.year }
                                   ).to_s}}
        when :pt720h, :pt4320h
          serie = serie.map{|t| {event: {events: t.send(@widget.aggregation)},
                                 timestamp: (t.timestamp)
                                 .change(
                                    month: in_time_machine { Time.zone.now.month },
                                    year: in_time_machine { Time.zone.now.year }
                                 ).to_s}}
        else
          serie = serie.map{|t| {event: {events: t.send(@widget.aggregation)},
                                 timestamp: (t.timestamp).to_s}}
        end

        timeseries.merge!(start_t_comp.to_i => serie)
      end
    end

    { timeseries: timeseries }
  end
end
