#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class WidgetType::Top < WidgetType::Base
  graphs [:multiserie, :pie, :horizontal_bar, :vertical_bar, :columnle, :table, :box]

  widget_fields [:range, :elements, :aggregation, :column, :client_profile]

  widget_after_initialize do
    self.elements ||= 5
  end

  # Validate the number of elements
  widget_validate :elements_num do
    self.elements = 2 if elements.to_i < 2
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def list(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_start_time = options[:start_time] || start_time
    ovrr_end_time = options[:end_time] || end_time

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)

    serie = source.where(
      query_type: :group_by,
      range: @widget.range,
      user: user,
      include_all_namespaces: options[:include_all_namespaces] || false,
      start_time: ovrr_start_time,
      end_time: ovrr_end_time,
      granularity: :all,
      limit: @widget.elements,
      sort: @widget.aggregation,
      group_by: [source::COLUMNS[column.to_sym][:columns].keys[0]],
      aggregation: @widget.aggregation,
      search: @widget.fetch_filters
    )

    # Check complementary columns
    c_columns = source::COLUMNS[column.to_sym][:complementary_columns]

    {
      list: serie.map { |s| { column.to_sym => s.send(column),
                              @widget.aggregation.to_sym => s.send(@widget.aggregation),
                              secondary_name:  (c_columns and column.to_sym != :sensor_name) ?
                                                    s.send(c_columns.keys.first.to_s) : s.send(column)
                            }
                      }
    }
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def timeseries(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_start_time = options[:start_time] || start_time
    ovrr_end_time = options[:end_time] || end_time
    ovrr_granularity = options[:granularity] || @widget.granularity

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)
    elements = list(user, options)
    # Init timeseries
    timeseries = {}
    search = @widget.fetch_filters

    elements[:list].each do |e|
      value = e[@widget.column.to_sym]
      search[@widget.column.to_sym] = { value => { operator: :is } }

      serie = source.timeserie(
        user: user,
        include_all_namespaces: options[:include_all_namespaces] || false,
        range: @widget.range,
        start_time: ovrr_start_time,
        end_time: ovrr_end_time,
        granularity: ovrr_granularity,
        aggregation: @widget.aggregation,
        search: search
      )

      if value.nil?
        timeseries.merge!('N/A' => serie)
      else
        timeseries.merge!(value => serie)
      end
    end

    # [ TODO ] Angel. Add filters
    {
      timeseries: timeseries
    }
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # An array or value with the data.
  #
  def total_timeserie(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_start_time = options[:start_time] || start_time
    ovrr_end_time = options[:end_time] || end_time
    ovrr_granularity = options[:granularity] || @widget.granularity

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)

    serie = source.timeserie(
      user: user,
      include_all_namespaces: options[:include_all_namespaces] || false,
      range: @widget.range,
      start_time: ovrr_start_time,
      end_time: ovrr_end_time,
      granularity: ovrr_granularity,
      aggregation: @widget.aggregation
    )

    { total_timeserie: serie }
  end

  #
  # Preload the data to show the widget. This method is used to generate
  # reports because WickedPDF can't call to the server when render.
  #
  # == Parameters:
  # user::
  #   The user who call to render the report. This is used to set permissions
  # options::
  #   A hash with multiple options to force on DruidQueries.
  #     - start_time
  #     - end_time
  #     - granularity
  #
  # == Returns:
  # A value with the total data of a topic in a range
  #
  def total(user = @widget.parent.owner, options = {})
    # Override the report block time if it's defined in the option
    ovrr_start_time = options[:start_time] || start_time
    ovrr_end_time = options[:end_time] || end_time

    # Use DruidSource to set the dataSource based on the product_type
    source = Redborder::DruidSource.get(source: @widget.product_type)

    total = source.timeserie(
      user: user,
      include_all_namespaces: options[:include_all_namespaces] || false,
      range: @widget.range,
      start_time: ovrr_start_time,
      end_time: ovrr_end_time,
      granularity: :all,
      aggregation: @widget.aggregation,
      search: @widget.fetch_filters
    )

    # [ TODO ] Angel. Add filters
    {
      total: total
    }
  end

  #
  # Fetch the color of severities. These no needs permissions.
  #
  # == Returns:
  # An array with severities
  #
  def colors
    {
      colors: Severity.all
    }
  end
end
