#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

include IncidentsHelper

module Incidents
  class IncidentDruidRetriever
    def self.get_data(incident, options = {}, user)
      elements = get_incident_druid_data(incident, options, user)
      elements
    end
  
    def self.get_incident_druid_data(incident, search_params, current_user)
      data = []
    
      # Mapping of fields that can change depending on the module or data source
      field_map = {
        priority: [:priority, :syslogseverity_text], # Priority can be one of these fields
        msg: [:msg, :message],                       # Msg can be one of these fields
      }
  
      @objects = RedborderObject.where("user_id = ? OR user_id IS NULL", current_user.id)
  
      incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
  
      incident_modules.each do |mod| 
        source = Redborder::DruidSource.get(source: mod.first)
        filter = Redborder::Filter.new(source, {})
        filter.add(:incident_uuid, incident.uuid)
  
        start_time = incident.created_at - 1.day
        end_time = incident.created_at + 1.day
  
        # group_by: @source.default_columns_raw,
        data += source.where({
          query_type: :group_by,
          range: :custom,
          start_time: start_time,
          end_time: end_time,
          granularity: :minute,
          aggregation: source.default_aggregation,
          group_by: source.default_columns_incidents,
          search: filter.get,
          user: current_user
        }).map do |t|
          url_filters = {}
          url_filters[:start_time] = start_time
          url_filters[:end_time] = end_time
  
          sig_id = 'N/A'
          if t.methods.include?(:sig_id)
            sig_id = t.send(:sig_id) || 'N/A'
          end

          classification = ''
          if t.methods.include?(:classification)
            classification = t.send(:classification) || ''
          end

          src = 'N/A'
          if t.methods.include?(:src)
            src = t.send(:src) || 'N/A'
          end

          src_port = 'N/A'
          if t.methods.include?(:src_port)
            src_port = t.send(:src_port) || 'N/A'
          end

          ethsrc = 'N/A'
          if t.methods.include?(:ethsrc)
            ethsrc = t.send(:ethsrc) || 'N/A'
          end

          dst = 'N/A'
          if t.methods.include?(:dst)
            dst = t.send(:dst) || 'N/A'
          end

          dst_port = 'N/A'
          if t.methods.include?(:dst_port)
            dst_port = t.send(:dst_port) || 'N/A'
          end

          ethdst = 'N/A'
          if t.methods.include?(:ethdst)
            ethdst = t.send(:ethdst) || 'N/A'
          end

          sensor_name = 'N/A'
          if t.methods.include?(:sensor_name)
            sensor_name = t.send(:sensor_name) || 'N/A'
          end

          sensor_uuid = 'N/A'
          if t.methods.include?(:sensor_uuid)
            sensor_uuid = t.send(:sensor_uuid) || 'N/A'
          end

          # We use the map_dynamic_fields method to map the fields correctly
          priority = map_dynamic_fields(t, priority: field_map[:priority])
          msg = map_dynamic_fields(t, msg: field_map[:msg])

          incident_source_name = mod.last[:incident_source_name] rescue "N/A"

          module_name = mod.first

          timestamp = t.send(:timestamp)

          # Observables are define in the columns of each module
          observables = []
          assets = []

          mod.last[:columns].select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
            url_filters[column] = t.send(column)

            next unless t.methods.include?(column) && t.send(column)

            mac_object = nil
            if cfg[:object_type] == "MacObject"
              mac_object = IncidentDruidRetriever.find_mac_object(current_user, t.send(column), cfg[:object_type])
            end

            if mac_object
              assets.push(mac_object.to_asset)
            else              
              observables.push( { t.send(column) => cfg[:observable_type] })
            end
          end

          # In case of vault alarm events
          # We try to extract observables/assets from the conditions of the alarm
          # Based on the product_type of the alarm
          # TODO: Refactor this thing
          if mod.first.to_s == "vault"

            alarm_product_type = nil
            if t.methods.include?(:alarm_product_type)
              alarm_product_type = t.send(:alarm_product_type) || nil
            end

            alarm_condition = nil
            if t.methods.include?(:alarm_condition)
              alarm_condition = t.send(:alarm_condition) || 'N/A'
            end

            columns = []
            if alarm_condition && alarm_product_type
              columns = IncidentDruidRetriever.get_columns_by_product_type(alarm_product_type)
            end

            unless columns.empty?
              alarm_condition_json = nil
              begin
                alarm_condition_json = JSON.parse(alarm_condition)
              rescue JSON::ParserError => e
                logger.error("Failed to parse alarm_condition JSON: #{e.message}")
                condition_json = nil
              end

              if alarm_condition_json
                condition_with_is_operator = alarm_condition_json.values.select { |item| item["operator"] == "is" }.map { |item| { "column" => item["column"], "value" => item["value"] } }

                columns.select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
                  # For each "is" condition
                  condition_with_is_operator.each do |c|
                    next unless column.to_s == c['column']

                    mac_object = nil
                    if cfg[:object_type] == "MacObject"
                      mac_object = IncidentDruidRetriever.find_mac_object(current_user, c['value'], cfg[:object_type])
                    end

                    if mac_object
                      assets.push(mac_object.to_asset)
                    else
                      observables.push( { c['value'] => cfg[:observable_type] })
                    end
                  end
                end
              end
            end
          end

          url_filters["incident_uuid"] = incident.uuid

          { timestamp: timestamp, priority: priority, msg: msg, incident_source_name: incident_source_name,
            assets: assets, observables: observables, module_name: module_name,
            url_filters: url_filters, sig_id: sig_id, classification: classification,
            src: src, src_port: src_port, dst: dst, dst_port: dst_port,
            sensor_name: sensor_name, sensor_uuid: sensor_uuid, ethsrc: ethsrc, ethdst: ethdst }
        end
      end
      
      # Filter by search
      if search_params && search_params[:source]
        data = data.select { |element| element[:incident_source_name] == search_params[:source] }
      end
  
      if search_params && search_params[:priority]
        data = data.select { |element| element[:priority] == search_params[:priority] }
      end
  
      data
    end

    def self.find_mac_object(current_user, value, object_type)
      RedborderObject.where("user_id = ? OR user_id IS NULL", current_user.id)
                     .where("value = ? AND type = ?", value, object_type)
                     .first
    end

    def self.get_columns_by_product_type(product_type)
      incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }

      return [] unless incident_modules.include? product_type.to_sym

      return incident_modules[product_type.to_sym][:columns]
    end
  end
end