#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
module Integrations
  class BaseStrategy
    def execute(action, params)
      raise NotImplementedError, "This method should be implemented by subclasses"
    end

    def actions
      raise NotImplementedError, "This method should be implemented by subclasses"
    end

    def self.strategy_name
      const_get(:NAME)
    end

    def self.action_params(action)
      const_get(:AVAILABLE_ACTIONS)[action]
    end
  end

  class OtrsStrategy < BaseStrategy
    # ----------------------------------------------
    # API Doc: https://otrscommunityedition.com/doc/api/otrs/6.0/Perl/index.html
    # ----------------------------------------------

    include HTTParty

    # ==========================================================
    # Constants
    # ==========================================================
    NAME = "otrs"
    AVAILABLE_ACTIONS = {
      "create_ticket" => {
        ticket_title: { required: true, type: :string },
        customer_user: { required: true, type: :string },
        queue_id: { required: false, type: :integer },
        priority_id: { required: false, type: :integer },
        state_id: { required: false, type: :integer },
        article_subject: { required: true, type: :string },
        article_body: { required: true, type: :string },
      },
    }

    def execute(action, params)
      if self.class.action_params(action)
        send(action, params)
      else
        { error: "Unknown action: #{action}" }
      end
    end

    def actions
      self.class.const_get(:AVAILABLE_ACTIONS).keys
    end

    private

    # Create ticket POST method via OTRS API
    def create_ticket(params)
      key_post = Integration.find_by(name: "otrs_key_post")
      web_service = Integration.find_by(name: "otrs_web_service")
      create_ticket_route_mapping = Integration.find_by(name: "create_ticket_route_mapping")
      username = Integration.find_by(name: "otrs_username")
      password = Integration.find_by(name: "otrs_password")

      unless key_post && web_service && username && password && create_ticket_route_mapping
        return { error: "OTRS create ticket API endpoint not found." }
      end

      base_url = "#{key_post.value}/otrs/nph-genericinterface.pl/Webservice/#{web_service.value}/#{create_ticket_route_mapping.value}?UserLogin=#{username.value}&Password=#{password.value}"

      ticket_params = {
        Ticket: {
          Title: params[:ticket_title],
          QueueID: params[:queue_id] || 3,
          PriorityID: params[:priority_id] || 3,
          StateID: params[:state_id] || 1,
          CustomerUser: params[:customer_user],
        },
        Article: {
          Subject: params[:article_subject],
          Body: params[:article_body],
          ContentType: "text/plain; charset=utf8",
        },
      }

      response = nil

      begin
        response = HTTParty.post(
          base_url,
          headers: { "Content-Type" => "application/json" },
          body: ticket_params.to_json,
        )
      rescue StandardError => e
        return { error: "HTTP request failed", message: e.message }
      end

      response.parsed_response
    end
  end

  class TrendMicroStrategy < BaseStrategy
    # ----------------------------------------------
    # API Doc: https://automation.deepsecurity.trendmicro.com/article/20_0/api-reference
    # ----------------------------------------------
    include HTTParty

    # ==========================================================
    # Constants
    # ==========================================================
    NAME = "trend_micro"
    AVAILABLE_ACTIONS = {
      "list_all_computers" => {},
      "list_firewall_rules_by_computer" => {
        computer_id: { required: true, type: :integer },
      },
      "add_ip_to_firewall" => {
        computer_id: { required: true, type: :integer },
        firewall_id: { required: true, type: :integer },
        ip_list: { request: true, type: :array, ip_list: { type: :string } },
      },
    }
    BASE_URL = Integration.find_by(name: "trend_micro_url")
    API_KEY = Integration.find_by(name: "trend_micro_api_key")

    def execute(action, params)
      if AVAILABLE_ACTIONS.keys.include?(action)
        send(action, params)
      else
        { error: "Unknown action: #{action}" }
      end
    end

    def actions
      AVAILABLE_ACTIONS.keys
    end

    def self.action_params(action)
      AVAILABLE_ACTIONS[action]
    end

    private

    # List all computers
    def list_all_computers(params)
      response = get_response("#{BASE_URL.value}/api/computers", "get")

      return response if response[:error]

      computers_info = response["computers"].map do |computer|
        {
          id: computer["ID"],
          ip: computer["lastIPUsed"],
        }
      end

      computers_info
    end

    # Get all firewall rules passed from a computer ID (deny and incomig type rules)
    def list_firewall_rules_by_computer(params)
      response = get_response("#{BASE_URL.value}/api/computers/#{params[:computer_id]}/firewall/rules", "get")

      return response if response[:error]

      rules_info = response["firewallRules"].select { |rule| rule["action"] == "deny" && rule["direction"] == "incoming" }.map do |rule|
        {
          name: rule["name"],
          ip: rule["ID"],
        }
      end

      rules_info
    end

    # Override an existing deny and incoming type firewall passed as a parameter from a computer ID
    def add_ip_to_firewall(params)
      available_rules = IntegrationService.new("trend_micro").execute("list_firewall_rules_by_computer", { computer_id: params[:computer_id] })
    
      if available_rules.is_a?(Array) && available_rules.empty?
        return "No Firewall type Deny and incoming registered"
      elsif available_rules.is_a?(Hash) && available_rules[:error]
        return available_rules
      end
    
      response_params = {
        sourceIPType: "multiple",
        sourceIPMultiple: params[:ip_list]
      }
    
      get_response("#{BASE_URL.value}/api/computers/#{params[:computer_id]}/firewall/rules/#{params[:firewall_id]}", "post", response_params)
    end

    def get_response(base_url, method, params = {})
      headers = {
        "Content-Type" => "application/json",
        "api-secret-key" => API_KEY.value,
        "api-version" => "v1",
      }

      begin
        response = case method.downcase
          when "get"
            HTTParty.get(base_url, headers: headers, verify: false)
          when "post"
            HTTParty.post(base_url, headers: headers, body: params.to_json, verify: false)
          else
            return { error: "Unknown method", message: "Error: Unknown method" }
          end

        if response.code != 200
          return { error: "HTTP request failed", status: response.code, message: response["message"] }
        end

        response.parsed_response
      rescue StandardError => e
        { error: "HTTP request failed", message: e.message }
      end
    end
  end
end
