#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class SyslogService

  # Sends the formatted message to the remote Syslog server via TCP or UDP
  # If we dont pass the parameter param_syslog then we asume we need to use
  # the params defined in the integration SyslogStrategy
  #
  # @param message [String] The message to be sent
  # @param_syslog [hash] Expected keys: syslog_server, syslog_port, syslog_protocol
  # @return [String] Returns the message if successfully sent
  def self.send_to_syslog(message, params_syslog = nil)
    udp_socket = nil
    tcp_socket = nil
    begin
      unless params_syslog
        syslog_strategy = Integrations::SyslogStrategy.new
        params_syslog = syslog_strategy.saved_params
      end

      required_keys = ["syslog_server"]
      required_keys.each do |key|
        raise StandardError, "Error: Missing require Syslog parameter: #{key}" unless params_syslog[key.to_sym]
      end

      syslog_server = params_syslog[:syslog_server]
      syslog_port = params_syslog[:syslog_port] || 514
      syslog_protocol = params_syslog[:syslog_protocol] || 'udp'

      # Send the message over TCP or UDP based on configuration
      if syslog_protocol.downcase == 'tcp'
        tcp_socket = TCPSocket.new(syslog_server, syslog_port)
        tcp_socket.write(message)
      else
        udp_socket = UDPSocket.new
        udp_socket.send(message, 0, syslog_server, syslog_port)
      end

      log_message = "Syslog message sent to #{syslog_server} on port #{syslog_port} with protocol #{syslog_protocol.downcase}: #{message}" rescue ''
      Rails.logger.info log_message
      Delayed::Worker.logger.info log_message 
    rescue => e
      Rails.logger.error "Error sending message to remote syslog: #{e.message}"
    ensure
      # Ensure sockets are closed properly
      udp_socket.close if udp_socket && !udp_socket.closed?
      tcp_socket.close if tcp_socket && !tcp_socket.closed?
    end
  end

  # We load the integration, if parameters are missing
  # syslog is disable, enable otherwise
  
  def self.syslog_integration_configured?
    syslog_strategy = Integrations::SyslogStrategy.new
    params = syslog_strategy.saved_params
    required_keys = ['syslog_server']

    required_keys.all? { |key| params[key.to_sym].present? }
  end
end
