#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

WickedPdf.config = {
  exe_path: '/usr/local/bin/wkhtmltopdf'
}

#
# Module to help the generation of PDF files with Wicked_pdf
#
module WickedPdfHelper
  #
  # Generate the script and stylesheets code to instert in the .pdf files
  # created with wicked_pdf. When create a pdf file we must insert all the code
  # in the HTML to avoid request to the server.
  #
  module Assets
    #
    # Create a stylesheet HTML tag and include the code of the selected css.
    #
    # == Parameters:
    # sources::
    #   String with the route of the source.
    #
    # == Returns:
    # HTML safe code with the text of the javascript file.
    #
    def wicked_pdf_stylesheet_link_tag(*sources)
      sources.collect do |source|
        asset = Rails.application.assets.find_asset("#{source}.css") ||
                Rails.application.assets.find_asset("#{source}.css.scss") ||
                Rails.application.assets.find_asset("#{source}.css.erb")

        "<style type='text/css'>\
          #{asset}\
        </style>"
      end.join("\n").html_safe
    end

    #
    # Set an image for image widgets on reports. When use wicked_pdf, we must
    # attach the code of the image or local uri of file. If the image is an URI
    # we try to retrieve it from S3 (see #StorageFile)
    #
    # == Parameters:
    # img::
    #   An URI or string with the url of the image.
    # options::
    #   Options to add to the HTML img tag like class...
    #
    # == Returns:
    # The local URL of the file or the PNG encoded on Base64.
    #
    def wicked_pdf_image_tag(img, options = {})
      if img.is_a?(URI)
        # Encode the image in Base64
        png = Base64.encode64(
                open(
                  img,
                  ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE
                ) { |io| io.read }
              )
        image_tag "data:image/png;base64,#{png}", options
      else
        asset = Rails.application.assets.find_asset(img)
        image_tag "file://#{asset.pathname}", options
      end
    end

    def wicked_pdf_javascript_src_tag(jsfile, options={})
      asset = Rails.application.assets.find_asset(jsfile)
      javascript_include_tag "file://#{asset.pathname.to_s}", options
    end

    #
    # Create a script HTML tag and include the code of the selected js.
    #
    # == Parameters:
    # sources::
    #   String with the route of the source. It can be a patter of routes.
    #
    # == Returns:
    # HTML safe code with the text of the javascript file.
    #
    def wicked_pdf_javascript_include_tag(*sources)
      js = []
      # Explode each source
      sources.each do |source|
        # Check if it's a file or a pattern
        if source.include? '*'
          # It's a pattern, find all the resources at the Rails assets folders
          Rails.application.assets.paths.each do |path|
            Dir.glob("#{path}/#{source}").each do |file|
              next unless file.include? '.js'
              js << create_script_tag(file)
            end
          end
        else
          # It's a file
          js << create_script_tag(source + '.js')
          
        end
      end

      js.join("\n").html_safe
    end

    #
    # Create the HTML tag code for a Sprosockets asset.
    #
    # == Parameters:
    # source::
    #   Source file to add to the script tag with .js
    #
    # == Returns:
    # An string with the code of the file.
    #
    def create_script_tag(asset)
      "<script type='text/javascript'>\
      #{Rails.application.assets.find_asset(asset)}\
      </script>"
    end
  end
end