#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# AP State DataSource. It takes the data from ap_state.
# The initialization traduce multiple parameters like ports, IP Address...
# to user friendly strings.
#
class ApState < Redborder::DataSource
  COLUMNS = Redborder::Module.get(:ap_state, :columns)
  SESSION_ID = Redborder::Module.get(:ap_state, :session_id).to_sym
  AGGREGATIONS = Redborder::Module.get(:ap_state, :aggregations)

  # Generate a reader attribute for the columns
  COLUMNS.map{ |k, v| v[:columns].keys }.flatten.each do |column|
    attr_reader column
  end

  # Define the column names to filter
  COLUMN_NAMES = { start_time: 'Start Time',
                   end_time: 'End Time'
                 }
  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  # Aggregations
  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  attr_reader :timestamp, :result

  #
  # Create an object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build a Traffic model.
  #
  # == Returns:
  # AP object
  #
  def initialize(user, args = {})
    if args.is_a?(Hash)
      args.each do |key, value|
        instance_variable_set("@#{key}", value)
      end

      @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp'].present?
    end

    # Round the aggregations that must be rounded
    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      unless current_value.nil?
        if aggregation_data[:round]
          instance_variable_set("@#{aggregation}", current_value.round)
        else
          instance_variable_set("@#{aggregation}", current_value.round(3))
        end
      end
    end
  end

  def self.controller
    :ap_state
  end

  #
  # Return the Symbol of id module
  #
  def self.id
    :ap_state
  end

  def self.default_aggregation
    :all
  end

  def self.sensors
    Sensor.where(type: [], is_deleted: false)
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  def self.aps_serie(filter, options)
    timeserie(
      include_all_namespaces: options[:include_all_namespaces] || false,
      granularity: options[:granularity],
      aggregation: :wireless_stations,
      user: options[:user],
      start_time: options[:start_time],
      end_time: options[:end_time],
      range: options[:range],
      search: filter.get
    )
  end

  def self.aps_count(filter, options)
    timeserie(
      include_all_namespaces: options[:include_all_namespaces] || false,
      granularity: :all,
      aggregation: :wireless_stations,
      user: options[:user],
      start_time: options[:start_time],
      end_time: options[:end_time],
      range: options[:range],
      search: filter.get
    )
  end

  def self.channel_load_serie(filter, options)
    filter_tmp = filter
    filter_tmp.add('type', "channel_load_#{options[:channel]}")
    timeserie(
      include_all_namespaces: options[:include_all_namespaces] || false,
      granularity: options[:granularity],
      aggregation: :load,
      user: options[:user],
      start_time: options[:start_time],
      end_time: options[:end_time],
      range: options[:range],
      search: filter_tmp.get
    )
  end

  #
  # Obtains a hash relating last known AP status to mac address.
  # 5 minutes query interval because that's the SNMP polling time.
  #
  # == Parameters:
  # mac_addresses::
  #   An array of strings of the AP mac addresses to filter
  #
  def self.get_access_points_status(mac_addresses, user = User.current_user)
    result = {}
    mac_addresses.each { |mac| result[mac] = 'off' }
    filter = Redborder::Filter.new(self)
    filter.add('wireless_station', mac_addresses)
    aps = current_aps_query([:wireless_station, :status], filter, user)
    aps.each { |ap| result[ap.wireless_station] = parse_status(ap.status) }
    result
  end

  #
  # Return active access points for given namespace
  #
  # == Parameters:
  # namespace::
  #   Sensor Namespace
  # user::
  #   User that make the query
  #
  # == Returns:
  # Array with MacAddress
  #
  def self.get_active_access_points_by_namespace(namespace_uuid,
                                            user = User.current_user)
    return if namespace_uuid.nil?
    filter = Redborder::Filter.new(self)
    filter.add('status', 'on')
    items_per_page = 1000
    aps = []
    pagination_aux = {}
    loop do
      pagination, events = select(
        namespaces: [namespace_uuid],
        ignore_user_namespaces: true,
        dimensions: [:wireless_station, :status],
        metrics: [:events],
        start_time: Time.now - 16.minutes,
        end_time: Time.now,
        search: filter.get,
        pagingIdentifiers: pagination_aux,
        threshold: items_per_page,
        raw_result: true,
        user: user
      )
      pagination_aux = pagination
      aps += events.map { |e| e['wireless_station'] }
      break if events.size != items_per_page
    end
    aps.uniq
  end

  def self.get_active_access_points(sensor, user = User.current_user)
    filter = Redborder::Filter.new(self)
    filter.add('status', 'on')
    filter.add('sensor_uuid', sensor.uuid)

    items_per_page = 1000
    aps = []
    pagination_aux = {}
    namespace_uuids = sensor.namespace_uuids_from_aps + [Sensor.root.uuid]

    loop do
      pagination, events = select(
        namespaces: namespace_uuids,
        ignore_user_namespaces: true,
        dimensions: [:wireless_station, :status],
        start_time: Time.now - 16.minutes,
        end_time: Time.now,
        search: filter.get,
        pagingIdentifiers: pagination_aux,
        threshold: items_per_page,
        user: user
      )
      pagination_aux = pagination
      aps += events.map(&:wireless_station)
      break if events.size != items_per_page
    end

    aps.uniq
  end

  #
  # Return the load of the channels (2.4GHz and 5GHz) based of
  # current sensor.
  #
  # == Parameters:
  # sensor_uuid::
  #
  # range::
  #
  def self.channels_load(filter, options)
    channels_load = {}
    %w(5 2_4).each do |channel|
      filter_tmp = filter
      # remove other channel_load filters
      filter_tmp.remove('type')
      filter_tmp.add('type', "channel_load_#{channel}")
      result = timeserie(
        ignore_user_namespaces: options[:ignore_user_namespaces] || false,
        granularity: :all,
        user: options[:user],
        start_time: options[:start_time],
        end_time: options[:end_time],
        range: options[:range],
        search: filter_tmp.get,
        aggregation: :load
      )
      channels_load[channel] = result.first.try(:[], :event).try(:[], :events) || nil
    end
    channels_load
  end

#
  # Return active access points with their clients associated based of
  # current sensor_uuid and namespaces.
  #
  # == Parameters:
  # sensor_uuid::
  #   String or Array of Strings with the UUID of sensors to get data.
  # namespaces::
  #   Selected namespaces of the user
  # user::
  #   User that make the query
  #
  # == Returns:
  # Hash with the format:
  #   { "MacAdddress" => number_clients,
  #     "MacAddress2" => number_clients2,
  #     ... }
  #
  def self.active_access_points_and_clients(sensor_uuid, namespaces, user = User.current_user)
    filter = Redborder::Filter.new(self)
    filter.add('status', 'on')
    filter.add('sensor_uuid', sensor_uuid)
    result = {}
    aps = current_aps_query(
      [:wireless_station, :status, :client_count], filter, user, namespaces)
    aps.each do |ap|
      result[ap.wireless_station] = ap.client_count || 0
    end
    result
  end

  #
  # Obtains a the APs (name, status and mac address) related to a sensor
  # by it's UUID
  # == Parameters:
  # sensor_uuid::
  #   The sensor UUID (number or string)
  #
  def self.get_access_points(sensor_uuid, user = User.current_user)
    filter = Redborder::Filter.new(self)
    filter.add('sensor_uuid', sensor_uuid.to_s)
    where(
        query_type: :group_by,
        granularity: :all,
        start_time: Time.now - 1.hour,
        end_time: Time.now,
        group_by: [:wireless_station, :wireless_station_name, :status],
        search: filter.get,
        user: user
    ).uniq{ |o| o.wireless_station }
  end

  #
  # Base query to get current AccessPoints
  #
  # == Parameters:
  # columns::
  #   Array of symbols with colums to group by.
  # filter::
  #   Filter to apply
  # user::
  #   User that make the query
  # namespaces::
  #   Selected namespaces of the user
  #
  # == Returns:
  # Druid response.
  #
  def self.current_aps_query(columns, filter, user, namespaces = nil)
    base = {
      query_type: :group_by,
      granularity: :all,
      start_time: Time.now - 16.minutes,
      end_time: Time.now,
      group_by: columns,
      search: filter.get,
      user: user
    }
    # Get from nampespaces
    base.merge!(namespaces: namespaces) unless namespaces.nil?
    where(base)
  end

  #
  # Query to get sensors uuids that have reported APs data to Druid
  #
  def self.checked_flow_sensors(namespaces, limit = 1000, user = User.current_user)
    topN(
      user: user,
      query_type: :group_by,
      start_time: Time.now - 16.minutes,
      end_time: Time.now,
      range: :custom,
      granularity: :all,
      limit: limit,
      namespaces: namespaces,
      ignore_user_namespaces: true,
      group_by: [:sensor_uuid],
      aggregation: :events
    ).to_a.map(&:sensor_uuid).uniq
  end

  def parse_status(status)
    if status == 'on'
      AccessPoint::STATUS[:UP]
    else
      AccessPoint::STATUS[:DOWN]
    end
  end

  # =================================================
  # QUERYING
  # =================================================

  #
  # Override method.
  #
  # Get the data from Druid in timeserie format.
  #
  # == Paramenters:
  # args::
  #   Hash with options to perform the query. The options are:
  #     - ...
  #
  # == Return:
  # Array with the data from Druid.
  #
  def self.timeserie(options = {})
    conditions = options.deep_dup
    user = options.fetch(:user, User.current_user)

    # Default values
    conditions[:query_type] = :timeseries
    conditions = defaults(conditions, user)
    # Set the aggregation
    aggregation = conditions[:aggregation]
    aggregations = Redborder::Module.get(id, :aggregations)
    aggregation_data = aggregations[aggregation.to_sym]

    # Perform the query
    result = perform_query(conditions)

    if result.is_a? Array
      result.map do |t|
        value = t[conditions[:aggregation].to_s]

        unless value.nil?
          if aggregation_data[:round]
            value = value.round
          else
            value = value.round(3)
          end
        end

        if conditions[:outliers]
          { expected: t['expected'],
            timestamp: t['timestamp'].to_time.in_time_zone.to_s }
        else
          { event: { events: value },
            timestamp: t['timestamp'].to_time.in_time_zone.to_s }
        end
      end
    else
      result
    end
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns, zones and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains, columns, sensors and zones
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user
              .sensor
              .self_and_descendants
              .where.not(id: 1)
              .where(type: Sensor::DOMAIN)
    organizations = user
                    .sensors_with_permission
                    .where.not(id: 1)
                    .where(domain_type: 2)

    f = super.merge!(
      ap_state_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by do |h|
        h[:value].to_s
      end,
      ap_state_sensors: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: [])
          .map { |c| { value: c.name, id: c.name } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      },
      organizations: {
        type: :dropdown,
        value: organizations.map { |c| { value: c.name, id: c.name } }
      }
    )
    # Check format
    json ? f.to_json : f
  end

end