#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# This job uses SNMP4EM, gem that extends Ruby-SNMP to use the
# asynchronous EventMachine library for added performance and
# scalabilit. This allows code to scale monitoring applications
# to access a very hight number of devices.
#
module Redborder
  module Jobs
    class CheckAPsStatusJob < Struct.new(:verbose)
      include Redborder::Jobs::JobHelper

      # Step to make bulk update query
      APS_QUERY_STEP = 300

      def perform
        # Get Array with all active APs
        @active_aps = active_aps
        # Mark APs as down
        all_aps_down
        # Mark Active APs to UP
        active_aps_up
        # Update_sensors last Check In
        update_sensors_check_in
        # Consolidate AP status
        consolidate_aps
        # Enqueue the Job
        enqueue_job
      end

      private

      #
      # Updates Flow/Meraki sensors that have reported
      # access points status data
      #
      def update_sensors_check_in
        # update all flow/meraki sensors matching uuids
        Sensor.where(
          type: [Sensor::FLOW, Sensor::MERAKI],
          is_deleted: false,
          uuid: sensors_uuids).each { |s| s.update(check_in_time: Time.now) }
      end

      # get all flow/meraki uuids from Druid
      # limit is number of flow/meraki sensors + extra threshold of 100
      def sensors_uuids
        ApState.checked_flow_sensors(
          namespaces_uuids,
          Sensor.where(type: [Sensor::FLOW, Sensor::MERAKI]).count + 100,
          checker_user
        )
      end

      #
      # Set All Access Point to down
      #
      def all_aps_down
        AccessPoint.update_all(next_status: AccessPoint::STATUS[:DOWN], clients: 0)
      end

      #
      # Mark Active APs to UP by blocks of APS_QUERY_STEP
      #
      def active_aps_up
        sensors_uuids.each do |sensor_uuid|
          active_aps_and_clients = active_aps_and_clients(sensor_uuid)

          active_aps_and_clients.each do |ap, client_count|
            AccessPoint.where(
            mac_address: ap
            ).update_all(
              next_status: AccessPoint::STATUS[:UP],
              clients: client_count,
              last_check_in: Time.now.utc
            )
          end
        end
      end

      #
      # Consolidate AP status from next_status to status attribute
      #
      def consolidate_aps
        AccessPoint.where(next_status: AccessPoint::STATUS[:DOWN])
          .where.not(status: AccessPoint::STATUS[:DOWN])
          .update_all(status: AccessPoint::STATUS[:DOWN])

        AccessPoint.where(next_status: AccessPoint::STATUS[:UP])
          .where.not(status: AccessPoint::STATUS[:UP])
          .update_all(status: AccessPoint::STATUS[:UP])
      end

      #
      # Divides Array of APs MACs in blocks of APS_QUERY_STEP
      #
      def active_aps_blocks
        @active_aps.each_slice(APS_QUERY_STEP).to_a
      end

      #
      # Returns Array of Active AP Mac Addresses
      #
      def active_aps
        aps = []
        namespaces_uuids.each { |uuid| aps += active_aps_by_namespace(uuid) }
        aps
      end

      #
      # Get Active APs by namespace from Druid
      #
      def active_aps_by_namespace(namespace_uuid)
        ApState.get_active_access_points_by_namespace(namespace_uuid,
                                                               checker_user)
      end

      #
      # Get Active APs by namespace from Druid
      #
      def active_aps_and_clients(sensor_uuid)
        ApState.active_access_points_and_clients(sensor_uuid, namespaces_uuids, checker_user)
      end

      #
      # Returns Namespaces uuids including root uuid
      #
      def namespaces_uuids
        Sensor.where(domain_type: Sensor::NAMESPACE).pluck(:uuid) << Sensor.root.uuid
      end

      #
      # Returns admin user on root
      #
      def checker_user
        User.find_by(admin: true, sensor: Sensor.root)
      end

      #
      # Enque Job to run after 5 minutes
      #
      def enqueue_job
        Delayed::Job.enqueue(Redborder::Jobs::CheckAPsStatusJob
          .new(true), priority: 10, run_at: Time.now + 5.minutes)
      end
    end
  end
end
