#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::AccessPointsController < Api::V1::BaseController
  # Validate permissions when create
  before_action :init_domain, only: [:create]
  before_action :set_access_point, only: [:update]
  before_action :manage_permission, only: [:update]

  skip_before_action :authenticate_user!, only: :import, raise: false
  skip_before_action :authenticate_user_from_token!, only: :import
  before_action :validate_cert, only: :import
  before_action :set_sensor, only: :import

  #
  # Get a list of access points.
  #
  def index
    # Get Ids of readable sensors
    sensor_ids = current_user.sensors_with_permission.map(&:id)
    # Get access points
    access_points = AccessPoint.where(
      'switch_id in (?) OR sensor_id in (?)',
      sensor_ids,
      sensor_ids
    ).where(ap_params)

    data = []
    access_points.each do |ap|
      data << ap.info
    end

    # Respond based on the format
    respond_to do |format|
      format.json do
        render json: {
          query: true,
          access_points: data
        }
      end
      format.csv { render text: AccessPoint.to_csv(data) }
    end
  end

  def destroy
      aps = params[:access_points]
      aps = [aps] if aps.is_a? Hash
      error = false
      destroyed = 0

      aps.each do |ap|
        if ap[:mac_address] || ap[:id]
          ap_to_delete = AccessPoint.find_by(mac_address: ap[:mac_address]) if ap[:mac_address]
          ap_to_delete = AccessPoint.find_by(id: ap[:id]) if ap[:id]
          if ap_to_delete && ap_to_delete.manage_permission?
            ap_to_delete.destroy
            destroyed += 1
          end
        else
          error = true
          break
        end
      end
      # Check if there was an error with the request
      if error
        error_400 "#{I18n.t('controllers.you_should')}"
      else
        render json: { query: true, destroyed_access_points: destroyed }
      end
  end

  #
  # Create an ap and add it to a floor and a flow sensor.
  #
  def create

    # Create the AP
    ap = AccessPoint.new(ap_params)
    ap.sensor_id = @domain.id
    if params[:flow_uuid]
      flow = Sensor.find_by(uuid: params[:flow_uuid], type: Sensor::FLOW)
      ap.switch_id = flow.id if flow
    end
    if ap.save
      respond_to do |format|
        format.json do
          render json: {
            query: true,
            ap: ap
          }
        end
      end
    else
      error_422("#{I18n.t('controllers.sensor_cant_create')}"\
                "#{ap.errors.full_messages.join(', ')}")
    end
  end

  #
  # Update an access point
  #
  def update
    para = ap_params
    para[:sensor_id] = Sensor.find_by(uuid: params[:sensor_uuid]).id if params[:sensor_uuid]
    para[:switch_id] = Sensor.find_by(uuid: params[:flow_uuid]).id if params[:flow_uuid]

    if @ap.update(para)
      respond_to do |format|
        format.json do
          render json: {
            query: true,
            ap: @ap
          }
        end
        # Render CSV
        format.csv { render text: @ap.to_csv }
      end
    else
      error_422("#{I18n.t('controllers.sensor_cant_update')}"\
                "#{@ap.errors.full_messages.join(', ')}")
    end
  rescue NoMethodError => e
    # Logger error in case that the sensor doesn't belong to the given uuid param
    Rails.logger.error e.inspect
    Rails.logger.error e.backtrace.join("\n")
    return error_404("#{I18n.t('controllers.sensor_not_found')}")
  end

  #
  # Creates / Updates Access Points for Meraki/Flow
  #
  def import
    # Cache loop variables
    access_points_json = (params[:access_points] || [])
    sensor_id = @sensor.id
    time_now = Time.now

    # Updates or insert Received APs
    Upsert.batch(AccessPoint.connection, AccessPoint.table_name) do |upsert|
      access_points_json.each do |ap|
        upsert.row({ mac_address: ap[:mac_address].downcase },
                   name: ap[:name], status: AccessPoint::STATUS[:UP], switch_id: sensor_id,
                   clients: ap[:clients], last_check_in: time_now)
      end
    end

    # Set not Received APs DOWN
    AccessPoint.where(switch_id: sensor_id)
      .where.not(mac_address: access_points_json.map { |ap| ap[:mac_address].downcase })
      .update_all(status: AccessPoint::STATUS[:DOWN])

    # Updates Last check in value
    @sensor.update(check_in_time: Time.now)

    Rails.logger.debug "Processed #{access_points_json.count} APs from " \
                       "#{@username}:#{@sensor.name} in #{(Time.now - time_now).round(2)}s."

    render json: { query: true }
  end

  private

  #
  # Instance sensor by uuid
  #
  def set_sensor
    @sensor = Sensor.find_by(type: [Sensor::FLOW, Sensor::MERAKI],
                             uuid: params[:sensor_uuid],
                             real_parent_id: @username.split('-').last)
    error_401("#{I18n.t('controllers.sensor_not_exist')}") unless @sensor
  end

  # Permitted parameters
  def ap_params
    if request.get?
      params.permit(:ip_address, :mac_address, :name, :status)
    elsif request.post? || request.patch? || request.put?
      params.permit(:ip_address, :mac_address, :name, :status,
                    :latitude, :longitude, :x, :y, :sensor_id,
                    :description, :switch_id,
                    enrichment: [:market_uuid, :campus_uuid, :building_uuid, :floor_uuid])
    end
  end

  def init_domain
    @domain  = Sensor.find_by(uuid: params[:sensor_uuid], type: Sensor::DOMAIN)
    check_domain_manage_permissions
  end

  def check_domain_manage_permissions
    return error_404("#{I18n.t('controllers.sensor_not_found2')}") if @domain.nil?
    return error_401("#{I18n.t('controllers.no_privileges')}") if current_user.cannot?(:manage, @domain)
  end

  #
  # Init and check sensors floors and sensors flow to create aps from API.
  #
  def init_sensors
    @floor  = Sensor.find_by(uuid: params[:sensor_uuid], domain_type: Sensor::FLOOR)
    @sensor = Sensor.find_by(uuid: params[:flow_uuid], type: Sensor::FLOW)

    check_create_permissions
  end

  #
  # Check permissions and the existence of sensors.
  #
  def check_create_permissions
    return error_404("#{I18n.t('controllers.specify_flow')}") if
      @floor.nil? && @sensor.nil?

    # Check permissions
    return error_401("#{I18n.t('controllers.no_privileges')}") if
      (@floor && current_user.cannot?(:manage, @floor)) ||
      (@sensor && current_user.cannot?(:manage, @sensor))
  end

  #
  # Check manage permissions on ap
  #
  def manage_permission
    return if @ap.manage_permission?(current_user)
    return error_401("#{I18n.t('controllers.no_privileges')}")
  end

  #
  # Init the access point based on params[:id]. It can be an ID or Mac Address
  #
  def set_access_point
    @ap = if params[:id].include?(':')
            AccessPoint.find_by_mac_address(params[:id])
          else
            AccessPoint.find_by_id(params[:id])
          end
    return error_404("#{I18n.t('controllers.requested_ap')}") if @ap.nil?
  end

end
