#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Provide an api to work with Flow sensors
#
class Api::V1::SensorsFlowController < Api::V1::BaseSensorsController
  # ----------------------------------------------
  # Before and after actions
  # ----------------------------------------------
  before_action :set_domain,
                only: [:flow, :mse, :ale, :arubacentral, :device, :pi, :meraki, :override_aps, :delete_aps,
                       :update_flow, :update_mse, :update_ale, :update_arubacentral, :update_device, :update_meraki]
  before_action :manage_domain,
                only: [:override_aps, :delete_aps]
  before_action :set_parent,
                only: [:create_flow, :create_mse, :create_ale, :create_arubacentral, :create_device, :create_meraki, :create_pi]
  before_action :manage_parent,
                only: [:create_flow, :create_mse, :create_ale, :create_arubacentral, :create_device, :create_meraki, :create_pi]

  # ----------------------------------------------
  # Actions
  # ----------------------------------------------

  # Work with sensors
  # ----------------------------------------------

  # For every type
  %w(flow mse meraki ale arubacentral device).each do |s|
    #
    # GET all sensors of a type for the specified domain
    #
    define_method(s) do
      render json: {
        query: true,
        s => @domain.send("get_#{s}")
      }
    end

    #
    # Update a sensor
    #
    define_method("update_#{s}") do
      # Check error when save
      return error_on_sensor(@domain) unless
        @domain.update(send("params_#{s}"))

      render json: {
        query: true,
        s => @domain
      }
    end

    #
    # Create a new sensor
    #
    define_method("create_#{s}") do
      @sensor = Sensor.new(send("params_#{s}"))
      @sensor.parent = @parent
      @sensor.type = Sensor.const_get("#{s.upcase}")

      # Check error when save
      return error_on_sensor(@sensor) unless @sensor.save
      @sensor.add_default_flow_monitors
      # Wakeup chef and return
      wakeup_chef
      render json: {
        query: true,
        sensor: send("#{s}_json")
      }
    end
  end

  # Work with APs
  # ----------------------------------------------

  #
  # Override all access points of the domain with the APs given on JSON payload
  #
  def override_aps
    @domain.delete_all_access_points
    # override with given ones
    @domain.create_access_points_from_json params[:access_points], @domain
    respond_to do |format|
      format.json do
        render json: { query: true }
      end
    end
  end

  #
  # Delete all the APs of the specified domain and it's children
  #
  def delete_aps
    @domain.delete_all_access_points
    render json: { query: true }
  end

  private

  # Params and renderization
  # ----------------------------------------------

  def params_flow
    params.permit(*Sensor.import_columns(Sensor::FLOW))
  end

  def params_mse
    params.permit(*Sensor.import_columns(Sensor::MSE))
  end

  def params_ale
    params.permit(*Sensor.import_columns(Sensor::ALE))
  end

  def params_arubacentral
    params.permit(*Sensor.import_columns(Sensor::ARUBACENTRAL))
  end

  def params_device
    params.permit(*Sensor.import_columns(Sensor::DEVICE))
  end

  def params_meraki
    params.permit(*Sensor.import_columns(Sensor::MERAKI))
  end

  # Return the sensors to be rendered
  def base_json
    { id: @sensor.id, uuid: @sensor.uuid, name: @sensor.name }
  end

  def flow_json
    base_json.merge(ip: @sensor.ip, nmsp: @sensor.nmsp,
                    nmsp_wireless_health: @sensor.nmsp_wireless_health,
                    nmsp_client_stats: @sensor.nmsp_client_stats,
                    dns_ptr_client: @sensor.dns_ptr_client,
                    dns_ptr_target: @sensor.dns_ptr_target,
                    homenets: @sensor.homenets,
                    routers_macs: @sensor.routers_macs,
                    spanport: @sensor.spanport,
                    captive_portal_name: @sensor.captive_portal_name,
                    captive_portal_name_color: @sensor.captive_portal_name_color,
                    captive_portal_name_color: @sensor.captive_portal_name_color                    
                    )
  end

  def mse_json
    base_json.merge(stream: @sensor.stream, version_mse: @sensor.version,
                    homenets: @sensor.homenets, send_http: @sensor.send_http)
  end

  def ale_json
    base_json.merge(stream: @sensor.stream, version_ale: @sensor.version,
                    controller_ip: @sensor.controler_ip, rest_api_user: @sensor.rest_api_user,
                    rest_api_key: @sensor.rest_api_key, name: @sensor.name,
                    homenets: @sensor.homenets, send_http: @sensor.send_http)
  end

  def arubacentral_json
    base_json.merge(arubacentral_username: @sensor.arubacentral_username,
                    arubacentral_password: @sensor.arubacentral_password,
                    arubacentral_client_id: @sensor.arubacentral_client_id,
                    arubacentral_client_secret: @sensor.arubacentral_client_secret,
                    arubacentral_customer_id: @sensor.arubacentral_customer_id,
                    name: @sensor.name,
                    homenets: @sensor.homenets, 
                    send_http: @sensor.send_http)
  end

  def device_json
    base_json.merge(stream: @sensor.stream, protocol: @sensor.version,
                    controller_ip: @sensor.controler_ip, rest_api_user: @sensor.rest_api_user,
                    rest_api_key: @sensor.rest_api_key, name: @sensor.name,
                    homenets: @sensor.homenets, send_http: @sensor.send_http)
  end

  def meraki_json
    base_json.merge(ip: @sensor.ip, validator: @sensor.meraki_validator,
                    secret: @sensor.meraki_secret,
                    captive_portal_name: @sensor.captive_portal_name,
                    captive_portal_name_color: @sensor.captive_portal_name_color,
                    captive_portal_background_color: @sensor.captive_portal_background_color
                    )
  end
end
