#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class TrafficsController < Redborder::Commons::CommonController
  # Storage of the filters
  include Redborder::Filters::Storage::Session
  # Includes
  include Redborder::Rbdruid

  self.view_paths = 'lib/modules/flow/app/views'
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  helper_method :sort_column, :sort_direction

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_saved_flow_filters,
                only: [:index, :tops, :map, :floor_map, :comparative,
                       :cardinality, :heatmap, :export_raw_to_file, :clients,
                       :client_points, :search_clients]

  before_action :set_time_params, only: [:aps_serie, :aps_metrics, :aps_count,
                                         :channels_load_serie]

  before_action :redirect_if_empty, only: [:index, :tops, :map, :comparative,
                                           :cardinality, :heatmap]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def lookup
    @lookup = Redborder::Rblookup.new(params[:address])
    render layout: false
  end

  #
  # => Represent zones in a map and data relative of these zones.
  #    Also represent floors.
  #    If no id is passed, the sensor is the Top Domain sensor of the
  #    current user.
  #
  def map
    # TODO: If no sensor_id is specified, top sensor must be initialized
    custom_sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
    @sensor = if custom_sensor && current_user.can?(:read, custom_sensor)
                custom_sensor
              else
                # Set top domain of the user
                current_user.sensor
              end

    all_sensor_ids = []

    # Store filters before add temp domain filters
    @filter.stash

    # Get the ids of the sensors the user can read. These sensors have the
    # domain filters (organization, market...) applied on domain_filter method.
    Sensor
    .where(id: domain_filter(@sensor, @filter.get, [Sensor::FLOW, Sensor::MERAKI])).each do |s|
      all_sensor_ids += [s.id] + s.all_children.map(&:id)
    end
    all_sensor_ids = all_sensor_ids.uniq

    if all_sensor_ids.empty?
      # Set all the data with empty values
      @total_access_points = 0
      @total_clients = 0
      @aps = 0
      @access_points = []
      @domain_zones = []
    else
      # Get all children
      children = Sensor.where(is_deleted: false,
                              type: Sensor::DOMAIN,
                              domain_type: Sensor::FLOOR,
                              id: all_sensor_ids)

      @access_points = if all_sensor_ids.empty?
                         @sensor.access_points
                       else
                         AccessPoint.where('sensor_id IN (?) OR switch_id IN (?)',
                                           all_sensor_ids, all_sensor_ids)
                       end

      # Get active Access Points from Druid.
      active_aps = ApState.active_access_points_and_clients(
        @sensor.self_and_descendants.map(&:uuid),
        @sensor.namespaces(current_user).map(&:uuid) &
        current_user.active_data_sources,
        current_user)

      @access_points.each do |ap|
        if active_aps.key? ap.mac_address
          ap.status = AccessPoint::STATUS[:UP]
          ap.clients = active_aps[ap.mac_address]
        else
          ap.status = AccessPoint::STATUS[:DOWN]
          ap.clients = 0
        end
      end

      @total_access_points = @access_points.count

      # Render the floor if there are more than 3000 aps
      if @total_access_points > 3000
        @floors = children
                  .select(&:gps_points)
                  .to_json(only: [:id, :name],
                           methods: [:gps_points, :icon])
      end

      # Prevent if clients attribute is nil
      @total_clients = @access_points.sum(:clients)
      @aps = @access_points.select { |ap| ap.clients > 0 }.size

      @domain_zones =
        Sensor.where(
          is_deleted: false,
          type: Sensor::DOMAIN,
          id: all_sensor_ids
        )
        .collect do |s|
          {
            domain_zones: JSON.parse(s.domain_zones),
            title: s.name,
            id: s.id,
            icon: Sensor::DOMAIN_TYPES[s.domain_type][:icon],
            type_icon: Sensor::DOMAIN_TYPES[s.domain_type][:type_icon],
            color: Sensor::DOMAIN_TYPES[s.domain_type][:color]
          } if s.domain_zones.present?
        end.compact

      # If @sensor can be read, all children can be read too.
      flow_sensors = Sensor
                     .where(type: Sensor::FLOW,
                            id: all_sensor_ids)
      add_filters_map(flow_sensors)
    end

    @end_time = Time.now
    @start_time = @end_time - 2.minutes

    @heatmap = @source.where(
      query_type: :group_by,
      start_time: @start_time,
      end_time: @end_time,
      range: :custom,
      granularity: :all,
      sort: [@aggregation,
             @source::COLUMNS[:coordinates_map][:columns].keys.first],
      limit: 1000,
      group_by: @source::COLUMNS[:coordinates_map][:columns].keys,
      aggregation: @aggregation,
      search: @filter.get
    )

    @clients = {}

    @clients[:associated] = @source.total(
      start_time: @start_time,
      end_time: @end_time,
      range: :custom,
      aggregation: :clients,
      search: @filter.add_tmp(:dot11_status, ['ASSOCIATED'])
    ) || 0

    @clients[:probing] = @source.total(
      start_time: @start_time,
      end_time: @end_time,
      range: :custom,
      aggregation: :clients,
      search: @filter.add_tmp(:dot11_status, ['PROBING'])
    ) || 0

    @bps = @source.total(
      start_time: @end_time - 5.minutes,
      end_time: @end_time,
      range: :custom,
      aggregation: :bps,
      search: @filter.get
    )

    # Restore back filters
    @filter.pop

    # Fix Associated clients if it's less than Active clients
    @total_clients = @clients[:associated] if @total_clients < @clients[:associated]

    respond_to do |format|
      format.html
      format.js
      format.json {
        render :json => {
                          total_access_points: @total_access_points,
                          total_clients: @total_clients,
                          access_points: @access_points,
                          domains: @domain_zones,
                          domain_zones: @domain_zones,
                          heatmap: @heatmap,
                          aps: @aps,
                          clients: @clients,
                          bps: @bps
                        }
      }
    end
  end

  #
  # => Return the mac address from clients.
  #
  def clients
    # Get the current sensor
    sensor = if params[:sensor_id]
               Sensor.find(params[:sensor_id])
             else
              # Set top domain of user
               current_user.sensor
             end

    # Check permissions in the sensor
    if current_user.cannot?(:read, sensor)
      render json: { status: :unauthorized }
      return
    end

    offset = params[:offset].present? ? (params[:offset].to_i) : 0

    # If @sensor can be read, all children can be read too.
    flow_sensors = sensor.all_children
                   .where(type: Sensor::FLOW)

    # Load the filters of the map
    add_filters_map(flow_sensors)

    @start_time =
      if params[:start_time]
        Time.zone.at(params[:start_time].to_i)
      else
        @start_time
      end

    @end_time =
      if params[:end_time]
        Time.zone.at(params[:end_time].to_i)
      else
        @end_time
      end

    clients = @source.where(
        query_type: :group_by,
        start_time: @start_time,
        end_time: @end_time,
        range: :custom,
        granularity: :all,
        limit: 30,
        offset: offset,
        group_by: @source::COLUMNS[:client_mac][:columns].keys,
        aggregation: :clients,
        # Ignore events without client_latlong
        search: @filter.exclude('client_latlong', [''])
      )

    render json: { clients: clients, client_status: status }
  end

  #
  # => Return the client points based on the mac address
  #
  def client_points
    # Get the current sensor
    sensor = if params[:sensor_id]
               Sensor.find(params[:sensor_id])
             else
               # Set top domain user
               current_user.sensor
             end

    # Check permissions in the sensor
    if current_user.cannot?(:read, sensor)
      render json: { status: :unauthorized }
      return
    end

    # Initialize search
    mac_address = params[:mac_address]

    # If @sensor can be read, all children can be read too.
    flow_sensors = sensor.all_children
                   .where(type: Sensor::FLOW)

    add_filters_map(flow_sensors)

    @start_time =
      if params[:start_time]
        Time.zone.at(params[:start_time].to_i)
      else
        @start_time
      end

    @end_time =
      if params[:end_time]
        Time.zone.at(params[:end_time].to_i)
      else
        @end_time
      end

    # Add filters
    @filter.add('client_mac', [mac_address])
    @filter.exclude('client_latlong', [''])

    positions = @source.topN(
      query_type: :group_by,
      start_time: @start_time,
      end_time: @end_time,
      range: :custom,
      granularity: :minute,
      limit: 1000,
      group_by: @source::COLUMNS[:client_latlong][:columns].keys,
      aggregation: :events,
      search: @filter.get
    )

    # Reverse the positions to hsow in order in the map
    render json: { positions: positions.reverse, client: mac_address }
  end

  def search_clients
    # Get the current sensor
    sensor = if params[:sensor_id]
               Sensor.find(params[:sensor_id])
             else
               # Set top domain user
               current_user.sensor
             end

    # Check permissions in the sensor
    if current_user.cannot?(:read, sensor)
      render json: { status: :unauthorized }
      return
    end

    # Get the string to search
    search = params[:search]

    # Initialize search
    clients = []
    offset = params[:offset].present? ? (params[:offset].to_i) : 0

    # Check if the search is a literal MAC value or a redborder object.
    if search.include? ':'
      @filter.contains('client_mac', [search])
    else
      mac = MacObject.where('(LOWER(name) like ?)',
                            "%#{params[:search].downcase.split.join('%')}%"
                            .downcase)

      if mac.empty?
        @filter.contains('client_id', [search])
      else
        @filter.add('client_mac', mac.map(&:value))
      end
    end

    # Check search
    if search.size > 2 &&
       (@filter.exist_any?('client_mac') && !@filter.get('client_mac').empty? ||
        @filter.exist_any?('client_id') && !@filter.get('client_id').empty?)
      # If @sensor can be read, all children can be read too.
      flow_sensors = sensor.all_children
                     .where(type: Sensor::FLOW)
      add_filters_map(flow_sensors)

      @end_time =
        if params[:end_time]
          Time.zone.at(params[:end_time].to_i)
        else
          @end_time
        end

      @start_time =
        if params[:start_time]
          Time.zone.at(params[:start_time].to_i)
        else
          @start_time
        end

      clients = @source.where(
        query_type: :group_by,
        start_time: @start_time,
        end_time: @end_time,
        range: :custom,
        granularity: :all,
        offset: offset,
        limit: 30,
        group_by: @source::COLUMNS[:client_mac][:columns].keys,
        aggregation: :events,
        search: @filter.add('client_latlong', [''], 'is_not')
      )
    end

    # Reverse the positions to hsow in order in the map
    render json: { clients: clients }
  end

  def get_points
    @filter.add('client_latlong', ['']) unless @filter.exist_any?(:client_latlong)

    result = @source.as_json(
      query_type: :group_by,
      granularity: params[:granularity],
      group_by: [:client_latlong],
      limit: 1000,
      aggregation: params[:aggregation],
      search: @filter.get,
      start_time: Time.at(params[:start_time].to_i),
      end_time: Time.at(params[:end_time].to_i)
    )
    render json: result
  end

  # TODO: check aggregation
  def get_maps_data
    tab = params[:id].to_sym
    limit = params.fetch(:limit, current_user.per_page_count)
    @start_time = params[:start_time].to_time
    @end_time = params[:end_time].to_time

    content = @source.where(
      query_type: :group_by,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      limit: limit,
      sort: @aggregation,
      group_by: @source::COLUMNS[tab][:columns].keys,
      search: @filter.get,
      aggregation: @aggregation
    )

    ips = content.collect(&@source::COLUMNS[tab][:ips])

    respond_to do |format|
      format.json { render json: Redborder::Geoip.latlong(ips) }
    end
  end

  def aps_metrics
    params[:user] = current_user
    params[:aggregation] = "wireless_stations"
    @aps_status = ApState.aps_serie(@filter, params)
    @channels_load = {}
    ['2_4', '5'].each do |channel|
      @channels_load[channel] = ApState.channel_load_serie(@filter, params.merge(channel: channel))
    end
    respond_to do |format|
      format.js {}
    end
  end

  def channels_load
    params[:user] = current_user
    channels_load = ApState.channels_load(@filter, params)
    render json: { query: true, channels_load: channels_load || [] }
  end

  def channels_load_serie
    params[:user] = current_user
    channels_load = ApState.channel_load_serie(@filter, params)
    render json: { query: true, channels_load: channels_load || [] }
  end

  def aps_serie
    params[:user] = current_user
    @serie = ApState.aps_serie(@filter, params)
    respond_to do |format|
      format.json { render json: { query: true, aps: @serie || [] } }
      format.js {}
    end
  end

  def aps_count
    params[:user] = current_user
    serie = ApState.aps_count(@filter, params)
    render json: { query: true, aps: serie || [] }
  end

  private

  def set_time_params
    unless params['w'] && params[:range].to_sym.eql?(:custom)
      params[:end_time] = in_time_machine { Time.now.change(sec: 0) }
      # Round time to lower multiple of 5 minutes
      step = Redborder::Rbdruid::GRANULARITY_TYPE[params[:granularity].to_sym][:step] / 1000 rescue 5.minutes
      params[:end_time] = Time.at((params[:end_time].to_f / step).floor * step)
      params[:start_time] = case params[:range].to_sym
                            when :last_1_hour
                              params[:end_time] - 1.hour
                            when :last_2_hour
                              params[:end_time] - 2.hour
                            when :last_3_hour
                              params[:end_time] - 3.hour
                            when :last_12_hour
                              (params[:end_time] - 12.hour).beginning_of_hour
                            when :last_24
                              (params[:end_time] - 24.hour).beginning_of_hour
                            when :last_week
                              (params[:end_time] - 1.week).beginning_of_hour
                            else
                              params[:end_time] - 1.hour
                            end
    end
    params[:granularity] = (Redborder::Rbdruid
                            .granularity_list(start_time: params[:start_time],
                            end_time: params[:end_time]) - [:minute, :pt2m, :pt5m]).first
  end

  def redirect_if_empty
    @sensors = Sensor.where type: [Sensor::FLOW, Sensor::MSE, Sensor::ALE, Sensor::DEVICE, Sensor::MERAKI, Sensor::CLIENT_PROXY]
    redirect_to help_traffics_path if @sensors.empty?
  end

  def set_saved_monitor_filters
    @filters = SavedMonitorFilter.where(user_id: current_user.id)
    @filters_public = SavedMonitorFilter.where
                      .not(user_id: current_user.id).select(&:public)
  end

  def get_scatterplot_tab
    columns = @source::COLUMNS[:scatterplot][:columns].keys

    content = @source.where(
      query_type: :group_by,
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      sort: @aggregation,
      limit: current_user.per_page_count,
      group_by: columns,
      aggregation: @aggregation,
      search: @filter.get
    )

    { content: content }
  end

  def get_trafficmap_tab
    columns = @source::COLUMNS[:trafficmap][:columns].keys

    content = @source.where(
      query_type: :group_by,
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      sort: @aggregation,
      limit: 1000,
      group_by: columns,
      aggregation: @aggregation,
      search: @filter.get
    )

    { content: content }
  end

  def get_gmap_tabs(tab)
    content = @source.where(
      query_type: :group_by,
      range: @range,
      start_time: @start_time,
      end_time: @end_time,
      granularity: :all,
      sort: @aggregation,
      limit: current_user.per_page_count,
      group_by: @source::COLUMNS[tab][:columns].keys,
      aggregation: @aggregation,
      search: @filter.get
    )

    ips = content.collect(&@source::COLUMNS[tab][:ips])

    { content: content, graph: Redborder::Geoip.latlong(ips) }
  end

  # -------------------------------------------------------------
  # Check Domain filters
  # -------------------------------------------------------------

  #
  # Search ids of domains an user can search from a Sensor and filters.
  # The sensors is the parent, the start point to search childs. Children
  # are filtered with domain filters inside the hash of filters.
  #
  # == Parameters:
  # sensor::
  #   Sensor object to get childs
  # filters::
  #   Hash of filters. Domain filters will be selected and applied to
  #   childs.
  # sensor_types::
  #   An array with sensor types to add to the query
  #
  # == Returns:
  # An array of ids with the filtered domains
  #
  def domain_filter(sensor, filters, sensor_types = [])
    all_sensors = sensor
                  .self_and_descendants
                  .where(is_deleted: false,
                         type: [Sensor::DOMAIN] + sensor_types)
    # Check if there are a domain filter
    return all_sensors.map(&:id) unless domain_filters?(filters)
    # Sensors after filtered
    filtered_sensors = []

    # Search on filters
    filters.each do |f|
      next unless concern_domain?(f.first)
      # Iterate over values of the filter
      f.last.keys.each do |name|
        case f.last[name][:operator]
        when 'is'
          filtered_sensors += all_sensors.select { |s| s.name == name }
        when 'is_not'
          sensor = all_sensors.select { |s| s.name == name }
          filtered_sensors.delete(sensor) unless sensor.nil?
        when 'contains'
          filtered_sensors += all_sensors.select { |s| s.name.include? name }
        end
      end
    end
    # Get the ids and return
    filtered_sensors.map(&:id)
  end

  #
  # Check if the filters has a domain filter
  #
  # == Parameters:
  # filters::
  #   A hash with filters
  #
  # == Returns:
  # True if there are a domain filter
  #
  def domain_filters?(filters)
    filters.keys.each do |f|
      return true if concern_domain?(f)
    end
    false
  end

  #
  # Check if the filter is a sensor domain type
  #
  # == Parameters:
  # filter::
  #   Symbol with the name of the filter
  #
  # == Returns:
  # True if the filter is a domain type
  #
  def concern_domain?(filter)
    [:organization, :floor, :market, :campus,
     :client_building, :sensor_name].include?(filter)
  end

  #
  # Add the a sensor name filter or access point to the current @filter var.
  # This will be used on clients tracking and map views.
  #
  # == Paramenters:
  # flow_sensors::
  #   An array of flow sensors.
  #
  def add_filters_map(flow_sensors)
    #
    # If we haven't got any flow sensor, we try to get clients
    # from the access points
    #
    if flow_sensors.empty?
      @access_points = @sensor.access_points

      if @access_points.empty?
        @filter.add('wireless_station', [''])
      else
        # Add all filters
        @filter.add('wireless_station', @access_points.map(&:mac_address))
      end
    else
      # Get the name of every flow sensor
      @filter.add('sensor_name', flow_sensors.map(&:name))
    end
  end

  #
  # Set the filters. Change current filters if params[:search] is defined.
  #
  def set_source
    @source = Redborder::DruidSource.get(source: 'flow')
    @filter = Redborder::Filter.new(@source, load_filters)
  end

  def set_saved_flow_filters
    @filters = SavedTrafficFilter.where(user_id: current_user.id)
    @filters_public = SavedTrafficFilter.where
                      .not(user_id: current_user.id).select(&:public)
  end
end
