#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Controls the creation, edition and destroy of access points
#    zones. We must get the parent domain and manage the permissions
#
class ZonesController < ApplicationController
  # =================================================
  # BEFORE ACTIONS
  # =================================================
  before_action :set_current_sensor, only: [:new, :create, :search]
  before_action :set_zone,
                only: [:update, :destroy, :add_access_point,
                       :remove_access_point, :edit,
                       :add_access_points, :remove_access_points]
  before_action :check_owner_permission,
                only: [:update, :destroy, :add_access_point,
                       :add_access_points, :remove_access_points,
                       :remove_access_point, :edit]

  #
  # => Show the UI to create a new Zone
  #
  def new
    @zone = Zone.new
    @sensors = @sensor.tree_domains(true).map do |s|
      [s.path_name, s.id] if current_user.can?(:manage, s)
    end.compact
    render layout: false
  end

  #
  # => Show the UI to edit a zone
  #
  def edit
    @sensors = @zone.sensor.tree_domains(true).map do |s|
      [s.path_name, s.id] if current_user.can?(:manage, s)
    end.compact
    render layout: false
  end

  #
  # => Create a new zone into the database
  #
  def create
    @zone = Zone.new(zone_params)

    # Check owner permissions
    if check_owner_permission
      @permission = true
      # Send the javascript
      respond_to do |format|
        if @zone.save
          format.html { redirect_to view_sensor_path(@sensor), notice: "#{I18n.t('controllers.zone_created')}" }
          format.js { flash[:notice] = "#{I18n.t('controllers.zone_created')}" }
          format.json { head :no_content }
        else
          format.html { redirect_to view_sensor_path(@sensor), alert: @zone.errors.full_messages.map { |msg| "<li>#{msg}</li>" }.join }
          format.js
          format.json { render json: @zone.errors, status: :unprocessable_entity }
        end
      end
    else
      @permission = false
    end
  end

  #
  # => Update a zone
  #
  def update
    respond_to do |format|
      if @zone.update(zone_params)
        format.html { redirect_to view_sensor_path(sensor), notice: "#{I18n.t('controllers.zone_updated')}" }
        format.js { flash[:notice] = "#{I18n.t('controllers.zone_updated')}" }
        format.json { head :no_content }
      else
        format.html { redirect_to view_sensor_path(@sensor), alert: @zone.errors.full_messages.map { |msg| "<li>#{msg}</li>" }.join }
        format.js
        format.json { render json: @zone.errors, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Destroy a zone if the user has permissions
  #
  def destroy
    # Save the id to delete in the UI
    @zone_id = @zone.id

    # Clear acccess points and destroy
    @zone.access_points.clear
    @zone.destroy
  end

  #
  # => Receive the click on access point into a zone event and add it.
  #    This action return a json with the info.
  #
  def add_access_point
    access_point = AccessPoint.find(params[:access_point_id])

    if access_point.nil?
      render json: { status: 'no', error: "#{I18n.t('controllers.ap_not_exist')}"}
    else
      # Save the access point
      if @zone.access_points.include? access_point
        render json: { status: 'no', error: "#{I18n.t('controllers.duplicated')}" }
      else
        @zone.access_points << access_point
        render json: { status: 'ok', access_point_id: access_point.id, zone_id: @zone.id, zone_has_problems: @zone.down_access_points? }
      end
    end
  end

  #
  # => Receive the click on access point into a zone event and remove it.
  #    This action return a json with the info.
  #
  def remove_access_point
    access_point = AccessPoint.find(params[:access_point_id])
    id = access_point.id

    if access_point.nil?
      render json: { status: 'no', error: "#{I18n.t('controllers.ap_not_exist')}"}
    else
      # Save the access point
      if @zone.access_points.include? access_point
        @zone.access_points.delete(access_point)
        render json: { status: 'ok', access_point_id: id, zone_id: @zone.id, zone_has_problems: @zone.down_access_points? }
      else
        render json: { status: 'no', error: "#{I18n.t('controllers.ap_not_zone')}" }
      end
    end
  end

  #
  # => Allow the user to add multiple access points
  #    to a zone in one request
  #
  def add_access_points
    access_points = AccessPoint.find(params[:access_point_ids]).select{|s| s if current_user.can?(:manage, s.sensor)}
    access_points_saved = []

    if access_points.empty?
      render json: { status: 'no', error: "#{I18n.t('controllers.ap_not_exist')}"}
    else
      # Save the access points
      access_points.each do |access_point|
        if !@zone.access_points.include? access_point
          @zone.access_points << access_point
          access_points_saved << access_point.id
        end
      end

      # We have some problems with the permissions of the aps
      if access_points.size < params[:access_point_ids].size
        status = 'warn'
      else
        status = 'ok'
      end

      render json: { status: status,
                     access_point_ids: access_points_saved,
                     zone_id: @zone.id,
                     zone_name: @zone.name,
                     zone_has_problems: @zone.down_access_points? }
    end
  end

  #
  # => Allow the user to remove multiple access points
  #    to a zone in one request
  #
  def remove_access_points
    access_points = AccessPoint.find(params[:access_point_ids]).select{|s| s if current_user.can?(:manage, s.sensor)}
    access_points_saved = []

    if access_points.empty?
      render json: { status: 'no', error: "#{I18n.t('controllers.ap_not_exist')}"}
    else
      # Save the access points
      access_points.each do |access_point|
        if @zone.access_points.include? access_point
          @zone.access_points.delete(access_point)
          access_points_saved << access_point.id
        end
      end

      # We have some problems with the permissions of the aps
      if access_points.size < params[:access_point_ids].size
        status = 'warn'
      else
        status = 'ok'
      end

      render json: { status: status,
                     access_point_ids: access_points_saved,
                     zone_id: @zone.id,
                     zone_name: @zone.name,
                     zone_has_problems: @zone.down_access_points? }
    end
  end

  #
  # => Search zones by the name of the zone in the current sensor
  #
  def search
    @search = params[:search]
    sensors_ids = []
    @sensor.tree_domains(true).map do |s|
      if current_user.can?(:read, s) and !s.zones.empty?
        sensors_ids << s.id
      end
    end
    @zones = Zone.where("LOWER(zones.name) LIKE :search",
                        {
                          search: "%#{@search.downcase.split.join('%')}%"
                        })
                 .where(sensor_id: sensors_ids)
                 .order('name ASC')
  end

  private

  def set_current_sensor
    id = params[:sensor_id] || params[:id]
    @sensor = Sensor.find(id)

    if @sensor.nil? or !current_user.can?(:read, @sensor)
      respond_to do |format|
        format.html { redirect_to view_sensor_path(@sensor), alert: "#{I18n.t('controllers.no_privileges')}" }
        format.json { render json: {error: "#{I18n.t('controllers.no_privileges')}"}, status: :unauthorized }
        format.js { render json: {error: "#{I18n.t('controllers.no_privileges')}"}, status: :unauthorized }
      end
    end
  end

  def set_zone
    @zone = Zone.find(params[:id])

    if @zone.nil? or !current_user.can?(:read, @zone.sensor)
      respond_to do |format|
        format.html { redirect_to view_sensor_path(@zone.sensor), alert: "#{I18n.t('controllers.no_privileges')}" }
        format.json { render json: {error: "#{I18n.t('controllers.no_privileges')}"}, status: :unauthorized }
        format.js { render json: {error: "#{I18n.t('controllers.no_privileges')}"}, status: :unauthorized }
      end
    end
  end

  #
  # Check if the user has permissions to edit/destroy
  # in the owner domain
  #
  def check_owner_permission
    if @zone.sensor.nil?
      no_privileges sensors_path
      return false
    elsif current_user.cannot?(:manage, @zone.sensor)
      no_privileges(view_sensor_path(@zone.sensor))
      return false
    end
    true
  end

  #
  # Render or redirect a non privilege view. If the selected format is JS,
  # wthis ONLY add an advice to the user.
  #
  def no_privileges(url)
    respond_to do |format|
      format.html { redirect_to url, alert: "#{I18n.t('controllers.no_privileges')}" }
      format.json { render json: {error: "#{I18n.t('controllers.no_privileges')}"}, status: :unauthorized }
      format.js { flash[:error] = "#{I18n.t('controllers.no_privileges')}" }
    end
  end

  def set_access_point
    @access_point = AccessPoint.find(params[:id])
  end

  def zone_params
    params.require(:zone)
      .permit(:name, :sensor_id, :description)
  end
end
