#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class AccessPoint < ApplicationRecord
  belongs_to :sensor
  belongs_to :switch, class_name: 'Sensor', foreign_key: :switch_id
  has_many :zones
  has_and_belongs_to_many :zones

  validates :name, :mac_address, presence: true
  validates :mac_address, uniqueness: { case_sensitive: false }

  validate :has_floor_or_switch

  before_save :downcase_mac_address
  before_save :clean_enrichment
  before_update :update_cache, if: (:name_changed? || :mac_address_changed?)
  before_update :unallocate, if: :sensor_id_changed?
  before_update :update_access_points_in_parent_node
  before_update :update_radius, if: :mac_address_changed?

  before_destroy :delete_cache
  before_destroy :delete_access_points_in_parent_node
  before_destroy :delete_radius

  after_create :create_cache
  after_create :create_access_points_in_parent_node
  after_create :create_radius

  store :enrichment, coder: JSON

  STATUS = {
    UNKNOWN: 0,
    UP: 1,
    DOWN: 2
  }.freeze

  # Return an AP list as CSV exportable
  #
  # @param content [Array] Access Points
  # @return [String] CSV content
  #
  def self.ap_list_to_csv(content)
    header = ['Service Provider ID', 'Deployment ID', 'Organization ID',
              'Market ID', 'Building ID', 'Campus ID', 'Floor ID', 'Mac Address']
    csv = header.join(',') + "\n"
    # Iterate in aps
    content.each do |ap|
      aux_array = []
      header.each do |head|
        aux_array << ap[head]
      end
      csv += aux_array.join(',') + "\n"
    end
    # Return CSV
    csv
  end

  #
  # Convert data from Hash or Array of Hash to CSV.
  #
  # == Parameters:
  # content::
  #   Hash or Array of hashes with data to convert
  #
  # == Returns:
  # String with CSV format.
  #
  def self.to_csv(content)
    return '' unless content.present?
    csv = ''
    if content.as_json.is_a? Hash
      header = content.as_json.keys
    elsif content.as_json.is_a? Array
      # Check if First hash of the Array isn't empty.
      return '' unless content.as_json.first.present?
      header = content.as_json.first.keys
    end
    csv += header.join(',') + "\n"

    content.as_json.each do |c|
      values = c.values
      values = values.map { |o| o.nil? ? 'null' : o }
      csv += values.join(',') + "\n"
    end
    csv
  end

  #
  # Convert AccessPoint to CSV.
  #
  # == Returns:
  # String with AccessPoint in CSV format
  #
  def to_csv
    converted = as_json
    converted.keys.join(',') + "\n" + converted.values.join(',') + "\n"
  end

  #
  # Return the curretn AccessPoint as Hash, ready to convert to JSON
  #
  # == Parameters:
  # options::
  #   Hash with options
  #
  # == Returns:
  # Hash with data of this AccessPoint
  #
  def as_json(options = {})
    result = super(options)
    result['sensor_name'] = sensor.name if sensor
    result['sensor_uuid'] = sensor.uuid if sensor
    result
  end

  #
  # Check if the user has permission to edi this access point.
  #
  # == Parameters:
  # user::
  #   User object to check the permissions. Current user by default.
  #
  # == Returns:
  # Boolean indicating if user has permission.
  #
  def manage_permission?(user = User.current_user)
    return false unless user.admin?
    if sensor.present? && switch.present?
      return true if user.can?(:manage, sensor) && user.can?(:manage, switch)
    elsif sensor.present?
      return true if user.can?(:manage, sensor)
    elsif switch.present?
      return true if user.can?(:manage, switch)
    end
    false
  end

  def info
    header = ['service_provider_uuid', 'deployment_uuid', 'organization_uuid',
              'market_uuid', 'building_uuid', 'campus_uuid' ,'floor_uuid',
              'mac_address', 'status', 'last_check_in', 'name']
    domain_types = {'organization_uuid' => Sensor::ORGANIZATION, 'market_uuid' => Sensor::MARKET, 'deployment_uuid' => Sensor::DEPLOYMENT,
                    'building_uuid' => Sensor::BUILDING, 'service_provider_uuid' => Sensor::SERVICE_PROVIDER , 'campus_uuid' => Sensor::CAMPUS, 'floor_uuid' => Sensor::FLOOR}

    ap = {}
    floor = Sensor.find_by(id: self.sensor_id)
    header.each do |head|
      case head
        when 'name'
          ap[head] = name
        when 'floor_uuid'
          floor ? ap[head] = floor.uuid : ap[head] = nil
        when 'status'
          self.status ? ap[head] = self.status : ap[head] = nil
        when 'last_check_in'
          self.last_check_in ? ap[head] = self.last_check_in : ap[head] = nil
        when 'mac_address'
          ap[head] = self.mac_address
        else
          # Search the ancestor with the corresponding type
          if enrichment.present? && enrichment[head]
            ap[head] = enrichment[head]
            ap[head[0..head.index('_uuid') - 1]] = nil
          elsif floor
            # If found
            ancestor = floor.ancestors.detect{ |w| w.domain_type == domain_types[head] }
            ancestor ? ap[head] = ancestor.uuid : ap[head] = nil
            ancestor ? ap[head[0..head.index('_uuid')-1]] = ancestor.name : ap[head[0..head.index('_uuid')-1]] = nil
          else
            ap[head] = nil
          end
      end
    end
    ap
  end

  #
  # Fill latitude and longitude if Access Point's floor is located
  #
  def recalculate_lnglat
    return unless sensor.present? && sensor.located? && x && y

    # Get the matrix transformation
    overlay_data = PrimeInfrastructure::Functions.overlay_data(
      sensor.floor_dimensions,
      sensor.gps_markers
    )
    trans = overlay_data[:matrix_trans]
    offset = overlay_data[:matrix_offset]
    point = [x, y]
    lnglat =
      PrimeInfrastructure::Functions.point_to_lnglat trans, offset, point
    self.longitude = lnglat[0, 0].round(6)
    self.latitude = lnglat[1, 0].round(6)
    save
  end

  #
  # Fill X, Y if Access Point's floor is located
  #
  def recalculate_xy
    return unless sensor.present? && sensor.located? && latitude && longitude

    # Get the matrix transformation
    overlay_data = PrimeInfrastructure::Functions.overlay_data(
      sensor.floor_dimensions,
      sensor.gps_markers
    )
    trans = overlay_data[:matrix_trans]
    offset = overlay_data[:matrix_offset]
    point = [longitude, latitude]
    xy = PrimeInfrastructure::Functions.lnglat_to_point trans, offset, point
    self.x = xy[0, 0].round(4)
    self.y = xy[1, 0].round(4)
    save
  end

  def status_string
    case status
    when 0
      'Unknown'
    when 1
      'Up'
    when 2
      'Down'
    end
  end

  #
  # Returns true if AP has latitude and longitude
  #
  def map_coordinates?
    latitude && longitude
  end

  def open_radius_connection
    config = YAML.load_file("#{Rails.root}/config/database.yml")      
      
    conn = PGconn.open( 
                        :host => config['radius']['host'],
                        :port => config['radius']['port'],
                        :dbname => config['radius']['database'],
                        :user => config['radius']['username'],
                        :password => config['radius']['password']                                   
                        )
  end

  def radius_query_sql(action="select", radius_secret=nil)
    query = nil
    case action.downcase
    when "insert"
      query = "INSERT INTO nas (nasname, shortname, secret, community) VALUES ('#{mac_address}', '#{mac_address}', '#{radius_secret}', '#{mac_address}');"
    when "update"
      if mac_address_changed?
        query = "UPDATE nas SET community = '#{mac_address}', secret = '#{radius_secret}', nasname = '#{mac_address}', shortname = '#{mac_address}' WHERE community = '#{mac_address_was}';"
      else
        query = "UPDATE nas SET community = '#{mac_address}', secret = '#{radius_secret}' WHERE community = '#{mac_address}';"
      end
    when "delete"
      query = "DELETE FROM nas where community like '#{mac_address}';"
    when "select"
      if mac_address_changed?
        query = "SELECT * from nas where community like '#{mac_address_was}';"
      else
        query = "SELECT * from nas where community like '#{mac_address}';"
      end
    end
    return query
  end

  #
  # Borrar si se le pasa el flag delete a true o si no hay radius_secret
  # Sino inserta si no hay ya uno previo, en otro caso update.
  #
  def update_radius(to_delete = false)
    owner = switch || sensor
    return unless owner.present? and owner.type != Sensor::DOMAIN and mac_address.present? and (owner.is_meraki? or owner.is_flow?)
  
    to_delete = (to_delete or owner.radius_secret.nil? or owner.radius_secret.empty?) ? true : false
    
    begin
      conn = open_radius_connection
      
      query = to_delete ? radius_query_sql("delete") : 
                 conn.exec(radius_query_sql("select")).none? ? radius_query_sql("insert", owner.radius_secret) : radius_query_sql("update", owner.radius_secret)

      
      conn.exec(query) if query

      conn.close

      return true
    rescue Exception => e 
        Rails.logger.info "\n[!] Error: database connection"
        Rails.logger.info e
        conn.close if conn
    end
  end

  
  def create_radius
    update_radius
  end

  def delete_radius
    update_radius(true)
  end

  # This function update or delete the attribute "access_points" in the parent sensor
  # where this access point was made, with the id and the mac address of the access point.
  def update_access_points_in_parent_node(to_delete = false)
    owner = switch || sensor

    return unless owner.present? and owner.type != Sensor::DOMAIN and mac_address.present? and (owner.is_meraki? or owner.is_flow?)
    if to_delete
      return unless owner.chef_node.normal['access_points']
      owner.chef_node.normal['access_points'].delete(id.to_s) # to_s because chef converted it to a string
    else
      owner.chef_node.normal['access_points'] = {} unless owner.chef_node.normal['access_points']
      owner.chef_node.normal['access_points'][id] = mac_address
    end
    owner.chef_node.save
  end

  def create_access_points_in_parent_node
    update_access_points_in_parent_node
  end

  def delete_access_points_in_parent_node
    update_access_points_in_parent_node(true)
  end


  private

  def update_cache
    if name_changed?
      Rails.cache.write "rb-webui:accesspoint:#{mac_address}", name
    elsif mac_address_changed?
      Rails.cache.delete "rb-webui:accesspoint:#{mac_address_was}"
      Rails.cache.write("rb-webui:accesspoint:#{mac_address}", name) if mac_address.present?
    end
  end

  def unallocate
    self.x = self.y = self.latitude = self.longitude = nil
  end

  def has_floor_or_switch
    return if sensor.present? || switch.present?
    errors.add(:floor_or_switch,
               "#{I18n.t('models.must_speficy')}")
  end

  def create_cache
    return unless mac_address.present?
    Rails.cache.write("rb-webui:accesspoint:#{mac_address}", name)
  end

  def delete_cache
    Rails.cache.delete "rb-webui:accesspoint:#{mac_address}"
  end

  def downcase_mac_address
    mac_address.downcase!
  end

  def clean_enrichment
    enrichment.reject! { |_k, v| v.nil? }
  end

end
