#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Traffic DataSource. It takes the data from flow traffic.
# The initialization traduce multiple parameters like ports, IP Address...
# to user friendly strings.
#
class Traffic < Redborder::DataSource
  COLUMNS = Redborder::Module.get(:flow, :columns)
  SESSION_ID = Redborder::Module.get(:flow, :session_id).to_sym
  AGGREGATIONS = Redborder::Module.get(:flow, :aggregations)

  # Generate a reader attribute for the columns
  COLUMNS.map{ |k, v| v[:columns].keys + (v[:complementary_columns].try('keys') || []) }.flatten.each do |column|
    attr_reader column
  end

  # Define the column names to filter
  COLUMN_NAMES = { start_time: 'Start Time',
                   end_time: 'End Time',
                   domain: 'Domain',
                   sensor_uuid: 'Sensor UUID'
                 }

  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  # Aggregations
  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  attr_reader :timestamp, :result

  #
  # Create a Traffic object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build a Traffic model.
  #
  # == Returns:
  # Traffic object
  #
  def initialize(user, args = {})
    if args.is_a?(Hash)
      args.each do |key, value|
        instance_variable_set("@#{key}", value)
      end

      @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp'].present?
    end

    @sensor_ip = get_sensor_ip(@sensor_name, [Sensor::FLOW, Sensor::MSE, Sensor::ALE,
                                              Sensor::DEVICE, Sensor::MERAKI, Sensor::CLIENT_PROXY])

    @lan_ip_name    = NetObject.fetch(@lan_ip, [user.id, nil], @lan_ip)
    @wan_ip_name    = NetObject.fetch(@wan_ip, [user.id, nil], @wan_ip)
    @public_ip_name = NetObject.fetch(@public_ip, user.id, @public_ip)

    @lan_l4_port_name = ServiceObject.fetch(@lan_l4_port, [user.id, nil], @lan_l4_port)
    @wan_l4_port_name = ServiceObject.fetch(@wan_l4_port, [user.id, nil], @wan_l4_port)

    @l4_proto_name = ProtocolObject.fetch(@l4_proto, nil, @l4_proto)

    @client_mac_name_two = get_mac_vendor_full(@client_mac)

    @wireless_station_name = get_ap_name(@wireless_station)

    # @wan_ip_country = get_country_name(@wan_ip_country_code)
    @ip_country = get_country_name(@ip_country_code)

    @client_rssi_value = get_client_rssi_value(@client_rssi)

    # VlanObject's are general values
    @lan_vlan_name = VlanObject.fetch(@lan_vlan, user.id, @lan_vlan)
    @wan_vlan_name = VlanObject.fetch(@wan_vlan, user.id, @wan_vlan)

    # It should be got by namespace
    @client_mac_name  = MacObject.fetch(@client_mac, [user.id, nil], @client_mac)

    # Client MAC type
    @client_mac_type = get_object_type(@client_mac)

    @client_mac_inventored = @client_mac_name != @client_mac

    # Round the aggregations that must be rounded
    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      next if current_value.nil?
      if aggregation_data[:round]
        instance_variable_set("@#{aggregation}", current_value.round)
      else
        instance_variable_set("@#{aggregation}", current_value.round(3))
      end
    end
  end

  #
  # Return the Symbol of id module
  #
  def self.id
    :flow
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :traffics
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: self.controller, action: 'tops' }
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.default_aggregation
    :bytes
  end

  #
  # Return default columns for RAW
  #
  def self.default_columns_raw
    [:lan_ip, :wan_ip, :l4_proto]
  end

  #
  # Return all flow sensors
  #
  # == Parameters:
  # name::
  #   Array or String of names. If its nil, return all flow sensors.
  #
  def self.sensors(name = nil)
    types = [Sensor::FLOW, Sensor::MSE, Sensor::ALE, Sensor::DEVICE, Sensor::MERAKI, Sensor::CLIENT_PROXY]
    return Sensor.where(type: types, is_deleted: false) if name.nil?
    # No name provided
    Sensor.where(type: types, name: name, is_deleted: false)
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns, zones and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains, columns, sensors and zones
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user.sensor
                  .self_and_descendants
                  .where.not(id: 1)
                  .where(type: Sensor::DOMAIN)
  
    organizations = user.sensors_with_permission
                        .where.not(id: 1)
                        .where(domain_type: 2)
  
    f = super.merge!(
      flow_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by { |h| h[:value].to_s },
      flow_sensors: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: [Sensor::FLOW, Sensor::MSE, Sensor::ALE, Sensor::DEVICE, Sensor::MERAKI, Sensor::CLIENT_PROXY])
          .map { |c| { value: c.name, id: c.name } }
      },
      flow_zones: {
        type: :dropdown,
        value: Zone.includes(:sensor)
          .where(sensors: { id: domains.map(&:id) })
          .map { |c| { value: c.name, id: c.id } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      },
      organizations: {
        type: :dropdown,
        value: organizations.map { |c| { value: c.name, id: c.name } }
      }
    )
    # Check format
    json ? f.to_json : f
  end

  #
  # Return the filters with zones
  #
  # == Parameters:
  # search::
  #   A hash with the current filters to apply
  # user::
  #   Current user
  #
  # == Returns:
  # A hash with the filters combination of current, user, domain and zones.
  #
  def self.user_filters(search, user)
    return {} if user.nil? || search.nil?
    filters = {}
    # Filters from user
    # TODO: Don't except zones
    # filters.merge! search.except(:domain, :zone)
    filters.merge! search.except(:domain)
    # Filters from domains
    filters.merge! domain_filters(search[:domain], search[:sensor_name], user)
    # Filters from zones

    # TODO: zone field comes from MSE and there is no need
    #       that the zones have to be created on rB
    # filters.merge! zone_filters(search[:zone], user)

    filters
  end

  #
  # This method transforms the zones filters on filters
  # based on the APs
  #
  # == Parameters:
  # zones::
  #   An array of zones
  # user::
  #   Current user
  #
  # == Returns:
  # A hash with the zones filters
  #
  def self.zone_filters(zones, user)
    return {} if zones.nil? || zones.empty?

    filters = {}
    filters[:wireless_station] = {}

    zones.each do |value, filter_type|
      # Select all the APs from the zones, based on the type of the
      # filter that the user has chosen

      # zone_query = Zone.includes(:sensor)
      #                  .where(sensors: { id: Sensor.where(type: Sensor::DOMAIN)
      #                                   .select { |s| user.can?(:read, s) }.map(&:id) })

      zones = if filter_type[:operator] == 'contains'
        Zone.includes(:sensor)
        .where('LOWER(zones.name) LIKE :search', search: '%#{value.downcase}%')
        .where(sensors: { id: Sensor.where(type: Sensor::DOMAIN)
                              .select { |s| user.can?(:read, s) }.map(&:id) })
      elsif filter_type[:operator] == 'is not'
        Zone.includes(:sensor)
        .where.not(id: value)
        .where(sensors: { id: Sensor.where(type: Sensor::DOMAIN)
                              .select { |s| user.can?(:read, s) }.map(&:id) })
      else
        Zone.includes(:sensor)
        .where(id: value)
        .where(sensors: { id: Sensor.where(type: Sensor::DOMAIN)
                              .select { |s| user.can?(:read, s) }.map(&:id) })
      end

      access_points = AccessPoint.includes(:zones).where(zones: {id: zones.map(&:id)})

      # If there are no access points on these zones, filter for none
      if access_points.empty?
        filters[:wireless_station]['00:00:00:00:00:00'] = { operator: 'is' }
      else
        access_points.map(&:mac_address).each do |s|
          filters[:wireless_station][s.to_s] = { operator: 'is' }
        end
      end
    end

    filters
  end

  private

  def get_client_rssi_value(client_rssi=nil)
    Rails.cache.fetch "rb-webui:client_rssi_value:#{client_rssi}" do
      client_rssi_values = YAML.load(File.read("#{Rails.root}/lib/modules/flow/config/rssi_values.yml")).symbolize_keys
      client_rssi_values[client_rssi.to_sym]['value']
    end
  rescue
    client_rssi
  end

  def get_ap_name(mac_address)
    Rails.cache.fetch "rb-webui:accesspoint:#{mac_address}", expires_in: 1.day do
      if mac_address.present?
        AccessPoint.find_by(mac_address: mac_address).try(:name) || mac_address
      else
        'N/A'
      end
    end
  end

  def get_object_type(mac_address)
    query = "SELECT object_type FROM redborder_objects WHERE value = '#{mac_address}' AND type = 'MacObject'"
    result = ActiveRecord::Base.connection.exec_query(query)
  
    if result.present? && result.first.present?
      result.first['object_type']
    else
      'N/A'
    end
  rescue ActiveRecord::StatementInvalid => e
    Rails.logger.error("Error executing SQL query: #{e.message}")
    'N/A'
  end

end
