#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'matrix'

module PrimeInfrastructure
  module Functions
    #
    # => Calculates the rotation of the image, the points where
    #    the image should be placed on the map and also saves the
    #    image as an overlay map
    #
    #    @param dimensions Array with width and height of the image
    #    in that order
    #
    #    @param gps Array with GpsMarkers objects
    #
    #    Note: The axises always start on the upper left corner.
    #
    def self.overlay_data(dimensions, gps)
      # We need 3 GPS markers and width and height to get
      # the rotation of the image and position that image
      # on an overlay map

      return if gps.length < 3 || dimensions.length != 2

      # Parse strings to floats
      # gps.each do |gps_marker|
      #  gps_marker.each { |k, v| gps_marker[k] = v.to_f }
      # end

      # Lets get the dimensions of the map in feets
      width = dimensions[0]
      height = dimensions[1]

      # We use these matrixes below to build an expression that lets us
      # transform every point in lat-lng coordinates
      matrix = Matrix[
        [gps[0].x, gps[0].y, 0,        0,        1, 0],
        [0,        0,        gps[0].x, gps[0].y, 0, 1],
        [gps[1].x, gps[1].y, 0,        0,        1, 0],
        [0,        0,        gps[1].x, gps[1].y, 0, 1],
        [gps[2].x, gps[2].y, 0,        0,        1, 0],
        [0,        0,        gps[2].x, gps[2].y, 0, 1]    
      ]

      lat_lng_column = Matrix.column_vector(
        [gps[0].longitude, gps[0].latitude, gps[1].longitude,
         gps[1].latitude, gps[2].longitude, gps[2].latitude]
      )

      # The solution of the system above
      sol = matrix.inverse * lat_lng_column
      
      # This matrix lets us transform every feet point in lat-lng points
      base_transf = Matrix[
        [sol[0, 0], sol[1, 0]],
        [sol[2, 0], sol[3, 0]]
      ]

      # The difference between the two reference systems
      diff = Matrix.column_vector([sol[4, 0], sol[5, 0]])

      # The points on the lng-lat system
      nw_lng_lat = point_to_lnglat(base_transf, diff, [0, 0])
      ne_lng_lat = point_to_lnglat(base_transf, diff, [width, 0])
      sw_lng_lat = point_to_lnglat(base_transf, diff, [0, height])
      se_lng_lat = point_to_lnglat(base_transf, diff, [width, height]) 
      diagonal_vector = ne_lng_lat - sw_lng_lat
      center = sw_lng_lat + diagonal_vector /  2

      # The vector that goes north to their image
      north_vector = ne_lng_lat - se_lng_lat

      # The image rotation
      rotation_rad = Math.acos(north_vector[1, 0] / Math.sqrt(north_vector[0, 0]**2 + north_vector[1, 0]**2))
      rotation_deg = (north_vector[0, 0] / north_vector[0, 0].abs) * (rotation_rad * 180 / Math::PI)

      # If GPS points are wrong, rotation_rad could be NaN
      return {rotation: 0, floor_ne: "0.0,0.0", floor_sw: "0.0,0.0"} if rotation_rad.nan?
      # Lets calculate the points with the image rotated
      # Longitude is X, and Latitude is Y
      # North-east point
      ne_vector = center - ne_lng_lat
      r = Math.sqrt(ne_vector[0, 0]**2 + ne_vector[1, 0]**2)
      alpha = Math.atan2(ne_vector[1, 0], ne_vector[0, 0])
      alpha2 = alpha + rotation_rad * (north_vector[0, 0] / north_vector[0, 0].abs)
      rot_ne_lng = (center[0, 0] - r * Math.cos(alpha2)).to_f.round(6)
      rot_ne_lat = (center[1 , 0] - r * Math.sin(alpha2)).to_f.round(6)

      # North-west point
      nw_vector = center - nw_lng_lat
      r = Math.sqrt(nw_vector[0, 0]**2 + nw_vector[1, 0]**2)
      alpha = Math.atan2(nw_vector[1, 0], nw_vector[0, 0])
      alpha2 = alpha + rotation_rad * (north_vector[0, 0] / north_vector[0, 0].abs)
      rot_nw_lng = (center[0, 0] - r * Math.cos(alpha2)).to_f.round(6)
      rot_nw_lat = (center[1 , 0] - r * Math.sin(alpha2)).to_f.round(6)

      # South-west point
      sw_vector = center - sw_lng_lat
      r = Math.sqrt(sw_vector[0, 0]**2 + sw_vector[1, 0]**2)
      alpha = Math.atan2(sw_vector[1, 0], sw_vector[0, 0])
      alpha2 = alpha + rotation_rad * (north_vector[0, 0] / north_vector[0, 0].abs)
      rot_sw_lng = (center[0, 0] - r * Math.cos(alpha2)).to_f.round(6)
      rot_sw_lat = (center[1, 0] - r * Math.sin(alpha2)).to_f.round(6)

      # South-east point
      se_vector = center - se_lng_lat
      r = Math.sqrt(se_vector[0, 0]**2 + se_vector[1, 0]**2)
      alpha = Math.atan2(se_vector[1, 0], se_vector[0, 0])
      alpha2 = alpha + rotation_rad * (north_vector[0, 0] / north_vector[0, 0].abs)
      rot_se_lng = (center[0, 0] - r * Math.cos(alpha2)).to_f.round(6)
      rot_se_lat = (center[1, 0] - r * Math.sin(alpha2)).to_f.round(6)

      # If the image needs to be rotated, ne and sw points need to be recalculated

      floor_ne = [
        [rot_ne_lat, rot_sw_lat, rot_se_lat, rot_nw_lat].max,
        [rot_ne_lng, rot_sw_lng, rot_se_lng, rot_nw_lng].max
      ].join(',')
      
      floor_sw = [
        [rot_ne_lat, rot_sw_lat, rot_se_lat, rot_nw_lat].min,
        [rot_ne_lng, rot_sw_lng, rot_se_lng, rot_nw_lng].min
      ].join(',')

      {
        floor_ne: floor_ne,
        floor_sw: floor_sw,
        rotation: rotation_deg,
        matrix_trans: base_transf.to_a,
        matrix_offset: diff.to_a
      }
    end

    def self.point_to_lnglat(m_trans, m_offset, point)
      trans = Matrix[*m_trans.to_a]
      offset = Matrix[*m_offset.to_a]
      trans * Matrix.column_vector([point[0].to_f, point[1].to_f]) + offset
    end

    def self.lnglat_to_point(m_trans, m_offset, point)
      trans = Matrix[*m_trans.to_a].inverse
      offset = Matrix[*m_offset.to_a]
      trans * (Matrix.column_vector([point[0].to_f, point[1].to_f]) - offset)
    end     
  end
end
