#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# Import required libraries
require 'resolv'
require 'rest-client'
require 'json'
require 'time'
require 'logger'
require 'active_record'
require 'zk'
require 'set'

# Class for application logging
class AppLogger
  def self.logger
    @logger ||= Logger.new(STDOUT).tap { |log| log.level = Logger::DEBUG }
  end
end

class Formatter
  def self.break_line
    "-" * 80 + "\n"
  end
end

# Main class responsible for asset discovery
class AssetsDiscovery
  # Retrieves the Druid broker URL
  def self.druid_url
    @druid_url ||= retrieve_druid_broker_url
  end

  # Fetches the Druid broker URL from Zookeeper
  def self.retrieve_druid_broker_url
    zk_data = YAML.load(File.read('config/rbdruid_config.yml'))[Rails.env].symbolize_keys
    zk_connect = zk_data[:zk_connect]
    discovery_path = '/druid/discoveryPath/broker'

    brokers = fetch_brokers(zk_connect, discovery_path)
    raise "No active Druid Broker found." if brokers.empty?

    brokers.first
  end

  # Queries Zookeeper for active brokers
  def self.fetch_brokers(zk_connect, discovery_path)
    brokers_names = []

    ZK.open(zk_connect) do |zk|
      zk.children(discovery_path).shuffle.each do |broker|
        zktdata = YAML.load(zk.get("#{discovery_path}/#{broker}").first)
        if zktdata["address"] && zktdata["port"]
          brokers_names << "http://#{zktdata['address']}:#{zktdata['port']}/druid/v2/?pretty"
        end
      end
    end

    brokers_names
  end

  def initialize
    @messages_printed = Hash.new(false)
    @mac_ip_mapping = {}
    @result = ""
  end

  # Main method to execute the asset inventory job
  def execute
    @result += "Starting Assets Inventory Job...\n"

    conn = establish_db_connection

    if conn
      data_sources = fetch_data_sources_from_sensor
      total_count = process_data(conn, data_sources)
      @result += "A total of #{total_count} records (MACs and IPs) have been processed.\n"
    else
      @result += "No database connection was established.\n"
    end

    # Closes the database connection if it was established
    conn.close if conn
    @result
  rescue StandardError => e
    AppLogger.logger.error("Error: #{e.message}")
    AppLogger.logger.error(e.backtrace.join("\n"))
    "Error: #{e.message}"
  end

  private

  # Establishes a connection to the PostgreSQL database.
  def establish_db_connection
    ActiveRecord::Base.connection_pool.with_connection do
      @result += "Connection to PostgreSQL established successfully.\n"
      ActiveRecord::Base.connection
    end
  rescue ActiveRecord::ConnectionNotEstablished => e
    @result += "Failed to connect to PostgreSQL: #{e.message}"
    nil
  end

  # Fetches a list of data sources from the 'Sensor' model.
  def fetch_data_sources_from_sensor
    sensors = Sensor.where(domain_type: Sensor::NAMESPACE).pluck(:uuid, :name)
    sensors.map { |uuid, name| ["rb_flow_#{uuid}", name] }.unshift(["rb_flow", "root"])
  end

  # Processes data by iterating through a list of data sources, fetching data, and recording results.
  def process_data(conn, data_sources)
    start_time, end_time = calculate_time_interval
    @result += "Time range for the query: #{start_time} to #{end_time}\n"
    @result += Formatter.break_line

    total_mac_insert_count, total_ip_insert_count = 0, 0

    all_mac_addresses = []
    all_ips = []

    data_sources.each do |data_source, name|
      AppLogger.logger.debug("Searching in domain: #{name}\n")
      responses = fetch_and_process_data(data_source, start_time, end_time)

      process_responses(responses, all_mac_addresses, all_ips, name)
    end

    total_mac_insert_count = insert_mac_records(conn, all_mac_addresses)
    total_ip_insert_count = insert_ip_records(conn, all_ips)

    @result += "Total MACs inserted: #{total_mac_insert_count}\n"
    @result += "Total IPs inserted: #{total_ip_insert_count}\n"
    @result += Formatter.break_line

    total_mac_insert_count + total_ip_insert_count
  end

  # Processes the responses fetched from the data sources.
  def process_responses(responses, all_mac_addresses, all_ips, organization_uuid)
    responses.each do |dimension, response_json|
      response_json.each do |item|
        client_mac = item.dig('event', 'client_mac')
        lan_ip = item.dig('event', 'lan_ip')
        wan_ip = item.dig('event', 'wan_ip')

        # Extract multiple UUIDs
        uuids = {
          sensor_uuid: item.dig('event', 'sensor_uuid'),
          organization_uuid: item.dig('event', 'organization_uuid'),
          campus_uuid: item.dig('event', 'campus_uuid'),
          building_uuid: item.dig('event', 'building_uuid'),
          service_provider_uuid: item.dig('event', 'service_provider_uuid'),
          floor_uuid: item.dig('event', 'floor_uuid'),
          namespace_uuid: item.dig('event', 'namespace_uuid'),
          market_uuid: item.dig('event', 'market_uuid'),
          deployment_uuid: item.dig('event', 'deployment_uuid')
        }

        if client_mac
          all_mac_addresses << { mac_address: client_mac, uuids: uuids }
        end

        all_ips << { ip_address: lan_ip, uuids: uuids } if lan_ip
        all_ips << { ip_address: wan_ip, uuids: uuids } if wan_ip
      end
    end
  end

  # Inserts the MAC records into the database.
  def insert_mac_records(conn, mac_data)
    count = 0
    mac_data.each do |data|
      mac_address = data[:mac_address]
      uuids = data[:uuids]
      resolved_name = resolve_mac(mac_address)
      count += 1 if insert_record(conn, mac_address, resolved_name, uuids, MacObject)
    end
    count
  end

  # Inserts the IP records into the database.
  def insert_ip_records(conn, ip_data)
    count = 0
    ip_data.each do |data|
      ip_address = data[:ip_address]
      uuids = data[:uuids]
      resolved_name = resolve_ip(ip_address)
      count += 1 if insert_record(conn, ip_address, resolved_name, uuids, NetObject)
    end
    count
  end

  # Inserts a record into the database for a given value (e.g., MAC or IP address),
  # ensuring it is associated with the nearest sensor UUID based on hierarchy.
  def insert_record(conn, value, resolved_name, uuids, model_class)
    # Check if the value is '00:00:00:00:00:00' (invalid MAC address)
    # If true, the method returns false and doesn't proceed with the insertion.
    return false if value.strip == '00:00:00:00:00:00'
    
    # Retrieves the nearest sensor UUID from the provided UUIDs hash
    nearest_id = get_nearest_sensor_id(uuids)

    # Check if a record with the same value and sensor already exists in the database.
    # If it exists, return false to avoid duplicate insertion.
    exists = model_class.where("value = ? AND sensor_id = ?", value, nearest_id).exists?
    return false if exists

    # Begin a database transaction to safely insert the new record.
    # This ensures that if anything goes wrong during the insertion, the entire operation can be rolled back.
    model_class.transaction do
      # Create a new instance of the model_class (e.g., MacObject or NetObject)
      # Populate its fields with the value, resolved name, and the nearest sensor ID.
      object = model_class.new(
        value: value,
        object_type: 'UnknownDevice',  # Default object type is set to 'UnknownDevice'
        name: resolved_name,           # Resolved name for the value (e.g., human-readable MAC or IP name)
        user_id: nil,                  # No user_id is assigned
        created_at: Time.current,      # Set the current time for the creation timestamp
        updated_at: Time.current,      # Set the current time for the updated timestamp
        sensor_id: nearest_id           # Associate the record with the nearest sensor ID
      )
      # Save the new object to the database
      object.save!
    end

    # Return true to indicate that the record was successfully inserted
    true
  end

  # Helper method to get the nearest sensor ID based on hierarchy
  def get_nearest_sensor_id(uuids)
    # Filter out nil UUIDs and put them in a list
    uuid_list = [
      uuids[:floor_uuid],
      uuids[:building_uuid],
      uuids[:campus_uuid],
      uuids[:organization_uuid],
      uuids[:service_provider_uuid],
      uuids[:market_uuid],
      uuids[:deployment_uuid],
      uuids[:sensor_uuid]
    ].compact

    # If there are no UUIDs in the list, return an empty string
    return '' if uuid_list.empty?

    # Find the nearest UUID based on the smallest `depth`
    nearest_id = Sensor.where(uuid: uuid_list).order(:depth).pluck(:uuid).first

    # If no UUID is found, return an empty string
    return '' unless nearest_id

    # Find the sensor corresponding to the nearest UUID
    sensor = Sensor.find_by(uuid: nearest_id)

    # If the sensor has no parent_id or the parent doesn't exist, return the sensor ID
    return sensor.id if sensor && !sensor.parent_id

    # Find the parent sensor
    parent_sensor = Sensor.find_by(id: sensor.parent_id)

    # If the parent sensor exists, return its ID; otherwise, return the nearest sensor ID
    parent_sensor ? parent_sensor.id : sensor.id
  end

  # Main function that fetches and processes data from a source.
  def fetch_and_process_data(data_source, start_time, end_time)
    # Define the dimensions to query
    dimensions = %w[lan_ip wan_ip]
    query_responses = {}

    # Iterate over each dimension to perform an individual query.
    dimensions.each do |dimension|
      # Build the query using the build_query function.
      query = build_query(data_source, dimension, start_time, end_time)
      druid_url = AssetsDiscovery.druid_url # Get the URL for the Druid service.

      begin
        # Perform the HTTP POST request to Druid with the query in JSON format.
        response = RestClient.post(druid_url, query.to_json, { content_type: :json })
        parsed_response = JSON.parse(response.body) # Parse the response into JSON.

        # Save the response only if it's not empty.
        query_responses[dimension.to_sym] = parsed_response unless parsed_response.empty?
      rescue RestClient::ExceptionWithResponse, JSON::ParserError => e
        # Log the error if one occurs.
        AppLogger.logger.error("Error querying Druid: #{e.message}")
        AppLogger.logger.error(e.response.body) if e.respond_to?(:response) # Log the response body if available.
        raise # Re-raise the exception.
      end
    end

    query_responses # Return the responses obtained for each dimension.
  end

  # Function that builds the query for Druid with the given parameters.
  def build_query(data_source, dimension, start_time, end_time)
    {
      queryType: "groupBy", # Type of query that groups results.
      dataSource: data_source, # Specifies the data source.
      granularity: "all", # Aggregation without granularity (all data).
      dimensions: [
        "client_mac", dimension, "sensor_uuid", "service_provider_uuid",
        "namespace_uuid", "organization_uuid", "campus_uuid", "building_uuid",
        "floor_uuid", "market_uuid", "deployment_uuid"
      ], # List of dimensions to include in the results.
      context: {
        timeout: 90000, # Maximum wait time in milliseconds (90 seconds).
        skipEmptyBuckets: true # Skip empty results.
      },
      limitSpec: {
        type: "default", # Type of result limitation.
        limit: 10000, # Maximum number of results to return.
        columns: [
          { dimension: "client_mac", direction: "ascending" } # Sort by client_mac in ascending order.
        ]
      },
      intervals: ["#{start_time}/#{end_time}"], # Time interval for the query.
      filter: {
        type: "regex", # Filter type based on a regular expression.
        dimension: dimension, # Dimension on which the filter is applied.
        pattern: /^(10\.|172\.(1[6-9]|2\d|3[01])\.|192\.168\.|fc[0-9a-fA-F]{2}:|fd[0-9a-fA-F]{2}:)/ # Regular expression to filter private IP addresses or specific prefixes.
      }
    }
  end

  # Function that calculates the time interval (last hour).
  # Returns two values: the start time (one hour ago) and the end time (now).
  def calculate_time_interval
    now = Time.now # Current time.
    start_time = (now - 3600).utc.iso8601 # Calculate the start time by subtracting one hour.
    end_time = now.utc.iso8601 # End time (current).
    [start_time, end_time] # Return the time interval.
  end

  # Function that attempts to resolve a MAC address.
  # If the MAC address cannot be resolved, it returns the original MAC.
  def resolve_mac(mac)
    @mac_ip_mapping[mac] || mac # Try to get the MAC from the mapping; if not found, return the original MAC.
  end

  # Function that attempts to resolve an IP address to its domain name.
  # If it cannot be resolved, it returns the original IP.
  def resolve_ip(ip)
    Resolv.getname(ip) # Try to resolve the domain name associated with the IP.
  rescue Resolv::ResolvError
    ip # If an error occurs while resolving, return the IP unchanged.
  end
end
