#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class AssetsInventoryJob < Struct.new(:user_id, :verbose)
      include Redborder::Jobs::JobHelper
      attr_accessor :content

      def perform
        content = "Node executing this job: #{`hostname`.chomp}\n\n"

        begin
          assets_discovery = AssetsDiscovery.new
          content += assets_discovery.execute.to_s

          check_success = $?.success?
          raise content unless check_success
          is_job_ready = true
        rescue => e
          content += "Error: #{e.message}\n"
          e.backtrace.each { |line| content += "#{line}\n" }
          logit "#{e.class}: #{e.message}" if verbose
          raise e
        ensure
          StoredDelayedJob.create(job: "AssetsInventoryJob", user_id: user_id, content: content)
          if is_job_ready
            attempts = Delayed::Job.where(handler: self.to_yaml).pluck(:attempts).first || 0
            next_run_time = reschedule_at(Time.now, attempts)
            Delayed::Job.enqueue(self.class.new(user_id, verbose), run_at: next_run_time)
          end
        end
      end

      def reschedule_at(current_time, attempts)
        if Setting.assets_update_periodicity == "weekly"
          hour, minute = Setting.assets_update_time.split(":").map(&:to_i)
          current_time.beginning_of_week.advance(weeks: 1, hours: hour, minutes: minute)
        elsif Setting.assets_update_periodicity == "daily"
          hour, minute = Setting.assets_update_time.split(":").map(&:to_i)
          current_time.beginning_of_day.advance(days: 1, hours: hour, minutes: minute)
        else
          current_time.advance(hours: 12)
        end
      end
    end
  end
end
