#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# This job uses SNMP4EM, gem that extends Ruby-SNMP to use the
# asynchronous EventMachine library for added performance and
# scalabilit. This allows code to scale monitoring applications
# to access a very hight number of devices.
#
# Supports only SNMPv2c
#

module Redborder
  module Jobs
    class CheckAccessPointsStatusJob < Struct.new(:verbose)
      include Redborder::Jobs::JobHelper
      # General OIDs
      NAME = '1.3.6.1.4.1.9.9.513.1.1.1.1.5'
      MAC = '1.3.6.1.4.1.9.9.513.1.2.3.1.1'
      CLIENTS = '1.3.6.1.4.1.9.9.513.1.1.1.1.54'
      UPTIME = '1.3.6.1.4.1.9.9.513.1.1.1.1.6'
      # Location OIDs
      LAT = '1.3.6.1.4.1.9.9.616.1.6.1.1.3'
      LNG = '1.3.6.1.4.1.9.9.616.1.6.1.1.4'
      # Meraki OIDs
      MERAKI = '1.3.6.1.4.1.29671.1.1.4.1'
      MERAKI_MAC = '1.3.6.1.4.1.29671.1.1.4.1.1.'
      MERAKI_NAME = '1.3.6.1.4.1.29671.1.1.4.1.2.'
      MERAKI_IP = '1.3.6.1.4.1.29671.1.1.4.1.7.'
      MERAKI_STATUS = '1.3.6.1.4.1.29671.1.1.4.1.3.'
      MERAKI_CLIENTS = '1.3.6.1.4.1.29671.1.1.4.1.5'

      def perform
        # Loop for every FLOW sensor
        Sensor.where(type: Sensor::FLOW, is_deleted: false)
          .where.not(ip: nil).each do |flow|
          snmp_community = get_snmp_community(flow)
          snmp_version = get_snmp_version(flow)
          next if snmp_version == :SNMPv1

          begin
            EM.run do
              snmp =
                SNMP4EM::Manager.new(
                  host: flow.ip,
                  community: snmp_community,
                  version: snmp_version
                )
              request =
              snmp.bulkwalk([NAME, MAC, CLIENTS,
                             UPTIME, LAT, LNG])
              request.errback do
                EM.stop
              end
              request.callback do |response|
                # All Access Points DOWN so we'll set UP SNMP detected APs
                if AccessPoint.where(switch_id: flow.id).count > 0
                  AccessPoint.where(switch_id: flow.id).update_all(
                    status: AccessPoint::STATUS[:DOWN],
                    clients: 0,
                    uptime: 0)
                  # Update Sensor last check in
                  flow.property[:check_in_time] = Time.now.to_i
                  flow.save
                  # Handle response
                  flow_response(flow, response)
                end
                EM.stop
              end
            end
          rescue SNMP::RequestTimeout
            next
          end
        end
        # Loop for every MERAKI sensor
        Sensor.where(type: Sensor::MERAKI, is_deleted: false)
          .where.not(ip: nil).each do |meraki|
          snmp_community = get_snmp_community(meraki)
          snmp_version = get_snmp_version(meraki)
          next if snmp_version == :SNMPv1
          begin
            EM.run do
              snmp = SNMP4EM::Manager.new(
                host: meraki.ip,
                port: 161_00,
                community: snmp_community,
                version: snmp_version
              )
              request = snmp.bulkwalk(MERAKI)
              request.errback do
                EM.stop
              end
              request.callback do |response|
                # Update Sensor last check in
                meraki.property[:check_in_time] = Time.now.to_i
                meraki.save
                # Handle response
                meraki_response(meraki, response)
                EM.stop
              end
            end
          rescue SNMP::RequestTimeout
            next
          end
        end
        # Enqueue the job again
        Delayed::Job.enqueue(Redborder::Jobs::CheckAccessPointsStatusJob
        .new(true), priority: 10, run_at: Time.now + 5.minutes)
      end

      #
      # Get Sensor's SNMP community
      #   Params:
      #     sensor: target sensor
      #   Returns:
      #     sensor's snmp community
      #
      def get_snmp_community(sensor)
        sensor.snmp_community.present? ? sensor.snmp_community : 'public'
      end

      #
      # Get Sensor's SNMP version
      #   Params:
      #     sensor: target sensor
      #   Returns:
      #     sensor's snmp version symbol
      #
      def get_snmp_version(sensor)
        sensor.snmp_version == '1' ? :SNMPv1 : :SNMPv2c
      end

      #
      # Get value for SNMP OID + Object
      #   Params:
      #     response: SNMP response
      #     oid: general OID
      #     obj: specific object's OID
      #   Returns:
      #     value
      #
      def obj_value(response, oid, obj)
        result = response[oid].detect { |k, _v| k.start_with?(oid + obj) }
        result.nil? ? nil : result.last
      end

      def meraki_obj_value(response, oid, obj)
        result = response[MERAKI].detect { |k, _v| k.start_with?(oid + obj) }
        result.nil? ? nil : result.last
      end

      #
      # Update AP or create it if it doesn't exist
      #   Params:
      #     ap_params: access point's attributes
      #
      def update_ap(ap_params)
        ap = AccessPoint.find_by(mac_address: ap_params[:mac_address])
        if ap.nil?
          AccessPoint.create(ap_params)
        else
          ap.update(ap_params)
        end
      end

      #
      # Handle Meraki SNMP response to BulkWalk request
      #   Params:
      #     meraki: Meraki Sensor
      #     response: SNMP response
      #
      def meraki_response(meraki, response)
        objects =
          response[MERAKI].keys.select { |k| k.start_with?(MERAKI_NAME) }
          .map { |k| k.gsub(MERAKI_NAME, '') }
        objects.each do |obj|
          ap_params = {
            mac_address: meraki_obj_value(response, MERAKI_MAC, obj)
                         .unpack('H2H2H2H2H2H2').join(':'),
            name: meraki_obj_value(response, MERAKI_NAME, obj),
            ip_address: meraki_obj_value(response, MERAKI_IP, obj),
            clients: meraki_obj_value(response, MERAKI_CLIENTS, obj),
            last_check_in: Time.now,
            switch_id: meraki.id
          }
          ap_params[:status] =
            if meraki_obj_value(response, MERAKI_STATUS, obj) == 1
              AccessPoint::STATUS[:UP]
            else
              AccessPoint::STATUS[:DOWN]
            end
          update_ap(ap_params)
        end
      end

      #
      # Handle Flow SNMP response to BulkWalk request
      #   Params:
      #     flo: Flow Sensor
      #     response: SNMP response
      #
      def flow_response(flow, response)
        objects = response[NAME].keys.map { |k| k.gsub(NAME, '') }
        objects.each do |obj|
          lat = obj_value(response, LAT, obj)
          lng = obj_value(response, LNG, obj)
          ap_params = {
            name: obj_value(response, NAME, obj),
            mac_address: obj_value(response, MAC, obj)
                         .unpack('H2H2H2H2H2H2').join(':'),
            clients: obj_value(response, CLIENTS, obj),
            uptime: obj_value(response, UPTIME, obj) / 100,
            last_check_in: Time.now,
            switch_id: flow.id,
            switch_name: flow.ip,
            status: AccessPoint::STATUS[:UP]
          }
          ap_params[:latitude] = lat.to_f if lat.present?
          ap_params[:longitude] = lng.to_f if lng.present?
          update_ap(ap_params)
        end
      end
    end
  end
end
