####################################################################### 
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class ZabbixInventoryJob < Struct.new(:user_id, :verbose)
      include Redborder::Jobs::JobHelper
      attr_accessor :content
  
      def perform
        content = "Node executing this job: #{`hostname`.chomp}\n\n"
  
        unless Integration.zabbix_params_present?
          logit("Zabbix parameters are missing. Job will not execute.") if verbose
          return
        end
  
        begin
          params_zabbix = get_params_zabbix
          logit("Params Zabbix: #{params_zabbix}") if verbose
  
          url = params_zabbix[:zabbix_url]
          user = params_zabbix[:zabbix_user]
          password = params_zabbix[:zabbix_password]
          token = params_zabbix[:zabbix_token]
  
          zabbix_interaction = ZabbixInteraction.new(url, user, password, token)
  
          content += zabbix_interaction.execute.to_s
  
          check_success = $?.success?
          raise content unless check_success
          is_job_ready = true
        rescue Exception => e
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          e.set_backtrace([])
          logit "#{e.class}: #{e.message}" if verbose
          raise e
        ensure
          StoredDelayedJob.create(job: "ZabbixInventoryJob", user_id: user_id, content: content)
  
          if Integration.zabbix_params_present?
            if is_job_ready
              attempts = Delayed::Job.where(handler: self.to_yaml).pluck(:attempts).first || 0
              next_run_time = reschedule_at(Time.now, attempts)
              Delayed::Job.enqueue(self.class.new(user_id, verbose), run_at: next_run_time)
            end
          else
            logit("Zabbix parameters are missing. Job will be stopped and removed.") if verbose
            delete_delayed_job
          end
        end
      end
  
      def reschedule_at(current_time, attempts)
        zabbix_params = get_params_zabbix
        periodicity = zabbix_params[:zabbix_update_periodicity]
        update_time = zabbix_params[:zabbix_update_time]
  
        case periodicity
        when "weekly"
          hour, minute = update_time.split(":").map(&:to_i)
          current_time.beginning_of_week.advance(weeks: 1, hours: hour, minutes: minute)
        when "daily"
          hour, minute = update_time.split(":").map(&:to_i)
          current_time.beginning_of_day.advance(days: 1, hours: hour, minutes: minute)
        when "hourly"
          current_time.advance(hours: 1)
        else
          current_time.advance(hours: 12)
        end
      end
  
      private
  
      def get_params_zabbix
        zabbix_strategy = Integrations::ZabbixStrategy.new
        params = zabbix_strategy.saved_params
  
        required_params = {
          zabbix_url: nil,
          zabbix_token: nil,
          zabbix_user: nil,
          zabbix_password: nil,
          zabbix_update_periodicity: nil,
          zabbix_update_time: nil
        }
  
        required_params.each_key do |key|
          unless params[key]
            raise StandardError, "Error: Missing required Zabbix parameter: #{key}"
          end
          required_params[key] = params[key]
        end
  
        required_params
      end
  
      def delete_delayed_job
        Delayed::Job.where(handler: self.to_yaml).each(&:destroy)
      end
    end
  end
  
end
