#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorAle
    # Id of the sensor
    ALE = 11

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [ALE]

        # Import values for ALE
        self.import_columns_for[ALE] =
          [:stream, :snmp_version, :version_ale,
           :send_http, { homenets: [:name, :value] }, :name, :controller_ip, :rest_api_user, :rest_api_key ].freeze

        # Alias the name of IP attribute
        alias_attribute :stream, :ip

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------

        # IPv4 Regex
        @ipv4_regex = %r{\A(#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})
                           (\/([0-9]|[0-2][0-9]?|3[0-2]))?\z}x

        validates :stream, presence: true,
                  uniqueness: { scope: :real_parent_id },
                  format: {
                    with: @ipv4_regex,
                    message: "#{I18n.t 'lib_flow.invalid'}"
                  },
                  if: proc { type == ALE }

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'lib_flow.only'}" },
                                   if: proc { type == ALE }

        validates :name, presence: true,
                 if: proc { type == ALE }

        validates :controller_ip, presence: true,
                format: {
                  with: @ipv4_regex,
                  message: "#{I18n.t 'lib_flow.invalid'}"
                },
                if: proc { type == ALE }

        validates :rest_api_user, presence: true,
                if: proc { type == ALE }

        validates :rest_api_key, presence: true,
                if: proc { type == ALE }
        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :ale

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_update :update_cache_ale,
                      if: proc { type == ALE && ip_changed? }
        after_create :create_cache_ale, if: proc { type == ALE }
        after_create :create_chef_node_ale,
                     if: proc { type == ALE }
        after_update :update_stream, if: proc { type == ALE }

        after_save :data_to_client_proxy_ale,
                   if: proc { type == ALE && real_parent_id }
        before_destroy :delete_data_from_client_proxy_ale,
                       if: proc { type == ALE && real_parent_id }

        # -------------------------------------------------------------
        # Chef
        # -------------------------------------------------------------

        has_chef_property :send_http, %w(redborder send_http)

        # -------------------------------------------------------------
        # Accessors
        # -------------------------------------------------------------
        store_accessor :property, :controller_ip
        store_accessor :property, :rest_api_user
        store_accessor :property, :rest_api_key
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{ALE}") do
        'sensors/ale_sensor'
      end
    end

    module InstanceMethods

      def is_ale?
        type == ALE
      end

      # Return the descendant ale sensors of the domain
      def get_ale
        self.all_children(true).where(type: Sensor::ALE)
      end

      define_method("icon_#{ALE}") do
        'icons/mse.png'
      end

      def version_ale
        property[:version_ale]
      end

      def version_ale=(ver)
        property[:version_ale] = ver
      end

      define_method("full_name_#{ALE}") do
        name
      end

      define_method("secondary_name_#{ALE}") do
        name
      end

      define_method("complete_name_#{ALE}") do
        name
      end

      define_method("type_to_s_#{ALE}") do
        "ale"
      end

      define_method("type_humanize_#{ALE}") do
        "ALE"
      end

      define_method("chef_node_#{ALE}") do
        Chef::Node.load("rbale-#{id}") rescue nil
      end

      define_method("chef_client_#{ALE}") do
        Chef::ApiClient.load("rbale-#{id}")
      end

      define_method("monitors_to_chef_#{ALE}") do
        node = chef_node
        node.normal[:redborder][:monitors] = monitor_categories_to_chef
        node.save
      end

      define_method("run_list_text_#{ALE}") do
        "role[ale-sensor]"
      end

      private

      def create_cache_ale
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def update_cache_ale
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def create_chef_node_ale
        node_json = {
          'name' => "rbale-#{id}",
          'run_list' => ['role[ale-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'stream' => stream,
            'controller_ip' => controller_ip,
            'rest_api_user' => rest_api_user,
            'rest_api_key' => rest_api_key,
            'redborder' => {
              'sensor_uuid' => uuid,
              'monitors' => [],
              'rbname' => name,
              'stream' => stream,
              'controller_ip' => controller_ip,
              'rest_api_user' => rest_api_user,
              'rest_api_key' => rest_api_key,
              'parent_id' => real_parent_id
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def save_version
        return if property[:version_ale].nil?
        chef_node.normal['rbversion'] = property[:version_ale]
        chef_node.save
      end

      def update_stream
        role = chef_role
        node = chef_node
        role.override_attributes["stream"] = stream
        role.override_attributes["controller_ip"] = controller_ip
        role.override_attributes["rest_api_user"] = rest_api_user
        role.override_attributes["rest_api_key"] = rest_api_key
        role.override_attributes["rbname"] = name
        node.normal["stream"] = stream
        node.normal["controller_ip"] = controller_ip
        node.normal["rest_api_user"] = rest_api_user
        node.normal["rest_api_key"] = rest_api_key
        node.normal["rbname"] = name
        
        node.normal["redborder"]["stream"] = stream
        node.normal["redborder"]["controller_ip"] = controller_ip
        node.normal["redborder"]["rest_api_user"] = rest_api_user
        node.normal["redborder"]["rest_api_key"] = rest_api_key
        node.normal["redborder"]["rbname"] = name
        

        role.save
        node.save
      end

      #
      # Return the sensor as json with variables from ALE
      #
      def as_json_with_ale(options = {})
        result = as_json_without_ale(options)
        if self.is_ale?
          result.merge!(
            {
              access_points: access_points,
              snmp_version: snmp_version,
              version_ale: version_ale,
              name: name,
              stream: stream,
              controller_ip: controller_ip,
              rest_api_user: rest_api_user,
              rest_api_key: rest_api_key,
              send_http: send_http
            }.as_json
          )
        end
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{ALE}") do
        update_chef_domain_values_in_node
        data_to_client_proxy_ale if real_parent
      end

      protected

      # Create the chef role for ALE sensors
      #
      define_method "create_chef_role_#{ALE}" do
        run_list = ['role[ale-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{ALE}" do
        chef_node.destroy
      end

      def data_to_client_proxy_ale
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['ale'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['ale'][name] = chef_node['redborder']
        role.save
      end

      def delete_data_from_client_proxy_ale
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['ale'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorAle)
