#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorArubacentral
    # Id of the sensor
    ARUBACENTRAL = 14

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [ARUBACENTRAL]

        # Import values for ARUBACENTRAL
        self.import_columns_for[ARUBACENTRAL] =
          [:snmp_version,
           :send_http,
           { homenets: [:name, :value] },
           :name,
           :arubacentral_url,
           :arubacentral_username,
           :arubacentral_password,
           :arubacentral_client_id,
           :arubacentral_client_secret,
           :arubacentral_customer_id ].freeze

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------

        # IPv4 Regex
        @ipv4_regex = %r{\A(#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})
                           (\/([0-9]|[0-2][0-9]?|3[0-2]))?\z}x

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'lib_flow.only'}" },
                                   if: proc { type == ARUBACENTRAL }

        validates :name, presence: true,
                 if: proc { type == ARUBACENTRAL }

        validates :arubacentral_url, presence: true,
                if: proc { type == ARUBACENTRAL }

        validates :arubacentral_username, presence: true,
                if: proc { type == ARUBACENTRAL }

        validates :arubacentral_password, presence: true,
                if: proc { type == ARUBACENTRAL }

        validates :arubacentral_client_id, presence: true,
                if: proc { type == ARUBACENTRAL }

        validates :arubacentral_client_secret, presence: true,
                if: proc { type == ARUBACENTRAL }

        validates :arubacentral_customer_id, presence: true,
                if: proc { type == ARUBACENTRAL }

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :arubacentral

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_update :update_cache_arubacentral,
                      if: proc { type == ARUBACENTRAL }
        after_create :create_cache_arubacentral,
                      if: proc { type == ARUBACENTRAL }
        after_create :create_chef_node_arubacentral,
                      if: proc { type == ARUBACENTRAL }
        after_update :update_stream,
                      if: proc { type == ARUBACENTRAL }
        after_save :data_to_client_proxy_arubacentral,
                    if: proc { type == ARUBACENTRAL && real_parent_id }
        before_destroy :delete_data_from_client_proxy_arubacentral,
                       if: proc { type == ARUBACENTRAL && real_parent_id }

        # -------------------------------------------------------------
        # Chef
        # -------------------------------------------------------------

        has_chef_property :send_http, %w(redborder send_http)

        # -------------------------------------------------------------
        # Accessors
        # -------------------------------------------------------------
        store_accessor :property, :arubacentral_url
        store_accessor :property, :arubacentral_username
        store_accessor :property, :arubacentral_password
        store_accessor :property, :arubacentral_client_id
        store_accessor :property, :arubacentral_client_secret
        store_accessor :property, :arubacentral_customer_id
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{ARUBACENTRAL}") do
        'sensors/arubacentral_sensor'
      end
    end

    module InstanceMethods

      def is_arubacentral?
        type == ARUBACENTRAL
      end

      # Return the descendant arubacentral sensors of the domain
      def get_arubacentral
        self.all_children(true).where(type: Sensor::ARUBACENTRAL)
      end

      define_method("icon_#{ARUBACENTRAL}") do
        'icons/mse.png'
      end


      define_method("full_name_#{ARUBACENTRAL}") do
        name
      end

      define_method("secondary_name_#{ARUBACENTRAL}") do
        name
      end

      define_method("complete_name_#{ARUBACENTRAL}") do
        name
      end

      define_method("type_to_s_#{ARUBACENTRAL}") do
        "arubacentral"
      end

      define_method("type_humanize_#{ARUBACENTRAL}") do
        "ARUBACENTRAL"
      end

      define_method("chef_node_#{ARUBACENTRAL}") do
        Chef::Node.load("rbarubacentral-#{id}") rescue nil
      end

      define_method("chef_client_#{ARUBACENTRAL}") do
        Chef::ApiClient.load("rbarubacentral-#{id}")
      end

      define_method("monitors_to_chef_#{ARUBACENTRAL}") do
        node = chef_node
        node.normal[:redborder][:monitors] = monitor_categories_to_chef
        node.save
      end

      define_method("run_list_text_#{ARUBACENTRAL}") do
        "role[arubacentral-sensor]"
      end

      private

      # TODO: Check if this is a good key or whe need to do url + username to be unique
      def create_cache_arubacentral
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", arubacentral_username
      end

      def update_cache_arubacentral
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", arubacentral_username
      end

      def create_chef_node_arubacentral
        node_json = {
          'name' => "rbarubacentral-#{id}",
          'run_list' => ['role[arubacentral-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'arubacentral_url' => arubacentral_url,
            'arubacentral_username' => arubacentral_username,
            'arubacentral_password' => arubacentral_password,
            'arubacentral_client_id' => arubacentral_client_id,
            'arubacentral_client_secret' => arubacentral_client_secret,
            'arubacentral_customer_id' => arubacentral_customer_id,
            'redborder' => {
              'sensor_uuid' => uuid,
              'monitors' => [],
              'rbname' => name,
              'arubacentral_url' => arubacentral_url,
              'arubacentral_username' => arubacentral_username,
              'arubacentral_password' => arubacentral_password,
              'arubacentral_client_id' => arubacentral_client_id,
              'arubacentral_client_secret' => arubacentral_client_secret,
              'arubacentral_customer_id' => arubacentral_customer_id,
              'parent_id' => real_parent_id
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def update_stream
        role = chef_role
        node = chef_node
        role.override_attributes["arubacentral_url"] = arubacentral_url
        role.override_attributes["arubacentral_username"] = arubacentral_username
        role.override_attributes["arubacentral_password"] = arubacentral_password
        role.override_attributes["arubacentral_client_id"] = arubacentral_client_id
        role.override_attributes["arubacentral_client_secret"] = arubacentral_client_secret
        role.override_attributes["arubacentral_customer_id"] = arubacentral_customer_id
        role.override_attributes["rbname"] = name
        node.normal["arubacentral_url"] = arubacentral_url
        node.normal["arubacentral_username"] = arubacentral_username
        node.normal["arubacentral_password"] = arubacentral_password
        node.normal["arubacentral_client_id"] = arubacentral_client_id
        node.normal["arubacentral_client_secret"] = arubacentral_client_secret
        node.normal["arubacentral_customer_id"] = arubacentral_customer_id
        node.normal["rbname"] = name

        node.normal["redborder"]["arubacentral_url"] = arubacentral_url
        node.normal["redborder"]["arubacentral_username"] = arubacentral_username
        node.normal["redborder"]["arubacentral_password"] = arubacentral_password
        node.normal["redborder"]["arubacentral_client_id"] = arubacentral_client_id
        node.normal["redborder"]["arubacentral_client_secret"] = arubacentral_client_secret
        node.normal["redborder"]["arubacentral_customer_id"] = arubacentral_customer_id
        node.normal["redborder"]["rbname"] = name


        role.save
        node.save
      end

      #
      # Return the sensor as json with variables from ARUBACENTRAL
      #
      def as_json_with_arubacentral(options = {})
        result = as_json_without_arubacentral(options)
        if self.is_arubacentral?
          result.merge!(
            {
              access_points: access_points,
              snmp_version: snmp_version,
              name: name,
              arubacentral_url: arubacentral_url,
              arubacentral_username: arubacentral_username,
              arubacentral_password: arubacentral_password,
              arubacentral_client_id: arubacentral_client_id,
              arubacentral_client_secret: arubacentral_client_secret,
              arubacentral_customer_id: arubacentral_customer_id,
              send_http: send_http
            }.as_json
          )
        end
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{ARUBACENTRAL}") do
        update_chef_domain_values_in_node
        data_to_client_proxy_arubacentral if real_parent
      end

      protected

      # Create the chef role for ARUBACENTRAL sensors
      #
      define_method "create_chef_role_#{ARUBACENTRAL}" do
        run_list = ['role[arubacentral-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{ARUBACENTRAL}" do
        chef_node.destroy
      end

      def data_to_client_proxy_arubacentral
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['arubacentral'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['arubacentral'][name] = chef_node['redborder']
        role.save
      end

      def delete_data_from_client_proxy_arubacentral
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['arubacentral'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorArubacentral)
