#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorDevice
    # Id of the sensor
    DEVICE = 12

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [DEVICE]

        # Import values for DEVICE
        self.import_columns_for[DEVICE] =
          [:device_ip, :snmp_version, :snmp_community, :protocol,
           :send_http, { homenets: [:name, :value] }, :name, :rest_api_user, :rest_api_password ].freeze

        # Alias the name of IP attribute
        alias_attribute :device_ip, :ip

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------

        # IPv4 Regex
        @ipv4_regex = %r{\A(#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})
                           (\/([0-9]|[0-2][0-9]?|3[0-2]))?\z}x

        validates :device_ip, presence: true,
                  #uniqueness: { scope: :real_parent_id },
                  format: {
                    with: @ipv4_regex,
                    message: "#{I18n.t 'lib_flow.invalid'}"
                  },
                  if: proc { type == DEVICE }

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'lib_flow.only'}" },
                                   if: proc { type == DEVICE }

        validates :name, presence: true,
                 if: proc { type == DEVICE }


        validates :rest_api_user, presence: false,
                if: proc { type == DEVICE }

        validates :rest_api_password, presence: false,
                if: proc { type == DEVICE }
        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :device

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_update :update_cache_device,
                      if: proc { type == DEVICE && ip_changed? }
        after_create :create_cache_device, if: proc { type == DEVICE }
        after_create :create_chef_node_device,
                     if: proc { type == DEVICE }
        after_update :update_device_ip, if: proc { type == DEVICE }

        after_save :data_to_client_proxy_device,
                   if: proc { type == DEVICE && real_parent_id }
        before_destroy :delete_data_from_client_proxy_device,
                       if: proc { type == DEVICE && real_parent_id }

        # -------------------------------------------------------------
        # Chef
        # -------------------------------------------------------------

        has_chef_property :send_http, %w(redborder send_http)

        # -------------------------------------------------------------
        # Accessors
        # -------------------------------------------------------------
        store_accessor :property, :device_ip
        store_accessor :property, :rest_api_user
        store_accessor :property, :rest_api_password
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{DEVICE}") do
        'sensors/device_sensor'
      end
    end

    module InstanceMethods

      def is_device?
        type == DEVICE
      end

      # Return the descendant device sensors of the domain
      def get_device
        self.all_children(true).where(type: Sensor::DEVICE)
      end

      define_method("icon_#{DEVICE}") do
        'icons/monitor_device.svg'
      end

      def protocol
        property[:protocol]
      end

      def protocol=(proto)
        property[:protocol] = proto
      end

      define_method("full_name_#{DEVICE}") do
        name
      end

      define_method("secondary_name_#{DEVICE}") do
        name
      end

      define_method("complete_name_#{DEVICE}") do
        name
      end

      define_method("type_to_s_#{DEVICE}") do
        "device"
      end

      define_method("type_humanize_#{DEVICE}") do
        "DEVICE"
      end

      define_method("chef_node_#{DEVICE}") do
        Chef::Node.load("rbdevice-#{id}") rescue nil
      end

      define_method("chef_client_#{DEVICE}") do
        Chef::ApiClient.load("rbdevice-#{id}")
      end

      define_method("monitors_to_chef_#{DEVICE}") do
        node = chef_node
        node.normal[:redborder][:monitors] = monitor_categories_to_chef
        node.save
      end

      define_method("run_list_text_#{DEVICE}") do
        "role[device-sensor]"
      end

      private

      def create_cache_device
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def update_cache_device
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def create_chef_node_device
        node_json = {
          'name' => "rbdevice-#{id}",
          'run_list' => ['role[device-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'ipaddress' => ip,
            'redborder' => {
              'monitors' => [],
              'sensor_uuid' => uuid,
              'ipaddress' => ip,
              'protocol' => protocol,
              'rest_api_user' => rest_api_user,
              'rest_api_password' => rest_api_password,
              'parent_id' => real_parent_id
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def save_protocol
        return if property[:protocol].nil?
        chef_node.normal['protocol'] = property[:protocol]
        chef_node.save
      end

      def update_device_ip
        role = chef_role
        node = chef_node
        role.override_attributes["ipaddress"] = ip
        node.normal["ipaddress"] = ip
        node.normal["redborder"]["ipaddress"] = ip

        node.normal["redborder"]["rest_api_user"] = rest_api_user
        node.normal["redborder"]["rest_api_password"] = rest_api_password
        node.normal["redborder"]["protocol"] = protocol

        role.save
        node.save
      end

      #
      # Return the sensor as json with variables from DEVICE
      #
      def as_json_with_device(options = {})
        result = as_json_without_device(options)
        if self.is_device?
          result.merge!(
            {
              access_points: access_points,
              snmp_community: snmp_community,
              snmp_version: snmp_version,
              protocol: protocol,
              name: name,
              device_ip: ip,
              rest_api_user: rest_api_user,
              rest_api_password: rest_api_password,
              send_http: send_http
            }.as_json
          )
        end
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{DEVICE}") do
        update_chef_domain_values_in_node
        data_to_client_proxy_device if real_parent
      end

      protected

      # Create the chef role for DEVICE sensors
      #
      define_method "create_chef_role_#{DEVICE}" do
        run_list = ['role[device-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{DEVICE}" do
        chef_node.destroy
      end

      def data_to_client_proxy_device
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['device'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['device'][name] = chef_node['redborder']
        role.save
      end

      def delete_data_from_client_proxy_device
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['device'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorDevice)
