#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero Public License for more details.
# You should have received a copy of the GNU Affero Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SensorFlow

    FLOW = 5

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do

        include StorageFile
        storage_file :captive_portal_background_image, path: '/rb-webui/:class/:id/:filename'
        storage_file :captive_portal_logo_image, path: '/rb-webui/:class/:id/logo/:filename'

        # A ssensor has many zones
        has_many :zones

        self.available_sensors += [FLOW]
        # Import values for flow
        self.import_columns_for[FLOW] =
          [:ip, :nmsp, :nmsp_client_stats, :nmsp_wireless_health, :spanport, 
           :captive_portal_background_image, :captive_portal_background_image_file_name, :captive_portal_background_image_content_type, :captive_portal_background_image_file_size, :captive_portal_background_image_updated_at, 
           :captive_portal_logo_image, :captive_portal_logo_image_file_name, :captive_portal_logo_image_content_type, :captive_portal_logo_image_file_size, :captive_portal_logo_image_updated_at, 
           :captive_portal_name, :captive_portal_name_color, :captive_portal_background_color,
           :dns_ptr_client, :dns_ptr_target, :snmp_community,
           :snmp_version, { homenets: [:name, :value] }, { routers_macs: [:name, :value] },
           :radius_secret, :observation_id, :license_id].freeze

        # IPv4 Regex
        @ipv4_regex = %r{\A(#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})\.
                           (#{Resolv::IPv4::Regex256})
                           (\/([0-9]|[0-2][0-9]?|3[0-2]))?\z}x
        # Validates IP
        validates :ip,
                  presence: true,
                  format: {
                    with: @ipv4_regex,
                    message: "#{I18n.t 'sensor_flow.invalid'}"
                  },
                  if: proc { type == FLOW }

        validate :check_observation_id, if: proc { type == FLOW }

        validate :uniqueness_ip_and_observation_id, if: proc { type == FLOW }

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'sensor_flow.only'}" },
                                   if: proc { type == FLOW }


        # TODO: VALIDATE IMAGE SIZE
        # validates_attachment :captive_portal_background_image, content_type: {
        #                                 content_type: %w(image/jpg image/jpeg image/png)
        #                         }, size: { in: 0..10.megabytes },
        #                         if: proc { type == FLOW }

        DittoCode::Exec.if 'ENTERPRISE' do
        validates_presence_of :license_id, if: proc { type == FLOW }
        validate :empty_license_buckets_for_flow, if: proc { type == FLOW }
        end

        before_save :check_ip_network, if: proc { type == FLOW }
        before_update :update_cache_flow,
                      if: proc { type == FLOW && ip_changed? }
        before_destroy :delete_access_points
        after_save :save_to_node, if: proc { type == FLOW }
        after_create :create_cache_flow, if: proc { type == FLOW }
        after_create :create_chef_node_flow,
                     if: proc { type == FLOW }
        #after_create :add_default_flow_monitors, if: proc { type == FLOW }

        after_update :update_ip_flow, if: proc { type == FLOW }
        after_save :data_to_client_proxy_flow,
                   if: proc { type == FLOW && real_parent_id }
        before_destroy :delete_data_from_client_proxy_flow,
                       if: proc { type == FLOW && real_parent_id }

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :aps

        # -------------------------------------------------------------
        # Chef
        # -------------------------------------------------------------
        has_chef_property :nmsp, %w(redborder nmsp)
        has_chef_property :routers_macs, %w(redborder routers_macs)
        has_chef_property :spanport, %w(redborder spanport)
        has_chef_property :captive_portal_name, %w(redborder captive_portal_name)
        has_chef_property :captive_portal_name_color, %w(redborder captive_portal_name_color)
        has_chef_property :captive_portal_background_color, %w(redborder captive_portal_background_color)

        store_accessor :property, :captive_portal_background_image_file_name 
        store_accessor :property, :captive_portal_background_image_content_type
        store_accessor :property, :captive_portal_background_image_file_size
        store_accessor :property, :captive_portal_background_image_updated_at
        store_accessor :property, :captive_portal_logo_image_file_name
        store_accessor :property, :captive_portal_logo_image_content_type
        store_accessor :property, :captive_portal_logo_image_file_size
        store_accessor :property, :captive_portal_logo_image_updated_at

        has_chef_property :nmsp_client_stats, %w(redborder nmsp_client_stats)
        has_chef_property :nmsp_wireless_health, %w(redborder nmsp_wireless_health)
        has_chef_property :dns_ptr_client, %w(redborder dns_ptr_client)
        has_chef_property :dns_ptr_target, %w(redborder dns_ptr_target)
        has_chef_property :radius_secret, %w(redborder radius_secret)
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{FLOW}") do
        'sensors/flow_sensor'
      end

      define_method("to_partial_path_#{Sensor::DOMAIN}_#{Sensor::FLOOR}") do
        'sensors/floor_sensor'
      end

      def add_aps_from_file(content_file, id, only_new_aps)
        file = File.open(content_file.tempfile)

        headers = file.first.gsub(/[\n]+|[\r]+/, "").split(/,|;/)
        column_names_common ||= []
        headers.each do |header|
          column_names_common << symbolize_header_name(header)
        end

        created_count = 0
        not_saved = 0
        wrong_header = column_names_common.include?(nil)

        file.each_with_index do |line, index|

          columns = line.gsub(/[\n]+|[\r]+/, "").split(/,|;/)

          values = {}
          column_names_common.each_with_index do |n, i|
            values[n] = columns[i].strip unless columns[i].blank? or n.nil?
          end

          sensor = Sensor.find(id)
          case sensor.type
          when Sensor::FLOW
            values.delete(:floor_name)
            values[:switch_id] = id
            values[:sensor_id] = sensor.parent.id if sensor.parent &&
                                                     sensor.parent.is_domain?
          when Sensor::DOMAIN
            floor_name = values[:floor_name].strip if values[:floor_name]
            switch_name = values[:switch_name].strip if values[:switch_name]

            values[:sensor_id] =
            sensor.self_and_descendants.where(
              domain_type: Sensor::FLOOR,
              name: floor_name).first.try(:id) || id
            sw = Sensor.find_by_ip(switch_name) if switch_name
            values.delete(:floor_name)
            if sw
              values[:switch_id] = sw.id
            elsif switch_name
              values.delete(:switch_name)
            end
          end
          ap = AccessPoint.where(mac_address: values[:mac_address].downcase).first if values[:mac_address]
          ap.destroy if (ap and only_new_aps == 'false')
          object = AccessPoint.new(values) if (ap.nil? || (ap and only_new_aps == 'false'))

          if object and object.save
            created_count += 1
          else
            not_saved += 1
          end
        end

        file.close

        return created_count, not_saved, wrong_header

      end

      # Return the database column for an access point based on a header
      # of the CSV.
      #
      # @param header [String] Name of the header
      # @return [Symbol] Column of the database
      #
      def symbolize_header_name(header)
        case header.to_s.strip.downcase
        when 'floor name', 'floor', 'floor plan'
          :floor_name
        when 'name'
          :name
        when 'ip address', 'ip'
          :ip_address
        when 'mac address', 'mac'
          :mac_address
        when 'sensor ip', 'sensor'
          :switch_name
        when 'latitude', 'lat'
          :latitude
        when 'longitude', 'long'
          :longitude
        when 'x'
          :x
        when 'y'
          :y
        when 'description'
          :description
        end
      end

    end

    module InstanceMethods

      def is_flow?
        type == FLOW
      end

      # Returns true if sensor is geolocated
      def located?
        gps_markers.count > 2
      end

      # Return the descendant flow sensors of the domain
      def get_flow
        self.all_children(true).where(type: Sensor::FLOW)
      end

      # Return last time when sensor was checked
      define_method("last_check_in_#{FLOW}") do
        Time.at(check_in_time) if check_in_time
      end

      define_method("icon_#{FLOW}") do
        'icons/flow.svg'
      end

      def version_flow
        property[:version_flow]
      end

      def version_flow=(ver)
        property[:version_flow] = ver
      end

      def observation_id
        property[:observation_id]
      end

      def observation_id=(observation_id)
        property[:observation_id] = observation_id
      end

      def ip_observation_id
        if observation_id.present?
          "#{ip} - #{observation_id}"
        else
          ip
        end
      end

      define_method("full_name_#{FLOW}") do
        name
      end

      define_method("secondary_name_#{FLOW}") do
        name
      end

      define_method("complete_name_#{FLOW}") do
        name
      end

      define_method("type_to_s_#{FLOW}") do
        "flow"
      end

      define_method("type_humanize_#{FLOW}") do
        "flow"
      end

      define_method("chef_node_#{FLOW}") do
        Chef::Node.load("rbflow-#{id}") rescue nil
      end

      define_method("chef_client_#{FLOW}") do
        Chef::ApiClient.load("rbflow-#{id}")
      end

      define_method("monitors_to_chef_#{FLOW}") do
        node = chef_node
        if !node.nil?
          begin
            node.normal[:redborder][:monitors] = monitor_categories_to_chef
            node.save
          rescue
            nil
          end
        end
      end

      define_method("run_list_text_#{FLOW}") do
        "role[flow-sensor]"
      end


      def captive_portal_background_image_uri
        return nil if captive_portal_background_image.path.nil?
        StorageFile.get_uri captive_portal_background_image.path
      end
    
      def captive_portal_background_image_path
        if StorageFile.s3_enabled?
          "s3://redborder" + captive_portal_background_image.path
        else
          captive_portal_background_image.path
        end
      end

      def captive_portal_logo_image_uri
        return nil if captive_portal_logo_image.path.nil?
        StorageFile.get_uri captive_portal_logo_image.path
      end
    
      def captive_portal_logo_image_path
        if StorageFile.s3_enabled?
          "s3://redborder" + captive_portal_logo_image.path
        else
          captive_portal_logo_image.path
        end
      end

      #
      # Return all zones under certain sensor and descendants
      #
      def zones
        Zone.where(sensor_id: self_and_descendants.where(type: Sensor::DOMAIN).map(&:id))
      end

      # Return the access points of the sensor
      #
      # @return [ActiveRecord::Relation] Access Points
      #
      def access_points
        if type == Sensor::FLOW
          AccessPoint.includes(:sensor).where(switch: self)
        else
          AccessPoint.includes(:sensor)
            .where('sensor_id in (:value) OR switch_id in (:value)',
                   value: self.all_children_ids(true))
        end
      end

      # Return only located access_points
      #
      # @return [ActiveRecord::Relation] Access Points
      #
      def located_access_points
        access_points.where.not(longitude: nil, latitude: nil)
      end

      def access_points_as_json
        access_points.as_json
      end

      #
      # Scan Sensor Access Points using Druid or SNMP and
      # updates DataBase
      #
      def scan_aps
        if access_points.present?
          property[:check_in_time] = Time.now.to_i
          save
        end
        # Flow & Meraki are the only ones capable of scanning APs (SNMP support)
        return unless is_flow? || is_meraki?
        aps = ApState.get_access_points(self.uuid)
        active_aps = aps.find_all{ |s| s.status == 'on' }

        # Set all APs to down
        access_points.update_all(status: AccessPoint::STATUS[:DOWN])
        # Set the ones that are up to up
        access_points.where(mac_address: active_aps.map(&:wireless_station))
          .update_all(status: AccessPoint::STATUS[:UP]) if active_aps.present?

        # Create the newly found ones
        if aps.present?
          new_aps_macs = aps.map{ |m| m.wireless_station.downcase }.uniq
          old_aps_macs = access_points.map{ |m| m.mac_address.downcase }.uniq
          macs_to_create = new_aps_macs - old_aps_macs
          to_create = []
          macs_to_create.each do |mac|
            ap = aps.find{ |s| s.wireless_station == mac }
            ap_to_create = {
             name: ap.wireless_station_name.nil? ? 'Unnamed' : ap.wireless_station_name,
             mac_address: ap.wireless_station,
             status: ap.status == 'on' ? AccessPoint::STATUS[:UP] : AccessPoint::STATUS[:DOWN],
             switch_id: id,
             switch_name: ip
            }
            to_create << ap_to_create
          end
          AccessPoint.create(to_create) if to_create.present?
        else
          # If not found on druid scan by SNMP
          Redborder::ScanAps.scan_snmp(self)
        end

      end

      #
      # Return the access points associated to this sensor
      # filtered by status
      #
      # == Parameters:
      # status::
      #    Integer with the status to filter.
      #
      # == Returns:
      # ActiveRelation with AccessPoint
      #
      def access_points_by_status(status)
        if type == Sensor::FLOW
          AccessPoint.where(switch: self, status: status.to_s.split(','))
        else
          AccessPoint
            .where("(sensor_id in (:value) OR switch_id in (:value)) \
                    AND status in (:status)",
                   value: all_children_ids(true),
                   status: status.to_s.split(','))
        end
      end

      #
      # Return count of sensors based on status. This method save in class
      # @aps_by_status, so databse it's only called one time.
      #
      # == Parameters:
      # status:
      #   Integer with the status of the AccessPoint. See AccessPoint#STATUS
      #
      # == Return:
      # Integer with the number of access points with status given.
      #
      def access_points_group_by_status(status)
        @aps_by_status ||=
          AccessPoint.where('(sensor_id in (:value) OR switch_id in (:value))',
                            value: all_children_ids(true)).group(:status).count
        @aps_by_status[status] || 0
      end

      #
      # Return access points of this sensor that have up status.
      #
      # == Returns:
      # ActiveRelation with AccessPoint
      #
      def up_access_points
        access_points_by_status(1)
      end

      #
      # Return access points of this sensor that have udown status.
      #
      # == Returns:
      # ActiveRelation with AccessPoint
      #
      def down_access_points
        access_points_by_status(2)
      end

      #
      # Return access points of this sensor that have unknown status.
      #
      # == Returns:
      # ActiveRelation with AccessPoint
      #
      def unknown_access_points
        access_points_by_status(0)
      end

      #
      # Return unallocated access points of this sensor.
      #
      # == Returns:
      # ActiveRelation with AccessPoint
      #
      def unallocated_access_points
        if type == Sensor::FLOW
          access_points.where('latitude is NULL OR longitude is NULL OR sensor_id is NULL')
        else
          access_points.where('latitude is NULL OR longitude is NULL')
        end
      end

      define_method("as_json_with_children_#{FLOW}") do
        access_points = AccessPoint.where(switch: self)
        self.as_json.merge!(access_points: access_points.as_json,
                            snmp_community: snmp_community,
                            snmp_version: snmp_version)
      end

      define_method("as_json_with_children_#{Sensor::DOMAIN}") do
        sensor = self
        result = sensor.as_json.merge!(children: sensor.children_json)
        result.merge!(zones: sensor.zones.as_json) if sensor.domain_type == Sensor::FLOOR
        result
      end

      # Return the floors of a sensor to represent in a map using infrastructureMap
      # javascript library.
      #
      # @return [Hash] Data to represent the floor
      #
      def to_floor_zone
        {
          name: name,
          id: id,
          uuid: uuid,
          floor_level: floor_level,
          domain_type_name: domain_type_name(to_key: true),
          dimensions: floor_dimensions,
          image: (overlay_map ? overlay_map.original_image_uri(controller: true).to_s : ''),
          access_points: access_points
        }
      end

      # Get the floor proportiones from the sensor
      # REFACTOR: Should this info come from the sensor? or the overlay?
      def floor_dimensions
        dimensions = property.fetch('dimensions', {})
        width = dimensions.fetch('width', 0)
        height = dimensions.fetch('height', 0)
        [width, height]
      end

      # Display the number of access points in domains in tree view
      #
      def show_aps?
        deployment? || campus? || building? || floor?
      end

      def geopositionate_aps
        # We just posicionate the aps if this sensor is a floor domain
        if type != Sensor::DOMAIN ||
           domain_type != Sensor::FLOOR ||
           access_points.empty?
          return false
        end

        # Posicionate each access point
        begin
          AccessPoint.transaction do
            access_points.each do |access_point|
              if access_point.x && access_point.y
                access_point.recalculate_lnglat
              elsif access_point.latitude && access_point.longitude
                access_point.recalculate_xy
              end
            end
          end

          return true
        rescue StandardError => e
          Rails.logger.error "Access Point Geopositionate #{e}"
          return false
        end
      end

      #
      # Creates access points from json by using redBorder API
      #
      # == Parameters:
      # access_points_json::
      #   APs data in json (hash or array)
      # top_domain::
      #   Top Domain Sensor
      #
      def create_access_points_from_json(access_points_json, top_domain)
        domain_levels = [
          :service_provider, :organization, :namespace, :deployment,
          :market, :campus, :building, :floor
        ]

        domain_correspondency = {
          service_provider: Sensor::SERVICE_PROVIDER,
          organization: Sensor::ORGANIZATION,
          namespace: Sensor::NAMESPACE,
          deployment: Sensor::DEPLOYMENT,
          market: Sensor::MARKET,
          campus: Sensor::CAMPUS,
          building: Sensor::BUILDING,
          floor: Sensor::FLOOR
        }

        # We need to iterate over an array, not over a hash
        access_points_json = [access_points_json] if access_points_json.is_a? Hash

        access_points_json.each do |ap_json|
          hierarchy = [top_domain]
          params_to_exclude = []

          domain_levels.each do |level|
            level_uuid = (level.to_s + "_uuid").to_sym
            level_name = (level.to_s + "_name").to_sym

            if ap_json[level_uuid]
              # uuid given, name may be given as well
              params_to_exclude << level_uuid.to_s
              params_to_exclude << level_name.to_s

              domain = top_domain.descendants.find_by(uuid: ap_json[level_uuid], domain_type: domain_correspondency[level])
              name = ap_json[level_name] ? ap_json[level_name] : level.to_s + "_" + ap_json[level_uuid]
              unless domain
                # does not exists, create it
                domain = Sensor.create(
                    parent_id: hierarchy.last.id,
                    name: name,
                    type: Sensor::DOMAIN,
                    domain_type: domain_correspondency[level]
                )
                unless domain.valid?
                  if domain.errors[:name]
                    # Name is already taken, update uuid to the given one
                    domain = top_domain.descendants.find_by(name: name)
                    domain.update(uuid: ap_json[level_uuid])
                  end
                end
              end
              hierarchy << domain

            elsif ap_json[level_name]
              # name given, uuid was not given
              params_to_exclude << level_name.to_s

              domain = top_domain.descendants.find_by(name: ap_json[level_name], domain_type: domain_correspondency[level])
              unless domain
                # market does not exists, create it
                domain = Sensor.create(
                    parent_id: hierarchy.last.id,
                    name: ap_json[level_name],
                    type: Sensor::DOMAIN,
                    domain_type: domain_correspondency[level]
                )
                unless domain.valid?

                end
              end
              hierarchy << domain
            end
          end

          para = { sensor_id: hierarchy.last.id }
          params_to_exclude.each do |e|
            ap_json.delete( e )
          end
          para.merge! ( ap_json )
          AccessPoint.create( para )
        end
      end

      #
      # Move APS from a sensor/switch to other if the user has permissions. If the
      # AP doesn't exist, it will be created.
      #
      # == Parameters:
      # aps::
      #   Parmas Hash with the info about aps
      # user::
      #   User who perform the query
      #
      def move_access_points(aps, user)
        return unless (is_domain? || is_flow?) && aps
        # Move or create aps
        aps.each do |ap|
          new_ap = AccessPoint.find_by_mac_address(ap[:mac_address])

          data = ap.permit(:name, :mac_address, :status, :ip_address, :latitude,
                           :longitude, :x, :y, :description)

          if new_ap.nil?
            new_ap = AccessPoint.new(data)
          elsif user.can?(:manage, new_ap.sensor) ||
                user.can?(:manage, new_ap.switch)
            new_ap.update(data)
          else
            # The user hasn't got privileges
            next
          end

          new_ap.sensor_id = id if is_domain?
          new_ap.switch_id = id if is_flow?
          new_ap.save
        end
      end

      # Force the destroy of access points that belongs to the current sensor
      # and all its descendants.
      #
      def delete_all_access_points
        return unless (is_flow? || is_meraki? || is_domain?) &&
                      User.current_user.can?(:manage, self)
        # consider sensor or switch depending on type
        current_sensor = (is_flow? || is_meraki?) ? :switch : :sensor
        other_sensor = (is_flow? || is_meraki?) ? :sensor : :switch
        # get ap ids to destroy
        ids = access_points.select do |ap|
          User.current_user.can?(:manage, ap.send(other_sensor)) ||
          ap.send(other_sensor).nil?
        end.map(&:id)

        # Update radius secret on radius database for each ap         
        AccessPoint.where(id: ids).each do |ap| 
          ap.delete_access_points_in_parent_node
          ap.delete_radius
        end

        # clear all aps sensor/switch
        access_points.update_all("#{current_sensor}_id" => nil)
        # delete aps
        ActiveRecord::Base.transaction do
          begin
            AccessPoint.where(id: ids).destroy_all
          rescue ActiveRecord::StatementInvalid => e
            Rails.logger.error e
            Rails.logger.error e.backtrace.join("\n")
          end
        end
      end

      def delete_selected_access_points(selected_ids)
        return unless (is_flow? || is_meraki? || is_domain?) &&
                      User.current_user.can?(:manage, self)
        # consider sensor or switch depending on type
        current_sensor = (is_flow? || is_meraki?) ? :switch : :sensor
        other_sensor = (is_flow? || is_meraki?) ? :sensor : :switch
        # get ap ids to destroy
        ids = access_points.where(id: selected_ids).select do |ap|
          User.current_user.can?(:manage, ap.send(other_sensor)) ||
          ap.send(other_sensor).nil?
        end.map(&:id)

        # Update radius secret on radius database for each ap         
        AccessPoint.where(id: ids).each do |ap|
          ap.delete_access_points_in_parent_node 
          ap.delete_radius
        end

        # clear all aps sensor/switch
        access_points.where(id: selected_ids).update_all("#{current_sensor}_id" => nil)

        # delete aps
        ActiveRecord::Base.transaction do
          begin
            AccessPoint.where(id: ids).destroy_all
          rescue ActiveRecord::StatementInvalid => e
            Rails.logger.error e
            Rails.logger.error e.backtrace.join("\n")
          end
        end
      end

      #
      # Get the UUIDs of the Namespaces that belongs its APs
      # This function is only valid for FLOW and Meraki sensors.
      #
      def namespace_uuids_from_aps
        return [] unless [Sensor::FLOW, Sensor::MERAKI].include?(type)
        query = "select distinct(namespace.uuid) \
                 from access_points as ap, \
                      sensors as namespace, \
                      sensors as domain
                 where ap.switch_id = #{id} and \
                       (ap.sensor_id = domain.id OR ap.switch_id = domain.id) and \
                       namespace.lft < domain.lft and \
                       namespace.rgt > domain.rgt and \
                       namespace.type = #{Sensor::DOMAIN} and \
                       namespace.domain_type = #{Sensor::NAMESPACE};"

        ActiveRecord::Base.connection.execute(query).map { |n| n['uuid'] }
      end

      def update_radius_access_points
        return unless type != Sensor::DOMAIN && (is_flow? || is_meraki? || is_domain?) && User.current_user &&
                    User.current_user.can?(:manage, self)
        # consider sensor or switch depending on type
        current_sensor = (is_flow? || is_meraki?) ? :switch : :sensor
        other_sensor = (is_flow? || is_meraki?) ? :sensor : :switch
        # Update radius secret on radius database for each ap         
        access_points.each_with_index do |ap,index| 
          if User.current_user.can?(:manage, ap.send(other_sensor)) || ap.send(other_sensor).nil?
            ap.update_radius 
          end
        end
      end

      def add_default_flow_monitors
         monitor_categories << MonitorCategory.where(default: true)
           .select { |mc| mc.sensor_types.include? type }
         save
         monitors_to_chef
      end

      private

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------
        def check_observation_id
          return if observation_id.blank?
          unless /\d+/ =~ observation_id && (0..2**32).include?(observation_id.to_i)
            errors.add(:observation_id, "#{I18n.t 'sensor_flow.between'}")
          end
        end

        def uniqueness_ip_and_observation_id
          # Check if exists other sensor flow with the same real_parent_id and observation_id
          sensors = Sensor.where(type: type, ip: ip, real_parent_id: real_parent_id)
          sensors = sensors.where.not(id: id) if id

          if sensors.any? && sensors.select { |s| s.observation_id == observation_id }.any?
            errors.add(:ip, "#{I18n.t 'sensor_flow.unique'}")
          end
        end

        # Check if license can be associated to current sensor
        #
        def empty_license_buckets_for_flow
          return unless license
          return if license_id_was == license_id
          errors.add(:license_id, "#{I18n.t 'sensor_flow.full'}") if license.in_use_flow >= license.limit_flow
        end

        def create_cache_flow
          Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
        end

        def update_cache_flow
          Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
        end

        def create_chef_node_flow
          node_json = {
            'name' => "rbflow-#{self.id}",
            'run_list' => ['role[flow-sensor]', "role[#{self.chef_name}]"],
            'attributes' => {
              'rbname' => name,
              'ipaddress' => ip,
              'redborder' => {
                'monitors' => [],
                'sensor_uuid' => uuid,
                'ipaddress' => ip,
                'parent_id' => real_parent_id
              }
            }
          }

          node = Chef::Node.from_hash(node_json)
          node.create
        end

        def check_ip_network
          self.ip = NetAddr::CIDR.create(self.ip).to_s.gsub(/\/32$/, "")
        end

        

        def update_ip_flow
          role = chef_role
          node = chef_node
          role.override_attributes['ipaddress'] = ip
          node.normal['ipaddress'] = ip
          node.normal['redborder']['ipaddress'] = ip
          role.save
          node.save
          update_radius_access_points
        end

        def save_to_node
          return if property[:version_flow].nil? && property[:observation_id].nil?
          chef_node.normal['redborder']['observation_id'] = property[:observation_id]
          chef_node.normal['rbversion'] = property[:version_flow]
          chef_node.save
        end

        # Update chef domain values of sensor flow.
        #
        define_method("update_chef_domain_values_#{FLOW}") do
          update_chef_domain_values_in_node
          data_to_client_proxy_flow if real_parent
        end

        def as_json_with_aps(options = {})
          result = as_json_without_aps(options)
          if self.is_domain?
            result.merge!(
              { access_points: AccessPoint.where(sensor_id: self.id) }.as_json
            )
          elsif self.is_flow?
            result.merge!(
              {
                access_points: access_points,
                snmp_community: snmp_community,
                snmp_version: snmp_version,
                nmsp: nmsp,
                nmsp_wireless_health: nmsp_wireless_health,
                nmsp_client_stats: nmsp_client_stats,
                dns_ptr_client: dns_ptr_client,
                dns_ptr_target: dns_ptr_target,
                spanport: spanport,
                captive_portal_name: captive_portal_name,
                captive_portal_name_color: captive_portal_name_color,
                captive_portal_background_color: captive_portal_background_color,

                captive_portal_background_image_file_name: captive_portal_background_image_file_name, 
                captive_portal_background_image_content_type: captive_portal_background_image_content_type, 
                captive_portal_background_image_file_size: captive_portal_background_image_file_size, 
                captive_portal_background_image_updated_at: captive_portal_background_image_updated_at,

                captive_portal_logo_image_file_name: captive_portal_logo_image_file_name, 
                captive_portal_logo_image_content_type: captive_portal_logo_image_content_type, 
                captive_portal_logo_image_file_size: captive_portal_logo_image_file_size, 
                captive_portal_logo_image_updated_at: captive_portal_logo_image_updated_at,

                radius_secret: radius_secret,
                routers_macs: routers_macs,
                homenets: homenets
              }.as_json
            )
          end
          result
        end

      protected

      # Create the chef role for Flow sensors
      #
      define_method "create_chef_role_#{FLOW}" do
        run_list = ['role[flow-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{FLOW}" do
        chef_node.destroy
      end

      #
      # Delelete access points if the sensor is a domain. If it's flow,
      # set the switch to unnasigned.
      #
      def delete_access_points
        if is_flow? || is_meraki?
          AccessPoint.where(switch_id: id).destroy_all
        elsif is_domain?
          AccessPoint.where(sensor_id: id).where.not(switch_id: nil).update_all(sensor_id: nil)
          AccessPoint.where(sensor_id: id, switch_id: nil).destroy_all
        end
      end

      # Store the current data of the Chef node in the client_proxy parent
      #
      def data_to_client_proxy_flow
        # Get the role of the parent
        role = real_parent.chef_role

        # Update attributes
        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['flow'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['flow'][name] =
          chef_node['redborder']
        # Save it
        role.save
      end

      def delete_data_from_client_proxy_flow
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['flow'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorFlow)
