#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorMeraki
    # Id of the sensor
    MERAKI = 9

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        include StorageFile
        storage_file :captive_portal_background_image, path: '/rb-webui/:class/:id/:filename'
        storage_file :captive_portal_logo_image, path: '/rb-webui/:class/:id/logo/:filename'

        # Increase available list of sensors
        self.available_sensors += [MERAKI]
        # Import values for Meraki
        self.import_columns_for[MERAKI] =
          %i(ip meraki_validator snmp_community snmp_version captive_portal_name captive_portal_name_color captive_portal_background_color radius_secret captive_portal_background_image captive_portal_background_image_file_name captive_portal_background_image_content_type captive_portal_background_image_file_size captive_portal_background_image_updated_at captive_portal_logo_image captive_portal_logo_image_file_name captive_portal_logo_image_content_type captive_portal_logo_image_file_size captive_portal_logo_image_updated_at).freeze

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------

        validate :check_snmp_community_meraki, if: proc { type == MERAKI }

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'lib_flow.only'}" },
                                   if: proc { type == MERAKI }

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_create :generate_secret, if: proc { type == MERAKI }
        before_update :update_cache_meraki,
                      if: proc { type == MERAKI && ip_changed? }
        after_create :create_cache_meraki, if: proc { type == MERAKI }
        after_create :create_chef_node_meraki,
                     if: proc { type == MERAKI }
        after_update :update_meraki_values, if: proc { type == MERAKI }
        after_update :update_radius_access_points, if: proc { type == MERAKI }

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :meraki

        # -------------------------------------------------------------
        # Accessors
        # -------------------------------------------------------------
        store_accessor :property, :meraki_validator
        store_accessor :property, :meraki_secret
        store_accessor :property, :captive_portal_name
        store_accessor :property, :captive_portal_name_color
        store_accessor :property, :captive_portal_background_color
        store_accessor :property, :radius_secret
        store_accessor :property, :captive_portal_background_image_file_name 
        store_accessor :property, :captive_portal_background_image_content_type
        store_accessor :property, :captive_portal_background_image_file_size
        store_accessor :property, :captive_portal_background_image_updated_at
        store_accessor :property, :captive_portal_logo_image_file_name
        store_accessor :property, :captive_portal_logo_image_content_type
        store_accessor :property, :captive_portal_logo_image_file_size
        store_accessor :property, :captive_portal_logo_image_updated_at
        
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{MERAKI}") do
        'sensors/meraki_sensor'
      end
    end

    module InstanceMethods
      def is_meraki?
        type == MERAKI
      end

      # Return the descendant mse sensors of the domain
      def get_meraki
        self.all_children(true).where(type: Sensor::MERAKI)
      end

      # Return last time when sensor was checked
      define_method("last_check_in_#{MERAKI}") do
        Time.at(check_in_time) if check_in_time
      end

      define_method("icon_#{MERAKI}") do
        'icons/meraki.png'
      end

      def version_meraki
        property[:version_meraki]
      end

      def version_meraki=(ver)
        property[:version_meraki] = ver
      end

      define_method("full_name_#{MERAKI}") do
        name
      end

      define_method("secondary_name_#{MERAKI}") do
        name
      end

      define_method("complete_name_#{MERAKI}") do
        name
      end

      define_method("type_to_s_#{MERAKI}") do
        'meraki'
      end

      define_method("type_humanize_#{MERAKI}") do
        'Meraki'
      end

      define_method("chef_node_#{MERAKI}") do
        Chef::Node.load("rbmeraki-#{id}") rescue nil
      end

      define_method("chef_client_#{MERAKI}") do
        Chef::ApiClient.load("rbmeraki-#{id}")
      end

      define_method("monitors_to_chef_#{MERAKI}") do
        node = chef_node
        if !node.nil?
          begin
            node.normal[:redborder][:monitors] = monitor_categories_to_chef
            node.save
          rescue
            nil
          end
        end
        node.save
      end

      define_method("run_list_text_#{MERAKI}") do
        'role[meraki-sensor]'
      end


      def captive_portal_background_image_uri
        return nil if captive_portal_background_image.path.nil?
        StorageFile.get_uri captive_portal_background_image.path
      end
    
      def captive_portal_background_image_path
        if StorageFile.s3_enabled?
          "s3://redborder" + captive_portal_background_image.path
        else
          captive_portal_background_image.path
        end
      end

      def update_radius_access_points
        return unless type != Sensor::DOMAIN && (is_flow? || is_meraki? || is_domain?) && User.current_user &&
                    User.current_user.can?(:manage, self)
        # consider sensor or switch depending on type
        current_sensor = (is_flow? || is_meraki?) ? :switch : :sensor
        other_sensor = (is_flow? || is_meraki?) ? :sensor : :switch
        # Update radius secret on radius database for each ap         
        access_points.each_with_index do |ap,index| 
          if User.current_user.can?(:manage, ap.send(other_sensor)) || ap.send(other_sensor).nil?
            ap.update_radius 
          end
        end
      end

      private

      #
      # The community of Meraki can't be public
      #
      def check_snmp_community_meraki
        return if snmp_community != 'public'
        errors.add(:snmp_community,
                   "#{I18n.t 'lib_flow.public'}")
      end

      def create_cache_meraki
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
      end

      def update_cache_meraki
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
      end

      def generate_secret
        # Check if the sensor has a meraki secret assigned (import)
        return if meraki_secret.present?

        sensors = true
        while sensors.present?
          secret = SecureRandom.hex(20)
          sensors = Sensor.where(type: MERAKI)
                    .where('property like ?', "%\"meraki_secret\":\"#{secret}\"%")
        end

        self.meraki_secret = secret
      end

      def create_chef_node_meraki
        node_json = {
          'name' => "rbmeraki-#{id}",
          'run_list' => ['role[meraki-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'ipaddress' => ip,
            'redborder' => {
              'meraki_validator' => meraki_validator,
              'meraki_secret' => meraki_secret,
              'monitors' => [],
              'sensor_uuid' => uuid
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def save_version
        return if property[:version_meraki].nil?
        chef_node.normal['rbversion'] = property[:version_meraki]
        chef_node.save
      end

      #
      # Return the sensor as json with variables from MSE
      #
      def as_json_with_meraki(options = {})
        result = as_json_without_meraki(options)
        if self.is_meraki?
          result.merge!(
            {
              access_points: access_points,
              snmp_community: snmp_community,
              snmp_version: snmp_version,
              meraki_secret: meraki_secret,
              meraki_validator: meraki_validator,
              captive_portal_name: captive_portal_name,
              captive_portal_name_color: captive_portal_name_color,
              captive_portal_background_color: captive_portal_background_color,
              radius_secret: radius_secret,

              captive_portal_background_image_file_name: captive_portal_background_image_file_name, 
              captive_portal_background_image_content_type: captive_portal_background_image_content_type, 
              captive_portal_background_image_file_size: captive_portal_background_image_file_size, 
              captive_portal_background_image_updated_at: captive_portal_background_image_updated_at,

              captive_portal_logo_image_file_name: captive_portal_logo_image_file_name, 
              captive_portal_logo_image_content_type: captive_portal_logo_image_content_type, 
              captive_portal_logo_image_file_size: captive_portal_logo_image_file_size, 
              captive_portal_logo_image_updated_at: captive_portal_logo_image_updated_at
            }.as_json
          )
        end
        result
      end

      def update_meraki_values
        role = chef_role
        node = chef_node
        role.override_attributes['ipaddress'] = ip
        role.override_attributes['redborder']['meraki_validator'] = meraki_validator
        role.override_attributes['redborder']['captive_portal_name'] = captive_portal_name
        role.override_attributes['redborder']['captive_portal_name_color'] = captive_portal_name_color
        role.override_attributes['redborder']['captive_portal_background_color'] = captive_portal_background_color
        role.override_attributes['redborder']['radius_secret'] = radius_secret
        node.normal['ipaddress'] = ip
        node.normal['redborder']['meraki_validator'] = meraki_validator
        node.normal['redborder']['captive_portal_name'] = captive_portal_name
        node.normal['redborder']['captive_portal_name_color'] = captive_portal_name_color
        node.normal['redborder']['captive_portal_background_color'] = captive_portal_background_color
        node.normal['redborder']['radius_secret'] = radius_secret
        role.save
        node.save
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{MERAKI}") do
        update_chef_domain_values_in_node
      end

      protected

      # Create the chef role for Meraki sensors
      #
      define_method "create_chef_role_#{MERAKI}" do
        run_list = ['role[meraki-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{MERAKI}" do
        chef_node.destroy
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorMeraki)
