#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorMse
    # Id of the sensor
    MSE = 8

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [MSE]

        # Import values for MSE
        self.import_columns_for[MSE] =
          [:stream, :snmp_version, :version_mse,
           :send_http, { homenets: [:name, :value] }].freeze

        # Alias the name of IP attribute
        alias_attribute :stream, :ip

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------
        validates :stream, presence: true,
                  uniqueness: { scope: :real_parent_id },
                  if: proc { type == MSE }

        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'lib_flow.only'}" },
                                   if: proc { type == MSE }

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :mse

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_update :update_cache_mse,
                      if: proc { type == MSE && ip_changed? }
        after_create :create_cache_mse, if: proc { type == MSE }
        after_create :create_chef_node_mse,
                     if: proc { type == MSE }
        after_update :update_stream, if: proc { type == MSE }

        after_save :data_to_client_proxy_mse,
                   if: proc { type == MSE && real_parent_id }
        before_destroy :delete_data_from_client_proxy_mse,
                       if: proc { type == MSE && real_parent_id }

        # -------------------------------------------------------------
        # Chef
        # -------------------------------------------------------------

        has_chef_property :send_http, %w(redborder send_http)
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{MSE}") do
        'sensors/mse_sensor'
      end
    end

    module InstanceMethods

      def is_mse?
        type == MSE
      end

      # Return the descendant mse sensors of the domain
      def get_mse
        self.all_children(true).where(type: Sensor::MSE)
      end

      define_method("icon_#{MSE}") do
        'icons/mse.png'
      end

      def version_mse
        property[:version_mse]
      end

      def version_mse=(ver)
        property[:version_mse] = ver
      end

      define_method("full_name_#{MSE}") do
        name
      end

      define_method("secondary_name_#{MSE}") do
        name
      end

      define_method("complete_name_#{MSE}") do
        name
      end

      define_method("type_to_s_#{MSE}") do
        "mse"
      end

      define_method("type_humanize_#{MSE}") do
        "MSE"
      end

      define_method("chef_node_#{MSE}") do
        Chef::Node.load("rbmse-#{id}") rescue nil
      end

      define_method("chef_client_#{MSE}") do
        Chef::ApiClient.load("rbmse-#{id}")
      end

      define_method("monitors_to_chef_#{MSE}") do
        node = chef_node
        if !node.nil?
          begin
            node.normal[:redborder][:monitors] = monitor_categories_to_chef
            node.save
          rescue
            nil
          end
        end
        node.save
      end

      define_method("run_list_text_#{MSE}") do
        "role[mse-sensor]"
      end

      private

      def create_cache_mse
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def update_cache_mse
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", stream
      end

      def create_chef_node_mse
        node_json = {
          'name' => "rbmse-#{id}",
          'run_list' => ['role[mse-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'stream' => stream,
            'redborder' => {
              'sensor_uuid' => uuid,
              'monitors' => []
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def save_version
        return if property[:version_mse].nil?
        chef_node.normal['rbversion'] = property[:version_mse]
        chef_node.save
      end

      def update_stream
        role = chef_role
        node = chef_node
        role.override_attributes["stream"] = stream
        node.normal["stream"] = stream
        role.save
        node.save
      end

      #
      # Return the sensor as json with variables from MSE
      #
      def as_json_with_mse(options = {})
        result = as_json_without_mse(options)
        if self.is_mse?
          result.merge!(
            {
              access_points: access_points,
              snmp_version: snmp_version,
              version_mse: version_mse,
              stream: stream,
              send_http: send_http
            }.as_json
          )
        end
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{MSE}") do
        update_chef_domain_values_in_node
        data_to_client_proxy_mse if real_parent
      end

      protected

      # Create the chef role for MSE sensors
      #
      define_method "create_chef_role_#{MSE}" do
        run_list = ['role[mse-sensor]', "role[#{parent.chef_name}]"]
        builder = Redborder::Chef::Role::Builder.new(self, run_list)
        builder.create
      end

      define_method "destroy_chef_node_#{MSE}" do
        chef_node.destroy
      end

      def data_to_client_proxy_mse
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['mse'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['mse'][name] = chef_node['redborder']
        role.save
      end

      def delete_data_from_client_proxy_mse
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['mse'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorMse)
