#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# => Module helper added to sensor helper. This module contains methods
#    to help the render of sensor's UI.
#
module RedBorder
  module SensorsHelperFlow
    def self.included(base)
      base.send(:include, InstanceMethods)

      base.class_eval do
        unloadable
        alias_method_chain :sensor_last_check_in, :flow_sensors

        #
        # => Return the HTML code for a sensor overlay map.
        #    The size of the image is small. If we don't find any
        #    image, create a mosaic based on child image (max 4 images)
        #
        def sensor_view_image(sensor)
          if sensor.overlay_map
            url = original_image_small_overlay_map_path(sensor.overlay_map)
            image_html = "<img class='sensor-image' src='#{url}'/>"
          else
            # Try to get children images
            num_images = 0
            images = ""
            url = ""

            sensor.all_children.where(domain_type: Sensor::FLOOR).each do |child|
              unless child.overlay_map.nil?
                url = original_image_small_overlay_map_path(child.overlay_map)
                images += "<div class='sensor-image' data-src='#{url}'></div>"
                num_images += 1

                # Stop if we have 4 images
                if num_images == 4 then break end
              end
            end

            if num_images == 0
              image_html =
                "<img class='sensor-image no-image' src='#{ asset_path('no-image.png')}'/>"
            elsif num_images == 1
              image_html = "<img class='sensor-image' src='#{url}'/>"
            else
              image_html =
                "<div class='sensor-mosaic mosaic-#{num_images}'>#{images}</div>"
            end
          end

          image_html.html_safe
        end

        # Return the number of access points and unclaimed access points for
        # tree sensors view
        #
        # @param sensor [Sensor] Sensor to find access_points
        # @return [String] with the format: number access points (unnasigned)
        #
        def count_access_points(sensor)
          aps = sensor.access_points.count
          return '0 APs' if aps == 0
          unnassiged = sensor.access_points.where(sensor_id: nil).count
          return "#{aps} #{I18n.t 'lib_flow.aps'}" if unnassiged == 0
          "#{aps} #{I18n.t 'lib_flow.aps'} (#{unnassiged} #{I18n.t 'lib_flow.unassigned'})"
        end

        #
        # => Return the HTML to represent the icon of
        #    a sensor.
        #
        #    @param sensor : the sensor to get the icon
        #
        def sensor_view_icon(sensor)
          # Check the type of the icon
          if sensor.icon_type == 'image'
            image_html =
              "<img src='#{ asset_path(sensor.icon)}'/>"
          else
            image_html =
              "<i class='#{sensor.icon}'></i>"
          end

          image_html.html_safe
        end

        #
        # => Return the HTML to adapt the font-size to
        #    the length of the name.
        #
        #    @param name : the name of the sensor
        #
        def sensor_view_name(name)

          # Check the size of the name
          if name.size > 21
            name =
              "<span style='font-size: .6em'>#{name}</span>"
          elsif name.size > 18
            name =
              "<span style='font-size: .7em'>#{name}</span>"
          elsif name.size > 15
            name =
              "<span style='font-size: .8em'>#{name}</span>"
          elsif name.size > 12
            name =
              "<span style='font-size: .9em'>#{name}</span>"
          end

          name.html_safe
        end

        def status_to_color(status)
          case status
          when 0
            'orange'
          when 1
            'green'
          when 2
            'red'
          end
        end

        #
        # Render a button to access to access points view.
        #
        # == Parameters:
        # sensor::
        #   Sensor to render access points
        #
        # == Returns:
        # String with HTML code of button.
        #
        def access_points_button(sensor)
          return unless sensor.is_domain? || sensor.is_meraki? ||
                        sensor.is_flow? || sensor.is_client_proxy?
          # Set the link
          link_to content_tag(:i, nil, class: 'fa fa-wifi'),
                  view_sensor_path(sensor), class: 'btn block-ui-but'
        end

        #
        # Return the IDs of the zones with AP in the sensor
        # or his children
        #
        def sensor_zones(sensor)
          zones = Zone.includes(:access_points).where(access_points:
            { sensor_id: sensor.all_children_ids(true) }).map(&:id)
          zones.uniq.join(',')
        end

        #
        # => Return the HTML that show the status of Access Points
        #    of a sensor, with number of them by status and color
        #
        def access_points_grouped_by_status(sensor)
          if sensor.access_points.count > 0
            aps_group_count = sensor.access_points.group(:status).count

              content_html = []

              if aps_group_count[AccessPoint::STATUS[:UP]] &&
                 aps_group_count[AccessPoint::STATUS[:UP]] > 0

                content_html << aps_group_count[AccessPoint::STATUS[:UP]].to_s + ' ' + content_tag(:i, nil, class: 'fa fa-circle',
                                                                                             style: 'color: green')
              end

              if aps_group_count[AccessPoint::STATUS[:DOWN]] &&
                 aps_group_count[AccessPoint::STATUS[:DOWN]] > 0
                content_html << aps_group_count[AccessPoint::STATUS[:DOWN]].to_s + ' ' + content_tag(:i, nil, class: 'fa fa-circle',
                                                                                             style: 'color: red')

              end

              if aps_group_count[AccessPoint::STATUS[:UNKNOWN]] &&
                 aps_group_count[AccessPoint::STATUS[:UNKNOWN]] > 0
                content_html << aps_group_count[AccessPoint::STATUS[:UNKNOWN]].to_s + ' ' + content_tag(:i, nil, class: 'fa fa-circle',
                                                                                             style: 'color: orange')
              end

              content_html.join('&nbsp;&nbsp;').html_safe

          else
            0
          end
        end

      end
    end

    module InstanceMethods

      #
      # => Check the last check in of a sensor and return an HTML with
      #    the current status and the associated icon.
      #
      def sensor_last_check_in_with_flow_sensors(sensor)
        return_text = ""

        if ![Sensor::DOMAIN, Sensor::FLOW, Sensor::MERAKI].include?(sensor.type) and sensor.last_check_in
          if sensor.last_check_in < Time.now - 20.minutes
            return_text += content_tag(:i, nil, class: "fa fa-circle red")
          else
            return_text += content_tag(:i, nil, class: "fa fa-circle green")
          end
          return_text += distance_of_time_in_words(Time.now, sensor.last_check_in)
        elsif ![Sensor::DOMAIN, Sensor::FLOW, Sensor::MERAKI].include?(sensor.type)
          if sensor.unclaimed
            return_text += content_tag(:i, nil, class: "fa fa-circle gray") + "#{I18n.t 'lib_flow.unclaimed'}"
          else
            return_text += content_tag(:i, nil, class: "fa fa-circle blue") + "#{I18n.t 'lib_flow.configuring'}"
          end
        elsif [Sensor::FLOW, Sensor::MERAKI].include?(sensor.type) && !sensor.property[:check_in_time].nil?
          time = sensor.check_in_time || sensor.created_at
          time = Time.at(time)
          if time < Time.now - 20.minutes
            return_text += content_tag(:i, nil, class: "fa fa-circle red")
          else
            return_text += content_tag(:i, nil, class: "fa fa-circle green")
          end
          return_text += distance_of_time_in_words(Time.now, time)
        end

        return_text.html_safe
      end
    end
  end
end
