#######################################################################
## Copyright (c) 2024 ENEO Tecnología S.L.
## This file is part of redBorder.
## redBorder is free software: you can redistribute it and/or modify
## it under the terms of the GNU Affero General Public License License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## redBorder is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Affero General Public License License for more details.
## You should have received a copy of the GNU Affero General Public License License
## along with redBorder. If not, see <http://www.gnu.org/licenses/>.
########################################################################

# Import required libraries for the script
require 'active_record'
require 'rest-client'
require 'json'
require 'yaml'
require 'logger'
require 'faraday'

# Logger for the application
class AppLogger
  # Create a logger instance, if not already created, that logs to STDOUT
  def self.logger
    @logger ||= Logger.new(STDOUT).tap do |log|
      log.level = Logger::DEBUG
    end
  end
end

# Class to manage file storage operations
class FileManager
  # Saves the data to a file in a human-readable JSON format
  def self.save_to_file(data, file_path)
    File.open(file_path, 'w') do |file|
      # Use JSON.pretty_generate for a nicely formatted JSON
      file.write(JSON.pretty_generate(data))
    end
    # Log that data has been successfully saved
    AppLogger.logger.info("Data saved to #{file_path}")
  rescue IOError => e
    # Log any error encountered during file saving and raise the exception
    AppLogger.logger.error("Failed to save data to file: #{e.message}")
    raise
  end
end

# Main class responsible for interacting with Zabbix
class ZabbixInteraction
  def initialize(api_url, user, password, token, logger = nil)
    @api_url = api_url
    @user = user
    @password = password
    @token = token
    @result = ""
  end

  # Method to execute the interaction with Zabbix
  def execute
    @result += "Starting Zabbix interaction\n"   # Log the start of the interaction

    # Instantiate ZabbixClient to interact with Zabbix API
    zabbix_client = ZabbixClient.new(@api_url, @user, @password, @token)

    # Fetch the list of hosts from Zabbix
    hosts = zabbix_client.fetch_hosts
    @result += "Fetched hosts..."

    # Save the fetched hosts data to a file
    save_data_to_file(hosts, '/tmp/data_zabbix.json')

    # Process each host in the fetched data
    hosts.each do |host|
      # Instantiate DeviceProcessor to process the host
      processor = DeviceProcessor.new(host, @result)
      processor.process  # Execute the processing of the host
    end

    @result
  end

  private

  # Save the data (hosts) to a file
  def save_data_to_file(data, file_path)
    FileManager.save_to_file(data, file_path)
  end

  # Database-related operations for Zabbix integration
  module Zabbix
    class DatabaseConfig
      # Set up the database connection using the production configuration
      def setup
        config = Rails.application.config.database_configuration[Rails.env]
        ActiveRecord::Base.establish_connection(config)
      rescue StandardError => e
        # Log any error that occurs during the connection setup
        AppLogger.logger.error("Database configuration error: #{e.message}")
        raise
      end

      # Verify that the database connection was successfully established
      def verify_connection
        ActiveRecord::Base.connection  # Attempt to connect to the database
        AppLogger.logger.info("Database connection established successfully.")
        @result += "Database connection established successfully.\n"
      rescue ActiveRecord::ConnectionNotEstablished => e
        # Log any error if the connection fails
        AppLogger.logger.error("Failed to connect to the database: #{e.message}")
        @result += "Failed to connect to the database: #{e.message}\n"
        raise
      end
    end
  end

  # Class to handle interaction with Zabbix API
  class ZabbixClient
    def initialize(api_url, user, password, token)
      @api_url = api_url
      @user = user
      @password = password
      @token = token
    end

    # Method to fetch the list of hosts from Zabbix API
    def fetch_hosts
      payload = build_payload  # Build the payload for the API request
      response = post_request(payload)  # Send the request
      parse_response(response)  # Parse and return the response
    rescue RestClient::ExceptionWithResponse => e
      # Raise an error if fetching hosts fails
      raise "Failed to fetch hosts: #{e.response}"
    end

    private

    # Build the payload (data structure) for the API request
    def build_payload
      {
        jsonrpc: "2.0",  # Zabbix API version
        method: "host.get",  # API method to fetch hosts
        params: {
          output: ["hostid", "name"],            # Request hostid and name in the response
          selectInterfaces: ["ip", "mac"],       # Include IP and MAC from interfaces
          selectParentTemplates: ["name"],       # Include templates linked to the host
          selectTags: ["tag", "value"]           # Include tags attached to the host
        },
        auth: @token,  # Authentication token for Zabbix API
        id: 2          # Request ID for tracking
      }
    end

    # Send a POST request to Zabbix API with the payload
    def post_request(payload)
      connection = Faraday.new(url: @api_url)  # Create a new Faraday connection
      response = connection.post do |req|
        req.url '/api_jsonrpc.php'  # Zabbix API endpoint
        req.headers['Content-Type'] = 'application/json'  # Set request content type
        req.body = payload.to_json  # Convert the payload to JSON format
      end
      response
    end

    # Parse the JSON response from Zabbix API
    def parse_response(response)
      JSON.parse(response.body)['result']  # Return the 'result' field from the JSON response
    end
  end

  # Class responsible for processing each device (host) retrieved from Zabbix
  class DeviceProcessor
    # Predefined list of device types and their corresponding names
    DEVICE_TYPES = {
      "router" => "Router",
      "switch" => "Switch",
      "firewall" => "Firewall",
      "server" => "Server",
      "printer" => "Printer",
      "workstation" => "Workstation",
      "nas" => "Network Attached Storage",
      "database" => "Database",
      "access_point" => "Access Point",
      "gateway" => "Gateway",
      "proxy" => "Proxy",
      "vpn" => "VPN",
      "load_balancer" => "Load Balancer",
      "hypervisor" => "Hypervisor",
      "virtual_machine" => "Virtual Machine",
      "storage" => "Storage",
      "monitor" => "Monitor",
      "scanner" => "Scanner",
      "projector" => "Projector",
      "camera" => "Camera",
      "security_camera" => "Security Camera",
      "ups" => "Uninterruptible Power Supply",
      "badge_reader" => "Badge Reader",
      "access_control" => "Access Control",
      "smart_tv" => "Smart TV",
      "ip_phone" => "IP Phone",
      "tablet" => "Tablet",
      "smartphone" => "Smartphone",
      "lab_equipment" => "Lab Equipment",
      "medical_imaging" => "Medical Imaging Device",
      "mri_machine" => "MRI Machine",
      "ct_scanner" => "CT Scanner",
      "x_ray_machine" => "X-Ray Machine",
      "ultrasound_machine" => "Ultrasound Machine",
      "patient_monitor" => "Patient Monitor",
      "ecg_machine" => "ECG Machine",
      "infusion_pump" => "Infusion Pump",
      "ventilator" => "Ventilator",
      "surgical_robot" => "Surgical Robot",
      "blood_analyzer" => "Blood Analyzer",
      "dna_sequencer" => "DNA Sequencer",
      "electronic_health_record_system" => "EHR System",
      "pacs" => "Picture Archiving and Communication System",
      "ris" => "Radiology Information System",
      "his" => "Hospital Information System",
      "lis" => "Laboratory Information System",
      "pharmacy_automation" => "Pharmacy Automation System",
      "telemedicine_equipment" => "Telemedicine Equipment",
      "nurse_call_system" => "Nurse Call System",
      "patient_bed" => "Smart Patient Bed",
      "temperature_sensor" => "Temperature Sensor",
      "humidity_sensor" => "Humidity Sensor",
      "smoke_detector" => "Smoke Detector",
      "biometric_scanner" => "Biometric Scanner",
      "glucose_meter" => "Glucose Meter",
      "insulin_pump" => "Insulin Pump",
      "wearable_device" => "Wearable Device",
      "smart_light" => "Smart Light",
      "smart_thermostat" => "Smart Thermostat",
      "iot_device" => "IoT Device",
      "body_temperature_sensor" => "Body Temperature Sensor",
      "oxygen_tank" => "Oxygen Tank",
      "emergency_button" => "Emergency Button",
      "wristband_tracker" => "Wristband Tracker",
      "kiosk" => "Patient Check-In Kiosk",
      "defibrillator" => "Defibrillator",
      "biomedical_equipment" => "Biomedical Equipment",
      "disinfectant_robot" => "Disinfectant Robot",
      "robotic_medication_dispenser" => "Robotic Medication Dispenser",
      "healthcare_data_analytics_system" => "Healthcare Data Analytics System",
      "clinical_decision_support_system" => "Clinical Decision Support System",
      "telemetry_device" => "Telemetry Device",
      "hemodialysis_machine" => "Hemodialysis Machine",
      "radiation_therapy_machine" => "Radiation Therapy Machine",
      "endoscope" => "Endoscope",
      "syringe_pump" => "Syringe Pump",
      "nebulizer" => "Nebulizer",
      "spirometer" => "Spirometer",
      "audiometer" => "Audiometer",
      "ophthalmoscope" => "Ophthalmoscope",
      "otoscope" => "Otoscope",
      "treadmill" => "Treadmill",
      "exercise_bike" => "Exercise Bike",
      "ergometer" => "Ergometer",
      "gait_trainer" => "Gait Trainer",
      "pulse_oximeter" => "Pulse Oximeter",
      "electronic_medical_record_system" => "EMR System",
      "hospital_bed_management_system" => "Hospital Bed Management System"
    }.freeze  # Freeze to prevent modification

    def initialize(host, result)
      @host = host
      @result = result || ""
    end

    # Main method to process a host and classify its interfaces
    def process
      @result += "Processing... \n"
      ActiveRecord::Base.transaction do
        begin
          name = normalize_host_name(@host['name'])
          interfaces = @host['interfaces']
          device_type = classify_device(@host)
    
          interfaces.each do |interface|
            ip = normalize_ip(interface['ip'])
            mac = normalize_mac(interface['mac'])
    
            if ip
              update_net_object(ip, name, device_type)
            elsif mac
              update_mac_object(mac, name, device_type)
            end
          end
        rescue ActiveRecord::RecordInvalid => e
          @result += "Failed to process device #{@host['name']}: #{e.message}\n"
          raise ActiveRecord::Rollback
        end
      end
    end
    

    private

    # Normalize the host name by removing trailing digits or underscores
    def normalize_host_name(name)
      name.gsub(/_?\d*$/, '')
    end

    # Normalize the IP address by removing trailing digits or underscores
    def normalize_ip(ip)
      ip.gsub(/_?\d*$/, '')
    end

    # Normalize the MAC address if provided, removing trailing digits or underscores
    def normalize_mac(mac)
      mac.gsub(/_?\d*$/, '') if mac
    end

    # Update or create a network object (IP-based)
    def update_net_object(ip, name, device_type)
      net_object = NetObject.find_or_create_by(value: ip) do |obj|
        obj.name = name
        obj.user_id = nil
        obj.object_type = device_type
      end

      net_object.update(
        name: name,
        user_id: nil,
        object_type: device_type
      )
    end

    # Update or create a MAC object (MAC-based)
    def update_mac_object(mac, name, device_type)
      mac_object = MacObject.find_or_create_by(value: mac) do |obj|
        obj.name = name
        obj.user_id = nil
        obj.object_type = device_type
      end

      mac_object.update(
        name: name,
        user_id: nil,
        object_type: device_type
      )
    end

    # Classify the device based on its Zabbix template or tags
    def classify_device(host)
      template_names = host['parentTemplates'].map { |t| t['name'].downcase } # Fetch template names
      tag_values = host['tags'].map { |t| t['value'].downcase if t['tag'].downcase == 'device_type' }.compact # Fetch device type tags

      # Match template names with device types, default to 'Other' if none matches
      DEVICE_TYPES.find(-> { ['Other'] }) { |key, _| template_names.any? { |name| name.include?(key) } }.last
    end
  end
end
