require 'test_helper'

class Api::V1::AccessPointsTest < ActionDispatch::IntegrationTest

  context 'A superadmin user' do
    setup do
      @user = users(:admin)
      @auth_token = users(:admin).authentication_token
      @uuid = sensors(:organization).uuid
      @ap1 = access_points(:ap1)
      @ap2 = access_points(:ap2)
      @ap_destroy = access_points(:ap_destroy)
      @ap_destroy_1 = access_points(:ap_destroy_1)
      @ap_destroy_2 = access_points(:ap_destroy_2)
      @flow = sensors(:flow)
      # Base api
      @api = 'api/v1'
    end

    should 'list all access points in JSON' do
      get "/#{@api}/access_points?auth_token=#{@auth_token}"
      assert_response :success
      # The number of elements must coincide
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(AccessPoint.count, json['access_points'].size,
                   'An admin user can see all sensors')
    end

    should 'list all access points in CSV' do
      get "/#{@api}/access_points.csv?auth_token=#{@auth_token}"
      assert_response :success
      # The number of elements must coincide
      csv = CSV.parse(response.body)
      # I subtract one row because it's CSV header
      assert_equal(AccessPoint.count, csv.size - 1,
                   'An admin user can see all sensors')
    end

    should 'get an existent AP (JSON)' do
      get "/#{@api}/access_points.json?auth_token=#{@auth_token}",
          mac_address: @ap1.mac_address
      assert_response :success
      # The number of elements must coincide
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(1, json['access_points'].size,
                   'An admin user see an specified access_point')
      assert_equal(@ap1.mac_address, json['access_points'][0]['mac_address'],
                   'MAC address must coincide!')
    end

    should 'get an existent AP (CSV)' do
      get "/#{@api}/access_points.csv?auth_token=#{@auth_token}",
          mac_address: @ap2.mac_address
      assert_response :success
      # The number of elements must coincide
      csv = CSV.parse(response.body)
      assert_equal(1, csv.size - 1,
                   'There aren\'t two access points with some MAC')
      mac_received = csv.last[csv.first.index('mac_address')]
      assert_equal(@ap2.mac_address, mac_received,
                   'MAC address must coincide!')
    end

    # should 'not found when try to get a nonexistent AP (JSON)' do
    #   mac_address = (1..6).map { "%0.2X"%rand(256) }.join(':')

    #   loop do
    #     break if AccessPoint.where(mac_address: mac_address).empty?
    #     mac_address = (1..6).map { "%0.2X"%rand(256) }.join(':')
    #   end

    #   get "/#{@api}/access_points.json?auth_token=#{@auth_token}",
    #       mac_address: mac_address
    #   assert_response :not_found
    #   # The number of elements must coincide
    #   json = JSON.parse(response.body)
    #   assert(json['query'], 'Query value must be true')
    #   assert_equal(0, json['access_points'].size,
    #                'There aren\'t any Access Point with these MAC')
    # end

    # should 'not found when try to get a nonexistent AP (CSV)' do
    #   mac_address = (1..6).map { "%0.2X"%rand(256) }.join(':')

    #   loop do
    #     break if AccessPoint.where(mac_address: mac_address).empty?
    #     mac_address = (1..6).map { "%0.2X"%rand(256) }.join(':')
    #   end

    #   get "/#{@api}/access_points.csv?auth_token=#{@auth_token}",
    #       mac_address: mac_address
    #   assert_response :not_found
    #   # The number of elements must coincide
    #   csv = CSV.parse(response.body)
    #   assert_equal(0, csv.size,
    #                'There aren\'t any access points with this MAC')
    # end

    should 'create an AP' do
      name = 'ANewAP'
      mac_address = '10:00:11:22:33:44'
      ip_address = '10.0.5.4'
      # Enrichment
      market_uuid = 'ASDK123adjhasd2143fsd'
      campus_uuid = 'ASDrewrtfdsasd2143fsd'
      floor_uuid = 'ASDK1225dfgsd2143fsd'
      building_uuid = 'Aadasd234fd2143fsd'

      post "/#{@api}/access_points?auth_token=#{@auth_token}",
           name: name, mac_address: mac_address, sensor_uuid: @uuid,
           ip_address: ip_address, flow_uuid: @flow.uuid,
           enrichment: { market_uuid: market_uuid, campus_uuid: campus_uuid,
                         floor_uuid: floor_uuid, building_uuid: building_uuid }

      assert_response :success
      # Test over JSON response
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      # Check all values
      %w(name mac_address ip_address).each do |property|
        assert_equal(eval(property), json['ap'][property],
                     "The attribute #{property} must coincide with request")
      end
      %w(market_uuid campus_uuid floor_uuid building_uuid).each do |property|
        assert_equal(eval(property), json['ap']['enrichment'][property],
                     "The attribute #{property} must coincide with request")
      end
    end

    should 'create N APs' do
      50.times do
        name = (0...20).map { ('a'..'z').to_a[rand(26)] }.join
        mac_address = (1..6).map { "%0.2X"%rand(256) }.join(':').downcase
        ip_address = Array.new(4) { rand(256) }.join('.')

        post "/#{@api}/access_points?auth_token=#{@auth_token}",
             name: name, mac_address: mac_address, sensor_uuid: @uuid,
             ip_address: ip_address

        assert_response :success
        # Test over JSON response
        json = JSON.parse(response.body)
        assert(json['query'], 'Query value must be true')
        # Check all values
        %w(name mac_address ip_address).each do |property|
          assert_equal(eval(property), json['ap'][property],
                       "The attribute #{property} must coincide with request")
        end
      end
    end

    should 'fail when create an incomplete AP' do
      post "/#{@api}/access_points?auth_token=#{@auth_token}",
             name: 'FailedAP', sensor_uuid: @uuid

      assert_response 422
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false when creation fails')
    end

    should 'update an AP by MAC address' do
      mac_address = @ap1.mac_address
      patch "/#{@api}/access_points/#{mac_address}?auth_token=#{@auth_token}",
            status: 1

      assert_response 200
      # Check if status is updated
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(1, json['ap']['status'], 'Status of AP must change in update')
    end

    should 'update an AP by ID' do
      id = @ap2.id
      patch "/#{@api}/access_points/#{id}?auth_token=#{@auth_token}",
            status: 1

      assert_response 200
      # Check if status is updated
      json = JSON.parse(response.body)

      assert(json['query'], 'Query value must be true')
      assert_equal(1, json['ap']['status'], 'Status of AP must change in update')
    end

    should 'not found when try to update a nonexistent AP' do
      mac_address = '00:00:00:00:00:00'
      patch "/#{@api}/access_points/#{mac_address}?auth_token=#{@auth_token}",
            status: 1

      assert_response 404
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false when creation fails')
    end

    should 'destroy an unique AP' do
      delete "/#{@api}/access_points?auth_token=#{@auth_token}",
             access_points: [{ mac_address: @ap_destroy.mac_address }]

      assert_response 200
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
    end

    should 'destroy an unique AP (Bad params)' do
      delete "/#{@api}/access_points?auth_token=#{@auth_token}",
             access_points: [{ mac_addres: @ap_destroy.mac_address }]

      assert_response 400
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false when destroy fails')
    end

    should 'destroy multiple APs' do
      delete "/#{@api}/access_points?auth_token=#{@auth_token}",
             access_points: [{ mac_address: @ap_destroy_1.mac_address },
                             { mac_address: @ap_destroy_2.mac_address }]

      assert_response 200
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(2, json['destroyed_access_points'],
                   'The number of destroyed access points must be 0')
    end
  end

  context 'An unauthorized user' do
    setup do
      @user = users(:user_domain)
      @auth_token = @user.authentication_token
      @uuid = sensors(:deployment).uuid
      @unauthorized_ap = access_points(:unauthorized_ap)
      # Base api
      @api = 'api/v1'
    end

    should 'not create an AP' do
      name = 'AnAP'
      mac_address = '10:00:11:22:33:54'

      post "/#{@api}/access_points?auth_token=#{@auth_token}",
           name: name, mac_address: mac_address, sensor_uuid: @uuid

      assert_response 401
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false when no permissions')
    end

    should 'not update an AP' do
      id = @unauthorized_ap.id
      patch "/#{@api}/access_points/#{id}?auth_token=#{@auth_token}",
            status: 1

      assert_response 401
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false when no permissions')
    end

    should 'not destroy an AP' do
      delete "/#{@api}/access_points?auth_token=#{@auth_token}",
             access_points: [{ mac_address: @unauthorized_ap.mac_address }]

      assert_response 200
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(0, json['destroyed_access_points'],
                   'The number of destroyed access points must be 0')
    end
  end
end
