require 'test_helper'
require 'json'

class Api::V1::SensorsFlowTest < ActionDispatch::IntegrationTest
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'an admin user' do
    setup do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'global'

      @user = users(:admin)
      @auth_token = users(:admin).authentication_token
      @uuid = sensors(:organization).uuid
      @namespace = sensors(:namespace)
      @campus = sensors(:campus)
      @flow = sensors(:flow)
      @mse = sensors(:mse)
      @meraki = sensors(:meraki)
      # Base api
      @api = 'api/v1'
    end

    %w(flow mse meraki).each do |s|
      should "render #{s} sensors of his top domain" do
        get "/#{@api}/sensors/#{s}?auth_token=#{@auth_token}"
        # Check status
        assert_response :success
        # Check response and number of items
        true_api_json response.body
      end

      should "render #{s} sensors of a domain" do
        get "/#{@api}/sensors/#{@uuid}/#{s}?auth_token=#{@auth_token}"
        # Check status
        assert_response :success
        # Check response and number of items
        true_api_json response.body
      end

      should "create #{s} sensors" do
        post "/#{@api}/sensors/#{s}?auth_token=#{@auth_token}",
             send("new_params_#{s}").merge(parent_id: @user.sensor.id,
                                           license_id: licenses(:two).id)
        # Check status
        assert_response :success
        # Check response and number of items
        true_api_json response.body
      end
    end

    should 'return stream value of a mse' do
      get "/#{@api}/sensors?auth_token=#{@auth_token}",
          parent_uuid: @namespace.uuid
      assert_response :success
      json = true_api_json response.body
      # Check if exists stream key
      check_stream_mse(
        json['sensors'].select { |s| s['id'] == @mse.id }.first)
    end

    should 'return stream value of a mse (mse)' do
      get "/#{@api}/sensors/mse?auth_token=#{@auth_token}"
      assert_response :success
      json = true_api_json response.body
      # Check if exists stream key
      check_stream_mse json['mse'].first
    end

    should 'return stream value of a mse (tree)' do
      get "/#{@api}/sensors/#{@campus.uuid}/tree?auth_token=#{@auth_token}"
      assert_response :success
      json = true_api_json response.body
      # Check if exists stream key
      check_stream_mse(json['sensor']['children'].find { |s| s['id'] == @mse.id })
    end

    should 'can update radius secret of a flow sensor' do
      new_secret = 'lsjkdasdj1231'
      patch "/#{@api}/sensors/#{@flow.uuid}/flow?auth_token=#{@auth_token}",
            radius_secret: new_secret
      assert_response :success
      true_api_json response.body
    end

    should 'move a sensor flow into organization' do
      patch "/#{@api}/sensors/#{@flow.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @uuid

      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
    end

    should 'can\'t move a domain into a sensor flow' do
      patch "/#{@api}/sensors/#{@uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @flow.uuid

      assert(response.response_code == 422,
             "Response code should be 422 : got #{response.response_code}")
    end

    should 'move a sensor mse into organization' do
      patch "/#{@api}/sensors/#{@mse.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @uuid

      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
    end

    should 'can\'t move a domain into a sensor mse' do
      patch "/#{@api}/sensors/#{@uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @mse.uuid

      assert(response.response_code == 422,
             "Response code should be 422 : got #{response.response_code}")
    end

    should 'move a sensor meraki into organization' do
      patch "/#{@api}/sensors/#{@meraki.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @uuid

      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
    end

    should 'can\'t move a domain into a sensor meraki' do
      patch "/#{@api}/sensors/#{@uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @meraki.uuid

      assert(response.response_code == 422,
             "Response code should be 422 : got #{response.response_code}")
    end
  end

  #
  # Check if the current JSON include stream value (for meraki)
  #
  def check_stream_mse(mse_json)
    assert(mse_json.keys.include?('stream'),
           'Stream value must be included in returned JSON')
    # Check stream value
    mse = Sensor.find(mse_json['id'])
    assert_equal(mse.stream, mse_json['stream'],
                 'Stream values must coincide')
  end

  def new_params_flow
    {
      name: "flow#{Random.rand(10_000)}",
      ip: "10.0.#{Random.rand(255)}.#{Random.rand(255)}",
      type: Sensor::FLOW
    }
  end

  def new_params_mse
    {
      name: "mse#{Random.rand(10_000)}",
      stream: (0...20).map { (65 + rand(26)).chr }.join
    }
  end

  def new_params_meraki
    {
      name: "meraki#{Random.rand(10_000)}",
      meraki_validator: (0...20).map { (65 + rand(26)).chr }.join
    }
  end
end
