require 'test_helper'

#
# Test methods added from SensorFlow to Sensor
#
class SensorFlowTest < ActiveSupport::TestCase
  context 'A Flow sensor' do
    setup do
      # Global licensing by default
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'global'

      @flow = sensors(:flow)
      @deployment = sensors(:deployment)
      @client_proxy = sensors(:client_proxy)
    end

    should 'be created' do
      assert @flow.valid?, 'Flow sensor is invalid.'
    end

    should 'not be created with invalid params' do
      name_not_valid = @flow.name
      s_not_valid_name = Sensor.new(name: name_not_valid, type: @flow.type, parent: @flow.root)
      assert_not s_not_valid_name.valid?, 'Flow sensor is not valid: duplicated name.'

      ip_not_valid = '10.0.15'
      s_not_valid_ip = Sensor.new(name: 'Flow not valid', type: @flow.type, parent: @flow.root, ip: ip_not_valid)
      assert_not s_not_valid_ip.valid?, 'Flow sensor is not valid: ip not valid.'

      ip_not_valid = @flow.ip
      s_not_valid_ip = Sensor.new(name: 'Flow not valid', type: @flow.type, parent: @flow.root, ip: ip_not_valid)
      assert_not s_not_valid_ip.valid?, 'Flow sensor is not valid: duplicated ip.'
    end

    should 'be created with the same name but different type or parent' do
      name_valid = @flow.name
      s_valid = Sensor.new(name: name_valid,
                           type: @flow.type,
                           parent: @deployment,
                           ip: '192.168.101.16',
                           license: licenses(:two))
      assert s_valid.valid?, 'Flow sensor is valid.'
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @flow.parent_id = market.id
      assert @flow.valid?, 'Flow moved into market'
    end

    should 'not contain children' do
      market = sensors(:market)

      market.parent_id = @flow.id
      assert_not market.valid?, 'Flow sensor can\'t contain children'
    end

    should 'be created with the same ip but different real_parent' do
      s_valid = Sensor.new(name: @flow.name,
                           type: @flow.type,
                           parent: @client_proxy,
                           real_parent: @client_proxy,
                           ip: @flow.ip,
                           license: licenses(:two))
      assert s_valid.valid?, 'Flow sensor is invalid.'
      s_valid.save

      s_not_valid = Sensor.new(name: 'Flow_not_valid',
                               type: @flow.type,
                               parent: @client_proxy,
                               real_parent: @client_proxy,
                               ip: @flow.ip)
      assert_not s_not_valid.valid?, 'Flow sensor is not valid: duplicated ip.'
    end

    should 'not be created with the same IP and real_parent and observation_id' do
      sensor = Sensor.new(name: 'sensor_flow',
                          type: @flow.type,
                          parent: @flow.parent,
                          ip: @flow.ip,
                          observation_id: nil)

      assert_not sensor.valid?
    end

    should 'be created with the same IP and real_parent but different observation_id' do
      sensor = Sensor.new(name: 'sensor_flow',
                          type: @flow.type,
                          parent: @flow.parent,
                          ip: @flow.ip,
                          observation_id: rand(0..2**32).to_s,
                          license: licenses(:two))

      assert sensor.valid?
    end

    should 'be associated with a organization license' do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'organization'
      license = licenses(:four)
      sensor = Sensor.new(name: 'sensor_license',
                          type: @flow.type,
                          parent: @flow.parent,
                          ip: '10.0.50.1',
                          observation_id: rand(0..2**32).to_s,
                          license: license)
      assert sensor.valid?
    end

    should 'be associated with a global license' do
      license = licenses(:two)
      sensor = Sensor.new(name: 'sensor_license',
                          type: @flow.type,
                          parent: @flow.parent,
                          ip: '10.0.50.1',
                          observation_id: rand(0..2**32).to_s,
                          license: license)
      assert sensor.valid?
    end

    should 'not be associated with an expired license' do
      license = licenses(:one)
      sensor = Sensor.new(name: 'sensor_license',
                          type: @flow.type,
                          parent: @flow.parent,
                          ip: '10.0.50.1',
                          observation_id: rand(0..2**32).to_s,
                          license_id: license.id)
      assert_not sensor.valid?
    end

    should 'not be associated with a license without empty bucket' do
      license = licenses(:two)
      sensor_1 = Sensor.new(name: 'sensor_license_1',
                            type: @flow.type,
                            parent: @flow.parent,
                            ip: '10.0.50.1',
                            observation_id: rand(0..2**32).to_s,
                            license_id: license.id)
      assert sensor_1.save

      sensor_2 = Sensor.new(name: 'sensor_license_2',
                            type: @flow.type,
                            parent: @flow.parent,
                            ip: '10.0.50.2',
                            observation_id: rand(0..2**32).to_s,
                            license_id: license.id)
      assert_not sensor_2.valid?
    end
  end

  context 'A MSE sensor' do
    setup do
      @mse = sensors(:mse)
      @client_proxy = sensors(:client_proxy)

      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @mse.parent_id = market.id
      assert @mse.valid?, 'MSE moved into market'
    end

    should 'can\'t contain children' do
      market = sensors(:market)

      market.parent_id = @mse.id
      assert_not market.valid?, 'MSE sensor can\'t contain children'
    end

    should 'be created with the same stream but different real_parent' do
      s_valid = Sensor.new(name: @mse.name,
                           type: @mse.type,
                           parent: @client_proxy,
                           real_parent: @client_proxy,
                           stream: @mse.stream)
      assert s_valid.valid?, 'MSE sensor is valid.'
      s_valid.save

      s_not_valid = Sensor.new(name: 'MSE_not_valid',
                               type: @mse.type,
                               parent: @client_proxy,
                               real_parent: @client_proxy,
                               stream: @mse.stream)
      assert_not s_not_valid.valid?, 'MSE sensor is not valid: duplicated stream.'
    end
  end

  context 'A Meraki sensor' do
    setup do
      @meraki = sensors(:meraki)

      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @meraki.parent_id = market.id
      assert @meraki.valid?, 'Meraki moved into market'
    end

    should 'can\'t contain children' do
      market = sensors(:market)

      market.parent_id = @meraki.id
      assert_not market.valid?, 'Meraki sensor can\'t contain children'
    end
  end
end
