#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ReputationPoliciesController < ApplicationController
  self.view_paths = "lib/modules/ips/app/views"
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_sensor
  before_action :check_sensor_privileges, only: [:new, :clone, :assign]
  before_action :set_rep_policy_and_check_privileges,
                only: [:destroy, :edit, :update, :import, :import_file]
  before_action :check_license, only: [:new]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    @reputation_policies = if @sensor
        @sensor.reputation_policies
      else
        ReputationPolicy.all
      end
  end

  def show
    @reputation_policy = ReputationPolicy.find(params[:id])
    respond_to do |format|
      format.html do
        @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
        @ip_net_reputations = @reputation_policy.reputations
          .where(reputation_type: Reputation::IP_NET_TYPE)
          .paginate(page: params[:ip_page], per_page: 10)
        @country_reputations = @reputation_policy.reputations
          .where(reputation_type: Reputation::COUNTRY_TYPE)
          .paginate(page: params[:country_page], per_page: 10)
        @continent_reputations =
          @reputation_policy.reputations
            .where(reputation_type: Reputation::CONTINENT_TYPE)
      end
      format.js do
        @reputation_policy = ReputationPolicy.find(params[:id])
        @reputations = @reputation_policy
          .reputations
          .where(reputation_type: params[:reputation_type])
          .paginate(page: params[:page], per_page: 10)
      end
      format.csv { export_csv }
    end
  end

  def new
    @reputation_policy = @sensor.owned_reputation_policies.build
    render layout: false
  end

  def create
    @sensor = Sensor.find(params[:reputation_policy][:sensor_id])

    if current_user.cannot?(:manage, @sensor)
      redirect_to sensors_path,
                  alert: "#{I18n.t("controllers.no_privileges")}"
    end

    return error_404("#{I18n.t("controllers.sensor_no_license")}") unless @sensor.can_assign_policy?

    @reputation_policy = ReputationPolicy.create(reputation_policy_params)
  end

  def edit
    render layout: false
  end

  def update
    return unless @reputation_policy.update(reputation_policy_params)
    flash[:notice] = "#{I18n.t("controllers.reputation_updated")}"
  end

  def destroy
    @reputation_policy.destroy
    redirect_to sensor_reputation_policies_path(@sensor)
  end

  def assign
    reputation_policy = ReputationPolicy.find(params[:id])

    return error_403("#{I18n.t("controllers.no_privileges")}") if current_user.cannot?(:manage, @sensor)

    # Update policy children
    @sensor.assign_reputation_policy_to_children(reputation_policy)

    redirect_to sensor_reputation_policies_path(@sensor)
  end

  ####################################################################################################
  # Funtion to clean the "new lines" in the CSV File
  ####################################################################################################
  def clean_blank_space_file(file)
    File.write(file, File.readlines(file).reject { |s| s.strip.empty? }.join)
  end

  ####################################################################################################
  # Funtion to clean the "new lines" in the CSV File
  ####################################################################################################
  def clean(input)
    input.gsub(/^[\s]*$\n/, "").strip
  end

  ####################################################################################################
  # Function that checks if the CSV file uploaded for the IPVARS and PORTVARS has the correct header
  # The two columns that are obligatory are "Variable" and "Values"
  # The "inherited" columns is filled with false when it has a nil value
  ####################################################################################################
  def check_header(file)
    begin
      headers = CSV.open(file.path, "r:ISO-8859-1:UTF-8") { |csv| csv.first }
      rule_type, action, value = headers
      if rule_type == "Rule_Type_Id" && action == "Action_Id" && value == "Value"
        return true
      else
        return false
      end
    rescue StandardError => e
      return false
    end
  end

  def import
    render layout: false
  end

  def import_file
    content_file = params[:reputation_policy][:csv_file]

    # Before deciding on adding or replacing variables first it checks if the header of the file is correct
    if check_header(content_file)
      # Clean the CSV file. Eliminate white space (new lines)
      clean_blank_space_file(content_file.path)
      if params[:add_rules] == "true"
        rules_count = @reputation_policy
          .add_rules_from_file(content_file,
                               params[:default_action_id])
        msg = "#{rules_count[0]} #{I18n.t("controllers.rules_added")} #{rules_count[1]} #{I18n.t("controllers.rules_updated")}"
      else
        rules_count = @reputation_policy
          .replace_rules_from_file(content_file,
                                   params[:default_action_id])
        msg = "#{rules_count} #{I18n.t("controllers.rules_added")}"
      end
      redirect_to sensor_reputation_policy_path(@sensor, @reputation_policy),
                  notice: msg
    else
      error_msg = "#{I18n.t "controllers.unknown_content_csv"}"
      redirect_to sensor_reputation_policy_path(@sensor, @reputation_policy), alert: error_msg
    end
  end

  def destroy_all
    @reputation_policy = ReputationPolicy.find(params[:id])
    @reputation_policy.reputations
      .where(reputation_type: params[:reputation_type])
      .destroy_all
    if params[:sensor_id]
      @sensor = Sensor.find(params[:sensor_id])
      redirect_to sensor_reputation_policy_path(@sensor, @reputation_policy),
                  notice: "#{I18n.t("controllers.rules_deleted")}"
    else
      redirect_to reputation_policy_path(@reputation_policy),
                  notice: "#{I18n.t("controllers.rules_deleted")}"
    end
  end

  #
  # Clone reputation policy on current sensor
  #
  def clone
    # Find and clone
    reputation_policy = ReputationPolicy.find(params[:id])
    reputation_policy.clone @sensor.id

    redirect_to sensor_reputation_policies_path(@sensor)
  end

  private

  def set_sensor
    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
  end

  def set_rep_policy_and_check_privileges
    @reputation_policy = ReputationPolicy.find(params[:id])
    return unless current_user.cannot? :manage, @reputation_policy.owner
    respond_to do |format|
      format.html do
        redirect_to sensors_path,
                    alert: "#{I18n.t("controllers.no_privileges")}"
      end
      format.js do
        render json: { error: "#{I18n.t("controllers.no_privileges")}" },
               status: :unauthorized
      end
    end
  end

  def check_sensor_privileges
    return unless current_user.cannot? :manage, @sensor
    respond_to do |format|
      format.html do
        redirect_to sensors_path,
                    alert: "#{I18n.t("controllers.no_privileges")}"
      end
      format.js do
        render json: { error: "#{I18n.t("controllers.no_privileges")}" },
               status: :unauthorized
      end
    end
  end

  def reputation_policy_params
    params.require(:reputation_policy).permit(:sensor_id, :name, :description)
  end

  def export_csv
    filename = in_time_machine do
      Time.now.strftime("%Y%m%d%H%M%S") +
      "-ReputationPolicy-" +
      @reputation_policy.clean_name +
      ".csv"
    end
    content = @reputation_policy.rules_to_csv_format
    send_data content, filename: filename
  end

  def check_license
    return error_404("#{I18n.t("controllers.sensor_no_license")}") unless @sensor.can_assign_policy?
  end
end
