#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class RuleDbVersionsController < ApplicationController
  self.view_paths = 'lib/modules/ips/app/views'
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_filter :require_administrative_privileges,
                except: [:index, :so_rules_file]
  before_filter :set_rule_source,
                only: [:edit_source, :update_source, :destroy_source]
  before_filter :set_rule_db_version, only: [:edit, :update]
  skip_before_action :authenticate_user!, only: [:so_rules_file], raise: false
  before_action :validate_cert, only: [:so_rules_file]
  before_action :check_updating, only: [:index]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    respond_to do |format|
      format.html do
        @rule_sources =
          RuleSource.where(is_deleted: false,
                           sensor_id: current_user.sensor.tree(true))
      end
      format.js do
        @rule_source = RuleSource.find(params[:rule_source_id])
        @rule_db_versions = @rule_source.rule_db_versions.where(completed: true)
                            .order(created_at: :desc)
                            .paginate(page: params[:page], per_page: 5)
      end
    end
  end

  def destroy
    rule_db_version = RuleDbVersion.find(params[:id])
    require_privileges_on_sensor(rule_db_version.rule_source)
    if rule_db_version.can_be_deleted?
      rule_db_version.self_and_olders.update_all(is_deleted: true)
      Delayed::Job.enqueue(
        Redborder::Jobs::DeleteRuleDbVersionJob.new(
          rule_db_version.id, current_user.id, true, I18n.locale
        ), priority: 10, run_at: Time.now
      )
    else
      flash[:error] = "#{I18n.t('controllers.rule_version')} #{rule_source.name} (#{rule_db_version_date})"
    end
    redirect_to rule_db_versions_path
  end

  def edit
    render layout: false
  end

  def update
    @rule_db_version.update(rule_db_version_params)
  end

  def new_source
    @rule_source = RuleSource.new(enable: true, has_code: false)
    render layout: false
  end

  def new_custom_source
    @rule_source = RuleSource.new(
      enable: true, has_code: false, is_custom: true
    )
    render layout: false
  end

  def create_custom_source
    @rule_source = RuleSource.new(rule_source_params)
    @rule_source.sensor_id = current_user.sensor.id
    @rule_source.is_custom = true
    @rule_source.cached_tgz = params[:rule_source][:cached_tgz]
    @rule_source.updating = true
    if @rule_source.save
      Delayed::Job.enqueue(
        Redborder::Jobs::CustomRuleUpdatedbJob.new(
          current_user.id, @rule_source.id, true
        ), priority: 10, run_at: Time.now
      )
      redirect_to rule_db_versions_path,
                  notice: 'Rule Source has been created'
    else
      redirect_to rule_db_versions_path,
                  alert: @rule_source.errors.messages
                    .map { |_item, value| "<li>#{value.first}</li>" }.join
    end
  end

  def create_source
    @rule_source = RuleSource.new(rule_source_params)
    @rule_source.sensor_id = current_user.sensor.id
    if @rule_source.save
      redirect_to rule_db_versions_path,
                  notice: 'Rule Source has been created'
    else
      redirect_to rule_db_versions_path,
                  alert: @rule_source.errors
                    .full_messages.map { |msg| "<li>#{msg}</li>" }.join
    end
  end

  def edit_source
    if @rule_source.is_custom
      render 'edit_custom_source', layout: false
    else
      render layout: false
    end
  end

  def update_source
    if @rule_source.update(rule_source_params)
      if @rule_source.is_custom && params[:rule_source][:cached_tgz]
        @rule_source.cached_tgz = params[:rule_source][:cached_tgz]
        @rule_source.updating = true
        @rule_source.save
        Delayed::Job.enqueue(
          Redborder::Jobs::CustomRuleUpdatedbJob.new(
            current_user.id, @rule_source.id, true
          ), priority: 10, run_at: Time.now
        )
      end
      respond_to do |format|
        format.html { redirect_to rule_db_versions_path, notice: 'Rule Source has been updated' }
        format.json { render json: @rule_source }
      end
    else
      respond_to do |format|
        format.html { redirect_to rule_db_versions_path, alert: @rule_source.errors.full_messages.map { |msg| "<li>#{msg}</li>" }.join }
        format.json { render json: @rule_source.errors }
      end
    end
  end

  def destroy_source
    if @rule_source.has_policies_in_use?
      redirect_to rule_db_versions_path,
                  alert: "#{I18n.t('controllers.the_source')} #{@rule_source.name} #{I18n.t('controllers.has_related')}"
    else
      @rule_source.update(is_deleted: true)
      Delayed::Job.enqueue(
        Redborder::Jobs::DeleteRuleSourceJob.new(
          @rule_source.id, current_user.id, true, I18n.locale
        ), priority: 10, run_at: Time.now
      )
      redirect_to rule_db_versions_path
    end
  end

  def force_update
    DelayedJob.force_rule_update
    redirect_to rule_db_versions_path
  end

  def check_update
    @rule_sources = RuleSource.where(
        is_deleted: false,
        sensor_id: current_user.sensor.tree(true)
    )
  end

  #
  # Receive request from IDS to download rules. BE CAREFUL with the
  # authentication of this method, an IDS is authenticated by certificate
  # (not user)
  #
  # == Returns:
  # A compressed file with rules
  #
  def so_rules_file
    rule_db_version = RuleDbVersion.find(params[:id])
    uri = rule_db_version.so_rules_uri

    DittoCode::Exec.if 'ENTERPRISE' do
      http = Net::HTTP.new(uri.host, uri.port)
      http.use_ssl = true
      # TODO: prevent use VERIFY_NONE
      http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      resp = http.get(uri.request_uri)

      send_data resp.body, filename: 'so_rules.tgz'
    end

    DittoCode::Exec.if 'COMMUNITY' do
      send_file uri
    end
  end

  def help

  end

  private

  def require_privileges_on_sensor(source)
    return true if current_user.can?(:manage, source.sensor)
    no_enough_privileges
  end

  def set_rule_source
    @rule_source = RuleSource.find(params[:rule_source_id])
    require_privileges_on_sensor(@rule_source)
  end

  def set_rule_db_version
    @rule_db_version = RuleDbVersion.find(params[:id])
    require_privileges_on_sensor(@rule_db_version.rule_source)
  end

  def rule_db_version_params
    params.require(:rule_db_version).permit(:tag, :checked)
  end

  def rule_source_params
    params.require(:rule_source).permit(
      :name, :description, :tgzurl, :md5url, :enable, :code
    )
  end

  def check_updating
    unless DelayedJob.rule_update && (DelayedJob.rule_update.running? ||
      DelayedJob.rule_update.run_at < (Time.now + 1.minute))
      RuleSource.where(is_custom: false).update_all(updating: false)
    end

    return if DelayedJob.custom_rule_update
    RuleSource.where(is_custom: true).update_all(updating: false)
  end
end
