#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class EventFilter < ApplicationRecord
  TRACK_TYPE = {
    1 => { id: 'by_src', name: "#{I18n.t 'models.by_source'}" },
    2 => { id: 'by_dst', name: "#{I18n.t 'models.by_destination'}" }
  }

  before_save :set_default_values

  belongs_to :policy
  belongs_to :user

  after_save :update_sensors
  before_destroy :update_sensors

  validate :uniqueness_event_filter_per_rule, if: Proc.new { |event_filter| ['threshold', 'limit', 'both'].include? event_filter.filter_type }
  validate :format_for_src_dst_ip, if: Proc.new { |event_filter| event_filter.filter_type == 'suppress' }

  def signature_name
    rule = (Rule.where(rule_sid: rule_sid, rule_gid: rule_gid, rule_db_version_id: policy.rule_db_versions.map(&:id), default_enabled: [true, false]).first or
            Rule.where(rule_sid: rule_sid, rule_gid: rule_gid, default_enabled: [true, false]).order(:rule_rev).last)
    rule.msg if rule
  end

  def threshold?
    filter_type == 'threshold'
  end

  def limit?
    filter_type == 'limit'
  end

  def both?
    filter_type == 'both'
  end

  def suppress?
    filter_type == 'suppress'
  end

  def rule
    Rule.find_by(rule_db_version_id: policy.rule_db_versions.map(&:id), rule_sid: rule_sid, rule_gid: rule_gid)
  end

  private

  def set_default_values
    self.position = (EventFilter.maximum(:position) || 0) + 1
    self.source_ip = nil if source_ip.blank?
    self.destination_ip = nil if destination_ip.blank?

    if source_ip.present?
      source_object = NetObject.find_by(name: source_ip)
      self.source_ip = source_object.value if source_object
    end

    if destination_ip.present?
      destination_object = NetObject.find_by(name: destination_ip)
      self.destination_ip = destination_object.value if destination_object
    end
  end

  def update_sensors
    Sensor.where(type: Sensor::IPS_BINDING, policy: policy).update_all(need_apply: true)
  end

  def uniqueness_event_filter_per_rule
    if new_record?
      event_filters = EventFilter.where(rule_sid: rule_sid, rule_gid: rule_gid, filter_type: ["threshold", "limit", "both"], policy_id: policy_id)
      if event_filters.present?
        errors.add(:filter_type, "#{I18n.t('models.only_one_event')}")
      end
    else
      event_filters = EventFilter.where(rule_sid: rule_sid, rule_gid: rule_gid, filter_type: ["threshold", "limit", "both"], policy_id: policy_id).where.not(id: id)
      if event_filters.present?
        errors.add(:filter_type, "#{I18n.t('models.only_one_event')}")
      end
    end
  end

  def format_for_src_dst_ip
    if source_ip.present?
      begin
        IPAddr.new(source_ip)
      rescue
        errors.add(:source_ip, "#{I18n.t('models.not_valid')}")
      end
    end

    if destination_ip.present?
      begin
        IPAddr.new(destination_ip)
      rescue
        errors.add(:destination_ip, "#{I18n.t('models.not_valid')}")
      end
    end
  end
end
