#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Policy < ApplicationRecord
  include Redborder::Audit

  # ==========================================================
  # Custom attributes
  # ==========================================================

  DEFAULT_ACTIONS = {
    1 => { text: "#{I18n.t('models.alert')}" },
    2 => { text: "#{I18n.t('models.alert_connectivity')}",
           factory_policy_name: 'connectivity-ips' },
    3 => { text: "#{I18n.t('models.alert_balanced')}",
           factory_policy_name: 'balanced-ips' },
    4 => { text: "#{I18n.t('models.alert_security')}",
           factory_policy_name: 'security-ips' },
    5 => { text: "#{I18n.t('models.recommended_connectivity')}",
           factory_policy_name: 'connectivity-ips' },
    6 => { text: "#{I18n.t('models.recommended_balanced')}",
           factory_policy_name: 'balanced-ips' },
    7 => { text: "#{I18n.t('models.recommended_security')}",
           factory_policy_name: 'security-ips' }
  }

  # ==========================================================
  # Relations
  # ==========================================================

  has_many :policy_rules, dependent: :destroy
  has_many :rules, through: :policy_rules
  has_many :sensors
  has_many :policy_rule_db_versions, dependent: :destroy
  has_many :rule_db_versions, through: :policy_rule_db_versions
  has_many :rule_sources, through: :rule_db_versions
  has_many :event_filters, dependent: :destroy
  has_many :stored_policies, dependent: :destroy

  belongs_to :owner, class_name: 'Sensor', foreign_key: 'sensor_id'

  # ================================================================
  # Callbacks
  # ================================================================

  after_create :create_event_filters

  # ================================================================
  # Validations
  # ================================================================

  validates :name, presence: true
  validate :uniqueness_name_for_policies
  validate :at_least_one_rule_db_version

  audited only: [:name, :description, :default_enabled_rules,
                 :auto_resolve_flowbits, :auto_add_new_rules, :block_updates,
                 :default_action, :updating],
          on: [:create, :update, :destroy]

  # ================================================================
  # Public methods
  # ================================================================

  #
  # => Get name without blank spaces
  #
  def clean_name
    name.gsub(/\s+|\.+/, '_')
  end

  #
  # => Clones the current policy on the specified sensor with a new name
  #
  def clone(to_sensor_id)
    matched = /(.*) \((\d+)\)$/.match(name)
    current_name = matched.nil? ? name : matched[1]
    current_number = matched.nil? ? 0 : matched[2].to_i

    loop do
      current_number += 1
      clone_name = "#{current_name} (#{current_number})"

      next if Policy.find_by(name: clone_name)

      created = Policy.create(name: clone_name, description: description,
                              sensor_id: to_sensor_id, updating: true,
                              default_enabled_rules: default_enabled_rules,
                              auto_resolve_flowbits: auto_resolve_flowbits,
                              rule_db_version_ids: rule_db_versions.map(&:id))

      return created if created.save
    end
  end

  #
  # => Get Event Filters from policy
  #    It can be filtered by msg
  #
  def filter_events_rules(msg = '')
    thresholds(%w(limit threshold both), msg)
  end

  #
  # => Get Event Suppresions from policy
  #    It can be filtered by msg
  #
  def suppress_events_rules(msg = '')
    thresholds('suppress', msg)
  end

  #
  # => Get Event Suppresions from policy.
  #    Discard the suppress events that
  #    has source and destination IPs.
  #    It can be filtered by msg
  #
  def suppress_events_rules_for_ips_v2(msg = '')
    thresholds('suppress', msg).reject do |er|
      er.source_ip && er.destination_ip
    end
  end

  #
  # => Get both Event Suppresion and Filter
  #
  def all_events_rules(msg = '')
    thresholds(%w(limit threshold both suppress), msg)
  end

  #
  # => Check if there is a rule update for the policy.
  #
  def update_available?
    return false if block_updates

    rule_db_versions.each do |rule_db_version|
      return true if rule_db_version.update_available?
    end

    false
  end

  #
  # => Update policy to a newest ruleset
  #
  def update_rule_db_versions
    return false unless update_available?

    rule_db_versions.each do |rule_db_version|
      next unless rule_db_version.update_available?

      last_rule_db_version = rule_db_version.rule_source
                             .last_rule_db_version_available

      # Active new rules if options is active
      if auto_add_new_rules

        new_rule_db_versions = RuleDbVersion.where("rule_source_id = ?
                                                    AND created_at > ?
                                                    AND created_at <= ?
                                                    AND completed = ?",
                                                    rule_db_version.rule_source_id,
                                                    rule_db_version.created_at,
                                                    last_rule_db_version.created_at,
                                                    true).order(created_at: :asc)

        if [1, 2, 3, 4].include? default_action
          rules = case default_action
                  when 1
                    Rule.where(is_new: true,
                               default_enabled: true,
                               rule_db_version_id: new_rule_db_versions.map(&:id))
                  else
                    Rule.includes(:factory_policies)
                        .where(is_new: true,
                               default_enabled: true,
                               rule_db_version_id: new_rule_db_versions.map(&:id),
                               factory_policies: { name: DEFAULT_ACTIONS[default_action][:factory_policy_name] })
                  end

          rules.each do |rule|
            new_rule = last_rule_db_version.rules.find_by(default_enabled: true,
                                                          rule_sid: rule.rule_sid,
                                                          rule_gid: rule.rule_gid)
            PolicyRule.create(policy_id: id,
                              rule_id: new_rule.id,
                              action: Action.alert) if new_rule
          end

        else

          factory_policy = FactoryPolicy.find_by name: DEFAULT_ACTIONS[default_action][:factory_policy_name]

          rules = Rule.includes(:factory_policy_rules)
                  .where(is_new: true,
                         default_enabled: true,
                         rule_db_version_id: new_rule_db_versions.map(&:id),
                         factory_policy_rules: { factory_policy_id: factory_policy.id })

          rules.each do |rule|
            action_id = FactoryPolicyRule.find_by(rule_id: rule.id,
                                                  factory_policy: factory_policy).action_id
            new_rule = last_rule_db_version.rules
                       .find_by(default_enabled: true,
                                rule_sid: rule.rule_sid,
                                rule_gid: rule.rule_gid)
            PolicyRule.create(policy_id: id,
                              rule_id: new_rule.id,
                              action_id: action_id) if new_rule
          end

        end

      end

      # Search for equal rules in new version
      self.rules.where(rule_db_version: rule_db_version).each do |rule|
        new_rule = Rule.find_by(rule_db_version: last_rule_db_version,
                                rule_sid: rule.rule_sid,
                                rule_gid: rule.rule_gid)
        next if new_rule.nil?

        action_id = PolicyRule.find_by(policy_id: id,
                                       rule_id: rule.id).action_id

        policy_rule = PolicyRule.find_by(policy_id: id, rule_id: new_rule.id)

        if policy_rule
          policy_rule.update(action_id: action_id)
        else
          PolicyRule.create(policy_id: id,
                            rule_id: new_rule.id,
                            action_id: action_id)
        end
      end

      # Destroy actions for old rules
      PolicyRule.includes(:rule)
        .where(policy_id: id,
               rules: { rule_db_version_id: rule_db_version.id }).destroy_all

      rule_db_versions.delete(rule_db_version)
      rule_db_versions << last_rule_db_version

    end
  end

  #
  # => Show policy's status
  #    It cant be 'Updating' or 'Update Available'
  #
  def status
    if updating
      'Updating'
    elsif update_available?
      'Update Available'
    end
  end

  #
  # => Show policy rules in csv format
  #    Action Rule_Sid Rule_Gid Rule_Msg
  #    Columns are separated by commas
  #
  def rules_to_csv_format
    policy_rules.includes(:rule, :action)
                .where(rules: { rule_db_version_id: rule_db_versions.map(&:id) })
                .map { |pr| pr.action.name + ',' + pr.rule.rule_sid.to_s + ',' + pr.rule.rule_gid.to_s + ',' + pr.rule.msg.gsub("\,", ' ') }.join("\n")
  end

  def filtered_rules_to_csv_format(is_query, not_query, msg, query_tables)
    query = if query_tables.is_a?(Symbol)
              policy_rules.includes(:rule, :action)
                          .where(is_query)
            else
              policy_rules.includes(:rule, :action)
                          .joins(rule: query_tables[:rules])
                          .where(is_query)
            end

    not_query.each do |item|
      query = query.where.not(item)
    end

    query = query.where('LOWER(rules.text) like ?', "%#{msg.downcase.split.join('%')}%") if msg.present?

    query.map { |pr| pr.action.name + ',' + pr.rule.rule_sid.to_s + ',' + pr.rule.rule_gid.to_s + ',' + pr.rule.msg.gsub("\,", ' ') }.join("\n")
  end

  #
  # => Show policy event filters in csv format
  #    Event_Filer_Type Rule_Sid Rule_Gid Count
  #    Seconds Src_Addr Dst_Addr Tracked_Type Comment
  #    Columns are separated by commas
  #
  def event_rules_to_csv_format
    csv = "Event_Filter_Type,Signature_Sid,Signature_Gid,Count,Seconds,Source,Destination,Tracked_Type,Comment\n"
    csv += filter_events_rules.map do |e|
      [e.filter_type, e.rule_sid, e.rule_gid, e.count,
       e.seconds, nil, nil,
       e.tracked_type == 1 ? 'by_src' : 'by_dst', e.comment].join(',')[0...-1]
    end.join("\n")
    csv += "\n" if filter_events_rules.present?
    csv += suppress_events_rules.map do |e|
      [e.filter_type, e.rule_sid, e.rule_gid, nil, nil, e.source_ip,
       e.destination_ip, nil, e.comment].join(',')[0...-1]
    end.join("\n")
  end

  #
  # => Add rules passed from csv file into policy.
  #    If the rule is present in policy, the action will be updated.
  #
  def add_rules_from_file(content_file)
    file = File.open(content_file, 'r')

    updated_count = created_count = 0

    rule_db_version_ids = rule_db_versions.map(&:id)

    PolicyRule.transaction do
      while line = file.gets
        action_name, rule_sid, rule_gid = line.sub(/[\n]+/, '').split(',')
        action = Action.find_by(name: action_name)

        policy_rule = PolicyRule.joins(:rule)
                      .where(policy_id: id, rules: { rule_sid: rule_sid,
                                                     rule_gid: rule_gid,
                                                     rule_db_version_id: rule_db_version_ids }).first

        if policy_rule && action_name == 'clear'
          policy_rule.delete
          updated_count += 1
        elsif policy_rule
          policy_rule.update(action: action)
          updated_count += 1 if policy_rule.errors.blank?
        elsif action_name != 'clear'
          rule = Rule.find_by(rule_sid: rule_sid, rule_gid: rule_gid,
                              rule_db_version_id: rule_db_version_ids)
          next unless rule

          policy_rule = PolicyRule.create(policy_id: id,
                                          rule_id: rule.id,
                                          action_id: action.id)

          created_count += 1 if policy_rule.errors.blank?
        end
      end
    end

    file.close

    [created_count, updated_count]
  end

  #
  # => Add rules passed from csv file into policy.
  #    The previous rules will be deleted.
  #
  def replace_rules_from_file(content_file)
    file = File.open(content_file, 'r')

    created_count = 0

    rule_db_version_ids = rule_db_versions.map(&:id)

    PolicyRule.transaction do
      policy_rules.delete_all

      while line = file.gets
        action_name, rule_sid, rule_gid = line.sub(/[\n]+/, '').split(',')
        action = Action.find_by(name: action_name)

        rule = Rule.find_by(rule_sid: rule_sid, rule_gid: rule_gid,
                            rule_db_version_id: rule_db_version_ids)

        next unless rule

        policy_rule = PolicyRule.create(policy_id: id,
                                        rule_id: rule.id,
                                        action_id: action.id)

        created_count += 1 if policy_rule.errors.blank?
      end
    end

    file.close

    created_count
  end

  #
  # => Add event filters and suppresions passed
  #    from csv file into policy.
  #    If the event filter is present, update options.
  #
  def add_event_rules_from_file(content_file, user_id)
    file = File.open(content_file.tempfile)

    updated_count = created_count = 0

    EventFilter.transaction do
      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\n]+/, '').split(',')

        filter_type, rule_sid, rule_gid,
        count, seconds, source_ip,
        destination_ip, tracked_type, comment = columns

        tracked_type = case tracked_type
                       when 'by_src'
                         1
                       when 'by_dst'
                         2
                       else
                         nil
                       end

        event_filter = event_filters.find_by(rule_sid: rule_sid,
                                             rule_gid: rule_gid,
                                             filter_type: %w('limit', 'threshold', 'both'))

        if event_filter && %w(limit threshold both).include?(filter_type)

          unless [event_filter.count, event_filter.seconds,
                  event_filter.filter_type, event_filter.tracked_type,
                  event_filter.comment].eql?([count.to_i, seconds.to_i,
                                              filter_type, tracked_type,
                                              comment])

            event_filter.update(count: count, seconds: seconds,
                                filter_type: filter_type,
                                tracked_type: tracked_type, user_id: user_id,
                                comment: comment)

            updated_count += 1 if event_filter.errors.blank?

          end

        elsif %w(limit threshold both).include?(filter_type)

          event_filter = event_filters.create(rule_sid: rule_sid, count: count,
                                              rule_gid: rule_gid,
                                              seconds: seconds,
                                              filter_type: filter_type,
                                              tracked_type: tracked_type,
                                              user_id: user_id,
                                              comment: comment)

          created_count += 1 if event_filter.errors.blank?

        else

          event_filter = event_filters.find_by(rule_sid: rule_sid,
                                               rule_gid: rule_gid,
                                               filter_type: 'suppress',
                                               source_ip: source_ip,
                                               destination_ip: destination_ip)

          if event_filter.nil?
            event_filter = event_filters.create(rule_sid: rule_sid,
                                                rule_gid: rule_gid,
                                                filter_type: 'suppress',
                                                source_ip: source_ip,
                                                destination_ip: destination_ip,
                                                user_id: user_id)
            created_count += 1 if event_filter.errors.blank?

          elsif event_filter.comment != comment
            event_filter.update(comment: comment)
            updated_count += 1 if event_filter.errors.blank?
          end
        end

      end
    end

    file.close

    [created_count, updated_count]
  end

  #
  # => Add event filters and suppresions passed
  #    from csv file into policy.
  #    The previous events will be deleted.
  #
  def replace_event_rules_from_file(content_file, user_id)
    file = File.open(content_file.tempfile)

    created_count = 0

    EventFilter.transaction do
      event_filters.delete_all

      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\n]+/, '').split(',')

        filter_type, rule_sid, rule_gid,
        count, seconds, source_ip, destination_ip,
        tracked_type, comment = columns

        tracked_type = case tracked_type
                       when 'by_src'
                         1
                       when 'by_dst'
                         2
                       else
                         nil
                       end

        event_filter = event_filters.create(rule_sid: rule_sid,
                                            rule_gid: rule_gid,
                                            filter_type: filter_type,
                                            count: count, seconds: seconds,
                                            source_ip: source_ip,
                                            destination_ip: destination_ip,
                                            tracked_type: tracked_type,
                                            user_id: user_id, comment: comment)

        created_count += 1 if event_filter.errors.blank?
      end
    end

    file.close

    created_count
  end

  #
  # => Returns isset flowbits from policy's rules
  #
  def isset_flowbits
    FlowbitState.includes(flowbit_rules: [:flowbit, rule: [:policy_rules]])
      .where(flowbits: { name: 'isset' },
             policy_rules: { policy_id: id })
  end

  #
  # => Returns set flowbits from policy's rules
  #
  def set_flowbits
    FlowbitState.includes(flowbit_rules: [:flowbit, rule: [:policy_rules]])
      .where(flowbits: { name: 'set' },
             policy_rules: { policy_id: id })
  end

  #
  # => Check unresolved flowbits and returns them.
  #
  def unresolved_flowbits
    isset_flowbits - set_flowbits
  end

  #
  # => Add rules to the policy to resolve flowbits.
  #    Rules will be added with alert action.
  #
  def resolve_flowbits
    unresolved_flowbits_count = unresolved_flowbits.count

    PolicyRule.transaction do

      unresolved_flowbits.each do |flowbit_state|

        rule_isset = Rule.includes(flowbit_rules: [:flowbit, :flowbit_state])
                     .where(rule_db_version_id: rule_db_versions.map(&:id),
                            default_enabled: default_enabled_rules? ? true : [true, false],
                            flowbit_states: { id: flowbit_state.id },
                            flowbits: { name: 'isset' }).first

        rule_set = Rule.includes(flowbit_rules: [:flowbit, :flowbit_state])
                   .where(rule_db_version_id: rule_db_versions.map(&:id),
                          default_enabled: default_enabled_rules? ? true : [true, false],
                          flowbit_states: { id: flowbit_state.id },
                          flowbits: { name: 'set' }).first

        next unless rule_set && PolicyRule.where(rule_id: rule_set.id,
                                                 policy_id: id).empty?

        PolicyRule.create(rule_id: rule_set.id,
                          policy_id: id, action_id: Action.alert.id,
                          dependency_rule_id: rule_isset.id)

      end

    end

    resolve_flowbits if unresolved_flowbits_count > unresolved_flowbits.count
  end

  # ================================================================
  # Protected methods
  # ================================================================

  protected

  #
  # => Search EventFilters from policy by filter_type and msg of the
  #    associated rule.
  #
  def thresholds(filter_type, msg)
    if msg.present?
      rule_db_version_ids = rule_db_versions.map(&:id)
      rules_sid_gid = Rule.where(rule_db_version_id: rule_db_version_ids)
                          .where('LOWER(rules.msg) like ?', "%#{msg.downcase.split.join('%')}%")
                          .map { |r| [r.rule_sid, r.rule_gid] }

      rules_sid_gid = (rules_sid_gid &
                       EventFilter.where(policy_id: id, filter_type: filter_type)
                                  .select(:rule_sid, :rule_gid)
                                  .map { |r| [r.rule_sid, r.rule_gid] })

      @query = nil
      rules_sid_gid.each_with_index do |rule_sid_gid, index|
        condition = EventFilter.arel_table[:rule_sid].eq(rule_sid_gid[0])
               .and(EventFilter.arel_table[:rule_gid].eq(rule_sid_gid[1]))

        @query = index.zero? ? condition : @query.or(condition)
      end

      if @query.nil?
        event_filters.where(id: nil)
      else
        event_filters.where(filter_type: filter_type)
                     .where(@query)
                     .order('position ASC')
      end
    else
      event_filters.where(filter_type: filter_type)
                   .order('position ASC')
    end
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  #
  # => Validate the uniqueness of policy's name
  #
  def uniqueness_name_for_policies
    return unless (new_record? || name_changed?) &&
                  Policy.where(sensor_id: sensor_id, name: name).count > 0
    errors.add(:name, "#{I18n.t('models.policy_already')} #{name} #{I18n.t('models.for')} #{owner.complete_name}")
  end

  #
  # => Each policy has to have at least one RuleDbVersion
  #
  def at_least_one_rule_db_version
    return unless new_record? && rule_db_version_ids.size == 0
    errors.add(:rule_db_versions, "#{I18n.t('models.have_check')}")
  end

  #
  # => Create the inline event filters.
  #    Some rulesets like Emerging Threats has threshold
  #    inside the rules.
  #
  def create_event_filters
    rule_db_versions.each do |rule_db_version|
      rule_db_version.rules.where(has_threshold: true).each do |rule|
        EventFilter.create(rule_sid: rule.rule_sid,
                           rule_gid: rule.rule_gid,
                           policy_id: id,
                           filter_type: rule.threshold_type,
                           tracked_type: (rule.threshold_track == 'by_src' ? 1 : 2),
                           count: rule.threshold_count,
                           seconds: rule.threshold_seconds)
      end
    end
  end
end
