#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Reputation < ApplicationRecord
  include Redborder::Audit

  IP_NET_TYPE = 1
  COUNTRY_TYPE = 2
  CONTINENT_TYPE = 3

  ACTION = {
    1 => { name: 'analyze', color: '#ffab2e', action_2: 'monitor' },
    2 => { name: 'bypass', color: 'green', action_2: 'white' },
    3 => { name: 'drop', color: 'red', action_2: 'block' }
  }

  belongs_to :reputation_policy

  validates :reputation_type, :action, numericality: { only_integer: true }
  validate :check_ip_value

  after_save :update_sensors
  before_destroy :update_sensors
  before_create :set_position

  audited on: [:create, :update, :destroy]

  def humanize_value
    case reputation_type
    when 1
      value
    when 2
      Country[value].name
    when 3
      Continent[value].name
    end
  end

  def value_to_alpha3
    return '' if reputation_type == IP_NET_TYPE
    Country[value].alpha3
  end

  def action_name
    ACTION[action][:name]
  end

  def action_color
    ACTION[action][:color]
  end

  def value=(v)
    begin
      super(NetAddr::CIDR.create(v).to_s.gsub(/\/32$/, ""))
    rescue
      super(v)
    end
  end

  private

  def check_ip_value
    return unless reputation_type == IP_NET_TYPE

    begin
      IPAddr.new(value)
    rescue ArgumentError => e
      errors.add(:value, e.message)
    end
  end

  def update_sensors
    Sensor.where(type: Sensor::IPS_GROUP, reputation_policy: self.reputation_policy).update_all(need_reputation_apply: true)
  end

  def set_position
    return if position?
    self.position = 1 + (reputation_policy.reputations.where(reputation_type: reputation_type).maximum(:position) || 0)
  end
end
