#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Define reputation policies
#
class ReputationPolicy < ApplicationRecord
  # Includes
  include Redborder::Audit

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------
  has_many :sensors
  has_many :reputations, -> { order('position ASC') }
  belongs_to :owner, class_name: "Sensor", foreign_key: "sensor_id"

  # -------------------------------------------------------------
  # Audited
  # -------------------------------------------------------------
  audited only: [:name, :description], on: [:create, :update, :destroy]

  # -------------------------------------------------------------
  # Methods
  # -------------------------------------------------------------

  #
  # Clean the name based changing spaces and . with underscores
  #
  def clean_name
    name.gsub(/\s+|\.+/, '_')
  end

  #
  # Return rules as CSV format
  def rules_to_csv_format
    header = %w(Rule_Type_Id Action_Id Value)
    csv = header.join(',') + "\n"
    csv += reputations.order(:reputation_type, :position).map{|r| r.reputation_type.to_s + "," + r.action.to_s + "," + r.value}.join("\n")
  end

  #
  # Clones current reputation policy on the specified sensor with a new name
  #
  # == Parameters:
  # sensor_id::
  #   Integer with the ID of the sensor to clone the reputaiton policies
  #
  # == Returns:
  # Reputation policy object or nil
  #
  def clone(to_sensor_id)
    matched = /(.*) \((\d+)\)$/.match(name)
    current_name = matched.nil? ? name : matched[1]
    current_number = matched.nil? ? 0 : matched[2].to_i

    loop do
      current_number += 1
      clone_name = "#{current_name} (#{current_number})"
      next if ReputationPolicy.find_by(name: clone_name)

      created = ReputationPolicy
                  .create(name: clone_name, description: description,
                          sensor_id: to_sensor_id)

      if created.save
        # Iterate over current policies
        reputations.each do |reputation|
          created.reputations
            .create(reputation_type: reputation.reputation_type,
                    value: reputation.value,
                    action: reputation.action,
                    position: reputation.action)
        end
        # Break and return
        return created
      end
    end
  end

  # -------------------------------------------------------------
  # Work with files
  # -------------------------------------------------------------

  #
  # Add rules passed from csv file into reputation policy. If the rule is
  # present in reputation policy, the action will be updated.
  #
  def add_rules_from_file(content_file, action=nil)
    file = File.open(content_file.tempfile)
    updated_count = 0
    created_count = 0

    Reputation.transaction do
      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\r\n]+|[\n]+|[\r]+/, '').split(',')

        # Check how many columns is the csv file.
        case columns.size
        when 1
          value = columns[0]
          reputation_type = (value.length == 2 ? Reputation::COUNTRY_TYPE : Reputation::IP_NET_TYPE)
        when 2
          action, value = columns
          reputation_type = (value.length == 2 ? Reputation::COUNTRY_TYPE : Reputation::IP_NET_TYPE)
        else
          reputation_type, action, value = columns
        end

        reputation = reputations.find_by(reputation_type: reputation_type, value: value)

        if reputation and action.to_i != reputation.action
          reputation.update(action: action)
          updated_count += 1 if reputation.errors.blank?
        elsif reputation.nil?
          reputation = Reputation.create(reputation_policy_id: self.id, action: action, reputation_type: reputation_type, value: value)
          created_count += 1 if reputation.errors.blank?
        end
      end
    end
    file.close
    return created_count, updated_count
  end

  #
  # Add rules passed from csv file into reputation policy.
  # The previous rules will be deleted.
  #
  def replace_rules_from_file(content_file, action=nil)
    file = File.open(content_file.tempfile)
    created_count = 0

    Reputation.transaction do
      reputations.delete_all

      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\r\n]+|[\n]+|[\r]+/, '').split(',')

        # Check how many columns is the csv file.
        case columns.size
        when 1
          value = columns[0]
          reputation_type = (value.length == 2 ? Reputation::COUNTRY_TYPE : Reputation::IP_NET_TYPE)
        when 2
          action, value = columns
          reputation_type = (value.length == 2 ? Reputation::COUNTRY_TYPE : Reputation::IP_NET_TYPE)
        else
          reputation_type, action, value = columns
        end

        reputation = Reputation.create(reputation_policy_id: self.id, action: action, reputation_type: reputation_type, value: value)
        created_count += 1 if reputation.errors.blank?
      end
    end
    file.close
    created_count
  end
end
