#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class RuleSource < ApplicationRecord

  include Redborder::Audit
  include StorageFile

  serialize :so_rules, Array
  serialize :search, Array

  has_many :rule_db_versions, dependent: :destroy

  validates :name, uniqueness: true
  validates :name, presence: { message: "#{I18n.t('controllers.name_required')}" }

  validates :cached_tgz, presence: { message: "#{I18n.t('controllers.file_required')}" }, if: :is_custom

  belongs_to :sensor

  audited only: [:name, :description, :tgzurl, :md5url, :enable],
          on: [:create, :update, :destroy]

  default_scope { order('created_at DESC') }

  storage_file :cached_tgz, path: 'rb-webui/:class/:id/:filename'

  before_create :default_values

  def code=(value)
    super(value.try(:strip))
  end

  def clean_name
    name.gsub(/\s+|\.+/, "_")
  end

  def last_rule_db_version_available
    rule_db_versions.where(completed: true, checked: true).order("created_at DESC").first
  end

  def to_partial_path
    "rule_db_versions/rule_source"
  end

  def full_tgzurl
    if has_code and code.present?
      tgzurl.sub('$$code$$', code)
    else
      tgzurl
    end
  end

  def full_md5url
    if md5url and has_code and code.present?
      md5url.sub('$$code$$', code)
    else
      md5url
    end
  end

  def cached_tgz_uri
    return nil if cached_tgz.path.nil?
    StorageFile.get_uri cached_tgz.path
  end

  def cached_tgz_path
    if StorageFile.s3_enabled?
      "s3://redborder" + cached_tgz.path
    else
      cached_tgz.path
    end
  end

  def compression_type
    if !self.is_custom or cached_tgz.path.match /\.tar.gz$|\.tgz$/
      "tgz"
    elsif cached_tgz.path.match /\.zip$/
      "zip"
    else
      "unknown"
    end
  end

  def has_policies_in_use?
    policy_ids = Sensor.where.not(policy_id: nil).select(:policy_id).map(&:policy_id).uniq
    PolicyRuleDbVersion.where(policy_id: policy_ids, rule_db_version_id: rule_db_versions.map(&:id)).present?
  end

  def rule_db_versions_completed_and_checked
    rule_db_versions.where(completed: true, checked: true)
  end

  private

  def default_values
    self.search = ["rules", "preproc_rules", "so_rules", "."] if self.search.blank?
  end

end
