#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class ApplyUpdateJob < Struct.new(:verbose, :sensor_ids, :user_id, :locale)
      include Redborder::Jobs::JobHelper

      def perform
        I18n.locale = locale
        sensors = Sensor.unscoped
                  .includes(:policy)
                  .where(id: sensor_ids,
                         type: [Sensor::IPS_BINDING, Sensor::IPS_GROUP, Sensor::IPS_GENERIC],
                         policies: { updating: false })

        if sensors.blank?
          Sensor.where(id: sensor_ids).update_all(applying: false)
          content = "#{I18n.t 'jobs.no_sensors'}"
          Notification.create(user_id: user_id,
                              comment: "#{I18n.t 'jobs.sensors'} #{Sensor.where(id: sensor_ids).map(&:full_name).join(', ')} #{I18n.t 'jobs.no_signatures'}")
          StoredDelayedJob.create(
            job: 'ApplyUpdateJob', user_id: user_id, content: content,
            options: {
              ids: sensor_ids,
              humanize: Sensor.where(id: sensor_ids).map(&:full_name).join(', ')
            }
          )
          return false
        end

        if sensors.first.type == Sensor::IPS_GENERIC
          apply_ips_generic(sensor_ids)
        elsif sensors.first.real_sensor_parent.type == Sensor::IPS
          apply_ips_v3(sensor_ids)
        elsif sensors.first.real_sensor_parent.type == Sensor::IPS_PROXY
          apply_ips_proxy(sensor_ids)
        end
      end

      def apply_ips_v3(sensor_ids)
        content = ''
        sensors_saved = []
        sensors_error = []

        sensors = Sensor.unscoped
                  .where(id: sensor_ids,
                         type: [Sensor::IPS_BINDING, Sensor::IPS_GROUP])

        sensors.group_by(&:group_id).each do |group_id, sensors_by_group|
          group_sensor = sensors_by_group.find(&:is_sensor_group?)
          binding_sensors = sensors_by_group.select(&:is_sensor_binding?)

          binding_sensors.each do |binding_sensor|
            # Resolve flowbits
            binding_sensor.policy.resolve_flowbits if (binding_sensor.policy.auto_resolve_flowbits && binding_sensor.policy.unresolved_flowbits.present?)
          end

          # Create the command to be executed
          cmd = "sudo -E /bin/env BOOTUP=none /usr/lib/redborder/bin/rb_get_sensor_rules.sh -s -g #{group_id}"
          cmd += ' -r' if group_sensor

          binding_sensors.each do |binding_sensor|
            rule_db_version_ids = binding_sensor.policy.rule_db_versions.map(&:id).join(',')
            cmd += " -b \"#{binding_sensor.binding_id}:#{rule_db_version_ids or 1}\""
          end

          # Get the sensor group
          sensor = (group_sensor || binding_sensors.first.real_parent)

          content += "############################################################################\n"
          content += "   #{sensor.complete_name}\n"
          content += "############################################################################\n\n"

          begin
            sensor_output = sensor.execute_output(cmd)

            if sensor_output.include? 'ERROR: '
              sensors_error.concat sensors_by_group
            else
              sensors_by_group.each do |s|
                s.update(
                  need_apply: false,
                  need_reputation_apply: false,
                  applied_at: Time.now
                )
              end
              sensors_saved.concat sensors_by_group
            end

            content += sensor_output
            content += "\n"

          rescue Timeout::Error
            content += "#{I18n.t 'jobs.timeout'}.\n\n"
            Notification.create(user_id: user_id,
                                comment: "#{I18n.t 'jobs.timeout'}: #{sensor.full_name}\n")
          end
        end

        if sensors_saved.present?
          Notification.create(user_id: user_id,
                              comment: "#{I18n.t 'jobs.rules'}: #{sensors_saved.map{|s| s.full_name + " (" + (s.is_sensor_group? ? s.reputation_policy.name : s.policy.name) + ")"}.join(", ")}\n")
        end

        if sensors_error.present?
          Notification.create(user_id: user_id,
                              comment: "#{I18n.t 'jobs.error_rules'}: #{sensors_error.map{|s| s.full_name + " (" + (s.is_sensor_group? ? s.reputation_policy.name : s.policy.name) + ")"}.join(", ")}\n")
        end
      rescue StandardError => e
        content += "#{e.class}: #{e.message}\n"
        e.backtrace.each do |line|
          content += "#{line}\n"
        end
        logit "#{e.class}: #{e.message}" if verbose
      ensure
        sensors.each { |s| s.update(applying: false) }

        if sensors_error.present? || sensors_saved.present?
          stored_job_humanize = ''
          stored_job_humanize += 'ERROR! ' if sensors_error.present?
          stored_job_humanize += sensors_saved.map(&:full_name).join(', ')

          StoredDelayedJob.create(
            job: 'ApplyUpdateJob',
            user_id: user_id,
            content: content,
            options: { ids: sensors_saved.map(&:id), humanize: stored_job_humanize }
          )
        end
      end

      def apply_ips_proxy(sensor_ids)
        sensors_saved = []

        sensors = Sensor.unscoped
                  .where(id: sensor_ids,
                         type: [Sensor::IPS_BINDING, Sensor::IPS_GROUP])

        sensors.group_by(&:group_id).each do |group_id, sensors_by_group|
          group_sensor = sensors_by_group.find(&:is_sensor_group?)
          binding_sensors = sensors_by_group.select(&:is_sensor_binding?)

          binding_sensors.each do |binding_sensor|
            # Resolve flowbits
            binding_sensor.policy.resolve_flowbits if (binding_sensor.policy.auto_resolve_flowbits && binding_sensor.policy.unresolved_flowbits.present?)
          end

          # Create the command to be executed
          cmd = "sudo /bin/env BOOTUP=none /usr/lib/redborder/bin/rb_get_sensor_rules.sh -s -g #{group_id}"
          cmd += ' -r' if group_sensor

          binding_sensors.each do |binding_sensor|
            rule_db_version_ids = binding_sensor.policy.rule_db_versions.map(&:id).join(',')
            cmd += " -b \"#{binding_sensor.binding_id}:#{rule_db_version_ids or 1}\""
          end

          # Get the sensor group
          sensor = (group_sensor || binding_sensors.first.real_parent)

          sensors_saved.concat sensors_by_group
          uuid = SecureRandom.uuid

          StoredDelayedJob.create(
            job: 'ApplyUpdateJob',
            user_id: user_id,
            content: nil,
            options: {
              ids: sensors_saved.map(&:id),
              uuid: uuid
            }
          )

          role = sensor.real_sensor_parent.chef_role
          role.override_attributes['redborder']['ipsrules'] ||= {}
          role.override_attributes['redborder']['ipsrules'][sensor.group_id] = {
            'timestamp' => Time.now.to_i,
            'command' => cmd,
            'uuid' => uuid
          }
          role.save
        end
      rescue StandardError => e
        content += "#{e.class}: #{e.message}\n"
        e.backtrace.each do |line|
          content += "#{line}\n"
        end
        logit "#{e.class}: #{e.message}" if verbose
      ensure
        sensors.each { |s| s.update(applying: false) }
      end

      def apply_ips_generic(sensor_ids)
        uuid = SecureRandom.uuid

        StoredDelayedJob.create(
          job: 'ApplyUpdateJob',
          user_id: user_id,
          content: nil,
          options: {
            ids: sensor_ids,
            uuid: uuid
          }
        )

        sensor = Sensor.find_by(id: sensor_ids)

        role = sensor.chef_role
        role.override_attributes['redborder']['ipsgeneric'] ||= {}
        role.override_attributes['redborder']['ipsgeneric']['rules'] ||= {}
        role.override_attributes['redborder']['ipsgeneric']['rules']['need_apply'] = true
        role.override_attributes['redborder']['ipsgeneric']['rules']['uuid'] = uuid
        role.save

        sensor.update(applying: false)
      end
    end
  end
end
