#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class ApplyUpdateJobV2 < Struct.new(:verbose, :sensors_v2_list, :user_id, :locale)
      include Redborder::Jobs::JobHelper

      def perform
        I18n.locale = locale
        content = ''
        user = User.find(user_id)
        sensor_v2_ids = sensors_v2_list.values.flatten.uniq
        sensors_saved = []
        sensors_error = []

        # Check sensors than need to apply policies
        sensors_v2 = Sensor.unscoped
                     .includes(:policy)
                     .where(id: sensor_v2_ids,
                         type: [Sensor::IPS_V2],
                         policies: { updating: false })
                     .select { |s| user.can? :manage, s }

        # If there are no sensors with policies to be applied
        if sensors_v2.blank?
          Sensor.where(id: sensor_v2_ids).update_all(applying: false)
          content += "#{I18n.t 'jobs.policies'}"
          Notification.create(user_id: user_id,
                              comment: "#{I18n.t 'jobs.sensors'} #{Sensor.where(id: sensor_v2_ids).map(&:full_name).join(', ')} #{I18n.t 'jobs.no_policies'}")
          StoredDelayedJob.create(
            job: 'ApplyUpdateJobV2', user_id: user_id, content: content,
            options: { ids: sensor_v2_ids, humanize: Sensor.where(id: sensor_v2_ids).map(&:full_name).join(', ') }
          )
          return false
        end

        begin
          # Iterate over the sensors v2 that need signatures and/or reputation
          # policies to be applied.
          sensors_v2.each do |sensor|
            update_hash = {}

            # Create the partial command to be executed
            cmd = 'sudo /bin/env BOOTUP=none /usr/lib/redborder/bin/rb_get_sensor_rules.sh'

            # If sensor v2 has signature policy to be applied
            if sensors_v2_list[:signature] && sensors_v2_list[:signature].include?(sensor.id.to_s)
              # Resolve flowbits
              sensor.policy.resolve_flowbits if (sensor.policy.auto_resolve_flowbits && sensor.policy.unresolved_flowbits.present?)
              # Signature policy flag
              cmd += ' -s'
              # Rule Versions flag
              rule_db_version_ids = sensor.policy.rule_db_versions.map(&:id).join(',')
              cmd += " -d \"#{rule_db_version_ids || 1}\""
              update_hash[:need_apply] = false
            end

            # If sensor v2 has reputation policy to be applied
            if sensors_v2_list[:reputation] && sensors_v2_list[:reputation].include?(sensor.id.to_s)
              cmd += ' -r'
              update_hash[:need_reputation_apply] = false
            end

            content += "############################################################################\n"
            content += "   #{sensor.complete_name}\n"
            content += "############################################################################\n\n"

            begin
              sensor_output = sensor.execute_output(cmd)

              if sensor_output.include? 'ERROR: '
                sensors_error << sensor
              else
                update_hash[:applied_at] = Time.now
                sensor.update(update_hash)
                sensors_saved << sensor
              end

              content += sensor_output
              content += "\n"

            rescue Timeout::Error
              content += "#{I18n.t 'jobs.timeout'}.\n\n"
              Notification.create(
                user_id: user_id,
                comment: "#{I18n.t 'jobs.timeout'}: #{sensor.full_name}\n"
              )
            end
          end
          # END sensor iteration

          if sensors_saved.present?
            Notification.create(
              user_id: user_id,
              comment: "#{I18n.t 'jobs.rules'} \
                        #{sensors_saved.map { |s| s.full_name }.join(', ')}\n"
            )
          end

          if sensors_error.present?
            Notification.create(
              user_id: user_id,
              comment: "#{I18n.t 'jobs.error_rules'} \
                        #{sensors_error.map { |s| s.full_name }.join(', ')}\n"
            )
          end

        rescue StandardError => e
          content += "#{e.class}: #{e.message}\n"
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          logit "#{e.class}: #{e.message}" if verbose
        ensure
          sensors_v2.each { |s| s.update(applying: false) }

          stored_job_humanize = ''
          stored_job_humanize += 'ERROR! ' if sensors_error.present?
          stored_job_humanize += sensors_saved.map(&:full_name).join(', ')

          StoredDelayedJob.create(
            job: 'ApplyUpdateJobV2',
            user_id: user_id,
            content: content,
            options: { ids: sensors_saved.map(&:id), humanize: stored_job_humanize }
          )
        end

      end

    end

  end

end
