#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SensorIPS

    IPS = 2
    IPS_GROUP = 3
    IPS_BINDING = 4

    # IPS v2
    IPS_V2 = 21

    IPS_PROXY = 32

    MODES = {
      "Inherited" => "",
      "IDS forwarding"       => "IDS_FWD",
      "IDS span"             => "IDS_SPAN",
      "IPS test mode"        => "IPS_TEST",
      "IPS"                  => "IPS"
    }

    CLUSTERMODES = {
      "2-tuple flow" => ['2', 'source ip - destination ip'],
      "4-tuple flow" => ['4', 'source ip: port - destination ip: port'],
      "5-tuple flow" => ['5', 'source ip: port - destination ip: port, protocol'],
      "6-tuple flow" => ['6', 'source ip: port - destination ip: port, protocol, vlan']
    }

    SNORT_VERSIONS = ['2.9.4', '2.9.6.1', '2.9.6.2', '2.9.7.0', '2.9.7.2', '2.9.7.6',
                      '2.9.8.0', '2.9.8.3']

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        self.available_sensors += [IPS, IPS_GROUP, IPS_BINDING, IPS_V2, IPS_PROXY]
        self.non_audited_columns += %w(need_reputation_apply need_apply)

        validate :uniqueness_name_for_groups_and_bindings,
                 if: proc { |sensor| [IPS_GROUP, IPS_BINDING].include? sensor.type }

        DittoCode::Exec.if 'ENTERPRISE' do
        validate :empty_license_buckets_for_ips, if: proc { [IPS, IPS_V2, IPS_PROXY].include? type }
        end

        has_many :owned_policies, class_name: "Policy", foreign_key: "sensor_id", dependent: :destroy
        has_many :owned_reputation_policies, class_name: "ReputationPolicy", foreign_key: "sensor_id", dependent: :destroy
        has_many :policy_rules, through: :policy
        has_many :rule_db_versions, through: :policy
        has_many :rule_sources

        belongs_to :policy
        belongs_to :reputation_policy

        before_create :default_values_for_group, if: proc { |sensor| sensor.type == IPS_GROUP }
        before_create :default_values_for_binding, if: proc { |sensor| sensor.type == IPS_BINDING }
        before_update :updated_policy_id, if: proc { |sensor| [IPS_BINDING, IPS_V2].include?(sensor.type) && (sensor.policy_id_changed? || sensor.reputation_policy_id_changed?) }
        before_update :updated_parent_id, if: (:parent_id_changed?)

        before_destroy :destroy_group, if: proc { |sensor| sensor.type == IPS_GROUP }
        before_destroy :destroy_binding, if: proc { |sensor| sensor.type == IPS_BINDING }
        before_destroy :move_sensors, if: proc { |sensor| sensor.type == Sensor::DOMAIN }
        before_destroy :restart_policies
        before_destroy :disassociate_sensor, if: proc { |sensor| [IPS, IPS_V2].include? sensor.type }

        after_create :recalculate_policies, if: proc { |sensor| [IPS_GROUP, IPS_BINDING].include? sensor.type }
        after_create :add_default_monitors, if: proc { |sensor| [IPS, IPS_GROUP, IPS_V2, IPS_PROXY, Sensor::IPS_GENERIC].include? sensor.type }
        after_create :create_cache_ips, if: proc { |sensor| [IPS, IPS_V2, IPS_PROXY, Sensor::IPS_GENERIC].include?(sensor.type) }

        after_commit :monitors_from_children, on: :create, if: proc { |sensor| [IPS, IPS_PROXY].include?(sensor.type) }
        after_commit :create_sensor_group, on: :create, if: proc { |sensor| [IPS, IPS_PROXY].include?(sensor.type) }
        after_commit :create_sensor_binding, on: :create, if: proc { |sensor| sensor.type == IPS_GROUP }

        after_update :update_name_in_role, if: proc { |sensor| [IPS_GROUP, IPS_BINDING].include? sensor.type }

        alias_method_chain :destroy_sensor, :ips
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{IPS_BINDING}") do
        'sensors/ips_binding_sensor'
      end

      define_method("to_partial_path_#{IPS_GROUP}") do
        'sensors/ips_group_sensor'
      end

      define_method("to_partial_path_#{IPS}") do
        'sensors/ips_sensor'
      end

      define_method("to_partial_path_#{IPS_PROXY}") do
        'sensors/ips_sensor_proxy'
      end

      define_method("to_partial_path_#{IPS_V2}") do
        'sensors/ips_sensor_v2'
      end
    end

    module InstanceMethods
      def to_partial_path
        if type == DOMAIN
          "sensors/domain_sensor"
        else
          Sensor.send("to_partial_path_#{type}")
        end
      end

      def is_sensor?
        type == IPS
      end

      def is_sensor_proxy?
        type == IPS_PROXY
      end

      def is_sensor_group?
        type == IPS_GROUP
      end

      def is_sensor_binding?
        type == IPS_BINDING
      end

      def is_sensor_v2?
        type == IPS_V2
      end

      define_method("icon_#{IPS_BINDING}") do
        'icons/binding.svg'
      end

      define_method("icon_#{IPS_GROUP}") do
        'icons/group.svg'
      end

      define_method("icon_#{IPS}") do
        'icons/sensor.svg'
      end

      define_method("icon_#{IPS_PROXY}") do
        'icons/sensor_proxy.svg'
      end

      define_method("icon_#{IPS_V2}") do
        'icons/sensor_v2.svg'
      end

      define_method("full_name_#{IPS_BINDING}") do
        "#{real_sensor_parent.name}/#{real_parent.name}/#{name}"
      end

      define_method("full_name_#{IPS_GROUP}") do
        "#{real_parent.name}/#{name}"
      end

      define_method("full_name_#{IPS}") do
        name
      end

      define_method("full_name_#{IPS_PROXY}") do
        name
      end

      define_method("full_name_#{IPS_V2}") do
        name
      end

      define_method("secondary_name_#{IPS_BINDING}") do
        "#{real_sensor_parent.name} - #{real_parent.name} - #{name}"
      end

      define_method("secondary_name_#{IPS_GROUP}") do
        "#{real_sensor_parent.name} - #{name}"
      end

      define_method("secondary_name_#{IPS}") do
        name
      end

      define_method("secondary_name_#{IPS_PROXY}") do
        name
      end

      define_method("secondary_name_#{IPS_V2}") do
        name
      end

      define_method("complete_name_#{IPS_BINDING}") do
        "Sensor: #{real_sensor_parent.name} - Group: #{real_parent.name} - Binding: #{name}"
      end

      define_method("complete_name_#{IPS_GROUP}") do
        "Sensor: #{real_parent.name} - Group: #{name}"
      end

      define_method("complete_name_#{IPS}") do
        name
      end

      define_method("complete_name_#{IPS_PROXY}") do
        name
      end

      define_method("complete_name_#{IPS_V2}") do
        name
      end

      define_method("type_to_s_#{IPS_BINDING}") do
        'sensor_binding'
      end

      define_method("type_to_s_#{IPS_GROUP}") do
        'sensor_group'
      end

      define_method("type_to_s_#{IPS}") do
        'sensor'
      end

      define_method("type_to_s_#{IPS_PROXY}") do
        'sensor_proxy'
      end

      define_method("type_to_s_#{IPS_V2}") do
        'sensor_v2'
      end

      define_method("type_humanize_#{IPS_BINDING}") do
        'binding'
      end

      define_method("type_humanize_#{IPS_GROUP}") do
        'group'
      end

      define_method("type_humanize_#{IPS}") do
        'sensor'
      end

      define_method("type_humanize_#{IPS_PROXY}") do
        'sensor_proxy'
      end

      #
      # IPS is a physical sensor.
      #
      # == Returns:
      # true
      #
      define_method("physical_#{IPS}?") do
        true
      end

      #
      # IPS is a physical sensor.
      #
      # == Returns:
      # true
      #
      define_method("physical_#{IPS_PROXY}?") do
        true
      end

      #
      # IPS V2 is a physical sensor.
      #
      # == Returns:
      # true
      #
      define_method("physical_#{IPS_V2}?") do
        true
      end

      define_method("chef_node_name_#{IPS_BINDING}") do
        "rbips-#{real_sensor_parent.id}"
      end

      define_method("chef_node_name_#{IPS_GROUP}") do
        "rbips-#{real_sensor_parent.id}"
      end

      define_method("chef_node_name_#{IPS}") do
        "rbips-#{id}"
      end

      define_method("chef_node_name_#{IPS_PROXY}") do
        "rbipscp-#{id}"
      end

      define_method("type_humanize_#{IPS_V2}") do
        'sensor_v2'
      end

      define_method("license_type_#{IPS}") do
        'event'
      end

      define_method("chef_node_#{IPS_BINDING}") do
        Chef::Node.load("rbips-#{real_sensor_parent.id}") rescue nil
      end

      define_method("chef_node_#{IPS_GROUP}") do
        Chef::Node.load("rbips-#{real_sensor_parent.id}") rescue nil
      end

      define_method("chef_node_#{IPS}") do
        Chef::Node.load("rbips-#{id}") rescue nil
      end

      define_method("chef_node_#{IPS_PROXY}") do
        Chef::Node.load("rbipscp-#{id}") rescue nil
      end

      define_method("chef_node_#{IPS_V2}") do
        Chef::Node.load("rbipsv2-#{id}") rescue nil
      end

      define_method("chef_client_#{IPS}") do
        Chef::ApiClient.load("rbips-#{id}")
      end

      define_method("chef_client_#{IPS_PROXY}") do
        Chef::ApiClient.load("rbipscp-#{id}") rescue nil
      end

      define_method("chef_client_#{IPS_V2}") do
        Chef::ApiClient.load("rbipsv2-#{id}")
      end

      define_method("last_check_in_#{IPS_BINDING}") do
        real_sensor_parent.send("last_check_in_#{IPS}")
      end

      define_method("last_check_in_#{IPS_GROUP}") do
        real_sensor_parent.send("last_check_in_#{IPS}")
      end

      define_method("last_check_in_#{IPS}") do
        begin
          node = chef_node
          Time.at(node[:ohai_time])
        rescue
          return false
        end
      end

      define_method("last_check_in_#{IPS_PROXY}") do
        begin
          node = chef_node
          Time.at(node[:ohai_time])
        rescue
          return false
        end
      end

      define_method("last_check_in_#{IPS_V2}") do
        begin
          node = chef_node
          Time.at(node[:ohai_time])
        rescue
          return false
        end
      end

      define_method("segments_#{IPS}") do
        node = chef_node
        return unless node && node['redborder']['segments']
        node['redborder']['segments'].keys.sort
      end

      define_method("segments_#{IPS_V2}") do
        node = chef_node
        return unless node && node['redborder']['segments']
        node['redborder']['segments'].keys.sort
      end

      define_method("cpu_number_#{IPS}") do
        node = chef_node
        return unless node && node['cpu']
        node['cpu']['total']
      end

      define_method("segments_#{IPS_PROXY}") do
        node = chef_node
        return unless node && node['redborder']['segments']
        node['redborder']['segments'].keys.sort
      end

      define_method("cpu_number_#{IPS_PROXY}") do
        node = chef_node
        return 1 unless node && node['cpu']
        node['cpu']['total']
      end

      define_method("monitors_to_chef_#{IPS_GROUP}") do
        role = real_parent.chef_role
        role.override_attributes["redborder"]["snort"]["groups"][group_id.to_s]["monitors"] = monitor_categories_to_chef
        role.save
      end

      define_method("monitors_to_chef_#{IPS}") do
        role = chef_role
        role.override_attributes["redborder"]["monitors"] = monitor_categories_to_chef
        role.save
      end

      define_method("monitors_to_chef_#{IPS_PROXY}") do
        role = chef_role
        role.override_attributes["redborder"]["monitors"] = monitor_categories_to_chef
        role.save
      end

      define_method("monitors_to_chef_#{IPS_V2}") do
        role = chef_role
        role.override_attributes["redborder"]["monitors"] = monitor_categories_to_chef
        role.save
      end

      define_method("real_sensor_#{IPS_BINDING}") do
        self.real_sensor_parent
      end

      define_method("real_sensor_#{IPS_GROUP}") do
        self.real_sensor_parent
      end

      #
      # => It represents the sensor that responds to wake up chef call.
      #    In case of Group and Binding, is the real sensor parent who
      #    responds to wake up chef method.
      #

      define_method("wakeupable_#{IPS}") do
        self
      end

      define_method("wakeupable_#{IPS_PROXY}") do
        self
      end

      define_method("wakeupable_#{IPS_V2}") do
        self
      end

      define_method("wakeupable_#{IPS_GROUP}") do
        self.real_sensor_parent
      end

      define_method("wakeupable_#{IPS_BINDING}") do
        self.real_sensor_parent
      end

      def real_sensor_parent
        if type == IPS_V2
          self
        elsif type == IPS_GROUP
          real_parent
        elsif type == IPS_BINDING && real_parent
          # In some cases real_parent can be nil because parent is destroyed.
          real_parent.real_parent
        else
          nil
        end
      end

      def need_apply
        if is_sensor_binding? || is_sensor_v2?
          super
        else
          all_children
            .where(type: [Sensor::IPS_BINDING, Sensor::IPS_V2])
            .map(&:need_apply)
            .include?(true)
        end
      end

      def policies
        sensor_ids = [1, self.id] + ancestors.map(&:id)
        Policy.where(sensor_id: sensor_ids, is_deleted: false)
      end

      def reputation_policies
        sensor_ids = [1, self.id] + ancestors.map(&:id)
        ReputationPolicy.where(sensor_id: sensor_ids)
      end

      def filter_events_rules
        thresholds("limit") + thresholds("threshold") + thresholds("both")
      end

      def suppress_events_rules
        thresholds("suppress")
      end

      def all_events_rules
        thresholds(["limit", "suppress"])
      end

      def download_rules
        msg = ''
        if !is_sensor_binding? && !is_sensor_v2?
          all_children.select { |s| s.is_sensor_binding? || s.is_sensor_v2? }.each do |sensor|
            sensor.download_rules
          end
        elsif is_sensor_binding?
          rule_db_version_ids = policy.rule_db_versions.map(&:id).join(",")
          msg += "############################################################################\n"
          msg += "   #{self.complete_name}\n"
          msg += "############################################################################\n\n"
          msg += self.execute_output("sudo /bin/env BOOTUP=none /usr/lib/redborder/bin/rb_get_sensor_rules.sh -f -g #{group_id} -b #{binding_id} -d \"#{rule_db_version_ids or 1}\"")
          self.update(need_apply: false, applied_at: Time.now)
        elsif is_sensor_v2?
          rule_db_version_ids = policy.rule_db_versions.map(&:id).join(',')
          msg += "############################################################################\n"
          msg += "   #{self.complete_name}\n"
          msg += "############################################################################\n\n"
          msg += self.execute_output("sudo /usr/lib/redborder/bin/rb_get_sensor_rules.sh -d \"#{rule_db_version_ids or 1}\"")
          self.update(need_apply: false, applied_at: Time.now)
        end
        msg
      end

      def sensor_group_childs
        return nil unless is_sensor?
        Sensor.where(type: Sensor::IPS_GROUP, real_parent: self)
      end

      def sensor_binding_childs
        return nil unless is_sensor_group?
        Sensor.where(type: Sensor::IPS_BINDING, real_parent: self)
      end

      define_method("version_#{IPS}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      define_method("version_#{IPS_PROXY}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      define_method("version_#{IPS_V2}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      define_method("reduced_version_#{IPS}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      define_method("reduced_version_#{IPS_PROXY}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      define_method("reduced_version_#{IPS_V2}") do
        chef_node["redborder"]["rpms"]["ips"] if chef_node
      end

      def change_bypass_mode(segment, mode)
        self.execute("sudo -E /usr/lib/redborder/bin/rb_bypass.sh -b #{segment} -s #{mode} -r")

        count = 0
        if is_sensor_v2?
          status = mode == "on" ? true : false
          while chef_node[:redborder][:segments][segment][:bypass] != status and count < 50
            sleep(0.5)
            count += 1
          end
        else
          status = mode == "on" ? "bypass" : "on"
          while chef_node[:redborder][:segments][segment][:status] != status and count < 50
            sleep(0.5)
            count += 1
          end
        end

        register_bypass_changes segment, mode
      end

      #
      # => Upload a file to the remote_file path of all IPS sensor childs
      #
      def upload(local_file, remote_file)
        all_children(true).select(&:is_sensor?).each do |sensor|
          Net::SSH.start(
            sensor.ip,
            'redborder',
            auth_methods: ['publickey'], # auth method
            paranoid: false,
            timeout: 5,
            keys: ["#{Rails.root}/config/#{RedBorder::CONFIG[:rsa_filename]}"],
            config: false
          ) do |ssh|
            File.open(local_file, 'r+') do |f|
              ssh.scp.upload!(f.path, remote_file, verbose: true)
            end
          end
        end

        register_license_changes self, local_file, remote_file
      end

      def vars_to_csv_format(type)
        role = chef_role
        csv = "Variable;Values;Inherited\n"
        csv += role.override_attributes["redborder"]["snort"]["default"][type].to_a.map{|v| "#{v[0]};#{v[1]};false"}.join("\n")
      end

      def all_vars_to_csv_format(type, sensor=nil)
        sensor_variables  = sensor.chef_role.override_attributes["redborder"]["snort"]["default"][type]
        role = chef_role
        csv = "Variable;Values;Inherited\n"
        csv += role.override_attributes["redborder"]["snort"]["default"][type].to_a.map{|v| "#{v[0]};#{sensor_variables.keys.include?("#{v[0]}") ? sensor_variables[v[0]] : v[1]};#{!sensor_variables.keys.include?("#{v[0]}")}"}.join("\n")
      end

      def add_vars_from_file(content_file, type)
        file = File.open(content_file.tempfile)
        root_variables = Sensor.root.chef_role.override_attributes["redborder"]["snort"]["default"][type].keys
        role = chef_role
        updated_count = 0
        created_count = 0

        file.each_with_index do |line, index|
          next if index.zero? #Here its assuming that the first line contains the column names and doesn't process new lines (\n) if they exist or eliminate them. That why Settings_controller_ips.rb CleanBlankSpaceFile and checkHeader whas implemented

          columns = line.sub(/[\n]+|[\r]+/, '').split(';')
          variable, value, inherited = columns

          next if (!is_root? and !root_variables.include?(variable))

          if type == "ipvars"
            next if !(value =~ /^(!?(\$HOME_NET)|(any)|((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/)
          else
            next if !(value =~ /^((any)|(((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?))(\s*,\s*((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?)))*))$/)
          end

          if inherited == "false" or inherited.nil?
            if role.override_attributes["redborder"]["snort"]["default"][type][variable] and role.override_attributes["redborder"]["snort"]["default"][type][variable] != value
              updated_count += 1
            elsif role.override_attributes["redborder"]["snort"]["default"][type][variable].nil?
              created_count += 1
            end

            role.override_attributes["redborder"]["snort"]["default"][type][variable] = value
          end
        end

        file.close
        role.save

        [created_count, updated_count]
      end

      def replace_vars_from_file(content_file, type)
        file = File.open(content_file.tempfile)
        root_variables = Sensor.root.chef_role.override_attributes["redborder"]["snort"]["default"][type].keys
        role = chef_role

        created_count = 0

        if is_root?
          role.override_attributes["redborder"]["snort"]["default"][type].delete_if{|key, value| !Setting.const_get(type.upcase).include?(key)}
        else
          role.override_attributes["redborder"]["snort"]["default"][type] = {}
        end

        file.each_with_index do |line, index|
          next if index.zero?

          columns = line.sub(/[\n]+|[\r]+/, '').split(';')
          variable, value, inherited = columns

          next if (!is_root? and !root_variables.include?(variable))

          if type == "ipvars"
            next if !(value =~ /^(!?(\$HOME_NET)|(any)|((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/)
          else
            next if !(value =~ /^((any)|(((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?))(\s*,\s*((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?)))*))$/)
          end

          if inherited == "false" or inherited.nil?
            role.override_attributes["redborder"]["snort"]["default"][type][variable] = value
            created_count += 1
          end
        end

        file.close
        role.save

        reload_variables(type) if is_root?

        created_count
      end

      #
      # => Receive the snort mode from the group table and assign it to the role
      #
      def update_snort_mode_from_group(value)
        role = self.chef_role

        if value.blank?

          role.override_attributes["redborder"]["snort"]["default"].delete("mode")

        else

          role.override_attributes["redborder"]["snort"]["default"]["mode"] = value

        end

        # Save the role
        role.save

        # Update values for group sensors
        role_real_sensor_parent = self.real_sensor_parent.chef_role
        role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{self.group_id}"]["mode"] = self.role_value(["redborder", "snort", "default", "mode"])
        role_real_sensor_parent.save
      end

      def update_chef_role_IPS(values)
        if [Sensor::DOMAIN, Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_GENERIC].include?(type) && !is_root?

          if values["redborder"]["proxy"]["enabled"].blank?
            values["redborder"]["proxy"].delete("enabled")
            values["redborder"]["proxy"].delete("address") if values["redborder"]["proxy"]["address"].blank?
            values["redborder"]["proxy"].delete("port") if values["redborder"]["proxy"]["port"].blank?
            values["redborder"]["proxy"].delete("user") if values["redborder"]["proxy"]["user"].blank?
            values["redborder"]["proxy"].delete("password") if values["redborder"]["proxy"]["password"].blank?
          elsif values["redborder"]["proxy"]["enabled"] == "true"
            values["redborder"]["proxy"]["enabled"] = true
          else
            values["redborder"]["proxy"] = { "enabled" => false }
          end

          values["redborder"]["rsyslog"]["servers"] = values["redborder"]["rsyslog"]["servers"].split(",")
          values["redborder"]["rsyslog"].delete("protocol") if values["redborder"]["rsyslog"]["protocol"].blank?
          values["redborder"]["rsyslog"].delete("mode") if values["redborder"]["rsyslog"]["mode"].blank?
        elsif type == Sensor::IPS_V2
          values['redborder']['rsyslog']['servers'] = values['redborder']['rsyslog']['servers'].split(',')
          values['redborder']['rsyslog'].delete('protocol') if values['redborder']['rsyslog']['protocol'].blank?
          values['redborder']['rsyslog'].delete('mode') if values['redborder']['rsyslog']['mode'].blank?
        end

        values["redborder"]["snort"]["default"]["ipvars"] = chef_role.override_attributes["redborder"]["snort"]["default"]["ipvars"]
        values["redborder"]["snort"]["default"]["portvars"] = chef_role.override_attributes["redborder"]["snort"]["default"]["portvars"]

        %w(mode pfring_watermark pfring_clustermode pfring_besteffort
           pfring_besteffort_minnumslots threshold_count threshold_seconds
           pfring_sbypass_profile).each do |key|
          values["redborder"]["snort"]["default"].delete(key) if values["redborder"]["snort"]["default"][key].blank?
        end

        if values["redborder"]["snort"]["default"]["autobypass"].blank?
          values["redborder"]["snort"]["default"].delete("autobypass")
        else
          values["redborder"]["snort"]["default"]["autobypass"] = (values["redborder"]["snort"]["default"]["autobypass"] == "true" ? true : false)
        end

        if is_sensor_v2?
          role = chef_role
          role.override_attributes['redborder']['snort']['save_pcap'] =
            values['redborder']['snort']['save_pcap']
          role.override_attributes['redborder']['snort']['default']['mode'] =
            values['redborder']['snort']['default']['mode']

          role.override_attributes['redborder']['rsyslog'] = values['redborder']['rsyslog']

          role.save
        elsif is_sensor_binding?
          role = chef_role
          role.override_attributes["redborder"]["snort"]["default"] = values["redborder"]["snort"]["default"]
          role.save

          role_real_sensor_parent = real_sensor_parent.chef_role
          role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]["bindings"]["#{binding_id}"]["threshold_count"] = role_value(["redborder", "snort", "default", "threshold_count"])
          role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]["bindings"]["#{binding_id}"]["threshold_seconds"] = role_value(["redborder", "snort", "default", "threshold_seconds"])
          role_real_sensor_parent.save
        elsif is_sensor_group?
          role = chef_role
          role.override_attributes["redborder"]["snort"]["default"] = values["redborder"]["snort"]["default"]
          role.save

          role_real_sensor_parent = real_sensor_parent.chef_role

          %w(mode pfring_watermark pfring_clustermode pfring_besteffort
             pfring_besteffort_minnumslots pfring_sbypass_profile).each do |key|
            role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"][key] = role_value(["redborder", "snort", "default", key])
          end
          role_real_sensor_parent.save
        elsif is_sensor? || is_sensor_proxy? || is_sensor_generic?
          role = chef_role
          role.override_attributes["redborder"]["snort"]["default"] = values["redborder"]["snort"]["default"]
          role.override_attributes["redborder"]["snort"]["groups"].keys.each do |group|
            role.override_attributes["redborder"]["snort"]["groups"][group]["segments"] = (values["redborder"]["snort"]["groups"][group].try(:[], "segments") or [])
            role.override_attributes["redborder"]["snort"]["groups"][group]["cpu_list"] = (values["redborder"]["snort"]["groups"][group].try(:[], "cpu_list") or [])
          end
          role.override_attributes["redborder"]["rsyslog"] = values["redborder"]["rsyslog"]
          role.override_attributes["redborder"]["proxy"] = values["redborder"]["proxy"]
          role.save
        else
          role = chef_role
          role.override_attributes["redborder"]["snort"]["default"] = values["redborder"]["snort"]["default"]
          role.override_attributes["redborder"]["rsyslog"] = values["redborder"]["rsyslog"]
          role.override_attributes["redborder"]["proxy"] = values["redborder"]["proxy"]
          role.save
        end

        # Update values for group sensors
        all_children.select{|s| s.is_sensor_group?}.each do |sensor|
          role_real_sensor_parent = sensor.real_sensor_parent.chef_role

          %w(mode autobypass pfring_watermark pfring_clustermode
             pfring_besteffort pfring_besteffort_minnumslots
             pfring_sbypass_profile).each do |key|
            role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{sensor.group_id}"][key] = sensor.role_value(["redborder", "snort", "default", key])
          end
          role_real_sensor_parent.save
        end

        # Update values for binding sensors
        all_children.select{|s| s.is_sensor_binding?}.each do |sensor|
          role_real_sensor_parent = sensor.real_sensor_parent.chef_role
          role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{sensor.group_id}"]["bindings"]["#{sensor.binding_id}"]["threshold_count"] = sensor.role_value(["redborder", "snort", "default", "threshold_count"])
          role_real_sensor_parent.override_attributes["redborder"]["snort"]["groups"]["#{sensor.group_id}"]["bindings"]["#{sensor.binding_id}"]["threshold_seconds"] = sensor.role_value(["redborder", "snort", "default", "threshold_seconds"])
          role_real_sensor_parent.save
        end
      end

      def restart_chef
        result = ""
        case type
        when Sensor::DOMAIN
          sensors = all_children.collect do |s|
            case s.type
            when IPS_GROUP, IPS_BINDING
              s.real_sensor_parent
            when IPS, IPS_V2, IPS_PROXY, Sensor::IPS_GENERIC
              s
            end
          end
          sensors.compact.uniq.each{|s| s.restart_chef}
        else
          result += execute_output("sudo -E /sbin/service chef-client restart; sleep 5")
        end
        result
      rescue StandardError
        Rails.logger.info $!.inspect
        Rails.logger.info $@.join("\n")
      end

      # TODO: IPS_V2 ?
      def update_rpms
        result = ""
        case type
        when Sensor::DOMAIN
          sensors = all_children.collect do |s|
            case s.type
            when IPS_GROUP, IPS_BINDING
              s.real_sensor_parent
            when IPS, IPS_PROXY, Sensor::IPS_GENERIC
              s
            end
          end
          sensors.compact.uniq.each{|s| s.update_rpms}
        else
          result += execute_output("sudo /opt/rb/bin/rb_update_redborder_rpms.sh -f")
        end
        result
      rescue StandardError
        Rails.logger.info $!.inspect
        Rails.logger.info $@.join("\n")
      end

      # TODO: IPS_V2 ?
      def view_raw_info params
        command = "sudo -E /usr/lib/redborder/bin/rb_u2pcap.sh -p"
        command += " -n #{params[:group_name]} " if params[:group_name]
        command += " -g #{params[:sig_generator]} " if params[:sig_generator]
        command += " -s #{params[:sig_id]} " if params[:sig_id]
        command += " -l #{params[:timestamp].to_datetime.to_i} -u #{params[:timestamp].to_datetime.end_of_minute.to_i} " if params[:timestamp]
        command += " -o #{params[:src]} " if params[:src]
        command += " -d #{params[:dst]} " if params[:dst]
        self.execute_output(command)
      rescue Net::SSH::AuthenticationFailed => e
        Rails.logger.error "Sensor: #{self.name} #{e.inspect}"
        Rails.logger.error e.backtrace.join("\n")
        result = "#{I18n.t 'sensors_ips.error'}"
      rescue StandardError
        Rails.logger.info $!.inspect
        Rails.logger.info $@.join("\n")
      end

      def audit_comment_on_update_with_ips(audit)
        object = audit.auditable_type.constantize.where(id: audit.auditable_id)
        name = object.first.try(:name)

        return if audit.audited_changes.nil?

        comment = "#{I18n.t 'sensors_ips.updated'} #{name}:"
        comment += audit.audited_changes.collect {|key, values|
          case key.to_sym
          when :reputation_policy_id
            " #{I18n.t 'sensors_ips.reputation'} '#{self.reputation_policy.try(:name)}'"
          when :parent_id
            " #{I18n.t 'sensors_ips.parent'} '#{self.parent.try(:name)}'"
          else
            " #{key} #{I18n.t 'sensors_ips.from'} '#{values[0]}' #{I18n.t 'sensors_ips.to'} '#{values[1]}'"
          end
        }.join(", ")
        comment
      end

      [IPS, IPS_GROUP, IPS_BINDING, IPS_V2, IPS_PROXY].each do |t|
        define_method("as_json_with_children_#{t}") do
          # Return nothing for IPS sensors
          return
        end
      end

      #
      # => Method used to complete values about ancestors on chef role
      #    For IPS, it is only useful on groups
      #
      define_method("update_chef_domain_values_#{IPS}") do
        role = chef_role
        return unless role
        # For each element in the hierarchy, try to get the values.
        (Sensor::HIERARCHY - [Sensor::FLOOR]).each do |domain_type|
          s = ancestors.find_by(domain_type: domain_type)
          next unless s
          dname = s.domain_type_name(to_key: true)
          children.each do |group|
            role.override_attributes['redborder']['snort']['groups'][group.group_id.to_s][dname] =
              s.try(:name)
            role.override_attributes['redborder']['snort']['groups'][group.group_id.to_s]["#{dname}_uuid"] =
              s.try(:uuid)
          end
        end
        # Save
        role.save
      end

      #
      # => Method used to complete values about ancestors on chef role
      #    For IPS_PROXY, it is only useful on groups
      #
      define_method("update_chef_domain_values_#{IPS_PROXY}") do
        role = chef_role
        return unless role
        # For each element in the hierarchy, try to get the values.
        (Sensor::HIERARCHY - [Sensor::FLOOR]).each do |domain_type|
          s = ancestors.find_by(domain_type: domain_type)
          next unless s
          dname = s.domain_type_name(to_key: true)
          children.each do |group|
            role.override_attributes['redborder']['snort']['groups'][group.group_id.to_s][dname] =
              s.try(:name)
            role.override_attributes['redborder']['snort']['groups'][group.group_id.to_s]["#{dname}_uuid"] =
              s.try(:uuid)
          end
        end
        # Save
        role.save
      end

      define_method("update_chef_domain_values_#{IPS_GROUP}") do
        real_parent_role = real_parent.chef_role
        return unless real_parent_role
        # For each element in the hierarchy, try to get the values.
        (Sensor::HIERARCHY - [Sensor::FLOOR]).each do |domain_type|
          s = ancestors.find_by(domain_type: domain_type)
          next unless s
          dname = s.domain_type_name(to_key: true)
          real_parent_role.override_attributes['redborder']['snort']['groups'][group_id.to_s][dname] =
            s.try(:name)
          real_parent_role.override_attributes['redborder']['snort']['groups'][group_id.to_s]["#{dname}_uuid"] =
            s.try(:uuid)
        end
        # Save
        real_parent_role.save
      end

      define_method("can_have_children_#{IPS}?") do
        true
      end

      define_method("can_have_children_#{IPS_PROXY}?") do
        true
      end

      define_method("can_have_children_#{IPS_GROUP}?") do
        true
      end

      define_method("can_have_children_#{IPS_BINDING}?") do
        false
      end

      define_method("update_chef_domain_values_#{IPS_V2}") do
        role = chef_role
        return unless role
        # For each element in the hierarchy, try to get the values.
        (Sensor::HIERARCHY - [Sensor::FLOOR]).each do |domain_type|
          s = ancestors.find_by(domain_type: domain_type)
          next unless s
          dname = s.domain_type_name(to_key: true)
          role.override_attributes['redborder']['snort'][dname] = s.try(:name)
          role.override_attributes['redborder']['snort']["#{dname}_uuid"] = s.try(:uuid)
        end
        # Save
        role.save
      end

      def assign_policy_to_children(policy)
        return unless self_and_ancestors_with_root.include?(policy.owner)

        self_and_descendants.each do |s|
          s.assign_policy(policy.id)
        end
      end

      def assign_reputation_policy_to_children(policy)
        return unless self_and_ancestors_with_root.include?(policy.owner)

        self_and_descendants.each do |s|
          s.assign_reputation_policy(policy.id)
        end
      end

      DittoCode::Exec.if 'ENTERPRISE' do
        def can_assign_policy?
          if is_sensor?
            license.try(:valid?)
          elsif [Sensor::IPS_GROUP, Sensor::IPS_BINDING].include?(type)
            real_sensor_parent.license.try(:valid?)
          else
            true
          end
        end
      end

      DittoCode::Exec.if 'COMMUNITY' do
        def can_assign_policy?
          true
        end
      end

      private
        #
        # => Process fired when create an IPS_GROUP or IPS_BINDING.
        #    groups and bindings need to include its policies
        #    and update need_apply and need_reputation_policy_id
        #
        def recalculate_policies
          parent_policy_id = parent.policy_id
          parent_reputation_policy_id = parent.reputation_policy_id

          all_children(true).each do |sensor|
            if (sensor.policy_id.nil? or (sensor.policy.sensor_id != sensor.id and !sensor.ancestors.include?(sensor.policy.owner)))
              if sensor.is_sensor_binding? and sensor.policy_id != parent_policy_id
                sensor.update_column(:need_apply, true)
                sensor.update_column(:policy_id, parent_policy_id)
              elsif sensor.policy_id != parent_policy_id
                sensor.update_column(:policy_id, parent_policy_id)
              end
            end
            if (sensor.reputation_policy_id.nil? or (sensor.reputation_policy.sensor_id != sensor.id and !sensor.ancestors.include?(sensor.reputation_policy.owner)))
              if sensor.is_sensor_group? and sensor.reputation_policy_id != parent_reputation_policy_id
                sensor.update_column(:need_reputation_apply, true)
                sensor.update_column(:reputation_policy_id, parent_reputation_policy_id)
              elsif sensor.reputation_policy_id != parent_reputation_policy_id
                sensor.update_column(:reputation_policy_id, parent_reputation_policy_id)
              end
            end
          end
        end

        def register_assign_policy policy_id
          audit = self.audits.create!(
            auditable_id: self.id,
            auditable_type: self.class.name,
            action: 'update',
            comment: "Policy with id #{policy_id} assign to #{name} with id #{id}",
            user_id: User.current_user.id
          )
          notify_audit_to_syslog(audit)
        end

        def register_bypass_changes segment, mode
          audit = self.audits.create!(
            auditable_id: self.id,
            auditable_type: self.class.name,
            action: 'update',
            comment: "#{I18n.t 'sensors_ips.changed'} #{segment}, #{I18n.t 'sensors_ips.to'} #{mode}",
            user_id: User.current_user.id
          )
          notify_audit_to_syslog(audit)
        end

        def create_sensor_group
          sensor_group =
            Sensor.create(name: 'default',
                          parent_id: id,
                          type: IPS_GROUP,
                          real_parent_id: id)
          role = chef_role
          role.override_attributes['redborder']['snort']['groups']["#{sensor_group.group_id}"]['group_uuid'] = sensor_group.uuid
          role.save
        end

        def create_sensor_binding
          sensor_binding = Sensor.create(name: 'default',
                                         parent_id: id,
                                         type: IPS_BINDING,
                                         real_parent_id: id,
                                         group_id: group_id)
          # Update real_sensor_parent role adding group_id
          role = real_sensor_parent.chef_role
          role.override_attributes['redborder']['snort']['groups']["#{group_id}"]['group_uuid'] = uuid
          role.save
        end

        def default_values_for_group
          group_ids = Sensor.where(type: Sensor::IPS_GROUP, real_parent_id: self.real_parent_id).map(&:group_id)

          if group_ids.present?
            self.group_id = group_ids.max + 1
            (0..group_ids.max).each{|x| (self.group_id = x and break) if !group_ids.include?(x)}
          else
            self.group_id = 0
          end

          ipvars = real_parent.role_value_for_vars("ipvars")
          portvars = real_parent.role_value_for_vars("portvars")
          mode = real_parent.role_value(["redborder", "snort", "default", "mode"])
          autobypass = real_parent.role_value(["redborder", "snort", "default", "autobypass"])

          role = real_parent.chef_role
          role.override_attributes["redborder"]["snort"]["groups"][self.group_id.to_s] = {"name" => "#{self.name}",
                                                                                          "mode" => mode,
                                                                                          "segments" => [],
                                                                                          "domain_id" => real_parent.first_domain.id,
                                                                                          "domain_name" => real_parent.first_domain.name,
                                                                                          "cpu_list" => [],
                                                                                          "autobypass" => autobypass,
                                                                                          "monitors" => [],
                                                                                          "bindings" => {"0" => {"name" => "default",
                                                                                                                 "ipvars" => ipvars,
                                                                                                                 "portvars" => portvars,
                                                                                                                 "save_pcap" => false,
                                                                                                                 "vlan_objects" => {},
                                                                                                                 "network_objects" => {}
                                                                                                                }
                                                                                                    }
                                                                                         } if role.override_attributes["redborder"]["snort"]["groups"][self.group_id.to_s].nil?

          role.save
        end

        def default_values_for_binding
          binding_ids = Sensor.where(type: Sensor::IPS_BINDING,
                                     real_parent_id: self.real_parent_id).map(&:binding_id)

          if binding_ids.present?
            self.binding_id = binding_ids.max + 1
            (0..binding_ids.max).each{|x| (self.binding_id = x and break) if !binding_ids.include?(x)}
          else
            self.binding_id = 0
          end

          ipvars = real_parent.role_value_for_vars('ipvars')
          portvars = real_parent.role_value_for_vars('portvars')
          role = real_sensor_parent.chef_role
          role.override_attributes['redborder']['snort']['groups'][self.group_id.to_s]['bindings'][self.binding_id.to_s] = {  'name' => self.name.to_s,
                                                                                                                              'ipvars' => ipvars,
                                                                                                                              'portvars' => portvars,
                                                                                                                              'save_pcap' => false,
                                                                                                                              'vlan_objects' => {},
                                                                                                                              'network_objects' => {} }

          role.save
        end

        def destroy_group
          role = real_sensor_parent.chef_role
          role.override_attributes["redborder"]["snort"]["groups"].delete("#{group_id}")
          role.save
        rescue NoMethodError
        end

        def destroy_binding
          ips_parent = real_sensor_parent
          return unless ips_parent
          role = ips_parent.chef_role
          role.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]["bindings"].delete("#{binding_id}") if role.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]
          role.save
        end

        def destroy_sensor_with_ips
          if is_domain?
            children.select{|sensor| sensor.is_sensor_group? or sensor.is_sensor_binding?}.each do |sensor|
              sensor.update(parent_id: sensor.real_parent_id)
            end
          end
          Sensor.where(policy_id: owned_policies.map(&:id)).update_all(policy_id: nil)

          destroy_sensor_without_ips
        end

        # TODO: it can be done with validates and scopes
        def uniqueness_name_for_groups_and_bindings
          if new_record?
            if is_sensor_group?
              errors.add(:name, "#{I18n.t 'sensors_ips.group'} #{name}") if Sensor.where(real_parent: real_parent, type: Sensor::IPS_GROUP, name: name).count > 0
            elsif is_sensor_binding?
              errors.add(:name, "#{I18n.t 'sensors_ips.binding'} #{name}") if Sensor.where(real_parent: real_parent, type: Sensor::IPS_BINDING, name: name).count > 0
            end
          else
            return true if parent_id_changed?

            if is_sensor_group? and name_changed?
              errors.add(:name, "#{I18n.t 'sensors_ips.group'} #{name}") if Sensor.where(real_parent: real_parent, type: Sensor::IPS_GROUP, name: name).count > 0
            elsif is_sensor_binding? and name_changed?
              errors.add(:name, "#{I18n.t 'sensors_ips.binding'} #{name}") if Sensor.where(real_parent: real_parent, type: Sensor::IPS_BINDING, name: name).count > 0
            end
          end
        end

        # Check if license can be associated to current sensor
        #
        def empty_license_buckets_for_ips
          return unless license
          return if license_id_was == license_id
          errors.add(:license_id, "#{I18n.t 'sensors_ips.full'}") if license.in_use_ips >= license.limit_ips
        end

        def create_cache_ips
          Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
        end

        def create_chef_node_ips
          node_json = {
            'name' => "rbips-#{id}",
            'run_list' => ['role[ips-sensor]', "role[#{chef_name}]"],
            'attributes' => {
              'rbname' => name,
              'ipaddress' => ip,
              'redborder' => {
                'segments' => {},
                'snort' => {},
                'monitors' => [],
                'sensor_id' => id
              }
            }
          }

          node = Chef::Node.from_hash(node_json)
          node.create
        end

        def create_chef_node_ips_v2
          node_json = {
            'name' => "rbipsv2-#{id}",
            'run_list' => ['role[ipsv2-sensor]', "role[#{chef_name}]"],
            'attributes' => {
              'rbname' => name,
              'ipaddress' => ip,
              'redborder' => {
                'segments' => {},
                'snort' => {
                  'segments' => []
                },
                'monitors' => [],
                'sensor_id' => id
              }
            }
          }

          node = Chef::Node.from_hash(node_json)
          node.create
        end

      protected
        #
        # => Assign the policy if the IPS sensor has a valid license or
        #    if it's a domain.
        #
        def assign_policy(policy_id)
          DittoCode::Exec.if 'ENTERPRISE' do
          return false if (is_sensor? && !license.try(:valid?)) ||
                          ([Sensor::IPS_GROUP, Sensor::IPS_BINDING].include?(type) && !real_sensor_parent.license.try(:valid?))
          end

          was_updated = update(policy_id: policy_id, need_apply: true)
          register_assign_policy(policy_id) if was_updated
          return was_updated
        end

        #
        # => Assign the reputation policy if the IPS sensor has a valid license or
        #    if it's a domain.
        #
        def assign_reputation_policy(policy_id)
          DittoCode::Exec.if 'ENTERPRISE' do
          return false if (is_sensor? && !license.try(:valid?)) ||
                          ([Sensor::IPS_GROUP, Sensor::IPS_BINDING].include?(type) && !real_sensor_parent.license.try(:valid?))
          end

          was_updated = update(reputation_policy_id: policy_id, need_reputation_apply: true)
          register_assign_policy(policy_id) if was_updated
          return was_updated
        end

        def thresholds(filter_type)
          result = event_filters.where(filter_type: filter_type).order("position ASC")
          conditions = result.map{|r| "not(rule_sid = #{r.rule_sid} AND rule_gid = #{r.rule_gid})"}.join(" AND ")

          ancestors.each do |sensor|
            if conditions.present?
              result += sensor.event_filters.where(filter_type: filter_type).where(conditions).order("position ASC")
            else
              result += sensor.event_filters.where(filter_type: filter_type).order("position ASC")
            end

            conditions = result.map{|r| "not(rule_sid = #{r.rule_sid} AND rule_gid = #{r.rule_gid})"}.join(" AND ")

          end

          result
        end

        def updated_policy_id
          self.need_apply = true
        end

        def updated_reputation_policy_id
          self.need_reputation_apply = true
        end

        def add_default_monitors
          monitor_categories << MonitorCategory.where(default: true)
            .select { |mc| mc.sensor_types.include? type }
          save
          monitors_to_chef
        end

        def monitors_from_children
          children.each(&:monitors_to_chef)
        end

        def reload_variables(type)
          root_variables = Sensor.root.chef_role.override_attributes["redborder"]["snort"]["default"][type].keys

          Sensor.where(type: [Sensor::DOMAIN, Sensor::IPS, Sensor::IPS_V2, Sensor::IPS_PROXY, Sensor::IPS_GENERIC]).where.not(id: 1).each do |sensor|
            role = sensor.chef_role
            role.override_attributes["redborder"]["snort"]["default"][type].delete_if{|key, _v| !root_variables.include?(key)}
            role.save
          end
        end

        def updated_parent_id
          if !is_domain?
            all_children(true).select(&:is_sensor_group?).each do |sensor|
              role = sensor.real_sensor_parent.chef_role
              role.override_attributes["redborder"]["snort"]["groups"]["#{sensor.group_id}"]["domain_uuid"] = self.first_domain.uuid
              role.override_attributes["redborder"]["snort"]["groups"]["#{sensor.group_id}"]["domain_name"] = self.first_domain.name
              role.save
            end
          end
        end

        def update_name_in_role
          role = real_sensor_parent.chef_role
          if is_sensor_group?
            role.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]["name"] = name
          elsif is_sensor_binding?
            role.override_attributes["redborder"]["snort"]["groups"]["#{group_id}"]["bindings"]["#{binding_id}"]["name"] = name
          end
          role.save
        end

        [IPS_BINDING, IPS_GROUP, IPS].each do |method|
          define_method "run_list_text_#{method}" do
            "role[ips-sensor]"
          end
        end

        define_method "run_list_text_#{IPS_V2}" do
          'role[ipsv2-sensor]'
        end

        define_method "run_list_text_#{IPS_PROXY}" do
          'role[ipscp-sensor]'
        end

        define_method("create_chef_role_#{IPS}") do
          # Create the chef node and client if them dont exists
          # since we are using them in this method
          create_chef_node_ips if chef_node.nil?

          run_list_roles = []
          run_list_roles << 'role[ips-sensor]' unless self.is_domain?
          run_list_roles << "role[#{parent.chef_name}]"

          role_json = { 'name' => chef_name, 'run_list' => run_list_roles }

          role_json['override_attributes'] =
            { 'redborder' => { 'role' => chef_name,
                               'sensor_uuid' => uuid,
                               'snort' => { 'default' => { 'ipvars' => {},
                                                           'portvars' => {},
                                                           'save_pcap' => nil,
                                                           'mode' => nil,
                                                           'autobypass' => nil } },
                               'rsyslog' => {},
                               'proxy' => {},
                               'snmp' => { 'trap_server' => [] },
                               'monitors' => [] } }

          ipvars = parent.role_value_for_vars('ipvars')
          portvars = parent.role_value_for_vars('portvars')
          mode = parent.role_value(%w(redborder snort default mode))
          autobypass = parent.role_value(%w(redborder snort autobypass))

          role_json['override_attributes']['redborder']['snort']['groups'] =
            { '0' => { 'name' => 'default',
                       'mode' => mode,
                       'domain_uuid' => Sensor.root.uuid,
                       'domain_name' => 'root',
                       # Assign, if exists, only the first segment.
                       # If there are more than 1, let them unassigned.
                       'segments' => (segments.nil? ? nil : segments.first(1)),
                       'cpu_list' => (cpus_number.nil? ? nil : cpus_number.times.collect(&:to_s)),
                       'autobypass' => autobypass,
                       'bindings' => { '0' => { 'name' => 'default',
                                                'ipvars' => ipvars,
                                                'portvars' => portvars,
                                                'save_pcap' => false,
                                                'vlan_objects' => {},
                                                'network_objects' => {} } } } }

          role = Chef::Role.from_hash(role_json)
          role.create
        end

        define_method("create_chef_role_#{IPS_PROXY}") do
          # Create the chef node and client if them dont exists
          # since we are using them in this method

          run_list_roles = []
          run_list_roles << 'role[ipscp-sensor]' unless self.is_domain?
          run_list_roles << "role[#{parent.chef_name}]"

          role_json = { 'name' => chef_name, 'run_list' => run_list_roles }

          role_json['override_attributes'] =
            { 'redborder' => { 'role' => chef_name,
                               'sensor_uuid' => uuid,
                               'snort' => { 'default' => { 'ipvars' => {},
                                                           'portvars' => {},
                                                           'save_pcap' => nil,
                                                           'mode' => nil,
                                                           'autobypass' => nil } },
                               'rsyslog' => {},
                               'proxy' => {},
                               'snmp' => { 'trap_server' => [] },
                               'monitors' => [] } }

          ipvars = parent.role_value_for_vars('ipvars')
          portvars = parent.role_value_for_vars('portvars')
          mode = parent.role_value(%w(redborder snort default mode))
          autobypass = parent.role_value(%w(redborder snort autobypass))

          role_json['override_attributes']['redborder']['snort']['groups'] =
            { '0' => { 'name' => 'default',
                       'mode' => mode,
                       'domain_uuid' => Sensor.root.uuid,
                       'domain_name' => 'root',
                       # Assign, if exists, only the first segment.
                       # If there are more than 1, let them unassigned.
                       'segments' => (segments.nil? ? nil : segments.first(1)),
                       'cpu_list' => (cpus_number.nil? ? nil : cpus_number.times.collect(&:to_s)),
                       'autobypass' => autobypass,
                       'monitors' => [],
                       'bindings' => { '0' => { 'name' => 'default',
                                                'ipvars' => ipvars,
                                                'portvars' => portvars,
                                                'save_pcap' => false,
                                                'vlan_objects' => {},
                                                'network_objects' => {} } } } }

          role = Chef::Role.from_hash(role_json)
          role.create
        end

        define_method("create_chef_role_#{IPS_V2}") do
          # Create the chef node and client if them dont exists
          # since we are using them in this method
          create_chef_node_ips_v2 if chef_node.nil?

          run_list_roles = []
          run_list_roles << 'role[ipsv2-sensor]' unless self.is_domain?
          run_list_roles << "role[#{parent.chef_name}]"

          role_json = { 'name' => chef_name, 'run_list' => run_list_roles }

          role_json['override_attributes'] =
            { 'redborder' => { 'role' => chef_name,
                               'sensor_uuid' => uuid,
                               'snort' => { 'default' => { 'ipvars' => {},
                                                           'portvars' => {},
                                                           'save_pcap' => nil,
                                                           'mode' => nil,
                                                           'autobypass' => nil },
                                            'segments' => (segments.nil? ? [] : segments),
                                            'initialized' => false
                                          },
                               'rsyslog' => {},
                               'proxy' => {},
                               'snmp' => { 'trap_server' => [] },
                               'monitors' => [] } }

          role = Chef::Role.from_hash(role_json)
          role.create
        end

        [IPS_BINDING, IPS_GROUP].each do |method|
          define_method "create_chef_role_#{method}" do
            run_list_roles = []
            run_list_roles << "role[ips-sensor]" unless self.is_domain?
            run_list_roles << "role[#{parent.chef_name}]"

            role_json = {"name" => self.chef_name, "run_list" => run_list_roles}

            role_json["override_attributes"] = {"redborder" => {"role" => self.chef_name,
                                                                  "snort" => {"default" => {"ipvars" => {},
                                                                                             "portvars" => {},
                                                                                             "save_pcap" => nil,
                                                                                             "mode" => nil,
                                                                                             "autobypass" => nil
                                                                                            }
                                                                              },
                                                                  "rsyslog" => {},
                                                                  "proxy" => {},
                                                                  "snmp" => {"trap_server" => []},
                                                                  "monitors" => []
                                                                 }
                                                  }

            role = Chef::Role.from_hash(role_json)
            role.create
          end
        end

        # When destroying a domain, move groups and bindings to her real_parent
        def move_sensors
          children.select{|sensor| sensor.is_sensor_group? or sensor.is_sensor_binding?}.each do |sensor|
            sensor.update(parent_id: sensor.real_parent_id)
          end
        end

        # Destroy references to owned policies
        def restart_policies
          Sensor.where(policy_id: owned_policies.map(&:id)).update_all(policy_id: nil)
        end

        def disassociate_sensor
          return unless (Sensor.where(type: [Sensor::IPS, Sensor::IPS_V2], ip: ip).size <= 1)
          rb_disassociate_sensor
        end

        def rb_disassociate_sensor
          self.execute('sudo -E /opt/rb/bin/rb_disassociate_sensor.sh -f')
        rescue Exception
          Rails.logger.info $!.inspect
          Rails.logger.info $@.join("\n")
        end

        define_method "destroy_chef_node_#{IPS}" do
          chef_node.destroy
        end

        define_method "destroy_chef_node_#{IPS_V2}" do
          chef_node.destroy
        end

        define_method "destroy_chef_node_#{IPS_PROXY}" do
          chef_node.destroy
        end
    end
  end
end

Sensor.send(:include, RedBorder::SensorIPS)
