#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SensorIPSGeneric

    IPS_GENERIC = 33

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        self.available_sensors += [IPS_GENERIC]

        self.import_columns_for[IPS_GENERIC] =
          [:rules_path, :filter_rules_path, :unified2_files, :unified2_name,
           :commands, :gen_file, :reference_file, :classification_file,
           :snort_stats_file].freeze

        has_chef_property :rules_path, %w(redborder ipsgeneric config rules_path)
        has_chef_property :filter_rules_path, %w(redborder ipsgeneric config filter_rules_path)
        has_chef_property :unified2_files, %w(redborder ipsgeneric config unified2_files)
        has_chef_property :unified2_name, %w(redborder ipsgeneric config unified2_name)
        has_chef_property :snort_stats_file, %w(redborder ipsgeneric config snort_stats_file)
        has_chef_property :commands, %w(redborder ipsgeneric config commands)
        has_chef_property :gen_file, %w(redborder ipsgeneric config gen_file)
        has_chef_property :reference_file, %w(redborder ipsgeneric config reference_file)
        has_chef_property :classification_file, %w(redborder ipsgeneric config classification_file)

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------
        validates :snmp_community, presence: false,
                                   format: { with: /\A[\/\w-]*\z/,
                                             message: "#{I18n.t 'sensors_controller_ips.only'}" },
                                   if: proc { type == IPS_GENERIC && chef_node }

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        after_save :update_need_apply, if: proc { |s| s.is_sensor_generic? }
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{IPS_GENERIC}") do
        'sensors/ips_sensor_generic'
      end
    end

    module InstanceMethods
      def is_sensor_generic?
        type == IPS_GENERIC
      end

      define_method("icon_#{IPS_GENERIC}") do
        'icons/sensor_generic.svg'
      end

      define_method("full_name_#{IPS_GENERIC}") do
        name
      end

      define_method("secondary_name_#{IPS_GENERIC}") do
        name
      end

      define_method("complete_name_#{IPS_GENERIC}") do
        name
      end

      define_method("type_to_s_#{IPS_GENERIC}") do
        'sensor_generic'
      end

      define_method("type_humanize_#{IPS_GENERIC}") do
        'sensor_generic'
      end

      #
      # IPS is a physical sensor.
      #
      # == Returns:
      # true
      #
      define_method("physical_#{IPS_GENERIC}?") do
        true
      end

      define_method("chef_node_name_#{IPS_GENERIC}") do
        "rbipsg-#{id}"
      end

      define_method("chef_node_#{IPS_GENERIC}") do
        Chef::Node.load("rbipsg-#{id}") rescue nil
      end

      define_method("chef_client_#{IPS_GENERIC}") do
        Chef::ApiClient.load("rbipsg-#{id}") rescue nil
      end

      define_method("last_check_in_#{IPS_GENERIC}") do
        begin
          node = chef_node
          Time.at(node[:ohai_time])
        rescue
          return false
        end
      end

      define_method("segments_#{IPS_GENERIC}") do
        node = chef_node
        return unless node && node['redborder']['segments']
        node['redborder']['segments'].keys.sort
      end

      define_method("cpu_number_#{IPS_GENERIC}") do
        node = chef_node
        return 1 unless node && node['cpu']
        node['cpu']['total']
      end

      define_method("monitors_to_chef_#{IPS_GENERIC}") do
        role = chef_role
        role.override_attributes["redborder"]["monitors"] = monitor_categories_to_chef
        role.save
      end

      define_method("wakeupable_#{IPS_GENERIC}") do
        self
      end

      define_method("version_#{IPS_GENERIC}") do
        chef_node["redborder"]["rpms"]["IPS-sensor"] if chef_node
      end

      define_method("reduced_version_#{IPS_GENERIC}") do
        'Generic'
      end

      #
      # => Method used to complete values about ancestors on chef role
      #    For IPS_GENERIC, it is only useful on groups
      #
      define_method("update_chef_domain_values_#{IPS_GENERIC}") do
        update_chef_domain_values_in_node
      end

      define_method("as_json_with_children_#{IPS_GENERIC}") do
        # Return nothing for IPS sensors
        return
      end

      define_method("can_have_children_#{IPS_GENERIC}?") do
        true
      end

      protected

        def update_need_apply
          self.need_apply = true

          role = self.chef_role
          role.override_attributes['redborder']['ipsgeneric'] ||= {}
          role.override_attributes['redborder']['ipsgeneric']['rules'] ||= {}
          role.override_attributes['redborder']['ipsgeneric']['rules']['need_apply'] = true
          role.save
        end

        define_method "run_list_text_#{IPS_GENERIC}" do
          'role[ipsg-sensor]'
        end

        define_method("create_chef_role_#{IPS_GENERIC}") do
          # Create the chef node and client if them dont exists
          # since we are using them in this method

          run_list_roles = []
          run_list_roles << 'role[ipsg-sensor]' unless self.is_domain?
          run_list_roles << "role[#{parent.chef_name}]"

          role_json = { 'name' => chef_name, 'run_list' => run_list_roles }

          role_json['override_attributes'] =
            { 'redborder' => { 'role' => chef_name,
                               'sensor_uuid' => uuid,
                               'snort' => { 'default' => { 'ipvars' => {},
                                                           'portvars' => {},
                                                           'save_pcap' => nil,
                                                           'mode' => nil,
                                                           'autobypass' => nil } },
                               'rsyslog' => {},
                               'proxy' => {},
                               'snmp' => { 'trap_server' => [] },
                               'monitors' => [] } }

          ipvars = parent.role_value_for_vars('ipvars')
          portvars = parent.role_value_for_vars('portvars')
          mode = parent.role_value(%w(redborder snort default mode))
          autobypass = parent.role_value(%w(redborder snort autobypass))

          role_json['override_attributes']['redborder']['snort']['groups'] =
            { '0' => { 'name' => 'default',
                       'mode' => mode,
                       'domain_uuid' => Sensor.root.uuid,
                       'domain_name' => 'root',
                       # Assign, if exists, only the first segment.
                       # If there are more than 1, let them unassigned.
                       'segments' => (segments.nil? ? nil : segments.first(1)),
                       'cpu_list' => (cpus_number.nil? ? nil : cpus_number.times.collect(&:to_s)),
                       'autobypass' => autobypass,
                       'monitors' => [],
                       'bindings' => { '0' => { 'name' => 'default',
                                                'ipvars' => ipvars,
                                                'portvars' => portvars,
                                                'save_pcap' => false,
                                                'vlan_objects' => {},
                                                'network_objects' => {} } } } }

          role = Chef::Role.from_hash(role_json)
          role.create
        end

        define_method "destroy_chef_node_#{IPS_GENERIC}" do
          chef_node.destroy
        end
    end
  end
end

Sensor.send(:include, RedBorder::SensorIPSGeneric)
