#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SensorsControllerIPS
    def self.included(base)

      base.class_eval do
        unloadable

        before_action :validate_cert_ips, only: [
          :classifications, :get_rule_db_version_ids, :unicode_map
        ]

        skip_before_action :verify_authenticity_token, only: [
          :associate, :associate_v2
        ], raise: false

        before_action :check_user_and_password, only: [
          :associate, :associate_v2
        ]

        before_action :set_sensor_ips, only: [
          :classifications, :unicode_map, :new_group, :create_group,
          :new_binding, :create_binding, :edit_variable, :update_variable,
          :delete_variable, :export_vars, :import_vars, :import_vars_file,
          :destroy_object, :new_vlan, :create_vlan, :new_network,
          :create_network, :configuration_update,
          :change_snort_mode, :change_group_name
        ]

        before_action :check_manage_privileges_ips, only: [
          :configuration_update
        ]

        before_action :require_administrative_privileges, only: [
          :new_group, :create_group, :new_binding, :create_binding,
          :edit_variable, :update_variable, :delete_variable, :import_vars,
          :import_vars_file, :edit_variable, :update_variable, :new_vlan,
          :create_vlan, :new_network, :create_network, :bypass,
          :segment_enabled, :change_group_name, :change_snort_mode
        ]

        before_action :recalculate_policies, only: [:update]
        before_action :set_ips_sensors_per_page, only: [:ips, :ips_load_more]
        before_action :set_default_sensor_view_ips, only: :ips

        def ips

          # The current page
          @page = 0;

          # The current letter
          @current_letter = ''

          if params[:id].nil?
            sensor_ids = Sensor.where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC],
                                      is_deleted: false)
                         .select { |s| current_user.can?(:read, s) }.map(&:id)
          else
            @sensor = Sensor.find(params[:id])
            sensor_ids = @sensor.all_children
                         .where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC],
                                is_deleted: false)
                         .select { |s| current_user.can?(:read, s) }.map(&:id)
          end
          @sensors = Sensor.unscoped.where(id: sensor_ids)
                     .order('LOWER(name) ASC').take(@sensors_page)
        end

        def classifications
          if @sensor.is_sensor? || @sensor.is_sensor_proxy? || @sensor.is_sensor_generic?
            group = Sensor.find_by(real_parent: @sensor, group_id: params[:group_id], type: Sensor::IPS_GROUP)
            policy_ids = Sensor.where(real_parent: group, type: Sensor::IPS_BINDING).map(&:policy_id)
          else
            policy_ids = @sensor.policy_id
          end

          @classifications = ClassType.includes(rule_db_version: [:policies]).where(policies: {id: policy_ids})
        end

        # GET /sensors/new_ips
        def new_ips
          @sensor = Sensor.new(type: Sensor::IPS, parent_id: 1)
          render layout: false
        end

        def new_group
          @sensor_group = Sensor.new
          render layout: false
        end

        # POST /sensors/create_fow
        def create_ips
          @sensor = Sensor.new(send("sensor_params_#{Sensor::IPS}"))
          @sensor.parent_id = current_user.super_admin? ? Sensor.root.id : current_user.sensor.id

          respond_to do |format|
            if @sensor.save
              Delayed::Job.enqueue(Redborder::Jobs::RbWakeupChefClusterJob.new(true, current_user.id), priority: 5) unless DelayedJob.rb_wakeup_chef_cluster?
              format.html { redirect_to sensors_path, notice: "#{I18n.t 'sensors_controller_ips.created'}" }
              format.json { head :no_content }
              format.js
            else
              format.html { render action: 'new_ips' }
              format.json { render json: @sensor.errors, status: :unprocessable_entity }
              format.js
            end
          end
        end

        def create_group
          @sensor_group = Sensor.new(name: params[:sensor][:name])
          @sensor_group.type = Sensor::IPS_GROUP
          @sensor_group.parent = @sensor
          @sensor_group.real_parent = @sensor

          respond_to do |format|
            if @sensor_group.save
              format.html { redirect_to edit_sensor_path(@sensor) }
              format.js
            else
              format.html { redirect_to edit_sensor_path(@sensor), error: @sensor_group.errors }
              format.js
            end
          end
        end

        #
        # => Change the group name inline with a $.post call
        #
        def change_group_name

          # Get the name
          name = params.require(:sensor).permit(:name)

          # Assign the name to the sernsor group. When the sensor is saved,
          # a callback assign the name to chef.
          @sensor.name = name[:name]

          respond_to do |format|

            if @sensor.save
              format.js {render json: { status: "ok", id: @sensor.id, name: @sensor.name } }
            else
              format.js {render json: {error: @sensor.errors.full_messages.join(","), id: @sensor.id}, status: :unprocessable_entity }
              format.json { render json: { error: @sensor.errors.full_messages.join(","), id: @sensor.id }, status: :unprocessable_entity }
            end

          end

        end

        #
        # => Change the Snort mode of a sensor inline
        #
        def change_snort_mode

          # Try to save the snort mode
          value = params.require(:sensor).permit(:snort_mode)[:snort_mode]

          respond_to do |format|

            if @sensor.update_snort_mode_from_group(value)
              format.js {render json: { status: "ok", id: @sensor.id } }
            else
              format.js {render json: {error: @sensor.errors.full_messages.join(","), id: @sensor.id}, status: :unprocessable_entity }
              format.json { render json: { error: @sensor.errors.full_messages.join(","), id: @sensor.id }, status: :unprocessable_entity }
            end

          end

        end

        # Return the recent unicode.map file for the group.
        def unicode_map
          group = Sensor.find_by(real_parent: @sensor, group_id: params[:group_id], type: Sensor::IPS_GROUP)
          rule_db_versions = RuleDbVersion.includes(policies: [:sensors]).where(sensors: {real_parent_id: group.id,
                                                                                          type: Sensor::IPS_BINDING}).order(created_at: :desc).all

          render text: rule_db_versions.map(&:unicode_map).uniq.join("\n")
        rescue
          render text: ""
        end

        #
        # => Search the user based on the first letter
        #    It's receive a letter.
        #
        def ips_letter_search
          # Get the letter
          @letter = params[:id]

          # Check the letter
          if [*"A".."Z"].include? @letter
            sensor_ids = if params[:sensor_id].nil?
              Sensor.where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC], is_deleted: false)
              .where('LOWER(name) LIKE ?', "#{@letter.downcase}%")
              .select { |s| current_user.can?(:read, s) }.map(&:id)
            else
              @sensor = Sensor.find(params[:sensor_id])
              @sensor.all_children
              .where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC], is_deleted: false)
              .where('LOWER(name) LIKE ?', "#{@letter.downcase}%")
              .select { |s| current_user.can?(:read, s) }.map(&:id)
            end
            @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
            render :ips_search
          else
            # Something went wrong
            redirect_to ips_sensors_path
          end
        end

        #
        # => Search a sensor based on a string
        #
        def ips_search
          @search = params[:search]
          sensor_ids = if params[:sensor_id].nil?
            Sensor.where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC], is_deleted: false)
            .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
            .select { |s| current_user.can?(:read, s) }.map(&:id)
          else
            @sensor = Sensor.find(params[:sensor_id])
            @sensor.all_children.where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC], is_deleted: false)
            .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
            .select { |s| current_user.can?(:read, s) }.map(&:id)
          end
          @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
        end

        def ips_load_more
          # Get the page
          @page = params[:page].to_i

          # Get the sensors
          if params[:sensor_id].nil?
            sensor_ids = Sensor.where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC],
                                      is_deleted: false)
                         .select { |s| current_user.can?(:read, s) }.map(&:id)
          else
            @sensor = Sensor.find(params[:sensor_id])
            sensor_ids = @sensor.all_children
                         .where(type: [Sensor::IPS, Sensor::IPS_PROXY, Sensor::IPS_V2, Sensor::IPS_GENERIC],
                                is_deleted: false)
                         .select { |s| current_user.can?(:read, s) }.map(&:id)
          end

          @sensors = Sensor.unscoped
                     .where(id: sensor_ids)
                     .offset(@page * @sensors_page)
                     .order('LOWER(name) ASC').take(@sensors_page)

          # Set the current letter empty
          before_sensor = Sensor.unscoped
                          .where(id: sensor_ids)
                          .order(:name)
                          .offset((@page * @sensors_page) - 1).first

          if before_sensor.nil?
            @current_letter = ''
          else
            @current_letter = before_sensor.secondary_name[0, 1].upcase
          end

          respond_to do |format|
           format.js {}
           format.json { render json: @sensors }
          end
        end

        def new_binding
          @sensor_binding = Sensor.new
          render layout: false
        end

        def create_binding
          @sensor_binding = Sensor.new(name: params[:sensor][:name])
          @sensor_binding.type = Sensor::IPS_BINDING
          @sensor_binding.parent = @sensor
          @sensor_binding.real_parent = @sensor
          @sensor_binding.group_id = @sensor.group_id

          respond_to do |format|
            if @sensor_binding.save
              format.html { redirect_to edit_sensor_path(@sensor) }
              format.js
            else
              format.html { redirect_to edit_sensor_path(@sensor), error: @sensor_binding.errors }
              format.js
            end
          end
        end

        def assign_policy
          @sensor = Sensor.find(params[:sensor_id])

          return error_403("#{I18n.t 'sensors_controller_ips.privileges'}") unless current_user.can?(:manage, @sensor)

          policy = Policy.find(params[:id])
          # Update policy children
          @sensor.assign_policy_to_children(policy)

          redirect_to sensor_policies_path(@sensor)
        end

        DittoCode::Exec.if 'COMMUNITY' do
        def apply_update
          if params[:sensor_ids]
            # Execute Job for standard ips v3
            if params[:sensor_ids][:standard]
              sensors = Sensor.where(id: params[:sensor_ids][:standard]).select { |s| current_user.can?(:manage, s) }
              sensors.each { |s| s.update(applying: true) }
              sensors_by_ips = sensors.group_by(&:real_sensor_parent).values
              @sensor = sensors.first

              if PolicyRule.where(policy_id: @sensor.policy_id).count <= RulesHelper::LIMIT_RULE
                sensors_by_ips.each_with_index do |sensor_by_ips, i|
                Delayed::Job.enqueue(
                  Redborder::Jobs::ApplyUpdateJob.new(true, sensor_by_ips.map(&:id), current_user.id, I18n.locale),
                  run_at: Time.now + (i * 5).seconds
                )
                end

                Delayed::Job.enqueue(
                  Redborder::Jobs::StoredPoliciesJob.new(true, sensors.map(&:id), current_user.id),
                  run_at: Time.now
                ) if sensors.select(&:is_sensor_binding?).present?
              end
            end
            # Execute Job for ips v2
            if params[:sensor_ids][:v2]
              sensors_v2 = params[:sensor_ids][:v2]
              list = Sensor.where(id: sensors_v2.values.flatten.uniq).select { |s| current_user.can? :manage, s }
              list.each { |s| s.update(applying: true) }
              Delayed::Job.enqueue(
                Redborder::Jobs::ApplyUpdateJobV2.new(true, sensors_v2, current_user.id, I18n.locale),
                run_at: Time.now + 5.seconds
              )
            end
          end
          redirect_to sensors_path
        end
        end

        DittoCode::Exec.if 'ENTERPRISE' do
        def apply_update
          if params[:sensor_ids]
            # Execute Job for standard ips v3
            if params[:sensor_ids][:standard]
              sensors = Sensor.where(id: params[:sensor_ids][:standard])
              .select { |s| current_user.can?(:manage, s) && s.real_sensor_parent.license.try('valid?') }
              sensors.each { |s| s.update(applying: true) }
              sensors_by_ips = sensors.group_by(&:real_sensor_parent).values

              sensors_by_ips.each_with_index do |sensor_by_ips, i|
              Delayed::Job.enqueue(
                Redborder::Jobs::ApplyUpdateJob.new(true, sensor_by_ips.map(&:id), current_user.id, I18n.locale),
                run_at: Time.now + (i * 5).seconds
              )
              end
              Delayed::Job.enqueue(
                Redborder::Jobs::StoredPoliciesJob.new(true, sensors.map(&:id), current_user.id),
                run_at: Time.now
              ) if sensors.select(&:is_sensor_binding?).present?
            end

            # Execute Job for ips v2
            if params[:sensor_ids][:v2]
              sensors_v2 = params[:sensor_ids][:v2]
              list = Sensor.where(id: sensors_v2.values.flatten.uniq).select { |s| current_user.can? :manage, s }
              list.each { |s| s.update(applying: true) }
              Delayed::Job.enqueue(
                Redborder::Jobs::ApplyUpdateJobV2.new(true, sensors_v2, current_user.id, I18n.locale),
                run_at: Time.now + 5.seconds
              )
            end
          end
          redirect_to sensors_path
        end
        end

        def get_rule_db_version_ids
          sensor = Sensor.find(params[:id])
          group = Sensor.find_by(real_parent: sensor, group_id: params[:group_id])

          rule_db_versions = Sensor.where(real_parent: group, type: Sensor::IPS_BINDING).map(&:rule_db_versions).flatten

          render json: rule_db_versions.map(&:id)
        end

        define_method("destroy_#{Sensor::IPS}") do
          add_delete_sensor_job
        end

        define_method("destroy_#{Sensor::IPS_PROXY}") do
          add_delete_sensor_job
        end

        define_method("destroy_#{Sensor::IPS_GENERIC}") do
          add_delete_sensor_job
        end

        define_method("destroy_#{Sensor::IPS_V2}") do
          add_delete_sensor_job
        end

        define_method("destroy_#{Sensor::IPS_GROUP}") do
          destroy_group_or_binding
        end

        define_method("destroy_#{Sensor::IPS_BINDING}") do
          destroy_group_or_binding
        end

        def destroy_group_or_binding
          if @sensor.is_sensor_binding? and @sensor.binding_id == 0
            respond_to do |format|
              format.html { redirect_to sensors_url, alert: "#{I18n.t 'sensors_controller_ips.binding'}" }
              format.json { head :no_content, status: 403 }
              format.js {render "", status: 403}
            end
          elsif (@sensor.real_parent.children.where(is_deleted: false).size <= 1)
            respond_to do |format|
              format.html { redirect_to sensors_url, alert: "#{I18n.t 'sensors_controller_ips.sensor'}" }
              format.json { head :no_content, status: 403 }
              format.js {render "", status: 403}
            end
          else
            if request.format.js?
              @real_sensor = @sensor.real_parent
            end

            add_delete_sensor_job
          end
        end

        def add_delete_sensor_job
          @sensor.update_column(:is_deleted, true)

          Delayed::Job.enqueue(Redborder::Jobs::DeleteSensorJob.new(@sensor.id, current_user.id, true, I18n.locale), priority: 10, run_at: Time.now)

          respond_to do |format|
            format.html { redirect_to sensors_url }
            format.json { render json: { query: true, id: @sensor.id } }
            format.js
          end
        end

        def associate
          sensor_name = params[:sensor_name]
          sensor_ip = params[:ipaddress]

          # create sensor
          @sensor = Sensor.create(name: sensor_name,
                                  type: Sensor::IPS,
                                  ip: sensor_ip,
                                  parent_id: @user.sensor_id)

          if @sensor.errors.present?
            render json: @sensor.errors, status: :unprocessable_entity and return
          end

          # Node is created by a callback after the sensor is created

          # create client
          node_name = "rbips-#{@sensor.id}"
          client_json = { 'name' => node_name }
          client = Chef::ApiClient.from_hash(client_json)
          client = client.save

          erchef_cert_file = "/etc/nginx/ssl/erchef.crt"

          rest = Chef::ServerAPI.new(Chef::Config[:chef_server_url])
          %w(read update).each do |perm|
            ace = rest.get_rest("nodes/#{node_name}/_acl")[perm]
            ace['actors'] << node_name unless ace['actors'].include?(node_name)
            rest.put_rest("nodes/#{node_name}/_acl/#{perm}", perm => ace)
          end

          render json: { sensor_id: @sensor.id,
                         client_api_key: client['private_key'].to_json,
                         encrypted_data_bag_secret: File.read("/etc/chef/encrypted_data_bag_secret").to_json,
                         erchef_cert: {
                           name: File.basename(erchef_cert_file),
                           content: File.read(erchef_cert_file).to_json
                         },
                         created: true },
                 status: :ok
        end

        def associate_v2
          sensor_name = params[:sensor_name]
          sensor_ip = params[:ipaddress]

          # create sensor
          @sensor = Sensor.create(name: sensor_name,
                                  type: Sensor::IPS_V2,
                                  ip: sensor_ip,
                                  parent_id: @user.sensor_id)

          if @sensor.errors.present?
            render json: @sensor.errors, status: :unprocessable_entity and return
          end

          # Node is created by a callback after the sensor is created

          # create client
          client_json = { 'name' => "rbipsv2-#{@sensor.id}" }
          client = Chef::ApiClient.from_hash(client_json)
          client = client.save

          erchef_cert_file = "#{Rails.root}/config/erchef.crt"

          render json: { sensor_id: @sensor.id,
                         client_api_key: client['private_key'].to_json,
                         #encrypted_data_bag_secret: File.read('/etc/chef/encrypted_data_bag_secret').to_json,
                         erchef_cert: {
                           name: File.basename(erchef_cert_file),
                           content: File.read(erchef_cert_file).to_json
                         },
                         created: true },
                 status: :ok
        end

        def edit_variable
          @variable = params[:variable]
          @type = params[:type]
          @value = @sensor.role_value(["redborder", "snort", "default", @type, @variable])

          render layout: false
        end

        def update_variable
          role = @sensor.chef_role
          @type = params[:type]
          @variable = params[:variable]
          @value = params[:value].gsub(/\s+/, "")

          if validate_variable(@type, @variable, @value)
            role.override_attributes['redborder']['snort']['default'][@type][@variable] = @value
            role.save
            invoke_binding_vars_job
          else
            @error = true
          end
        end

        def delete_variable
          role = @sensor.chef_role
          type = params[:type]
          variable = params[:variable]

          role.override_attributes['redborder']['snort']['default'][type].delete(variable)
          role.save
          invoke_binding_vars_job

          render type
        end

        def export_vars
          type = params[:type]
          export_vars_csv type
        end

        def import_vars
          @type = params[:type]
          render layout: false
        end
        ####################################################################################################
        # Funtion to clean the "new lines" in the CSV File
        ####################################################################################################
        def CleanBlankSpaceFile(file)
          File.write(file, File.readlines(file).reject { |s| s.strip.empty? }.join)
        end

        ####################################################################################################
        # Funtion to clean the "new lines" in the CSV File
        ####################################################################################################
        def clean(input)
          input.gsub(/^[\s]*$\n/, "").strip
        end

        ####################################################################################################
        # Function that checks if the CSV file uploaded for the IPVARS and PORTVARS has the correct header
        # The two columns that are obligatory are "Variable" and "Values"
        # The "inherited" columns is filled with false when it has a nil value
        ####################################################################################################
        def checkHeader(file)
          CSV.foreach(file.path, headers: true) do |csv|
            columns = csv.headers[0].sub(/[\n]+|[\r]+/, '').split(';')
            variable , values, inherited = columns
            if variable == "Variable" && values == "Values"
                return true
            else 
              return false
            end
          end
        end
        def import_vars_file
          type = params[:sensor][:type]
          content_file = params[:sensor][:csv_file]

          # Clean the CSV file. Eliminate white space (new lines)
          CleanBlankSpaceFile(content_file.path)

          # Before deciding on adding or replacing variables first it checks if the header of the file is correct
          if checkHeader(content_file)
            if params[:add_vars] == 'true'
              vars_count = @sensor.add_vars_from_file(content_file, type)
              msg = "#{vars_count[0]} #{I18n.t 'sensors_controller_ips.added'} #{vars_count[1]} #{I18n.t 'sensors_controller_ips.updated'}"
            else
              vars_count = @sensor.replace_vars_from_file(content_file, type)
              msg = "#{vars_count} #{I18n.t 'sensors_controller_ips.added'}"
            end
            invoke_binding_vars_job
            redirect_to edit_sensor_path(@sensor), notice: msg
          else
            error_msg = "#{I18n.t 'sensors_controller_ips.unknown_content_csv'}"
            redirect_to edit_sensor_path(@sensor), alert: error_msg
          end
        end

        #
        # Destroy NET and VLAN objects on a binding/group
        #
        def destroy_object
          @type = params[:type]
          @id = params[:object_id]
          role = @sensor.real_sensor_parent.chef_role
          role.override_attributes["redborder"]["snort"]["groups"]["#{@sensor.group_id}"]["bindings"]["#{@sensor.binding_id}"]["#{@type}_objects"].delete(@id)
          role.save
        end

        #
        # Show the form to associate a new VLAN to a binding/group
        #
        def new_vlan
          vlans_in_role =
            @sensor
            .real_sensor_parent
            .chef_role
            .override_attributes["redborder"]["snort"]["groups"]["#{@sensor.group_id}"]["bindings"]["#{@sensor.binding_id}"]["vlan_objects"]
            .map { |id, vlan| vlan[vlan.keys.first] }
          @vlans =
            current_user.vlan_objects
            .select { |vlan| !vlans_in_role.include?(vlan[:value]) }
            .map { |vlan| { name: vlan.name, value: vlan.value, id: vlan.id } }
          # Show the form
          render layout: false
        end

        #
        # Save the VLAN on a group/binding
        #
        def create_vlan
          @vlan_object = current_user.vlan_objects.find(params["vlan_object_id"])

          if @vlan_object.nil?
            @err = true
          else
            role = @sensor.real_sensor_parent.chef_role
            role.override_attributes["redborder"]["snort"]["groups"]["#{@sensor.group_id}"]["bindings"]["#{@sensor.binding_id}"]["vlan_objects"][@vlan_object.id] = {@vlan_object.name => @vlan_object.value}
            role.save
          end
        end

        #
        # Show the form to associate a new Net Object to a binding/group
        #
        def new_network
          nets_in_role =
            @sensor
            .real_sensor_parent
            .chef_role
            .override_attributes["redborder"]["snort"]["groups"]["#{@sensor.group_id}"]["bindings"]["#{@sensor.binding_id}"]["network_objects"]
            .map { |id, net| net[net.keys.first] }
          @networks =
            current_user.net_objects
            .select { |net| !nets_in_role.include?(net[:value]) }
            .map { |net| { name: net.name, value: net.value, id: net.id } }
          # Show the form
          render layout: false
        end

        #
        # Save the Net Object on a group/binding
        #
        def create_network
          @net_object = current_user.net_objects.find(params["net_object_id"])
          if @net_object.nil?
            @err = true
          else
            role = @sensor.real_sensor_parent.chef_role
            role.override_attributes["redborder"]["snort"]["groups"]["#{@sensor.group_id}"]["bindings"]["#{@sensor.binding_id}"]["network_objects"][@net_object.id] = {@net_object.name => @net_object.value}
            role.save
          end
        end

        #
        # Apply the configuration to the current sensor and his childrens.
        #
        def configuration_update
          # We have checked the privileges on check_manage_privileges_ips
          @sensors_v2 = get_ips_children(@sensor, Sensor::IPS_V2)
          @sensors_group = get_ips_children(@sensor, Sensor::IPS_GROUP)
          @sensors_binding = get_ips_children(@sensor, Sensor::IPS_BINDING)
          @sensors_generic = get_ips_children(@sensor, Sensor::IPS_GENERIC)

          redirect_to :back, alert: "#{I18n.t 'sensors_controller_ips.no_sensors'}" if
            @sensor_v2 && @sensors_group.count.zero? && @sensors_binding.count.zero? && @sensors_generic.count.zero?
        end

        def bypass
          @sensor.change_bypass_mode(params[:segment], params[:mode])
          @error = false
          # Run Chef if it's not running now
          unless DelayedJob.rb_wakeup_chef_cluster?
            Delayed::Job.enqueue(Redborder::Jobs::RbWakeupChefClusterJob
              .new(true, current_user.id), priority: 5)
          end
        rescue Net::SSH::AuthenticationFailed => e
          @error = true
          Rails.logger.error "Sensor: #{@sensor.name} #{e.inspect}"
          Rails.logger.error e.backtrace.join("\n")
          flash[:error] = "#{I18n.t 'sensors_controller_ips.create_error'}"
        end

        #
        # => Action used to enabled or disabled segments for IPS V2.
        #
        def segment_enabled
          role = @sensor.chef_role
          node = @sensor.chef_node

          if role.override_attributes['redborder']['snort']['initialized']
            if params[:mode] == 'true'
              role.override_attributes['redborder']['snort']['segments'] << params[:segment]
            else
              role.override_attributes['redborder']['snort']['segments'].delete(params[:segment])
            end
          # First time segments are disabled
          else
            role.override_attributes['redborder']['snort']['segments'] = node['redborder']['segments'].keys.sort
            role.override_attributes['redborder']['snort']['segments'].delete(params[:segment])
          end

          # Always mark initialized as true
          role.override_attributes['redborder']['snort']['initialized'] = true

          role.save
          @error = false
        rescue StandardError => e
          @error = true
          Rails.logger.error "Sensor: #{@sensor.name} #{e.inspect}"
          Rails.logger.error e.backtrace.join("\n")
          flash[:error] = "#{I18n.t 'sensors_controller_ips.update_error'}"
        end

        def update_rpms
          sensor = Sensor.find(params[:id])
          update_rpms_ips sensor if ([Sensor::DOMAIN, Sensor::IPS].include?(sensor.type) and User.current_user.super_admin?)
          redirect_to :back
        end

        def raw
          sensor = Sensor.find_by name: params[:name]
          @result = sensor.view_raw_info params

          render layout: false
        end

        private

        #
        # => Check
        #
        def check_user_and_password
          # Check if username and password are present
          render json: { msg: "#{I18n.t 'sensors_controller_ips.invalid'}" },
                 status: :conflict and return unless params[:username] &&
                                                     params[:password]

          # Check if user and password are correct.
          @user = User.where(['lower(login) = :value OR lower(email) = :value',
                             { value: params[:username].downcase }]).first

          render json: { msg: "#{I18n.t 'sensors_controller_ips.invalid'}" },
                 status: :conflict and return unless @user &&
                                                     @user.valid_password?(params[:password])

          render json: { msg: "#{I18n.t 'sensors_controller_ips.user_admin'}" },
                 status: :forbidden and return unless @user.admin?
        end

        def set_default_sensor_view_ips
          session[:sensor_view] = params[:action].to_sym
        end

        def update_rpms_ips sensor
          begin
            Delayed::Job.enqueue(Redborder::Jobs::UpdateRPMSJob.new(true, current_user.id, params[:id])) unless DelayedJob.update_rpms?
          rescue Exception
            Rails.logger.info $!.inspect
            Rails.logger.info $@.join("\n")
          end
        end

        def authenticate_user_ips
          authenticate_user
        end

        def validate_cert_ips
          validate_cert
        end

        def set_sensor_ips
          set_sensor
        end

        #
        # Check if the current user has manage privileges on current
        # initialized sensor.
        #
        def check_manage_privileges_ips
          manage_privileges
        end

        #
        # Get the children of a sensor based on sensor type. Includes policies
        # and order.
        #
        # == Parameters:
        # sensor::
        #   Sensor to get the children
        # sensor_type::
        #   Integer specify the type of sensor.
        #
        # == Returns:
        # ActiveRecord association with results.
        #
        def get_ips_children(sensor, sensor_type)
          sensor
            .all_children(true)
            .where(type: sensor_type)
            .includes(:policy)
            .sort do |s1, s2|
              s1.real_sensor_parent.name <=> s2.real_sensor_parent.name
            end
        end

        def set_ips_sensors_per_page
          @sensors_page = Sensor::SENSORS_PAGE
        end

        define_method("sensor_params_#{Sensor::IPS}") do
          params.require(:sensor)
            .permit(:parent_id, :type, :name, :ip, :general_description, :license_id)
        end

        define_method("sensor_params_#{Sensor::IPS_PROXY}") do
          params.require(:sensor)
            .permit(:parent_id, :general_description, :license_id)
        end

        define_method("sensor_params_#{Sensor::IPS_GENERIC}") do
          params.require(:sensor)
            .permit(:parent_id, :type, :name, :ip, :general_description,
                    :rules_path, :unified2_files, :unified2_name, :commands,
                    :gen_file, :reference_file, :classification_file,
                    :snmp_community, :snort_stats_file, :filter_rules_path, :license_id)
        end

        def export_vars_csv(type)
          filename = in_time_machine { Time.now.strftime("%Y%m%d%H%M%S") + "-#{type.capitalize}-" + @sensor.name + ".csv" }
          content = Sensor.root.all_vars_to_csv_format(type, @sensor)
          send_data content, filename: filename
        end

        def invoke_binding_vars_job
          Delayed::Job.enqueue(Redborder::Jobs::BindingVarsJob.new(true, @sensor.id, current_user.id), priority: 15, run_at: Time.now + 5.seconds)
        end

        def recalculate_policies
          if [Sensor::IPS, Sensor::IPS_GROUP, Sensor::IPS_BINDING, Sensor::DOMAIN].include? @sensor.type
            @sensor.recalculate_policies if params[:sensor][:parent_id]
          end
        end

        define_method("new_sensor_params_#{Sensor::IPS}") do
          params.require(:sensor).permit(:type, :name, :ip, :license_id)
        end

        def validate_variable(type, variable, value)
          if type == 'ipvars' && variable == '$HOME_NET'
            value.match /^(((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/
          elsif type == 'ipvars'
            value.match /^(!?(\$HOME_NET)|((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/
          elsif type == 'portvars'
            value.match /^((((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?))(\s*,\s*((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?)))*))$/
          end
        end

        def skip_authenticate_user_with_ips
          return true if [:classifications, :get_rule_db_version_ids,
                          :unicode_map, :associate,
                          :associate_v2].include?(action_name.to_sym)
          skip_authenticate_user_without_ips
        end

        alias_method_chain :skip_authenticate_user, :ips
      end
    end
  end
end
