#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SettingsControllerIPS
    def self.included(base)
      base.class_eval do
        unloadable

        def new_variable
          @type = params[:type]
          render layout: false
        end

        def create_variable
          role = Sensor.root.chef_role
          type = params[:type]
          variable = params[:variable]
          value = params[:value]
          if role.override_attributes['redborder']['snort']['default'][type][variable]
            @error = "#{I18n.t 'settings_controller_ips.variable'} #{variable} #{I18n.t 'settings_controller_ips.defined'}"
          elsif validate_variable(type, variable, value)
            role.override_attributes['redborder']['snort']['default'][type][variable] = value
            role.save
            invoke_binding_vars_job
          else
            @error = "#{I18n.t 'settings_controller_ips.value'} #{variable} #{I18n.t 'settings_controller_ips.not_correct'}"
          end

          render type
        end

        def delete_variable
          type = %w(ipvars portvars).find do |var_type|
            var_type == params[:type]
          end
          

          return error_422("#{I18n.t 'settings_controller_ips.unknown'}") unless type

          variable = params[:variable]
          if Setting.const_get(type.upcase).include?(variable)
            nil
          else
            role = Sensor.root.chef_role
            role.override_attributes['redborder']['snort']['default'][type].delete(variable)
            role.save
            delete_variable_sensors(type, variable)
            invoke_binding_vars_job
          end

          render type
        end

        def edit_variable
          @variable = params[:variable]
          @type = params[:type]
          @value = Sensor.root.chef_role.override_attributes['redborder']['snort']['default'][@type][@variable]

          render layout: false
        end

        def update_variable
          role = Sensor.root.chef_role
          @type = params[:type]
          @variable = params[:variable]
          @value = params[:value].gsub(/\s+/, "")
          if validate_variable(@type, @variable, @value)
            role.override_attributes['redborder']['snort']['default'][@type][@variable] = @value
            role.save
            invoke_binding_vars_job
          else
            @error = true
          end
        end

        def export_vars
          type = params[:type]
          export_vars_csv type
        end

        def import_vars
          @type = params[:type]
          render layout: false
        end

        ####################################################################################################
        # Funtion to clean the CSV File
        ####################################################################################################
        def clean_file(file)

          unless File.exist?(file)
            fail "#{I18n.t 'settings_controller_ips.exists_error'}"
          end

          clean_file = File.read(file).gsub("\"","").gsub("\'","").gsub("\r\n","\n")

          File.write(file, clean_file)
         
        end

        ####################################################################################################
        # Function that checks if the CSV file uploaded for the IPVARS and PORTVARS has the correct header
        # The two columns that are obligatory are "Variable" and "Values"
        # The "inherited" columns is filled with false when it has a nil value
        ####################################################################################################
        def check_header(file)

          imported_file = File.read(file.path).strip.gsub("\n",";").gsub("\r","").gsub("\"","").gsub("\'","")
          file_splited = imported_file.split(';')
        
          if file_splited[0] == "Variable" && file_splited[1] == "Values" && file_splited[2] == "Inherited"
            return true
          else
            return false
          end
          
        end

        def import_vars_file

          type = params[:setting][:type]
          
          content_file = params[:setting][:csv_file]
          
          temp_file = params[:setting][:csv_file].tempfile.path

          filename = params[:setting][:csv_file].original_filename

          fail "The file: #{filename} #{I18n.t 'settings_controller_ips.empty_error'}" if temp_file.nil? || temp_file.empty? || File.zero?(temp_file)

          #
          # Management of allowed extensions, where files with extension tgz, gz and tar are allowed.
          #
          
          unless filename.match(/(csv)$/)
            fail "#{I18n.t 'settings_controller_ips.ext_error'}"
          end

          
          # Before deciding on adding or replacing variables first it checks if the header of the file is correct
          if check_header(content_file)

             # Clean the CSV file. Eliminate white space (new lines)
            clean_file(content_file.path)

            if params[:add_vars] == "true"
              vars_count = Sensor.root.add_vars_from_file(content_file, type)
              msg = "#{vars_count[0]} #{I18n.t 'settings_controller_ips.added'} #{vars_count[1]} #{I18n.t 'settings_controller_ips.updated'}"
            else
              vars_count = Sensor.root.replace_vars_from_file(content_file, type)
              msg = "#{vars_count} #{I18n.t 'settings_controller_ips.added'}"
            end
            invoke_binding_vars_job
            redirect_to settings_path, notice: msg

          else 
            error_msg = "#{I18n.t 'settings_controller_ips.unknown_content_csv'}"
            redirect_to settings_path, alert: error_msg
          end
        rescue => e
          rb_flash e.message, type: :error
          redirect_to settings_path
        end

        def tos
        end

        private

          def export_vars_csv(type)
            filename = in_time_machine { Time.now.strftime("%Y%m%d%H%M%S") + "-#{type.capitalize}-GeneraSettings.csv" }
            content = Sensor.root.vars_to_csv_format(type)
            send_data content, filename: filename
          end

          def delete_variable_sensors(type_, variable)
            Sensor.where(type: [Sensor::DOMAIN, Sensor::IPS]).where.not(id: 1).each do |sensor|
              role = sensor.chef_role
              role.override_attributes['redborder']['snort']['default'][type_].delete(variable)
              role.save
            end
          end

          def invoke_binding_vars_job
            Delayed::Job.enqueue(
              Redborder::Jobs::BindingVarsJob.new(true, 1, current_user.id),
              priority: 15,
              run_at: Time.now + 5.seconds
            )
          end

          def validate_variable(type, variable, value)
            if type == 'ipvars' && variable == '$HOME_NET'
              value.match /^(((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/
            elsif type == 'ipvars'
              value.match /^(!?(\$HOME_NET)|((!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?)(\s*,\s*(!)?(((?:(25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|[1][0-9][0-9]|[1-9][0-9]|[0-9]))(\/(?:3[0-2]|[12][0-9]|[0-9]))?))*))$/
            elsif type == 'portvars'
              value.match /^((((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?))(\s*,\s*((!)?((?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])(:(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9][0-9]|6[0-4][0-9][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[1-9])?)?)))*))$/
            end
          end
      end
    end
  end
end
