require 'test_helper'

#
# Test methods added from SensorIPS to Sensor
#
class SensorIpsTest < ActiveSupport::TestCase
  context 'an IPS sensor' do
    setup do
      # Global licensing by default
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'global'

      @ips = sensors(:ips)
    end

    should 'be created' do
      assert @ips.valid?, 'IPS sensor is invalid.'
    end

    should 'respond true to physical? method' do
      assert(@ips.physical?, 'IPS has a physical node')
    end

    should 'can reregister the private key' do
      @private_key = Struct.new(:private_key).new('SDADSQASDASDASDQWDAS')
      Chef::ApiClient.stubs(:reregister).returns(@private_key)

      assert_equal(@private_key.private_key, @ips.reregister,
                   'Private key msut be regenerated on cloud proxy')
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @ips.parent_id = market.id
      assert @ips.valid?, 'IPS moved into market'
    end

    should 'be associated with a organization license' do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'organization'

      license = licenses(:four)
      sensor = Sensor.new(name: 'IPS_Sensor_2',
                          type: @ips.type,
                          parent: @ips.parent,
                          ip: '192.168.101.21',
                          license: license)
      assert sensor.valid?
    end

    should 'be associated with a global license' do
      license = licenses(:two)
      sensor = Sensor.new(name: 'IPS_Sensor_2',
                          type: @ips.type,
                          parent: @ips.parent,
                          ip: '192.168.101.21',
                          license: license)
      assert sensor.valid?
    end

    should 'not be associated with an expired license' do
      license = licenses(:one)
      sensor = Sensor.new(name: 'IPS_Sensor_2',
                          type: @ips.type,
                          parent: @ips.parent,
                          ip: '192.168.101.21',
                          license: license)
      assert_not sensor.valid?
    end

    should 'not be associated with a license without empty bucket' do
      license = licenses(:two)
      sensor_1 = Sensor.new(name: 'sensor_license_1',
                            type: @ips.type,
                            parent: @ips.parent,
                            ip: '10.0.50.1',
                            license_id: license.id)
      assert sensor_1.save

      sensor_2 = Sensor.new(name: 'sensor_license_2',
                            type: @ips.type,
                            parent: @ips.parent,
                            ip: '10.0.50.2',
                            license_id: license.id)
      assert_not sensor_2.valid?
    end

    # should 'can\'t contain domains' do
    #   market = sensors(:market)

    #   market.parent_id = @ips.id
    #   assert_not market.valid?, 'IPS sensor can\'t contain domain'
    # end
  end

  context 'an IPS Group sensor' do
    setup do
      @ips_group = sensors(:ips_group)
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @ips_group.parent_id = market.id
      assert @ips_group.valid?, 'IPS Group moved into market'
    end

    # should 'can\'t contain domains' do
    #   market = sensors(:market)

    #   market.parent_id = @ips_group.id
    #   assert_not market.valid?, 'IPS Group sensor can\'t contain domain'
    # end
  end

  context 'an IPS Binding sensor' do
    setup do
      @ips_binding = sensors(:ips_binding)
    end

    should 'be moved into a market' do
      market = sensors(:market)

      @ips_binding.parent_id = market.id
      assert @ips_binding.valid?, 'IPS Binding moved into market'
    end

    should 'can\'t contain domains' do
      market = sensors(:market)

      market.parent_id = @ips_binding.id
      assert_not market.save, 'IPS Binding sensor can\'t contain domain'
    end

    # should 'can\'t be moved into IPS' do
    #   ips = sensors(:ips)

    #   @ips_binding.parent_id = ips.id
    #   assert_not @ips_binding.valid?, 'IPS Binding can\'t be moved into IPS'
    # end
  end
end
