#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module LocationHelper
  def view_selector_dropdown(current)
    content_tag(:ul, class: 'dropdown-menu') do
      current = "Compare" if current == "Comparative"
      (["Tops", "Map", "Movements", "Movements Map", "Sessions", "Compare", "Weather"] - [current]).each do |view|
        action = (view.downcase == "compare") ? :comparative : view.sub(" ","_").downcase.to_sym
        concat content_tag(:li, link_to(view.to_s, { controller: :location, action: action }, class: 'block-ui-but'))
      end
    end
  end

  #
  # Helper for retrieving and processing filter based on source and filter object.
  #
  def filters_for_source(source, filter)
    if filter.filters[source.id].present?
      filter.filters[source.id]
        .try(:except, :client_profile)
        .try(:merge, filter.filters[:shared].slice(:start_time, :end_time, :client_mac, :client_mac_vendor, :calendar_item_name))
    else
      {}
    end
  end

  #
  # Helper to convert minutes to X Hours, Y Minutes format
  #
  def parse_minutes(minutes)
    min = (minutes % 60).to_i
    hours = (minutes / 60).to_i
    result = ''
    result += "#{hours} Hours." if hours > 0
    result += " #{min} Mins." if min > 0
    result
  end

  #
  # Helper to sum all values for certain aggregation at data array
  #
  def total_aggregation_value(data, aggregation)
    data.reduce(0) { |sum, x| sum + x.send(Location::AGGREGATIONS[aggregation][:value]) }.round
  end

  #
  # Helper to set Tooltip Text for Location Custom Button
  #
  def custom_tooltip(range, start_time, end_time)
    return unless (start_time && end_time) && range.to_s == 'custom'
    "From: #{@start_time_for_slider.strftime('%Y/%m/%d %H:%M:%S')} <br>" \
    "To: #{@end_time_for_slider.strftime('%Y/%m/%d %H:%M:%S')}"
  end

  # Helper to saturate a num to a limit
  def saturate(num, limit)
    num > limit ? limit : num
  end

  def location_cast_num(count, aggregation)
    unit = Location::AGGREGATIONS[aggregation][:unit]
    case count
    when 1e12..Float::INFINITY
      "#{(count.to_f / 1e12).round(2)} T#{unit}"
    when 1e9..1e12
      "#{(count.to_f / 1e9).round(2)} G#{unit}"
    when 1e6..1e9
      "#{(count.to_f / 1e6).round(2)} M#{unit}"
    when 1e3..1e6
      "#{(count.to_f / 1e3).round(2)} K#{unit}"
    else
      "#{count} #{unit}"
    end
  end

  def weather_menu_options
    [
      { name: I18n.t('weather.menu_options.temperature'), data_type: 'temperature' },
      { name: I18n.t('weather.menu_options.wind_speed'), data_type: 'wind_speed' },
      { name: I18n.t('weather.menu_options.rain_1h'), data_type: 'rain_1h' },
      { name: I18n.t('weather.menu_options.snow_1h'), data_type: 'snow_1h' }
    ]
  end

  # Processes and condenses weather data from multiple locations.
  # For locations with the same timestamp, it averages the rain and temperature
  # and picks the most frequent weather icon and description.
  def process_and_condense_weather_data(locations)
    locations
      .group_by { |location| location.timestamp.strftime('%Y-%m-%d %H:%M:%S') }
      .map do |timestamp, group|
        # Use the first location as a template for the representative location.
        representative = group.first.dup

        # If there's more than one location in the group, calculate and set averages
        # and the most frequent weather characteristics.
        if group.size > 1
          set_weather_attributes(representative, group)
        end

        representative
      end
  end

  # Sets averaged weather attributes for a representative location based on a group of locations.
  def set_weather_attributes(representative, group)
    averages = calculate_averages(group)
    weather_icon, weather_description = calculate_most_frequent_weather_attributes(group)

    representative.instance_variable_set(:@rain_1h, averages[:rain_1h])
    representative.instance_variable_set(:@temperature, averages[:temperature])
    representative.instance_variable_set(:@weather_icon, weather_icon)
    representative.instance_variable_set(:@weather_description, weather_description)
  end

  # Calculates the average rain and temperature for a group of locations.
  # Returns a hash with the averaged values.
  def calculate_averages(group)
    {
      rain_1h: average(group, :rain_1h),
      temperature: average(group, :temperature)
    }
  end

  # Helper method to calculate the average of a given attribute for a group.
  def average(group, attribute)
    total = group.sum { |loc| loc.send(attribute).to_f }
    (total / group.size).round(2)
  end

  # Calculates the most frequent weather icon and description in a group of locations.
  # Returns the most frequent icon and description as a two-element array.
  def calculate_most_frequent_weather_attributes(group)
    icon = most_frequent(group, :weather_icon)
    description = most_frequent(group, :weather_description)
    [icon, description]
  end

  # Helper method to find the most frequent value of a given attribute in a group.
  def most_frequent(group, attribute)
    group.group_by { |loc| loc.send(attribute) }.max_by { |_, v| v.size }.first
  end


end
