#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Location < Redborder::DataSource
  COLUMNS = Redborder::Module.get :location, :columns
  TABS = Redborder::Module.get :location, :tabs
  COMPARE_TABS = [:pt24h, :pt168h, :pt720h, :pt4320h]
  AGGREGATIONS = Redborder::Module.get(:location, :aggregations)
  DOMAIN_TYPES = [Sensor::CAMPUS, Sensor::BUILDING, Sensor::FLOOR].freeze
  SESSION_ID = Redborder::Module.get(:location, :session_id).to_sym

  TRANSITION = {
    STATIC: 0,
    MOVEMENT: 1
  }.freeze

  COLUMN_NAMES = {
    start_time: 'Start Time',
    end_time: 'End Time'
  }

  STATUS = {
    'ASSOCIATED' => 'Associated',
    'PROBING' => 'Probing'
  }.freeze

  CLIENT_PROFILE = {
    hard: {
      name: 'tr.location.options_client_profile.hard',
      filters: %w(hard)
    },
    medium: {
      name: 'tr.location.options_client_profile.medium',
      filters: %w(hard medium)
    },
    soft: {
      name: 'tr.location.options_client_profile.soft',
      filters: []
    }
  }.freeze

  COLUMNS.map{ |k, v| v[:columns].keys + (v[:complementary_columns].try('keys') || []) }.flatten.each do |column|
    attr_reader column
  end

  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  AGGREGATIONS.keys.each { |aggregation| attr_accessor aggregation }

  attr_reader :timestamp, :result

  attr_accessor :secondary_name

  #
  # Create a Location object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build a Location model.
  #
  # == Returns:
  # Location object
  #
  def initialize(_, args)
    return unless args.is_a?(Hash)
    args.each { |key, value| instance_variable_set("@#{key}", value) }
    @client_mac_name_two = get_mac_vendor_full(@client_mac)
    @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp'].present?

    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      next if current_value.nil?
      if aggregation_data[:round]
        instance_variable_set("@#{aggregation}", current_value.try('round'))
      else
        instance_variable_set("@#{aggregation}", current_value.try('round', 3))
      end
    end
  end

  #
  # Merge user top domain filter into query filters
  #
  def self.user_filters(search, user)
    return search if user.super_user?
    sensor = user.sensor
    domain_key = "#{sensor.domain_type_name(to_key: true)}_uuid".to_sym
    search[domain_key] = (search[domain_key] || {}).merge!(sensor.uuid => { operator: 'is' })
    search
  end

  #
  # Override method to avoid including sensor_uuid filter.
  #
  def self.readable_sensors(_); end

  #
  # Return the Symbol of id module
  #
  def self.id
    :location
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :location
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: controller, action: 'index' }
  end

  def self.default_aggregation
    :clients
  end

  def self.default_view
    :tops
  end

  def self.default_chart_type
    :line
  end

  def self.default_tab
    tab_names.first.to_sym
  end

  def self.compare_tabs
   COMPARE_TABS
  end

  def self.default_aggregations
    :clients
  end

  #
  # Return Module Tabs as strings
  #
  def self.tab_names
    TABS.keys
  end

  def tab_column(tab)
    @tab_column ||= Location::TABS[tab][:filter][:id]
  end

  def tab_column_value(tab)
    tab_column_value = send(tab_column(tab)) || "N/A"
    @tab_column_value ||= tab_column_value
  end

  def client_mac_name_two(tab)
    @client_mac_name_two ||= Location::Tabs[tab][:filter][:id]
  end

  def parse?(tab)
    @must_parse ||= TABS[tab][:parse]
  end

  def self.filter_to_json(user = User.current_user, json = true)
    locations = user.locations_with_permission
    zones = user.zones_with_permission

    f = super.merge!(
      location_columns: [{
        value: 'Campus',
        id: :campus_uuid,
        type: :dropdown
      },
      {
        value: 'Building',
        id: :building_uuid,
        type: :dropdown
      },
      {
        value: 'Floor',
        id: :floor_uuid,
        type: :dropdown
      },
      {
        value: 'Zone',
        id: :zone_uuid,
        type: :dropdown
      },
      {
        value: 'Start Time',
        id: :start_time,
        type: :datetime
      },
      {
        value: 'End Time',
        id: :end_time,
        type: :datetime
      },
      {
        value: 'Calendar Item Name',
        id: :calendar_item_name,
        type: :text_input
      },
      {
        value: 'Client MAC',
        id: :client_mac,
        type: :text_input
      },
      {
        value: 'Client MAC Vendor',
        id: :client_mac_vendor,
        type: :text_input
      }],
      campuses: {
        type: :dropdown,
        value: locations.where(domain_type: Sensor::CAMPUS).map do |domain|
          { value: domain.path_name, id: domain.uuid }
        end
      },
      buildings: {
        type: :dropdown,
        value: locations.where(domain_type: Sensor::BUILDING).map do |domain|
          { value: domain.path_name, id: domain.uuid }
        end
      },
      floors: {
        type: :dropdown,
        value: locations.where(domain_type: Sensor::FLOOR).map do |domain|
          { value: domain.path_name, id: domain.uuid }
        end
      },
      zones: {
        type: :dropdown,
        value: zones.map do |zone|
          { value: zone.name, id: zone.uuid }
        end
      }
    )
    # Check format
    json ? f.to_json : f
  end
end
