#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base Redborder Module
#
module Redborder
  #
  # Base Location Module
  #
  module Location
    #
    # Location QueryManager implement methods to perform module queries
    #
    module QueryManager
      NUM_SECONDS = 600_00

      private

      def movements_data
        set_type_filter
        parser = Redborder::Location::MovementsParser.new(movements_query, @locations)
        { chord: parser.to_chord, table: parser.to_movements_table }
      end

      def total_data
        total_query
      end

      def tops_data
        @filter = @filter.dup

        if params[:w]
          set_client_profile_filters
        end

        @filter.add(:transition, @source::TRANSITION[:STATIC])

        set_type_filter
        data = params[:w] == 'true' ? tops_query_to_widget : tops_query
        if @aggregation == :cumulative_dwell_time
          minutes = Redborder::Rbdruid::GRANULARITY_TYPE[@granularity][:step] / NUM_SECONDS
          data.each { |d| d.cumulative_dwell_time /= minutes }
        elsif @aggregation == :clients
          data.each { |d| d.clients = d.clients.round }
        end
        data
      end

      def movements_map_data
        @filter = @filter.dup
        
        if params[:w]
          set_client_profile_filters
        end
        
        @filter.add(:transition, @source::TRANSITION[:STATIC])

        set_type_filter

        data = tops_query
        if @aggregation == :cumulative_dwell_time
          minutes = Redborder::Rbdruid::GRANULARITY_TYPE[@granularity][:step] / NUM_SECONDS
          data.each { |d| d.cumulative_dwell_time /= minutes }
        elsif @aggregation == :clients
          data.each { |d| d.clients = d.clients.round }
        end
        data
      end

      def serie_data
        @filter = @filter.dup
        
        if params[:w]
          set_client_profile_filters
        end

        set_type_filter

        @filter.add(:transition, @source::TRANSITION[:STATIC])

        # Set filter for #value (or @uuid in tops view)
        value_to_filter = @uuid || @value
        if value_to_filter
          if %w[campus building floor zone].include?(@tab.to_s)
            column = (@tab.to_s + "_uuid").to_sym
          elsif "status" == @tab.to_s
            column = :dot11_status
          else
            column = @tab.to_sym
          end
          @filter.remove(column)
          @filter.add(column, value_to_filter)
        end
        
        data = serie_query

        # TODO: idea, refactor this
        if @aggregation == :cumulative_dwell_time && @granularity != :minute
          minutes = Redborder::Rbdruid::GRANULARITY_TYPE[@granularity][:step] / NUM_SECONDS
          data.each { |d| d[:event][:events] /= minutes }
        end

        data
      end

      def map_data
        @filter.add(:transition, @source::TRANSITION[:STATIC])
        @filter.exclude(:client_latlong, nil)
        map_query
      end

      def client_latlong_data
        @filter = @filter.dup

        if params[:w]
          set_client_profile_filters
        end
        
        @filter.add(@tab.to_sym, @value)
        @filter.add(:transition, @source::TRANSITION[:STATIC])

        @filter.add('client_latlong', [''], 'is_not') unless @filter.exist_any?(:client_latlong)

        data = client_latlong_query
      end

      ##
      ## Queries Methods
      ##

      def total_query
        @source.total(
          start_time: @start_time,
          end_time: @end_time,
          aggregation: @aggregation,
          search: @filter.get
        )
      end

      def serie_query
        @source.timeserie(
          start_time: @start_time,
          end_time: @end_time,
          range: @range || :custom,
          granularity: @granularity,
          search: @filter.get,
          aggregation: @aggregation
        )
      end
      
      def client_latlong_query
        @source.where(
          query_type: :group_by,
          range: @range,
          start_time: @start_time,
          end_time: @end_time,
          granularity: @granularity,
          group_by: [:client_mac, :client_latlong],
          search: @filter.get,
          direction: :ASCENDING,
          aggregation: :clients
        )
      end

      def map_query
        @source.where(
          query_type: :group_by,
          range: @range,
          start_time: @start_time,
          end_time: @end_time,
          granularity: :all,
          group_by: [:client_latlong],
          search: @filter.get,
          aggregation: :clients
        )
      end

      def movements_query
        @source.where(
          query_type: :group_by,
          range: @range,
          start_time: @start_time,
          end_time: @end_time,
          granularity: :all,
          group_by: [:old, :new],
          search: @filter.get,
          aggregation: :clients
        )
      end

      def tops_query
        @source.topN(
          query_type: :group_by,
          start_time: @start_time,
          end_time: @end_time,
          range: @range || :custom,
          granularity: :all,
          limit: (@load_more_offset || 1) * current_user.per_page_count,
          group_by: @source::TABS[@tab][:columns],
          search: @filter.get,
          aggregation: @aggregations
        )
      end

      def tops_query_to_widget
        @source.topN(
          query_type: :group_by,
          start_time: @start_time,
          end_time: @end_time,
          range: @range || :custom,
          granularity: :all,
          limit: (@load_more_offset || 1) * current_user.per_page_count,
          group_by: @source::TABS[@tab][:columns],
          search: @filter.get,
          aggregation: @aggregation
        )
      end

      def set_type_filter
        # Clean filters type first
        @filter.remove(:type)

        if %w[status client_mac client_mac_vendor calendar_item_name].include?(@tab.to_s)
          # filter by Floor at Status Tab as it's the most restrictive
          @filter.add(:type, 'floor')
        else
          @filter.add(:type, @tab.to_s)
        end
      end

      private

      def set_client_profile_filters
        case @client_profile
        when 'hard'
          @filter.add(:client_profile, 'hard')
        when 'medium'
          @filter.add(:client_profile, ['medium', 'hard'])
        when 'soft'
          # No hacer nada si es 'soft'
        end
      end
    end
  end
end
