#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Api::V1::CustomTabsController < Api::V1::BaseController

    before_action :set_custom_tab, only: [:update]

    skip_before_action :authenticate_user!, only: [:import, :index], raise: false
    skip_before_action :authenticate_user_from_token!, only: [:import, :index]
    #before_action :validate_cert, only: :import

    #
    # Get a list of access points.
    #
    def index

      # Get access points
      custom_tabs = CustomTab.where(custom_tab_params)

      data = []
      custom_tabs.each do |custom_tab|
        data << custom_tab.info
      end

      # Respond based on the format
      respond_to do |format|
        format.json do
          render json: {
            query: true,
            custom_tabs: data
          }
        end
        format.csv { render text: CustomTab.to_csv(data) }
      end
    end

     def destroy
      custom_tabs = params[:custom_tabs]
      custom_tabs = [custom_tabs] if custom_tabs.is_a? Hash
      error = false
      destroyed = 0
      custom_tabs.each do |custom_tab|
        if custom_tab[:name] || custom_tab[:id]
          custom_tab_to_delete = CustomTab.find_by(name: custom_tab[:name]) if custom_tab[:name]
          custom_tab_to_delete = CustomTab.find_by(id: custom_tab[:id]) if custom_tab[:id]
          if custom_tab_to_delete
            custom_tab_to_delete.destroy
            destroyed += 1
          end
        else
          error = true
          break
        end
      end
      # Check if there was an error with the request
      if error
        error_400 "#{I18n.t 'controllers.should_either'}"
      else
        render json: { query: true, destroyed_custom_tabs: destroyed }
      end
    end

    #
    # Create an Custom Tab
    #
    def create

      # Create the Custom Tab
      custom_tab = CustomTab.new(custom_tabs_params)
      if custom_tab.save
        respond_to do |format|
          format.json do
            render json: {
              query: true,
              custom_tab: custom_tab
            }
          end
        end
      else
        error_422("#{I18n.t 'controllers.custom_tab'}"\
                  "#{custom_tab.errors.full_messages.join(', ')}")
      end
    end

    #
    # Update an Custom Tab
    #
    def update
      para = custom_tabs_params

      if @custom_tab.update(para)
        respond_to do |format|
          format.json do
            render json: {
              query: true,
              custom_tab: @custom_tab
            }
          end
          # Render CSV
          format.csv { render text: @custom_tab.to_csv }
        end
      else
        error_422("#{I18n.t 'controllers.custom_tab2'}"\
                  "#{@custom_tab.errors.full_messages.join(', ')}")
      end
    rescue NoMethodError => e
      Rails.logger.error e.inspect
      Rails.logger.error e.backtrace.join("\n")
      return error_404("#{I18n.t 'controllers.custom_tab3'}")
    end

    #
    # Creates / Updates Custom Tabs
    #
    def import
      # Cache loop variables
      custom_tabs_json = (params[:custom_tabs] || [])
      Rails.logger.info "Processing: "
      Rails.logger.info custom_tabs_json
      time_now = Time.now

      error = false
      imported = 0
      # Updates or insert Received Custom Tabs
      Upsert.batch(CustomTab.connection, CustomTab.table_name) do |upsert|
        custom_tabs_json.each do |custom_tab|
          if %i[name enabled percentage category units].all? {|field| custom_tab.key? field.to_s}
            upsert.row({ name: custom_tab[:name] },
                      enabled: custom_tab[:enabled], percentage: custom_tab[:percentage],
                      category: custom_tab[:category], units: custom_tab[:units],
                      created_at: time_now, updated_at: time_now)
              imported += 1
          else
            error = true
            break
          end
        end
      end

      Rails.logger.debug "Processed #{custom_tabs_json.count} CustomTab in #{(Time.now - time_now).round(2)}s."

      # Check if there was an error with the request
      if error
        error_400 "#{I18n.t 'controllers.should_either3'}"
      else
        render json: { query: true, imported_custom_tabs: imported }
      end
    end

    private

    # Permitted parameters
    def custom_tab_params
      params.permit(:name, :enabled, :percentage, :category, :units)
    end

    #
    # Init the custom tab based on params[:id]. It can be an ID or Name
    #
    def set_custom_tab
      @custom_tab = if params[:id].include?(':')
              CustomTab.find_by_name(params[:id])
            else
              CustomTab.find_by_id(params[:id])
            end
      return error_404("#{I18n.t 'controllers.requested_ap'}") if @custom_tab.nil?
    end

  end
