#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ProfilesController < ApplicationController
  self.view_paths = 'lib/modules/scanner/app/views'
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_sensor
  before_action :check_sensor_privileges, only: [:new, :destroy, :clone]
  before_action :check_license, only: [:new]

  before_action :set_profile, only: [:edit, :update, :destroy, :clone]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  def index
    @profiles = if @sensor
                  @sensor.profiles.includes(:owner).order('created_at ASC')
                else
                  Profile.includes(:owner).order('created_at ASC')
                end
  end

  def new
    @profile = @sensor.owned_profiles.build()
    @owners = @profile.owner
              .self_and_ancestors_with_root
              .select { |s| current_user.can?(:manage, s) }.map { |s| [s.name, s.id] }
    render layout: false
  end

  def create
    @sensor = Sensor.find(params[:profile][:sensor_id])
    Rails.logger.info @sensor
    if current_user.cannot?(:manage, @sensor)
      redirect_to sensors_path,
                  alert: "#{I18n.t('controllers.no_privileges')}"
    end
    return error_404("#{I18n.t('controllers.sensor_no_license')}") unless @sensor.can_create_profile?

    unless params[:profile][:run_at].nil?
      transform_run_at
    end

    Rails.logger.info "\n\n"
    Rails.logger.info params
    Rails.logger.info profile_params
    # @profile = Profile.create(profile_params)
    #
    @profile = Profile.new
    @profile.update(profile_params)
    @owners = @profile.owner.self_and_ancestors_with_root
                      .select { |s| current_user.can?(:manage, s) }.map { |s| [s.name, s.id] }
    if @profile.save
      # Unless the periodicity is on demand, create the scan object
      @profile.create_scan unless @profile.on_demand?
      flash[:notice] = "Scanner Profile created."
    end

  end

  def edit
    @owners = @profile.owner.self_and_ancestors_with_root
              .select { |s| current_user.can?(:manage, s) }.map { |s| [s.name, s.id] }

    if current_user.cannot? :manage, @profile.owner
      redirect_to sensors_path,
                  alert: "#{I18n.t('controllers.no_privileges')}"
    end

    render layout: false
  end

  def update
    @sensor = @profile.owner

    if current_user.cannot? :manage, @sensor
      redirect_to sensors_path,
                  alert: "#{I18n.t('controllers.no_privileges')}"
    end

    unless params[:profile][:run_at].nil?
      transform_run_at
    end

    if @profile.update(profile_params)
      @profile.rebuild_scan
      flash[:notice] = "Scanner Profile updated."
    end
  end

  def search
    @search = params[:search]
    @profiles = if @sensor
                  @sensor.profiles.includes(:owner)
                  .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
                  .order('LOWER(name) ASC')
                else
                  Profile.includes(:owner)
                  .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
                  .order('LOWER(name) ASC')
                end
  end

  def destroy
    @profile.destroy

    redirect_to sensor_profiles_path(@sensor),
                notice: "Scan profile successfully deleted"
  end

  def clone
    new_profile = @profile.clone @sensor.id
    new_profile.save

    redirect_to sensor_profiles_path(@sensor)
  end

  def check_state
    profiles_status = {}
    profiles = Profile.where(sensor_id: params[:sensor_id]) #all profiles of one sensor
    profiles.each do |profile|
      state = "default"
      state = "cancelling" if profile.cancelling?
      state = "running" if profile.running?
      profiles_status.merge!({"#{profile.id}" => state})
    end
    render :json => profiles_status
  end

  private

  # Use callbacks to share common setup or constraints between actions.
  def set_sensor
    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
  end

  def set_profile
    @profile = Profile.find(params[:id])
  end

  def profile_params
    params.require(:profile)
      .permit(:sensor_id, :name, :description, :profile_type,
              :target_addr, :target_port, :periodicity, :time, :run_at
              )
  end

  #
  # Transform datepicker string format to datetime while periodicity is on demand or once
  # Transform user timezone to redBoder timezone while periodicity is schedule
  #
  def transform_run_at
    if params[:profile][:periodicity].to_i == 1
      Rails.logger.info Time.strptime(params[:profile][:run_at], '%m/%d/%Y %T %z').to_time
      params[:profile][:run_at] = Time.strptime(params[:profile][:run_at], '%m/%d/%Y %T %z').to_time
    end
  end

  #
  # Check sensor permissions or redirect with an error.
  #
  def check_sensor_privileges
    return if current_user.admin? && current_user.can?(:manage, @sensor)
    redirect_to sensors_path, alert: "#{I18n.t('controllers.no_privileges')}"
  end

  def check_license
    return error_404("#{I18n.t('controllers.sensor_no_license')}") unless @sensor.can_create_profile?
  end

end
