#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class Profile < ApplicationRecord
    include Redborder::Audit

    # ==========================================================
    # Custom attributes
    # ==========================================================
    PERIODICITY_RANGE = { 0 => "On Demand",
                          1 => "Once",
                          2 => "Daily",
                          3 => "Weekly",
                          4 => "Monthly" }


    PROFILE_TYPE =      { 0 => "Vulnerabilities Scan",
                          1 => "Host Discovery" }

    # ==========================================================
    # Relations
    # ==========================================================

    has_many :scans, dependent: :destroy
    has_many :sensors
    belongs_to :owner, class_name: 'Sensor', foreign_key: 'sensor_id'

    # ================================================================
    # Callbacks
    # ================================================================


    # ================================================================
    # Validations
    # ================================================================

    # IPv4 Regex
    @ipv4_regex = %r{\A(#{Resolv::IPv4::Regex256})\.
                        (#{Resolv::IPv4::Regex256})\.
                        (#{Resolv::IPv4::Regex256})\.
                        (#{Resolv::IPv4::Regex256})
                        (\/([0-9]|[0-2][0-9]?|3[0-2]))?\z}x

    # https://rubular.com/r/epioBhMWBa  :3
    @target_port_regex = %r{
      \A(((\d{1,5})(\s*)(-)(\s*)(\d{1,5})|(\d{1,5}))((\s*)(,)(\s*)((\d{1,5})(\s*)(-)(\s*)(\d{1,5})|(\d{1,5})))+|(\d{1,5})(\s*)(-)(\s*)(\d{1,5})(\s*)|(\d{1,5})(\s*))\z
      }x
      # ((\d{1,5})(\s*)(-)(\s*)(\d{1,5})|(\d{1,5}))((\s*)(,)(\s*)((\d{1,5})(\s*)(-)(\s*)(\d{1,5})|(\d{1,5})))+| #sequence of ranges/single    6,9-66,96-666 | 6,  9 - 66   ,      96- 666
      # (\d{1,5})(\s*)(-)(\s*)(\d{1,5})(\s*)|                                                                   #single range                 6-9 | 6-  9 | 6  -9
      # (\d{1,5})(\s*)                                                                                          #single port                  69  | 6 | 9
    @time_regex = %r{\A([0-1]?[0-9]|2[0-3]):[0-5][0-9]\z}

    validate :uniqueness_name_for_profiles
    validates :name, presence: true
    validates :target_addr, presence: true, format: { with: @ipv4_regex, message: "Invalid Target" }
    validates :target_port, allow_blank: true, format: { with: @target_port_regex, message: "Invalid Ports" }, length: {in: 0..1000000}
    validates :time, allow_blank: true, format: { with: @time_regex, message: "Invalid Time Format [HH:MM]" }

    audited only: [:name, :target_addr, :target_port, :description, :run_at, :time],
            on: [:create, :update, :destroy]

    # ================================================================
    # Public methods
    # ================================================================

    #
    # => Get name without blank spaces
    #
    def clean_name
      name.gsub(/\s+|\.+/, '_')
    end

      #
    # => Clones the current policy on the specified sensor with a new name
    #
    def clone(to_sensor_id)
      matched = /(.*) \((\d+)\)$/.match(name)
      current_name = matched.nil? ? name : matched[1]
      current_number = matched.nil? ? 0 : matched[2].to_i

      loop do
        current_number += 1
        clone_name = "#{current_name} (#{current_number})"

        next if Profile.find_by(name: clone_name)

        created = Profile.create(name: clone_name, description: description,
                                sensor_id: to_sensor_id,
                                target_addr: target_addr, target_port: target_port,
                                 periodicity: periodicity, time: time, run_at: run_at
                                )
        if created.save
          unless created.on_demand?
            created.create_scan
          end
          return created
        end
        # return created if created.save
      end
    end

    def running?   #if any scan in scans is running
      scans.any?{|scan| scan.running?}
    end

    def cancelling?
      scans.any?{|scan| scan.cancelling?}
    end

    def on_queue?
      scans.any?{|scan| scan.on_queue?}
    end

    def profile_type_humanize
      (PROFILE_TYPE.key?profile_type) ? PROFILE_TYPE[profile_type] : PROFILE_TYPE.values.first
    end

    def vulnerability_type?
      profile_type_humanize == "Vulnerabilities Scan" ? true : false
    end

    def hostdiscovery_type?
      profile_type_humanize == "Host Discovery" ? true : false
    end

    def periodicity_humanize
      (PERIODICITY_RANGE.key?periodicity) ? PERIODICITY_RANGE[periodicity] : PERIODICITY_RANGE.values.first
    end

    def on_demand?
      periodicity_humanize == "On Demand" ? true : false
    end

    def once?
      periodicity_humanize == "Once" ? true : false
    end

    def daily?
      periodicity_humanize == "Daily" ? true : false
    end

    def weekly?
      periodicity_humanize == "Weekly" ? true : false
    end

    def monthly?
      periodicity_humanize == "Monthly" ? true : false
    end

    def rebuild_scan
      if on_queue?
        scan = Scan.find_by(profile_id: self.id, status: "on queue")
        scan.destroy
      end
      unless self.on_demand?
        self.create_scan
      end
    end

    def create_scan
      scan = Scan.new
      scan.profile = self

      if self.on_demand?
        scan.running!
      elsif self.once?
        scan.on_queue!
        scan.run_at = run_at
        self.run_at = run_at
      elsif self.daily?
        scan.run_at = time < Time.now ? Time.parse(time) + 1.day : Time.parse(time)
        scan.on_queue!
        self.run_at = scan.run_at
      elsif self.weekly?
        scan.run_at = time < Time.now ? Time.parse(time) + 1.week : Time.parse(time)
        scan.on_queue!
        self.run_at = scan.run_at
      elsif self.monthly?
        scan.run_at = time < Time.now ? Time.parse(time) + 1.month : Time.parse(time)
        scan.on_queue!
        self.run_at = scan.run_at
      end
      self.save
      scan.save
    end

    def has_scans?
      return Scan.where(profile_id: self.id) != nil ? true : false
    end

    # ================================================================
    # Protected methods
    # ================================================================

    protected


    # ================================================================
    # Private methods
    # ================================================================

    private

    #
    # => Validate the uniqueness of policy's name
    #
    def uniqueness_name_for_profiles
      return unless (new_record? || name_changed?) &&
                    Profile.where(sensor_id: sensor_id, name: name).count > 0
      errors.add(:name, "Name was already taken")
    end

  end
