#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Scanner DataSource. It takes the data from scanner traffic.
#
class Scanner < Redborder::DataSource
  COLUMNS = Redborder::Module.get(:scanner, :columns)
  SESSION_ID = Redborder::Module.get(:scanner, :session_id).to_sym
  AGGREGATIONS = Redborder::Module.get(:scanner, :aggregations)

  # Generate a reader attribute for the columns
  COLUMNS.map{ |k, v| v[:columns].keys + (v[:complementary_columns].try('keys') || []) }.flatten.each do |column|
    attr_reader column
  end

  # Define the column names to filter
  COLUMN_NAMES = { start_time: "#{I18n.t 'scanner2.start_time'}",
                   end_time: "#{I18n.t 'scanner2.end_time'}",
                   domain: "#{I18n.t 'scanner2.domain'}"
                 }
  COLUMNS.each { |k, v| COLUMN_NAMES[k] = v[:columns][k] }

  # Aggregations
  AGGREGATIONS.each do |aggregation, _aggregation_data|
    attr_reader aggregation
  end

  attr_reader :timestamp, :result

  #
  # Create a Scanner object with the data returned from Druid.
  #
  # == Parameters:
  # args::
  #   Hash with the data to build a Scanner model.
  #
  # == Returns:
  # Scanner object
  #
  def initialize(_user, args = {})
    if args.is_a?(Hash)
      args.each do |key, value|
        instance_variable_set("@#{key}", value)
      end

      @timestamp = args['timestamp'].to_time.in_time_zone if args['timestamp'].present?
    end

    @hostname = NetObject.fetch(@ipv4, _user.id, @ipv4)
    @vendor = get_mac_vendor_full(@mac)

    # Round the aggregations that must be rounded
    AGGREGATIONS.each do |aggregation, aggregation_data|
      current_value = instance_variable_get("@#{aggregation}")
      next if current_value.nil?
      if aggregation_data[:round]
        instance_variable_set("@#{aggregation}", current_value.round)
      else
        instance_variable_set("@#{aggregation}", current_value.round(3))
      end
    end
  end

  #
  # Return the Symbol of id module
  #
  def self.id
    :scanner
  end

  #
  # Return the Symbol of the session of module
  #
  def self.session
    SESSION_ID
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.controller
    :scanner
  end

  #
  # Return the a hash with default action to redirect
  #
  def self.redirect
    { controller: controller, action: 'tops' }
  end

  #
  # Return the Symbol identifies the controller
  #
  def self.default_aggregation
    :events
  end

  #
  # Return default columns for RAW
  #
  def self.default_columns_raw
    [:pri, :hostname]
  end

  #
  # Return all scanner sensors
  #
  # == Parameters:
  # name::
  #   Array or String of names. If its nil, return all flow sensors.
  #
  def self.sensors(name = nil)
    return Sensor.where(type: Sensor::SCANNER) if name.nil?
    # No name provided
    Sensor.where(type: Sensor::SCANNER, name: name)
  end

  # =================================================
  # FILTERS
  # =================================================

  #
  # Add the columns and domains to build filter form
  # It's overrides the filter_to_json from filtrable
  #
  # == Parameters:
  # user::
  #   User who get filters
  # json::
  #   If true, return filters as JSON. True by default.
  #
  # == Returns:
  # A Hash with the basic operators and the domains and columns
  # to build the filter form
  #
  def self.filter_to_json(user = User.current_user, json = true)
    domains = user
              .sensor
              .self_and_descendants
              .where.not(id: 1)
              .where(type: Sensor::DOMAIN)
    
    organizations = user
                    .sensors_with_permission
                    .where.not(id: 1)
                    .where(domain_type: 2)

    f = super.merge!(
      scanner_columns: COLUMN_NAMES.to_a.map do |c|
        { value: c[1],
          id: c[0],
          type: ([:start_time, :end_time].include?(c[0].to_sym) ? :datetime : :text_input) }
      end.compact.sort_by do |h|
        h[:value].to_s
      end,
      scanner_sensors: {
        type: :dropdown,
        value: user
          .sensor
          .self_and_descendants
          .where.not(id: 1)
          .where(type: [Sensor::SCANNER])
          .map { |c| { value: c.name, id: c.name } }
      },
      domains: {
        type: :dropdown,
        value: domains.map { |c| { value: c.path_name, id: c.id } }
      },
      organizations: {
        type: :dropdown,
        value: organizations.map { |c| { value: c.name, id: c.name } }
      }
    )
    # Check format
    json ? f.to_json : f
  end

  #
  # Return the filters with zones
  #
  # == Parameters:
  # search::
  #   A hash with the current filters to apply
  # user::
  #   Current user
  #
  # == Returns:
  # A hash with the filters combination of current, user, domain and zones.
  #
  def self.user_filters(search, user)
    return {} if user.nil? || search.nil?
    filters = {}
    # Filters from user
    # TODO: Don't except zones
    # filters.merge! search.except(:domain, :zone)
    filters.merge! search.except(:domain)
    # Filters from domains
    filters.merge! domain_filters(search[:domain], search[:sensor_name], user)
    # Filters from zones
    filters
  end
end
