#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorScanner
    # Id of the sensor
    SCANNER = 17

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [SCANNER]

        # Import values for SCANNER
        self.import_columns_for[SCANNER] = [:license_id].freeze

       

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------

        DittoCode::Exec.if 'ENTERPRISE' do
        validates_presence_of :license_id, if: proc { type == SCANNER }
        validate :empty_license_buckets_for_scanner, if: proc { type == SCANNER }
        end

        has_many :owned_profiles, class_name: "Profile", foreign_key: "sensor_id", dependent: :destroy

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :scanner

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        after_create :create_cache_scanner, if: proc { type == SCANNER }
        after_create :create_chef_node_scanner,
                     if: proc { type == SCANNER }
        after_save :data_to_client_proxy_scanner,
                   if: proc { type == SCANNER && real_parent_id }
        before_destroy :delete_data_from_client_proxy_scanner,
                       if: proc { type == SCANNER && real_parent_id }
        before_destroy :restart_profiles
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{SCANNER}") do
        'sensors/scanner_sensor'
      end
    end

    module InstanceMethods

      def is_scanner?
        type == SCANNER
      end

      # Return the descendant scanner sensors of the domain
      def get_scanner
        self.all_children(true).where(type: Sensor::SCANNER)
      end

      #Should check is nmap is installed in related machine
      def nmap_installed?
        parent_sensor = Sensor.find(parent_id)
        return parent_sensor.is_client_proxy? ? parent_sensor.chef_node["redborder-scanner"]["is_nmap_installed"] : Manager.nmap_installed?     
        #TODO: check nmap_installed in the related manager   
      end

      # Return last time when sensor was checked
      define_method("last_check_in_#{SCANNER}") do
        Time.at(check_in_time) if check_in_time
      end

      define_method("icon_#{SCANNER}") do
        'icons/scanner.png'
      end

      define_method("full_name_#{SCANNER}") do
        name
      end

      define_method("secondary_name_#{SCANNER}") do
        name
      end

      define_method("complete_name_#{SCANNER}") do
        name
      end

      define_method("type_to_s_#{SCANNER}") do
        'scanner'
      end

      define_method("type_humanize_#{SCANNER}") do
        'SCANNER'
      end

      define_method("chef_node_#{SCANNER}") do
        Chef::Node.load("rbscanner-#{id}") rescue nil
      end

      define_method("chef_client_#{SCANNER}") do
        Chef::ApiClient.load("rbscanner-#{id}")
      end

      DittoCode::Exec.if 'ENTERPRISE' do
        def can_run_profile?
          if is_sensor?
            is_scanner? ? license.try(:valid?) : false
          else
            is_scanner? ? true : false
          end
        end

        def can_create_profile?
          if is_sensor?
            license.try(:valid?)
          else
            true 
          end
        end

      end

      DittoCode::Exec.if 'COMMUNITY' do
        def can_run_profile?
          is_scanner? ? true : false
        end
        def can_create_profile?
            true 
        end
      end

      def profiles
        sensor_ids = [1, self.id] + ancestors.map(&:id)
        Profile.where(sensor_id: sensor_ids)
      end

      #
      # Return the pending running jobs
      # 
      def running_and_cancelling_scans
        scans = []
        check_on_queue
        profiles.each{ |profile| scans += profile.scans.select{ |s| s.running? or s.cancelling? } }
        return scans
      end

      private

      def check_on_queue
        scans = []
        profiles.each{ |profile| scans += profile.scans.select{ |s| s.on_queue? } }
        scans.each { |scan| scan.running! if scan.run_at <= Time.now }

        Rails.logger.info "\n\nRetrieving update list"
        Rails.logger.info scans
      end

      # Check if license can be associated to current sensor
      #
      def empty_license_buckets_for_scanner
        return unless license
        return if license_id_was == license_id
        errors.add(:license_id, "#{I18n.t 'lib_scanner.full'}") if license.in_use_scanner >= license.limit_scanner
      end

      def create_cache_scanner
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", name
      end

      def update_cache_scanner
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", name
      end

      def create_chef_node_scanner
        node_json = {
          'name' => "rbscanner-#{id}",
          'run_list' => ['role[scanner-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'redBorder' => {
              'sensor_uuid' => uuid,
              'monitors' => [],
              'parent_id' => real_parent_id
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end


      #
      # Return the sensor as json with variables from SCANNER
      #
      def as_json_with_scanner(options = {})
        result = as_json_without_scanner(options)
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{SCANNER}") do
        update_chef_domain_values_in_node
      end

      # Create the chef role for SCANNER sensors
      #
      define_method "create_chef_role_#{SCANNER}" do
        run_list = ['role[scanner-sensor]', "role[#{parent.chef_name}]"]
        scanner_attributes = {
          'redBorder' => {
            'rsyslog' => {},
            'proxy' => {},
            'snmp' => { 'trap_server' => [] },
            'monitors' => []
          }
        }
        builder = Redborder::Chef::Role::Builder.new(
                    self, run_list, override_attributes: scanner_attributes
                  )
        builder.create
      end

      define_method "destroy_chef_node_#{SCANNER}" do
        chef_node.destroy
      end

      # Store the current data of the Chef node in the client_proxy parent
      #
      def data_to_client_proxy_scanner
        # Get the role of the parent
        role = real_parent.chef_role

        # Update attributes
        role.override_attributes['redBorder']['sensors_mapping'] ||= {}
        role.override_attributes['redBorder']['sensors_mapping']['scanner'] ||= {}
        role.override_attributes['redBorder']['sensors_mapping']['scanner'][name] =
          chef_node['redBorder']
        # Save it
        role.save
      end

      # Destroy references to owned profiles
      def restart_profiles
        Sensor.where(profile_id: owned_profiles.map(&:id)).update_all(profile_id: nil)
      end

      def delete_data_from_client_proxy_scanner
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redBorder']['sensors_mapping']['scanner'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorScanner)
