#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  #
  # Enrich with location
  #
  module SensorVault
    # Id of the sensor
    VAULT = 10

    def self.included(base)
      base.extend(ClassMethods)
      base.send(:include, InstanceMethods)

      base.class_eval do
        # Increase available list of sensors
        self.available_sensors += [VAULT]

        # Import values for VAULT
        self.import_columns_for[VAULT] = [:ip, :license_id].freeze

        # -------------------------------------------------------------
        # Validations
        # -------------------------------------------------------------
        validates :ip, presence: true,
                    #   uniqueness: { scope: :real_parent_id },
                       if: proc { type == VAULT }

        validate :uniqueness_ip, if: proc { type == VAULT }

        DittoCode::Exec.if 'ENTERPRISE' do
        validates_presence_of :license_id, if: proc { type == VAULT }
        validate :empty_license_buckets_for_vault, if: proc { type == VAULT }
        end

        # -------------------------------------------------------------
        # JSON serialization
        # -------------------------------------------------------------

        alias_method_chain :as_json, :vault

        # -------------------------------------------------------------
        # Before and after methods
        # -------------------------------------------------------------
        before_update :update_cache_vault,
                      if: proc { type == VAULT && ip_changed? }
        after_create :create_cache_vault, if: proc { type == VAULT }
        after_create :create_chef_node_vault,
                     if: proc { type == VAULT }
        after_update :update_ip_vault, if: proc { type == VAULT }
        after_save :data_to_client_proxy_vault,
                   if: proc { type == VAULT && real_parent_id }
        before_destroy :delete_data_from_client_proxy_vault,
                       if: proc { type == VAULT && real_parent_id }
      end
    end

    module ClassMethods
      define_method("to_partial_path_#{VAULT}") do
        'sensors/vault_sensor'
      end
    end

    module InstanceMethods

      def is_vault?
        type == VAULT
      end

      # Return the descendant vault sensors of the domain
      def get_vault
        self.all_children(true).where(type: Sensor::VAULT)
      end

      # Return last time when sensor was checked
      define_method("last_check_in_#{VAULT}") do
        Time.at(check_in_time) if check_in_time
      end

      define_method("icon_#{VAULT}") do
        'icons/vault.png'
      end

      define_method("full_name_#{VAULT}") do
        name
      end

      define_method("secondary_name_#{VAULT}") do
        name
      end

      define_method("complete_name_#{VAULT}") do
        name
      end

      define_method("type_to_s_#{VAULT}") do
        'vault'
      end

      define_method("type_humanize_#{VAULT}") do
        'VAULT'
      end

      define_method("chef_node_#{VAULT}") do
        Chef::Node.load("rbvault-#{id}") rescue nil
      end

      define_method("chef_client_#{VAULT}") do
        Chef::ApiClient.load("rbvault-#{id}")
      end

      private

      def uniqueness_ip
        sensors = Sensor.where(type: type, ip: ip, real_parent_id: real_parent_id)
        sensors = sensors.where.not(id: id) if id
        if sensors.any?
          errors.add(:ip, "#{I18n.t 'lib_vault.unique'}")
        end
      end

      # Check if license can be associated to current sensor
      #
      def empty_license_buckets_for_vault
        return unless license
        return if license_id_was == license_id
        errors.add(:license_id, "#{I18n.t 'lib_vault.full'}") if license.in_use_vault >= license.limit_vault
      end

      def create_cache_vault
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
      end

      def update_cache_vault
        Rails.cache.write "rb-webui:sensor:#{type}:#{name}", ip
      end

      def create_chef_node_vault
        node_json = {
          'name' => "rbvault-#{id}",
          'run_list' => ['role[vault-sensor]', "role[#{chef_name}]"],
          'attributes' => {
            'rbname' => name,
            'ipaddress' => ip,
            'redborder' => {
              'sensor_uuid' => uuid,
              'ipaddress' => ip,
              'monitors' => [],
              'parent_id' => real_parent_id
            }
          }
        }

        node = Chef::Node.from_hash(node_json)
        node.create
      end

      def update_ip_vault
        role = chef_role
        node = chef_node
        role.override_attributes['ipaddress'] = ip
        node.normal['ipaddress'] = ip
        node.normal['redborder']['ipaddress'] = ip
        role.save
        node.save
      end

      #
      # Return the sensor as json with variables from VAULT
      #
      def as_json_with_vault(options = {})
        result = as_json_without_vault(options)
        result.merge!({ ip: ip }.as_json) if self.is_vault?
        result
      end

      # Method used to complete values about ancestors on chef node
      define_method("update_chef_domain_values_#{VAULT}") do
        update_chef_domain_values_in_node
      end

      # Create the chef role for VAULT sensors
      #
      define_method "create_chef_role_#{VAULT}" do
        run_list = ['role[vault-sensor]', "role[#{parent.chef_name}]"]
        vault_attributes = {
          'redborder' => {
            'rsyslog' => {},
            'proxy' => {},
            'snmp' => { 'trap_server' => [] },
            'monitors' => []
          }
        }
        builder = Redborder::Chef::Role::Builder.new(
                    self, run_list, override_attributes: vault_attributes
                  )
        builder.create
      end

      define_method "destroy_chef_node_#{VAULT}" do
        chef_node.destroy
      end

      # Store the current data of the Chef node in the client_proxy parent
      #
      def data_to_client_proxy_vault
        # Get the role of the parent
        role = real_parent.chef_role

        # Update attributes
        role.override_attributes['redborder']['sensors_mapping'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['vault'] ||= {}
        role.override_attributes['redborder']['sensors_mapping']['vault'][name] =
          chef_node['redborder']
        # Save it
        role.save
      end

      def delete_data_from_client_proxy_vault
        client_proxy = real_parent
        role = client_proxy.chef_role

        role.override_attributes['redborder']['sensors_mapping']['vault'].delete(name)
        role.save
      end
    end
  end
end

Sensor.send(:include, RedBorder::SensorVault)
