#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  module SensorsControllerVault
    def self.included(base)

      base.class_eval do
        unloadable

        # -------------------------------------------------------------
        # Before and after actions
        # -------------------------------------------------------------

        before_action :privileges_on_top_domain_vault,
                      only: [:create_vault, :new_vault]
        before_action :set_parent_id_vault, only: :new_vault
        before_action :set_vault_sensors_per_page,
                      only: [:vault, :vault_load_more]
        before_action :set_default_sensor_view_vault, only: :vault

        # -------------------------------------------------------------
        # Methods
        # -------------------------------------------------------------

        def vault
          # The current page
          @page = 0

          # The current letter
          @current_letter = ''

          if params[:id].nil?
            sensor_ids = Sensor.where(type: Sensor::VAULT, is_deleted: false)
                         .select{|s| current_user.can?(:read, s)}.map(&:id)
          else
            @sensor = Sensor.find(params[:id])
            sensor_ids = @sensor.all_children
                         .where(type: Sensor::VAULT, is_deleted: false)
                         .select{|s| current_user.can?(:read, s)}.map(&:id)
          end

          # Use unscoped to remove the ordering setted in sensor.rb
          @sensors = Sensor.unscoped
                     .where(id: sensor_ids)
                     .order('LOWER(name) ASC')
                     .take(@sensors_page)
        end

        # -------------------------------------------------------------
        # Sensors methods
        # -------------------------------------------------------------

        # New method
        # --------------------
        def new_vault
          @sensor = Sensor
                    .new(type: Sensor::VAULT,
                         parent_id: @parent_id)
          render layout: false
        end

        # Destroy method
        # --------------------
        define_method("destroy_#{Sensor::VAULT}") do
          @sensor.update_column(:is_deleted, true)
          Delayed::Job.enqueue(
            Redborder::Jobs::DeleteSensorJob.new(
              @sensor.id, current_user.id, true, I18n.locale),
            priority: 10, run_at: Time.now)

          respond_to do |format|
            format.html { redirect_to sensors_url }
            format.json { render json: { query: true, id: @sensor.id } }
            format.js
          end
        end

        # -------------------------------------------------------------
        # End Sensors methods
        # -------------------------------------------------------------

        #
        # => Search a vault sensor based on the first letter of its name
        #
        def vault_letter_search
          # Get the letter
          @letter = params[:id]

          # Check the letter
          if [*"A".."Z"].include? @letter
            sensor_ids = if params[:sensor_id].nil?
              Sensor.where(type: Sensor::VAULT, is_deleted: false)
                    .where('LOWER(name) LIKE ?', "#{@letter.downcase}%")
                    .select{|s| current_user.can?(:read, s)}.map(&:id)
            else
              @sensor = Sensor.find(params[:sensor_id])
              @sensor.all_children.where(type: Sensor::VAULT, is_deleted: false)
                     .where('LOWER(name) LIKE ?', "#{@letter.downcase}%")
                     .select{|s| current_user.can?(:read, s)}.map(&:id)
            end
            @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
            render :vault_search
          else
            # Something went wrong
            redirect_to vault_sensors_path
          end

        end

        def vault_search
          @search = params[:search]
          sensor_ids = if params[:sensor_id].nil?
            Sensor.where(type: Sensor::VAULT, is_deleted: false)
                  .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
                  .select{|s| current_user.can?(:read, s)}.map(&:id)
          else
            @sensor = Sensor.find(params[:sensor_id])
            @sensor.all_children.where(type: Sensor::VAULT, is_deleted: false)
                   .where('LOWER(name) LIKE ?', "%#{@search.downcase.split.join('%')}%")
                   .select{|s| current_user.can?(:read, s)}.map(&:id)
          end
          @sensors = Sensor.where(id: sensor_ids).order('LOWER(name) ASC')
        end

        def vault_load_more

          # Get the page
          @page = params[:page].to_i

          # Get the sensors
          if params[:sensor_id].nil?
            sensor_ids = Sensor.where(type: Sensor::VAULT, is_deleted: false)
                         .select{|s| current_user.can?(:read, s)}.map(&:id)
          else
            @sensor = Sensor.find(params[:sensor_id])
            sensor_ids = @sensor.all_children
                         .where(type: Sensor::VAULT, is_deleted: false)
                         .select{|s| current_user.can?(:read, s)}.map(&:id)
          end


          @sensors = Sensor.unscoped.where(id: sensor_ids)
                           .order('LOWER(name) ASC')
                           .offset(@page * @sensors_page)
                           .take(@sensors_page)

          # Set the current letter empty
          before_sensor = Sensor.unscoped.where(id: sensor_ids).order(:name)
                                .offset((@page * @sensors_page) - 1).first

          if before_sensor.nil?
            @current_letter = ''
          else
            @current_letter = before_sensor.secondary_name[0,1].upcase
          end

          respond_to do |format|
           format.js {}
           format.json { render json: @sensors }
          end

        end

        # POST /sensors/create_vault
        def create_vault
          @sensor = Sensor.new(send("new_sensor_params_#{Sensor::VAULT}"))

          if params[:sensor][:parent_id].blank? && current_user.super_admin?
            @parent_id = @sensor.parent_id = Sensor.root.id
          elsif params[:sensor][:parent_id].blank?
            @parent_id = @sensor.parent_id = current_user.sensor.id
          else
            @sensor.parent_id = params[:sensor][:parent_id]
            @parent_id = params[:sensor][:parent_id]
          end

          @sensor.real_parent_id = @parent_id if @sensor.parent.is_client_proxy?

          @permission = true

          respond_to do |format|
            if current_user.cannot?(:manage, @sensor.parent)
              @permission = false
              # No permissions on parent.
              format.html do
                redirect_to sensors_path,
                            alert: "#{I18n.t 'lib_vault.privileges'}"
              end
              format.json do
                render json: { error: "#{I18n.t 'lib_vault.privileges'}" },
                       status: :unauthorized
              end
              format.js do
                flash[:error] = "#{I18n.t 'lib_vault.privileges2'}"
              end
            elsif @sensor.save
              # You need to reload the user sensor rgt value before rendering menus
              # in order to set manage privileges for the current user
              current_user.reload
              Delayed::Job.enqueue(Redborder::Jobs::RbWakeupChefClusterJob.new(true, current_user.id), priority: 5) unless DelayedJob.rb_wakeup_chef_cluster?
              format.html { redirect_to sensors_path, notice: "#{I18n.t 'lib_vault.created'}" }
              format.json { head :no_content }
              format.js
            else
              format.html { render action: 'new_vault' }
              format.json { render json: @sensor.errors, status: :unprocessable_entity }
              format.js
            end
          end
        end

        private

        def set_default_sensor_view_vault
          session[:sensor_view] = params[:action].to_sym
        end

        define_method("sensor_params_#{Sensor::VAULT}") do
          params
            .require(:sensor)
            .permit(:parent_id, :ip, :general_description, :license_id)
        end

        define_method("new_sensor_params_#{Sensor::VAULT}") do
          params
            .require(:sensor)
            .permit(:type, :name, :ip, :parent_id, :license_id)
        end

        def set_parent_id_vault
          set_parent_id
        end

        def privileges_on_top_domain_vault
          privileges_on_top_domain
        end

        def set_vault_sensors_per_page
          @sensors_page = Sensor::SENSORS_PAGE
        end
      end
    end
  end
end
