#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'syslog'
require 'json'

module Redborder
  # This class notificate an alarm to the users using several methods. Now,
  # available methods are:
  #   - Syslog
  #   - GCM push
  #   - Email
  #
  class AlarmNotificator
    attr_reader :alarm, :result, :type

    # Receive the alarm to notify and the result of query
    #
    # @param alarm [Alarm] Alarm to notify
    # @param result [Hash] Result of the query based on alarm conditions
    #
    def initialize(alarm, result)
      @alarm = alarm
      @result = result
      # We can define an alarm with an upper, a bottom or both limits. In the case we
      # define two limits, the alarm will be fired only by one of those limits.
      # That the reason I set here the "type" of the alarm.
      @type = @result[:where].to_s
    end

    # Send a notification based on the options from alarm
    #
    def notify
      notify_email if @alarm.notify_to_email
      notify_syslog if @alarm.notify_to_syslog
      notify_slack if @alarm.notify_to_slack
      notify_vault if @alarm.incident_enabled && @alarm.incident && @alarm.sensor
    end

    # Send a notificaiton using the email
    #
    def notify_email
      AlarmMailer.alarm(alarm, result).deliver_now
    end

    # Send a notificaiton using syslog
    #
    def notify_syslog
      ::Syslog.open('redBorder') do |s|
        s.alert(
          "Alarm %s fired because it exceeded the #{@type} limit of %s %s with %s %s in %s <-> %s",
          alarm.name, limit.to_s, units, value, units, result[:start_time].to_s,
          result[:end_time].to_s
        )
      end
    end

    def notify_slack
      require "uri"
      require "json"
      require "openssl"
      require "net/http"
      if @alarm
        begin
          uri = URI(@alarm.slack_webhook)
          http = Net::HTTP.new(uri.host, uri.port)
          http.use_ssl = true
          http.verify_mode = OpenSSL::SSL::VERIFY_PEER
          request = Net::HTTP::Post.new(uri.request_uri)
          request.add_field('Content-Type', 'application/json')
          request.body = { text:"[#{Setting.company}] Alarm #{@alarm.name} #{@alarm.product_type} Severity #{@alarm.severity.name} fired! Domain Alarm Triggered: #{Setting.webpage_url} " }.to_json
          response = http.request(request)

        rescue Exception => e
          raise "#{Time.now} [ERROR] Error on Slack request: #{e}"
        end
      end
    end

    def notify_vault
      return unless @alarm.incident_enabled

      return unless @alarm.incident

      return unless @alarm.sensor

      begin
        syslog_message = Rfc5424SyslogMessage.new
        syslog_message.app_name = 'alarmsjobs'
        syslog_message.msg_id = 'alarm-triggered'
        syslog_message.structured_data.push(alarm_event_structured_data)
        syslog_message.severity = severity_alarms_to_rfc5424_severity(@alarm.severity)

        syslog_message.msg = "Alarm #{@alarm.name} fired because it exceeded the #{@type} limit of #{limit.to_s}#{units} with #{value}#{units} between #{result[:start_time].to_s} and #{result[:end_time].to_s}"
        syslog_message.msg += " with conditions: #{humanize_conditions(@alarm.condition).to_s}" if @alarm.condition && !@alarm.condition.empty?

        params_syslog = { :syslog_server => '127.0.0.1' }
        SyslogService.send_to_syslog(syslog_message.to_s, params_syslog)

      rescue => e
        Rails.logger.error "Error while sending Syslog alert: #{e.message}"
      end
    rescue => e
      Rails.logger.error "Error in notify_vault: #{e.message} - Backtrace: #{e.backtrace.join("\n")}"
    end

    def humanize_conditions(conditions)
      conditions_text = ''
      conditions = JSON.parse(conditions, symbolize_names: true)

      conditions.each do |key, value|
        column = value[:column]
        operator = value[:operator]
        condition_value = value[:value]
        enabled = value[:enabled]

        if column && operator && condition_value && enabled
          conditions_text += "#{column} #{operator} #{condition_value}, "
        end
      end

      conditions_text.gsub(/, $/, '')
    end

    def alarm_event_structured_data
      sd_id = "alarm-event@#{RedBorder::PEN}"

      fields = {
        "incident-uuid" => @alarm.incident.uuid,
        "incident-uri" => @alarm.incident.uri,
        "incident-domain" => @alarm.incident.domain,
        "sensor-name" => @alarm.sensor.name,
        "sensor-uuid" => @alarm.sensor.uuid,
        "alarm-id" => @alarm.id,
        "alarm-user" => @alarm.user.name,
        "alarm-name" => @alarm.name,
        "alarm-product-type" => @alarm.product_type,
        "alarm-condition" => @alarm.condition.gsub('"', '\"'),
        "alarm-severity" => @alarm.severity.name
      }

      service_provider = @alarm.sensor.self_and_ancestors.find_by(domain_type: Sensor::SERVICE_PROVIDER)
      deployment       = @alarm.sensor.self_and_ancestors.find_by(domain_type: Sensor::DEPLOYMENT)
      namespace        = @alarm.sensor.self_and_ancestors.find_by(domain_type: Sensor::NAMESPACE)
      organization     = @alarm.sensor.self_and_ancestors.find_by(domain_type: Sensor::ORGANIZATION)
      market           = @alarm.sensor.self_and_ancestors.find_by(domain_type: Sensor::MARKET)

      %w(service_provider deployment namespace organization market).each do |v|
        fields["#{v}"] = eval(v).try('name')
        fields["#{v}_uuid"] = eval(v).try('uuid')
      end

      "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
    end

    # Map severiy alarm value with tradictional syslog priority and severity of rfc5424
    # Severity RFC 5424                                           | Severiy Alarms
    # 0       Emergency: system is unusable                       |
    # 1       Alert: action must be taken immediately             |
    # 2       Critical: critical conditions - Critical            | 14     Critical
    # 3       Error: error conditions - Very High                 | 13     Very high
    # 4       Warning: warning conditions - High                  | 1      High
    # 5       Notice: normal but significant condition - Medium   | 2      Medium
    # 6       Informational: informational messages  - Low        | 3      Low
    # 7       Debug: debug-level messages - Very Low              | 4      Very low
    def severity_alarms_to_rfc5424_severity(severity)
      return 6 unless severity.respond_to?(:sig_id)

      case severity.sig_id
      when 4 # Very Low
        7 # Debug
      when 3 # Low
        6 # Informational
      when 2 # Medium
        5 # Notice
      when 1 # High
        4 # Warning
      when 13 # Very high
        3 # Error
      when 14 # Critical
        2 # Critical
      else
        6 # Default: Informational
      end
    end

    private

    # Get the value of the result
    #
    def value
      @value ||= (result[:value] / (10.0**alarm.send("multiplier_#{@type}"))).round(2)
    end

    # Return the alarm unit
    #
    def units
      @units ||= alarm.units("multiplier_#{@type}".to_sym)
    end

    # Return the limit of the alarm
    #
    def limit
      @limit ||= alarm.send("#{@type}_limit")
    end

    # Return the severity color of the alarm
    #
    def severity_colour
      @alarm.severity.try(:bg_color) || '#777777'
    end

    # Build the message of the alarm
    #
    def mobile_message
      # Base message
      message = {
        name: @alarm.name,
        start_time: @result[:start_time].to_i,
        end_time: @result[:end_time].to_i,
        type: @type,
        value: value,
        server_key: Setting.server_key,
        severity_colour: severity_colour
      }

      # Merge limits
      message.merge(
        "#{@type}_limit".to_sym => limit,
        "#{@type}_units".to_sym => units
      )
    end
  end
end
