#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder #:nodoc:
  #
  # Extend ActiveRecord models to easy cache of values.
  #
  module Cacheable
    attr_accessor :cache_value, :cache_key, :cache_namespace, :cache_expiration
    #
    # This method define the model as cacheable. Several options as allowed.
    #
    # == Parameters:
    # options::
    #   (Optional) Hash with multiple optional options to customize the cache
    #     - Symbol with the key to isolate. By default: nil
    #     - value: Symbol with the value to store. By default: :value
    #     - key: Symbol with the key to act as identifier. By default: :id
    #     - namespace: An string to prefix cache
    #     - expiration: by default, 1.hour
    #
    def cacheable(options = {})
      cache_parse_options! options

      # Include the cache methods
      include CacheableMethods

      # Set callbacks
      cacheable_callbacks!
    end

    #
    # Define the callbacks to create/destroy cache
    #
    def cacheable_callbacks!
      after_create :create_cache
      before_update :update_cache
      before_destroy :delete_cache
    end

    def cache_default_options
      {
        isolate: nil,
        value: :value,
        key: :id,
        namespace: '',
        expiration: 1.hour
      }.freeze
    end

    def cache_parse_options!(options)
      options = cache_default_options.merge(options)

      class_attribute :cache_options
      self.cache_options = options
    end

    # ------------------------------------------------
    # Get method to load cache or create it
    # ------------------------------------------------

    #
    # Try to get this model by cache or create the cache
    #
    # == Parameters:
    # key::
    #   String with the value for the key to search.
    # isolation::
    #   (Optional) String or integer with the value to isolate.
    # default::
    #   (Optional) Default value to sent back if the key doesn't exist
    #   or key is nil
    #
    # == Returns:
    # The value stored on cache
    #
    def fetch(key, isolation = nil, default = nil)
      return default if key.nil?
      if isolation.is_a? Array
        cache_routes = isolation.map { |i| cache_route_with_key(key, i) }
        cache_routes << { expires_in: cache_options[:expiration] }
        values = Rails.cache.fetch_multi(*cache_routes) do
          if cache_options[:isolate]
            object = where(cache_options[:key] => key,
                           cache_options[:isolate] => isolation)
                     .order("#{cache_options[:isolate]} DESC NULLS LAST").first
          else
            object = find_by(cache_options[:key] => key)
          end
          # Get the value
          object ? object.try(cache_options[:value]) : default
        end
        # Parse the value
        values = values.values - [default]
        values.present? ? values.first : default
      else
        Rails.cache.fetch cache_route_with_key(key, isolation),
                          expires_in: cache_options[:expiration] do
          if cache_options[:isolate]
            object = find_by(cache_options[:key] => key,
                             cache_options[:isolate] => isolation)
          else
            object = find_by(cache_options[:key] => key)
          end
          # Get the value
          object ? object.try(cache_options[:value]) : default
        end
      end
    end

    #
    # Get the route for cache the element
    #
    # == Paramters:
    # key::
    #   String with a key with the user try to fetch a cache value
    # isolation::
    #   (Optional) String or integer with the value to isolate.
    #
    # == Returns:
    # String with the full identifier
    #
    def cache_route_with_key(key, isolation)
      if cache_options[:isolate]
        "#{cache_options[:namespace]}:#{ancestors.first.name.downcase}:"\
        "#{isolation}:#{key}"
      else
        "#{cache_options[:namespace]}:#{ancestors.first.name.downcase}:#{key}"
      end
    end

    #
    # Is the model cacheable?
    #
    # == Returns:
    # True
    #
    def cacheable?
      included_modules.include? Redborder::Cacheable::CacheableMethods
    end

    #
    # Is the cache isolated?
    #
    # == Returns:
    # Boolean
    #
    def isolated?
      cacheable? && cache_options[:isolate].present?
    end

    #
    # Module that define the methods to interact with cache on an instance
    #
    module CacheableMethods
      # ------------------------------------------------
      # This model is cacheable
      # ------------------------------------------------

      #
      # Is the model cacheable?
      #
      # == Returns:
      # True
      #
      def cacheable?
        self.class.cacheable?
      end

      #
      # Is the cache isolated?
      #
      # == Returns:
      # Boolean
      #
      def isolated?
        self.class.isolated?
      end

      # ------------------------------------------------
      # Create, update and destroy cache
      # ------------------------------------------------

      #
      # Cache current instance
      #
      def create_cache
        Rails.cache.write cache_route, send(cache_options[:value])
      end

      #
      # Delete cache of current instance
      #
      # == Parameters
      # older::
      #   Boolean to delete an older value (using attr_was from ActiveRecord).
      #   False by default.
      #
      def delete_cache(older = false)
        if older
          route = cache_route_older
        else
          route = cache_route
        end

        Rails.cache.delete route
      end

      #
      # Update cache of current instance
      #
      def update_cache
        if send("#{cache_options[:key]}_changed?")
          create_cache
        elsif send("#{cache_options[:value]}_changed?")
          delete_cache(true)
          create_cache
        end
      end

      private

      #
      # Get the route for cache the element
      #
      # == Returns:
      # String with the full identifier
      #
      def cache_route
        isolate = if cache_options[:isolate].nil?
                    ''
                  else
                    "#{send(cache_options[:isolate])}:"
                  end

        "#{cache_options[:namespace]}:#{self.class.name.downcase}:"\
        "#{isolate}#{send(cache_options[:key])}"
      end

      #
      # Get the route for an older cache element. (Using attr_was of
      # ActiveRecord)
      #
      # == Returns:
      # String with the full identifier
      #
      def cache_route_older
        isolate = if cache_options[:isolate].nil?
                    ''
                  else
                    "#{send(cache_options[:isolate])}:"
                  end

        "#{cache_options[:namespace]}:#{self.class.name.downcase}:"\
        "#{isolate}#{send("#{cache_options[:key]}_was")}"
      end
    end
  end
end
