#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Chef
    module Role
      # Provide methods to work with Chef role sensors. There are a lot of common
      # code between different sensors, so we must unify it.
      #
      class Builder
        include Redborder::Chef::Role::OverrideAttributes

        # Accessors
        attr_reader :response, :name, :role
        attr_accessor :run_list, :override_attributes

        # Initialize the helper
        #
        # @param sensor [Sensor] sensor that wants to create its Chef Role
        # @param run_list [Array] Name of inherited sensors
        # @param override_attributes [Hash] Optional. Force override_attributes.
        #
        def initialize(sensor, run_list, override_attributes: {}, domain: false)
          @name = sensor.chef_name
          @run_list = run_list
          @override_attributes = fetch_override_attributes(override_attributes)
          # Ensure sensor UUID and role
          @override_attributes['redborder']['sensor_uuid'] = sensor.uuid
          @override_attributes['redborder']['role'] = @name
          # Add the domain
          domain(sensor) if domain
        end

        # Create the Chef role
        #
        # @return [Chef::Role] Response returned from Chef
        #
        def create
          @role = ::Chef::Role.from_hash(to_h)
          @response = @role.create
        end

        # Check if the current role has been created
        #
        # @return [Boolean]
        #
        def created?
          @response.is_a? ::Chef::Role
        end

        # Return the objects as Hash. This can be used to create the Chef role
        #
        # @return [Hash] All properties, name and run list of Role
        #
        def to_h
          base = {
            'name' => @name,
            'override_attributes' => @override_attributes
          }
          @run_list.empty? ? base : base.merge('run_list' => @run_list)
        end

        private

        # Include domain in override attributes. This method will include a
        # "#{domain_type}" and "#{domain_type}_uuid" properties in
        # override attributes. Other domain types will be inherithed from
        # ancestors.
        #
        # @param domain [Sensor] Domain to include in override_attributes
        #
        def domain(domain)
          return false unless domain.is_a?(Sensor) &&
                              domain.respond_to?(:domain_type_name)
          name = domain.domain_type_name(to_key: true)
          @override_attributes['redborder'][name] = domain.name
          @override_attributes['redborder']["#{name}_uuid"] = domain.uuid
          true
        end
      end
    end
  end
end
