#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
  #
  # Define the basic methods to call to Druid.
  #
  class DataSource < Filters::Filtrable
    # Cache helpers
    include Redborder::DataSourceCache

    # =================================================
    # Interface. Define the methods the children
    # must define.
    # =================================================

    #
    # Throw an error when a method is undefined
    #
    def self.undefined_method
      fail NotImplementedError, "#{I18n.t 'lib_redborder.datasource'}" \
                                "#{I18n.t 'lib_redborder.all'}"
    end

    def self.controller
      undefined_method
    end

    def self.id
      undefined_method
    end

    def self.default_aggregation
      undefined_method
    end

    def self.sensors
      undefined_method
    end

    #
    # Combine the current conditions with defaults to perform a Druid query.
    #
    # == Parameters:
    # conditions::
    #   Hash with conditions from the client
    # == Returns:
    # Hash with default conditions for a query
    #
    def self.defaults(conditions = {}, user)
      # Merge defaults conditions prioritizing last.
      conditions.merge!(
        default_conditions
      ) { |_, v1, v2| v1.nil? ? v2 : v1 }

      conditions[:sort] = conditions_sort(conditions[:sort])

      if conditions[:limit]
        conditions[:limit] = conditions.fetch(:offset, 0).to_i +
                             conditions[:limit].to_i
      end

      conditions[:timezone] ||= timezone(conditions)
      conditions[:search] = user_filters conditions[:search], user
      conditions[:sensors] = readable_sensors(user) unless user.super_admin? ||
                                                           user.super_reader?

      # Load datasources
      namespaces = if conditions[:include_all_namespaces]
                     user.data_sources_ids
                   elsif conditions[:namespaces] &&
                         conditions[:ignore_user_namespaces]
                     # Limit namespaces to user active data sources
                     conditions[:namespaces]
                   elsif conditions[:namespaces]
                     conditions[:namespaces] & user.active_data_sources
                   else
                     user.active_data_sources
                   end
      # Get real data sources and delete namespaces key
      conditions[:data_sources] =
        data_sources(namespaces, conditions[:controller], user)
      conditions.delete(:namespaces)
      conditions.delete(:ignore_user_namespaces)

      # Return the new conditions hash
      conditions
    end

    #
    # Specify the default conditions. This method can be override on
    # DataSources.
    #
    # == Returns
    # A Hash with the default conditions.
    #
    def self.default_conditions
      { granularity: :minute,
        sort: :timestamp,
        direction: :DESCENDING,
        range: :last_1_hour,
        aggregation: default_aggregation,
        group_by: self::COLUMNS.select { |_, value| value[:enabled] }.keys,
        remove_zeros: false,
        skip_empty: true,
        controller: controller }
    end

    # =================================================
    # ATTRIBUTES
    # =================================================

    #
    # Get the value of the instance attribute
    #
    # == Parameters:
    # attribute::
    #   String representing the attribute
    #
    # == Returns:
    # Value of the attribute. If it doesn't exists, nil is returned.
    #
    def read_attribute(attribute)
      instance_variable_get("@#{attribute}")
    end

    # =================================================
    # DATASOURCES
    # =================================================

    #
    # Get the dataSources based on user.
    #
    # == Parameters:
    # namespaces::
    #   Array of Strings with the UUID of the namespace
    # controller::
    #   Symbol of controller to send the data.
    # user::
    #   User to fetch the data_sources based on namespaces
    #
    # == Returns:
    # Array of strings or String
    #
    def self.data_sources(namespaces, controller, user)
      # Get base dataSources for this query
      base = RedBorder::DRUID_CONFIG[controller]['datasource']
      # Dats sources
      sources = []

      # Add the id to datasource
      root_uuid = Sensor.root.uuid
      namespaces.each do |uuid|
        # For root, we use base
        ds = (uuid == root_uuid) ? "#{base}" : "#{base}_#{uuid}"
        sources << ds
      end

      # Return data sources
      sources.size == 1 ? sources.first : sources
    end

    # =================================================
    # QUERIES
    # =================================================

    #
    # Get the data from Druid in timeserie format.
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - ...
    #
    # == Return:
    # Array with the data from Druid.
    #
    def self.timeserie(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)

      # Default values
      conditions[:query_type] = :timeseries
      conditions = defaults(conditions, user)
      # Set the aggregation
      aggregation = conditions[:aggregation]
      aggregations = Redborder::Module.get(id, :aggregations)
      aggregation_data = aggregations[aggregation.to_sym]

      # Perform the query
      result = perform_query(conditions)

      if result.is_a? Array
        result.map do |t|
          value = t[conditions[:aggregation].to_s]

          unless value.nil?
            if aggregation_data[:round]
              value = value.round
            else
              value = value.round(3)
            end
          end

          if conditions[:outliers]
            { expected: t['expected'],
              timestamp: t['timestamp'].to_time.in_time_zone.to_s }
          else
            { event: { events: value },
              timestamp: t['timestamp'].to_time.in_time_zone.to_s }
          end
        end
      elsif result.is_a? Hash
        p = result.map { |k,v|
           [ k,
               if ( k.to_s == "predicted" )
                   v.map { |t|
                      { forecast: t['forecast'],
                        timestamp: t['timestamp'].to_time.in_time_zone.to_s }
                   }
               elsif ( k.to_s == "anomalies" )
                   v.map { |t|
                      { expected: t['expected'],
                        timestamp: t['timestamp'].to_time.in_time_zone.to_s }
                   }
               else
                   v
               end
           ]}.to_h
        p

      else
        result
      end
    end

    #
    # Get the data from a general DruidQuery
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - ...
    #
    # == Return:
    # Array with the data from Druid.
    #
    def self.where(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)

      # Default values
      conditions = defaults(conditions, user)
      # Perform the query
      result = perform_query(conditions)

      if conditions[:limit]
        result =
          result[conditions.fetch(:offset, 0).to_i..(conditions[:limit].to_i - 1)]
      end

      if result.is_a?(Array)
        conditions[:outliers] ? result : result.map { |item| new(user, item) }
      elsif result.nil?
        []
      else
        [new(user, result)]
      end
    end

    #
    # Get the data from a total DruidQuery
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - ...
    #
    # == Return:
    # Array with the data from Druid.
    #
    def self.total(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)

      # Default values
      conditions[:query_type]    = :timeseries
      conditions[:granularity] ||= :all
      conditions = defaults(conditions, user)
      # Set the aggregation
      aggregation = conditions[:aggregation]
      aggregations = Redborder::Module.get(id, :aggregations)
      aggregation_data = aggregations[aggregation.to_sym]

      # Perform the query
      result = perform_query(conditions)

      if result.blank?
        0
      elsif aggregation_data[:round]
        result.first[aggregation.to_s].round
      else
        result.first[aggregation.to_s].round(3)
      end
    end

    #
    # Use select query to get diferent elements in the interval
    # paginating the queries.
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - dimension: dimension to map by
    #     - pagingIdentifiers: Hash with the next paging identifier.
    #                          Empty by default (first query)
    #     - threshold: number of elements to return. 25 by default
    #     - dimensions: dimensions for return. If not present, it returns all
    #     - metrics: metric for returns. If not present, it returns all
    #     - ...
    #
    # == Return:
    # Array with the different dimension data
    #
    def self.select_distinct(options)
      # get dimension to map by
      dimension = options.delete(:dimension)
      data = []
      loop do
        # init at first page
        options[:pagingIdentifiers] = {}
        # use select
        pagination, data_ = select(options)
        # exit loop when no new data received
        break if data_.empty?
        data += data_m.map { |el| el[dimension] }.compact
        # increase pagination
        pagination.each { |k,_| pagination[k] += 1 }
        options[:pagingIdentifiers] = pagination
      end
      # return unique data
      data.uniq
    end

    #
    # Use select query to get elements in the interval and paginate the results.
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - pagingIdentifiers: Hash with the next paging identifier.
    #                          Empty by default (first query)
    #     - threshold: number of elements to return. 25 by default
    #     - dimensions: dimensions for return. If not present, it returns all
    #     - metrics: metric for returns. If not present, it returns all
    #     - ...
    #
    # == Return:
    # Array with the data from Druid and the next paging
    #
    def self.select(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)
      # Default values
      conditions[:query_type]    = :select
      conditions[:granularity]   = :all
      conditions[:pagingSpec]    = {}
      conditions[:pagingSpec][:pagingIdentifiers] =
        conditions.delete(:pagingIdentifiers) || {}
      conditions[:pagingSpec][:threshold] = conditions.delete(:threshold) || 25
      conditions = defaults(conditions, user)
      # Delete unused
      to_delete = %i(group_by aggregation range limit)
      conditions.delete_if { |k, _| to_delete.include? k }
      # Perform query
      result = perform_query(conditions)

      if result.empty?
        next_pagination = {}
        events = []
      elsif conditions[:raw_result]
        next_pagination = result.first['pagingIdentifiers']
        events = result.first['events'].map { |item| item['event'] }
      else
        next_pagination = result.first['pagingIdentifiers']
        events = result.first['events'].map do |item|
          new(user, item['event'])
        end
      end
      # Return next pagination and events
      [next_pagination, events]
    end

    def self.selectmonitor(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)
      # Default values
      conditions[:query_type]    = :select
      conditions[:granularity]   = :all
      conditions[:pagingSpec]    = {}
      conditions[:pagingSpec][:pagingIdentifiers] =
        conditions.delete(:pagingIdentifiers) || {}
      conditions[:pagingSpec][:threshold] = conditions.delete(:threshold) || 25
      conditions = defaults(conditions, user)
      # Delete unused
      to_delete = %i(group_by aggregation range limit)
      conditions.delete_if { |k, _| to_delete.include? k }
      # Perform query
      result = perform_query(conditions)

      if result.empty?
        events = []
      else
        events = []
        result.first['events'].map do |item|
          item['event'].delete('timestamp')
          item['event'].delete("min_value")
          item['event'].delete("sum_value")
          item['event'].delete("events")
          item['event'].delete("max_value")
          events.push(item['event'])
        end
      end
      events.uniq
    end

    #
    # Get the data from a TopN DruidQuery. Its organized by a timestamp and
    # the multiple results.
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - ...
    #
    # == Return:
    # Array with the data from Druid.
    #
    def self.topN(options = {})
      conditions = options.deep_dup
      user = options.fetch(:user, User.current_user)

      # Default values
      conditions = defaults(conditions, user)
      # Perform the query
      result = perform_query(conditions)

      if result.is_a?(Array)
        if conditions[:granularity].to_sym == :all
          result.map { |item| new(user, item) }
        else
          result.each_with_object([]) do |period, memo|
            next if period.empty?
            memo << {
              timestamp: period.first['timestamp'],
              result: period.map { |top| new(user, top) }
            }
          end
        end
      elsif result.nil?
        []
      else
        [new(user, result)]
      end
    end

    #
    # Get the data from a general DruidQuery to parse as JSON
    #
    # == Paramenters:
    # args::
    #   Hash with options to perform the query. The options are:
    #     - ...
    #
    # == Return:
    # Array with the data from Druid.
    #
    def self.as_json(options = {})
      conditions = options.deep_dup

      # Default values
      conditions = defaults(conditions,
                            options.fetch(:user, User.current_user))

      # Perform the query
      result = perform_query(conditions)

      result.each_with_object([]) do |period, memo|
        next if period.empty?
        memo << {
          timestamp: period.first['timestamp'],
          result: period
        }
      end
    end

    #
    # Perform the DruidQuery and return the result. Throws and exception
    # if it fails.
    #
    # == Parameters:
    # conditions::
    #   A Hash with the conditions specify in other methods.
    #
    # == Returns:
    # An array with the restuls. If there are a problem, it will be empty.
    #
    def self.perform_query(conditions = {})
      user = conditions.fetch(:user, User.current_user)

      # Check the interval range for the user.
      conditions[:start_time] = [conditions[:start_time], Time.at(user.oldest_timestamp_allowed)].max
      conditions[:end_time] = [conditions[:end_time], Time.at(user.oldest_timestamp_allowed)].max

      begin
        result = Redborder::Rbdruid.query conditions
        # Check if we must reverse the data
        if conditions[:sort] == :timestamp &&
           conditions[:direction] == :DESCENDING &&
           conditions[:granularity] != :all
           result.reverse! if result.is_a? Array
        end
        # Return the result
        result
      rescue RuntimeError => e
        Rails.logger.error e.inspect
        Rails.logger.error e.backtrace.join("\n")
        []
      end
    end

    # =================================================
    # HELPERS
    # =================================================

    #
    # Return the top limit based on the limit and the offset given.
    #
    # == Parameters:
    # args::
    #   Hash with the info. The relevant data is offset and limit.
    #
    # == Returns:
    # Number of the top_limit.
    #
    def self.top_limit(conditions = {})
      conditions.fetch(:offset, 0).to_i + (conditions[:limit].to_i - 1)
    end

    #
    # Return the default timezone based on the conditions of the query
    #
    def self.timezone(conditions = {})
      Time.zone.name if conditions[:timezone] == 'auto' ||
                        conditions[:timezone].nil?
    end

    # =================================================
    # FILTERS
    # =================================================

    #
    # Return the filters
    #
    # == Parameters:
    # search::
    #   A hash with the current filters to apply
    # user::
    #   Current user
    #
    # == Returns:
    # A hash with the filters combination of current, user, domain and zones.
    #
    def self.user_filters(search, user)
      return {} if user.nil? || search.nil?
      filters = {}
      # Filters from user
      filters.merge! search.except(:domain)
      # Filters from domains
      filters.merge! domain_filters(search[:domain], search[:sensor_uuid], user)

      filters
    end

    #
    # This method transforms the domains filters on filters
    # based on the sensors
    #
    # == Parameters:
    # domains::
    #   An array of domains
    # sensor_ids::
    #   An array with the name of the sensors
    # user::
    #   Current user
    #
    # == Returns:
    # A hash with the domains filters
    #
    def self.domain_filters(domains, sensor_uuids, user)
      return {} if domains.nil? || domains.empty?

      filters = {}
      filters[:sensor_uuid] = sensor_uuids || {}

      domains.each do |value, filter_type|
        # Select all the sensors from the domains, based on the type of the
        # filter that the user chose
        selected_domains = domain_filtered_sensors(value, filter_type)

        # Only select sensors that are Flow Sensors and that
        # the user can read
        # TODO: FLOW can't be used here
        sensors = selected_domains.collect do |d|
          d.all_children.where(type: RedBorder::SensorFlow::FLOW)
        end.flatten

        sensors = sensors.select do |sensor|
          user.can?(:read, sensor)
        end

        # If there are no sensors on these domains, filter for none
        sensors = [Sensor.new(name: '0', uuid: '0')] if sensors.empty?

        # Apply the selected filter
        sensors.map(&:uuid).each do |s|
          filters[:sensor_uuid][s.to_s] = { operator: 'is' }
        end
      end

      filters
    end

    #
    # Return an array of sensors in a domain based on filters operators.
    #
    # == Parameters:
    # value::
    #   The value of the filter
    # filter_type::
    #   A Hash with the options of the filter
    #
    def self.domain_filtered_sensors(value, filter_type)
      case filter_type[:operator]
      when 'contains'
        selected_domains = Sensor.where(type: Sensor::DOMAIN)
                           .where('LOWER(name) LIKE :search',
                                  search: "%#{value.downcase}%")
      when 'is_not'
        selected_domain = Sensor.find(value)
        all_domains = Sensor.where(type: Sensor::DOMAIN)
        readable_domains = all_domains.select do |domain|
          user.can? :read, domain
        end

        selected_domains = (readable_domains - [selected_domain])
      when 'is'
        selected_domains = [Sensor.find(value)]
      end

      selected_domains || []
    end

    #
    # Check the sensors current user can read.
    #
    # == Parameters:
    # user::
    #   User to check the permissions
    #
    # == Returns:
    # A hash with the sensor_id key contains all readable sensor uuids.
    def self.readable_sensors(user)
      sensor_uuids = (sensors.select do |sensor|
        user.can?(:read, sensor)
      end).map(&:uuid)

      { sensor_uuid: sensor_uuids }
    end

    def self.set_custom_values(params)
      unless params[:granularity].to_sym.eql?(:all)
        params[:start_time] = params[:custom_filter][:timestamp].to_s.to_time
      end

      case params[:granularity].to_sym
      when :all
        params[:granularity] = :pt720h
      when :pt720h
        params[:granularity] = :pt168h
        params[:end_time] = params[:start_time] + 720.hour
      when :pt168h
        params[:granularity] = :pt120h
        params[:end_time] = params[:start_time] + 168.hour
      when :pt120h
        params[:granularity] = :pt24h
        params[:end_time] = params[:start_time] + 120.hour
      when :pt24h
        params[:granularity] = :pt8h
        params[:end_time] = params[:start_time] + 24.hour
      when :pt8h
        params[:granularity] = :pt2h
        params[:end_time] = params[:start_time] + 8.hour
      when :pt2h
        params[:granularity] = :hour
        params[:end_time] = params[:start_time] + 2.hour
      when :hour
        params[:granularity] = :thirty_minute
        params[:end_time] = params[:start_time] + 1.hour
      when :thirty_minute
        params[:granularity] = :fifteen_minute
        params[:end_time] = params[:start_time] + 30.minute
      when :fifteen_minute
        params[:granularity] = :pt5m
        params[:end_time] = params[:start_time] + 15.minute
      when :pt5m
        params[:granularity] = :pt2m
        params[:end_time] = params[:start_time] + 5.minute
      when :pt2m
        params[:granularity] = :minute
        params[:end_time] = params[:start_time] + 2.minute
      when :minute
        params[:granularity] = :minute # couchbase
        params[:end_time] = params[:start_time] + 1.minute
      else
        params[:granularity] = :minute
        params[:end_time] = params[:start_time] + 24.hour
      end

      params[:range] = :custom
    end

    # =================================================
    # CSV
    # =================================================

    #
    # Convert the content of a tab to an exportable CSV.
    #
    # == Parameters:
    # content::
    #   Data to export
    # tab::
    #   Column to export
    # aggregation::
    #   Symbol of aggregation to export
    #
    # == Returns:
    # An string with the CSV code
    #
    def self.to_csv(content, tab, aggregations, locations=nil)
      source_columns = Redborder::Module.get(id, :columns)
      columns_keys   = source_columns[tab.to_sym][:columns].keys +
        (source_columns[tab.to_sym][:complementary_columns].try('keys') || []) +
        aggregations.map(&:to_s)
      columns_values = source_columns[tab.to_sym][:columns].values +
        (source_columns[tab.to_sym][:complementary_columns].try('values') || []) +
        aggregations.map(&:to_s)
      parse_csv(content, columns_keys, columns_values, locations)
    end

    #
    # Convert RAW content to an exportable CSV.
    #
    # == Parameters:
    # content::
    #   Data to export
    # columns::
    #   The columns to export
    # aggregation::
    #   Symbol of aggregation to export
    #
    # == Returns:
    # An string with the CSV code
    #
    def self.raw_to_csv(content, columns, aggregations)
      source_columns = Redborder::Module.get(id, :columns)
      columns_keys   = columns + aggregations << :timestamp
      columns_values = []
      columns.each do |c|
        columns_values << (source_columns[c.to_sym][:name] ||
                            source_columns[c.to_sym][:columns].first[1])
      end
      columns_values << aggregations + [:timestamp]

      parse_csv(content, columns_keys, columns_values)
    end

    def self.resolve_location(magic_key, locations, aggregation)
      return magic_key if locations[:lookup]

      aggregation = aggregation.to_s.downcase.to_sym

      data = locations[:locations][magic_key]

      if locations[:aggregations].key?(aggregation)
        data = magic_key.round
      end

      data
    end

    # Resolv an ip
    def self.resolv_ip(ip)
      name = ip
      begin
        ip_aux = IPAddr.new(ip)

        name = Rails.cache
              .fetch "rb-webui:ip:#{ip}", expires_in: 30.minutes do
          begin
            NetObject.find_by(value: ip).try('name') ||
            timeout(0.5) { Socket.getnameinfo(Socket.sockaddr_in(80, ip))[0] }
          rescue Timeout::Error
            ip_aux.to_s
          end
        end
      rescue
        name = ip
      end
      name
    end

    #
    # Join all the data into the correct format of CSV.
    #
    # == Parameters:
    # content::
    #   Data to export
    # columns_keys::
    #   The columns keys for the CSV
    # aggregation::
    #   The value for the columns.
    #
    # == Returns:
    # An string with the CSV code
    #
    def self.parse_csv(content, columns_keys, columns_values, locations=nil)
      resolve_private_dns = RedBorder::CONFIG[:resolve_private_dns]
      columns_to_resolve_dns = [:lan_ip_name, :wan_ip_name, :src_name, :dst_name]

      CSV.generate(col_sep: ',', force_quotes: true) do |csv|
        csv << columns_values

        content.each do |line|
          row = columns_keys.map do |column|
            value = line.send(column.to_s)
            
            if resolve_private_dns && columns_to_resolve_dns.include?(column.to_sym)
              resolv_ip(value) || 'N/A'
            elsif locations
              resolve_location(value, locations, column) || 'N/A'
            elsif column.to_sym == :client_mac_inventored
              value ? "#{I18n.t 'redborder.commons.common.index.inventored'}" : "NO"
            else
              value || 'N/A'
            end
          end
          csv << row
        end
      end
    end
    #
    # Modify sort condition if it's a virtual aggregation
    # sort condition can be an array or a single value
    #
    # == Parameters:
    # attributes::
    #   Columns used to sort the query.
    #
    # == Returns:
    # Array with the correct sort values.
    #
    def self.conditions_sort(attributes = [])
      attributes_tmp = []
      [*attributes].each do |s|
        if self::AGGREGATIONS.keys.include? s
          attributes_tmp += self::AGGREGATIONS[s][:aggregations].map { |a| a[:name] }
        else
          attributes_tmp << s
        end
      end

      attributes_tmp.one? ? attributes_tmp.first : attributes_tmp
    end
  end
end
