#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
    #
    # Class to work with Chef databags.
    #
    class Databag
      # Exceptions
      class NoSecretError < StandardError; end
      class ChefConnectionError < StandardError; end
      class DatabagNotFoundError < StandardError; end
  
      #
      # Load a databag and return it.
      #
      # == Parameters:
      # db_name::
      #   String with the store of the databag
      # databag::
      #   String with the name of the databag
      # exception::
      #   (Optional) Boolean indicating if we wan't to launch exceptions.
      #   True by default.
      # create::
      #   (Optional) Boolean indicating if you want to create the databag if it
      #   doesn't exist. True by default
      #
      # == Returns:
      # The databag or nil if we can't recover it.
      #
      def self.load(db_name, databag, exception = true, create = true)
        # Try to load it
        return ::Chef::DataBagItem.load(db_name, databag)
      rescue Errno::ECONNREFUSED, ArgumentError
        exception_or_nil(
          exception, ChefConnectionError,
          'ERROR: when access databag: cannot contact erchef server')
      rescue Net::HTTPServerException
        if create
          create_databag_item(db_name, databag, exception)
        else
          exception_or_nil(
            exception, DatabagNotFoundError,
            "Databag #{db_name} #{databag} doesn't exist")
        end
      end
  
      #
      # Create a Databag in Chef.
      #
      # == Parameters:
      # db_name::
      #   String with the store of the databag
      # databag::
      #   String with the name of the databag
      # exception::
      #   (Optional) Boolean indicating if you want to launch exceptions or only
      #   return nil. True by default
      #
      # == Returns:
      # Databag, nil or an exception
      #
      def self.create_databag_item(db_name, databag, exception = true)
        Rails.logger.info "INFO: Creating Databag #{db_name} #{databag}"
        data = Chef::DataBagItem.new
        data.data_bag db_name
        data['id'] = databag
        data['created'] = false
        if data.save
          Rails.logger.info "INFO: Databag #{db_name} #{databag} created."
          data
        else
          exception_or_nil(
            exception, ChefConnectionError,
            'ERROR: when access databag, cannot contact erchef server')
        end
      end
  
      private
  
      #
      # Return an exception or nil based on exception boolean
      #
      # == Parameters:
      # exception::
      #   True if you want to launch an exception
      # message::
      #   Message of the exception
      #
      # == Returns:
      # Nil or an exception
      #
      def self.exception_or_nil(exception, exception_type, message)
        return fail(exception_type, message) if exception
        # Only log and return
        Rails.logger.error message
        nil
      end
    end
end
  