#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
  #
  # Works with filters. This module helps to build and parse filters.
  # The objective is to send an intuitive filters to Druid.
  #
  class Filter
    # Includes
    include Redborder::Filters::FilterValues
    include Redborder::Filters::FilterParse
    include Redborder::Filters::FilterTemporal

    #
    # Initialize the class with the current filters. This class is independent
    # about the store of filters
    #
    # == Parameters:
    # source::
    #   Filtrable class. It can be nil if we wants all filters
    # filters::
    #   Hash with current filters
    #
    def initialize(source, filters = {})
      # Check the source
      fail ArgumentError, "#{I18n.t 'lib_redborder.source'}" if source.nil?

      @source = source
      # Check base filters
      filters = {} if filters.nil?
      filters.merge!(shared: {}) { |_, v1, _| v1 }
      @filters = filters
      @count = count_filters
    end

    # -------------------------------------------------------------
    # Methods to load, save and delete all filters
    # -------------------------------------------------------------

    #
    # Get the filter's source
    #
    # == Returns:
    # Source
    #
    def source
      @source
    end

    #
    # Load all the filters.
    #
    # == Returns:
    # Hash with the filters
    #
    def filters
      # Get the filters for the source
      @filters
    end

    #
    # Load all the filters for the source.
    #
    # == Parameters:
    # column::
    #   Symbol with column to get the filters. If it's nil,
    #   return all filters
    #
    # == Returns:
    # Hash with the filters
    #
    def get(column = nil)
      # Get the filters for the source
      return @source.filters(@filters) if column.nil?
      @source.filters(@filters)[column.to_sym]
    end

    #
    # Load shared filters.
    #
    # == Parameters:
    # column::
    #   Symbol with column to get the filters. If it's nil,
    #   return all shared filters
    #
    # == Returns:
    # Hash with the filters
    #
    def shared(column = nil)
      # Return the session hash or an empty hash
      return @filters[:shared][column] || {} unless column.nil?
      @filters[:shared]
    end

    #
    # Remove all filters stored
    #
    def remove_all
      @count = 0
      @filters = { shared: {} }
    end

    #
    # Replace the filters
    #
    def replace(filters)
      @filters = filters
      @count = 0
      @count = count_filters
    end

    #
    # Return the number of filters is applied
    #
    # == Returns:
    # Integer wwith the number of filters
    #
    def size
      @count
    end

    alias_method :count, :size

    #
    # Check if exist a filter for a column
    #
    # == Parameters:
    # column::
    #   A symbol with the column to check
    #
    # == Returns:
    # True if the column exists on filters
    #
    def exist_any?(column)
      column = column.to_sym
      @filters[:shared].keys.include?(column) ||
        (@filters[@source.id].present? &&
        @filters[@source.id].keys.include?(column))
    end

    #
    # Check if exist a filter on the column with given value. Nil values
    # are converted to an empty string
    #
    # == Parameters:
    # column::
    #   A symbol with the column to check
    # value::
    #   An string of value. By default, it's nil
    #
    # == Returns:
    # True if the column exists on filters
    #
    def exist?(column, value)
      column = column.to_sym
      value = '' if value.nil?
      # Check for values
      (@filters[:shared][column].present? &&
       @filters[:shared][column][value].present?) ||
        (@filters[@source.id].present? &&
         @filters[@source.id][column].present? &&
         @filters[@source.id][column][value].present?)
    end

    #
    # Check if there are time filters applied
    #
    # == Returns:
    # True if there are a time filter appied
    #
    def time_filters?
      exist_any?(:start_time) && exist_any?(:end_time)
    end

    # -------------------------------------------------------------
    # Methods to operate on filters
    # -------------------------------------------------------------

    #
    # Apply the filter. It's call to add_filter to add a filter for evevy value
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    # operator::
    #   The operator to apply. 'Is' is the operator by default
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def add(column, values, operator = 'is')
      # Init the variables
      column = column.to_sym
      values = format_values(column, values)
      # Check comparative filters
      if comparative? column
        values = comparative_values(column, values)
        add_filter(:start_time, values.first, operator)
        add_filter(:end_time, values.last, operator)
      else
        # Add a filter for every simple filter.
        values.each do |value|
          add_filter(column, value, operator)
        end
      end
      # Return and save filters
      get
    end

    #
    # Alias to add an empty filter in the column
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def add_empty(column)
      add(column, [''], 'is_not')
    end

    #
    # Alias method to add a contains filter. It's call to add
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def contains(column, values)
      add(column, values, 'contains')
      ## Add domains if column is sensor_name
      #add(:domain, values, 'contains') if column.to_sym == :sensor_name
      # Check objects and add them to filters
      @source.parse_objects(column.to_sym, values).each do |value|
        add(column, value)
      end
      # Return filters
      get
    end

    #
    # Alias method to add an exclude filter. It's call to add
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def exclude(column, values)
      add(column, values, 'is_not')
    end

    #
    # Alias to exclude empty values
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def exclude_empty(column)
      exclude(column, [''])
    end

    #
    # Remove current values of the column from the filter
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def remove(column, values = nil)
      # Init the variables
      column = column.to_sym
      # Remove only the column if values is nil
      if values.nil?
        remove_column(column)
      else
        # Format values
        values = format_values(column, values)
        # Add a filter for every simple filter
        values.each do |value|
          remove_filter(column, value)
        end
      end

      # Remove time filters value
      remove_time_filters(column)
      # Return and save filters
      get
    end

    # ======================================================================
    #  Helpers
    # ======================================================================

    def dump(index: 0)
      dump = {}
      self.get.each do |col, f|
        f.each do |column_value, operator|
          dump[index]             = {}
          dump[index]['column']   = col
          dump[index]['value']    = column_value
          dump[index]['operator'] = operator[:operator]
          dump[index]['enable']   = true
          index += 1
        end
      end
      dump
    end

    private

    #
    # Add a unique filter to current if the column is allowed
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    # filters::
    #   Hash with previous filters
    # operator::
    #   The operator to apply. 'Is' is the operator by default
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def add_filter(column, value, operator)
      # Skip if the source can't be filtered by the column
      return unless @source.can_filter?(column)
      # Only add a new filter if it doesn't exist
      @count += 1 unless exist?(column, value)
      build = build(column, value, operator)

      # Add the filter to current values. If it exists, it's added as a hash
      # If it doesn't exist, the column is added to the hash.
      # The second merge condition ({ |_, _, n| n }), is needed when we have a
      # previous is_not operator. This condition override it with a 'is'
      if !time?(column) && @source.private_filter?(column)
        @filters[@source.id] = {} if @filters[@source.id].nil?
        @filters[@source.id]
          .merge!(build) { |_, v1, v2| v1.merge(v2) { |_, _, n| n } }
      else
        @filters[:shared]
          .merge!(build) { |_, v1, v2| v1.merge(v2) { |_, _, n| n } }
      end
    end

    #
    # Remove a filter and the column if there aren't any filter more.
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    # values::
    #   Array of values for the filter
    # filters::
    #   Hash with previous filters
    #
    # == Returns:
    # A Hash with without removed filters
    #
    def remove_filter(column, value)
      # Skip if the source can't be filtered by the column or the column
      # doesn't exist in the filter
      return unless @source.can_filter?(column) && exist_any?(column)
      @count -= 1

      # Delete de filter from filters
      dest = if !time?(column) && @source.private_filter?(column)
               @source.id
             else
               :shared
             end
      dest = :shared unless @filters[dest] && @filters[dest][column]

      @filters[dest][column].delete(value)
      @filters[dest].delete(column) if @filters[dest][column].empty?
    end

    #
    # Remove a column from filters
    #
    # == Parameters:
    # column::
    #   A string of the column to filter
    #
    # == Returns:
    # A Hash with the new filters applied.
    #
    def remove_column(column)
      # Init the variables
      column = column.to_sym
      # Check if the column exists
      return get unless exist_any?(column)
      # To fetch the value, use the keys of the column, these will be delete
      # all the column
      dest = @source.private_filter?(column) ? @source.id : :shared
      dest = :shared if @filters[:shared] and @filters[:shared][column] and !@filters[dest].key?column
      return get if @filters[dest].nil?

      @count -= @filters[dest][column].keys.size
      @filters[dest].delete(column)
      # Return and save filters
      get
    end

    #
    # If column is is a time column, we must delete another time columns
    #
    # == Parameters:
    # column::
    #   The column the user want's to delete
    #
    def remove_time_filters(column)
      return unless time?(column)
      TIME_COLUMNS.each do |tcolumn|
        # Column has deleted previously
        if exist_any?(tcolumn) && tcolumn != column
          # Time filters are always shared
          @filters[:shared].delete(tcolumn)
          @count -= 1
        end
      end
    end

    #
    # Count the number of filters. Iterate over shared and source filters.
    #
    # == Returns:
    # Integer with the number of filters
    #
    def count_filters
      @count = 0 if @count.nil?
      get.each do |column, values|
        @count += values.size
      end
      # Return the integer
      @count
    end

    #
    # Build the hash of a filter
    #
    # == Parameters:
    # column::
    #   DataSource column
    # value::
    #   String with the value
    # operator::
    #   A valid operator to Druid
    #
    # == Returns:
    # Hash with the builded filter
    #
    def build(column, value, operator)
      # Return the builded filter
      { column => { value => { operator: operator } } }
    end
  end
end
