#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
  module Filters
    #
    # Module to parse and serialize filters.
    #
    module FilterParse
      #
      # Parse the filters from String and add them to the current filters.
      #
      # == Parameters:
      # data::
      #   An string contains a JSON object with the syntax:
      #     { "0": { 'column': "...", 'operator': "...", 'value': "..." },
      #       "1": { .... } }
      # replace::
      #   Boolean to add the filters or replace them
      #
      # == Returns:
      # A Hash with the filter for the current source.
      #
      def parse(data, replace = false)
        # TODO: A controller can try to parse filters from params[:search]. In
        # some situations, params[:search] can be a simple string called from a
        # search box. We must change this behaviour on controllers, but now,
        # when we can't parse JSON, return current filters
        begin
          data = JSON.parse(data || '{}')
        rescue JSON::ParserError
          return get
        end

        # Check if we must replace all widgets
        remove_all if replace

        # Loop inside data
        data.each do |f|
          f = f.last
          # The data is stored on f.last
          add(f['column'], Array(f['value']), f['operator'])
        end
        # Return the filters for this
        get
      end

      #
      # Serialize the current methods in a hash to convert to JSON.
      #
      # == Parameters:
      # json::
      #   Boolean to indeicate if return a Hash or JSON String.
      # exclude_time::
      #   Boolean to indicate if we must ignore time filters. Store by default
      #
      # == Returns:
      # A has with the correct format to sendo to Javascript App or the JSON
      # string if json is true. The format is:
      #   { "0": { 'column': "...", 'operator': "...", 'value': "..." },
      #     "1": { .... } }
      #
      def serialize(json = false, exclude_time = false, parse_time = true)
        # Init vars
        items = {}
        index = 0

        # Get from Filter values
        time_columns = Redborder::Filters::FilterValues::TIME_COLUMNS

        # Loop on every filter
        get.each do |column, filter|
          filter.each do |value, operator|
            next if exclude_time && time_columns.include?(column)
            items[index] = build_json(column.to_s, value,
                                      operator[:operator].to_s,
                                      parse_time)
            index += 1
          end
        end
        # Return as string or json
        json ? items.to_json : items
      end

      #
      # Build a JSON filter.
      #
      # == Parameters:
      # column::
      #   DataSource column
      # value::
      #   String with the value
      # operator::
      #   A valid operator to Druid
      #
      # == Returns:
      # Hash with the builded JSON filter
      #
      def build_json(column, value, operator, parse_time)
        # Check the value
        value = if time?(column.to_sym) && parse_time
                  Time.zone.at(value).strftime('%Y-%m-%d %H:%M:%S')
                elsif time?(column.to_sym)
                  value
                else
                  value.to_s
                end
        # Build the hash
        { column: column, value: value, operator: operator, enabled: true }
      end
    end
  end
end
