#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
  module Filters
    #
    # Module to add, delete and store temporal filters. This is used when you
    # wants to add a tmp filter or store an snapshot of current filters.
    #
    module FilterTemporal
      # -------------------------------------------------------------
      # Add temp filters
      # -------------------------------------------------------------

      #
      # Added a temp filter. This add method doesn't affect to filters.
      #
      # == Parameters:
      # column::
      #   A string of the column to filter
      # values::
      #   Array of values for the filter
      # operator::
      #   The operator to apply. 'Is' is the operator by default
      #
      # == Returns:
      # A Hash with the new filters applied.
      #
      def add_tmp(column, values, operator = 'is')
        # Stash current filters add and pop!
        stash
        tmp = add(column, values, operator)
        pop

        # Return tmp
        tmp
      end

      #
      # Remove current values of the column from the filter. This method
      # doesn't affecto to stored filters
      #
      # == Parameters:
      # column::
      #   A string of the column to filter
      # values::
      #   Array of values for the filter
      #
      # == Returns:
      # A Hash with the new filters applied.
      #
      def remove_tmp(column, values = nil, operator = 'is')
        # Stash current filters add and pop!
        stash
        tmp = remove(column, values, operator)
        pop

        # Return tmp
        tmp
      end

      # -------------------------------------------------------------
      # Store snapshots
      # -------------------------------------------------------------

      #
      # Create a new snapshot. This store the current status of the filters
      #
      # == Parameters:
      # name::
      #   (Optional) String with name of snapshot.
      #
      # == Returns:
      # Filter object
      #
      def stash(name = nil)
        # Set the name
        name = random_name unless name

        # Store the snapshot
        @snapshots = {} if @snapshots.nil?
        @snapshots[name] = @filters.deep_dup

        # Return the object
        self
      end

      #
      # Return and delete a snapshot.
      #
      # == Parameters:
      # name::
      #   (Optional) String with name of snapshot.
      #
      # == Returns:
      # Filter object
      #
      def pop(name = nil)
        # We must have an snapshot
        return empty_snapshots if @snapshots.nil? || @snapshots.empty?

        # Set the name
        name = @snapshots.keys.last unless name

        # Store the snapshot
        replace(@snapshots.delete(name))

        # Return the object
        self
      end

      #
      # Generate a 8 character random string.
      #
      # == Returns:
      # String.
      def random_name
        (0...8).map { (65 + rand(26)).chr }.join
      end

      #
      # Throw an error because the user try to get an snapshot and any snapshot
      # is declared.
      #
      def empty_snapshots
        fail IndexError, 'There aren\'t any snapshot stored'
      end
    end
  end
end
