#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder
  module Filters
    #
    # Module to works with time filters like comparative or time columns.
    #
    module FilterValues
      # Time columns
      TIME_COLUMNS = [:start_time, :end_time]
      # Comparative columns (Loaded from Rbdruid)
      COMPARATIVE_COLUMNS = Redborder::Rbdruid::COMPARATIVE_COLUMNS.keys
      COMPARATIVE_DATA = Redborder::Rbdruid::COMPARATIVE_COLUMNS
      # Filters with multiple columns
      MULTIPLE_COLUMNS = [:conversation, :filter_monitor]

      # ======================================================================
      #  Format helpers
      # ======================================================================

      #
      # Receive an array of values and asnitize and format them.
      #
      # == Parameters:
      # column::
      #   Symbol of column to apply the filter
      # values::
      #   Array of values
      #
      # == Return:
      # An array of values.
      #
      def format_values(column, values)
        # Check if values is an array
        values = array_of_values(values)
        # Check if it's multiple
        values = MULTIPLE_COLUMNS.include?(column) ? values.values : values
        # Check types
        if time?(column) || comparative?(column)
          # Time filters must be integer
          values.map do |v|
            v.is_a?(String) && !(v.match(/^\d+$/)) ? v.to_time.to_i : v.to_i
          end
        else
          # Sanitize to druid
          values.map { |v| druid_clear_value v }
        end
      end

      #
      # Return the values for a comparative filter
      #
      # == Parameters:
      # column::
      #   Symbol of column to apply the filter
      # values::
      #   An array of values
      #
      # == Returns:
      # An array of two values. [Min time, Max time]
      #
      def comparative_values(column, values)
        [values.min,
         values.size == 1 ? values.min + COMPARATIVE_DATA[column][:step] : values.max]
      end

      # ======================================================================
      #  Type helpers
      # ======================================================================
      
      #
      # Check if column is a time Column
      #
      # == Parameters:
      # column::
      #   A symbol with the column to check
      #
      # == Returns:
      # True if it's a time column
      #
      def time?(column)
        TIME_COLUMNS.include?(column)
      end

      #
      # Check if column is a comparative time Column
      #
      # == Parameters:
      # column::
      #   A symbol with the column to check
      #
      # == Returns:
      # True if it's a comparative time column
      #
      def comparative?(column)
        COMPARATIVE_COLUMNS.include?(column)
      end

      # ======================================================================
      #  Format helpers
      # ======================================================================

      #
      # Check if values are an array or init it.
      #
      # == Parameters:
      # values::
      #   A string or array
      #
      # == Returns:
      # Array of values
      #
      def array_of_values(values)
        return values if values.is_a?(Array)
        [values]
      end

      #
      # Clear the value to send it to Druid.
      #
      # == Parameters:
      # value::
      #   Value to send to Druid
      #
      # == Returns:
      # Sanitized value
      #
      def druid_clear_value(value)
        (value.blank? || value.eql?('N/A')) ? '' : value
      end
    end
  end
end
