#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Base redBorder module
#
module Redborder

  module Filters
    #
    # Define the operators to work with Druid. Every module extend this class on
    # his DataSource must specify the allowed columns and the shared columns.
    #
    class Filtrable
      #
      # Traduce the Druid Operator to a symbol operator.
      #
      OPERATOR = {
        is: '=',
        is_not: '!=',
        not: '!=',
        contains: 'LIKE',
        starts_with: '^=',
        ends_with: '=$',
        contains_not: 'NOT LIKE',
        regex: 'regex',
        empty: '==',
        not_empty: '!=',
        gte: '>=',
        lte: '<=',
        lt: '<',
        gt: '>'
      }

      #
      # Show all the operators and search variables
      #
      # == Parameters:
      # user::
      #   The user who require the filters
      #
      # == Returns:
      # A Hash operators and variables required
      #
      def self.filter_to_json(_ = User.current_user, _ = true)
        {
          operators: {
            text_input: [
              { id: :contains,      value: I18n.t('filters.contains') },
              { id: :contains_not,  value: I18n.t('filters.not_contains') },
              { id: :starts_with,   value: I18n.t('filters.starts_with') },
              { id: :ends_with,     value: I18n.t('filters.ends_with') },
              { id: :regex,         value: I18n.t('filters.regex') },
              { id: :is,            value: I18n.t('filters.is') },
              { id: :is_not,        value: I18n.t('filters.is_not') },
              { id: :empty,         value: I18n.t('filters.NA') },
              { id: :not_empty,     value: I18n.t('filters.not_NA') }
            ],
            integer_input: [
              { id: :is, value: "#{I18n.t 'lib_redborder.is'}" },
              { id: :is_not, value: "#{I18n.t 'lib_redborder.is_not'}" },
              { id: :lt, value: "#{I18n.t 'lib_redborder.less_than'}" },
              { id: :gt, value: "#{I18n.t 'lib_redborder.greater_than'}" },
              { id: :lte, value: "#{I18n.t 'lib_redborder.less_or'}" },
              { id: :gte, value: "#{I18n.t 'lib_redborder.greater_or'}" }
            ],
            select: [
              { id: :is,            value: "#{I18n.t 'lib_redborder.is'}" },
              { id: :is_not,        value: "#{I18n.t 'lib_redborder.is_not'}" },
              { id: :contains,      value: "#{I18n.t 'lib_redborder.contains'}" }
            ],
            datetime: [
              { id: :is,            value: "#{I18n.t 'lib_redborder.is'}" },
              { id: :is_not,        value: "#{I18n.t 'lib_redborder.is_not'}" },
              { id: :gt,            value: "#{I18n.t 'lib_redborder.greater_than'}" },
              { id: :gte,           value: "#{I18n.t 'lib_redborder.greater_or'}" },
              { id: :lt,            value: "#{I18n.t 'lib_redborder.less_than'}" },
              { id: :lte,           value: "#{I18n.t 'lib_redborder.less_or'}" }
            ]
          }
        }
      end

      #
      # Return all filters based on the current dataSource
      #
      # == Parameters:
      # filters::
      #   The current filters
      #
      # == Returns:
      # Hash with filters for the dataSource
      #
      def self.filters(filters)
        columns = self::COLUMN_NAMES.keys
        shared = filters[:shared].select { |k, _| columns.include? k }
        # Return all the filters
        return shared if filters[id].nil?
        shared.merge(filters[id])
      end

      #
      # Check if the DataSource can be filtered by the column
      #
      # == Parameters:
      # column::
      #   The column user want's to filter
      #
      # == Returns:
      # True if the column is supported
      #
      def self.can_filter?(column)
        self::COLUMN_NAMES.keys.include? column
      end

      #
      # Check if a filter is private
      #
      # == Parameters:
      # column::
      #   The column user want's to filter
      #
      # == Returns:
      # True if the column is privated
      #
      def self.private_filter?(column)
        if self::COLUMNS[column]
          self::COLUMNS[column][:shared] == false
        else
          false
        end
      end

      #
      # Get the columns availables to filter.
      #
      # == Returns:
      # Array of Strings
      #
      def self.allowed_columns
        # Return the allowed columns
        self::COLUMN_NAMES.keys
      end

      #
      # Load the objects of the source to return new values.
      #
      # == Parameters:
      # column::
      #   A symbol with the column
      # values::
      #   Array with values
      #
      # == Returns:
      # Array of Strings
      #
      def self.parse_objects(column, values)
        return [] if self::COLUMNS[column].nil? ||
                      self::COLUMNS[column][:object_type].nil?
        # Get the values from the objects
        object_type = self::COLUMNS[column.to_sym][:object_type]
        # Check the number of elements to modify the query
        if values.size == 1
          RedborderObject.where(
            'LOWER(name) LIKE :search AND type = :type AND user_id = :user',
            search: "%#{values.first.downcase}%",
            type: "#{object_type}",
            user: User.current_user
          ).map(&:value)
        else
          RedborderObject.where(
            'LOWER(name) SIMILAR TO :search AND type = :type AND user_id = :user',
            search: "(#{values.join('|').downcase})%",
            type: "#{object_type}",
            user: User.current_user
          ).map(&:value)
        end
      end

      # -------------------------------------------------------------
      # Conversion methods
      # -------------------------------------------------------------

      #
      # Get a column and return the tab name for that column. This function is
      # used when redirect from applied filter to display the tab in the UI.
      #
      # == Parameters:
      # column::
      #   An String with the current column
      #
      # == Returns:
      # String of the tab
      #
      def self.fetch_tab(dimension)
        self::COLUMNS.each do |tab, values|
          values[:columns].each do |column, _|
            return tab.to_s if column.to_s == dimension
          end
        end
        # Return an empty string if we can't found the tab
        ''
      end

      #
      # TODO: Refactor this
      #
      def self.data_column(column, value)
        return 'N/A' if value.blank?
        send("convert_#{column}", value)
      rescue NoMethodError
        value
      end

      def self.convert_severity(value)
        ## Discard convert if Severity is related to Vulnerability Scanner
        if %w[CRITICAL HIGH MEDIUM LOW].include?(value)
          return value
        else
          Severity.find(value.to_i).name
        end
      end

      def self.convert_start_time(value)
        Time.zone.at(value.to_i).strftime('%y-%m-%d %H:%M')
      end

      def self.convert_end_time(value)
        Time.zone.at(value.to_i).strftime('%y-%m-%d %H:%M')
      end

      def self.convert_domain(value)
        Sensor.where(id: value.to_i).blank? ? value : Sensor.find(value.to_i).name
      end

      def self.convert_group_id(value)
        Sensor.find(value.to_i).complete_name
      end
    end
  end
end
