#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'geoip'

module Redborder
  module Geoip

    PATH   = File.new('/usr/share/GeoIP/city.dat') if File.exist?('/usr/share/GeoIP/city.dat')
    PATHv6 = File.new('/usr/share/GeoIP/cityv6.dat') if File.exist?('/usr/share/GeoIP/cityv6.dat')

    FAKE_DATA = {
      :country_code2 => "N/A"
    }

    def self.databaseCity?
      File.exists?(PATH) and !File.zero?(PATH)
    end

    #
    # Parse an array of IPs to get the positions of these devices based on
    # GeoIP database.
    #
    # == Parameters:
    # collection::
    #   An array of Strings, contain the IPs.
    #
    # == Returns:
    # An array of hashes. The format of objects are:
    #   {
    #     lat: Latitude of the IP
    #     lng: longitude of the IP
    #     description: an HTML code of the tooltip to set the info in the map.
    #   }
    #
    def self.latlong(collection)
      databaseCity   = GeoIP.new(PATH)
      databaseCityv6 = GeoIP.new(PATHv6)
      var = []
      collection.each do |ip|
        next if ip.nil?
        desc = ''

        ip_data = if IPAddr.new(ip).ipv4?
                    databaseCity.city(ip)
                  else
                    databaseCityv6.city(ip)
                  end

        next if ip_data.nil?

        ip_data.to_hash.each do |key, value|
          next unless [:ip, :latitude, :longitude, :country_name,
                       :city_name, :timezone].include?(key)
          # HTML code to display the tooltip
          desc += "<b>#{key}:</b> #{value.to_s.encode('utf-8')}<br/>"
        end

        var << {
                 lat: ip_data[:latitude],
                 lng: ip_data[:longitude],
                 description: "<html>#{desc}</html>"
               }
      end
      # Return the data
      var
    rescue
      []
    end

    #
    # Get the positions of a given IP based on
    # GeoIP database.
    #
    # == Parameters:
    # string::
    #   A String, contain the IP.
    #
    # == Returns:
    # A hash. The format of hash is:
    #   {
    #     lat: Latitude of the IP
    #     lng: longitude of the IP
    #   }
    #
    def self.geolocate(ip)
      var = {}
      return var unless ip

      databaseCity   = GeoIP.new(PATH)
      databaseCityv6 = GeoIP.new(PATHv6)

      ip_data = if IPAddr.new(ip).ipv4?
                  databaseCity.city(ip)
                else
                  databaseCityv6.city(ip)
                end

      return var unless ip_data

      var[:latitude] = ip_data[:latitude] if ip_data[:latitude].present?
      var[:longitude] = ip_data[:longitude] if ip_data[:longitude].present?

      # Return the data
      var
    rescue StandardError
      {}
    end

    def self.check_database(file, type)
      case type.to_sym
      when :city, :cityv6
        GeoIP.new(file).city("8.8.8.8")
      when :country, :countryv6
        GeoIP.new(file).country("8.8.8.8")
      when :asn
        GeoIP.new(file).asn("8.8.8.8")
      end
      return true
    rescue Exception
      Rails.logger.info $!.inspect
      Rails.logger.info $@.join("\n")
      return false
    end
  end
end
