#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Miguel Ávarez <malvarez@redborder.com>
#######################################################################

#
# Base redBorder module
#
module Redborder
  #
  # Incidents with outliers
  # The objective is to create new incidents based on outliers
  # Written by a lone developer who swore they’d "never forget" this code... 
  # Now only God truly understands what’s happening here.
  #
  class IncidentsOutliers    
    #
    # Creates incidents for all flow sensors in the system.
    #
    # == Returns:
    # The number of incidents created.
    #
    def self.create_incidents_for_all_sensors
      log_content = "Executing create_incidents_for_all_sensors.. \n"

      sensors = Sensor.get_all_flow_sensors
      incident_count = 0

      sensors.each do |sensor|
        log_content += "Checking sensor #{sensor[:sensor_name]} \n"
        log_function_call = ""
        
        result, log_function_call = create_incident_of_outlier(sensor[:sensor_name], sensor[:sensor_uuid], sensor[:namespace_uuid], sensor[:domain_uuid])
        
        log_content += log_function_call

        incident_count += 1 if result && result[:success]
      end

      return incident_count, log_content
    end

    private

    #
    # Creates and saves an incident based on the provided data.
    #
    # == Parameters:
    # sensor_name::
    #   The name of the sensor for the incident.
    # outlier_data::
    #   The data related to the outliers.
    # domain_uuid::
    #   The domain UUID to associate with the incident.
    #
    # == Returns:
    # A hash containing success status and the incident ID if created successfully.
    #
    def self.create_and_save_incident(sensor_name, sensor_uuid, outlier_data, domain_uuid)
      Incident.transaction do
        incident = build_incident(sensor_name, domain_uuid)
    
        if incident.save
          outlier_data[:outliers_data].each do |data|
            begin
              Outlier.create!(
                sensor_name: sensor_name,
                sensor_uuid: sensor_uuid,
                lan_ip_regex: outlier_data[:lan_ip_regex],
                wan_ip_regex: outlier_data[:wan_ip_regex],
                observables: outlier_data[:observables],
                priority: incident.priority.name,
                timestamp: DateTime.parse(data["timestamp"]),
                range: outlier_data[:range],
                data: data.to_json,
                incident: incident
              )
            rescue => e
              raise ActiveRecord::Rollback
            end
          end
          { success: true, id: incident.id }
        else
          { success: false, errors: incident.errors.full_messages }
        end
      end
    end

    #
    # Builds an incident object from the outliers data.
    #
    # == Parameters:
    # sensor_name::
    #   The name of the sensor for the incident.
    # outliers_data::
    #   The outliers data to associate with the incident.
    # domain_uuid::
    #   The domain UUID to associate with the sensor.
    #
    # == Returns:
    # An incident object.
    #
    def self.build_incident(sensor_name, domain_uuid)
      sensor = Sensor.find_by(uuid: domain_uuid) || Sensor.root
      Incident.new(
        name: "Network Traffic Anomalies on sensor #{sensor_name}",
        priority: Priority.find_by(name: Setting.outliers_incidents_severity) || Priority.find_by(name: IncidentOutliersUtils::PRIORITY_UNKNOWN),
        uuid: SecureRandom.uuid,
        incident_status: IncidentStatus.find_by(name: IncidentStatus::NEW),
        source: Redborder::Module.registered_modules[:flow][:incident_source_name],
        sensor_id: sensor.id,
        visible: true
      )
    end

    #
    # Creates an incident based on outliers for a specific sensor.
    #
    # == Parameters:
    # sensor_name::
    #   The name of the sensor for which the incident is being created.
    # sensor_uuid::
    #   The sensor UUID.
    # namespace_uuid::
    #   The namespace UUID for the sensor.
    # domain_uuid:: UUID of the sensor to create the incident
    # priority::
    #   The priority of the incident (default is 'unknown').
    #
    # == Returns:
    # A hash containing success status and the incident ID if created successfully.
    #
    def self.create_incident_of_outlier(sensor_name, sensor_uuid, namespace_uuid, domain_uuid, priority: IncidentsOutliersUtils::PRIORITY_UNKNOWN)
      content_log = "Executing create_incident_of_outlier..\n"
      
      range = Setting.outliers_incidents_range
      
      content_log += "Range: #{range} \n" if range

      content_log += "Calculating incident outlier range ..\n"
      
      start_time, end_time = IncidentsOutliersUtils::RangeUtils.calculate_incident_outlier_range(nil, range)
      
      content_log += "The start_time: #{start_time} \n" if start_time
      content_log += "The end_time: #{end_time} \n" if end_time

      content_log += "Calculating top lan ips and tops wan ips..\n"
      
      tops_lan_ips, tops_wan_ips = IncidentsOutliersUtils::DataRetrieval.fetch_top_ips(sensor_uuid, start_time, end_time, range)
      
      content_log += "The tops lan ips: #{tops_lan_ips} \n" if tops_lan_ips
      content_log += "The tops wan ips: #{tops_wan_ips} \n" if tops_wan_ips
      
      return nil, content_log if tops_lan_ips.empty? && tops_wan_ips.empty?

      content_log += "Calling build ip regex.. \n"
      
      lan_ip_regex, wan_ip_regex = IncidentsOutliersUtils::IPUtils.build_ip_regex(tops_lan_ips, tops_wan_ips)

      content_log += "Fetching outlier data..\n"

      outlier_data = IncidentsOutliersUtils::Outliers.fetch_outlier_data(namespace_uuid, sensor_uuid, start_time, end_time, lan_ip_regex, wan_ip_regex, range)

      return nil, content_log if outlier_data[:outliers_data].nil? || outlier_data[:outliers_data].empty?

      outliers_by_dimension = {
        IncidentsOutliersUtils::LAN_IP_DIMENSION => [],
        IncidentsOutliersUtils::WAN_IP_DIMENSION => []
      }
      
      ip_ips = [tops_lan_ips, tops_wan_ips]

      [IncidentsOutliersUtils::LAN_IP_DIMENSION, IncidentsOutliersUtils::WAN_IP_DIMENSION].each_with_index do |dimension, index|
        ips = ip_ips[index]
        outliers_by_dimension[dimension] = IncidentsOutliersUtils::Outliers.process_outliers_for_ip_type(
          ips, start_time, end_time, range, sensor_uuid, dimension, outlier_data[:outliers_data]
        )
      end

      merged_outliers = IncidentsOutliersUtils::IPUtils.merge_ip_arrays(
        outliers_by_dimension[IncidentsOutliersUtils::LAN_IP_DIMENSION], outliers_by_dimension[IncidentsOutliersUtils::WAN_IP_DIMENSION]
      )

      return nil, content_log if merged_outliers.empty?

      outlier_data[:wan_ip_regex] = wan_ip_regex
      outlier_data[:lan_ip_regex] = lan_ip_regex
      outlier_data[:observables] << merged_outliers

      content_log += "Creating and saving incident.. \n"

      return create_and_save_incident(sensor_name, sensor_uuid, outlier_data, domain_uuid), content_log
    end
  end
end
