#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Miguel Ávarez <malvarez@redborder.com>
#######################################################################

#
# Base redBorder module
#
module Redborder
  #
  # Incidents outliers Utils
  # Written by a lone developer who swore they’d "never forget" this code... 
  # Now only God truly understands what’s happening here.
  #
  module IncidentsOutliersUtils
    
    # ==========================================================
    # Constants
    # ==========================================================

    OUTLIERS_FLOW_LIMIT = 100
    PRIORITY_UNKNOWN = 'unknown'.freeze
    OUTLIERS_DRUID_DEFAULT_TIMEOUT = 90_000
    OUTLIERS_DRUID_TOPS_LIMIT = 10
    LAN_IP_DIMENSION = "lan_ip"
    WAN_IP_DIMENSION = "wan_ip"
    FLOW_SOURCE = 'flow'

    # ==========================================================
    # Modules and methods
    # ==========================================================

    module RangeUtils
      # Maps a range to its appropriate time granularity setting.
      # 
      # == Parameters:
      # - range:: The time range to convert into granularity.
      #
      # == Returns:
      # A string representing the granularity for the specified range.
      def self.range_to_granularity(range)
        {
          'last_1_hour' => 'pt1m',
          'last_2_hour' => 'pt1m',
          'last_3_hour' => 'pt1m',
          'last_12_hour' => 'pt5m',
          'last_24_hour' => 'pt5m',
          'last_week' => 'pt60m'
        }.fetch(range, 'pt1m')
      end

      # Calculates the start and end times for an outlier incident based on a specified range.
      #
      # == Parameters:
      # - incident:: The existing incident or nil if calculating for a new incident.
      # - range:: The time range to use for calculation.
      #
      # == Returns:
      # An array containing the calculated start and end times.
      def self.calculate_incident_outlier_range(incident, range)
        end_time = incident ? incident.created_at : Time.now
        start_time = end_time

        case range.to_sym
        when :live
          start_time -= 3.minutes
        when :last_5_minute
          start_time -= 5.minutes
        when :last_15_minute
          start_time -= 15.minutes
        when :last_30_minute
          start_time -= 30.minutes
        when :last_1_hour
          start_time -= 1.hour
        when :last_2_hour
          start_time -= 2.hours
        when :last_3_hour
          start_time -= 3.hours
        when :last_12_hour
          start_time = (end_time - 12.hours).beginning_of_hour
        when :last_24_hour
          start_time = (end_time - 24.hours).beginning_of_hour
        when :last_week
          start_time = (end_time - 1.week).beginning_of_hour
        else
          start_time -= 1.hour
        end
        [start_time, end_time]
      end
    end

    # Module for handling IP-related utilities, such as merging IP arrays and building regex for IP lists.
    module IPUtils
      # Merges two arrays of IPs, avoiding duplicates, and tags each with the specified type.
      #
      # == Parameters:
      # - arr1:: First array of IPs.
      # - arr2:: Second array of IPs.
      # - type:: Tag type for the IPs, default is 'ip'.
      #
      # == Returns:
      # An array of hashes where each IP is tagged with the type.
      def self.merge_ip_arrays(arr1, arr2, type = "ip")
        (arr1 + arr2).uniq.map { |ip| { ip => type } }
      end

      # Converts an array of IP addresses into a regular expression string.
      #
      # == Parameters:
      # - ips:: Array of IPs to convert into a regex pattern.
      #
      # == Returns:
      # A string containing the regex pattern for the provided IPs.
      def self.ips_array_to_regex(ips)
        ips = ips.any? ? "(?i)^.*(#{ips.join('|')}).*$" : nil
        ips
      end

      # Builds regex patterns for LAN and WAN IPs.
      #
      # == Parameters:
      # - tops_lan_ips:: Array of top LAN IPs.
      # - tops_wan_ips:: Array of top WAN IPs.
      #
      # == Returns:
      # An array of regex patterns for the LAN and WAN IPs.
      def self.build_ip_regex(tops_lan_ips, tops_wan_ips)
        [tops_lan_ips, tops_wan_ips].map { |ips| IPUtils.ips_array_to_regex(ips) }
      end
    end

    # Module for traffic data formatting and retrieval.
    module TrafficData
      # Formats traffic data for a specific IP address.
      #
      # == Parameters:
      # - traffic_data:: Array of traffic data.
      # - ip_address:: The IP address for which traffic data is formatted.
      #
      # == Returns:
      # A hash containing the formatted traffic data for the specified IP.
      def self.format_traffic_data(traffic_data, ip_address)
        { ip_address => traffic_data.map do |traffic|
            {
              'timestamp' => traffic.timestamp.iso8601,
              'result' => { 'bytes' => traffic.bytes }
            }
          end }
      end

      # Fetches and formats IP data for the top IPs within a specified time range.
      #
      # == Parameters:
      # - top_ips:: Array of top IPs.
      # - start_time:: Start time for data retrieval.
      # - end_time:: End time for data retrieval.
      # - range:: Time range identifier.
      # - sensor_uuid:: UUID of the sensor.
      # - ip_type:: Type of the IP (LAN or WAN).
      #
      # == Returns:
      # A hash containing formatted IP data for the specified IPs.
      def self.fetch_and_format_ip_data(top_ips, start_time, end_time, range, sensor_uuid, ip_type)
        ip_data = {}
        top_ips.each do |ip|
          ip_data[ip] = format_traffic_data(
            DataRetrieval.get_all_ips_data(start_time, end_time, range, sensor_uuid, ip, ip_type),
            ip
          )[ip]
        end
        ip_data
      end
    end

    # Module for data retrieval tasks, such as fetching IP data and top IP lists.
    module DataRetrieval
      # Retrieves traffic data for all IPs over a specified time range.
      #
      # == Parameters:
      # - start_date:: Start date for data retrieval.
      # - end_date:: End date for data retrieval.
      # - range:: Time range for data granularity.
      # - sensor_uuid:: Sensor UUID.
      # - ip:: IP address for the data.
      # - dimension:: Dimension identifier (LAN or WAN).
      #
      # == Returns:
      # An array of traffic data for the specified IP.
      def self.get_all_ips_data(start_date, end_date, range, sensor_uuid, ip, dimension)
        source = Redborder::DruidSource.get(source: FLOW_SOURCE)
        filter = Redborder::Filter.new(source, {})
        filter.add(:sensor_uuid, sensor_uuid)
        filter.add(dimension.to_sym, ip)
        source.where(
          query_type: :timeseries,
          start_time: start_date,
          end_time: end_date,
          granularity: RangeUtils.range_to_granularity(range).to_sym,
          range: range,
          search: filter.get,
          aggregation: :bytes,
          user: User.current_user || User.first
        )
      end

      # Retrieves the top IP addresses for a sensor over a specified time range.
      #
      # == Parameters:
      # - start_date:: Start date for IP retrieval.
      # - end_date:: End date for IP retrieval.
      # - range:: Time range identifier.
      # - sensor_uuid:: Sensor UUID.
      # - tab:: IP type tab (e.g., LAN or WAN).
      #
      # == Returns:
      # An array of top IPs for the specified sensor and time range.
      def self.get_tops_ips(start_date, end_date, range, sensor_uuid, tab)
        source = Redborder::DruidSource.get(source: FLOW_SOURCE)
        filter = Redborder::Filter.new(source, {})
        filter.add(:sensor_uuid, sensor_uuid)
        ips = source.topN(
          query_type: :group_by,
          start_time: start_date,
          end_time: end_date,
          granularity: :all,
          range: range,
          limit: OUTLIERS_DRUID_TOPS_LIMIT,
          group_by: [source::COLUMNS[tab.to_sym][:columns].keys[0]],
          search: filter.get,
          aggregation: :bytes,
          user: User.current_user || User.first
        )
        ips.map { |traffic| traffic.public_send(tab) }
      end

      #
      # Fetches the top LAN and WAN IPs for a sensor.
      #
      # == Parameters:
      # sensor_uuid::
      #   The sensor uuid.
      # start_time::
      #   The start time for fetching data.
      # end_time::
      #   The end time for fetching data.
      # range::
      #   The time range for the analysis.
      #
      # == Returns:
      # An array containing the top LAN IPs and top WAN IPs.
      #
      def self.fetch_top_ips(sensor_uuid, start_time, end_time, range)
        [LAN_IP_DIMENSION, WAN_IP_DIMENSION].each_with_object([]) do |dimension, result|
          result << get_tops_ips(start_time, end_time, range, sensor_uuid, dimension)
        end
      end
    end

    # Module for outlier detection, processing, and query construction.
    module Outliers
      # Processes outliers for a specified IP type by fetching and analyzing traffic data.
      #
      # == Parameters:
      # - top_ips:: Array of top IPs.
      # - start_time:: Start time for analysis.
      # - end_time:: End time for analysis.
      # - range:: Time range identifier.
      # - sensor_uuid:: Sensor uuid.
      # - ip_type:: IP type (LAN or WAN).
      # - outliers_data:: Data used for outlier detection.
      #
      # == Returns:
      # A list of causes for the detected outliers.
      def self.process_outliers_for_ip_type(top_ips, start_time, end_time, range, sensor_uuid, ip_type, outliers_data)
        all_top_ip_data = TrafficData.fetch_and_format_ip_data(top_ips, start_time, end_time, range, sensor_uuid, ip_type)
        extract_outliers(outliers_data, top_ips, all_top_ip_data)
      end

      # Extracts the causes for the detected outliers based on provided data.
      #
      # == Parameters:
      # - outliers_data:: Data used to identify outliers.
      # - top_ips:: Array of top IPs.
      # - all_top_ip_data:: Complete traffic data for the top IPs.
      #
      # == Returns:
      # An array of reasons for each outlier.
      def self.extract_outliers(outliers_data, top_ips, all_top_ip_data)
        outliers = Redborder::AIoutliers.get_outliers_ips(outliers_data, top_ips, all_top_ip_data)
        outliers["ips"].map { |ip_data| ip_data["caused_by"] }.flatten
      end

      # Retrieves outliers for a specified sensor by querying historical data.
      #
      # == Parameters:
      # - query:: Query parameters for outlier search.
      # - range:: Time range identifier (defaults to system setting).
      #
      # == Returns:
      # A hash containing the detected outliers and observables.
      def self.fetch_outliers_for_sensor(query, range = Setting.outliers_incidents_range)
        outliers_data = []
        begin
          outlier_data = Redborder::AIoutliers.outlier_r(query, {})
          outliers_data = outlier_data["anomalies"] if outlier_data
        rescue StandardError => e
          Rails.logger.error("Error calculating outliers: #{e.message}")
        end
        { outliers_data: outliers_data, observables: [], range: range, priority: Setting.outliers_incidents_severity }
      end

      #
      # Fetches outlier data for a specific sensor.
      #
      # == Parameters:
      # sensor_uuid::
      #   The uuid of the sensor.
      # start_time::
      #   The start time for fetching data.
      # end_time::
      #   The end time for fetching data.
      # lan_ip_regex::
      #   The regex pattern for LAN IPs.
      # wan_ip_regex::
      #   The regex pattern for WAN IPs.
      # range::
      #   The time range for the analysis.
      #
      # == Returns:
      # A hash containing the outlier data.
      #
      def self.fetch_outlier_data(namespace_uuid, sensor_uuid, start_time, end_time, lan_ip_regex, wan_ip_regex, range)
        query = build_outlier_query(start_time, end_time, lan_ip_regex, wan_ip_regex, namespace_uuid, sensor_uuid, range)
        fetch_outliers_for_sensor(query, range)
      end

      # Builds a query for outlier data retrieval based on provided IP patterns and range.
      #
      # == Parameters:
      # - start_date:: Start date for the query.
      # - end_date:: End date for the query.
      # - lan_ip_regex:: Regular expression for LAN IPs.
      # - wan_ip_regex:: Regular expression for WAN IPs.
      # - namespace_uuid:: Namespace UUID.
      # - sensor_uuid:: Sensor UUID.
      # - range:: Time range identifier.
      #
      # == Returns:
      # A hash representing the query for outlier detection.
      def self.build_outlier_query(start_date, end_date, lan_ip_regex, wan_ip_regex, namespace_uuid, sensor_uuid, range)
        data_source = namespace_uuid.present? ? "rb_flow_#{namespace_uuid}" : "rb_flow"
        {
          dataSource: data_source,
          granularity: { type: 'period', period: RangeUtils.range_to_granularity(range), origin: start_date.utc.iso8601 },
          intervals: ["#{start_date.utc.iso8601}/#{end_date.utc.iso8601}"],
          filter: {
            type: 'and',
            fields: [
              { type: 'regex', dimension: LAN_IP_DIMENSION, pattern: lan_ip_regex },
              { type: 'regex', dimension: WAN_IP_DIMENSION, pattern: wan_ip_regex },
              { type: 'selector', dimension: 'sensor_uuid', value: sensor_uuid }
            ]
          },
          queryType: 'timeseries',
          context: { timeout: OUTLIERS_DRUID_DEFAULT_TIMEOUT, skipEmptyBuckets: true },
          limitSpec: { type: 'default', limit: OUTLIERS_FLOW_LIMIT, columns: [] },
          aggregations: [{ type: 'longSum', name: 'bytes', fieldName: 'sum_bytes' }],
          postAggregations: []
        }
      end
    end
  end
end
