#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'open-uri'
require 'digest/md5'

module Redborder
  module Jobs
    class CheckPluginsJob < Struct.new(:user_id, :force)
      include Redborder::Jobs::JobHelper

      def perform
        @content = "#{I18n.t('lib_redborder.node')} #{`hostname`.chomp!}\n\n"

        begin
          @config = load_plugins_config
          @content += "#{I18n.t('lib_redborder.checking_plugins')}\n\n"

          download_files = false
          clean_name = "plugins.tar.gz"
          md5_file = "#{@config['cache_dir']}/#{clean_name}.md5"
          md5url = @config['md5url']
          tgzurl = @config['tgzurl']
          global_plugins_dir = @config['global_plugins_dir'] || "/opt/rb/var/www/plugins"
          plugins_dir = @config['plugins_dir']

          md5, md5_file_content = check_md5(md5_file, md5url)
          download_files = md5_file_content.nil? || md5 != md5_file_content || force

          if download_files
            @content += "#{I18n.t('lib_redborder.update_found')}\n\n"
            tgz_tmp_file = download_tgz(tgzurl, "#{@config['cache_dir']}/#{clean_name}-tmp.tar.gz", md5_file_content)
          end

          if download_files && plugins_dir
            @content += "Preparing to install plugins..}\n\n"
            install_plugins(tgz_tmp_file, plugins_dir, global_plugins_dir)
          end

          @content += "#{I18n.t('lib_redborder.finish')}\n\n"

        rescue StandardError => e
          log_exception(e)
        ensure
          stored_job = StoredDelayedJob.create(job: "CheckPluginsJob", user_id: user_id, content: @content)
          reschedule_job
        end
      end

      private

      def load_plugins_config
        config_file = File.join(Rails.root, 'config', 'plugins_config.yml')
        YAML.load_file(config_file)
      end

      def check_md5(md5_file, md5_url)
        return nil, nil unless md5_url

        @content += "Checking md5 of #{md5_file} and #{md5_url}\n\n"
        md5 = File.open(md5_file) {|f| f.readline}.gsub(/^"/, '').chomp.split(" ").first if File.exist?(md5_file)

        @content += "Downloading #{md5_url} into #{md5_file}.. \n\n"
        download_file(md5_url, md5_file)
        @content += "Download of #{md5_url} finished.\n\n"

        md5_file_content = File.open(md5_file) {|f| f.readline}.gsub(/^"/, '').chomp.split(" ").first
        return md5, md5_file_content
      end

      def download_tgz(tgz_url, tgz_tmp_file, md5_file_content)
        @content += "Preparing to download #{tgz_url} into #{tgz_tmp_file}\n\n"
      
        download_file(tgz_url, tgz_tmp_file)
        @content += "Download of #{tgz_url} finish.\n\n"
        new_md5 = Digest::MD5.hexdigest(File.read(tgz_tmp_file))
        
        @content += "Checking if valid download (md5 check)..\n\n"
        if md5_file_content && md5_file_content != new_md5
          @content += "Downloaded file is corrupted.\n\n"
          raise "Downloaded file is corrupted"
        end

        @content += "File was successfully downloaded.\n\n"
      
        tgz_tmp_file
      end

      def install_plugins(tgz_tmp_file, plugins_dir, global_plugins_dir)
        managers = Manager.all.map(&:name)
        myhost = `hostname -s`.strip
        managers.unshift(managers.delete(myhost))

        @content += "#{I18n.t('lib_redborder.extracting_files')}\n\n"
        if plugins_dir and !plugins_dir.empty?
          managers.each do |manager|
            @content += `/usr/lib/redborder/bin/rb_manager_utils.sh -e -n #{manager} -s "rm -Rf #{plugins_dir}"`
          end
        end

        system('tar', 'xzf', tgz_tmp_file, '-C', plugins_dir)
        system('sudo chmod -R +x', "#{plugins_dir}/extensions/*")

        count = 0
        plugins = Dir.open("#{plugins_dir}/extensions/").to_a.select { |file| file.start_with?("configure-") }

        plugins.each do |plugin|
          plugin_info = plugin.gsub('configure-', '').gsub('.sh', '').split("-")

          next if plugin_info.count != 2

          name, version = plugin_info

          @content += "#{I18n.t('lib_redborder.processing')} #{name} #{version}..\n\n"

          plugin = Plugin.find_or_create_by(name: name)
          plugin.plugin_versions.find_or_create_by(version: version, installed: false)
        end

        @content += "#{I18n.t('lib_redborder.sync_cluster')}\n\n"
        managers.each do |manager|
          @content += `/usr/lib/redborder/bin/rb_manager_utils.sh -e -n #{manager} -s "-Rf #{plugins_dir}"`
          @content += `/usr/lib/redborder/bin/rb_manager_utils.sh -e -n #{manager} -s "chown -R rb-webui. #{global_plugins_dir}"`
          @content += `/usr/lib/redborder/bin/rb_manager_utils.sh -e -n #{manager} -s "chmod -R +x #{global_plugins_dir}/plugins/extensions/*"`
        end
      end

      def download_file(url, local_file)
        File.delete(local_file) if File.exist?(local_file)
        url = "http://#{url}" unless url.start_with?("http://") || url.start_with?("https://") || url.start_with?("ftp://")
      
        proxy_opts = proxy_settings(url) || {} # Set proxy_opts to an empty hash if no proxy is used
        @content += "#{I18n.t('lib_redborder.downloading_from')} #{url} ..\n\n"
        open(url, proxy_opts.merge(ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE, read_timeout: 10)) do |remote_file|
          File.open(local_file, "wb") do |local_file|
            local_file.write(remote_file.read)
          end
        end
      
        @content += "#{I18n.t('lib_redborder.finish_download')}\n\n"
      end

      def proxy_settings(url)
        return false if !Setting.proxy? || Setting.proxy_address.blank?

        proxy_port = Setting.proxy_port.to_i
        proxy_port = 8080 if proxy_port <= 0
        proxy_addr = Setting.proxy_address
        proxy_addr = "http://#{proxy_addr}" unless proxy_addr.start_with?("http://") || proxy_addr.start_with?("https://")
        proxy = URI.parse("#{proxy_addr}:#{proxy_port}")

        if Setting.proxy_user.present?
          { proxy_http_basic_authentication: [proxy, Setting.proxy_user, Setting.proxy_password] }
        else
          { proxy: proxy }
        end
      end

      def log_exception(exception)
        logit(exception)
        logit(exception.backtrace.join("\n"))
        @content += "#{exception.inspect}\n"
      end

      def reschedule_job
        Delayed::Job.enqueue(Redborder::Jobs::CheckPluginsJob.new(), priority: 10, run_at: Time.now + 1.week)
      end
    end
  end
end
