#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class ImportTreeJob < Struct.new(:verbose, :parent_id, :file_name, :override, :user_id)
      include Redborder::Jobs::JobHelper

      def perform
        content = ('Node executing this job: ' + `hostname`.chomp! + "\n\n")

        @user = User.find(user_id)

        obj = download_s3(file_name)

        content += "Extracting file\n"
        dir = extract_file(file_name)

        content += create_from_file(dir, parent_id, override)
        content += "Import finished successfully\n"

      rescue StandardError => se

        content += "#{se.class}: #{se.message}\n"
        se.backtrace.each do |line|
          content += "#{line}\n"
        end
        logit "#{se.class}: #{se.message}" if verbose

      ensure

        clean_dir(verbose, obj, dir, file_name)

        logit 'Creating StoredDelayedJob' if verbose
        StoredDelayedJob.create(
          job: 'ImportTreeJob',
          user_id: user_id,
          content: content
        )
      end

      def create_from_file(dir, parent_id, override)
        if File.exist? "#{dir}/jsonFiles/sensors.json"
          create_from_json_file(dir, parent_id, parent_id, override)
        else
          create_from_prime_file(dir, parent_id)
        end
      end

      def create_from_json_file(dir, parent_id, top_parent_id, override)
        content = "Creating domains and images\n"
        sensors_file = File.read("#{dir}/jsonFiles/sensors.json")
        sensors_data = JSON.parse(sensors_file)

        sensors_data.each do |sensor|
          existing_sensor = Sensor.find_by(uuid: sensor['uuid'])
          if override && existing_sensor && !existing_sensor.is_root?
            begin
              content += override_sensor(sensor, existing_sensor, dir, parent_id, top_parent_id)
            rescue StandardError => e
              content += "Cant override sensor #{sensor['name']}, with uuid #{sensor['uuid']}.\n"
            end
          elsif existing_sensor.nil?
            begin
              content += create_sensor(sensor, dir, parent_id, top_parent_id)
            rescue StandardError => e
              content += "Cant create sensor #{sensor['name']}, with uuid #{sensor['uuid']}.\n"
            end
          else
            content += "The sensor #{sensor['name']}, with uuid #{sensor['uuid']} could not be created.\n"
          end
        end

        # Locate unassigned APs
        AccessPoint.where(switch_id: nil).where.not(switch_name: nil).each do |ap|
          switch = Sensor.find_by(ip: ap.switch_name, type: Sensor::FLOW)
          ap.update(switch_id: switch.id) if switch
        end

        content
      end

      def create_from_prime_file(dir, parent_id)
        content = "Creating domains and GPS points\n"
        content += process_csv_gps(dir, parent_id)

        content += "Creating images\n"
        content += process_csv_span(dir, parent_id)

        content
      end

      def process_csv_gps(dir, parent_id, content = '')
        csv_file = CSV.read("#{dir}/otherCsv/Gps.csv", headers: true)

        csv_file.each do |row|
          floor_info = create_floor(parent_id, row)
          floor = floor_info[:floor]
          content += floor_info[:content]

          property = floor.property
          property[:gps_points] ||= []
          property[:gps_points] << { x: row['X'], y: row['Y'],
                                     latitude: row['Latitude'],
                                     longitude: row['Longitude'] }

          content += "      * GPS point -> x: #{row['X']}, y: #{row['Y']}, latitude: #{row['Latitude']}, longitude: #{row['Longitude']}\n"

          floor.update(property: property)
        end
        content
      end

      def process_csv_span(dir, parent_id, content = '')
        csv_file = CSV.read("#{dir}/otherCsv/Span.csv", headers: true)

        csv_file.each do |row|
          floor_info = create_floor(parent_id, row)
          floor = floor_info[:floor]
          content += floor_info[:content]

          image_name   = row['Img']
          floor_length = row['Length'].to_f
          floor_width  = row['Width'].to_f

          if floor.property[:gps_points].present?
            overlay_data = overlay_data(floor, floor_length, floor_width)
            update_floor(floor, floor_length, floor_width, overlay_data)
            content += process_overlay(dir, image_name, floor, row)
          elsif File.exist?("#{dir}/images/#{image_name}")
            begin
              update_floor(floor, floor_length, floor_width)
              content += "   * Processing image #{image_name}\n"
              create_overlay_map(dir, image_name, floor, row)
            rescue StandardError => e
              content += "     #{image_name} cant be processed\n"
              content += "     #{e.inspect}\n"
            end
          end
        end
        content
      end

      def create_access_points_from_json(sensor, sensor_json)
        sensor_json['access_points'].try('each') do |access_point|
          if sensor_json['type'].to_i == Sensor::DOMAIN
            ap = AccessPoint.create_with(
              name: access_point['name'],
              ip_address: access_point['ip_address'],
              switch_name: access_point['switch_name'],
              latitude: access_point['latitude'],
              longitude: access_point['longitude'],
              sensor_id: sensor.id,
              mac_address: access_point['mac_address'],
              x: access_point['x'],
              y: access_point['y'],
              switch_id: nil,
              enrichment: access_point['enrichment']
            ).find_or_create_by(
              mac_address: access_point['mac_address']
            )

            values = {}
            %w(latitude longitude x y).each do |i|
              values[i.to_sym] = access_point[i] if ap.send(i).nil?
            end

            # Assign sensor flow to AP (if it exists)
            switch = Sensor.find_by(ip: ap.switch_name, type: Sensor::FLOW)
            values[:switch_id] = switch.id if switch

            ap.update(values) if values.present?

          elsif sensor_json['type'].to_i == Sensor::FLOW ||
                sensor_json['type'].to_i == Sensor::MERAKI
            ap = AccessPoint.create_with(
              name: access_point['name'],
              ip_address: access_point['ip_address'],
              switch_name: access_point['switch_name'],
              latitude: access_point['latitude'],
              longitude: access_point['longitude'],
              sensor_id: nil,
              mac_address: access_point['mac_address'],
              x: access_point['x'],
              y: access_point['y'],
              switch_id: sensor.id,
              enrichment: access_point['enrichment']
            ).find_or_create_by(
              mac_address: access_point['mac_address']
            )

            values = {}
            %w(latitude longitude x y).each do |i|
              values[i.to_sym] = access_point[i] if ap.send(i).nil?
            end

            ap.update(values) if values.present?

          end
        end
      end

      def create_zones_from_json(sensor, sensor_json)
        sensor_json['zones'].try('each') do |zone|
          if sensor_json['type'].to_i == Sensor::DOMAIN &&  sensor_json['domain_type'].to_i == Sensor::FLOOR
            z = Zone.create(
              name: zone['name'],
              description: zone['description'],
              sensor_id: sensor.id
            )

            # Everytime we create a zone, we check if its associated with an access_points
            # and if it is, we create the Relation (like in add_access_point in zones_controller)
            if !zone['mac_address'].empty?
              ap = AccessPoint.where(mac_address: zone['mac_address'])
              if !ap.nil?
                if !z.access_points.include? ap
                  z.access_points << ap
                end
              end
            end
          end
        end
      end

      def create_overlay_maps(dir, sensor_id, sensor)
        maps_file = File.read("#{dir}/jsonFiles/maps.json")
        maps_data = JSON.parse(maps_file)

        map_json = maps_data[sensor_id.to_s]

        return unless map_json
        return unless File.exist?("#{dir}/images/rotated/#{sensor_id}")
        return unless File.exist?("#{dir}/images/#{sensor_id}")
        overlay_map = OverlayMap.create_with(ne: map_json['ne'],
                                             sw: map_json['sw'],
                                             name: map_json['name'],
                                             rotation: map_json['rotation'],
                                             sensor_id: sensor.id,
                                             image: File.open("#{dir}/images/rotated/#{sensor_id}"),
                                             original_image: File.open("#{dir}/images/#{sensor_id}"))
                      .find_or_create_by(sensor_id: sensor.id)

        # Create GPS Markers
        map_json['gps_markers'].each do |gps_marker|
          GpsMarker.find_or_create_by(
            overlay_map_id: overlay_map.id,
            x: gps_marker['x'],
            y: gps_marker['y'],
            latitude: gps_marker['latitude'],
            longitude: gps_marker['longitude']
          )
        end
      end

      def overlay_data(floor, floor_length, floor_width)
        gps_markers = floor.property[:gps_points].map do |point|
          GpsMarker.new point
        end

        PrimeInfrastructure::Functions.overlay_data(
          [floor_width, floor_length],
          gps_markers
        )
      end

      def process_image(dir, image_name, floor)
        image = Magick::ImageList.new("#{dir}/images/#{image_name}")
        image.background_color = 'none'
        image = image.rotate(floor.property[:rotation])
        image.write("#{dir}/images/rotated/#{image_name}")
      end

      def process_overlay(dir, image_name, floor, row)
        return '' unless File.exist?("#{dir}/images/#{image_name}")

        process_image(dir, image_name, floor)

        overlay_map = create_overlay_map(dir, image_name, floor, row)

        return '' unless floor.property[:gps_points].present? && overlay_map.errors.empty?
        floor.property[:gps_points].each do |point|
          GpsMarker.create(
            x: point[:x],
            y: point[:y],
            longitude: point[:longitude],
            latitude: point[:latitude],
            overlay_map_id: overlay_map.id
          )
        end

        "   * Processing image #{image_name}\n"
      end
    end
  end
end
