#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder

  module Jobs

    module JobHelper

      def logit(msg, show_sensor = true)
        if show_sensor and @sensor
          STDOUT.puts "Sensor #{@sensor.id}: #{msg}"
        else
          STDOUT.puts msg.to_s
        end
      end

      def override_sensor(sensor, existing_sensor, dir, parent_id, top_parent_id)
        # Check if there is no other sensor/domain with the
        # same parent_id, type and domain_type.
        # It is a model's validation.
        cant_destroy = Sensor.where.not(uuid: sensor['uuid'])
                             .where(name: sensor['name'],
                                    parent_id: sensor['parent_id'],
                                    type: sensor['type'],
                                    domain_type: sensor['domain_type'])

        if cant_destroy.empty? && @user.can?(:manage, existing_sensor)
          # Destroy the existing sensor and its descendants,
          # and update users
          ancestor = existing_sensor
                     .ancestors
                     .where(type: Sensor::DOMAIN).reverse.first || Sensor.root
          destroyed_ids = existing_sensor.self_and_descendants.map(&:id)
          User.unscoped.where(sensor_id: destroyed_ids)
              .update_all(sensor_id: ancestor, enabled: false)

          existing_sensor.destroy

          # Create the new sensor with the uuid
          create_sensor(sensor, dir, parent_id, top_parent_id)
        else
          "The sensor #{sensor['name']}, with uuid #{sensor['uuid']} could not be created.\n"
        end
      end

      def create_sensor(sensor_json, dir, parent_id, top_parent_id)
        content = ''

        sensor_id = sensor_json['id']

        sensor = create_sensor_by_type(sensor_json, parent_id)

        if sensor.errors.present?
          content += "Sensor/Domain: #{sensor_json['name']} - #{sensor.errors.full_messages.join('. ')}\n"
          return content
        elsif (sensor.is_flow? || sensor.is_meraki?) && @user.cannot?(:manage, sensor)
          return content
        end

        # If it has access points
        # Access Points can belongs to Flow Sensors or Floor Sensors
        create_access_points_from_json(sensor, sensor_json)

        # If it has zones
        # Zones can belongs  Floor Sensors
        create_zones_from_json(sensor, sensor_json)

        # Create overlay map
        create_overlay_maps(dir, sensor_id, sensor)

        # Create children
        sensor_json['children'].try('each') do |sensor_child|
          content += create_sensor(sensor_child, dir, sensor.id, top_parent_id)
        end

        content
      end

      def create_sensor_by_type(sensor_json, parent_id)
        fields = {
          name: sensor_json['name'],
          parent_id: parent_id,
          ip: sensor_json['ip'],
          latitude: sensor_json['latitude'],
          longitude: sensor_json['longitude'],
          type: sensor_json['type'],
          property: sensor_json['property'],
          domain_type: sensor_json['domain_type'],
          unassigned: sensor_json['unassigned'],
          uuid: sensor_json['uuid']
        }

        sensor_type = sensor_json['type'].to_i

        case sensor_type
        when Sensor::DOMAIN
          sensor = Sensor.create_with(fields).find_or_create_by(
            name: sensor_json['name'],
            parent_id: parent_id,
            type: sensor_json['type'],
            domain_type: sensor_json['domain_type']
          )

        when Sensor::FLOW, Sensor::VAULT, Sensor::ARUBACENTRAL, Sensor::ALE, Sensor::SCANNER
          additional_fields = {}
          Sensor.import_columns_for[sensor_json['type'].to_i].each do |field|
            if field.is_a? Symbol
              additional_fields[field] = sensor_json[field.to_s] if sensor_json.key?(field.to_s)
            elsif field.is_a? Hash
              field.keys.each do |field_|
                additional_fields[field_] = sensor_json[field_.to_s] if sensor_json.key?(field_.to_s)
              end
            end
          end
          fields.merge!(additional_fields)

          find_by_attributes = {
            name: sensor_json['name'],
            type: sensor_json['type']
          }

          if sensor_type == Sensor::SCANNER
            find_by_attributes[:parent_id] = parent_id
          else
            find_by_attributes[:ip] = sensor_json['ip']
          end

          sensor = Sensor.create_with(fields).find_or_create_by(find_by_attributes)
        else
          sensor = Sensor.create_with(fields).find_or_create_by(
            name: sensor_json['name'],
            parent_id: parent_id,
            type: sensor_json['type']
          )
        end
      end

      def create_floor(parent_id, row)
        content = ''
        # Find or create Campus
        campus = create_domain_from_type('Campus', parent_id, row)
        content += "   * Campus: #{campus.name}\n" if campus

        # Find or create Building
        building = create_domain_from_type('Building', campus.id, row)
        content += "   * Building: #{building.name}\n" if building

        # Find or create Floor
        floor = create_domain_from_type('Floor', building.id, row)
        content += "   * Floor: #{floor.name}\n" if floor

        { floor: floor, content: content }
      end

      def create_domain_from_type(type, parent_id, row)
        Sensor.create_with(type: Sensor::DOMAIN,
                           parent_id: parent_id,
                           domain_type: "Sensor::#{type.upcase}".constantize,
                           unassigned: false,
                           name: row[type])
              .find_or_create_by(type: Sensor::DOMAIN,
                                 domain_type: "Sensor::#{type.upcase}".constantize,
                                 name: row[type],
                                 parent_id: parent_id)
      end

      def update_floor(floor, floor_length, floor_width, overlay_data = nil)
        properties = floor.property

        properties[:dimensions] = {
          width: floor_width,
          height: floor_length
        }

        if overlay_data
          properties[:coordinates] = overlay_data.except(:rotation)
          properties[:rotation] = overlay_data[:rotation]
          properties[:matrix_trans]  = overlay_data[:matrix_trans]
          properties[:matrix_offset] = overlay_data[:matrix_offset]
        end

        floor.update(property: properties)
      end

      def create_overlay_map(dir, image_name, floor, row)
        gps_points = floor.property[:gps_points].present?
        OverlayMap.create(
          ne: gps_points ? floor.property[:coordinates][:floor_ne] : '0.0,0.0',
          sw: gps_points ? floor.property[:coordinates][:floor_sw] : '0.0,0.0',
          original_image: File.new("#{dir}/images/#{image_name}"),
          image: gps_points ? File.new("#{dir}/images/rotated/#{image_name}") : File.new("#{dir}/images/#{image_name}"),
          rotation: gps_points ? floor.property[:rotation] : 0,
          name: row['Img'],
          sensor_id: floor.id
        )
      end

      def extract_file(file_name)
        # Create unique dir to prevent conflicts with other ImportTreeJobs
        dir = "#{Rails.root}/tmp/imported_files/#{file_name}"

        FileUtils.mkdir_p dir
        system('tar', 'xzf', "/tmp/#{file_name}", '-C', dir)

        # Delete tgz file
        FileUtils.rm_r "/tmp/#{file_name}"
        FileUtils.mkdir_p "#{dir}/images/rotated"

        dir
      end

      def download_s3(file_name)
        s3_config = YAML.load_file("#{Rails.root}/config/aws.yml")
        s3 = AWS::S3.new(access_key_id: s3_config[Rails.env]['access_key_id'],
                         secret_access_key: s3_config[Rails.env]['secret_access_key'],
                         s3_endpoint: s3_config[Rails.env]['s3_host_name'])

        obj = s3.buckets[s3_config[Rails.env]['bucket']]
              .objects["imports/#{file_name}"]

        # Download s3 file
        File.open("/tmp/#{file_name}", 'wb') do |file|
          file.write(obj.read)
        end

        obj
      end

      def clean_dir(verbose, obj, dir, file_name)
        # Delete s3 file after import tree
        logit "Deleting s3 file #{file_name}" if verbose
        obj.delete if obj

        logit "Deleteting directory #{dir}" if verbose
        FileUtils.rm_r dir
      end
    end
  end
end
