#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    # This jobs generates content: titles, description, etc. depending of the function
    class UpdateContentIncidentsJob < Struct.new(:verbose, :user_id)
      include Redborder::Jobs::JobHelper
      include IncidentsHelper

      def perform
        content = ''
        syslog_configured = false
        incidents = []
        incidents_to_process = 0
        begin
          content += "Searching for non visible incidents...\n" if verbose
          incidents = Incident.where(visible: false)
          incidents_to_process = incidents.count

          if incidents_to_process > 50
            limit = (incidents_to_process * 0.10).ceil
            content += "#{incidents_to_process} incidents found (processing 10% = #{limit}).\n" if verbose
            incidents = Incident.where(visible: false).limit(limit)
            incidents_to_process = incidents.count
          else
            content += "#{incidents_to_process} incidents found (processing all).\n" if verbose
          end

          user = User.find_by(id: user_id)

          content += user_id.to_s

          if incidents_to_process > 0
            syslog_configured = SyslogService.syslog_integration_configured?

            incidents.each do |incident|
              content += "\nGetting data for incident #{incident.id}...\n" if verbose
              alerts = incident.get_data({}, user)

              if alerts.nil? || alerts.empty?
                incident.increment!(:failure_count)
                content += "\nNo data found for incident #{incident.id}. Failure count: #{incident.failure_count}. This incident will be skipped...\n" if verbose

                if incident.failure_count >= 24
                  content += "Incident #{incident.id} has failed 24 times and will be deleted\n" if verbose
                  incident.destroy
                end
                next
              end

              # If redborder-llm is enabled
              if Redborder::LLM.is_llm_service_enabled
                content += "\n\nredborder-llm service is enabled. Generating title and description for incident #{incident.id}\n" if verbose
                params = { alerts: alerts, incident_name: incident.name }

                title = Redborder::LLM.generate_incident_title(params)

                if title.is_a?(String)
                  incident.name = title[0, 255].gsub("\"",'').strip
                  content += "Title for incident #{incident.id}: #{title}\n\n" if verbose
                  create_llm_generation_record('name', incident.id)
                else
                  content +="Error generating title for incident #{incident.id}\n" if verbose
                  content +="Title error: #{title}\n\n" if verbose
                end

                description = Redborder::LLM.generate_incident_description(params)
                if description.is_a?(String)
                  incident.description = description[0, 1000]
                  content += "Description for incident #{incident.id}: #{description}\n\n" if verbose
                  create_llm_generation_record('description', incident.id)
                else
                  content += "Error generating description for incident #{incident.id}\n" if verbose
                  content += "Description error: #{description}\n" if verbose
                end
              else
                content += "redborder-llm service is not enabled. LLM title and description will not be generated.\n" if verbose
              end

              content += "Saving incident with ID: #{incident.id}\n" if verbose
              save_success = false
              if incident.save
                save_success = incident.update(failure_count: 0, visible: true)
              end

              unless save_success
                incident.increment!(:failure_count)
                content += "Error Saving incident #{incident.id}. Failure count: #{incident.failure_count}\n" if verbose
                content += "Validation errors: #{incident.errors.full_messages.join(', ')}\n" if verbose

                # TODO: From a security perspective we shold never delete incident
                #       we need to find a way to avoid this.
                if incident.failure_count >= 12
                  content += "Incident #{incident.id} has failed 12 times and will be deleted\n" if verbose
                  incident.destroy
                end
              end
            end
          end
        rescue Exception => e
          content += "#{e.class}: #{e.message}\n"
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          logit "#{e.class}: #{e.message}"
        ensure
          if syslog_configured && !DelayedJob.incident_to_syslog?
            content += "\nStarting IncidentToSyslogJob..\n"
            # Re-enqueue the job to run again after 10 minutes
            Delayed::Job.enqueue(Redborder::Jobs::IncidentToSyslogJob.new(user_id, verbose), run_at: Time.now + 2.minutes)
          end

          # No need to make the storejob of jobs that did nothing to avoid filling up the stored jobs list
          if incidents_to_process > 0
            StoredDelayedJob.create(job: "UpdateContentIncidentsJob", user_id: user_id, content: content)
          end
          
          if DelayedJob.where('handler like ?',
            '%!ruby/struct:Redborder::Jobs::UpdateContentIncidentsJob%').count <= 1

            # Enqueue the job again
            Delayed::Job.enqueue(
            Redborder::Jobs::UpdateContentIncidentsJob.new(verbose, user_id), priority: 10, run_at: Time.now + 5.minutes)
          end
        end
      end

      private

      # Method to create a new entry in llm_generation_records table
      def create_llm_generation_record(field_name, incident_id)
        LlmGenerationRecords.create(
          table_name: 'incidents',
          field_name: field_name,
          record_id: incident_id,
          record_type: 'Incident'
        )
      rescue ActiveRecord::RecordInvalid => e
        Rails.logger.error "Failed to create LLM generation record: #{e.message}"
      end
    end
  end
end