#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class UpdateIncidentsJob < Struct.new(:verbose)
      include Redborder::Jobs::JobHelper

      def perform
        begin
          Incident.transaction do 
            settings = Setting.where(name: ["incidents_autoclose_enabled", "incidents_deletion_enabled"]).pluck(:name, :value).to_h
            autoclose if settings["incidents_autoclose_enabled"] == "1"
            autodelete if settings["incidents_deletion_enabled"] == "1"
          end
        rescue StandardError => e
          logit("#{e.class}: #{e.message}") if verbose
        ensure
          Delayed::Job.enqueue(self.class.new(true), priority: 10, run_at: 5.minutes.from_now)
        end
      end

      private

      def autoclose
        duration, unit = fetch_duration_and_unit("autoclose")
        cutoff_date = calculate_cutoff_date(duration, unit)
        incidents_to_close = Incident.where('created_at < ?', cutoff_date)

        if incidents_to_close.exists?
          incidents_to_close.update_all(incident_status_id: 1) 
        end
      end

      def autodelete
        duration, unit, status_id = fetch_duration_unit_and_status
        cutoff_date = calculate_cutoff_date(duration, unit)
        incidents_to_delete = Incident.where('created_at < ? AND incident_status_id = ?', cutoff_date, status_id)

        if incidents_to_delete.exists?
          incidents_to_delete.destroy_all
        end

      end

      def fetch_duration_and_unit(setting_type)
        duration = Setting.find_by(name: "#{setting_type}_incidents_duration").value.to_i
        unit = Setting.find_by(name: "#{setting_type}_incidents_unit").value
        [duration, unit]
      end

      def fetch_duration_unit_and_status
        duration, unit = fetch_duration_and_unit("delete")
        status_name = Setting.find_by(name: "delete_incidents_status").value
        status_id = IncidentStatus.find_by(name: status_name).id
        [duration, unit, status_id]
      end

      def calculate_cutoff_date(duration, unit)
        case unit
        when 'days'
          Date.today - duration.days
        when 'weeks'
          Date.today - duration.weeks
        when 'months'
          Date.today - duration.months
        when 'years'
          Date.today - duration.years
        else
          raise "Invalid time unit: #{unit}"
        end
      end
    end
  end
end