#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Miguel Álvarez <malvarez@redborder.com>
#######################################################################

module Redborder
  module Jobs
    class UpdateThreatIntelJob < Struct.new(:verbose, :user_id)

      include Redborder::Jobs::JobHelper

      MEMCACHED_CONFIG = (YAML.load(File.read('config/memcached_config.yml'))[Rails.env].symbolize_keys rescue {}).freeze
      MEMCACHED_NAMESPACE = 'rbti'
      TMP_FILE = '/tmp/ti_tmp'

      def perform
        content = ""
        entities = []
        begin
          entities = ThreatIntelSource.all
      
          return if entities.empty?
      
          entities.each do |entity|
            if entity.enable
              data_bag = Redborder::Databag.load('passwords', 'threat_intel_sources')
              if data_bag[entity.id.to_s]
                password = data_bag[entity.id.to_s]
              else
                password = ""
              end
      
              content += "Starting download for #{entity.path}...\n"
      
              content += "Initializing client for entity #{entity.id} (#{entity.path})...\n"
      
              client = Redborder::ThreatIntel::Client.new(
                protocol: ThreatConnType.find_by(id: entity.threat_conn_types_id).name,
                host: entity.host,
                port: entity.port,
                user: entity.anonymous ? "anonymous" : entity.user,
                password: entity.anonymous ? "" : password,
                dalli_options: {
                  servers: MEMCACHED_CONFIG[:servers],
                  options: {
                    namespace: MEMCACHED_NAMESPACE,
                    compress: true
                  }
                }
              )
      
              content += "Client initialized for entity #{entity.id} (#{entity.path})...\n"
              content += "Assets type for download is #{entity.threat_src_type}...\n"
              begin
                content += "Starting download and parsing for #{entity.path}...\n"
                counts = client.download_and_inject(remote_path: "#{entity.path}", local_path: TMP_FILE, type: entity.threat_src_type)
      
                content += "Downloaded #{entity.path} and saved in system cache...\n"
                content += "Valid keys: #{counts[:valid]}, Invalid keys: #{counts[:invalid]}, Skipped keys: #{counts[:skipped]}\n"
              rescue StandardError => download_error
                content += "Error during download/inject for #{entity.path}: #{download_error.class}: #{download_error.message}\n"
                download_error.backtrace.each { |line| content += "#{line}\n" }
              end
            else
              content += "Entity #{entity.id} is disabled, skipping download.\n"
            end
          end
      
        rescue StandardError => e
          content += "#{e.class}: #{e.message}\n"
          e.backtrace.each { |line| content += "#{line}\n" }
        ensure
          if DelayedJob.where('handler like ?',
            '%!ruby/struct:Redborder::Jobs::UpdateThreatIntelJob%').count <= 1
            periodicity_minutes = Setting.threat_intel_sources_download_periodicity.to_i || 10
            Delayed::Job.enqueue(self.class.new(user_id, verbose), run_at: Time.now + periodicity_minutes.minutes, priority: 10)
          end

          StoredDelayedJob.create(job: "UpdateThreatIntelJob", user_id: user_id, content: content) unless entities.empty?
        end
      end

      def reschedule_at(current_time, attempts)
        periodicity_minutes = Setting.threat_intel_sources_download_periodicity.to_i || 10
        current_time.advance(minutes: periodicity_minutes)
      end
    end
  end
end