#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  # This module generates content by an LLM model
  # Can be used to generate titles, description, etc.
  # All the methods defined here are prepared to be executed by llm_generate_content.rb delayed job
  module LLM
    # ==========================================================
    # Constants
    # ==========================================================

    LLM_CONFIG = YAML.load_file("#{Rails.root}/config/llm_models.yml")
    API_URL = LLM_CONFIG["api_url"]

    # ==========================================================
    # Methods
    # ==========================================================

    # Check if llm service is enabled or not
    def self.is_llm_service_enabled
      Manager.all.sort_by(&:name).any? do |manager|
        manager.services['redborder-llm'] && manager.services['redborder-llm'][:status] && manager.services['redborder-llm'][:ok]
      end
    end

    def self.generate(prompt, regex = nil, timeout = 600)
      response = nil
      begin
        response = HTTParty.post(
          API_URL,
          headers: { 'Content-Type' => 'application/json', 'Authorization' => 'Bearer no-key' },
          body: prompt.to_json,
          timeout: timeout
        )
      rescue Net::ReadTimeout => e
        Rails.logger.error "Net::ReadTimeout while generating LLM content.\nError: #{e.message}"
        return { error: 'HTTP request failed due to timeout', message: e.message }
      rescue StandardError => e
        Rails.logger.error "HTTP request failed.\nError: #{e.message}"
        return { error: 'HTTP request failed', message: e.message }
      end

      if response && response.success?
        # Extract the content of the response
        raw_content = response.dig('choices', 0, 'message', 'content')

        if raw_content
          # Clean up the response by removing unnecessary tags
          cleaned_response = raw_content.gsub(/<\/s>/, '').strip

          # Apply regex if provided as parameter
          if regex
            return cleaned_response.match(regex).to_s.strip
          else
            return cleaned_response
          end
        else
          Rails.logger.error 'Unexpected response format, missing expected keys'
          return { error: 'Unexpected response format', message: 'Response missing expected content' }
        end
      elsif response && response.code
        Rails.logger.error "HTTP request failed with response code: #{response.code}"
        return { error: 'HTTP request failed', message: "Response code: #{response.code}" }
      else
        Rails.logger.error 'HTTP request failed without any response code'
        return { error: 'HTTP request failed', message: 'Unknown error' }
      end
    end

    # Creates an incident title
    def self.generate_incident_title(params)
      Rails.logger.info "Generating title of an incident...\n"
      alerts = params[:alerts]
      regex = /(?<=Title:\s).+/

      policies = []
      alerts.each do |alert|
        policies.push(alert[:msg])
      end

      policies = if policies.count > 0
                   policies.uniq.join(', ')
                 else
                   params[:incident_name].to_s
                 end

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 255,
        messages: [
          {
            role: 'system',
            content: LLM_CONFIG['generate_title_system_prompt']
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_title_user_prompt']}#{policies}."
          }
        ]
      }

      generate(prompt, regex)
    end

    # Creates the prompt of incident description
    def self.generate_incident_description(params)
      Rails.logger.info "Generating description of an incident...\n"
      alerts = params[:alerts]
      regex = /(?<=Description:\s).+/

      policies = []
      alerts.each do |alert|
        policies.push(alert[:msg])
      end

      policies = if policies.count > 0
                   policies.uniq.join(', ')
                 else
                   params[:incident_name].to_s
                 end

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 4000,
        messages: [
          {
            role: 'system',
            content: "#{LLM_CONFIG['generate_description_system_prompt']}"
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_description_user_prompt']}#{policies}."
          }
        ]
      }

      generate(prompt, regex)
    end

    # Creates the prompt of incident report executive summary
    def self.generate_incident_executive_summary(params)
      Rails.logger.info "Generating incident report executive summary...\n"

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 4000,
        messages: [
          {
            role: 'system',
            content: "#{LLM_CONFIG['generate_executive_summary_system_prompt']}"
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_executive_summary_user_prompt']}#{params[:content]}."
          }
        ]
      }

      generate(prompt, nil, 900)
    end

    # Creates the prompt of incident report event summary
    def self.generate_incident_event_summary(params)
      Rails.logger.info "Generating incident report event summary...\n"

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 4000,
        messages: [
          {
            role: 'system',
            content: "#{LLM_CONFIG['generate_event_summary_system_prompt']}"
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_event_summary_user_prompt']}#{params[:content]}."
          }
        ]
      }

      generate(prompt, nil, 900)
    end
    # Creates the prompt of incident report impact
    def self.generate_incident_impact(params)
      Rails.logger.info "Generating incident report impact...\n"

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 4000,
        messages: [
          {
            role: 'system',
            content: "#{LLM_CONFIG['generate_impact_system_prompt']}"
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_impact_user_prompt']}#{params[:content]}."
          }
        ]
      }

      generate(prompt, nil, 900)
    end
    # Creates the prompt of incident report mitigation
    def self.generate_incident_mitigation(params)
      Rails.logger.info "Generating incident report mitigation...\n"

      prompt = {
        model: 'LLaMA_CPP',
        max_tokens: 4000,
        messages: [
          {
            role: 'system',
            content: "#{LLM_CONFIG['generate_mitigation_system_prompt']}"
          },
          {
            role: 'user',
            content: "#{LLM_CONFIG['generate_mitigation_user_prompt']}#{params[:content]}."
          }
        ]
      }

      generate(prompt, nil, 900)
    end
  end
end
