#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  #
  # Manage menus on the platform
  #
  module MenuManager
    class MenuError < StandardError #:nodoc:
    end

    module MenuHelper
      #
      # Render the product_menu. This is an alias method.
      #
      def render_product_menu
        render_menu(:product_menu, nil)
      end

      def render_product_menu_with_intro(step)
        render_menu(:product_menu,
                    nil,
                    class: 'nav main-menu',
                    data: {
                      step: step,
                      intro: "#{I18n.t 'lib_redborder.installed'}"\
                             "#{I18n.t 'lib_redborder.graphs'}"
                    }
                   )
      end

      #
      # Render a menu
      #
      # == Parameters:
      # menu::
      #   Symbol with the menu to render
      # sensor::
      #   Current sensor if it's needed
      # manage::
      #   Boolean indicates if user has manage privileges
      # modules::
      #   Array of string with the modules availables for an user
      #
      # == Returns:
      # Safe string with the code of the menu
      #
      def render_menu(menu, sensor = nil, options = {})
        links = []
        menu_items_for(menu, sensor) do |node|
          links << render_menu_node(node, sensor)
        end
        links.empty? ? nil : links.join("\n").html_safe
      end

      #
      # Render node for a menu (li element). This method can render a node with
      # children (ul inside) or a single node (only a link)
      #
      # == Parameters:
      # node::
      #   A node object with the info of element
      # sensor::
      #   Current sensor if it's needed
      # bold::
      #   Show text as bold
      #
      # == Returns:
      # Safe string with the code of element
      #
      def render_menu_node(node, sensor = nil, bold = nil, opts = {})
        return unless allowed_node?(node, User.current_user, sensor, opts)
        if node.children.present? || !node.child_menus.nil?
          return render_menu_node_with_children(node, sensor)
        else
          caption, url = extract_node_details(node, sensor, opts)
          return content_tag('li',
                               render_single_menu_node(node, caption, url, bold))
        end
      end

      #
      # Render node with a children for a menu (li element)
      #
      # == Parameters:
      # node::
      #   A node object with the info of element
      # sensor::
      #   Current sensor if it's needed
      #
      # == Returns:
      # Safe string with the code of element
      #
      def render_menu_node_with_children(node, sensor=nil)
        caption, url = extract_node_details(node, sensor)

        html = [].tap do |html|
          html << "<li class='dropdown-submenu align-left'>"
          # Parent
          html << render_single_menu_node(node, caption, url)

          # Standard children
          standard_children_list = "".html_safe.tap do |child_html|
            node.children.each do |child|
              child_html << render_menu_node(child, sensor)
            end
          end

          html << content_tag(:ul, standard_children_list, :class => 'dropdown-menu') unless standard_children_list.empty?

          # Unattached children
          unattached_children_list = render_unattached_children_menu(node, sensor)
          html << content_tag(:ul, unattached_children_list, :class => 'menu-children unattached') unless unattached_children_list.blank?

          html << '</li>'
        end
        return html.join("\n").html_safe
      end

      # Returns a list of unattached children menu items
      def render_unattached_children_menu(node, sensor)
        return nil unless node.child_menus

        "".html_safe.tap do |child_html|
          unattached_children = node.child_menus.call(sensor)
          # Tree nodes support #each so we need to do object detection
          if unattached_children.is_a? Array
            unattached_children.each do |child|
              child_html << content_tag(:li, render_unattached_menu_item(child, sensor))
            end
          else
            raise MenuError, ":child_menus #{I18n.t 'lib_redborder.must be'}"
          end
        end
      end

      #
      # Render a single element (anchor) from node
      #
      # == Parameters:
      # item::
      #   A node object with the info of element
      # caption::
      #   String with the text to show on anchor
      # url::
      #   String with URL of the anchor
      # bold::
      #   Show text as bold
      #
      # == Returns:
      # Safe string with the code of element
      #
      def render_single_menu_node(item, caption, url, bold = nil)
        render_caption(item, caption, url, bold)
      end

      def render_unattached_menu_item(menu_item, sensor)
        raise MenuError, ":child_menus #{I18n.t 'lib_redborder.must_be'}" unless menu_item.is_a? MenuItem

        if User.current_user.allowed_to?(menu_item.url, sensor)
          link_to(h(menu_item.caption),
                  menu_item.url,
                  menu_item.html_options)
        end
      end

      #
      # Return items for a menu
      #
      # == Parameters:
      # menu::
      #   A symbol specifies the menu
      # sensor::
      #   Current sensor if it's needed
      #
      # == Returns:
      # Array of items or nothing if a block given
      #
      def menu_items_for(menu, sensor = nil)
        items = []
        Redborder::MenuManager.items(menu).root.children.each do |node|
          if allowed_node?(node, User.current_user, sensor, {})
            if block_given?
              yield node
            else
              items << node  # TODO: not used?
            end
          end
        end
        return block_given? ? nil : items
      end

      def menu_lis_for(menu)
        lis = []
        Redborder::MenuManager.items(menu).root.children.each do |node|
          content_tag(:li, render_caption(node, node.caption, node.url, nil))
        end
        lis.join("\n").html_safe
      end

      # -------------------------------------------------------------
      # Captions
      # -------------------------------------------------------------

      #
      # Render a caption. The caption can have an icon, image or
      # font-awesome icon.
      #
      # == Parameters:
      # item::
      #   MenuItem object
      # caption::
      #   String with the caption to show
      # url::
      #   Url for the link
      # bold::
      #   Symbol of the item who must be bold
      #
      # == Returns:
      # String with the HTML code
      #
      def render_caption(item, caption, url, bold)
        if item.icon
          item.icon.include?('.') ? caption_with_image(item, caption, url) :
                                    caption_with_fa(item, caption, url)
        else
          html_options = item.html_options.dup
          html_options[:class] += ' bold' if item.url.try(:keys) &&
                                             item.url.keys.include?(:action) &&
                                             item.url[:action].to_sym.eql?(bold)
          if caption.start_with? "tr."
            link_to(h("#{I18n.t caption[3..-1]}"), url, html_options)
          else
            link_to(h(caption), url, html_options)
          end
        end
      rescue
        nil
      end

      #
      # Create a caption with an image
      #
      # == Parameters:
      # item::
      #   MenuItem object
      # caption::
      #   String with the caption to show
      # url::
      #   Url for the link
      #
      # == Returns:
      # String with the HTML code
      #
      def caption_with_image(item, caption, url)
        img = ActionController::Base.helpers.asset_path(item.icon)
        link_to(content_tag(:img, nil, class: 'custom-icon', src: img) +
                            " #{h(caption)}",
                url,
                item.html_options)
      end

      #
      # Create a caption with a font-awesome icon
      #
      # == Parameters:
      # item::
      #   MenuItem object
      # url::
      #   Url for the link
      #
      # == Returns:
      # String with the HTML code
      #
      def caption_with_fa(item, caption, url)
        link_to(content_tag(:i, nil, class: "#{item.icon}") +
                            " #{h(caption)}",
                url,
                item.html_options)
      end

      def extract_node_details(node, sensor = nil, opts = {})
        item = node
        url = case item.url
        when Hash
          sensor.nil? ? item.url : {item.param => sensor}.merge(item.url)
        when Symbol
          send(item.url)
        else
          item.url
        end
        url = url.merge(opts) if opts.present?
        caption = item.caption(sensor)
        caption = (caption.to_s.start_with? "tr.") ? (I18n.t caption.to_s[3..-1]) : (caption)
        return [caption, url]
      end

      #
      # Checks if a user is allowed to access the menu item by:
      #
      # * Checking the url target (sensor only)
      # * Checking the conditions of the item
      #
      # == Parameters:
      # node::
      #   The item to render
      # user::
      #   Current user
      # sensor::
      #   Sensor object
      #
      # == Return:
      # True if the user can see the node.
      #
      def allowed_node?(node, user = User.current_user, sensor, opts)
        return false if node.admin && !user.admin?
        return false if node.super_admin && !user.super_admin?
        return false if node.super_reader && !user.super_reader?
        # Condition that doesn't pass
        return false if node.condition && !node.condition.call(sensor)
        # Check modules
        # user.present? is needed for tests.
        return false if node.module.present? && user.present? &&
                        !user.modules.include?(node.module)
        # Don't add if it's a view from the common menu, for a specific module
        # and the current module view is not its owner.
        return false if node.common && node.module &&
                        !node.module.to_sym.eql?(controller_name.to_sym)
        # Don't add if it's a view from the common menu and it is the current one.
        return false if node.common && opts[:controller] &&
                        node.url[:action].to_sym.eql?(action_name.to_sym) &&
                        opts[:controller].to_sym.eql?(controller_name.to_sym)
        true
      end
    end

    class << self
      def map(menu_name)
        @items ||= {}
        mapper = Mapper.new(menu_name.to_sym, @items)
        if block_given?
          yield mapper
        else
          mapper
        end
      end

      def items(menu_name)
        @items[menu_name.to_sym] || MenuNode.new(:root, {})
      end
    end

    class Mapper
      attr_reader :menu, :menu_items

      def initialize(menu, items)
        items[menu] ||= MenuNode.new(:root, {})
        @menu = menu
        @menu_items = items[menu]
      end

      # Adds an item at the end of the menu. Available options:
      # * param: the parameter name that is used for the object id (default is :id)
      # * if: a Proc that is called before rendering the item, the item is displayed only if it returns true
      # * caption that can be:
      #   * a localized string Symbol
      #   * a String
      #   * a Proc that can take the object as argument
      # * before, after: specify where the menu item should be inserted (eg. :after => :activity)
      # * parent: menu item will be added as a child of another named menu (eg. :parent => :issues)
      # * children: a Proc that is called before rendering the item. The Proc should return an array of MenuItems, which will be added as children to this item.
      #   eg. :children => Proc.new {|object| [Redborder::MenuManager::MenuItem.new(...)] }
      # * last: menu item will stay at the end (eg. :last => true)
      # * html_options: a hash of html options that are passed to link_to
      def push(name, url, options={})
        options = options.dup

        if options[:parent]
          subtree = self.find(options[:parent])
          if subtree
            target_root = subtree
          else
            target_root = @menu_items.root
          end

        else
          target_root = @menu_items.root
        end

        # menu item position
        if first = options.delete(:first)
          target_root.prepend(MenuItem.new(name, url, options))
        elsif before = options.delete(:before)

          if exists?(before)
            target_root.add_at(MenuItem.new(name, url, options), position_of(before))
          else
            target_root.add(MenuItem.new(name, url, options))
          end

        elsif after = options.delete(:after)

          if exists?(after)
            target_root.add_at(MenuItem.new(name, url, options), position_of(after) + 1)
          else
            target_root.add(MenuItem.new(name, url, options))
          end

        elsif options[:last] # don't delete, needs to be stored
          target_root.add_last(MenuItem.new(name, url, options))
        else
          target_root.add(MenuItem.new(name, url, options))
        end
      end

      # Removes a menu item
      def delete(name)
        if found = self.find(name)
          @menu_items.remove!(found)
        end
      end

      # Checks if a menu item exists
      def exists?(name)
        @menu_items.any? {|node| node.name == name}
      end

      def find(name)
        @menu_items.find {|node| node.name == name}
      end

      def position_of(name)
        @menu_items.each do |node|
          if node.name == name
            return node.position
          end
        end
      end
    end

    class MenuNode
      include Enumerable
      attr_accessor :parent
      attr_reader :last_items_count, :name

      def initialize(name, content = nil)
        @name = name
        @children = []
        @last_items_count = 0
      end

      def children
        if block_given?
          @children.each {|child| yield child}
        else
          @children
        end
      end

      # Returns the number of descendants + 1
      def size
        @children.inject(1) {|sum, node| sum + node.size}
      end

      def each &block
        yield self
        children { |child| child.each(&block) }
      end

      # Adds a child at first position
      def prepend(child)
        add_at(child, 0)
      end

      # Adds a child at given position
      def add_at(child, position)
        #raise "Child already added" if find {|node| node.name == child.name}
        return nil if find {|node| node.name == child.name}
        @children = @children.insert(position, child)
        child.parent = self
        child
      end

      # Adds a child as last child
      def add_last(child)
        add_at(child, -1)
        @last_items_count += 1
        child
      end

      # Adds a child
      def add(child)
        position = @children.size - @last_items_count
        add_at(child, position)
      end
      alias :<< :add

      # Removes a child
      def remove!(child)
        @children.delete(child)
        @last_items_count -= +1 if child && child.last
        child.parent = nil
        child
      end

      # Returns the position for this node in it's parent
      def position
        self.parent.children.index(self)
      end

      # Returns the root for this node
      def root
        root = self
        root = root.parent while root.parent
        root
      end
    end

    #
    # Define a menu item (aka node). This class store the name, caption,
    # parent... and the conditions to show the Item.
    #
    class MenuItem < MenuNode
      attr_reader :name, :url, :param, :condition, :parent,
                  :child_menus, :last, :icon, :module, :admin,
                  :super_admin, :super_reader, :common

      #
      # Initialize the variables for an item
      #
      def initialize(name, url, options)
        # Perform some validations
        # ----------------------------
        validate(name, options)

        # Initialize all variables
        @name = name
        @url = url
        @condition = options[:if]
        @param = options[:param] || :id
        @caption = options[:caption]
        @html_options = options[:html] || {}
        # Adds a unique class to each menu item based on its name
        @html_options[:class] = [@html_options[:class], @name.to_s.dasherize]
                                .compact.join(' ')
        @parent = options[:parent]
        @child_menus = options[:children]
        @admin = options[:admin] || false
        @super_admin = options[:super_admin] || false
        @super_reader = options[:super_reader] || false
        @common = options[:common] || nil
        @last = options[:last] || false
        @icon = options[:icon]
        @module = options[:module]
        super @name.to_sym
      end

      #
      # Caption for the link.
      #
      # == Parameters:
      # sensor::
      #   (Optional) Sensor object to render.
      #
      # == Returns:
      # An String with the caption
      #
      def caption(sensor = nil)
        if @caption.is_a?(Proc)
          c = @caption.call(sensor).to_s
          c = @name.to_s.humanize if c.blank?
          c
        else
          if @caption.nil?
            l_or_humanize(name, prefix: 'label_')
          else
            @caption.is_a?(Symbol) ? l(@caption) : @caption
          end
        end
      end

      def html_options(options={})
        @html_options
      end

      def url=(url_to_save)
        @url = url_to_save
      end

      #
      # Check variables and raise an exception if there are a problem
      #
      # == Parameters:
      # name::
      #   Name of the node
      # options::
      #   Hash with options for the MenuItem
      #
      def validate(name, options)
        fail ArgumentError,
             "#{I18n.t 'lib_redborder.invalid_if'} '#{name}'" if
             options[:if] && !options[:if].respond_to?(:call)
        fail ArgumentError,
             "#{I18n.t 'lib_redborder.invalid_html'} '#{name}'" if
             options[:html] && !options[:html].is_a?(Hash)
        fail ArgumentError,
             "#{I18n.t 'lib_redborder.cannot_parent'}" if
             options[:parent] == name.to_sym
        fail ArgumentError,
             "#{I18n.t 'lib_redborder.invalid_children'} '#{name}'" if
             options[:children] && !options[:children].respond_to?(:call)
      end
    end
  end
end
