#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  # Register modules from lib/modules folder. You can configure several
  # parameters of the module like:
  #
  # - Alarm creation
  # - Custom widgets
  # - Menus
  # - ...
  #
  class Module
    mattr_accessor :registered_modules
    self.registered_modules = {}

    CONTROLLER_MAPPING = {
      traffics: 'flow',
      events: 'ips'
    }.freeze

    def self.register(id, &block)
      m = new(id)
      m.instance_eval(&block)

      # Prepends the app/views directory of the plugin to the view path
      view_path = File.join(Rails.root, 'lib', 'modules', id.to_s, 'app', 'views')
      if File.directory?(view_path)
        ActionController::Base.prepend_view_path(view_path)
        ActionMailer::Base.prepend_view_path(view_path)
      end

      # Copy the assets files to app/assets
      copy_module_folder id, :javascripts
      copy_module_folder id, :stylesheets
      copy_module_folder id, :images
    end

    def self.available
      modules = Dir.entries("#{Rails.root}/lib/modules/")
                .reject { |entry| ['.', '..'].include? entry }

      if File.exist? "#{Rails.root}/config/modules.yml"
        available_modules = YAML.load(File.read('config/modules.yml'))['modules']
        modules &= available_modules unless available_modules.include?('all')
      end

      modules
    end

    def self.available?(mod)
      available.include?(mod)
    end

    def self.get(id, name)
      registered_modules.fetch(id.to_sym, {})[name.to_sym]
    end

    def self.get_from_all(name)
      registered_modules.each_with_object({}) do |(mod, data), result|
        result[mod] = data[name]
      end
    end

    def self.each
      registered_modules.each do |k, v|
        yield k, v
      end
    end

    # Return the registered modules with UI like flow or ips.
    # For example, API hasn't interface. By default, interface_presence
    # are true.
    def self.each_with_interface
      registered_modules.each do |k, v|
        yield k, v if v.fetch(:interface_presence, true)
      end
    end

    def self.keys_with_interface
      registered_modules.each_with_object([]) do |(k, v), memo|
        memo << k if v.fetch(:interface_presence, true)
      end
    end

    def self.keys_with_selectable_widget
      registered_modules.each_with_object([]) do |(k, v), memo|
        memo << k if v.fetch(:selectable_widget, false)
      end
    end

    #
    # Return the registered modules with alarms enabled
    #
    def self.each_with_alarm
      registered_modules.each do |k, v|
        yield k, v if v.fetch(:interface_presence, true) &&
                      v.fetch(:alarm, true)
      end
    end

    def self.copy_module_folder(id, folder)
      src = "#{Rails.root}/lib/modules/#{id}/app/assets/#{folder}"
      dst = "#{Rails.root}/lib/assets"

      begin
        FileUtils.cp_r(src, dst, remove_destination: true) if File.exist? src
      rescue StandardError => e
        Rails.logger.error e
      end
    end

    #
    # Returns the module associated to a controller or nil
    #
    def self.from_controller(controller)
      return nil unless controller
      return controller if registered_modules.keys.include? controller.to_sym
      CONTROLLER_MAPPING[controller.to_sym]
    end

    def self.columns_with_modules(hide_global: false)
      columns = {}

      Redborder::Module.registered_modules.each do |mod, info|
        next if info[:columns].nil? ||
                (info[:hide_global] == true && hide_global)

        info[:columns].each do |column, prop|
          next if prop[:hide_raw] == true || prop[:columns][column].nil? ||
                  (prop[:hide_global] == true && hide_global)

          unless columns[column.to_sym].present?
            columns[column.to_sym] = { name: prop[:columns][column],
                                       modules: [mod] }
          else
            columns[column.to_sym][:modules] << mod
          end
        end
      end
      columns
    end

    def initialize(id)
      registered_modules[id.to_sym] = {}
      @id = id.to_sym
    end

    def method_missing(name, *args)
      registered_modules[@id][name] = args[0]
    end

    def columns(values)
      registered_modules[@id][:columns] ||= {}
      registered_modules[@id][:columns].merge! values
    end

    def aggregations(values)
      registered_modules[@id][:aggregations] ||= {}
      registered_modules[@id][:aggregations].merge! values
      # REFACTOR: Using rbdruid here is inappropriate
      Redborder::Rbdruid::AGGREGATIONS.merge! values
    end

    # TODO: Move it to flow module
    def ap_aggregations(values)
      registered_modules[@id][:ap_aggregations] ||= {}
      registered_modules[@id][:ap_aggregations].merge! values
      # REFACTOR: Using rbdruid here is inappropriate
      Redborder::Rbdruid::AGGREGATIONS.merge!(values)
    end

    def tabs(values)
      registered_modules[@id][:tabs] ||= {}
      registered_modules[@id][:tabs].merge! values
    end

    #
    # Allow a module to prevent alarm creation for it
    #
    # == Parameters:
    # value::
    #   True or false to enable or disable
    #
    def alarm(value)
      registered_modules[@id][:alarm] = value
    end

    # Specifies the widgets that will appear on the widgets form
    # in the section with the name of the module.
    # This widgets will be added automatically when the user clicks on them
    # without any more interaction from the user.
    #
    # == Parameters:
    # values::
    #   Array of hashes that specifies the widgets that will appear on
    #   the section. This hash can have the following properties:
    #     - name (required): the name that will appear on the widget
    #       selector box
    #     - icon (required): the icon that will apear on the widget
    #       selector box. This accepts both fa icons names or image URIs
    #     - auto (default true): set it to false if you want to display
    #       the widget form on click instead of creating the widget with
    #       the default settings. Useful when a widget can't be rendered
    #       with default settings. If it's true, then the id property is
    #       required, as it will be the parameter passed to the widget form
    #       to render the required form.
    #     - fields: hash with key value pairs specifying every property of the
    #       widget that will be saved when the user clicks on the selector box
    #     - id: symbol that specifies both graph_type and widget_type fields.
    #       Useful to create default concrete widgets when both graph_type
    #       and widget_type have the same name.
    def widgets(values = [], custom: true)
      name = registered_modules[@id][:name]

      Widget.sections[@id] = {
        name: name,
        elements: values,
        custom: custom
      }
    end

    # Widgets defined for overview dashboard
    #
    def overview_widgets(values = [])
      registered_modules[@id][:overview_widgets] = values
    end

    # Adds one or more widgets to a certain section in the widgets menu
    # If the section with the given id doesn't exists, then it does nothing
    #
    # == Parameters:
    # id::
    #   The id of the section in which the widget will be added
    # widgets_data::
    #   A hash or an array of hashes that describes the way in which
    #   the widgets will be presented on the menu
    #
    # == Returns:
    # nil if nothing was done or the full array of widgets data if the
    # widgets were added to the section
    def widgets_section(id, widgets_data)
      return if Widget.sections[id].nil?
      Widget.sections[id][:elements] += Array(widgets_data)
    end

    def menu(menu, item, url, options = {})
      if available?(@id.to_s)
        Redborder::MenuManager.map(menu).push(item, url, options.merge(module: @id.to_s))
      end
    end

    alias add_menu_item menu
  end
end
