#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder

  class AIoutliers

    def self.outlier_r(query, conditions)
      uri = URI.parse('http://rb-aioutliers.service/api/v1/outliers')

      http = Net::HTTP.new(uri.host, uri.port)

      request = Net::HTTP::Post.new(uri.path)
      request.add_field('Content-Type', 'application/x-www-form-urlencoded')

      query = Base64.strict_encode64(query.to_json)
      request.body = URI.encode_www_form(query: query)

      response = http.request(request)
      result = JSON.parse(response.body)
    end
    
    def self.get_outliers_ips(outliers, ips, all_ips_data)
      uri = URI.parse('http://rb-aioutliers.service/api/v1/ip_identifier')
    
      http = Net::HTTP.new(uri.host, uri.port)
    
      request = Net::HTTP::Post.new(uri.path)
      request.add_field('Content-Type', 'application/x-www-form-urlencoded')
    
      payload = {
        ips: ips,
        all_ips_data: all_ips_data,
        outliers: outliers
      }
    
      json_payload = payload.to_json
      request.body = URI.encode_www_form(payload: json_payload)
    
      begin
        response = http.request(request)
        if response.is_a?(Net::HTTPSuccess)
          return JSON.parse(response.body.gsub(/\A"|"\Z/, '').gsub('\\"', '"'))
        else
          return { error: "Request failed with status: #{response.code}" }
        end
      rescue StandardError => e
         return { error: e.message }
      end
    end

  end

  class Sherlock

    def self.timeserie_range(predict_interval, granularity)
      # Converts a prediction interval into the specified granularity, such as minutes, hours, or weeks.
      # Returns the count of intervals based on the chosen granularity. If granularity is not recognized,
      # the original interval is returned.
      case granularity.to_sym
      when :minute
        return predict_interval / 60
      when :hour
        return predict_interval / 60 / 60
      when :week
        return predict_interval / 60 / 60 / 7
      else
        return predict_interval
      end
    end

    def self.granularity_range(granularity)
      # Maps various granularity identifiers to corresponding minute counts.
      # This function takes a granularity identifier as input and returns the
      # equivalent minute count for that granularity. If the input granularity
      # is not recognized, a default value of 1 is returned.
      case granularity.to_s
      when 'all'
        return "1"
      when /\Apt?(\d+)m\z/
        return $1
      when /\Apt?(\d+)h\z/
        hours = $1.to_i
        return (hours * 60).to_s
      when /\AP(\d+)D\z/
        days = $1.to_i
        return (days * 1440).to_s
      when 'fifteen_minute', 'pt15m'
        return "15"
      when 'thirty_minute', 'pt30m'
        return "30"
      when 'hour'
        return "60"
      else
        return "1"
      end
    end

    def self.sherlock_granularity(gran)
      # Converts various granularity identifiers used in the "Sherlock" context to a higher-level time unit.
      # Given a granularity identifier, this function returns a corresponding higher-level time unit,
      # such as "minute", "hour", or "day". If the input granularity is not recognized,
      # "minute" is returned by default.
      case gran.to_sym
      when :minute, :pt1m, :pt2m, :pt5m, :fifteen_minute, :pt15m, :thirty_minute, :pt30m, :pt60m, :p1m
        return "minute"
      when :hour, :pt2h, :pt8h, :pt24h, :pt120h, :pt168h, :pt720h
        return "hour"
      when :P1D
        return "day"
      else
        return "minute"
      end
    end

    def self.calculate_query_length(start_time, end_time)
      # Calculates the length of a time query in minutes between the given start and end times.
      # The function takes two arguments:
      # - start_time: The starting time of the query in a parsable time format.
      # - end_time: The ending time of the query in a parsable time format.
      # It calculates the time difference in minutes between the end time and the start time
      # and returns the resulting value. The calculation is based on converting the given times
      # to Unix timestamps and measuring the difference in minutes.
      (Time.parse(end_time).to_i / 60) - (Time.parse(start_time).to_i / 60)
    end

    def self.hook_druid_query(query, only_return_calculated_start_time)
      # Modifies a Druid query by adjusting the start time and granularity based on specified criteria.
      # Arguments:
      # - query: Original Druid query in JSON format.
      # - only_return_calculated_start_time: If true, only the calculated start time is returned.
      # Returns a modified Druid query with adjusted start time and granularity, if applicable.
      hooked_query = JSON.parse(query.to_json)
      start_time = hooked_query['intervals'][0].split('/').first
      end_time = hooked_query['intervals'][0].split('/').last
      granularity = granularity_to_minutes(hooked_query['granularity']['period'])
      train_length = train_length(
          granularity,
          calculate_query_length(start_time, end_time),
          Redborder::Rbdruid.get_druid_first_interval(
            hooked_query['dataSource']
          ),
          Time.parse(end_time).to_i
      )
      new_start_time = Time.parse(end_time).to_i - train_length * 60
      new_start_time = new_start_time - math_mod(new_start_time, granularity * 60)
      return new_start_time if only_return_calculated_start_time
      hooked_query['intervals'][0] = convert_to_iso8601(Time.at(new_start_time).to_s) + "/" + end_time
      hooked_query['granularity']['origin'] = convert_to_iso8601(Time.at(new_start_time).to_s)
      hooked_query
    end

    def self.granularity_to_minutes(granularity)
      # Converts a Druid query granularity into an equivalent duration in minutes.
      # Arguments:
      # - granularity: The granularity value to be converted, such as "minute", "hour", or "day".
      # Returns the duration in minutes corresponding to the provided granularity.
      minutes_per_unit = {
          "minute" => 1,
          "hour" => 60,
          "day" => 1440
      }
      unit = sherlock_granularity(granularity.to_sym)
      numeric_value = granularity[/\d+/]
      if numeric_value
          minutes = minutes_per_unit[unit] * numeric_value.to_i
      else
          minutes = minutes_per_unit[unit]
      end
      minutes
   end

    def self.convert_to_iso8601(input_str)
      # Converts a given time input string into ISO 8601 format.
      # Arguments:
      # - input_str: The input time string to be converted.
      # Returns the input time in ISO 8601 format (with timezone offset and colon-separated hours and minutes).
      input_time = Time.parse(input_str)
      iso8601_format = input_time.strftime("%Y-%m-%dT%H:%M:%S%z")
      iso8601_format.insert(-3, ':')
      return iso8601_format
    end

    def self.math_mod(num, mod)
      # Performs a modulo operation on a given number using a potentially optimized approach
      # for improved performance. Implemented by Malvads.
      # Arguments:
      # - num: The number to be divided.
      # - mod: The modulus value.
      # Returns the result of the modulo operation, utilizing a faster implementation
      # that takes advantage of bit manipulation and avoids division when possible.
      if mod != 0
        if (mod & (mod - 1) == 0)
          return num & (mod - 1)
        else
          return num % mod
        end
      end
    end

    def self.train_length(granularity, query_length, first_interval, query_end_time)
      # Calculates the optimal training length for a predictive model based on various settings.
      # Arguments:
      # - granularity: Duration in minutes for each time interval.
      # - query_length: Duration in minutes of the query being trained.
      # - first_interval: The timestamp of the first interval in minutes since Unix epoch.
      # Returns the recommended training length in minutes, considering configured settings
      # for outlier cycles and intervals while accounting for granularity and recorded time.
      recorded = (query_end_time / 60) - first_interval
      min_cycles = query_length * Setting.outliers_min_cycles.to_i
      max_cycles = query_length * Setting.outliers_max_cycles.to_i
      min_intervals = Setting.outliers_min_intervals.to_i * granularity + query_length
      max_intervals = Setting.outliers_max_intervals.to_i * granularity + query_length
      naive_min = [min_cycles, min_intervals].max
      naive_max = [max_cycles, max_intervals].min
      if naive_max < naive_min
          naive_min, naive_max = (naive_min + naive_max) / 2, (naive_min + naive_max) / 2
      end
      naive_max = naive_max - math_mod(naive_max, granularity)
      naive_min = naive_min - math_mod(naive_min, granularity)
      [recorded, naive_max].min
    end

    def self.filter_predicted_data(data, interval)
      # Filters predicted data entries within a specified time interval and updates the data accordingly.
      # Arguments:
      # - data: A hash containing predicted data entries.
      # - interval: The time interval (start/end) in ISO 8601 format to filter the predicted data.
      # Returns the updated data hash containing only entries within the specified interval.
      start_time, end_time = interval.split('/')
      start_time = Time.parse(start_time)
      end_time = Time.parse(end_time)

      filtered_data = data["predicted"].select do |entry|
        timestamp = Time.parse(entry["timestamp"])
        timestamp >= start_time && timestamp <= end_time
      end

      data["predicted"] = filtered_data
      return data
    end

    def self.get_cluster_deviation
      # Retrieves the cluster deviation value for outlier detection.
      #
      # This method fetches the sensitivity setting for outlier detection from a Chef role's
      # override attributes. If the sensitivity setting is not available or an error occurs,
      # it defaults to a value of "3". Any encountered errors are logged using Rails.logger.
      cluster_deviation = "3"
      begin
        manager_role = Chef::Role.manager
        cluster_deviation = manager_role.override_attributes['redBorder']['outliers']['AUTO_SENSITIVITY_SD']
      rescue StandardError => error
        Rails.logger.error "ERROR Sherlock: #{error}"
        Rails.logger.error error.backtrace.join("\n")
      end
      cluster_deviation
    end

    def self.request_body(query, conditions, broker)
      # Constructs and returns a request body for querying an anomaly detection system.
      # Arguments:
      # - query: Original Druid query in JSON format.
      # - conditions: A hash containing various conditions and settings for the request.
      # - broker: A broker object containing host and port information.
      # Returns a request body hash that includes the modified query, granularity details,
      # detection window, thresholds, and other configuration parameters required for anomaly detection.
      a = hook_druid_query(query, true)
      predict_interval = conditions[:end_time] - (Time.at(hook_druid_query(query, true)) rescue conditions[:start_time])
      detectionWindowTimes = JSON.parse(query.to_json)['intervals'][0].split('/')
      request_body = {}
      request_body["query"] = hook_druid_query(query, false).to_s rescue query.to_json
      request_body["granularity"] = "minute"
      request_body["granularityRange"] = granularity_range(conditions[:granularity])
      request_body["timeseriesRange"] = timeserie_range(predict_interval, request_body["granularity"]).to_i
      request_body["frequency"] = "day"
      queryEndTime = conditions[:end_time].to_i - math_mod(conditions[:end_time].to_i, request_body["granularityRange"].to_i * 60)
      request_body["queryEndTimeText"] = Time.at(queryEndTime).to_datetime.strftime('%Y-%m-%dT%H:%M')
      request_body["detectionWindow"] = self.calculate_query_length(detectionWindowTimes[0], detectionWindowTimes[1])
      request_body["hoursOfLag"] = "0"
      request_body["clusterId"] = "1"
      request_body["tsFramework"] = "Egads"
      request_body["adModels"] = Setting.anomaly_model
      request_body["sigmaThreshold"] = get_cluster_deviation
      request_body["tsModels"] = Setting.time_series_model
      request_body["brokerHost"] = broker.host
      request_body["brokerPort"] = broker.port
      return request_body
    end

    def self.outlier_r(query, conditions)
      # Executes an outlier detection query using the Sherlock anomaly detection system.
      # Arguments:
      # - query: Original Druid query in JSON format.
      # - conditions: A hash containing various conditions and settings for the query.
      # Returns a hash containing the results of the outlier detection process, including status,
      # anomalies, and predicted data, if available.
      outlier = {}
      config = sherlock_config
      begin
        broker = Redborder::Rbdruid.extract_broker_uri(query)
        http = Net::HTTP.new(config['host'], config['port'])
        response_body = send_request(http, request_body(query, conditions, broker))
        process_response(response_body, JSON.parse(query.to_json)['intervals'][0], outlier)
      rescue StandardError => error
        Rails.logger.error "2.ERROR Sherlock: #{error}"
        Rails.logger.error error.backtrace.join("\n")
        outlier
      end
      outlier
    end

    def self.process_response(response_body, interval, outlier)
      # Processes the response from the Sherlock anomaly detection system.
      # Arguments:
      # - response_body: The JSON response body received from the Sherlock system.
      # - interval: The time interval (start/end) for which the query was executed.
      # - outlier: A hash to store the processed response data.
      # Modifies the outlier hash to include status, anomalies, and predicted data (if successful),
      # or logs an error message if the response status indicates failure.
      outlier[:status] = response_body["status"]
      if response_body["status"] == "success"
        response_body = filter_predicted_data(response_body, interval)
        outlier[:anomalies] = response_body["anomalies"]
        outlier[:predicted] = response_body["predicted"]
      else
        Rails.logger.error "1.ERROR Sherlock: #{response_body['error']}"
      end
    end

    def self.sherlock_config
      # Loads and returns the Sherlock configuration based on the Rails environment.
      # Returns a hash containing the Sherlock configuration parameters.
      YAML.load_file("#{Rails.root}/config/sherlock.yml")[Rails.env]
    end

    def self.send_request(http, request_body)
      # Sends an HTTP request to the Sherlock system for executing an anomaly detection job.
      # Arguments:
      # - http: An initialized Net::HTTP object.
      # - request_body: The request body in JSON format for the anomaly detection job.
      # Returns a parsed JSON response received from the Sherlock system.
      request = Net::HTTP::Post.new('/Rb-Query/ProcessRbInstantAnomalyJob')
      request.add_field('Content-Type', 'application/json')
      request.body = request_body.to_json
      response = http.request(request)
      JSON.parse(response.body)
    end

  end
end
