#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'druid'
require 'zk'

#
# RedBorder utility module
#
module Redborder
  #
  # RedBorder wrapper to Druid ruby gem. It's provide high level functions
  # to return the data from Druid accordly to our app.
  #
  module Rbdruid
    @client = {}
    # Load Timeout from DruidConfig (20 seconds by default)
    @timeout        = RedBorder::DRUID_CONFIG[:timeout] || 20
    @timeout_millis = @timeout * 1000

    # rubocop:disable

    COMPARATIVE_COLUMNS = {
      pt24h: {
        name: "tr.report_blocks.form.hours",
        enabled: true,
        granularity: { table: :hour, comparative: :hour, graph: :minute },
        start: Proc.new { |time| time.beginning_of_hour },
        step: 1.hour,
        look_back: 1.month,
        format: '%H:%S h - %d/%m',
        tooltip_format: 'Minute %M'
      },
      pt168h: {
        name: "tr.report_blocks.form.days",
        enabled: true,
        granularity: { table: :pt24h, comparative: :hour, graph: :pt5m },
        start: Proc.new { |time| time.beginning_of_day },
        step: 1.day,
        look_back: 3.months,
        format: '%A - %d/%m',
        tooltip_format: 'Hour %H:%M'
      },
      pt720h: {
        name: "tr.report_blocks.form.weeks",
        enabled: false,
        granularity: { table: :pt168h, comparative: :pt24h, graph: :hour },
        start: Proc.new { |time| time.beginning_of_week },
        step: 1.week,
        look_back: 6.months,
        format: 'Week %V - %d/%m',
        tooltip_format: '%A, %H:%M'
      },
      pt4320h: {
        name: "tr.report_blocks.form.months",
        enabled: false,
        granularity: { table: :p1m, comparative: :pt168h, graph: :pt8h },
        start: Proc.new { |time| time.beginning_of_month },
        step: 1.month,
        look_back: 1.year,
        format: '%B',
        tooltip_format: 'Day %d - %H:%M'
      }
    }

    AGGREGATIONS = {}

    GRANULARITY_TYPE = {
      minute:          { long: '1 Minute',     short: '1m',    step: 60_000        },
      pt1m:            { long: '1 Minute',     short: '1m',    step: 60_000        },
      pt2m:            { long: '2 Minutes',    short: '2m',    step: 120_000       },
      pt5m:            { long: '5 Minutes',    short: '5m',    step: 300_000       },
      fifteen_minute:  { long: '15 Minutes',   short: '15m',   step: 900_000       },
      pt15m:           { long: '15 Minutes',   short: '15m',   step: 900_000       },
      thirty_minute:   { long: '30 Minutes',   short: '30m',   step: 1_800_000     },
      pt30m:           { long: '30 Minutes',   short: '30m',   step: 1_800_000     },
      hour:            { long: '1 Hour',       short: '1h',    step: 3_600_000     },
      pt60m:           { long: '1 Hour',       short: '1h',    step: 3_600_000     },
      pt2h:            { long: '2 Hours',      short: '2h',    step: 7_200_000     },
      pt8h:            { long: '8 Hours',      short: '8h',    step: 28_800_000    },
      pt24h:           { long: '1 Day',        short: '1d',    step: 86_400_000    },
      P1D:             { long: '1 Day',        short: '1d',    step: 86_400_000    },
      pt120h:          { long: '5 Days',       short: '5d',    step: 432_000_000   },
      pt168h:          { long: '1 Week',       short: '1w',    step: 604_800_000   },
      pt720h:          { long: '30 Days',      short: '30d',   step: 2_592_000_000 },
      p1m:             { long: '30 Days',      short: '30d',   step: 2_592_000_000 },
      all:             { long: 'all',          short: 'all'                        }

    }

    # rubocop:enable

    def self.get_clients
      @client
    end

    #
    # Reconnect to Druid based on Zookeeper. The timeout is
    # setted to 20 seconds.
    #
    def self.druid_reconnect controller
      @client[controller] = Druid::Client.new(
                              RedBorder::DRUID_CONFIG[:zk_connect],
                              discovery_path: '/druid/discoveryPath',
                              zk_keepalive: true,
                              http_timeout: @timeout
                            )
    end

    #
    # Send a query to Druid to get the current and predicted data
    #
    # == Parameters:
    # conditions::
    #   Hash with options:
    #     - range: range of data time
    #     - granularity: time between samples
    #     - controller: the controller to send the data
    #     - query_type: mode of the query. It can be group_by,
    #                   cardinality, search, group_by_cardinality
    #                   or timeseries.
    #
    # == Returns:
    # Return an array of data with the current interval data and the predicted
    # (based on R statistics)
    #
    def self.query(conditions)
      @range           = conditions[:range]
      @granularity     = conditions[:granularity]
      query_json       = get_query_from_params(conditions)
      controller       = conditions[:controller]

      # Data to return
      result = []

      # Gruad clause to connect if the connection is not defined
      druid_reconnect controller unless @client[controller]

      begin
        query = @client[controller].query(conditions[:data_sources])
      rescue RuntimeError
        druid_reconnect controller
        retry unless (attempts ||= 0 and attempts += 1) > 1
      end

      # Perform the query
      if query
        if @end_time.to_time > Time.now
          if @end_time > @start_time
            diff = @end_time - @start_time
          else
            diff = @start_time - @end_time
          end
          @end_time = Time.now
          @start_time = @end_time - diff
        end

        # Common query modificators
        query.interval(@start_time, @end_time)
        set_granularity(query, conditions[:granularity], conditions)
        query.instance_eval { @properties }[:filter] = query_json
        result = case conditions[:query_type]
                 when :cardinality
                   cardinality_query query, conditions
                 when :group_by
                   if conditions[:group_by].one?
                     (topn_query query, conditions).flatten(1)
                   else
                     group_by_query query, conditions
                   end
                 when :group_by_cardinality
                   group_by_cardinality_query query, conditions
                 when :search
                   search_query query, conditions
                 when :select
                   select_query query, conditions
                 else
                   timeseries_query query, conditions
                 end
      end

      # Return the results
      if result.is_a? Array
        result
      elsif result.is_a? Hash
        result
      else
        result.reverse
      end
    end

    #
    # Send a GroupBy query to Druid and return the data.
    #
    # These types of queries take a groupBy query object and
    # return an array of JSON objects where each object
    # represents a grouping asked for by the query.
    # From {http://druid.io/docs/0.7.0/GroupByQuery.html Druid.io Druid.io}
    #
    # == Parameters:
    # query::
    #   Object to perform the query
    # conditions::
    #   Hash with options:
    #     - range: range of data time
    #     - granularity: time between samples
    #     - controller: the controller to send the data
    #     - group_by: group conditions
    #
    # == Returns:
    # Reurn an array with the data filtered by the conditions.
    #
    def self.group_by_query(query, conditions)
      aggregation = set_aggregation(conditions[:aggregation])
      query.group_by(conditions[:group_by]) unless conditions[:group_by].blank?
      query.query_type(:groupBy)
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis
      aggregation(aggregation, query)

      conditions = sort_conditions(query, conditions)

      begin
        start = Time.now
        result =  if conditions[:outliers]
                    AIoutliers.outlier_r(query, conditions[:aggregation])
                  else
                    @client[conditions[:controller]].send(query)
                    .each do |element|
                      element['timestamp'] = element.timestamp
                    end
                  end
        # Log the query info to Rails
        log_query query, 'GroupBy', start, Time.now

      rescue Timeout::Error, Errno::ECONNREFUSED => e
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        raise Exception.new("The system could not process your data") if conditions[:raise_error]
        log_error query, 'GroupBy', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        raise Exception.new("The system could not process your data") if conditions[:raise_error]
        log_error query, 'GroupBy', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.topn_query(query, conditions)
      aggregation = set_aggregation(conditions[:aggregation])
      if aggregation.is_a?(Array)
        query.topn(conditions[:group_by][0], aggregation.first, conditions[:limit] || 100)
      else
        query.topn(conditions[:group_by][0], aggregation, conditions[:limit] || 100)
      end
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis
      aggregation(aggregation, query)

      begin
        start = Time.now
        result = @client[conditions[:controller]].send(query)
        # Log the query info to Rails
        log_query query, 'TopN', start, Time.now

        unless result.nil?
          result = result.each do |period|
            next if period.nil? || period.empty?
            period.each do |element|
              element['timestamp'] = period.timestamp
            end
          end
        end

        result = result.first if conditions[:granularity].to_sym == :all
      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'TopN', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'TopN', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.cardinality_query query, conditions
      query.query_type(:timeseries)
      query.instance_eval{@properties}[:aggregations] = [{type: 'cardinality', name: 'result', fieldNames: [conditions[:cardinality_column]]}]
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis

      begin
        start = Time.now
        result = @client[conditions[:controller]].send(query)
        # Log the query info to Rails
        log_query query, 'Cardinality', start, Time.now
        result = result.each do |element|
          element['timestamp'] = element.timestamp
          element['result'] = element['result'].round
        end
      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'Cardinality', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'Cardinality', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.group_by_cardinality_query query, conditions
      query.topn(conditions[:group_by][0], 'result', conditions[:limit])
      query.instance_eval{@properties}[:aggregations] = [{type: 'cardinality', name: 'result', fieldNames: [conditions[:cardinality_column]]}]
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis

      begin
        start = Time.now
        result = @client[conditions[:controller]].send(query)
        # Log the query info to Rails
        log_query query, 'GroupByCardinality', start, Time.now

        result = result.first unless result.blank?
        result.each do |element|
          element['result'] = element['result'].round
        end
      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'GroupByCardinality', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'GroupByCardinality', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.timeseries_query(query, conditions)
      aggregation = set_aggregation(conditions[:aggregation])
      query_properties = query.instance_eval { @properties }

      query.query_type(:timeseries)
      query_properties[:context] ||= {}
      query_properties[:context][:timeout] = @timeout_millis

      conditions = sort_conditions(query, conditions)

      aggregation(aggregation, query)

      # Dont fill with zeros
      query_properties[:context]['skipEmptyBuckets'] =
        if conditions[:skip_empty].nil?
          'true'
        else
          conditions[:skip_empty].to_s
        end

      begin
        start = Time.now
        result =  if conditions[:outliers]
                    AIoutliers.outlier_r(query, conditions[:aggregation])
                  else
                    @client[conditions[:controller]].send(query)
                    .each do |element|
                      element['timestamp'] = element.timestamp
                    end
                  end
        # Log the query info to Rails
        log_query query, 'TimeSeries', start, Time.now

      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'TimeSeries', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'TimeSeries', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.search_query query, conditions
      query.instance_eval{@properties}[:queryType] = :search
      query.instance_eval{@properties}[:query] = {type: 'insensitive_contains', value: conditions[:search_value]}
      query.instance_eval{@properties}[:sort] = {type: 'lexicographic'}
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis

      begin
        start = Time.now
        result = @client[conditions[:controller]].send(query)
        # Log the query info to Rails
        log_query query, 'Search', start, Time.now

        result = result.first unless result.nil?
      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'Search', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'Search', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.select_query query, conditions
      query.instance_eval{@properties}[:queryType] = :select
      query.instance_eval{@properties}[:query] = {type: 'insensitive_contains', value: conditions[:search_value]}
      query.instance_eval{@properties}[:sort] = {type: 'lexicographic'}
      query.instance_eval{@properties}[:context] = {}
      query.instance_eval{@properties}[:context][:timeout] = @timeout_millis
      # Conditions of pagination
      query.instance_eval{@properties}[:pagingSpec] = conditions[:pagingSpec]

      begin
        start = Time.now
        result = @client[conditions[:controller]].send(query)
        # Log the query info to Rails
        log_query query, 'Select', start, Time.now
      rescue Timeout::Error, Errno::ECONNREFUSED
        druid_reconnect conditions[:controller]
        retry unless (attempts ||= 0 and attempts += 1) > 1
        log_error query, 'Select', start, Time.now
      rescue RuntimeError => e
        Rails.logger.error e
        Rails.logger.error e.backtrace.join("\n")
        log_error query, 'Select', start, Time.now
      ensure
        result ||= []
      end

      result
    end

    def self.set_granularity(query, gran, params)
      gran = case gran.to_s
      when 'minute'
        'pt1m'
      when 'fifteen_minute'
        'pt15m'
      when 'thirty_minute'
        'pt30m'
      when 'hour'
        'pt60m'
      when 'day'
        'P1D'
      when 'none', 'all'
        query.instance_eval{@properties}[:granularity] = gran.to_s
        return
      else
        gran.to_s
      end

      timezone = (ActiveSupport::TimeZone[params.fetch(:timezone, 'UTC')] or Time.zone)

      query.instance_eval{@properties}[:granularity] = {
        type: 'period',
        period: gran,
        #timeZone: timezone.tzinfo.name,
        origin: timezone.at(params[:start_time].to_time).beginning_of_hour.utc.iso8601
      }
    end

    def self.set_aggregation(aggregation = :bytes)
      if aggregation.is_a? Array
        aggregation.map { |a| AGGREGATIONS[a.to_sym][:value] }
      elsif AGGREGATIONS[aggregation.to_sym].blank?
        :bytes
      else
        AGGREGATIONS[aggregation.to_sym][:value]
      end
    end

    def self.set_defaults(params)
      @start_time = params[:start_time]
      @end_time = params[:end_time]

      if @range.to_sym == :raw_column and @granularity.to_sym != :all
        @start_time = [(params[:start_time] or 0), @end_time - GRANULARITY_TYPE[@granularity.to_sym][:step] / 1000].max
      end
    end

    # TODO must be tested and modified
    def self.get_query_from_params(params)
      @range = params[:range].blank? ? 'last_1_hour' : params[:range]
      self.set_defaults(params)
      if params[:monitor_tab].present?
        #Monitor
        query_json  = {
          type: 'and',
          fields: [{ type: 'selector', dimension: 'monitor', value: params[:monitor_tab]}]
        }

        #TODO: Check if we need to do this or its added on the filters....
        if params[:bulkstats].present?
          query_json[:fields] << { type: 'selector', dimension: 'bulkstats', value: params[:bulkstats] }
        end

        if params[:monitor_sensor].present?
          query_json[:fields] << { type: 'selector', dimension: 'sensor_name', value: params[:monitor_sensor]}
          query_json[:fields] << { type: 'selector', dimension: 'group_name', value: params[:group_name]} if params[:group_name]
          if params[:instance].present?
            query_json[:fields] << { type: 'selector', dimension: 'instance', value: params[:instance]}
          end
        elsif params.has_key?(:search)
          params[:search].each do |k, v|
            if [:sensor_name, :group_name, :group_id, :sensor_id, :organization_uuid,
                :campus, :campus_uuid, :building, :building_uuid,
                :deployment, :deployment_uuid, :market, :market_uuid, :organization,
                :service_provider, :service_provider_uuid, :namespace, :namespace_uuid].include?(k.to_sym)
              sub_query_json = self.value_sub_query_json(v.values.first[:operator])
              v.each do |key, val|
                sub_query_json[:fields] << self.parse_filter_list(val[:operator].to_s, k, key)
              end
              query_json[:fields] << ((sub_query_json[:fields].size > 1) ? sub_query_json : sub_query_json[:fields][0] )
            end
          end
        end
        query_json = query_json[:fields][0] unless query_json[:fields].size > 1
      else
        #IPS, FLOW
        query_json = nil
        if params.has_key?(:search)
          query_json  = {
            type: 'and',
            fields: []
          }

        if params[:search] != nil
          params[:search].each do |column, v|

            if ![:start_time, :end_time].include?(column.to_sym)
              if v.size > 1
                # Multiple filter (different values for same column - include/or || exclude/and)
                sub_query_json = self.value_sub_query_json(v.values.first[:operator])
                merge_filters_and = { type: 'and', fields: [] }
                merge_filters_or = { type: 'or', fields: [] }
                v.each do |key, val|
                  if ['contains_not', 'is_not'].include?(val[:operator]) ||
                     (column == :window && val[:operator] != 'is')
                    merge_filters_and[:fields] << self.parse_filter_list(val[:operator], column, key)
                  else
                    merge_filters_or[:fields] << self.parse_filter_list(val[:operator], column, key)
                  end
                end
                if merge_filters_and[:fields].blank?
                  sub_query_json[:type] = 'or'
                  sub_query_json[:fields] = merge_filters_or[:fields]
                else
                  # merge_filters_or[:fields].each{|h| merge_filters_and[:fields] << h}
                  merge_filters_and[:fields] << merge_filters_or if merge_filters_or[:fields].present?
                  sub_query_json[:type] = 'and'
                  sub_query_json[:fields] = merge_filters_and[:fields]
                end
                query_json[:fields] << ((sub_query_json[:fields].size > 1) ? sub_query_json : sub_query_json[:fields][0] )
              else
                # Simple filters
                v.each do |key, val|
                  sub_query_json = self.value_sub_query_json(val[:operator].to_s)
                  columns_list = {}

                  Redborder::Module.each_with_interface do |id, properties|
                    if properties[:customizable_tabs]
                      if properties[:columns].respond_to? :keys
                        columns_list.merge! properties[:columns]
                      else
                        columns_list.merge! properties[:tabs]
                      end
                    end
                  end

                  if columns_list.keys.include?(column.to_sym)
                    columns_list[column.to_sym][:columns].keys.each do |l|
                      sub_query_json[:fields] << self.parse_filter_list(val[:operator], l, key)
                    end
                  else
                    sub_query_json[:fields] << self.parse_filter_list(val[:operator], column.to_sym, key)
                  end
                  query_json[:fields] << ((sub_query_json[:fields].size > 1) ? sub_query_json : sub_query_json[:fields][0] )
                end
              end
            end
          end
        end

          if params.has_key?(:custom_filter)
            params[:custom_filter].each do |k, v|
              query_json[:fields] << {type: 'selector', dimension: k, value: v} unless v.blank? or AGGREGATIONS.keys.include?(k.to_sym)
            end
          end

          query_json = query_json[:fields][0] unless query_json[:fields].size > 1
        end
        if query_json.nil? and params.has_key?(:custom_filter)
          query_json = self.custom_filter(params)
        end
      end

      unless params[:sensors_monitor].nil?
        sensors_filters = { type: :or, fields: [] }

        # Add IPS sensors
        params[:sensors_monitor][:ips].each do |item|
          sensors_filters[:fields] << {
            type: :and,
            fields: [
              { type: :selector,
                dimension: :sensor_name,
                value: item[:sensor_name] },
              { type: :selector,
                dimension: :group_name,
                value: item[:group_name] }
            ]
          }
        end

        # Add Flow sensors
        params[:sensors_monitor][:flow].each do |sensor|
          sensors_filters[:fields] << {
            type: :selector,
            dimension: :sensor_name,
            value: sensor
          }
        end

        unless sensors_filters[:fields].empty?
          if query_json.nil?
            query_json = sensors_filters
          else
            tmp = query_json
            query_json = { type: :and, fields: [] }
            query_json[:fields] << tmp
            query_json[:fields] << sensors_filters
          end
        end

      end

      unless params[:sensors].nil?
        sensors_filters = { type: :or, fields: [] }

        params[:sensors].each do |filter_field_name, sensors|
          if sensors.empty?
            sensors_filters[:fields] << {
              type: :selector,
              dimension: filter_field_name,
              value: '0'
            }
          else
            sensors.each do |sensor|
              sensors_filters[:fields] << {
                type: :selector,
                dimension: filter_field_name,
                value: sensor
              }
            end
          end
        end

        unless sensors_filters[:fields].empty?
          if query_json.nil?
            query_json = sensors_filters
          else
            tmp = query_json
            query_json = { type: :and, fields: [] }
            query_json[:fields] << tmp
            query_json[:fields] << sensors_filters
          end
        end
      end

      query_json
    end

    def self.parse_filter_list(operator, column, value)
      case operator.to_s
      when 'is'
        {type: 'selector', dimension: column, value: value}
      when 'is_not', 'not'
        {type: 'not', field: {type: 'selector', dimension: column, value: value}}
      when 'contains'
        {type: 'regex', dimension: column, pattern: '(?i)^.*' + value + '.*$' }
      when 'contains_not'
        {type: 'regex', dimension: column, pattern: '(?i)^((?!' + value + ').)*$' }
      when 'starts_with'
        {type: 'regex', dimension: column, pattern: '(?i)^' + value }
      when 'ends_with'
        {type: 'regex', dimension: column, pattern: '(?i)' + value + '$' }
      when 'regex'
        {type: 'regex', dimension: column, pattern: '(?i)' + value}
      when 'empty'
        {type: 'selector', dimension: column, value: ''}
      when 'not_empty'
        {type: 'not', field: {type: 'selector', dimension: column, value: ''}}
      when 'lte'
        {type: "javascript", dimension: column, function: "function(x) { return(x <= #{value}) }"}
      when 'gte'
        {type: "javascript", dimension: column, function: "function(x) { return(x >= #{value}) }"}
      when 'lt'
        {type: "javascript", dimension: column, function: "function(x) { return(x < #{value}) }"}
      when 'gt'
        {type: "javascript", dimension: column, function: "function(x) { return(x > #{value}) }"}
      end
    end

    def self.value_sub_query_json(value)
      case value
      when 'is_not', 'not', 'contains_not', 'lte', 'gte', 'lt', 'gt'
        { type: 'and', fields: [] }
      else
        { type: 'or', fields: [] }
      end
    end

    def self.custom_filter(params)
      query_json  = {
        type: 'and',
        fields: []
      }
      params[:custom_filter].each do |k, v|
        query_json[:fields] << {type: 'selector', dimension: k, value: v} unless v.blank? or AGGREGATIONS.keys.include?(k.to_sym)
      end
      query_json = query_json[:fields][0] unless query_json[:fields].size > 1
      return query_json
    end

    def self.granularity_list(params = {})
      time_difference = params[:end_time] - params[:start_time]

      if time_difference < 2.hour
        [:minute, :pt2m, :pt5m, :fifteen_minute, :thirty_minute, :all]
      elsif time_difference < 8.hours
        [:minute, :pt2m, :pt5m, :fifteen_minute, :thirty_minute, :hour, :pt2h, :all]
      elsif time_difference < 2.day
        [:pt5m, :fifteen_minute, :thirty_minute, :hour, :pt2h, :pt8h, :all]
      elsif time_difference < 5.days
        [:fifteen_minute, :thirty_minute, :hour, :pt2h, :pt8h, :pt24h, :all]
      elsif time_difference < 1.week
        [:fifteen_minute, :thirty_minute, :hour, :pt2h, :pt8h, :pt24h, :pt120h, :all]
      elsif time_difference < 20.days
        [:hour, :pt2h, :pt8h, :pt24h, :pt120h, :pt168h, :all]
      elsif time_difference < 40.days
        [:pt2h, :pt8h, :pt24h, :pt120h, :pt168h, :all]
      elsif time_difference < 100.days
        [:pt8h, :pt24h, :pt120h, :pt168h, :pt720h, :all]
      elsif time_difference < 200.days
        [:pt24h, :pt120h, :pt168h, :pt720h, :all]
      else
        [:pt120h, :pt168h, :pt720h, :all]
      end
    end

    def self.slow_granularity_list(params = {})
      time_difference = params[:end_time] - params[:start_time]

      if time_difference < 2.hour
        []
      elsif time_difference < 8.hours
        []
      elsif time_difference < 2.day
        [:minute, :pt2m]
      elsif time_difference < 5.days
        [:pt2m, :pt5m]
      elsif time_difference < 1.week
        [:pt2m, :pt5m]
      elsif time_difference < 20.days
        [:fifteen_minute, :thirty_minute]
      elsif time_difference < 40.days
        [:thirty_minute, :hour]
      elsif time_difference < 100.days
        [:hour, :pt2h]
      elsif time_difference < 200.days
        [:pt2h, :pt8h]
      else
        [:pt8h, :pt24h]
      end
    end

    def self.get_druid_first_interval(datasource)
      times = []
      datasource = datasource.is_a?(String) ? [datasource] : datasource["dataSources"]
      uri = get_all_coordinator_info.sample

      datasource.each do |ds|
        begin
          url = "http://#{uri['hostname']}:#{uri['port']}/druid/coordinator/v1/datasources/#{ds}/"
          uri = URI(url)
          response = Net::HTTP.get(uri)
          data = JSON.parse(response)
          times.push(Time.parse(data["segments"]["minTime"]).to_i / 60)
        rescue StandardError => e
          Rails.logger.error "Error in get_druid_first_interval, there is no segments or something went wrong..."
        end
      end

      times.min
    end

    def self.get_all_coordinator_info
      zk = ZK.new RedBorder::DRUID_CONFIG[:zk_connect], chroot: :check
      watchPath = '/druid/discoveryPath/coordinator'
      live = zk.children(watchPath, watch: true)

      coordinator_info = []

      live.each do |name|
        info = zk.get("#{watchPath}/#{name}")
        node = JSON.parse(info[0])
        coordinator_info << { 'hostname' => node['address'], 'port' => node['port'] }
      end

      zk.close
      coordinator_info
    end

    def self.extract_broker_uri(query)
      broker_uri = query.instance_eval { @client }.instance_eval { @zk }
                        .instance_eval { @brokers }[0][:uri]
      URI(broker_uri)
    end

    def self.get_dimensions(data_type, start_time, end_time)
      datasource = RedBorder::DRUID_CONFIG[data_type.to_sym]['datasource']
      start_time = start_time.strftime('%Y-%m-%dT%H:%M:%S.000Z')
      end_time   = end_time.strftime('%Y-%m-%dT%H:%M:%S.000Z')

      zk = ZK.new RedBorder::DRUID_CONFIG[:zk_connect], chroot: :check
      watchPath = '/druid/discoveryPath/master'
      live = zk.children(watchPath, watch: true)

      dimensions = Array.new

      live.each do |name|
        info = zk.get '#{watchPath}/#{name}'
        node = JSON.parse(info[0])
        uri =  "http://#{node['address']}:#{node['port']}"

        response = Net::HTTP.get_response(URI('#{uri}/druid/v2/datasources/#{datasource}/?interval=#{start_time}/#{end_time}'))

        if response.code == '200'
          dimensions += JSON.parse(response.body)['dimensions']
        end
      end

      zk.close
      dimensions.uniq
    end

    def self.timeseries_aggregation(aggregation, query)
      query_properties = query.instance_eval { @properties }
      query_properties[:aggregations] ||= []
      query_properties[:postAggregations] ||= []

      aggregation_data = []

      if aggregation.is_a?(Array)
        aggregation.each do |agg|
          aggregation_data << AGGREGATIONS[agg.to_sym]
        end
      else
        aggregation_data << AGGREGATIONS[aggregation.to_sym]
      end

      aggregation_data.each do |agg_data|
        aggregation_timeseries = agg_data[:timeseries]

        if aggregation_timeseries
          if @granularity.to_sym != :all
            if aggregation_timeseries[:per_sec]
              from = aggregation_timeseries[:per_sec][:from]
              to = aggregation_timeseries[:per_sec][:to]
              step_secs = GRANULARITY_TYPE[@granularity.to_sym][:step] / 1000

              query_properties[:postAggregations] << {
                type: 'arithmetic',
                fn: '/',
                name: to,
                fields: [
                  {
                    type: 'fieldAccess',
                    fieldName: from
                  },
                  {
                    type: 'constant',
                    value: step_secs
                  }
                ]
              }
            end

            if aggregation_timeseries[:multiplier]
              from = aggregation_timeseries[:multiplier][:from]
              to = aggregation_timeseries[:multiplier][:to]
              value = aggregation_timeseries[:multiplier][:value]

              query_properties[:postAggregations] << {
                type: 'arithmetic',
                fn: '*',
                name: to,
                fields: [
                  {
                    type: 'fieldAccess',
                    fieldName: from
                  },
                  {
                    type: 'constant',
                    value: value
                  }
                ]
              }
            end
          else
            if aggregation_timeseries[:default]
              from = aggregation_timeseries[:default][:from]
              to = aggregation_timeseries[:default][:to]

              # We multiply by one, since we cant use druid's fieldAccess
              # as a single postAggregator.
              query_properties[:postAggregations] << {
                type: 'arithmetic',
                fn: '*',
                name: to,
                fields: [
                  {
                    type: 'fieldAccess',
                    fieldName: from
                  },
                  {
                    type: 'constant',
                    value: 1
                  }
                ]
              }
            end
          end
        end

        if agg_data[:timeserie_aggregations].present?
          query_properties[:aggregations] += agg_data[:timeserie_aggregations].dup
        elsif agg_data[:aggregations].present?
          query_properties[:aggregations] += agg_data[:aggregations].dup
        end

        # It fails if there are more than one aggregation with same value
        query_properties[:aggregations].uniq! { |a| a[:name] }

        if agg_data[:post_aggregations].present?
          query_properties[:postAggregations] += agg_data[:post_aggregations].dup
        end
      end
    end

    # TODO: Improve aggregation when is an Array. Improve code.
    def self.aggregation(aggregation, query)
      @time_value = granularity_value(query)

      query_properties = query.instance_eval { @properties }
      query_properties[:aggregations] ||= []
      query_properties[:postAggregations] ||= []

      aggregation_data = {}
      aggregation_data[:aggregations] = []
      aggregation_data[:post_aggregations] = []

      [*aggregation].each do |a|
        aggregation_data[:aggregations] |= AGGREGATIONS[a.to_sym][:aggregations]
        aggregation_data[:post_aggregations] |= resolve_aggregation(AGGREGATIONS[a.to_sym][:post_aggregations].deep_dup, @time_value) if AGGREGATIONS[a.to_sym][:post_aggregations]

        aggregation_timeseries = AGGREGATIONS[a.to_sym][:timeseries]
        next unless aggregation_timeseries
        next unless aggregation_timeseries[:default]

        from = aggregation_timeseries[:default][:from]
        to = aggregation_timeseries[:default][:to]

        # We multiply by one, since we cant use druid's fieldAccess
        # as a single postAggregator.

        # query_properties[:postAggregations] << {
        #   type: 'arithmetic',
        #   fn: '*',
        #   name: to,
        #   fields: [
        #     {
        #       type: 'fieldAccess',
        #       fieldName: from
        #     },
        #     {
        #       type: 'constant',
        #       value: 1
        #     }
        #   ]
        # }
      end

      unless aggregation_data[:aggregations].nil?
        query_properties[:aggregations] += aggregation_data[:aggregations].dup
      end

      unless aggregation_data[:post_aggregations].nil?
        query_properties[:postAggregations] += aggregation_data[:post_aggregations].dup
      end
    end

    def self.resolve_aggregation(aggregation, value)
      if aggregation.is_a? Array
        aggregation.flatten.each { |x| resolve_aggregation(x, value) }
      elsif aggregation.is_a? Hash
        aggregation.each do |k, v|
          if v.is_a? Proc
            aggregation[k] = v.call(value)
          else
            resolve_aggregation(v, value)
          end
        end
      end
    end

    def self.granularity_value(query)
      if query.properties[:granularity] == 'all'
        query.properties[:intervals].inject(0) do |a, b|
          a + b.split('/').map { |t| DateTime.parse(t) }.each_cons(2).map { |t1, t2| t2.to_i - t1.to_i }[0]
        end
      elsif query.properties[:granularity].is_a? Hash
        period = query.properties[:granularity][:period].to_sym
        GRANULARITY_TYPE[period][:step] / 1000
      else
        GRANULARITY_TYPE[query.properties[:granularity].to_sym][:step] / 1000
      end
    end

    def self.search_event_dimension_name(dimension)
      Event::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return name
          end
        end
      end

      ''
    end

    def self.search_event_tab_by_dimension(dimension)
      Event::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return tab.to_s
          end
        end
      end

      nil
    end

    def self.search_traffic_dimension_name(dimension)
      Traffic::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return name
          end
        end
      end

      ''
    end

    def self.search_traffic_tab_by_dimension(dimension)
      Traffic::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return tab.to_s
          end
        end
      end

      nil
    end

    def self.search_social_dimension_name(dimension)
      Social::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return name
          end
        end
      end

      ''
    end

    def self.search_social_tab_by_dimension(dimension)
      Social::COLUMNS.each do |tab, values|
        values[:columns].each do |column, name|
          if column.to_s == dimension
            return tab.to_s
          end
        end
      end

      nil
    end

    #
    # Set a condition to the query to sort the results
    #
    # == Parameters:
    # query::
    #   Druid query object
    # conditions::
    #   Hash with the conditions of the query
    #
    # == Returns:
    # Hash with conditions updated
    #
    def self.sort_conditions(query, conditions)
      if conditions[:sort] #&& [*conditions[:sort]].any?{ |sort| sort.to_sym != :timestamp }
        query.instance_eval{@properties}[:limitSpec] = {
          'type' => 'default',
          'limit' => conditions[:limit] || 100,
          'columns' => []
        }

        [*conditions[:sort]].each do |sort|
          query.instance_eval{ @properties }[:limitSpec]['columns'] << {
            'dimension' => sort.to_s,
            'direction' => conditions[:direction].to_s.upcase
          } unless sort.to_sym == :timestamp
        end
      end

      conditions
    end

    #
    # Log the query to Rails
    #
    # == Parameters:
    # query::
    #   The query with the params to log
    # type::
    #   String with the name of the query
    # start::
    #   Time object with start time
    # end_time::
    #   Time object with end time
    #
    def self.log_query(query, type, start, finish)
      info = green("DRUID #{type} (#{((finish - start) * 1000.0).round(2)}ms)")
      Rails.logger.debug "  #{info} #{query.to_json}"
    end

    #
    # Color a text on log
    #
    # == Parameters:
    # str::
    #   String to color
    #
    # == Returns:
    # String escaped with colors
    #
    def self.green(str)
      "\033[32m#{str}\033[0m"
    end

    #
    # Color a text red on log
    #
    # == Parameters:
    # str::
    #   String to color
    #
    # == Returns:
    # String escaped with colors
    #
    def self.red(str)
      "\033[31m#{str}\033[0m"
    end

    #
    # Log an error in a query
    #
    def self.log_error(query, type, start, finish)
      info = "DRUID #{type} (#{((finish - start) * 1000.0).round(2)}ms)"
      Rails.logger.error red("  #{info} #{query.to_json}")
    end
  end
end
