#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Miguel Álvarez <malvarez@redborder.com>
#######################################################################

require 'net/ssh'
require 'net/sftp'
require 'net/ftp'
require 'net/http'
#
# Base redBorder module
#
module Redborder
  #
  # Module to interecat with ftp, ftps, and sftp
  #
  module TIConnections
    class Client
      #
      # Initialize the FTP/HTTP client with the specified protocol, host, user, and password.
      #
      # == Parameters:
      # protocol::
      #   A symbol representing the protocol to use (:ftp, :sftp, :ftps, :http, :https).
      # host::
      #   A string representing the host of the FTP server.
      # user::
      #   A string representing the username for authentication.
      # password::
      #   A string representing the password for authentication.
      #
      # == Returns:
      # A new instance of Redborder::TIConnections::Client.
      #
      def initialize(protocol:, host:, port:, user:, password:)
        @protocol = protocol.downcase.to_sym
        @host = host
        @port = port
        @user = user
        @password = password
        @connection = nil
        connect
      end

      #
      # Establish a connection to the FTP, SFTP, or FTPS server based on the provided protocol.
      #
      # == Raises:
      # ArgumentError::
      #   If the provided protocol is unsupported.
      # Net::FTPError::
      #   If there is an error establishing an FTP connection.
      # Net::SSH::AuthenticationFailed::
      #   If the authentication fails for SFTP.
      # SocketError::
      #   If there is a network error.
      # IOError::
      #   If there is an I/O error during connection.
      #
      def connect
        @connection = case @protocol
                      when :ftp then connect_ftp
                      when :sftp then connect_sftp
                      when :ftps then connect_ftps
                      when :http then connect_http
                      when :https then connect_https
                      else raise ArgumentError, "Unsupported protocol: #{@protocol}"
                      end
      rescue ArgumentError => e
        Rails.logger.error "Configuration error: #{e.message}"
        raise
      rescue Net::FTPError => e
        Rails.logger.error "FTP connection error: #{e.message}"
        raise
      rescue Net::SSH::AuthenticationFailed => e
        Rails.logger.error "SFTP authentication error: #{e.message}"
        raise
      rescue SocketError => e
        Rails.logger.error "Network error: #{e.message}"
        raise
      rescue IOError => e
        Rails.logger.error "I/O error during connection: #{e.message}"
        raise
      end

      #
      # Download a file from the remote FTP/SFTP server to the local machine.
      #
      # == Parameters:
      # remote_path::
      #   A string representing the remote file path to download.
      # local_path::
      #   A string representing the local file path to store the downloaded file.
      #
      # == Returns:
      # None.
      #
      # == Raises:
      # ArgumentError::
      #   If the protocol does not support downloading.
      # Net::FTPError::
      #   If there is an error during the FTP download.
      # Net::SFTP::StatusException::
      #   If there is an error during the SFTP download.
      # IOError::
      #   If there is an I/O error during the download.
      #
      def download_file(remote_path, local_path)
        with_connection do
          case @protocol
          when :ftp, :ftps
            @connection.getbinaryfile(remote_path, local_path)
          when :sftp
            @connection.download!(remote_path, local_path)
          when :http, :https
            download_file_http(remote_path, local_path)
          else
            raise ArgumentError, "Unsupported protocol for download: #{@protocol}"
          end
        end
      rescue Net::FTPError => e
        Rails.logger.error "FTP error during download: #{e.message}"
        raise
      rescue Net::SFTP::StatusException => e
        Rails.logger.error "SFTP error during download: #{e.message}"
        raise
      rescue IOError => e
        Rails.logger.error "I/O error during file download: #{e.message}"
        raise
      end
  
      #
      # Download a file over HTTP or HTTPS.
      #
      # == Parameters:
      # remote_path::
      #   A string representing the remote file path (URL) to download.
      # local_path::
      #   A string representing the local file path to store the downloaded file.
      #
      # == Raises:
      # IOError::
      #   If the HTTP/HTTPS download fails.
      #
      def download_file_http(remote_path, local_path)
        uri = URI.join("http://#{@host}", remote_path) if @protocol == :http
        uri = URI.join("https://#{@host}", remote_path) if @protocol == :https

        Net::HTTP.start(uri.host, uri.port, use_ssl: @protocol == :https) do |http|
          request = Net::HTTP::Get.new(uri)

          http.request(request) do |response|
            case response
            when Net::HTTPSuccess
              File.open(local_path, 'wb') do |file|
                response.read_body do |chunk|
                  file.write(chunk)
                end
              end
            else
              raise IOError, "HTTP/HTTPS download failed: #{response.message} (#{response.code})"
            end
          end
        end
      rescue StandardError => e
        raise IOError, "HTTP/HTTPS download failed: #{e.message}"
      end

      #
      # Upload a file from the local machine to the remote FTP/SFTP server.
      #
      # == Parameters:
      # local_path::
      #   A string representing the local file path to upload.
      # remote_path::
      #   A string representing the remote file path to upload to.
      #
      # == Returns:
      # None.
      #
      # == Raises:
      # ArgumentError::
      #   If the protocol does not support uploading.
      # Net::FTPError::
      #   If there is an error during the FTP upload.
      # Net::SFTP::StatusException::
      #   If there is an error during the SFTP upload.
      # IOError::
      #   If there is an I/O error during the upload.
      #
      def upload_file(local_path, remote_path)
        with_connection do
          case @protocol
          when :ftp, :ftps
            @connection.putbinaryfile(local_path, remote_path)
          when :sftp
            @connection.upload!(local_path, remote_path)
          else
            raise ArgumentError, "Unsupported protocol for upload: #{@protocol}"
          end
        end
      rescue Net::FTPError => e
        Rails.logger.error "FTP error during upload: #{e.message}"
        raise
      rescue Net::SFTP::StatusException => e
        Rails.logger.error "SFTP error during upload: #{e.message}"
        raise
      rescue IOError => e
        Rails.logger.error "I/O error during file upload: #{e.message}"
        raise
      end

      #
      # Close the connection to the FTP/SFTP server.
      #
      # == Returns:
      # None.
      #
      # == Raises:
      # ArgumentError::
      #   If the protocol is unsupported when closing.
      # IOError::
      #   If there is an I/O error during closing.
      #
      def close_connection
        return unless @connection

        case @protocol
        when :ftp, :ftps
          @connection.close
        when :sftp
          @connection.close unless @connection.closed?
        when :http, :https
          nil
        else
          raise ArgumentError, "Unsupported protocol for closing connection: #{@protocol}"
        end
      rescue IOError => e
        Rails.logger.error "I/O error during connection close: #{e.message}"
        raise
      end

      private

      #
      # Establish an FTP connection using the provided credentials.
      #
      # == Returns:
      # An instance of Net::FTP representing the connection.
      #
      # == Raises:
      # Net::FTPError::
      #   If the FTP connection fails.
      #
      def connect_ftp
        ftp = Net::FTP.new
        ftp.connect(@host, @port)
        ftp.login(@user, @password)
        ftp
      rescue Net::FTPError => e
        raise Net::FTPError, "FTP connection failed: #{e.message}"
      end

      #
      # Establish an SFTP connection using the provided credentials.
      #
      # == Returns:
      # An instance of Net::SFTP::Session representing the connection.
      #
      # == Raises:
      # Net::SSH::AuthenticationFailed::
      #   If the SFTP authentication fails.
      # SocketError::
      #   If there is a network error.
      #
      def connect_sftp
        ssh = Net::SSH.start(@host, @user, {
          :password => @password,
          :port => @port
        })
        ssh.sftp
      rescue Net::SSH::AuthenticationFailed => e
        raise Net::SSH::AuthenticationFailed, "SFTP authentication failed: #{e.message}"
      rescue SocketError => e
        raise SocketError, "Failed to reach SFTP server: #{e.message}"
      end

      #
      # Establish an FTPS connection using the provided credentials.
      #
      # == Returns:
      # An instance of Net::FTP representing the FTPS connection.
      #
      # == Raises:
      # Net::FTPError::
      #   If the FTPS connection fails.
      #
      def connect_ftps
        ftp = Net::FTP.new
        ftp.connect(@host, @port)
        ftp.login(@user, @password)
        ftp.encrypted = true
        ftp
      rescue Net::FTPError => e
        raise Net::FTPError, "FTPS connection failed: #{e.message}"
      end

      #
      # Establish an HTTP connection using the provided credentials.
      #
      # == Returns:
      # An instance of Net::HTTP representing the connection.
      #
      # == Raises:
      # IOError::
      #   If the HTTP connection fails.
      #
      def connect_http
        http = Net::HTTP.new(@host, @port)
        http.use_ssl = false
        response = http.start do |conn|
          request = Net::HTTP::Get.new('/')
          conn.request(request)
        end
        unless response.is_a?(Net::HTTPSuccess)
          raise IOError, "HTTP connection failed: #{response.message} (#{response.code})"
        end
        http
      rescue IOError => e
        raise IOError, "HTTP connection failed: #{e.message}"
      end

      #
      # Establish an HTTPS connection using the provided credentials.
      #
      # == Returns:
      # An instance of Net::HTTP representing the connection.
      #
      # == Raises:
      # IOError::
      #   If the HTTPS connection fails.
      #
      def connect_https
        https = Net::HTTP.new(@host, @port)
        https.use_ssl = true
        https.verify_mode = OpenSSL::SSL::VERIFY_PEER
        response = https.start do |conn|
          request = Net::HTTP::Get.new('/')
          conn.request(request)
        end
        unless response.is_a?(Net::HTTPSuccess)
          raise IOError, "HTTPS connection failed: #{response.message} (#{response.code})"
        end
        https
      rescue IOError => e
        raise IOError, "HTTPS connection failed: #{e.message}"
      end
      #
      # Ensures that a valid connection is established before performing an action.
      #
      # == Raises:
      # RuntimeError::
      #   If no connection is established.
      #
      def with_connection
        raise RuntimeError, "No active connection. Call connect first." unless @connection

        yield
      rescue RuntimeError => e
        Rails.logger.error "Runtime error: #{e.message}"
        raise
      end
    end
  end
end