#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module RedBorder
  # This class manages the mapping of intrusion rules to Mitre tactics and techniques.
  class IntrusionRulesToMitre
    # Start the process of assigning intrusion rules to Mitre tactics and techniques.
    def start
      ActiveRecord::Base.transaction do
        # Retrieve distinct intrusion rules where Mitre tactic or technique IDs are present.
        @rules = Rule.where("mitre_tactic_id != 'N/A' OR mitre_technique_id != 'N/A'")
                     .select(:rule_sid, :mitre_tactic_id, :mitre_technique_id)
                     .distinct
                     .to_a
        # Assign the rules to corresponding tactics and techniques.
        assign
      end
    end

    # Return the number of rules that have been affected.
    def get_affected_rules
      @rules.count
    end

    private

    # Assign rules to tactics and techniques based on the rule data.
    def assign
      @rules.each do |rule|
        update_tactic(rule)
        update_technique(rule)
      end
    end

    # Update the tactics with the rule information.
    def update_tactic(rule)
      # Find tactics that match the rule's tactic ID and are flagged as tactics.
      tactics = Mitre.where(tactic_id: rule.mitre_tactic_id, is_tactic: true)

      tactics.each do |tactic|
        # Ensure the rule SID is included in the tactic's list of rule IDs.
        tactic.rule_ids ||= []
        tactic.rule_ids |= [rule.rule_sid]
        tactic.save!
      end
    end

    # Update the techniques with the rule information.
    def update_technique(rule)
      # Find techniques that match the rule's technique ID and are flagged as techniques.
      techniques = Mitre.where(technique_id: rule.mitre_technique_id, is_technique: true)

      techniques.each do |technique|
        # Find the corresponding tactics for the technique.
        tactics = Mitre.where(is_tactic: true, tactic_id: technique.tactic_id)
        tactics.each do |tactic|
          # Add the rule SID to the tactic's list of rule IDs if it's not already included.
          unless tactic.rule_ids.include?(rule.rule_sid)
            tactic.rule_ids << rule.rule_sid
            tactic.save!
          end
        end

        # Ensure the rule SID is included in the technique's list of rule IDs.
        technique.rule_ids ||= []
        technique.rule_ids |= [rule.rule_sid]
        technique.save!
      end
    end
  end
end