#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#!/opt/rb/var/www/rb-rails/script/rails runner
#
# Copyright (c) 2011 redBorder Networks
# Authors:
#   Pablo Nebrera Herrera  pablonebrera@eneotecnologia.com
#   Juan Jesus Prieto:     jjprieto@eneotecnologia.com
#   Jose Antonio Parra:    japarra@eneotecnologia.com
#
#  Description:
# It reads the configuration from config/rules2rbIPS.yml. Format:
#   verbose: false
#   cache_dir: "/var/www/snortrules/cache"
#
###########################################################################################################

require 'yaml'
require 'net/http'
require 'tempfile'
require 'digest/md5'

module Redborder

  class RulesUpdate

    # If 'path_rules_dir' is set, it will be executed reading the rules from the directory

    def self.start(verbose=false, initial_tgz=nil, source_id=nil)

      @content = ""

      now = Time.now

      sources = source_id.nil? ? RuleSource.where(is_custom: false, enable: true) : RuleSource.where(id: source_id)

      @config = YAML.load_file("#{Rails.root}/config/rules2rbIPS.yml")

      logit "#{I18n.t 'lib_redborder.node'} " + `hostname`.chomp!

      logit ""

      logit "#{I18n.t 'lib_redborder.starting_rules'}"

      # Download rules from each source
      sources.each do |source|

        begin

          new_rules      = false    # if true we have new rules and we will create a new RuleDbVersion.
          download_files = false
          @warn_version   = false

          logit ""
          logit "############################################################################"
          logit "       #{I18n.t 'lib_redborder.rule_source'}: #{source.clean_name}"
          logit "############################################################################"

          if source.has_code and source.code.blank?
            # Print no code for source
            logit "INFO: #{I18n.t 'lib_redborder.the_source'} #{source.name} #{I18n.t 'lib_redborder.needs'}."
            next
          end

          if source.is_custom
            tgz_tmp_file = "#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}"
            logit "#{I18n.t 'lib_redborder.downloading'} #{source.cached_tgz_path}" if verbose
            logit "  * #{I18n.t 'lib_redborder.saving'} #{tgz_tmp_file}" if verbose

            if StorageFile.s3_enabled?
              download_file(source.cached_tgz_uri.to_s, tgz_tmp_file)
            else
              FileUtils.copy_file(source.cached_tgz_uri.to_s, tgz_tmp_file)
            end

            if source.md5 != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
              new_rules = true
            else
              logit "  * #{I18n.t 'lib_redborder.the_upload'}." if verbose
            end
          else
            if source.cached_tgz.nil? # if no cached_tgz is stored, force download files
              download_files = true
            else
              md5_file = "#{@config["cache_dir"]}/#{source.clean_name}.md5"
              File.delete(md5_file) if File.exists?(md5_file)
              if source.md5url.present? # check downloaded md5sum with md5sum of current rule_db_version
                logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_md5url}" if verbose
                logit "  * #{I18n.t 'lib_redborder.saving'} #{md5_file}" if verbose

                download_file(source.full_md5url, md5_file)
                md5_file_content = File.open(md5_file) {|f| f.readline}.gsub(/^"/, '').chomp

                if md5_file_content != source.md5 # if md5sum is different, force download files
                  logit "     - Cached tgz MD5 -> #{source.md5}" if verbose
                  logit "     - #{source.full_tgzurl} MD5 -> #{md5_file_content}" if verbose
                  download_files = true
                else
                  logit "  * #{I18n.t 'lib_redborder.cached_file'}" if verbose
                end
              else
                download_files = true
              end
            end

            if download_files
              tgz_tmp_file = "#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}"

              logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_tgzurl}" if verbose
              logit "  * #{I18n.t 'lib_redborder.saving'} #{tgz_tmp_file}" if verbose

              download_file(source.full_tgzurl, tgz_tmp_file)

              #check if the file is correct
              if defined?(md5_file_content) and md5_file_content and md5_file_content != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
                logit "  * #{I18n.t 'lib_redborder.not_match'}" if verbose
                next # go to next source
              elsif defined?(md5_file_content) and md5_file_content
                new_rules = true
              elsif source.cached_tgz.nil? # source has no md5url and no cached_tgz
                new_rules = true
              elsif source.md5 != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
                new_rules = true
              end
            end
          end

          next if !new_rules

          FileUtils.mv(tgz_tmp_file, "#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}")

          last_rule_db_version = source.rule_db_versions.where(completed: true).last
          rule_db_version = RuleDbVersion.create(rule_source: source, completed: false)
          rule_db_version.update(checked: true) if (last_rule_db_version.nil? or source.is_custom)

          parse_rules(source.id, rule_db_version.id, verbose)

          Object::Rule.transaction do |t|
            begin

              if last_rule_db_version.present?

                last_rule_db_version.rules.where(default_enabled: [true, false]).find_each do |r|
                  r.update(is_deprecated: true) if Object::Rule.where(rule_sid: r.rule_sid, rule_gid: r.rule_gid, rule_db_version: rule_db_version, default_enabled: [true, false]).last.nil?
                end

                rule_db_version.rules_deprecated_count = last_rule_db_version.rules.where(is_deprecated: true, default_enabled: [true, false]).count

              end

              source.cached_tgz = File.open("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}") unless source.is_custom
              source.md5 = Digest::MD5.hexdigest(File.read("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}"))
              source.save

              rule_db_version.completed              = true
              rule_db_version.rules_count            = rule_db_version.rules.where(default_enabled: [true, false]).count
              rule_db_version.rules_new_count        = rule_db_version.rules.where(is_new: true, default_enabled: [true, false]).count
              rule_db_version.rules_modified_count   = rule_db_version.rules.where(is_modified: true, default_enabled: [true, false]).count

              rule_db_version.save

            rescue Exception => e
              logit "#{I18n.t 'lib_redborder.an_error'} #{e.to_s}. #{I18n.t 'lib_redborder.rolling'}"
              raise ActiveRecord::Rollback
              return
            end
          end

        rescue Exception => e
          logit "#{I18n.t 'lib_redborder.error_source'} #{source.name} (#{e})"
          e.backtrace.each do |b|
            logit b
          end

        ensure
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}.md5")
          FileUtils.rm_rf("#{@config["cache_dir"]}/#{source.clean_name}")
        end

      end

      logit "\n#{I18n.t 'lib_redborder.warning'} #{Sensor::SNORT_VERSIONS.join(",")}\n" if @warn_version

      logit "#{I18n.t 'lib_redborder.exiting'}" if verbose

      return @content

    rescue Exception => e
      logit "#{I18n.t 'lib_redborder.error_updating'} (ERROR: #{e})"
      return @content
    end

    def self.find(dir, filename="*.*", subdirs=true)
      Dir[ subdirs ? File.join(dir.split(/\\/), "**", filename) : File.join(dir.split(/\\/), filename) ]
    end

    def self.logit(msg)
      STDOUT.puts "#{msg}"
      @content += msg + "\n"
    end

    def self.download_file(url, local_file)

      File.delete(local_file) if File.exists?(local_file)

      url = ("http://" + url) if !url.start_with?"http://" and !url.start_with?"https://" and !url.start_with?"ftp://"

      writeOut = open(local_file, "wb")

      # If the url contains s3, the request musn't go over proxy
      if Setting.proxy? and Setting.proxy_address.present? and !url.start_with?"https://redborder.s3"
         proxy_port = Setting.proxy_port.to_i
         proxy_port = 8080 if proxy_port <= 0
         proxy_addr = Setting.proxy_address
         proxy_addr = ("http://" + proxy_addr) if !proxy_addr.start_with?"http://" and !proxy_addr.start_with?"https://"
         proxy = URI.parse(proxy_addr +":" +proxy_port.to_s)

         if Setting.proxy_user.present?
           writeOut.write(open(url, proxy_http_basic_authentication: [proxy, Setting.proxy_user, Setting.proxy_password], ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE, read_timeout: 10).read)
         else
           writeOut.write(open(url, proxy: proxy, ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE, read_timeout: 10).read)
         end

      else
         writeOut.write(open(url, proxy: false, ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE, read_timeout: 10).read)
      end

      writeOut.close

    end

    def self.download_cached_tgz(uri, cached_tgz_file)
      http = Net::HTTP.new(uri.host, uri.port)
      http.use_ssl = true
      # TODO: prevent use VERIFY_NONE
      http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      resp = http.get(uri.request_uri)

      open(cached_tgz_file, "wb") do |file|
        file.write(resp.body)
      end
    end

    def self.parse_rules(source_id, rule_db_version_id, verbose)
      source = RuleSource.find(source_id)
      last_rule_db_version = source.rule_db_versions.where(completed: true).last

      # it is possible we have deleted or inserted new sources

      logit "#{I18n.t 'lib_redborder.new_rules'}" if verbose

      tgzLocalFile = "#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}"

      logit "#{I18n.t 'lib_redborder.extracting'} #{tgzLocalFile}" if verbose
      dir = "#{@config['cache_dir']}/#{source.clean_name}"

      FileUtils.rm_rf(dir)
      FileUtils.mkdir_p dir

      case source.compression_type
      when 'zip'
        system('unzip', tgzLocalFile, '-d', dir)
      when 'tgz'
        system('tar', 'xzf', tgzLocalFile, '-C', dir)
      end

      rule_db_version = RuleDbVersion.find(rule_db_version_id)

      so_rules_dir  = "#{@config['so_rules_dir']}/dbversion-#{rule_db_version.id}"

      FileUtils.mkpath so_rules_dir

      # We process first the so_rules
      logit "#{I18n.t 'lib_redborder.processing'} #{source.name} so_rules" if verbose

      so_rules = (source.so_rules || ['so_rules/precompiled/RHEL-6-0/x86-64'])

      so_rules.each do |rulesDir|
        logit "  * #{I18n.t 'lib_redborder.directory'}: #{rulesDir}" if verbose
        if !Dir.exist?("#{dir}/#{rulesDir}")
          logit "    - #{I18n.t 'lib_redborder.directory'} #{dir}/#{rulesDir} #{I18n.t 'lib_redborder.ignored'}" if verbose
        else

          Dir.open("#{dir}/#{rulesDir}") do |rulesDirectory|
            count = 0
            total = rulesDirectory.to_a.delete_if { |x| ['..', '.'].include? x }.size

            rulesDirectory.to_a.delete_if { |x| ['..', '.'].include? x }.each do |file|
              count += 1
              next if file == '.' || file == '..'

              @warn_version = true unless Sensor::SNORT_VERSIONS.include?(file)

              logit "       - Reading #{file} ( #{(count*100.0/total).round(2)}% )" if verbose
              `cd #{dir}/#{rulesDir}/#{file} && if [ -f #{so_rules_dir}/snort-so_rules-#{file}.tar ]; then tar rf #{so_rules_dir}/snort-so_rules-#{file}.tar .; else tar cf #{so_rules_dir}/snort-so_rules-#{file}.tar .; fi`
            end
          end

        end
      end

      fake_so_rules = "#{so_rules_dir}/snort-so_rules.tar"

      system("rm -f /tmp/.fake; touch /tmp/.fake; tar cf #{fake_so_rules} -C /tmp .fake &>/dev/null; rm -f /tmp/.fake")
      system("gzip -q #{fake_so_rules};")

      # Compressing the so_rules tar files
      Dir.open(so_rules_dir) do |so_rulesCompressed|
        so_rulesCompressed.to_a.each do |file|
          next if file == '.' || file == '..'

          next unless file.ends_with? 'tar'

          `gzip -q #{so_rules_dir}/#{file};`

          so_rule_match = /^snort-so_rules-(?<version>[\d\.]+).tar$/.match(file)

          next if so_rule_match.nil?

          so_versions = so_rule_match[:version].split('.')

          if so_versions.length > 2
            `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0].to_s}.#{so_versions[1].to_s}.tar.gz ] &&
               ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0].to_s}.#{so_versions[1].to_s}.tar.gz`
          end

          next unless so_versions.size > 3 && so_versions[3].to_i > 0

          `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.')}.tar.gz ] &&
             ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.')}.tar.gz`
          (0..so_versions[3].to_i - 1).each do |i|
            `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.') + '.' + i.to_s}.tar.gz ] &&
               ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.') + '.' + i.to_s}.tar.gz`
          end
        end
      end

      # Compress so_rules to a uniq tgz and upload it
      system("cd #{so_rules_dir}; tar czf #{@config["cache_dir"]}/#{source.clean_name}/so_rules.tgz *")
      rule_db_version.update(so_rules: File.open("#{@config["cache_dir"]}/#{source.clean_name}/so_rules.tgz"))
      FileUtils.rm_rf(so_rules_dir)

      # We process the unicode.map files
      logit "#{I18n.t 'lib_redborder.processing'} #{source.name} unicode.map" if verbose
      unicode_maps = find(dir, 'unicode.map')
      unicode_maps.each do |file|
        File.open(file, 'r') do |f|
          rule_db_version.update(unicode_map: f.read)
        end
      end

      # Processing classification.config file.
      logit "#{I18n.t 'lib_redborder.processing'} #{source.name} classification.config" if verbose

      begin
        csFiles = find("#{@config["cache_dir"]}/#{source.clean_name}", 'classification.config')

        csFiles.each do |f|
          logit "  * #{I18n.t 'lib_redborder.found'} #{f}" if verbose
          File.open(f) do |classificationFile|
            while classtypeLine = classificationFile.gets

              next if classtypeLine.starts_with?('#') || classtypeLine.starts_with?('\n')

              classtypeMatch = /^config classification:\s*(?<name>[^,]+)\s*,\s*(?<description>[^,]+)\s*,\s*(?<severity_id>[\d]+)$/.match(classtypeLine)

              unless classtypeMatch.nil?
                rc2 = ClassType.where(name: classtypeMatch[:name], rule_db_version_id: rule_db_version.id).first_or_create
                rc2.update(description: classtypeMatch[:description], severity_id: classtypeMatch[:severity_id])
              end

            end
          end
        end

      rescue => e
        logit "  *  #{I18n.t 'lib_redborder.classification'}" if verbose
      end

      logit "#{I18n.t 'lib_redborder.processing'} #{source.name} #{I18n.t 'lib_redborder.rules'}" if verbose

      source.search.each do |rulesDir|
        logit "  * #{I18n.t 'lib_redborder.directory'}: #{rulesDir}" if verbose

        dir = "#{@config["cache_dir"]}/#{source.clean_name}/#{rulesDir}"

        if !Dir.exist?(dir)
          logit "    - #{I18n.t 'lib_redborder.directory'} #{dir} #{I18n.t 'lib_redborder.ignored'}" if verbose
        else

          Object::Rule.transaction do |t|
            begin

              Dir.open(dir) do |rulesDirectory|

                pathRulesDir   = rulesDirectory.to_path
                count          = 0
                total          = rulesDirectory.to_a.size

                rulesDirectory.to_a.each do |file|
                  count += 1
                  next if file == '.' || file == '..'

                  if file =~ /\.rules$/
                    parse_rules_file(file, pathRulesDir, source, rule_db_version, last_rule_db_version, total, count, verbose)
                  elsif file == 'file_magic.conf'
                    parse_file_capture_rules_file(file, pathRulesDir, source, rule_db_version, last_rule_db_version, total, count, verbose)
                  end
                end
              end

            rescue StandardError => e

              logit "#{I18n.t 'lib_redborder.an_error'} #{e.to_s}. #{I18n.t 'lib_redborder.rolling'}"
              raise ActiveRecord::Rollback
              return

            end
          end

        end
      end
    end

    def self.parse_rules_file(file, pathRulesDir, source, rule_db_version, last_rule_db_version, total, count, verbose)
      logit "       - #{I18n.t 'lib_redborder.reading'} #{file} ( #{(count*100.0/total).round(2)}% )" if verbose
      ruleFile  = pathRulesDir + '/' + file
      category1 = file.gsub('.rules','').strip.downcase
      category1 = category1.gsub(/#{source.regexp_group_ignore}/, '') unless source.regexp_group_ignore.nil?

      category1_tmp = category1

      unless @config['category1_groups'].nil?
        @config['category1_groups'].each do |catgr|
          unless /#{catgr["regexp"]}/.match(category1).nil?
            category1_tmp = catgr['name']
            break
          end
        end
      end

      if category1_tmp != "deleted"

        rc1 = Group.where(name: category1_tmp).first_or_create

        category4 = "others"
        category4_desc = nil

        unless @config["category4_groups"].nil?
          @config["category4_groups"].each do |catgr|
            unless /#{catgr["regexp"]}/.match(category1_tmp).nil?
              category4      = catgr["name"]
              category4_desc = catgr["desc"]
              break
            end
          end
        end

        rc4 = Category.where(name: category4).first_or_create
        rc4.update(description: category4_desc) unless category4_desc.nil?

        File.open(ruleFile) do |ruleFile|

          while ruleLine = ruleFile.gets

            ruleLine.chop!
            ruleMatch = /^(?<disabled>[#]\s*)?(?<action>\w+)\s+(?<l3l4>(?<protocol>\w+)\s+(?<source_addr>[^\s]+)\s+(?<source_port>[^\s]+)\s+(?<separator>->|<>)\s+(?<target_addr>[^\s]+)\s+(?<target_port>[^\s]+)\s+)?\((?<rule>.+)\)\s*$/.match(ruleLine)

            unless ruleMatch.nil?

              sidRuleMatch = /;\s*sid:\s*(?<sid>\d+);/.match(ruleMatch[:rule])

              unless sidRuleMatch.nil?

                msgRuleMatch = /msg:\s*"(?<msg>[^"]+)";/.match(ruleMatch[:rule])

                unless msgRuleMatch.nil?

                  classtypeRuleMatch = /; classtype:\s*(?<classtype>[^;]+);/.match(ruleMatch[:rule])

                  if classtypeRuleMatch.nil?
                    category2 = "unknown"
                  else
                    category2 = classtypeRuleMatch[:classtype].strip.downcase
                    category2 = category2.gsub /#{source.regexp_classtype_ignore}/, "" unless source.regexp_classtype_ignore.nil?
                  end

                  rc2 = ClassType.where(name: category2, rule_db_version_id: rule_db_version.id).first_or_create

                  rule_msg = msgRuleMatch[:msg].downcase
                  rule_msg = rule_msg.gsub /#{source.regexp_family_ignore}/, "" unless source.regexp_family_ignore.nil?
                  rule_msg = rule_msg.gsub /^#{category1} /, ""

                  category3 = nil

                  unless @config["category3_groups"].nil?
                    @config["category3_groups"].each do |catgr|
                      unless /#{catgr["regexp"]}/.match(rule_msg).nil?
                        category3 = catgr["name"]
                      end
                    end
                  end

                  if category3.nil?
                    category3RuleMatch = /([^ _-]*([ _-][^ _-]*)?([ _-][^ _-]*)?)/.match(rule_msg)
                    if category3RuleMatch.nil?
                      category3 = "unknown"
                    else
                      category3 = category3RuleMatch[0][0..63].strip
                    end
                  end

                  rc3 = Family.where(name: category3).first_or_create

                  revRuleMatch = /; rev:\s*(?<value>\d+);/.match(ruleMatch[:rule])
                  rev = revRuleMatch.nil? ? 0 : revRuleMatch[:value].to_i

                  gidRuleMatch = /; gid:\s*(?<value>\d+);/.match(ruleMatch[:rule])
                  gid = gidRuleMatch.nil? ? 1 : gidRuleMatch[:value].to_i

                  mitreTacticIdMatch = /mitre_tactic_id\s(?<value>TA\d+)/.match(ruleMatch[:rule])
                  mitre_tactic_id = mitreTacticIdMatch.nil? ? "N/A" : mitreTacticIdMatch[:value]
                                    
                  mitreTechniqueIdMatch = /mitre_technique_id\s(?<value>T\d+)/.match(ruleMatch[:rule])
                  mitre_technique_id = mitreTechniqueIdMatch.nil? ? "N/A" : mitreTechniqueIdMatch[:value]

                  r = Object::Rule.new

                  r.rule_sid            = sidRuleMatch[:sid].to_i
                  r.rule_gid            = gid.to_i
                  r.rule_rev            = rev.to_i
                  r.msg                 = msgRuleMatch[:msg]
                  r.short_msg           = msgRuleMatch[:msg].gsub( /^#{rc1.name.upcase} /, '' )
                  r.group_id            = rc1.id
                  r.class_type_id       = rc2.id
                  r.family_id           = rc3.id
                  r.category_id         = rc4.id
                  r.protocol            = ruleMatch[:protocol]
                  r.source_addr         = ruleMatch[:source_addr]
                  r.source_port         = ruleMatch[:source_port]
                  r.target_addr         = ruleMatch[:target_addr]
                  r.target_port         = ruleMatch[:target_port]
                  r.rule_db_version_id  = rule_db_version.id
                  r.text                = ruleMatch[:rule]
                  r.default_enabled     = ruleMatch[:disabled].nil?
                  r.mitre_tactic_id     = mitre_tactic_id
                  r.mitre_technique_id  = mitre_technique_id

                  #We need to discover if this rule is new or modified.
                  r.is_deprecated = false

                  if last_rule_db_version.nil?
                    r.is_new      = true
                    r.is_modified = false
                  else
                    rlast = Object::Rule.where(rule_sid: r.rule_sid, rule_gid: r.rule_gid, rule_db_version: last_rule_db_version, default_enabled: [true, false]).last
                    if rlast.nil?
                      r.is_new      = true
                      r.is_modified = false
                    else
                      r.is_new      = false
                      if r.rule_rev != rlast.rule_rev
                        r.is_modified = true
                      else
                        r.is_modified = false
                      end
                    end
                  end

                  thresholdMatch = /; threshold:(?<threshold>[^;]+);/.match(r.text)
                  if thresholdMatch
                    r.has_threshold = true
                    r.threshold_options = thresholdMatch[:threshold].lstrip
                    r.text = r.text.gsub(/\s*threshold:[^;]+;/, "")
                  end

                  if r.save

                    #METADATA
                    metadataRuleMatch = /; metadata:\s*(?<value>[^;]+);/.match(ruleMatch[:rule])

                    unless metadataRuleMatch.nil?
                      metadataRuleMatch[:value].split(/,\s*/).each do |x|
                        policyRuleMatch = /^policy\s+(?<name>[^ ]+)\s+(?<action>[^ ]+)/.match(x)
                        if policyRuleMatch.nil?
                          serviceRuleMatch = /^service\s+(?<name>[^;]+)\s*/.match(x)
                          if serviceRuleMatch.nil?
                            impactRuleMatch = /^impact_flag\s+(?<name>[^;]+)\s*/.match(x)
                            unless impactRuleMatch.nil?
                              impact = Impact.where(name: impactRuleMatch[:name]).first_or_create
                              #RuleImpact.create(rule_id: r.id, impact_id: impact.id)
                              impact.rules << r
                            end
                          else
                            service = Service.where(name: serviceRuleMatch[:name]).first_or_create
                            #RuleService.create(rule_id: r.id, service_id: service.id)
                            service.rules << r
                          end
                        else
                          policy = FactoryPolicy.where(name: policyRuleMatch[:name]).first_or_create
                          action = Action.where(name: policyRuleMatch[:action]).first
                          FactoryPolicyRule.create(rule_id: r.id, factory_policy_id: policy.id, action_id: action.id) unless action.nil?
                        end
                      end
                    end

                    #REFERENCE
                    referenceSplit = ruleMatch[:rule].split(/;\s*reference:\s*/)

                    referenceSplit.each_with_index do |x, i|
                      if i > 0
                        referenceRuleMatch = /^(?<name>[^,]+)\s*,\s*(?<url>[^;]+)\s*/.match(x)
                        unless referenceRuleMatch.nil?
                          reference = RbReference.where(name: referenceRuleMatch[:name]).first_or_create(humanize: referenceRuleMatch[:name])
                          RbReferenceRule.create(rule_id: r.id, rb_reference_id: reference.id, url: referenceRuleMatch[:url])
                        end
                      end
                    end

                    #FLOWBITS
                    flowbitSplit = ruleMatch[:rule].split(/;\s*flowbits:\s*/)

                    flowbitSplit.each_with_index do |x, i|

                      if i > 0
                        flowbitRuleMatch = /^(?<action>[^,; ]+)\s*(,\s*(?<state>[^,; ]+)\s*(,\s*(?<group>[,; ]))?)?\s*/.match(x)
                        unless flowbitRuleMatch.nil?
                          flowbit = Flowbit.where(name: flowbitRuleMatch[:action]).first_or_create
                          flowbit_state = nil
                          flowbit_group = nil
                          unless flowbitRuleMatch[:state].nil?
                            flowbit_state = FlowbitState.where(name: flowbitRuleMatch[:state]).first_or_create

                            unless flowbitRuleMatch[:group].nil?
                              flowbit_group = FlowbitGroup.where(name: flowbitRuleMatch[:group]).first_or_create
                            end
                          end
                          FlowbitRule.create(rule_id: r.id, flowbit_id: flowbit.id, flowbit_state: flowbit_state, flowbit_group: flowbit_group)
                        end
                      end

                    end

                  else
                    logit "#{I18n.t 'lib_redborder.rule_not_saved'} #{ruleLine}; #{I18n.t 'lib_redborder.errors'} #{r.errors.inspect}" if verbose
                  end

                else
                  logit "MSG #{I18n.t 'lib_redborder.not_found'} -> #{ruleLine}" if verbose
                end

              else
                logit "SID #{I18n.t 'lib_redborder.not_found'} -> #{ruleLine}" if verbose
              end

            else
              #logit "rule line doesn't match -> ruleLine" if verbose
            end

          end

        end

      end
    end

    def self.parse_file_capture_rules_file(file, pathRulesDir, source, rule_db_version, last_rule_db_version, total, count, verbose)
      logit "       - #{I18n.t 'lib_redborder.reading'} #{file} ( #{(count*100.0/total).round(2)}% )" if verbose
      ruleFile = pathRulesDir + '/' + file

      category = Category.where(name: 'file_capture').first_or_create(description: "#{I18n.t 'lib_redborder.capture_rules'}")
      class_type = ClassType.where(name: 'file_capture', rule_db_version_id: rule_db_version.id, severity_id: 1).first_or_create

      File.open(ruleFile) do |f|
        f.each do |rule_text|
          rule_match = rule_text.match /file type:(?<family>\w+); id:(?<id>\d+); category:(?<group>.+); msg:"(?<msg>.+)"; rev:(?<rev>\d+);/

          next if rule_match.nil?

          begin
            group = Group.where(name: rule_match[:group]).first_or_create
            family = Family.where(name: rule_match[:family]).first_or_create

            rule = Object::Rule.create(
              rule_sid: rule_match[:id],
              rule_gid: 146,
              rule_rev: rule_match[:rev],
              msg: rule_match[:msg],
              short_msg: rule_match[:msg],
              text: rule_text,
              default_enabled: true,
              is_new: true,
              is_deprecated: false,
              is_modified: false,
              rule_db_version_id: rule_db_version.id,
              category_id: category.id,
              group_id: group.id,
              family_id: family.id,
              class_type_id: class_type.id
            )

            # We need to discover if this rule is new or modified.
            rule.is_deprecated = false

            if last_rule_db_version.nil?
              rule.is_new      = true
              rule.is_modified = false
            else
              rlast = Object::Rule.where(rule_sid: rule.rule_sid, rule_gid: rule.rule_gid, rule_db_version: last_rule_db_version, default_enabled: [true, false]).last
              if rlast.nil?
                rule.is_new      = true
                rule.is_modified = false
              else
                rule.is_new      = false
                if rule.rule_rev != rlast.rule_rev
                  rule.is_modified = true
                else
                  rule.is_modified = false
                end
              end
            end
          rescue NoMethodError
          end
        end
      end
    end
  end
end
