#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
module Redborder
  module ScanAps
    # WLC OIDs
    NAME = '1.3.6.1.4.1.9.9.513.1.1.1.1.5'
    MAC = '1.3.6.1.4.1.9.9.513.1.2.3.1.1'
    CLIENTS = '1.3.6.1.4.1.9.9.513.1.1.1.1.54'

    # MERAKI OIDs
    MERAKI = '1.3.6.1.4.1.29671.1.1.4.1'
    MERAKI_MAC = '1.3.6.1.4.1.29671.1.1.4.1.1.'
    MERAKI_NAME = '1.3.6.1.4.1.29671.1.1.4.1.2.'

    #
    # Scan Sensor via SNMP
    # == Parameters:
    # sensor::
    #    Sensor to scan
    #
    def self.scan_snmp(sensor)
      scan_wlc(sensor) if sensor.is_flow?
      scan_meraki(sensor) if sensor.is_meraki?
    end

    private 

    #
    # Update AP or create it if it doesn't exist
    #   Params:
    #     ap_params: access point's attributes
    #
    def self.update_ap(ap_params)
      ap = AccessPoint.find_by(mac_address: ap_params[:mac_address])
      if ap.nil?
        AccessPoint.create(ap_params)
      else
        ap.update(ap_params)
      end
    end

    #
    # Scan Sensor via SNMP
    # == Parameters:
    # flow::
    #    Sensor FLOW to scan
    #
    def self.scan_wlc(flow)
      snmp_version = :SNMPv2c
      snmp_community = 
        flow.snmp_community.present? ? flow.snmp_community : 'public'
      begin
        EM.run do
          snmp =
            SNMP4EM::Manager.new(
              host: flow.ip,
              community: snmp_community,
              version: snmp_version
            )
        request =
          snmp.bulkwalk([NAME, MAC, CLIENTS])
          request.errback do
            EM.stop
          end
          request.callback do |response|
            flow_response(flow, response)
            EM.stop
          end
        end
      rescue SNMP::RequestTimeout
        return
      end
    end

    #
    # Scan Sensor via SNMP
    # == Parameters:
    # meraki::
    #    Sensor MERAKI to scan
    #
    def self.scan_meraki(meraki)
      snmp_version = :SNMPv2c
      snmp_community = 
        meraki.snmp_community.present? ? meraki.snmp_community : 'public'
      begin
        EM.run do
          snmp =
            SNMP4EM::Manager.new(
              host: meraki.ip,
              port: 161_00,
              community: snmp_community,
              version: snmp_version
            )
        request =
          snmp.bulkwalk([MERAKI])
          request.errback do
            EM.stop
          end
          request.callback do |response|
            meraki_response(meraki, response)
            EM.stop
          end
        end
      rescue SNMP::RequestTimeout
        return
      end
    end

    #
    # Get value for SNMP OID + Object
    #   Params:
    #     response: SNMP response
    #     oid: general OID
    #     obj: specific object's OID
    #   Returns:
    #     value
    #
    def self.obj_value(response, oid, obj)
      result = response[oid].detect { |k, _v| k.start_with?(oid + obj) }
      result.nil? ? nil : result.last
    end

    def self.meraki_obj_value(response, oid, obj)
      result = response[MERAKI].detect { |k, _v| k.start_with?(oid + obj) }
      result.nil? ? nil : result.last
    end

    #
    # Handle Meraki SNMP response to BulkWalk request
    #   Params:
    #     meraki: Meraki Sensor
    #     response: SNMP response
    #
    def self.meraki_response(meraki, response)
      objects =
        response[MERAKI].keys.select { |k| k.start_with?(MERAKI_NAME) }
        .map { |k| k.gsub(MERAKI_NAME, '') }
      objects.each do |obj|
        ap_params = {
          mac_address: meraki_obj_value(response, MERAKI_MAC, obj)
                       .unpack('H2H2H2H2H2H2').join(':'),
          name: meraki_obj_value(response, MERAKI_NAME, obj),
          switch_id: meraki.id
        }
        update_ap(ap_params)
      end
    end

    #
    # Handle Flow SNMP response to BulkWalk request
    #   Params:
    #     flow: Flow Sensor
    #     response: SNMP response
    #
    def self.flow_response(flow, response)
      objects = response[NAME].keys.map { |k| k.gsub(NAME, '') }
      objects.each do |obj|
        ap_params = {
          name: obj_value(response, NAME, obj),
          mac_address: obj_value(response, MAC, obj)
                       .unpack('H2H2H2H2H2H2').join(':'),
          switch_id: flow.id,
          switch_name: flow.ip,
        }
        update_ap(ap_params)
      end
    end

  end
end