#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Miguel Ávarez <malvarez@redborder.com>
#######################################################################

#
# Base Redborder module
#
module Redborder
  #
  # Validate TI data in order to inject in memcached
  #
  module TIDataValidator
    #
    # A hash mapping data types to their corresponding validation methods.
    # This allows for dynamic selection of validation logic based on type.
    #

    VALIDATORS = {
      "ip" => :match_ip,
      "domain" => :match_domain,
      "sha_1" => :match_sha1,
      "sha_2" => :match_sha2,
      "url" => :match_url
    }.freeze

    #
    # Validate an IP address (IPv4 or IPv6).
    #
    # == Parameters:
    # ip::
    #   A string representing the IP address to validate.
    #
    # == Returns:
    # A string of the matched IP address if valid, otherwise nil.
    #
    def self.match_ip(ip)
      ipv4_regex = /\A((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\z/
      ipv6_regex = /\A([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}\z/
      match = ip.match(ipv4_regex) || ip.match(ipv6_regex)
      match ? match.to_s : nil
    end

    #
    # Validate a domain name.
    #
    # == Parameters:
    # domain::
    #   A string representing the domain name to validate.
    #
    # == Returns:
    # A string of the matched domain if valid, otherwise nil.
    #
    def self.match_domain(domain)
      domain_regex = /\A[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(\.[a-zA-Z]{2,})+\z/
      match = domain.match(domain_regex)
      match ? match.to_s : nil
    end

    #
    # Validate a SHA-1 hash.
    #
    # == Parameters:
    # hash::
    #   A string representing the SHA-1 hash to validate.
    #
    # == Returns:
    # A string of the matched SHA-1 hash if valid, otherwise nil.
    #
    def self.match_sha1(hash)
      sha1_regex = /\A[a-fA-F0-9]{40}\z/
      match = hash.match(sha1_regex)
      match ? match.to_s : nil
    end

    #
    # Validate a SHA-2 hash.
    #
    # == Parameters:
    # hash::
    #   A string representing the SHA-2 hash to validate.
    #
    # == Returns:
    # A string of the matched SHA-2 hash if valid, otherwise nil.
    #
    def self.match_sha2(hash)
      sha2_regex = /\A[a-fA-F0-9]{64}\z/
      match = hash.match(sha2_regex)
      match ? match.to_s : nil
    end

    #
    # Validate a URL.
    #
    # == Parameters:
    # url::
    #   A string representing the URL to validate.
    #
    # == Returns:
    # A string of the matched URL if valid, otherwise nil.
    #
    def self.match_url(url)
      url_regex = URI.regexp
      match = url.match(url_regex)
      match ? match.to_s : nil
    end
  end
  #
  # Module to download TI data from remote FTP server and inject
  # into memcached for later enrichment in logstash pipelines
  #
  module ThreatIntel
    class Client < Redborder::TIConnections::Client
      #
      # Magic numbers
      #
      CACHE_EXPIRATION_TIME = 3600
      CACHE_DEFAULT_VALUE = 1

      #
      # Initialize the ThreatIntel client with FTP/HTTP connection and Dalli caching client.
      #
      # == Parameters:
      # protocol::
      #   A string representing the protocol to use for FTP/HTTP connection (http, https, ftp, sftp, or ftps).
      # host::
      #   A string representing the host of the FTP server.
      # user::
      #   A string representing the username for authentication.
      # password::
      #   A string representing the password for authentication.
      # dalli_options::
      #   A hash containing the Dalli client options. Must include :server and :options.
      #
      # == Returns:
      # A new instance of Redborder::ThreatIntel::Client.
      #
      def initialize(protocol:, host:, port:, user:, password:, dalli_options: {})
        super(protocol: protocol, host: host, port: port, user: user, password: password)
        @dalli_client = Dalli::Client.new(dalli_options[:servers], dalli_options[:options])
      end

      #
      # Download a file from the remote FTP server and inject its content into a Dalli cache.
      #
      # == Parameters:
      # remote_path::
      #   A string representing the remote file path to download.
      # local_path::
      #   A string representing the local file path to store the downloaded file.
      #
      # == Returns:
      # None.
      #
      # == Raises:
      # Net::FTPError::
      #   If there is an FTP error during the download process.
      # IOError::
      #   If there is an I/O error during the download process.
      # ArgumentError::
      #   If there is an argument error during the download and injection process.
      #
      def download_and_inject(remote_path:, local_path:, type:)
        validate_type(type)
        download_file(remote_path, local_path)
        inject(local_path, type)
      rescue Net::FTPError => e
        Rails.logger.error "FTP error in download_and_inject: #{e.message}"
        raise
      rescue IOError => e
        Rails.logger.error "I/O error in download_and_inject: #{e.message}"
        raise
      rescue ArgumentError => e
        Rails.logger.error "Argument error in download_and_inject: #{e.message}"
        raise
      end

      private

      #
      # Validate that the provided type is valid.
      #
      # == Parameters:
      # type::
      #   A string or symbol representing the type to validate.
      #
      # == Raises:
      # ArgumentError::
      #   If the type is invalid.
      #
      def validate_type(type)
        unless Redborder::TIDataValidator::VALIDATORS.key?(type)
          Rails.logger.error "Invalid type provided: #{type}"
          raise ArgumentError, "Unsupported type: #{type}"
        end
      end

      #
      # Process the file line by line and inject each line as a key into the Dalli cache.
      #
      # == Parameters:
      # file_path::
      #   A string representing the path to the local file to inject into Dalli cache.
      #
      # == Returns:
      # None.
      #
      # == Raises:
      # Dalli::DalliError::
      #   If there is an error while setting keys into Dalli cache.
      # IOError::
      #   If there is an I/O error during the file processing.
      # ArgumentError::
      #   If there is an argument error during the injection process.
      #
      def inject(file_path, type)
        counts = { valid: 0, invalid: 0, skipped: 0 }
        
        validator_method = Redborder::TIDataValidator::VALIDATORS[type]
      
        File.foreach(file_path) do |line|
          begin
            key = line.strip

            if key.blank?
              counts[:skipped] += 1
              next
            end
      
            valid_key = Redborder::TIDataValidator.public_send(validator_method, key)
      
            if valid_key
              counts[:valid] += 1
              @dalli_client.set(valid_key, CACHE_DEFAULT_VALUE, CACHE_EXPIRATION_TIME)
            else
              counts[:invalid] += 1
              Rails.logger.warn "Invalid #{type} key skipped: #{key}"
            end
          rescue Dalli::DalliError => e
            Rails.logger.error "Dalli error while injecting key '#{key}': #{e.message}"
            raise
          end
        end
        counts
      rescue IOError => e
        Rails.logger.error "I/O error during file processing: #{e.message}"
        raise
      rescue ArgumentError => e
        Rails.logger.error "Argument error in inject: #{e.message}"
        raise
      end     
    end
  end
end