#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

def colorize(text, color_code)
  printf "\e[#{color_code}m#{text}\e[0m"
end

def red(text); colorize(text, 31); end
def green(text); colorize(text, 32); end
def e_ok()
  printf "[  "
  green("OK")
  printf "  ]\n"
end
def e_fail()
  red("[ FAIL ]\n")
end

namespace :redBorder do
  task environment_options: :environment do
    @worker_options = {
      min_priority: ENV['MIN_PRIORITY'] || 1,
      max_priority: ENV['MAX_PRIORITY'] || 10,
      queues: (ENV['QUEUES'] || ENV['QUEUE'] || '2').split(','),
      quiet: false
    }
  end

  #TODO: use clear_jobs task instead of this one.
  #      it will be deleted in future releases.
  desc "Clear the delayed_job queue."
  task clear: :environment do
    Delayed::Job.delete_all
  end

  desc 'Clear the delayed_job queue.'
  task clear_jobs: :environment_options do
    printf "Removing jobs                                              "
    Delayed::Job.delete_all
    e_ok
  end

  desc 'Restart Worker/Jobs'
  task restart_worker: :environment_options do
    if DelayedJob.running?
      printf "Stopping rb-workers                                        "
      DelayedJob.stop
    else
      printf "Stopping rb-workers (already stopped)                      "
    end

    sleep 1

    if DelayedJob.running?
      e_fail
    else
      e_ok

      printf "Starting rb_worker                                         "
      DelayedJob.start
      e_ok

      sleep 1

      if DelayedJob.running?
        printf "Adding default jobs to the queue                           "
        DelayedJob.run_now
        e_ok
      else
        e_fail
      end
    end
  end

  desc 'Stop Worker/Jobs'
  task stop_worker: :environment do
    if DelayedJob.running?
      printf "Stopping rb-workers                                        "
      DelayedJob.stop
    else
      printf "Stopping rb-workers (already stopped)                      "
    end
    sleep 1
    if DelayedJob.running?
      e_fail
    else
      e_ok
    end
  end

  desc 'Start Worker/Jobs'
  task start_worker: :environment_options do
    if DelayedJob.running?
      printf "Starting rb_worker (already started)                       "
    else
      printf "Starting rb_worker                                         "
      DelayedJob.start
    end

    sleep 2

    if DelayedJob.running?
      e_ok
      printf "Adding default jobs to the queue                           "
      DelayedJob.run_now
      e_ok
    else
      e_fail
    end
  end

  desc 'Generate a random string to identify the server'
  task :generate_server_key => :environment do

    # Get a random string with length = 30
    key = (0...30).map { (('a'..'z').to_a + (0..9).to_a + ('A'..'Z').to_a)[rand(62)] }.join

    # Save it
    Setting.server_key = key

  end

  # Usage: rake redBorder:admin_email['This is the subject','This is the body']
  desc 'Send email to super admin users'
  task :admin_email, [:subject, :body] => :environment do |_task, args|
    if args.subject.nil? || args.body.nil?
      puts "Usage: rake redBorder:admin_email['This is the subject','This is the body']"
    else
      AdminMailer.admin_email(args.subject, args.body).deliver_now
      puts 'Email was sent successfully.'
    end
  end

  # Usage: rake redBorder:admin_email['This is the subject','This is the body']
  desc 'Send email to super users'
  task :super_user_email, [:subject, :body] => :environment do |_task, args|
    if args.subject.nil? || args.body.nil?
      puts "Usage: rake redBorder:super_user_email['This is the subject','This is the body']"
    else
      AdminMailer.super_user_email(args.subject, args.body).deliver_now
      puts 'Email was sent successfully.'
    end
  end

  desc 'Copy assets from module to lib/assets directory'
  task :move_assets_module do
    # iterate over the modules
    Dir.entries("#{Rails.root}/lib/modules/").each do |dir|
      next if [".", ".."].include? dir
      # copy javascript files

      # Check if javascripts exists for this module
      if Dir.exists? "#{Rails.root}/lib/modules/#{dir}/app/assets/javascripts"
        Dir.entries("#{Rails.root}/lib/modules/#{dir}/app/assets/javascripts/").each do |file|
          FileUtils.mv("#{Rails.root}/lib/modules/#{dir}/app/assets/javascripts/#{file}", "#{Rails.root}/lib/assets/javascripts/#{file}", force: true) if File.file?("#{Rails.root}/lib/modules/#{dir}/app/assets/javascripts/#{file}")
        end
      end

      if Dir.exists? "#{Rails.root}/lib/modules/#{dir}/app/assets/stylesheets"
        # copy stylesheet files
        Dir.entries("#{Rails.root}/lib/modules/#{dir}/app/assets/stylesheets/").each do |file|
          FileUtils.mv("#{Rails.root}/lib/modules/#{dir}/app/assets/stylesheets/#{file}", "#{Rails.root}/lib/assets/stylesheets/#{file}", force: true) if File.file?("#{Rails.root}/lib/modules/#{dir}/app/assets/stylesheets/#{file}")
        end
      end

    end
  end

  desc 'Fix users that have no top domains assigned'
  task :fix_top_domain_user => :environment do
    users = User.where(sensor_id: nil)

    if users.present?
      print "Fixing #{users.count} users..."
      users.update_all(sensor_id: Sensor.root.id)
      puts "   [OK]\n"
      puts ""
    else
      puts "All users have their Top Domain assigned.\n"
      puts ""
    end
  end

  desc 'Request trial license'
  task request_trial_license: :environment do
    begin
      if License.any?
        puts 'A license for this cluster already exists.'
      else
        info = YAML.load_file("#{Rails.root}/config/licenses.yml")[Rails.env]
        uri = URI.parse(info['uri'])

        http = Net::HTTP.new(uri.host, uri.port)
        request = Net::HTTP::Post.new(uri.request_uri)
        request.add_field('Content-Type', 'application/json')
        request.body = { cluster_uuid: Setting.cluster_uuid }.to_json

        response = http.request(request)
        hash = JSON.parse(response.body)

        if hash['message'].present?
          puts hash['message']
        elsif hash['license'].present?
          license = License.new_or_update(encoded_license: hash['license'])
        end

        if license && license.save
          puts 'A new trial license has been loaded.'
        else
          puts 'There was a problem loading the trial license.'
        end
      end
    rescue StandardError => e
      puts 'There was a problem loading the trial license.'
    end
  end

  desc 'Migrates sensors serialized property to a rails store'
  task :migrate_sensor_property => :environment do
    # Sensor property should be a store at this point
    Sensor.all.each do |sensor|
      coder = ActiveRecord::Coders::YAMLColumn.new ActiveSupport::HashWithIndifferentAccess
      hash = coder.load sensor.property rescue nil

      if hash.nil?
        coder = ActiveRecord::Coders::YAMLColumn.new Hash
        hash = coder.load sensor.property rescue nil
      end

      if hash.nil?
        puts "Couldnt read property column from Sensor #{sensor.id}"
      else
        sensor.property = hash
        sensor.save
      end
    end
  end

  desc 'Return a regular expresion of the current emojis if they are available'
  task :emoji_regex do
    # Check if the folder exists
    emoji_folder = "#{Rails.root}/lib/modules/social/app/assets/images/emojis"

    if Dir.exists?(emoji_folder)
      # Open the regex
      regex = "/(["
      # Get all the emoji names
      emojis = Dir.glob("#{emoji_folder}/*.png").select{ |e| File.file? e }
      # Group range emojis
      initial = emojis.first.split("/").last.split(".").first.hex
      last = initial

      # Get the unicode for each item
      emojis.each do |item|

        # Check if is a multicode emoji
        unicode = item.split("/").last.split(".").first
        multicode = unicode.split("-")

        if initial == 0 && multicode.size == 1
          initial = unicode.hex
          last = initial
        end

        if multicode.size == 2 && multicode.first.size > 2
          # Multicode
          dif = multicode.last.hex - multicode.first.hex
          new_emojis = []

          dif.times do |index|
            new_emoji = (multicode.first.hex + index).to_s(16)
            unless File.exists?("#{emoji_folder}/#{new_emoji}.png")
              FileUtils.cp(item,"#{emoji_folder}/#{new_emoji}.png")
            end
          end

          # Group items
          regex = "#{regex}\\u{#{multicode.first}-\\u{#{multicode.last}}|"

          FileUtils.rm(item)

        elsif multicode.size == 2
          # Multicode but, we mus't split the files
          unless File.exists?("#{emoji_folder}/#{multicode.last}.png")
            FileUtils.cp(item,"#{emoji_folder}/#{multicode.last}.png")
          end

          FileUtils.rm(item)

          regex = "#{regex}\\u{#{multicode.last}}|"
        else
          # No multicode

          # Check if we must group
          if (unicode.hex == last + 1) or last == 0
            last = unicode.hex
          else
            if initial != 0 && initial != unicode.hex
              regex = "#{regex}\\u{#{(initial + 1).to_s(16)}}-\\u{#{unicode}}|"
            else
              regex = "#{regex}\\u{#{unicode}}|"
            end

            initial = 0
            last = 0
          end

        end

      end

      # Close the regex
      regex = "#{regex[0, regex.length - 1]}])/"
      # Show the regex
      puts regex
    end
  end

  desc 'Sets the correct values for widget positions based on dashboard layout'
  task :check_widget_positions => :environment do
    POSITIONS = {
      position_x: {
        columns: 1,
        absolute: '50px'
      },
      position_y: {
        columns: 1,
        absolute: '100px'
      },
      position_width: {
        columns: 2,
        absolute: '400px'
      },
      position_height: {
        columns: 2,
        absolute: '180px'
      }
    }

    Dashboard.all.each do |dashboard|
      dashboard.widgets.each do |widget|
        POSITIONS.each do |position, defaults|
          if dashboard.columns?
            if widget.send(position).to_s[-2..-1] == 'px'
              widget.send(position.to_s + '=', defaults[:columns])
              widget.save
            end
          else
            if widget.send(position).to_s[-2..-1] != 'px'
              widget.send(position.to_s + '=', defaults[:absolute])
              widget.save
            end
          end
        end
      end
    end
  end

  desc 'Update redBorder. It will content all the task that need to be run when updating the application.'
  task :update do
    Rake::Task['redBorder:update_cluster_databag_uuid'].invoke
    Rake::Task['redBorder:remove_generic_domains'].invoke
    Rake::Task['redBorder:update_organization_active_modules'].invoke
    Rake::Task['redBorder:fix_typos'].invoke
    Rake::Task['redBorder:create_license_databag'].invoke
    Rake::Task['redBorder:create_applications_databag'].invoke
    Rake::Task['redBorder:fix_sensor_on_licenses'].invoke
  end

  desc 'Assign users without sensor to root (and disable them)'
  task repair_isolated_users: :environment do
    root_id = Sensor.root.id
    n = 0
    User.find_each do |u|
      next if Sensor.find_by_id(u.sensor_id).present?
      u.update(sensor_id: root_id, enabled: false)
      n += 1
    end
    puts "#{n} Users updated"
  end

  desc 'Upgrade the widgets to new version (with parent)'
  task :upgrade_widgets_to_parent => :environment do
    Widget.all.each do |widget|
      next if widget.dashboard_id.nil?
      widget.parent_id = widget.dashboard_id
      widget.parent_type = 'Dashboard'
      widget.save
    end
    puts 'Done'
  end

  desc 'Swaps widgets x and y positions on columns dashboards'
  task :swap_columns_widgets_positions => :environment do
    Widget.all.each do |widget|
      next unless widget.dashboard.columns?
      aux = widget.position_x
      widget.position_x = widget.position_y
      widget.position_y = aux
      widget.save
    end
  end

  desc 'Rebuild sensors to fill awesome_nested_set columns'
  task :rebuild_sensors => :environment do
    Sensor.rebuild!
  end

  desc 'Assigns uuid to sensors that have it to nil'
  task :uuid_to_sensors => :environment do
    Sensor.where(uuid: nil).each do |sensor|
      rand_int = rand(1..9_223_372_036_854_775_807)
      sensor.update(uuid: rand_int)
    end
  end

  desc 'Add domain information to sensor Chef Role'
  task :add_domain_to_chef_role => :environment do
    sensors = Sensor.where(type: Sensor::DOMAIN)
    puts "#{sensors.size} domains will be updated"
    puts 'Processing...'

    sensors.each do |s|
      dname = s.domain_type_name to_key: true
      role = s.chef_role
      role.override_attributes['redborder'][dname.to_s] = s.name
      role.override_attributes['redborder']["#{dname}_uuid"] = s.uuid
      role.save
      puts "Updated #{s.name} (UUID: #{s.uuid})"
    end
    puts 'Finish'
  end

  desc 'Creates four namespaces for testing purposes'
  task :create_testing_namespaces => :environment do
    Sensor.create(name: 'namespace1', type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE, uuid: '11111111', parent: Sensor.root)
    Sensor.create(name: 'namespace2', type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE, uuid: '22222222', parent: Sensor.root)
    Sensor.create(name: 'namespace3', type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE, uuid: '33333333', parent: Sensor.root)
    Sensor.create(name: 'namespace4', type: Sensor::DOMAIN, domain_type: Sensor::NAMESPACE, uuid: '44444444', parent: Sensor.root)
  end

  desc 'Upgrade objects by user'
  task :upgrade_objects => :environment do
    # For every object, duplicate it on users. Current objects will be
    # assigned to first admin user
    admin = User.find_by(admin: true, id: 1) || User.where(admin: true).first
    # Fetch all users ids
    user_ids = User.where(admin: true).map(&:id)

    # For every object...
    VlanObject.where(user_id: nil).each do |vlan|
      vlan.user = admin
      vlan.save

      user_ids.each do |user_id|
        nvlan = vlan.dup
        nvlan.user_id = user_id
        nvlan.save
      end
    end

    MacObject.where(user_id: nil).each do |mac|
      mac.user = admin
      mac.save

      user_ids.each do |user_id|
        nmac = mac.dup
        nmac.user_id = user_id
        nmac.save
      end
    end

    NetObject.where(user_id: nil).each do |net|
      net.user = admin
      net.save

      user_ids.each do |user_id|
        nnet = net.dup
        nnet.user_id = user_id
        nnet.save
      end
    end
  end

  desc 'Update active_data_sources property of users.'
  task :update_active_data_sources => :environment do
    User.all.each do |user|
      user.active_data_sources = user.data_sources_ids
      user.save
    end
  end

  desc 'Update old organization active_modules with nil values.'
  task :update_organization_active_modules => :environment do
    # Find old organizations without active_modules field
    organizations = Sensor
                    .select{ |s| s.domain_type.eql?(Sensor::ORGANIZATION) &&
                      s.active_modules.nil? }

    # Update them
    organizations.each do |sensor|
      # Set the value of its namespace parent if exists
      namespace = sensor.ancestors.where(domain_type: Sensor::NAMESPACE).first
      if namespace && namespace.active_modules
        sensor.update(active_modules: namespace.active_modules)
        next
      end

      # Set the value from config files if it has no namespace parent with active_modules
      sensor.active_modules = Redborder::Module.available.join(',')

      sensor.save
    end
  end

  desc 'Update (or create) a cluster databag with a new uuid if it does not exist.'
  task :update_cluster_databag_uuid => :environment do
    data = Redborder::Databag.load('rBglobal', 'cluster')
    if data && data['uuid'].nil?
      data = Chef::DataBagItem.new
      data.data_bag 'rBglobal'
      data['id'] = 'cluster'
      data['uuid'] = SecureRandom.uuid
      data.save
    end
  end

  desc 'Create the license databag if it does not exist.'
  task create_license_databag: :environment do
    begin
      data = Chef::DataBagItem.load('rBglobal', 'licenses')
      data['licenses'] ||= {}
      data['sensors'] ||= {}
      data.save
    rescue Net::HTTPServerException
      data = Chef::DataBagItem.new
      data.data_bag 'rBglobal'
      data['id'] = 'licenses'
      data['licenses'] = {}
      data['sensors']  = {}
      data.save
    end
  end

  desc 'Create the applications databag if it does not exist.'
  task create_applications_databag: :environment do
    begin
      data = Chef::DataBagItem.load('rBglobal', 'applications')
      data['id'] = 'applications'
      # For elder versions it was an array, so we must change the items to hash
      if data['list'].is_a?(Array)
        data['list'] = data['list'].map { |l| [l[:value_int].to_s, l] }.to_h
      else
        data['list'] ||= {}
      end
      data.save
    rescue Net::HTTPServerException
      data = Chef::DataBagItem.new
      data.data_bag 'rBglobal'
      data['id'] = 'applications'
      data['list'] = {}
      data.save
    end
  end

  desc 'Assign root to licenses with no sensor relation'
  task fix_sensor_on_licenses: :environment do
    License.where(sensor_id: nil).update_all(sensor_id: Sensor.root.id)
  end

  desc 'Fix database typos'
  task :fix_typos => :environment do
    Severity.find_by(name: 'extremly high').try(:update , name: 'extremely high')
  end

  desc 'Reload sensor path name cache'
  task :reload_sensor_path_names => :environment do
    Sensor.reload_path_names!
  end

  desc 'Remove generic domains and convert them to another domain_type'
  task :remove_generic_domains => :environment do
    # 1 is the old value of generic sensors
    generic = Sensor.where(domain_type: 1, type: Sensor::DOMAIN)
    allowed = [Sensor::MARKET, Sensor::CAMPUS]

    generic.each do |domain|
      if domain.name == 'Unclaimed'
        new_type = Sensor::SERVICE_PROVIDER
      else
        used_types = domain.ancestors.where(type: Sensor::DOMAIN).map!(&:domain_type) +
                     domain.descendants.where(type: Sensor::DOMAIN).map!(&:domain_type)
        used_types.uniq!
        new_type = (allowed - used_types).first || Sensor::DEPLOYMENT
      end
      domain.domain_type = new_type
      domain.save(validate: false)
    end
    # Update root
    Sensor.root.update_columns(domain_type: Sensor::ROOT)
    # Log
    puts "Finished. Updated #{generic.size} domains"
  end

  desc 'Upgrade serialized properties to store on sensors'
  task :upgrade_sensor_properties => :environment do
    query = 'SELECT id, property FROM sensors'
    # Save number of updated sensors.
    updated = 0
    results = ActiveRecord::Base.connection.execute(query)
    # Check every sensor
    results.each do |res|
      next if valid_json?(res['property'])
      # Update with a valid JSON
      begin
        property = YAML.load(res['property']).to_json
      rescue Exception => e
        property = '{}'
      end
      # Update on database
      id = res['id']
      query = "UPDATE sensors SET property='#{property}'"\
              " WHERE id='#{id}';"
      ActiveRecord::Base.connection.execute(query)
      updated += 1
    end
    # Show some info
    puts "Done. Updated #{updated} sensors"
  end

  #
  # Check if the current string is a valid JSON
  #
  # == Returns:
  # Boolean
  #
  def valid_json?(json)
    begin
      JSON.parse(json)
      return true
    rescue Exception => e
      return false
    end
  end

  desc 'Upgrade old to new widgets'
  task :upgrade_widgets => :environment do
    # Save base x and y position
    x = 1
    y = 1
    last_dashboard = nil

    Widget.order(dashboard_id: :asc).all.each do |ow|
      dashboard_id = ow.dashboard_id || ow.parent_id
      dashboard = Dashboard.find_by_id(dashboard_id)

      # Reset x and y when change dashboard
      if last_dashboard.present? && last_dashboard != dashboard
        x = 1
        y = 1
      end

      last_dashboard = dashboard

      # Check if the widget is updated ormust be ignored
      next if dashboard_id.nil? || ow.property['show_name'].nil? ||
              %w(date video webcam).include?(ow.product_type)

      # Convert it
      widget = Widget.new(
        name: ow.property['show_name'] == '1' ? ow.name : '',
        transparent: ow.transparent,
        click_action: ow.click_action,
        click_dashboard: ow.click_dashboard,
        filters: ow.filters,
        parent: dashboard
      )

      widget = convert_widget(widget, ow)

      # Set position
      # Calc width:
      if dashboard.layout == 'columns'
        width = 4
        # Set positions
        widget.position_width = width
        widget.position_height = 2

        if (x + width) > 9
          x = 1
          y = y + 2
        end

        widget.position_x = x
        widget.position_y = y

        x += width
      else
        widget.position_x = ow.position_x
        widget.position_y = ow.position_y
        widget.position_width = ow.position_width
        widget.position_height = ow.position_height
      end

      widget.save
    end
  end

  #
  # Convert the widget based on product type
  #
  def convert_widget(widget, old_widget)
    case old_widget.product_type
    when 'flow', 'social', 'ips', 'malware', 'monitor'
      widget.product_type = old_widget.product_type
      widget =
          convert_widget_type(widget, old_widget, old_widget.widget_type)
      widget =
          convert_widget_graph_type(widget, old_widget, old_widget.widget_type)
    when 'sensors'
      widget.product_type = 'none'
      widget.widget_type = 'sensor'
      widget.graph_type = old_widget.sensor_table_type.include?('map') ?
                          'map' : 'sensor_tree'
    when 'cluster'
      widget.product_type = 'none'
      widget.widget_type = 'cluster'
      widget.graph_type = old_widget.cluster_type == 'status' ?
                          'cluster_table' : 'cluster_diagram'
      widget.size = 'small'
    when 'custom', 'text', 'square', 'image'
      widget.product_type = 'none'
      widget.widget_type = 'shape'
      widget.graph_type = old_widget.product_type

      if widget.graph_type == 'custom'
        widget.height = old_widget.height
        widget.url = old_widget.url
      elsif widget.graph_type == 'image'
        # Store the image
        widget.media = old_widget.image if old_widget.image_file_name.present?
        widget.keep_aspect_ratio = old_widget.keep_aspect_ratio
      elsif widget.graph_type == 'square'
        widget.square_border = old_widget.square_border
        widget.square_fill = old_widget.square_fill
        widget.square_opacity = old_widget.square_opacity
        widget.square_border_enabled = old_widget.square_border_enabled
      else
        widget.text_alignment = '1'
        widget.font_size = '24px'
        widget.name = old_widget.name
      end
    when 'alarms'
      widget.product_type = 'none'
      widget.widget_type = 'alarm'
      widget.graph_type = 'alarm'
    end

    widget
  end

  desc 'Upgrade Report blocks to widgets'
  task :upgrade_report_blocks => :environment do
    # We must convert Report Blocks to widgets
    x = 1
    y = 1

    ReportBlock.all.each do |rblock|
      report = Report.find_by_id(rblock.report_id)

      # The report dont exist!
      next if report.nil?

      # Load page
      page = rblock.page ? (rblock.page + 1) : 1
      page = report.number_pages if page > report.number_pages

      widget = Widget.new(
        name: rblock.is_custom? ? rblock.name : '',
        description: rblock.description,
        transparent: rblock.transparent,
        page: page,
        click_action: 'none',
        filters: rblock.filters,
        parent: report
      )

      next if rblock.product_type == 'cluster' ||
                rblock.report_block_type == 'Aps' ||
                (rblock.product_type == 'monitor' &&
                 rblock.report_block_type != 'Top') ||
                rblock.graph_type == 'table' ||
                rblock.is_separator?

      widget = convert_report_block(widget, rblock)

      # Calc width:
      if rblock.width < 4
        width = rblock.width + 1
      else
        width = 6
      end

      # Set positions
      widget.position_width = width
      widget.position_height = 2

      if (x + width) > 6
        x = 1
        y = y + 2
      end

      widget.position_x = x
      widget.position_y = y

      x += rblock.width

      widget.save
    end
  end

  desc 'Get next modules migration id available'
  task :next_module_migration_id do
    begin
      last_migration_name = Dir["lib/modules/*/db/migrate/*.rb"].collect { |file| File.basename(file) }.sort.last
      result = /\A(?<id>\d+)_.+\z/.match(last_migration_name)
      id = result[:id].to_i + 1
      puts id < 100 ? "0#{id}" : id
    rescue StandardError
      puts 'There was an error trying to get the last module migration'
    end
  end

  #
  # Conver the report blcok to a new widget
  #
  def convert_report_block(widget, rblock)
    # We must check the report block type (Common or custom). Common blocks
    # are now shape widgets
    if(rblock.is_custom?)
      # Store product and widget type
      widget.product_type = rblock.product_type
      widget =
        convert_widget_type(widget, rblock, rblock.report_block_type)
      widget =
        convert_widget_graph_type(widget, rblock, rblock.report_block_type)
    else
      # Common report blocks
      # Set the other values
      widget.product_type = 'none'
      widget.widget_type = 'shape'

      # Convert the common report block type to widget graph type
      if rblock.is_image?
        widget.graph_type = 'image'
        # Store the image
        widget.media = rblock.image if !rblock.image_file_name.nil?
        widget.keep_aspect_ratio = rblock.keep_aspect_ratio
      else
        # Titles and texts
        widget.graph_type = 'text'
        widget.text_alignment = rblock.text_alignment
        widget.font_size = '24px'
        widget.name = rblock.is_title? ? rblock.title : rblock.text
      end
    end

    widget
  end

  #
  # Convert the report block type and widget to new widget_type
  #
  def convert_widget_type(widget, rblock, type)
    if rblock.graph_type == 'map'
      widget.widget_type = 'map'
      widget.range = rblock.range
      widget.column = rblock.column
      widget.aggregation = rblock.aggregation
    elsif rblock.product_type == 'monitor'
      widget.widget_type = 'monitor'
      widget.elements = rblock.property['top_num']
      widget.range = rblock.range
      widget.category = rblock.category
      widget.aggregation = rblock.aggregation
    else
      case type.downcase
      when 'cardinality'
        if rblock.property['cardinality_group_by'] == 'none' ||
           rblock.property['cardinality_group_by'] == :none
          widget.widget_type = 'single_unique'
        else
          widget.widget_type = 'grouped_unique'
          widget.group_by = rblock.property['cardinality_group_by']
          widget.elements = rblock.property['top_num']
        end
        widget.column = rblock.column
        widget.range = rblock.range
      when 'top'
        widget.widget_type = 'top'
        widget.elements = rblock.property['top_num']
        widget.range = rblock.range
        widget.column = rblock.column
        widget.aggregation = rblock.aggregation
      when 'raw'
        widget.widget_type = 'raw'
        widget.range = rblock.range
        widget.aggregation = rblock.aggregation
      when 'compare'
        widget.widget_type = 'compare'
        widget.elements = rblock.property['top_num']
        widget.range = rblock.compare_range
        widget.aggregation = rblock.aggregation
      end
    end

    widget
  end

  #
  # Convert the report block and widget graph type tp graph_type
  #
  def convert_widget_graph_type(widget, rblock, type)
    case rblock.graph_type
    when 'stacked', 'area', 'line', 'bars', 'sbars'
      if type == 'Top'
        widget.graph_type = 'multiserie'
        widget.show_total = rblock.show_total
      else
        widget.graph_type = 'serie'
        widget.color = rblock.color
      end
      widget.serie_type = rblock.graph_type
      widget.granularity = rblock.granularity
      widget.timezone = rblock.timezone
    when 'horbars'
      widget.graph_type = 'horizontal_bar'
    when 'verbars'
      widget.graph_type = 'vertical_bar'
    when 'pie'
      widget.graph_type = rblock.graph_type
    when 'ball'
      widget.graph_type = rblock.graph_type
      widget.ball_max = rblock.ball_max
    when 'value'
      widget.graph_type = rblock.graph_type
      widget.reference = rblock.reference
    when 'streams'
      widget.graph_type = 'stream'
      widget.elements = rblock.property['top_num']
      widget.stream_columns = rblock.stream_columns
    when 'map'
      widget.graph_type = 'map'
      widget.map_type = rblock.map_type
    end

    widget
  end
end
