require 'test_helper'
require 'json'

class Api::V1::SensorsControllerTest < ActionDispatch::IntegrationTest
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'an admin user' do
    setup do
      @user = users(:admin)
      @auth_token = users(:admin).authentication_token
      # Base api
      @api = 'api/v1'
      @uuid1 = 'adashjkanfsjbnads453asdashjk'
      @uuid2 = 'uijofkld798y2hiugfbdjksi1u21'
      # Example sensor
      @organization = sensors(:organization)
      @organization_uuid = @organization.uuid
      # Organization license
      @license = licenses(:four)
      # To test MAC Hashing salt
      @service_provider = sensors(:service_provider)
      # Domain to override
      @market = sensors(:market)
      # Floor
      @floor = sensors(:floor)

      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'organization'
    end

    #
    # Tested method: index
    #
    should 'render all sensors' do
      get "/#{@api}/sensors?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(json['sensors'].size > 0, 'The API must return some sensors')
      assert_equal(@user.sensor.self_and_descendants.size, json['sensors'].size,
                   'The user only must see his own sensors')
    end

    #
    # Tested method: tree
    #
    should 'render all sensors as tree' do
      get "/#{@api}/sensors/tree?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(json['sensor'].size > 0, 'The API must return some sensors')
    end

    #
    # Tested method: tree
    #
    should 'render all sensors as tree from one given sensor' do
      get "/#{@api}/sensors/#{@organization_uuid}/tree?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(json['sensor'].size > 0, 'The API must return some sensors')
      assert_equal(
        @organization_uuid, json['sensor']['uuid'],
        'Returned sensor by tree method must be sensor sent on URL (UUID)')
      assert_equal(
        @organization.children.count, json['sensor']['children'].size,
        'The number of children must coincide with organization children')
    end

    should 'create a new domain (DEPRECATED)' do
      name = 'DomainAPI'
      post "/#{@api}/sensors/create_domain?auth_token=#{@auth_token}",
           name: name, domain_type: Sensor::SERVICE_PROVIDER, uuid: @uuid1
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(!json['sensor'].empty?, 'The API must return created sensor')
      assert(json['sensor']['name'] == name,
             "Created sensor must has name attribute equal to #{name}")
      assert(json['sensor']['uuid'] == @uuid1,
             "Created sensor must has uuid attribute equal to #{@uuid1}")
    end

    should 'create a new domain' do
      name = 'DomainAPI2'
      range = 123
      post "/#{@api}/sensors/domain?auth_token=#{@auth_token}",
           name: name, domain_type: Sensor::SERVICE_PROVIDER, uuid: @uuid2,
           visualization_range: range
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(!json['sensor'].empty?, 'The API must return created sensor')
      assert_equal(
        name, json['sensor']['name'],
        "Created sensor must has name attribute equal to #{name}")
      assert_equal(
        @uuid2, json['sensor']['uuid'],
        "Created sensor must has uuid attribute equal to #{@uuid2}")
    end

    should 'create a new domain without domain_type in service provider' do
      name = 'DomainAPI3'
      post "/#{@api}/sensors/domain?auth_token=#{@auth_token}",
           name: name, parent_uuid: @service_provider.uuid
      # Domain type must be...
      domain_type = Sensor::HIERARCHY[Sensor::HIERARCHY.index(@service_provider.domain_type) + 1]
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(!json['sensor'].empty?, 'The API must return created sensor')
      assert_equal(domain_type, json['sensor']['domain_type'])
    end

    should 'update a domain' do
      name = 'NewNameForDomain'
      patch "/#{@api}/sensors/#{@organization_uuid}?auth_token=#{@auth_token}",
            name: name
      assert_response :success
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(name, json['sensor']['name'], 'Name must be updated')
    end

    should 'update MAC Hashing salt on Service Provider' do
      mac_hashing_salt = '87923yruihsfjkd'
      patch "/#{@api}/sensors/#{@service_provider.uuid}?auth_token=#{@auth_token}",
            mac_hashing_salt: mac_hashing_salt
      assert_response :success
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(mac_hashing_salt, json['sensor']['mac_hashing_salt'],
                   'MAC Hashing Salt must be updated')
    end

    should 'update visualization range in an organization' do
      range = 15
      patch "/#{@api}/sensors/#{@organization_uuid}?auth_token=#{@auth_token}",
            visualization_range: range
      assert_response :success
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_equal(range, json['sensor']['visualization_range'].to_i,
                   'Range must be updated')
    end

    should 'destroy a domain' do
      delete "/#{@api}/sensors/#{@organization_uuid}?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
    end

    should 'can\'t destroy his own top domain' do
      delete "/#{@api}/sensors/#{@user.sensor.uuid}?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false')
    end

    should 'override all of a domain (big query)' do
      data = JSON.parse(
        File.read(Rails.root.join('test', 'data', 'override_all.json')))

      post "/#{@api}/sensors/#{@market.uuid}/override_all?auth_token=#{@auth_token}",
           data

      assert_response 200
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
    end

    #
    # Tested method: move
    #
    should 'move a market into organization' do
      patch "/#{@api}/sensors/#{@market.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @organization_uuid

      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
    end

    should 'can\'t move a domain into a floor' do
      patch "/#{@api}/sensors/#{@market.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @floor.uuid

      assert(response.response_code == 422,
             "Response code should be 422 : got #{response.response_code}")
    end

    #
    # Tested method: licenses
    #
    should 'render all licenses' do
      get "/#{@api}/sensors/#{@organization_uuid}/licenses?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert_equal(json.first['id'], @license.id, 'Wrong id license')
      assert_equal(json.first['uuid'], @license.uuid, 'Wrong uuid license')
      assert_equal(json.first['cluster_uuid'], @license.cluster_uuid, 'Wrong cluster_uuid license')
    end

    #
    # Tested method: licenses (available by type)
    #
    should 'render licenses by type' do
      get "/#{@api}/sensors/#{@organization_uuid}/licenses?auth_token=#{@auth_token}&available_by_type=flow"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert_equal(json.count, 1, 'Wrong number of licenses')
      assert_equal(json.first['uuid'], @license.uuid, 'Wrong uuid license')
      assert_equal(json.first['cluster_uuid'], @license.cluster_uuid, 'Wrong cluster_uuid license')
      assert(json.first['sensors']['flow'] > 0, 'There is no available license for this type of sensor')
    end

    #
    # Tested method: license
    #
    should 'render a license' do
      get "/#{@api}/sensors/#{@organization_uuid}/licenses/#{@license.id}?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert_equal(json['id'], @license.id, 'Wrong id license')
      assert_equal(json['uuid'], @license.uuid, 'Wrong uuid license')
      assert_equal(json['cluster_uuid'], @license.cluster_uuid, 'Wrong cluster_uuid license')
    end
  end

  context 'an user without manage permissions' do
    setup do
      @user = users(:user_domain)
      @auth_token = users(:user_domain).authentication_token
      # Base api
      @api = 'api/v1'
      @domain_uuid = sensors(:deployment).uuid
      @market = sensors(:market)
      @organization = sensors(:organization)
    end

    should 'render his sensors' do
      get "/#{@api}/sensors?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(json['sensors'].size > 0, 'The API must return some sensors')
      assert_equal(@user.sensor.self_and_descendants.size, json['sensors'].size,
                   'The user only must see his own sensors')
    end

    should 'render his sensors as tree' do
      get "/#{@api}/sensors/tree?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 200,
             "Response code should be 200 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert(json['sensor'].size > 0, 'The API must return some sensors')
    end

    should 'can\'t create a new domain (DEPRECATED)' do
      post "/#{@api}/sensors/create_domain?auth_token=#{@auth_token}",
           name: 'DomainAPI', domain_type: Sensor::ORGANIZATION, uuid: @uuid1
      # Check status
      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false')
    end

    should 'can\'t create a new domain' do
      name = 'DomainAPI2'
      post "/#{@api}/sensors/domain?auth_token=#{@auth_token}",
           name: name, domain_type: Sensor::ORGANIZATION, uuid: @uuid2
      # Check status
      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
      # Check response and number of items
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false')
    end

    should 'can\'t destroy a domain' do
      delete "/#{@api}/sensors/#{@domain_uuid}?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false')
    end

    should 'can\'t destroy his own top domain' do
      delete "/#{@api}/sensors/#{@user.sensor.uuid}?auth_token=#{@auth_token}"
      # Check status
      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false')
    end

    should 'not override all of a domain (big query)' do
      data = JSON.parse(
        File.read(Rails.root.join('test', 'data', 'override_all.json')))

      post "/#{@api}/sensors/#{@market.uuid}/override_all?auth_token=#{@auth_token}",
           data

      assert_response 401
      json = JSON.parse(response.body)
      assert(!json['query'], 'Query value must be false (no permission)')
    end

    should 'can\'t move domains into another' do
      patch "/#{@api}/sensors/#{@market.uuid}/move?auth_token=#{@auth_token}",
            parent_uuid: @organization_uuid

      assert(response.response_code == 401,
             "Response code should be 401 : got #{response.response_code}")
    end
  end

  context 'an admin user' do
    setup do
      @user = users(:admin_domain)
      @auth_token = users(:admin_domain).authentication_token
      # Base api
      @api = 'api/v1'
      @client_proxy = sensors(:client_proxy)
      @client_proxy_2 = sensors(:client_proxy_2)
      @organization = sensors(:organization)
      @hash = 'b736c4d8-ae91-48ff-853e-b4a2c0e2465a'
    end

    should 'register a client proxy' do
      post "/#{@api}/sensors",
           { order: 'register', type: Sensor::CLIENT_PROXY, hash: @hash },
           { 'X-Real-IP' => '10.0.30.5' }
      assert_response :success, "Response should be 200 OK : got #{response.response_code}"
      json = JSON.parse(response.body)
      assert_equal 'registered', json['status'], 'Status response should be registered'
      client_proxy = Sensor.find_by(uuid: json['uuid'])
      assert_equal client_proxy.ip, '10.0.30.5', 'IP is not correct'
    end

    should "can't register an already registered client proxy" do
      post "/#{@api}/sensors", order: 'register', type: Sensor::CLIENT_PROXY, hash: @client_proxy.mac_address
      assert_response :forbidden, "Response should be 403 Forbidden: got #{response.response_code}"
    end

    should "can't verify an unclaimed client proxy" do
      post "/#{@api}/sensors", order: 'verify', hash: @client_proxy_2.mac_address, uuid: @client_proxy_2.uuid
      assert_response :success, "Response should be 200 OK : got #{response.response_code}"
      json = JSON.parse(response.body)
      assert_equal 'registered', json['status'], 'Status response should be registered'
    end

    should 'can verify a claimed client proxy' do
      post "/#{@api}/sensors", order: 'verify', hash: @client_proxy.mac_address, uuid: @client_proxy.uuid
      assert_response :success, "Response should be 200 OK : got #{response.response_code}"
      # { status: 'claimed', cert: @client['private_key'].to_json, nodename: "rbproxy-#{@sensor.id}" }
      json = JSON.parse(response.body)
      assert_equal 'claimed', json['status'], 'Status response should be claimed'
    end

    should "can't update visualization range in an organization" do
      range = 15
      patch "/#{@api}/sensors/#{@organization.uuid}?auth_token=#{@auth_token}",
            visualization_range: range
      assert_response :success
      json = JSON.parse(response.body)
      assert(json['query'], 'Query value must be true')
      assert_not(json['sensor'].key?('visualization_range'), "Range musn't be showed")
      assert_not_equal(range, @organization.reload.visualization_range, "Range musn't be update")
    end
  end
end
