require 'test_helper'

class LicensesControllerTest < ActionController::TestCase
  include Warden::Test::Helpers
  Warden.test_mode!
  include Devise::TestHelpers

  setup do
    @root = sensors(:root)
    @organization = sensors(:organization)
  end

  context 'In Global licensing' do
    setup do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'global'
    end

    context 'stubing consumption_data method' do
      setup do
        SensorMonitor.stubs(:consumption_data).returns([])
      end

      context 'Super admin user' do
        setup do
          @user = users(:admin)
          sign_in @user

          init_license_files
        end

        should 'access root licenses view' do
          get :index
          assert_response :success
        end

        should 'access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_response :not_found
        end

        should 'not access deployment licenses view' do
          @deployment = sensors(:deployment)
          get :index, sensor_id: @deployment.id
          assert_response :not_found
        end

        should 'add global license' do
          assert_difference 'License.count', 1 do
            post :create, license: { file: @upload_global_license }
          end
        end

        should 'add organization license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization.id, license: { file: @upload_organization_license }
          end
        end

        should 'not add license to a wrong organization' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization.id, license: { file: @upload_global_license }
          end
        end

        should 'delete global license' do
          @license = licenses(:three)

          assert_difference 'License.count', -1 do
            delete :destroy, id: @license.id
          end
        end

        should 'get consumption CSV file' do
          namespaces_count = Sensor.root.namespaces.count

          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :cardinality &&
              args[:data_sources].length == namespaces_count
          end.returns([]).at_least_once

          get :export_consumption_to_file, month: Time.now.utc.beginning_of_month
          assert_response :success
        end
      end

      context 'Organization admin user' do
        setup do
          @user = users(:admin_domain)
          sign_in @user

          init_license_files
        end

        should 'not access root licenses view' do
          get :index, sensor_id: @root.id
          assert_redirected_to root_path
        end

        should 'not access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_response :not_found

          get :index # if no sensor_id parameter is passed, use the user top domain
          assert_response :not_found
        end

        should 'not add global license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @root.id, license: { file: @upload_global_license }
          end
        end

        should 'not add organization license' do
          assert_no_difference 'License.count' do
            post :create, license: { file: @upload_organization_license }
          end
        end

        should 'not add license to another organization' do
          @organization_2 = sensors(:organization_2)

          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization_2.id, license: { file: @upload_organization_license }
          end
        end

        should 'not delete global license' do
          @license = licenses(:three)

          assert_no_difference 'License.count' do
            delete :destroy, sensor_id: @root.id, id: @license.id
          end
        end
      end

      context 'Normal user' do
        setup do
          @user = users(:user_domain)
          @organization = sensors(:organization)
          sign_in @user
        end

        should 'not access global licenses view' do
          get :index
          assert_redirected_to root_path
        end

        should 'not access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_redirected_to root_path
        end

        should 'not add global license' do
          assert_no_difference 'License.count' do
            post :create, license: { file: @upload_global_license }
          end
        end

        should 'not add organization license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization.id, license: { file: @upload_organization_license }
          end
        end
      end
    end

    context 'without stubing consumption_data method' do
      context 'Super admin user' do
        setup do
          @user = users(:admin)
          sign_in @user

          init_license_files
        end

        should 'access root licenses view' do
          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :timeseries &&
              args[:data_sources] == 'rb_monitor'
          end.returns([]).once

          get :index
          assert_response :success
        end
      end
    end
  end

  context 'In Organization licensing' do
    setup do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'organization'
    end

    context 'stubing consumption_data method' do
      setup do
        SensorMonitor.stubs(:consumption_data).returns([])
      end

      context 'a super admin user' do
        setup do
          @user = users(:admin)
          sign_in @user

          init_license_files
        end

        should 'not access root licenses view' do
          get :index, sensor_id: @root.id
          assert_response :not_found
        end

        should 'access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_response :success
        end

        should 'not access deployment licenses view' do
          @deployment = sensors(:deployment)
          get :index, sensor_id: @deployment.id
          assert_response :not_found
        end

        should 'not add global license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @root.id, license: { file: @upload_global_license }
          end

          assert_no_difference 'License.count' do
            post :create, license: { file: @upload_global_license }
          end
        end

        should 'add organization license' do
          assert_difference 'License.count', 1 do
            post :create, sensor_id: @organization.id, license: { file: @upload_organization_license }
          end
        end

        should 'not add organization license without sensor param' do
          assert_no_difference 'License.count' do
            post :create, license: { file: @upload_organization_license }
          end
        end

        should 'not add license to a wrong organization' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization.id, license: { file: @upload_global_license }
          end
        end

        should 'delete organization license' do
          @license = licenses(:four)

          assert_difference 'License.count', -1 do
            delete :destroy, sensor_id: @organization.id, id: @license.id
          end
        end

        should 'not delete license from a wrong organization' do
          @license = licenses(:four)
          @organization_2 = sensors(:organization_2)

          assert_no_difference 'License.count' do
            delete :destroy, sensor_id: @organization_2, id: @license.id
          end
        end

        should 'get consumption CSV file' do
          namespace_uuid = @organization.namespaces.first.uuid

          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :cardinality &&
              args[:data_sources] =~ /_#{namespace_uuid}/ &&
              args[:search].try(:[], :organization_uuid).try(:[], @organization.uuid)
          end.returns([]).at_least_once

          get :export_consumption_to_file, month: Time.now.utc.beginning_of_month, sensor_id: @organization.id
          assert_response :success
        end
      end

      context 'an organization admin user' do
        setup do
          @user = users(:admin_domain)
          sign_in @user

          init_license_files
        end

        should 'not access root licenses view' do
          get :index, sensor_id: @root.id
          assert_response :not_found
        end

        should 'access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_response :success

          get :index # if no sensor_id parameter is passed, use the user top domain
          assert_response :success
        end

        should 'not add global license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @root.id, license: { file: @upload_global_license }
          end
        end

        should 'add organization license' do
          assert_difference 'License.count', 1 do
            post :create, license: { file: @upload_organization_license }
          end
        end

        should 'not add license to another organization' do
          @organization_2 = sensors(:organization_2)

          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization_2.id, license: { file: @upload_organization_license }
          end
        end

        should 'delete organization license' do
          @license = licenses(:four)

          assert_difference 'License.count', -1 do
            delete :destroy, id: @license.id
          end
        end

        should 'get consumption CSV file' do
          namespace_uuid = @organization.namespaces.first.uuid

          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :cardinality &&
              args[:data_sources] =~ /#{namespace_uuid}/ &&
              args[:search].try(:[], :organization_uuid).try(:[], @organization.uuid)
          end.returns([]).at_least_once

          get :export_consumption_to_file, month: Time.now.utc.beginning_of_month, sensor_id: @organization.id
          assert_response :success
        end
      end

      context 'a normal user' do
        setup do
          @user = users(:user_domain)
          @organization = sensors(:organization)
          sign_in @user
        end

        should 'not access root licenses view' do
          get :index
          assert_redirected_to root_path
        end

        should 'not access organization licenses view' do
          get :index, sensor_id: @organization.id
          assert_redirected_to root_path
        end

        should 'not add global license' do
          assert_no_difference 'License.count' do
            post :create, license: { file: @upload_global_license }
          end
        end

        should 'not add organization license' do
          assert_no_difference 'License.count' do
            post :create, sensor_id: @organization.id, license: { file: @upload_organization_license }
          end
        end
      end
    end

    context 'without stubing consumption_data method' do
      setup do
        init_license_files
      end

      context 'a super admin user' do
        setup do
          @user = users(:admin)
          sign_in @user
        end

        should 'access organization licenses view' do
          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :timeseries &&
              args[:search][:organization_uuid][@organization.uuid] &&
              args[:data_sources] == 'rb_monitor'
          end.returns([]).once

          get :index, sensor_id: @organization.id
          assert_response :success
        end
      end

      context 'an organization admin user' do
        setup do
          @user = users(:admin_domain)
          sign_in @user
        end

        should 'access organization licenses view' do
          Redborder::Rbdruid.expects(:query).with do |args|
            args[:query_type] == :timeseries &&
              args[:search][:organization_uuid][@organization.uuid] &&
              args[:data_sources] == 'rb_monitor'
          end.returns([]).once

          get :index, sensor_id: @organization.id
          assert_response :success
        end
      end
    end
  end

  private

  def init_license_files
    global_license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid,
                                    expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)
    
    # TODO: This is a workarround as stubs(:read) looks like is not working from the controller side
    file = File.open("files/license.lic", "w") { |f| f.write global_license }

    @upload_global_license = fixture_file_upload('files/license.lic')
    @upload_global_license.stubs(:read).returns(global_license)

    organization_license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid,
                                          organization_uuid: @organization.uuid, expire_at: (Time.now + 30.days).to_i,
                                          limit_bytes: 500_000_000)

    # TODO: This is a workarround as stubs(:read) looks like is not working from the controller side
    file = File.open("files/license.lic", "w") { |f| f.write organization_license }

    @upload_organization_license = fixture_file_upload('files/license.lic')
    @upload_organization_license.stubs(:read).returns(organization_license)

    # Leave files/license.lic as empty
    file = File.open("files/license.lic", "w") { |f| f.write "" }
  end
end
