require 'test_helper'

#
# Test users controller
#
class UsersControllerTest < ActionController::TestCase
  include Warden::Test::Helpers
  Warden.test_mode!
  include Devise::TestHelpers

  context 'non admin user' do
    setup do
      @user = users(:user_domain)
      @another_user =
        User.where(sensor_id: @user.domains_with_permission.map(&:id)).sample
      sign_in @user
    end

    %i(edit show update).each do |method|
      should "be redirected when try to #{method} an user" do
        get method, id: @another_user.id
        assert_response 404
      end
    end

    should 'be redirected when try to destroy an user' do
      get :destroy, id: @another_user.id
      assert_redirected_to root_path
    end

    should 'be redirected when try to create an user' do
      get :new
      assert_redirected_to root_path
    end

    should 'be redirected when try to see the users list' do
      get :index
      assert_redirected_to root_path
    end
  end

  context 'admin domain' do
    setup do
      @user = users(:admin_domain)
      @another_user = users(:user_domain)
      @service_provider_2 = sensors(:service_provider_2)
      sign_in @user
    end

    should 'not update another user with a non permitted top domain' do
      patch :update, id: @another_user.id, user: { sensor_id: @service_provider_2.id }
      assert_not_equal @service_provider_2.id, @another_user.reload.sensor_id
      assert_redirected_to users_path
    end

    should 'update admin privilege to another user under his domain' do
      patch :update, id: @another_user.id, user: { admin: true, sensor_id: @user.sensor_id }
      assert_equal true, @another_user.reload.admin
      assert_redirected_to users_path
    end
  end

  context 'admin user' do
    setup do
      @user = users(:admin)
      @another_user = users(:user_domain)
      sign_in @user
    end

    should 'see user list' do
      get :index
      assert_response :success
    end

    should 'create a new user' do
      get :new
      assert_response :success
    end

    should 'edit itself registration' do
      get :edit, id: @user.id
      assert_redirected_to edit_user_registration_path
    end

    should 'edit an user' do
      get :edit, id: @another_user.id
      assert_response :success
    end

    should 'be updated' do
      patch :update, id: @user.id, user: { name: 'Name', sensor_id: @user.sensor_id }
      assert_equal 'Name', @user.reload.name
      assert_redirected_to users_path
    end

    should 'not be updated without name' do
      patch :update, id: @user.id, user: { name: '', sensor_id: @user.sensor_id  }
      assert_not_equal '', @user.reload.name
      assert_redirected_to edit_user_path(@user)
    end

    should 'destroy an user' do
      assert_difference('User.count', -1) do
        delete :destroy, id: @another_user.id
      end
      assert_redirected_to users_path
      assert flash[:notice].include? 'User successfully deleted'
    end

    should 'list users to share' do
      get :list
      assert_response :success
    end

    should 'load more users into the table' do
      xhr :get, :load_more, format: :js, page: 1
      assert_response :success
    end

    should 'see users by a valid letter' do
      xhr :get, :letter_search, format: :js, id: 'A'
      assert_response :success
    end

    should 'not see users by a non valid letter' do
      xhr :get, :letter_search, format: :js, id: '4'
      assert_redirected_to users_path
    end

    should 'search users with the search box' do
      xhr :get, :search, format: :js, search: 'Name'
      assert_response :success
    end

    should 'recreate api key for an user' do
      xhr :patch, :recreate_api_key, format: :js
      assert_response :success
    end

    should 'see time selector' do
      xhr :get, :time_selector, format: :js
      assert_response :success
    end

    should 'save_columns' do
      xhr :patch, :save_columns, format: :js, url: { controller: 'traffics' }
      assert_response :success
    end
  end
end
