require 'test_helper'

#
# Check the sessions.
#
class DashboardsTest < ActionDispatch::IntegrationTest
  # To login users
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'Creating dashboard' do
    # Get first user
    setup do
      @admin = users(:admin)
      @new_dashboard = 'AnotherDashboard'
      # Login him
      login_as(@admin, scope: :user)
    end

    should 'fail when the dashboard name is empty' do
      fail_create_dashboard
    end

    should 'fail when the dashboard name is empty (mobile)' do
      Capybara.current_driver = :mobile
      fail_create_dashboard('_mobile')
      Capybara.use_default_driver
    end

    should 'create a dashboard' do
      visit '/dashboard'
      click_link('Default', match: :first)
      link = page.find_link('Add Dashboard')
      assert(link, 'The user must have a link to create a dashboard')
      # Click on the link
      click_link('Add Dashboard')
      # Create the dashboard
      within('#new_dashboard') do
        fill_in 'dashboard[name]', with: @new_dashboard
      end
      click_button 'Create Dashboard'
      # Check the redirection.
      assert_equal(dashboards_path, page.current_path,
                   'When create a dashboard, user must redirect to new'\
                   ' dashboard')
      # Check if the new dashboard is created
      link = page.find_link(@new_dashboard, match: :first)
      assert(link, 'New dashboard must be presento on the menu')
    end
  end

  context 'Sharing a dashboard' do
    # Get first user
    setup do
      @admin = users(:admin)
      # Login him
      login_as(@admin, scope: :user)
    end

    should 'Share with a domain' do
      dashboard = @admin.dashboards.first
      dashboard.dashboard_sensors.destroy_all
      visit edit_dashboard_path(dashboard)
      # Click on modal
      find('.share-domains .panel-button', match: :first).click
      # Get a random domain
      domain = @admin.sensor.children.where(type: Sensor::DOMAIN).sample
      button_css = "tr[data-tt-id=\"#{domain.id}\"] td.last .btn"
      find(button_css).click
      click_link 'Add the domains', wait: 0.1
      # Check if it's added
      page.has_content?(domain.name)
      page.has_selector?('.share-domains table tbody tr', count: 1)
    end

    should 'Share with an user' do
      dashboard = @admin.dashboards.first
      dashboard.dashboard_users.destroy_all
      visit edit_dashboard_path(dashboard)
      # Click on modal
      find('.share-users .panel-button', match: :first).click
      # Get a random domain
      user = User.last
      user_css = ".users_selector .user[data-user-id=\"#{user.id}\"]"
      find(user_css).click
      click_link 'Return to form', wait: 0.1
      # Check if it's added
      page.has_content?(user.name)
      page.has_selector?('.share-users table tbody tr', count: 1)
    end
  end

  context 'A dashboard' do
    setup do
      @admin = users(:admin)
      @col_dashboard = dashboards(:admin_col_dashboard)
      @abs_dashboard = dashboards(:admin_abs_dashboard)
      # Login him
      login_as(@admin, scope: :user)
    end

    should 'be imported' do
      visit "/dashboard/#{@col_dashboard.id}"
      find('#dashboard-selector', match: :first).click
      wait_for_ajax page
      click_link 'Import Dashboard'
      wait_for_ajax page
      assert(page.has_css?('.dashboard_file', count: 1))
    end

    should 'clone an existing widget to a fixed dashboard from a columnar one' do
      visit "/dashboard/#{@col_dashboard.id}"
      widget = find('.widget', match: :first)
      widget.hover
      widget.find('.widget-toolbar').click
      click_link 'Clone'
      wait_for_ajax page
      fill_in('widget[name]', with: 'TestClone')
      page.select(@abs_dashboard.name, from: 'widget[parent_id]')
      click_button 'Clone widget'

      n_widgets = @abs_dashboard.widgets.size
      visit "/dashboard/#{@abs_dashboard.id}"
      wait_for_ajax page
      assert_equal(n_widgets + 1, page.all('.widget').size,
                  'Unable to find the cloned widget')
    end

    should 'clone an existing widget to a columnar dashboard from a fixed one' do
      visit "/dashboard/#{@abs_dashboard.id}"
      widget = find('.widget', match: :first)
      widget.hover
      widget.find('.widget-toolbar').click
      click_link 'Clone'
      wait_for_ajax page
      fill_in('widget[name]', with: 'TestClone')
      page.select(@col_dashboard.name, from: 'widget[parent_id]')
      click_button 'Clone widget'

      n_widgets = @col_dashboard.widgets.size
      visit "/dashboard/#{@col_dashboard.id}"
      wait_for_ajax page, true
      assert_equal(n_widgets + 1, page.all('.widget').size,
                  'Unable to find the cloned widget')
    end

    should 'clone a widget to itself' do
      visit "/dashboard/#{@col_dashboard.id}"

      widget = find('.widget', match: :first)
      widget.hover
      widget.find('.widget-toolbar').click

      click_link 'Clone'
      wait_for_ajax page
      fill_in('widget[name]', with: 'TestClone')
      page.select(@col_dashboard.name, from: 'widget[parent_id]')
      click_button 'Clone widget'

      wait_for_ajax page
      assert(page.has_content?('TestClone'), 'Unable to find the cloned widget')
    end
  end
end

#
# Try to create a dashboard
#
def fail_create_dashboard(screen = nil)
  visit '/dashboard/new'
  click_button 'Create Dashboard'
  # Check if the new dashboard is created
  error = page.has_css?('.dashboard_name.error')
  assert(error, 'Dashboard form must show an error beacuse name is not'\
                ' present')
end
