require 'test_helper'

class ReportsTest < ActionDispatch::IntegrationTest
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'an user' do
    setup do
      @user = users(:user1)
      @another_user = users(:admin)

      # Login him
      login_as(@user, scope: :user)
      # Stub chef
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    ### Reports
    should 'see it\'s report' do
      visit '/reports'
      assert_equal(@user.reports.size, all('#reports table tbody tr').size,
                   'The user should see all his reports')
    end

    should 'not edit someone else\'s report' do
      visit "/reports/#{@another_user.reports.first.id}/edit"
      assert page.has_content?('You don\'t have enough privileges'),
             'Forbidden alert was expected but not found'
    end

    should 'edit it\'s report' do
      report = @user.reports.first
      visit "/reports/#{report.id}/edit"
      assert page.has_content?("Edit #{report.name}"),
             'Edition page not detected'
    end

    should 'share it\'s report with another user' do
      report = @user.reports.first
      visit "/reports/#{report.id}/edit"
      page.find('a[data-target="#add_user_modal"]').click
      wait_for_ajax page, true
      page.find('#add_user_modal .users_selector .user', match: :first).click
      wait_for_ajax page
      assert_equal(
        report.report_users.size + 1,
        page.all('table#users_table tbody tr').size,
        'Number of users the report has been shared with is unexpected')
    end

    should 'share it\'s report with another domain' do
      report = @user.reports.first
      visit "/reports/#{report.id}/edit"
      page.find('a[data-target="#add_domain_modal"]').click
      wait_for_ajax page, true

      assert page.has_content?('Add domains'), 'Domain\'s modal not detected'
      assert page.has_css?('.modal-dialog'), 'Domain\'s modal not detected'

      page.find('.modal-dialog .table tbody tr td.last a', match: :first).click
      page.find('.modal-dialog .modal-footer .btn.btn-primary').click
      wait_for_ajax page
      assert_equal(
        report.report_sensors.size + 1,
        all('table#sensors_table tbody tr').size,
        'Number of domains the report has been shared with is unexpected')
    end

    should 'edit it\'s report composition' do
      report = @user.reports.first
      visit "/reports/#{report.id}/composition"

      assert page.has_content?("Composition #{report.name}"),
             'Composition page not detected'
      assert page.has_css?('.page-toolbar .add-block.btn', visible: false),
             'Add block button not found'
      assert page.has_css?('.new-page.btn'), 'Add page button not found'

      page.find('.new-page.btn').click
      wait_for_ajax page
      assert_equal(
        page.all('.page-content').size, report.reload.number_pages,
        'The number of pages for the report is unexpected')
    end

    should 'delete the reports' do
      visit reports_path
      assert page.has_css?('table tbody .btn[data-method="delete"] .fa-trash'),
             'Delete button not found'

      n_reports = page.all('#reports table tbody tr').size
      page.all('table tbody .btn[data-method="delete"]').first.click
      wait_for_ajax page
      assert_equal(
        page.all('#reports table tbody tr').size,
        n_reports - 1,
        'Unexpected number of reports showing')
    end

    should 'be able to create a new report' do
      visit '/reports/new'
      assert page.has_css?('.report-form'), 'Report form not found'
      assert page.has_css?('.panel-actions .btn[type=submit]'),
             'Submit button not found'
    end

    should 'create a new report' do
      visit '/reports/new'
      fill_in 'report[name]', with: 'Test'
      click_button 'Create Report'

      assert_equal "/reports/#{Report.order('created_at').last.id}/composition",
                   page.current_path,
                   'The current path is unexpected'
    end
  end

end
