require 'test_helper'

class SensorsTest < ActionDispatch::IntegrationTest
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'not logged user' do
    should 'be redirected to login when access to sensors' do
      visit '/sensors/tree'
      assert_equal(new_user_session_path, page.current_path,
                   'The user must be redirected to login')
    end
  end

  context 'admin user' do
    # Get first user
    setup do
      @user = users(:admin)
      @domain = sensors(:deployment)
      @second_domain = sensors(:campus)
      # Login him
      login_as(@user, scope: :user)
      # Stub chef
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
      SensorsController.any_instance.stubs(:initialize_namespace_rules).returns(nil)
    end

    should 'search sensor in tree' do
      visit '/sensors/tree'
      fill_in('search', with: @second_domain.name)
      page.execute_script("$('form.sensor-search').submit()")
      assert(page.has_css?("#tree tbody tr[data-tt-id='#{@second_domain.id}'].selected",
        count: 1), 'User must find a sensor using search bar')
    end

    should 'see root sensors (domain, flow, ips) and can delete them' do
      visit '/sensors/tree'
      # There are 3 root elements
      assert(page.has_css?('#tree tbody tr[data-tt-parent-id=""]',
             count: Sensor.root.children.size),
             'The user must see root sensors')
      assert(page.has_css?('#tree tbody tr td .btn.sensor-delete',
             count: Sensor.root.children.size),
             'The admin can delete sensors')
    end

    should 'edit a sensor' do
      get edit_sensor_path(@domain)
      assert_response(200, 'A user with privileges must can see edit form')
    end
  end

  context 'non admin user' do
    # Get first user
    setup do
      @user = users(:user_domain)
      @domain = sensors(:deployment)
      # Login him
      logout(:user)
      login_as(@user, scope: :user)
      # Stub chef
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'not see its root sensor (domain) and can\'t delete it' do
      visit '/sensors/tree'
      # There are 3 root elements
      assert(!page.has_css?('#tree tbody tr[data-tt-parent-id=""]', count: 1),
             'The user must not see its root sensor (domain)')
      assert(!page.has_css?('#tree tbody tr td .btn.sensor-delete'),
             'An user without privileges can\'t delete a sensor')
    end

    should 'edit a sensor' do
      get edit_sensor_path(@domain)
      assert_redirected_to(sensors_path, 'A user without privileges '\
                                         'can\'t see edit form')
      assert_match /You don't have enough privileges to do that./, flash[:alert]
    end
  end
end
