require 'test_helper'

#
# Check users view to list, edit, create and destroy users.
#
class UsersTest < ActionDispatch::IntegrationTest
  # To login users
  include Warden::Test::Helpers
  Warden.test_mode!

  context 'A super admin user' do
    # Get first user
    setup do
      @admin = users(:admin)
      @password = 'redborder'
      # Login him
      login_as(@admin, scope: :user)
      User.any_instance.stubs(:send_creation_instructions).returns(true)
    end

    should 'list all users' do
      visit '/users'
      assert(page.has_css?('#users table tbody tr.user', count: User.count),
             'Admin user can see all users')
    end

    should 'create an user' do
      # New user params
      user = user_params
      # Go to create
      visit '/users'
      click_link('New User')
      # Create the user
      within('#new_user') do
        fill_in 'user[name]', with: user.name
        fill_in 'user[login]', with: user.login
        fill_in 'user[email]', with: user.email
        fill_in 'user[password]', with: user.password
        fill_in 'user[password_confirmation]', with: user.password
      end
      # Create
      click_button 'Create'
      # Check the redirection.
      assert_equal(users_path, page.current_path,
                   'When create an user, we must be redirected to users list')
      # Check user created
      assert(page.has_css?('#users table tbody td.name', text: user.name),
             'The user must be added in the list')
    end

    # Add errors to form
    should 'add an error to the form when a param is not correct' do
      user = user_params
      visit new_user_path
      form_with_error('User', '#new_user', 'Create', 'name') do
        fill_in 'user[login]', with: user.login
        fill_in 'user[email]', with: user.email
        fill_in 'user[password]', with: user.password
        fill_in 'user[password_confirmation]', with: user.password
      end
    end

    should 'check format of email' do
      user = user_params
      visit new_user_path
      form_with_error('User', '#new_user', 'Create', 'email') do
        fill_in 'user[login]', with: user.login
        fill_in 'user[email]', with: 'asdasdasdasd'
        fill_in 'user[password]', with: user.password
        fill_in 'user[password_confirmation]', with: user.password
      end
    end

    should 'check longitude of password' do
      user = user_params
      visit new_user_path
      form_with_error('User', '#new_user', 'Create', 'password') do
        fill_in 'user[login]', with: user.login
        fill_in 'user[login]', with: user.login
        fill_in 'user[email]', with: user.email
        fill_in 'user[password]', with: 'aaaa'
        fill_in 'user[password_confirmation]', with: 'aaaa'
      end
    end

    should 'edit an user' do
      user = users(:user3)
      new_name = 'NewUser3'
      visit '/users'
      find("#users table tbody tr.user[data-user-id='#{user.id}'] .edit-user")
        .click
      # Edit the name
      within('form.edit_user') do
        fill_in 'user[name]', with: new_name
      end
      # Edit
      click_button 'Update'
      wait_for_ajax(page)
      assert_equal(users_path, page.current_path,
                   'When update an user, we must be redirected to users list')
      assert(page.has_css?('#users table tbody td.name', text: new_name),
             'The user must be updated')
    end

    should 'recreate API key of an user' do
      user = users(:user3)
      last_api = user.authentication_token
      visit '/users'
      find("#users table tbody tr.user[data-user-id='#{user.id}'] .edit-user")
        .click
      click_link 'Recreate API Access Key'
      wait_for_ajax(page)
      user.reload
      assert_not_equal(last_api, user.authentication_token,
                       'The user must change its API key')
    end

    should 'edit himself' do
      visit '/users/edit'
      new_login = 'administrator'
      within('#edit_user') do
        fill_in 'user[login]', with: new_login
        fill_in 'user[current_password]', with: @password
      end
      click_button 'Update'
      assert_equal(new_login, @admin.reload.login,
                   'Login must be updated')
    end

    should 'can not delete himself' do
      visit '/users'
      page.has_css?("#users table tbody tr.user[data-user-id='#{@admin.id}']"\
                    ' .destroy-user.disabled')
    end

    should 'delete an user' do
      visit '/users'
      user = users(:user2)
      find("#users table tbody tr.user[data-user-id='#{user.id}'] .destroy-user")
        .click
      wait_for_ajax(page)
      assert(page.has_no_css?('#users table tbody td.name', text: user.name),
             'The user must be deleted')
    end
  end

  context 'A non admin user' do
    # Get first user
    setup do
      @user = users(:user_domain)
      @password = 'redborder'
      # Login him
      login_as(@user, scope: :user)
      User.any_instance.stubs(:send_creation_instructions).returns(true)
    end

    should 'list all users' do
      visit '/users'
      assert(page.has_css?(
               '#users table tbody tr.user',
               count: User.where(sensor_id: @user.domains_with_permission.map(&:id)).count),
             'An user can see only users in own top domain')
    end

    should 'can not see new user button' do
      # Go to create
      visit '/users'
      has_no_css? '.nav.pull-right.navbar-buttons.li.a', text: 'New User'
    end

    should 'see all remove buttons as disabled' do
      visit '/users'
      page.all(:css, '.destroy_user').each do |el|
        el[:class].include? 'disabled'
      end
    end

    should 'see all edit buttons as disabled except himself' do
      visit '/users'
      page.all(:css, '#users table tbody tr.user').each do |el|
        if el['data-user-id'].to_i == @user.id
          assert_not(el.find('.edit-user')[:class].include?('disabled'),
                     'The user can edit his own profile')
        else
          assert(el.find('.edit-user')[:class].include?('disabled'),
                 'User can not edit another user profile')
        end
      end
    end

    should 'edit himself' do
      visit '/users'
      find("#users table tbody tr.user[data-user-id='#{@user.id}'] .edit-user")
        .click
      new_login = 'new_login_for_user'
      within('#edit_user') do
        fill_in 'user[login]', with: new_login
        fill_in 'user[current_password]', with: @password
      end
      click_button 'Update'
      assert_equal(new_login, @user.reload.login,
                   'Login must be updated')
    end

    should 'edit its own password' do
      visit '/users/edit'
      find("a.edit_password_button").click
      page.has_selector?('#edit_password', visible: true)
      new_pass = 'qwasdqwd'
      within('#edit_user') do
        fill_in 'user[password]', with: new_pass
        fill_in 'user[password_confirmation]', with: new_pass
        fill_in 'user[current_password]', with: @password
      end
      click_button 'Update'
      assert_equal(new_user_session_path, page.current_path,
                   'User must be redirected to login when change password')
    end
  end

  #
  # Return params for a new user.
  #
  def user_params
    Struct.new(:name, :login, :email, :password).new(
      'new_user',
      'new_user_login',
      'new_user@user.net',
      '1234asdasdasd'
    )
  end
end
