require 'test_helper'

#
# DataSources are the models to retrieve the data from Druid. All DataSource
# must implements an interface methods defined on DruidSource. We must test it.
#
class FilterTest < ActiveSupport::TestCase
  #
  # Setup the filter
  #
  context 'A filter' do
    setup do
      # Create the var with the data to test
      @data = {
        src: {
          column: 'src',
          values: ['98.11.22.55']
        },
        dst: {
          column: 'dst',
          values: ['80.2.11.4', '90.1.44.3']
        },
        to_parse: "{\"0\":{\"column\":\"src\",\"operator\":\"is\",\"value\":"\
                  "\"10.0.30.58\",\"enabled\":true},\"1\":{\"column\":\"src\""\
                  ",\"operator\":\"is\",\"value\":\"10.0.30.152\",\"enabled\""\
                  ":true},\"2\":{\"column\":\"application_id_name\",\"operato"\
                  "r\":\"is\",\"value\":\"http\",\"enabled\":true},\"3\":{\"c"\
                  "olumn\":\"application_id_name\",\"operator\":\"contains\","\
                  "\"value\":\"google\",\"enabled\":true}}",
        serialized: { 
                      src: {
                        "10.0.30.58" => { operator: "is" },
                        "10.0.30.152" => { operator: "is" } 
                      },
                      application_id_name: {
                        "http" => { operator: "is" },
                        "google" => { operator: "contains" }
                      }
                    }
      }
      # Select a data source available for testing. I prefer flow.
      source = 'ips' if Sensor::IPS
      source = 'flow' if Sensor::FLOW
      # In it the filters and data sources
      data_source = Redborder::DruidSource.get(source: source)
      @filter = Redborder::Filter.new(data_source)
    end
    #
    # Add Commons filters
    #
    should 'Add a common filter to an empty previous filters' do
      @filter.remove_all
      data = @data[:src]
      sol = { src: { '98.11.22.55' => { operator: 'is' } } }
      # Add the filter
      @filter.add(data[:column], data[:values])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                 'expected output.')
    end

    should 'Add a common filter with a previous filter (another key)' do
      @filter.remove_all
      data = @data[:dst]
      prev = { src: { '98.11.22.55' => { operator: 'is' } } }
      @filter.replace(shared: prev)
      sol = {
              src: { '98.11.22.55' => { operator: 'is' } },
              dst: { '80.2.11.4' => { operator: 'is' },
                     '90.1.44.3' => { operator: 'is' } }
            }
      # Add the filter
      @filter.add(data[:column], data[:values])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                 'expected output.')
    end

    should 'Add a common filter with a previous filter (same is_not)' do
      @filter.remove_all
      data = @data[:src]
      prev = { src: { '98.11.22.55' => { operator: 'is_not' } } }
      @filter.replace(shared: prev)
      sol = {
              src: { '98.11.22.55' => { operator: 'is' } }
            }
      # Add the filter
      @filter.add(data[:column], data[:values])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                 'expected output.')
    end

    #
    # Contain Commons filters
    #
    should 'Add a contain filter. It\'s the only in the filter hash' do
      @filter.remove_all
      sol = { dst: { '11.' => { operator: 'contains' } } }
      # Add the filter
      @filter.contains('dst', ['11.'])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                     'expected output.')
    end

    #
    # Delete Commons filters
    #
    should 'Delete a common filter. It\'s the only in the filter hash' do
      @filter.remove_all
      data = @data[:src]
      prev = { src: { '98.11.22.55' => { operator: 'is' } } }
      @filter.replace(shared: prev)
      # Add the filter
      @filter.remove(data[:column], data[:values])
      # It must coincide with sol
      assert_equal({}, @filter.get, 'Filters don\'t coincide with the ' \
                                    'expected output.')
    end

    should 'Delete a common filter. There are another filter with same key' do
      @filter.remove_all
      data = @data[:src]
      prev = { src: { '98.11.22.55' => { operator: 'is' },
                      '90.1.44.3' => { operator: 'is' } } }
      sol = { src: { '90.1.44.3' => { operator: 'is' } } }
      # Add the filter
      @filter.replace(shared: prev)
      @filter.remove(data[:column], data[:values])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                     'expected output.')
    end

    should 'Delete a common filter. There are another filter' do
      @filter.remove_all
      data = @data[:src]
      prev = { src: { '98.11.22.55' => { operator: 'is' } },
               dst: { '90.1.44.3' => { operator: 'is' } } }
      sol = { dst: { '90.1.44.3' => { operator: 'is' } } }
      @filter.replace(shared: prev)
      # Add the filter
      @filter.remove(data[:column], data[:values])
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                 'expected output.')
    end

    should 'Count filters' do
      data = @data[:src]
      prev = { src: { '98.11.22.55' => { operator: 'is' } },
               dst: { '90.1.44.3' => { operator: 'is' } } }
      @filter.replace(shared: prev)
      # Add the filter
      @filter.remove(data[:column], data[:values])
      @filter.add(data[:column], data[:values])
      @filter.add(:dst, ['5.5.5.5', '12.23.3.4', '23.4.4.4'])
      @filter.contains(:dst, ['5.5.'])
      @filter.add(:l4_proto, ['TCP'])
      @filter.remove(:l4_proto)
      # It must coincide with sol
      assert_equal(6, @filter.count, 'The number of filters don\'t coincide ' \
                                     'with expected output.')
    end

    should 'Parse filters from string' do
      data = @data[:to_parse]
      sol = @data[:serialized]
      # Parse the data and remove it
      @filter.parse(data, true)
      # It must coincide with sol
      assert_equal(sol, @filter.get, 'Filters don\'t coincide with the ' \
                                     'expected output.')
    end

    should 'Serialize filters' do
      data = @data[:serialized]
      # Load filters.
      @filter.replace(shared: data)
      # Serialize the data as JSON string
      serialized = @filter.serialize(true)
      # Load serialized data
      @filter.parse(serialized, true)
      # Check if init filters and serialized -> parsed coincide    
      assert_equal(data, @filter.get, 'Serialized filters don\'t coincide ' \
                                      'with the expected output.')
    end

    should 'Works with time filters' do
      # Load filters.
      @filter.remove_all
      # Add a time filter
      @filter.add(:start_time, [(Time.now - 3.hours).to_i])
      @filter.add(:end_time, [Time.now.to_i])
      # Check if there are time filters   
      assert(@filter.time_filters?, 'Filters must detect there are time ' \
                                    'filters applied.')

      # Delete one time filter, the other must deleted too.
      @filter.remove(:start_time)
      # Check if init filters and serialized -> parsed coincide    
      assert_equal({}, @filter.get, 'End_time must be deleted when delete ' \
                                    'start_time.')
    end

    should 'Checks the existence of filters' do
      # Clean filters and initialize
      @filter.remove_all
      @filter.add(@data[:src][:column], @data[:src][:values])
      @filter.add(@data[:dst][:column], @data[:dst][:values])

      # Check columns
      assert(@filter.exist_any?(:src), 'SRC Address column exist in filters')
      assert(@filter.exist_any?(:dst), 'DST Address column exist in filters')
      assert(!@filter.exist_any?(:l4_proto), "L4 Proto column doesn\'t exist "\
                                             'in filters.')
      # Check columns and values
      assert(@filter.exist?(:src, @data[:src][:values].first),
              'SRC Address and this value exist in the filters')
      assert(@filter.exist?(:dst, @data[:dst][:values].last),
              'DST Address and this value exist in the filters')
      assert(!@filter.exist?(:src, @data[:dst][:values].last),
              'SRC Address exist, but this value don\'t exist in the filters')
      assert(!@filter.exist?(:l4_proto, @data[:dst][:values].last),
              'L4 proto and this value don\'t exist in the filters')
    end
  end
end
