require File.expand_path('../data/randomizer', __FILE__)
require 'json'

#
# This class allow the user to stub calls on to druid
#
class DruidStub
  include Randomizer

  # Steps
  STEPS = {
    pt1m: 60,
    pt2m: 120,
    pt5m: 300,
    fifteen_minute: 900,
    thirty_minute: 1800,
    hour: 3600,
    pt2h: 7200,
    pt8h: 28800,
    pt24h: 86400,
    pt120h: 432000,
    pt168h: 604800,
    pt720h: 2592000
  }

  # Time format
  TIME_FORMAT = '%Y-%m-%dT%H:%M:%S:00.000Z'

  attr_accessor :query
  attr_reader :stub_data

  #
  # Initialize the mock Druid. This method load the Hash from stub data to
  # generate the correct data.
  #
  def initialize
    @stub_data = JSON.parse(
      File.read(
        Rails.root.join('test', 'mocha', 'druid', 'data', 'stub_data_type.json')
      )
    )
  end

  #
  # Return a mock druid to query
  #
  # == Parameters:
  # data_source::
  #   String with the datasource
  #
  # == Returns:
  # DruidStubQuery object
  #
  def query(data_source)
    DruidStubQuery.new(data_source)
  end

  #
  # Return random data with the query provided
  #
  # == Parameters:
  # query::
  #   A DruidStubQuery object.
  #
  # == Returns:
  # Random data
  #
  def send(query)
    @query = query
    build_data
  end

  #
  # Build the data to return based on data source and the options sent
  # to DruidStubQuery object
  #
  # == Returns:
  # Stub data of Druid
  #
  def build_data
    # Fill base elements
    data = []
    step = calc_step

    case @query.store_query_type
    when :timeseries, :topN, :groupBy
      num_elements(step).times do
        time = Time.at(@query.start_time + step)
        data << fill(time,
                     @query.properties[:aggregations].first[:name],
                     @query.properties[:postAggregations])
      end
    end

    data
  end

  #
  # Return the step based on the granularity
  #
  def calc_step
    if @query.properties[:granularity] == 'all'
      0
    else
      STEPS[@query.properties[:granularity][:period].to_sym]
    end
  end

  #
  # Return the number of elements to stub
  #
  def num_elements(step)
    return 1 if step.zero?
    (@query.end_time.to_i - @query.start_time.to_i) / step
  end

  #
  # Fill the main aggregation with random elements
  #
  def fill(time, aggregation, post_aggregations)
    # Check postaggregations
    post_aggregations = [] if post_aggregations.nil?

    # Get the value for the aggregation
    agg_value = value(aggregation)
    result = { aggregation => agg_value }

    if @query.properties[:dimension]
      dim = @query.properties[:dimension]
      result = [{ aggregation => agg_value,
                  @query.properties[:dimension].to_s => value(dim) }]
    elsif @query.properties[:dimensions]
      result = { aggregation => agg_value }
      @query.properties[:dimensions].each do |dimension|
        result[dimension.to_s] = value(dimension)
      end
    end

    # Load postaggregations
    data = {
      timestamp: time.strftime(TIME_FORMAT),
      result: result
    }

    # now, we only will work with static post_aggregations
    post_aggregations.each do |post_agg|
      if data[:result].is_a?(Array)
        data[:result].first[post_agg[:name]] =
          fill_post(agg_value, post_agg[:fn], post_agg[:fields].last[:value])
      else
        data[:result][post_agg[:name]] =
          fill_post(agg_value, post_agg[:fn], post_agg[:fields].last[:value])
      end
    end

    DruidStubResponse.new(data)
  end

  #
  # Get a random value based on dimension/aggregation
  #
  # == Parameters:
  # key::
  #   String with the aggregation, dimension... to get random data
  #
  # == Returns:
  # Integer or string with random data
  #
  def value(key)
    # Get the main value
    case @stub_data[key.to_s]
    when 'integer'
      random_integer
    when 'float'
      random_float
    when 'ip'
      random_ip
    when 'mac'
      random_mac
    when 'signature'
      random_signature
    else
      # By default...
      random_string
    end
  end

  #
  # Fill post_aggregation elements
  #
  def fill_post(value, operator, constant)
    # Eval the math elements
    eval("#{value}.to_f #{operator} #{constant}")
  rescue SyntaxError, StandardError
    random_integer
  end

  #
  # Class with the base result
  #
  class DruidStubResponse
    (instance_methods + private_instance_methods).each do |method|
      unless method.to_s =~ /^(__|object_id|initialize)/
        undef_method method
      end
    end

    attr_reader :timestamp, :row

    def initialize(data)
      @timestamp = data[:timestamp]
      @row = data[:result]
    end

    def method_missing(name, *args, &block)
      @row.send name, *args, &block
    end

    def to_s
      "#{@timestamp.to_s}:#{@row.to_s}"
    end

    def inspect
      "#{@timestamp.inspect}:#{@row.inspect}"
    end
  end

  #
  # Class to mock Druid query objects
  #
  class DruidStubQuery
    attr_reader :data_source, :properties, :store_query_type, :start_time,
                :end_time

    def initialize(ds)
      @data_source = ds
      @properties = {}
    end

    def interval(start_time, end_time)
      @properties[:intervals] = ["#{start_time.iso8601}/#{end_time.iso8601}"]
      @start_time = start_time
      @end_time = end_time
    end

    def query_type(type)
      @store_query_type = type
    end

    def topn(dimension, metric, threshold)
      query_type(:topN)
      @properties[:dimension] = dimension
      @properties[:metric] = metric
      @properties[:threshold] = threshold
    end

    def group_by(*dimensions)
      query_type(:groupBy)
      @properties[:dimensions] = dimensions.flatten
      self
    end
  end
end
