require 'test_helper'

#
# Testing for Alarm model. It should test validations, and basic operations
# like create, update and destroy.
#
class AlarmTest < ActiveSupport::TestCase
  # Test relations
  should belong_to :user
  should belong_to :severity

  # Presence
  %w(name interval time_window product_type severity).each do |col|
    should validate_presence_of col.to_sym
  end

  # Inclusion
  should validate_inclusion_of(:interval).in_range(Alarm::INTERVAL_RANGE)
  should validate_inclusion_of(:time_window).in_range(Alarm::TIME_WINDOW_RANGE)
  %w(multiplier_upper multiplier_bottom).each do |col|
    should validate_inclusion_of(col.to_sym).in_range(Alarm::MULTIPLIER_RANGE)
  end

  context 'An alarm' do
    setup do
      @alarm = alarms(:one)
    end
  
    should 'be enabled and disabled' do
      @alarm.disable
      assert_not @alarm.enabled, 'The alarm must be disabled'
      @alarm.enable
      assert @alarm.enabled, 'The alarm must be enbled'
    end      

    should 'be created and valid with upper_limit' do
      alarm = Alarm.new(alarm_params_upper)
      assert(alarm.valid?, 'The alarm must be valid')
      assert(alarm.upper_limit_enabled?,
             'Upper limit must be enabled')
      assert(alarm.save, 'The alarm must be saved')
    end

    should 'be created and valid with bottom_limit' do
      alarm = Alarm.new(alarm_params_bottom)
      assert(alarm.valid?, 'The alarm must be valid')
      assert(alarm.bottom_limit_enabled?,
             'Bottom limit must be enabled')
      assert(alarm.save, 'The alarm must be saved')
    end

    should 'fail without limit specified' do
      alarm = Alarm.new(alarm_params)
      assert_not(alarm.valid?, 'The alarm must be invalid')
      assert_equal(:upper_limit, alarm.errors.keys.first,
                   'An error must be added when there aren\'t any limit '\
                   'present')
    end
    should 'be valid with a Slack Webhook' do
      alarm = Alarm.new(alarm_params_slack)
      assert(alarm.valid?, 'The alarm must be valid')
      assert(alarm.save, 'The alarm must be saved')
    end
    should 'fail if notify to Slack is active but webhook field is empty' do
      alarm = Alarm.new(alarm_params_upper.merge(notify_to_slack: 1))
      assert_not(alarm.valid?, 'The alarm must be invalid')
      assert_equal(:slack_webhook, alarm.errors.keys.first,
                   'An error must be added when there is no Slack webhook '\
                   'and notify to Slack is active')
      assert_equal('must be present', alarm.errors[:slack_webhook].first,
                   'An error must be added when there is no Slack webhook '\
                   'and notify to Slack is active')
    end

    should 'fail if a Slack Webhook is invalid' do
      alarm = Alarm.new(alarm_params_slack_invalid)
      assert_not(alarm.valid?, 'The alarm must be invalid')
      assert_equal(:slack_webhook, alarm.errors.keys.first,
                   'An error must be added when there are a wrong email '\
                   'in the list')
      assert_equal('is not correctly formatted',
                   alarm.errors[:slack_webhook].first,
                   'An error must be added when there aren\'t email list '\
                   'and notify to email is active')
    end
    should 'be valid with a email list' do
      alarm = Alarm.new(alarm_params_email)
      assert(alarm.valid?, 'The alarm must be valid')
      assert(alarm.save, 'The alarm must be saved')
    end

    should 'fail if notify to email is active but email list is empty' do
      alarm = Alarm.new(alarm_params_upper.merge(notify_to_email: 1))
      assert_not(alarm.valid?, 'The alarm must be invalid')
      assert_equal(:email_list, alarm.errors.keys.first,
                   'An error must be added when there aren\'t email list '\
                   'and notify to email is active')
      assert_equal('must be present', alarm.errors[:email_list].first,
                   'An error must be added when there aren\'t email list '\
                   'and notify to email is active')
    end

    should 'fail if an email is invalid' do
      alarm = Alarm.new(alarm_params_email_invalid)
      assert_not(alarm.valid?, 'The alarm must be invalid')
      assert_equal(:email_list, alarm.errors.keys.first,
                   'An error must be added when there are a wrong email '\
                   'in the list')
      assert_equal('is not correctly formatted',
                   alarm.errors[:email_list].first,
                   'An error must be added when there aren\'t email list '\
                   'and notify to email is active')
    end
  end

  context 'Invalid alarm' do
    should 'fail if product type is wrong' do
      a = alarms(:four)
      a.valid?
      assert_match(/is not included in the list/,
                   a.errors[:product_type].join(','),
                   'Invalid product type validation')
    end

    should 'fail if monitor tab is wrong' do
      a = alarms(:five)
      a.valid?
      assert_match(/is not included in the list/,
                   a.errors[:monitor_tab].join(','),
                   'Invalid monitor tab validation')
    end
  end

  context 'Firing alarms' do
    setup do
      @alarm = alarms(:one)
    end

    should 'be fired' do
      @alarm.launch
      assert(@alarm.fired?, 'Alarm was fired')
      assert(@alarm.last_fired?, 'Alarm was fired')
    end

    should 'disabled alarms musn\'t be checked' do
      @alarm.disable
      assert_not(@alarm.must_be_checked?,
                 'Disabled alarms musn\'t be checked')
    end

    should 'wait interval time before check it another time' do
      @alarm.launch
      assert(@alarm.fired?, 'Alarm was fired')
      assert_not(@alarm.must_be_checked?,
                 'Alarm musn\'t be checked because it is in waiting '\
                 'period.')
      # Now, we will step waiting period and recheck it
      @alarm.update(fired: Time.now - @alarm.interval - 1000)
      assert(@alarm.must_be_checked?,
             'Alarm mus\'t be checked because it finish waiting '\
             'period.')
    end
  end
end

#
# Return params for a invalid alarm (need uppper or bottom limit)
#
# == Returns:
# Hash with params
#
def alarm_params
  {
    name: 'NewAlarm',
    interval: Alarm::INTERVAL_RANGE.last,
    time_window: Alarm::TIME_WINDOW_RANGE.last,
    product_type: 'flow',
    severity: severities(:one),
    condition: '{}' # JSON filters
  }
end

#
# Return params for a valid alarm (with emails)
#
# == Returns:
# Hash with params
#
def alarm_params_email
  alarm_params_upper.merge(
    notify_to_email: 1,
    email_list: 'asdasd@gmail.com,bbb@gmail.com'
  )
end

#
# Return params for a valid alarm (with Slack Webhook)
#
# == Returns:
# Hash with params
#
def alarm_params_slack
  alarm_params_upper.merge(
    notify_to_slack: 1,
    slack_webhook: 'https://hooks.slack.com/services/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'
  )
end

def alarm_params_slack_invalid
  alarm_params_upper.merge(
    notify_to_slack: 1,
    slack_webhook: 'gkgkgkg;lhk'
  )
end
#
# Return params for a valid alarm (with emails)
#
# == Returns:
# Hash with params
#
def alarm_params_email_invalid
  alarm_params_upper.merge(
    notify_to_email: 1,
    email_list: 'asdasd,bbb@gmail.com,asd@asdk.com'
  )
end

#
# Return params for a valid alarm (upper limit)
#
# == Returns:
# Hash with params
#
def alarm_params_upper
  alarm_params.merge(
    upper_limit_enabled: 1,
    upper_limit: 5000
  )
end

#
# Return params for a valid alarm (bottom limit)
#
# == Returns:
# Hash with params
#
def alarm_params_bottom
  alarm_params.merge(
    bottom_limit_enabled: 1,
    bottom_limit: 5000
  )
end
