require 'test_helper'

# Tests for Dashboard models
#
class DashboardTest < ActiveSupport::TestCase
  should validate_presence_of(:name)
  should validate_inclusion_of(:layout).in_array(%w(columns absolute))
  should belong_to(:owner)
  should have_many(:all_widgets)

  context 'Convert dashboards to reports' do
    setup do
      @dashboard = dashboards(:columnar)
      @dashboard_col = dashboards(:admin_col_dashboard)
      @dashboard_abs = dashboards(:admin_abs_dashboard)
      @admin = users(:admin)
      User.current_user = @admin
    end

    should 'convert a column dashboard to a report' do
      report = @dashboard_col.to_report
      # Check the report is created and the number of widgets is correct.
      check_basic_report(report, @dashboard_col)
      # Check position of widgets
      [*1..20].each do |w|
        widget = report.widgets.find_by(name: "text widget test#{w}")
        dwidget = @dashboard_col.all_widgets.find_by(name: "text widget test#{w}")
        assert_not_nil(widget)
        assert_equal(1, widget.position_x)
        assert_equal(1, widget.position_y)
        assert_equal(4, widget.position_width)
        assert_equal(dwidget.position_height, widget.position_height)
      end
    end

    should 'calculate new pages' do
      report = @dashboard.to_report
      # Check the report is created and the number of widgets is correct.
      check_basic_report(report, @dashboard)
      assert_equal(report.number_pages, 2)
    end

    should 'convert an absolute dashboard to a report' do
      report = @dashboard_abs.to_report
      check_basic_report(report, @dashboard_abs)
      # Check position of widgets
      position_x = 1
      position_y = 1
      page = 1
      per_page = 0

      report.widgets.sort_by { |w| [w.id.to_i, w.position_y] }.each do |w|
        assert_equal(position_x, w.position_x)
        assert_equal(position_y, w.position_y)
        assert_equal(2, w.position_width)
        assert_equal(2, w.position_height)
        # Increase position
        position_x += w.position_width
        per_page += 1

        if per_page >= 18
          per_page = 0
          position_y = 1
          position_x = 1
          page += 1
        end
        if position_x > 5
          position_y += 2
          position_x = 1
        end

      end
    end
  end

  context 'An admin user' do
    setup do
      @dashboard_col = dashboards(:admin_col_dashboard)
      @dashboard_abs = dashboards(:admin_abs_dashboard)
      @admin_default = dashboards(:admin_default)
      @admin = users(:admin)
      User.current_user = @admin
    end

    should 'duplicate a dashboard' do
      new_dashboard = @dashboard_col.dup
      assert_equal(@dashboard_col.name, new_dashboard.name)
      assert_equal(@dashboard_col.description, new_dashboard.description)
      assert_equal(@dashboard_col.layout, new_dashboard.layout)
    end

    should 'check if is a dashboard or a report' do
      assert_equal(true, @dashboard_col.dashboard?)
      assert_equal(false, @dashboard_col.report?)
    end

    should 'check if a dashboard is absolute or columns based' do
      assert_equal(true, @dashboard_abs.absolute?)
      assert_equal(true, @dashboard_col.columns?)
    end

    should 'check if an user has permissions over a dashboard' do
      assert_equal(true, @admin_default.permission?(@admin))
      assert_equal(true, @admin_default.read_permission?(@admin))
    end

    should 'change layout to absolute or to columns' do
      @dashboard_col.update(layout: 'absolute')
      assert_equal('absolute', @dashboard_col.reload.layout)

      @dashboard_col.update(layout: 'columns')
      assert_equal('columns', @dashboard_col.reload.layout)
    end

    should 'reset default dashboards' do
      @admin_default.update(name: 'New name')
      assert_equal(false, @admin_default.reload.default)
    end

    should 'export a dashboard to json' do
      dashboard = @admin_default.export_json
      assert_equal @admin_default.name,
                   ActiveSupport::JSON.decode(dashboard)['name']
    end

    should 'add a colaborator' do
      collaborators = [{'user_id' => 1, 'can_edit' => 'on'}]
      @admin_default.update(collaborators: collaborators)
      assert_not_equal nil, @admin_default.dashboard_users.where(user_id: 1)
    end
  end

  # Check if the report is created.
  def check_basic_report(report, dashboard)
    # Check the report is created and the number is widgets is correct.
    assert(report.persisted?)
    assert_equal(dashboard.widgets.count(&:can_report?),
                 report.widgets.size)
    # Check properties
    check_type_properties(report)
  end

  # Check the porperties are integers
  def check_type_properties(report)
    report.widgets.each do |w|
      %w(x y width height).each do |property|
        assert w.send("position_#{property}").is_a? Fixnum
      end
    end
  end
end
