require 'test_helper'
require 'yaml'

#
# Testing DruidSegments class
#
class DruidSegmentsTest < ActiveSupport::TestCase
  # ----------------------------------------------------
  # Stub segments
  # ----------------------------------------------------

  # Create an struct for the payload
  SegmentStruct =
    Struct.new(:data_source, :interval, :version, :loadSpec, :dimensions,
               :metrics, :shardSpec, :binaryVersion, :size, :identifier)

  SegmentData = Struct.new(:id, :datasource, :created_date, :start, :end,
                           :partitioned, :version, :used, :payload) do
    #
    # Load the payload of the segment
    #
    # == Returns::
    # Hash with data
    #
    def payload_hash
      return @cached_payload if @cached_payload
      @cached_payload = JSON.parse [payload[2..-1]].pack('H*')
    rescue JSON::ParserError
      @cached_payload = JSON.parse payload
    end

    #
    # Load the payload of the segment and initialize structs
    #
    # == Returns::
    # Struct with data
    #
    def payload_struct
      SegmentStruct.new(*payload_hash.values)
    end
  end

  # ----------------------------------------------------
  # Tests
  # ----------------------------------------------------

  context 'basic functionality' do
    setup do
      @druid_segments = DruidSegments.new(Time.now, sensors(:root))
      data =
        YAML.load(
          File.open("#{Rails.root}/test/ext_fixtures/druid_segments.yml"))
      @loaded_data = data.map do |_, value|
        SegmentData.new(*value.values)
      end
      # Load total for default
      @total_default = @loaded_data.map do |e|
        /^(rb_[a-zA-Z]+)$/.match(e.datasource) ? e.payload_struct.size : 0
      end.sum
      # Load total for rb_flow
      @total_rb_flow = @loaded_data.map do |e|
        e.datasource.include?('rb_flow') ? e.payload_struct.size : 0
      end.sum
      DruidSegments.any_instance.stubs(:month_data).returns(@loaded_data)
    end

    should 'set a month' do
      @druid_segments.month = Time.now - 1.month
      assert_equal(
        (Time.now - 1.month).utc.beginning_of_month, @druid_segments.month,
        'The month must be updated')
    end

    should 'get the data from the month' do
      assert(@druid_segments.respond_to?(:month_data),
             'DruidSegments must define month_data')
    end

    should 'get total' do
      assert(@druid_segments.respond_to?(:total),
             'DruidSegments must define total')
      total = @druid_segments.month_data.map do |seg|
        seg.payload_struct.size
      end.sum
      assert_equal(total, @druid_segments.total,
                   'The total is bad calculated')
    end

    should 'get timeserie' do
      assert(@druid_segments.respond_to?(:timeserie),
             'DruidSegments must define timeserie')
      row = @druid_segments.timeserie.first
      # Check format
      assert(row.key?(:event), 'Event is the key that has event number')
      assert(row.key?(:timestamp), 'The time of the event happens')
    end

    should 'get accumulated timeserie' do
      assert(@druid_segments.respond_to?(:timeserie),
             'DruidSegments must define timeserie')
      accumulated = @druid_segments.timeserie_accumulated
      row = accumulated.first
      # Check format
      assert(row.key?(:event), 'Event is the key that has event number')
      assert(row.key?(:timestamp), 'The time of the event happens')
      # Check values are accumulative
      last = 0
      count = 50
      accumulated.reverse.each do |el|
        assert(el[:event][:events] >= last, "Accumulated values can't decrease")
        last = el[:event][:events]
        break if count == 0
        count -= 1
      end
    end

    should 'get data from the month based on nampespaces' do
      assert(@druid_segments.respond_to?(:by_namespaces),
             'DruidSegments must define by_namespaces')
      data = @druid_segments.by_namespaces
      assert(data.is_a?(Hash), 'Returned data must be a Hash')
      assert_equal(
        data.keys.sort, %w(default NS-12390uijoadsl123).sort,
        'Keys must coincide')
      # Sum for default must coincide
      assert_equal(@total_default, data['default'],
                   'Values must be the same')
    end

    should 'get data from the month of a nampespace' do
      assert(@druid_segments.respond_to?(:by_namespace),
             'DruidSegments must define by_namespace')
      data = @druid_segments.by_namespace('default')
      assert_equal(@total_default, data,
                   'Values must be the same')
    end

    should 'get data from the month based on topics' do
      assert(@druid_segments.respond_to?(:by_topics),
             'DruidSegments must define by_topics')
      data = @druid_segments.by_topics
      assert(data.is_a?(Hash), 'Returned data must be a Hash')
      assert_equal(
        data.keys.sort, %w(Flow Monitor Social Hashtag).sort,
        'Keys must coincide')
      # Sum for rb_hashtag must coincide
      assert_equal(@total_rb_flow, data['Flow'],
                   'Values must be the same')
    end

    should 'get data from the month of a topic' do
      assert(@druid_segments.respond_to?(:by_topic),
             'DruidSegments must define by_topic')
      data = @druid_segments.by_topic('Flow')
      assert_equal(@total_rb_flow, data,
                   'Values must be the same')
    end
  end
end
