require 'test_helper'

class LicenseTest < ActiveSupport::TestCase
  context 'Valid root license' do
    setup do
      @license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid,
                                expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)

      data_bag = DataBagStub.new('licenses' => {}, 'sensors' => {})
      Chef::DataBagItem.stubs(:load).returns(data_bag)
    end

    should 'be added' do
      license = License.new(encoded_license: @license)
      assert license.save
      assert_equal(license.sensor, Sensor.root)
      assert_includes(Sensor.root.licenses, license)
    end
  end

  context 'Valid organization license' do
    setup do
      @organization = sensors(:organization)
      @license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid,
                                organization_uuid: @organization.uuid, expire_at: (Time.now + 30.days).to_i,
                                limit_bytes: 500_000_000)

      data_bag = DataBagStub.new('licenses' => {}, 'sensors' => {})
      Chef::DataBagItem.stubs(:load).returns(data_bag)
    end

    should 'be added' do
      license = License.new(encoded_license: @license)
      assert license.save
      assert_equal(license.sensor, @organization)
      assert_includes(@organization.licenses, license)
    end
  end

  context 'Existing license' do
    setup do
      uuid = SecureRandom.uuid
      @license_1 = create_license(uuid: uuid, cluster_uuid: Setting.cluster_uuid, expire_at: (Time.now + 10.days).to_i, limit_bytes: 500_000_000)
      @license_2 = create_license(uuid: uuid, cluster_uuid: Setting.cluster_uuid, expire_at: (Time.now + 40.days).to_i, limit_bytes: 600_000_000)
    end

    should 'be updated' do
      assert_difference 'License.count', 1 do
        license_1 = License.new(encoded_license: @license_1)
        assert license_1.save
      end

      assert_no_difference 'License.count' do
        license_2 = License.new_or_update(encoded_license: @license_2)
        assert license_2.save
      end
    end
  end

  context 'License with wrong signature' do
    setup do
      @license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid, expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)

      data_bag = DataBagStub.new('licenses' => {})
      Chef::DataBagItem.stubs(:load).returns(data_bag)
    end

    should 'not be added' do
      license = License.new(encoded_license: @license)

      # modify info
      info = license.info
      info['expire_at'] = (Time.now + 10.days).to_i

      license.encoded_info = Base64.urlsafe_encode64(info.to_json)

      assert_not license.save
      assert license.errors[:signature].any?
    end
  end

  context 'License with wrong cluster UUID' do
    setup do
      @license = create_license(uuid: SecureRandom.uuid, cluster_uuid: SecureRandom.uuid, expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)
    end

    should 'return an error' do
      license = License.new(encoded_license: @license)

      assert_not license.save
      assert license.errors[:cluster_uuid].any?
    end
  end

  context 'License with wrong UUID' do
    should 'return an error' do
      encoded_license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid, expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)

      valid_license = License.new(encoded_license: encoded_license)
      valid_license.save

      wrong_license = License.new(encoded_license: encoded_license)

      assert_not wrong_license.save
      assert wrong_license.errors[:uuid].any?
    end
  end

  context 'Invalid format license' do
    should 'raise an exception' do
      assert_raises(License::InvalidFormat) { License.new(encoded_license: (0...8).map { (65 + rand(26)).chr }.join) }
    end
  end

  context 'Invalid organization' do
    should 'raise an exception' do
      encoded_license = create_license(uuid: SecureRandom.uuid, cluster_uuid: Setting.cluster_uuid,
                                       organization_uuid: SecureRandom.uuid,
                                       expire_at: (Time.now + 30.days).to_i, limit_bytes: 500_000_000)
      assert_raises(License::InvalidOrganization) do
        License.new(encoded_license: encoded_license)
      end
    end
  end
end
