require 'test_helper'

#
# Check the creation and cache of redborder objects
#
class RedborderObjectTest < ActiveSupport::TestCase
  should validate_presence_of(:name)
  should validate_presence_of(:value)

  #
  # Test personal objects of users like MacObjects, NetObjects...
  #
  context 'Private Objects' do
    setup do
      Rails.cache.clear
      @user1 = users(:admin)
      @user2 = users(:admin_domain)
      @net_user1 = redborder_objects(:net1)
      @net_user2 = redborder_objects(:net3)
      @net_user1_destroy = redborder_objects(:net_destroy)

      # Stub chef
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'Read only his own NetObjects' do
      assert_equal(3, @user1.net_objects.size, 'User1 only have 3 Objects')
      assert_equal(1, @user2.net_objects.size, 'User2 only have 1 Objects')

      # Check fetch method
      value_ok = NetObject.fetch(@net_user1.value, @user1.id, nil)
      value_nil = NetObject.fetch(@net_user2.value, @user1.id, nil)

      assert_equal(@net_user1.name, value_ok, 'User1 must read the value of '\
                                              'his object.')
      assert_nil value_nil, 'User1 musn\'t read the value of User2 object.'
    end

    should 'Objects must be cached' do
      Rails.cache.clear
      # Return the element and create the cache
      NetObject.fetch(@net_user1.value, @user1.id, nil)
      # Read direct on the cache
      act = Rails.cache.read(
        NetObject.cache_route_with_key(@net_user1.value, @user1.id)
      )

      assert_equal(@net_user1.name, act, 'The Object must be cached!')
    end

    # should 'When create a new Object, create the cache' do
    #   new_object = NetObject.create(value: '127.0.0.1', name: 'CreatedForTest',
    #                                 user_id: @user1.id)
    #   rte = NetObject.cache_route_with_key(new_object.value, @user1.id)
    #   # Read direct on the cache
    #   act = Rails.cache.read(rte)
    #   assert_equal(new_object.name, act, 'Cache must be created when save a '\
    #                                      'new Object')
    # end

    # should 'When update an Object, cache must change' do
    #   Rails.cache.clear
    #   # Return the element and create the cache
    #   NetObject.fetch(@net_user1.value, @user1.id, nil)
    #   # Update the element
    #   new_value = 'UpdatingObject'
    #   @net_user1.update(name: new_value)
    #   # Fetch the element after update
    #   res = NetObject.fetch(@net_user1.value, @user1.id, nil)

    #   assert_equal(new_value, res, 'The Object must update the cache too')
    # end

    # should 'When destroy an Object, cache must be deleted too' do
    #   Rails.cache.clear
    #   # Return the element and create the cache
    #   NetObject.fetch(@net_user1_destroy.value, @user1.id, nil)
    #   # Destroy the object
    #   rte = NetObject.cache_route_with_key(@net_user1_destroy.value, @user1.id)
    #   @net_user1_destroy.destroy
    #   # Read direct on the cache
    #   act = Rails.cache.read(rte)
    #   assert_nil act, 'The Object must delete the cache when is destroyed'
    # end

    should 'has search method' do
      # Check method
      assert_respond_to(RedborderObject, :search)
    end

    should 'be searched by name' do
      # If the object is isolated we must pass the user
      assert_equal([], NetObject.search('MyP'),
                   'The user is not send and the Object is isolated')
      assert_equal(1, NetObject.search('MyP', @user1).size,
                   'There are 1 element for admin user with MyP in name')
      # You can not see objects of another user.
      assert_equal([],  NetObject.search(@net_user2.name, @user1),
                   'Net3 object is an object of admin_domain')
    end

    should 'be searched by value' do
      assert_equal(1, NetObject.search('0.0.1', @user1).size,
                   'There are 1 element with 0.0.1 in value')
    end
  end

  context 'Service Objects' do
    setup do
      Rails.cache.clear
      @user1 = users(:admin)
      @user2 = users(:admin_domain)
      @service1 = redborder_objects(:service1)
      @service2 = redborder_objects(:service2)
      @service3 = redborder_objects(:service3)
    end

    should 'Read his own ServiceObjects and non assigned ones, if super admin' do
      assert_equal(3, @user1.service_objects.size, 'User1 can read 3 Objects')

      # Check fetch method
      user_list = @user1.super_admin? ? [@user1.id, nil] : @user1.id
      value_ok = ServiceObject.fetch(@service1.value, user_list, nil)
      value_ok_2 = ServiceObject.fetch(@service3.value, user_list, nil)
      value_nil = ServiceObject.fetch(@service2.value, user_list, nil)

      assert_equal(@service1.name, value_ok, 'User1 must read the value of '\
                                             'his object.')
      assert_equal(@service3.name, value_ok_2, 'User1 must read the value of '\
                                               'a non assigned object.')
      assert_nil value_nil, 'User1 musn\'t read the value of User2 object.'
    end

    should 'Read only his own ServiceObjects, if non super admin' do
      assert_equal(1, @user2.service_objects.size, 'User2 only have 1 Objects')

      # Check fetch method
      user_list = @user2.super_admin? ? [@user2.id, nil] : @user2.id
      value_ok = ServiceObject.fetch(@service2.value, user_list, nil)
      value_nil = ServiceObject.fetch(@service1.value, user_list, nil)

      assert_equal(@service2.name, value_ok, 'User2 must read the value of '\
                                              'his object.')
      assert_nil value_nil, 'User2 musn\'t read the value of User1 object.'
    end
  end

  #
  # Test global objects of users: protocol and services
  #
  context 'Global Objects' do
    setup do
      Rails.cache.clear
      @user1 = users(:admin)
      @user2 = users(:admin_domain)
      @protocol1 = redborder_objects(:protocol1)
      @protocol2 = redborder_objects(:protocol2)
      @protocol_destroy = redborder_objects(:protocol_destroy)
    end

    should 'Read all Service and Protocol Objects' do
      # Check how many objects has
      assert_equal(4, @user1.protocol_objects.size, 'User1 only have 3 Objects')
      assert_equal(4, @user2.protocol_objects.size, 'User2 only have 1 Objects')

      # Check fetch method
      value3 = ProtocolObject.fetch(@protocol1.value, nil, nil)
      value4 = ProtocolObject.fetch(@protocol2.value, nil, nil)

      # Check
      assert_equal(@protocol1.name, value3, 'User must read the value of '\
                                            'global object.')
      assert_equal(@protocol2.name, value4, 'User must read the value of '\
                                            'global object.')
    end

    should 'has search method' do
      # Check method
      assert_respond_to(RedborderObject, :search)
    end

    should 'be searched by name' do
      # Check method
      assert_respond_to(RedborderObject, :search)
      # If the object is isolated we must pass the user
      assert_equal([], ProtocolObject.search('qwdasqwdas'),
                   "There aren't any object with this name or value")
      assert_equal(1, ProtocolObject.search('icmp').size,
                   'There are 1 element with icmp in name')
    end

    should 'be searched by value' do
      assert_equal(1, ProtocolObject.search('13').size,
                   'There are 1 element with 13 in value')
    end

    should 'Objects must be cached' do
      Rails.cache.clear
      # Return the element and create the cache
      ProtocolObject.fetch(@protocol1.value, nil, nil)
      ProtocolObject.fetch(@protocol2.value, nil, nil)
      # Read direct on the cache
      value3 = Rails.cache.read(
        ProtocolObject.cache_route_with_key(@protocol1.value, nil)
      )
      value4 = Rails.cache.read(
        ProtocolObject.cache_route_with_key(@protocol2.value, nil)
      )

      # Check
      assert_equal(@protocol1.name, value3, 'The Object must be cached!')
      assert_equal(@protocol2.name, value4, 'The Object must be cached!')
    end

    # should 'When update an Object, cache must change' do
    #   Rails.cache.clear
    #   # Return the element and create the cache
    #   ProtocolObject.fetch(@protocol1.value, nil, nil)
    #   # Update the element
    #   new_value = 'UpdatingObject2'
    #   @protocol1.update(name: new_value)

    #   # Fetch the element after update
    #   value = ProtocolObject.fetch(@protocol1.value, nil, nil)

    #   # Check
    #   assert_equal(new_value, value, 'The Object must update the cache too')
    # end

    # should 'When destroy an Object, cache must be deleted too' do
    #   Rails.cache.clear
    #   # Return the element and create the cache
    #   ProtocolObject.fetch(@protocol_destroy.value, nil, nil)
    #   # Destroy the object
    #   route = ProtocolObject
    #            .cache_route_with_key(@protocol_destroy.value, nil)
    #   @protocol_destroy.destroy
    #   # Read direct on the cache
    #   value = Rails.cache.read(route)

    #   # Check
    #   assert_nil value, 'The Object must delete the cache when is destroyed'
    # end
  end
end
