require 'test_helper'

class SensorTest < ActiveSupport::TestCase
  # Store accessors for property
  STORE_ACCESSORS = %w(gps_points domain_zones
                       general_description floor_level active_modules company
                       email path mac_hashing_salt)
  should have_db_column(:property)
  # Before apply this valdiation we must improve check_snmp_version
  # validates
  # should validate_presence_of(:name)
  # should validate_length_of(:name).is_at_most(50)

  context '' do
    setup do
      RedBorder.send(:remove_const, 'LICENSING_MODE')
      RedBorder::LICENSING_MODE = 'global'
    end

    should 'Rebuild tree' do
      Sensor.rebuild!
    end
  end

  context 'Basic sensors' do
    setup do
      @root = sensors(:root)
      # Mock
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'have properties in property column' do
      s = sensors(:root)
      STORE_ACCESSORS.each do |prop|
        assert(s.respond_to?(prop), "Sensor must define a getter for: #{prop}")
        assert(s.respond_to?("#{prop}="),
               "Sensor must define a setter for: #{prop}")
      end
    end

    should 'create root sensor' do
      s = sensors(:root)
      assert s.valid?, 'Root sensor is not valid.'
    end

    should 'create domain' do
      s = sensors(:deployment)
      assert s.valid?, 'Domain is not valid.'
    end

    should "can't create a sensors with Root name or domain_type" do
      assert_not(
        Sensor.new(name: 'root', type: Sensor::DOMAIN, domain_type: Sensor::ROOT).valid?)
      assert_not(
        Sensor.new(name: 'asdasd', type: Sensor::DOMAIN, domain_type: Sensor::ROOT).valid?)
    end

    should 'prevent change the domain type' do
      s = sensors(:deployment)
      assert_not s.update(domain_type: Sensor::ORGANIZATION)
      # Check error
      assert s.errors.messages.keys.include? :domain_type
    end

    should 'accept unicode characters in the name' do
      allowed_name = "ѺѼ_.-,||@+=/!!:#>&"
      not_allowed_name = "_.%%\\/!!:#''>'&"
      not_allowed_unicode = "ѺѼ_.-,||@ՖՖ"

      s_valid =
        Sensor.new(name: allowed_name, type: 1, domain_type: @root.children_type, parent: @root)
      s_not_valid =
        Sensor.new(name: not_allowed_name, type: 1, domain_type: @root.children_type,
                   parent: @root)
      s_not_valid_unicode =
        Sensor.new(name: not_allowed_unicode, type: 1, domain_type: @root.children_type,
                   parent: @root)

      assert(s_valid.valid?, 'These characters must be allowed')
      assert_not(s_not_valid.valid?, 'This sensor has not allowed characters')
      assert_not(s_not_valid_unicode.valid?,
                 'This sensor has not allowed unicode characters')

      # Save the valid and test find by method
      s_valid.save
      assert(Sensor.exists?(name: allowed_name),
             'We must can find this sensor by name')
    end

    should 'get recommended children type' do
      Sensor::HIERARCHY.each_with_index do |domain_type, i|
        s = Sensor.new(name: 'Testing', parent: @root, type: Sensor::DOMAIN,
                       domain_type: domain_type)
        children_type = Sensor::HIERARCHY[i + 1] || Sensor::FLOOR
        assert_equal(children_type, s.children_type)
      end
    end

    should 'not create a sensor with an existing uuid' do
      s = sensors(:deployment)
      s_not_valid = Sensor.new(name: 'Not valid', type: s.type, parent: s.parent, uuid: s.uuid)
      assert_not s_not_valid.valid?, 'Domain not valid: duplicated uuid.'
    end

    should 'initialize uuid if no one has been provided' do
      s = Sensor.create(name: 'Domain_uuid', type: Sensor::DOMAIN,
                        domain_type: Sensor::ORGANIZATION, parent: sensors(:namespace))
      assert s, 'Sensor not valid'
      assert_match(/([a-f\d]{8}(-[a-f\d]{4}){3}-[a-f\d]{12}?)/i, s.uuid, 'Sensor\'s uuid is not uuid format')
    end

    should 'not create a domain with the same name, domain_type and parent' do
      s = sensors(:deployment)
      s_not_valid = Sensor.new(name: s.name, type: s.type, domain_type: s.domain_type, parent: s.parent)
      assert_not s_not_valid.valid?, 'Domain not valid: duplicated params.'
    end

    should 'create a domain with the same name, but different domain_type or parent' do
      s = sensors(:deployment)

      s_valid = Sensor.new(name: s.name, type: s.type, domain_type: Sensor::FLOOR, parent: s.parent)
      assert s_valid.valid?, 'Domain is valid.'

      s_valid = Sensor.new(name: s.name, type: s.type, domain_type: Sensor::FLOOR, parent: s)
      assert s_valid.valid?, 'Domain is valid.'
    end

    should 'move market into organization' do
      organization = sensors(:organization)
      market = sensors(:market)

      market.parent_id = organization.id
      assert market.valid?, 'Move market into organization invalid.'
    end

    should 'can\'t move market into floor' do
      market = sensors(:market)
      floor = sensors(:floor)

      market.parent_id = floor.id
      assert_not market.valid?, 'Floor can\'t have children'
    end
  end

  context 'An organization' do
    setup do
      @organization = sensors(:organization)
      # Mock
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'has active_modules initialized' do
      am = @organization.active_modules
      assert_not_nil(am, 'Active modules must be initialized')
      assert_not_empty(am, 'Active modules must be initialized')
    end

    should 'check if a module is active' do
      # Active modules must be initialized
      mod = @organization.active_modules.split(',').last
      assert_respond_to(@organization, :active_module?,
                        'A sensor must define active_module? method')
      assert(@organization.active_module?(mod),
             'The module is active on this sensor')
    end

    should 'update active modules' do
      new_active = 'gg,dd,vv,ee'
      @organization.update(active_modules: new_active)
      @organization.reload
      # Check if there aren't any callback changing this update
      assert_equal(new_active, @organization.active_modules,
                   'Organization must keep changes when update active modules')
    end

    should 'active_modules must be initialized' do
      s = Sensor.create(name: 'NewOrganization', type: Sensor::DOMAIN,
                        domain_type: Sensor::ORGANIZATION, parent: sensors(:namespace))
      assert_not_nil(s.active_modules, 'Active modules must be initialized')
    end
  end

  context 'any sensor' do
    setup do
      @ips = sensors :ips
      @mse = sensors :mse
      @mse_cp = sensors :mse_cp
      # Stub chef configuration
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'return a last_check_in from a timestamp' do
      assert(@ips.check_in_time.is_a?(Fixnum), "#{@ips.last_check_in} is not a Fixnum")
    end

    should 'return a last_check_in from a valid String' do
      assert(@mse.check_in_time.is_a?(Fixnum), "#{@mse.last_check_in} is not a Fixnum")
    end

    should 'return a nil last_check_in from an invalid String' do
      assert(@mse_cp.check_in_time.nil?, "#{@mse_cp.last_check_in} is not a nil")
    end
  end

  context 'request from API' do
    setup do
      @organization = sensors(:organization)
      # Stub chef configuration
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'respond to to_api method with a hash of the properties of organization' do
      assert_respond_to(@organization, :to_api,
                        'Sensors must respondo to to_api method')
      data = @organization.to_api
      assert(data.is_a?(Hash), 'To api method must return a Hash')
      # Visualization range and limits is in organization
      %w(id name domain_type type uuid).each do |column|
        assert(data.key?(column), "Returned sensor must contains #{column}")
      end
      # Mac hashing salt is only for Service Providers
      assert_not(data.key?(:mac_hashing_salt),
                 'An organization sensor hasn\'t got MAC Hashing salt')
    end

    should 'add mac_hashing_salt on response when it\'s a service provider' do
      mac_hashing_salt = 'ASDASDASDASD123'
      s = Sensor.new(name: 'aaa', type: Sensor::DOMAIN,
                     domain_type: Sensor::SERVICE_PROVIDER,
                     mac_hashing_salt: mac_hashing_salt)

      assert(s.to_api.key?('mac_hashing_salt'),
             'A service provider must return mac_hashing_salt')
      assert_equal(mac_hashing_salt, s.mac_hashing_salt,
                   'MAC Hashing salt must coincide with created')
    end
  end

  context 'a Client Proxy' do
    setup do
      @client_proxy = sensors(:client_proxy)
      # Stub chef configuration
      chef = ChefStub.new
      Chef::Node.stubs(:load).returns(chef)
      Chef::Role.stubs(:load).returns(chef)
      Chef::Node.stubs(:from_hash).returns(chef)
      Chef::Role.stubs(:from_hash).returns(chef)
    end

    should 'respond true to physical? method' do
      assert(@client_proxy.physical?, 'Client Proxy has a physical node')
    end

    should 'can reregister the private key' do
      @private_key = Struct.new(:private_key).new('SDADSQASDASDASDQWDAS')
      Chef::ApiClient.stubs(:reregister).returns(@private_key)

      assert_equal(@private_key.private_key, @client_proxy.reregister,
                   'Private key msut be regenerated on client proxy')
    end

    should 'change its location' do
      latitude = '0'
      longitude = '0'
      @client_proxy.update(latitude: latitude, longitude: longitude)
      @client_proxy.reload
      assert_equal(@client_proxy.latitude, latitude, 'Latitude has not been updated')
      assert_equal(@client_proxy.latitude, longitude, 'Longitude has not been updated')
    end
  end
end
