require 'test_helper'

#
# Test user model and permissions
#
class UserTest < ActiveSupport::TestCase
  should validate_presence_of(:name)
  # available_modules =
  #   (Redborder::Module.registered_modules.keys - [:location]).map { |m| "/#{m}" }
  # should validate_inclusion_of(:root_page).in_array(%w(/dashboard /sensors) + available_modules)
  should validate_inclusion_of(:per_page_count).in_array([25, 45, 65, 100])

  # Check basic functionality of users
  context 'User' do
    setup do
      @user = users(:admin)
    end

    should 'recreate auth token' do
      # The length of the auth token msut be 30
      first_auth = @user.authentication_token
      @user.reset_authentication_token!
      new_auth = @user.authentication_token

      # Check tokens
      assert(first_auth != new_auth, 'Auth tokens mus\'t be different')
      assert_equal(40, new_auth.length, 'Recreated auth token must has 30 ' \
                                        'characters')
    end
  end

  # Check permissions of admin
  context 'Superadmin user' do
    setup do
      # Init Admin user
      @user = users(:admin)
      @domain_sensor = sensors(:deployment)
      @flow_sensor = sensors(:flow)
    end

    should 'be superdamin' do
      assert(@user.super_admin?, 'The user Admin is superadmin!')
    end

    should 'has permission on sensors' do
      assert(@user.can?(:manage, @domain_sensor),
             'The user can manage a domain sensor!')
      assert(@user.can?(:manage, @flow_sensor),
             'The user can manage a flow sensor!')
    end

    should 'see and manage all sensors using user methods' do
      assert_equal(Sensor.count, @user.sensors_with_permission.size,
                   'The user can see all sensors!')
      assert_equal(Sensor.count, @user.sensors_with_manage_permission.size,
                   'The user can manage all sensors!')
    end
  end

  # Normal user with top domain on domain and admin active
  context 'Admin on domain' do
    setup do
      # Init Admin user
      @user = users(:admin_domain)
      @domain_sensor = sensors(:deployment)
      # This is a child of domain
      @organization_sensor = sensors(:organization)
      @market_sensor = sensors(:market)
      # Flow sensor
      @flow_sensor = sensors(:flow)
    end

    should 'be admin only on domain and his children' do
      # Is admin
      assert(!@user.super_admin?, 'The user AdminDomain isn\'t superadmin!')
      assert(@user.admin_on?(@domain_sensor),
             'The user is admin on domain')
      assert(@user.admin_on?(@organization_sensor),
             'The user is admin on organization')
      # The user can read and manage on flow_sensor
      assert(@user.admin_on?(@flow_sensor),
             'The user is admin on flow_sensor')
    end

    should 'can read only domain and his children' do
      # Can read
      assert(@user.can?(:read, @domain_sensor),
             'The user can read a domain sensor!')
      assert(@user.can?(:read, @organization_sensor),
             'The user can read a organization sensor!')
      assert(@user.can?(:read, @market_sensor),
             'The user can read a market sensor!')
      # Can read
      assert(@user.can?(:read, @flow_sensor),
             'The user can read a flow sensor!')
    end

    should 'has permission only on domain and his children' do
      # Can manage
      assert(@user.can?(:manage, @domain_sensor),
             'The user can manage a domain sensor!')
      assert(@user.can?(:manage, @organization_sensor),
             'The user can manage a organization sensor!')
      assert(@user.can?(:manage, @market_sensor),
             'The user can manage a market sensor!')
      # Can manage
      assert(@user.can?(:manage, @flow_sensor),
             'The user can manage a flow sensor!')
    end

    should 'see and manage only on domain and his children '\
           'sensors using user methods' do
      assert_equal(@user.sensor.all_children(true).size, @user.sensors_with_permission.size,
                   'The user can see 4 sensors!')
      assert_equal(@user.sensor.all_children(true).size, @user.sensors_with_manage_permission.size,
                   'The user can manage 4 sensors!')
    end
  end

  # Normal user with top domain on domain and only can read
  context 'User on domain' do
    setup do
      # Init Admin user
      @user = users(:user_domain)
      @domain_sensor = sensors(:deployment)
      # This is a child of domain
      @organization_sensor = sensors(:organization)
      @market_sensor = sensors(:market)
      # Flow sensor
      @flow_sensor = sensors(:flow)
    end

    should 'be admin only on domain and his children' do
      # Is admin
      assert(!@user.super_admin?, 'The user AdminDomain isn\'t superadmin!')
      assert(!@user.admin_on?(@domain_sensor),
             'The user is admin on domain')
      assert(!@user.admin_on?(@organization_sensor),
             'The user is admin on organization')
      # The user can't read and manage on flow_sensor
      assert(!@user.admin_on?(@flow_sensor),
             'The user isn\'t admin on flow_sensor')
    end

    should 'can read only domain and his children' do
      # Can read
      assert(@user.can?(:read, @domain_sensor),
             'The user can read a domain sensor!')
      assert(@user.can?(:read, @organization_sensor),
             'The user can read a organization sensor!')
      assert(@user.can?(:read, @market_sensor),
             'The user can read a market sensor!')
      # Can read
      assert(@user.can?(:read, @flow_sensor),
             'The user can read a flow sensor!')
    end

    should 'hasn\'t permissions to manage' do
      # Can manage
      assert(!@user.can?(:manage, @domain_sensor),
             'The user can manage a domain sensor!')
      assert(!@user.can?(:manage, @organization_sensor),
             'The user can manage a organization sensor!')
      assert(!@user.can?(:manage, @market_sensor),
             'The user can manage a market sensor!')
      # Can't manage
      assert(!@user.can?(:manage, @flow_sensor),
             'The user can\'t manage a flow sensor!')
    end

    should 'see only on domain and his children '\
           'sensors using user methods' do
      assert_equal(@user.sensor.all_children(true).size, @user.sensors_with_permission.size,
                   'The user can see 4 sensors!')
      assert_equal(0, @user.sensors_with_manage_permission.size,
                   'The user can\'t manage any sensors!')
    end
  end
end
