#
# Added some helpers to work with integration tests
#
module IntegrationSupport
  #
  # Check if the UI is blocked.
  #
  # == Parameters:
  # page::
  #   The current page (capybara session)
  #
  def ui_blocked?
    assert(page.has_css?('div.blockUI', count: 2), 'UI must be blocked')
  end

  #
  # Wait the UI to finish AJAX calls
  #
  # == Parameters:
  # page::
  #   The current page (capybara session)
  # sleep::
  #   True if you want to sleep sometimes to leave
  #   time to javascript
  #
  def wait_for_ajax(page, sleep = false)
    Timeout.timeout(Capybara.default_max_wait_time) do
      loop until finished_all_ajax_requests?(page)
      sleep(0.5)
    end
  end

  #
  # Check if there are any AJAX active
  #
  # == Parameters:
  # page::
  #   The current page (capybara session)
  #
  # == Returns:
  # True if all AJAX request finished
  #
  def finished_all_ajax_requests?(page)
    res = page.evaluate_script('jQuery.active')
    res.nil? ? true : res.zero?
  end

  #
  # Check status codes
  #
  %w(200 301 400 401 404 422 500).each do |code|
    define_method("code_#{code}?") do
      assert(page.status_code == code.to_i,
             "Response code should be #{code} : got #{page.status_code}")
    end
  end

  #
  # Click a hidden button. This method show the button and click it.
  #
  # == Parameters:
  # path:
  #   String with CSS path of the element
  #
  def click_hidden_button(path)
    page.execute_script("$('#{path}').removeClass('hidden');"\
                        "$('#{path}').show()")
    page.find(path).click
  end

  # Check errors in the new user form
  #
  # == Parameters:
  # model::
  #   Strign with model to test
  # form::
  #   String with CSS ID of the form
  # create_button::
  #   String with text of create button
  # field::
  #   String with field with error
  # yield::
  #   Block to fill the form
  #
  # == Returns:
  # Perform some asserts
  #
  def form_with_error(model, form, create_button, field)
    # Create the user without name
    within(form) do
      yield
    end
    # Create
    click_button create_button
    wait_for_ajax(page)
    # Check the error
    el = page.find("#{form} .#{model.downcase}_#{field}")
    assert(el[:class].include?('error'),
           "We must add an error on #{field} field")
  end

  #
  # Check general info table in tops view. This method perfrom multiple asserts
  #
  def check_general_info_tops(page)
    # Check the box
    page.has_css?('.tab-pane.active .table-info table.wolo')
    page.has_css?('.tab-pane.active .table-info td.total')
    page.has_css?('.tab-pane.active .table-info td.total-selected',
                  visible: false)
    page.has_css?('.tab-pane.active .table-info td.rows')
    page.has_css?('.tab-pane.active .table-info td.percentage')
    # Count rows
    rows = page.find('.tab-pane.active .table-info td.rows')
           .text.try(:to_i, 0)
    page.has_css?('.tab-pane.active table.table tbody tr.selectable-serie',
                  count: rows)
  end

  #
  # Check general info table in RAW view.
  #
  def check_general_info_raw(page)
    page.has_css?('.table-info table')
    page.has_css?('.table-info td.total')
    page.has_css?('.table-info td.rows')
    page.has_css?('.table-info td.percentage')
    # Count rows
    rows = page.find('.table-info td.rows').text.try(:to_i, 0)
    page.has_css?('table.content tbody tr.raw-row', count: rows)
  end
end
